/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SNAPLIB_H
#define SNAPLIB_H

#include <complex.h>
#include <gmp.h>
#include <mpfr.h>
#include <mpc.h>
typedef double complex dcomplex;

#define SNAPLIB_VERSION 0.02

/* if your clapack_zgeqrf returns the array TAU correctly */
#define SNAPLIB_USE_PATCHED_CLAPACK
// #undef SNAPLIB_USE_PATCHED_CLAPACK

/* rounding mode for the entire routines in libsnap. */
#define SNAPLIB_MPFR_RND MPFR_RNDN
#define SNAPLIB_MPC_RND MPC_RNDNN

/* snap_misc_test.c */
double snap_getrusageinsec();

/* snap basic types */

/*!
 * @struct supz_t
 * @brief The type for univariate polynomial over integers.
 * @warning The coefficients are in ascending order.
 */
typedef struct {
  /*! the coefficients (ascending order) */
  mpz_t *coefficients;
  /*! the size of the array of mpz_t */
  unsigned int size;
  /*! the degree of the polynomial */
  unsigned int degree;
} supz_struct;
typedef supz_struct supz_t[1];

/*!
 * @struct supd_t
 * @brief The type for univariate polynomial over reals.
 * @warning The coefficients are in ascending order.
 */
typedef struct {
  /*! the coefficients (ascending order) */
  double *coefficients;
  /*! the size of the array of double */
  unsigned int size;
  /*! the degree of the polynomial */
  unsigned int degree;
} supd_struct;
typedef supd_struct supd_t[1];

/*!
 * @struct supfr_t
 * @brief The type for univariate polynomial over reals with multi-precision float.
 * @warning The coefficients are in ascending order.
 */
typedef struct {
  /*! the coefficients (ascending order) */
  mpfr_t *coefficients;
  /*! the size of the array of mpfr_t */
  unsigned int size;
  /*! the degree of the polynomial */
  unsigned int degree;
} supfr_struct;
typedef supfr_struct supfr_t[1];

/*!
 * @struct supdc_t
 * @brief The type for univariate polynomial over complexs.
 * @warning The coefficients are in ascending order.
 */
typedef struct {
  /*! the coefficients (ascending order) */
  dcomplex *coefficients;
  /*! the size of the array of double complex */
  unsigned int size;
  /*! the degree of the polynomial */
  unsigned int degree;
} supdc_struct;
typedef supdc_struct supdc_t[1];

/*!
 * @struct supfc_t
 * @brief The type for univariate polynomial over complexes with multi-precision float.
 * @warning The coefficients are in ascending order.
 */
typedef struct {
  /*! the coefficients (ascending order) */
  mpc_t *coefficients;
  /*! the size of the array of mpc_t */
  unsigned int size;
  /*! the degree of the polynomial */
  unsigned int degree;
} supfc_struct;
typedef supfc_struct supfc_t[1];

/*!
 * @struct sdmz_t
 * @brief The type for integer matrix.
 * @warning This matrix is ROW-MAJOR.
 */
typedef struct {
  /*! the elements (row-major) */
  mpz_t *elements;
  /*! the number of rows used */
  unsigned int rows;
  /*! the number of rows allocated */
  unsigned int rsize;
  /*! the number of cols used */
  unsigned int cols;
  /*! the number of cols allocated */
  unsigned int csize;
} sdmz_struct;
typedef sdmz_struct sdmz_t[1];

/*!
 * @struct sdmd_t
 * @brief The type for real matrix with double.
 * @warning This matrix is ROW-MAJOR.
 */
typedef struct {
  /*! the elements (row-major) */
  double *elements;
  /*! the number of rows used */
  unsigned int rows;
  /*! the number of rows allocated */
  unsigned int rsize;
  /*! the number of cols used */
  unsigned int cols;
  /*! the number of cols allocated */
  unsigned int csize;
} sdmd_struct;
typedef sdmd_struct sdmd_t[1];

/*!
 * @struct sdmdc_t
 * @brief The type for complex matrix with double.
 * @warning This matrix is ROW-MAJOR.
 */
typedef struct {
  /*! the elements (row-major) */
  dcomplex *elements;
  /*! the number of rows used */
  unsigned int rows;
  /*! the number of rows allocated */
  unsigned int rsize;
  /*! the number of cols used */
  unsigned int cols;
  /*! the number of cols allocated */
  unsigned int csize;
} sdmdc_struct;
typedef sdmdc_struct sdmdc_t[1];

/*!
 * @struct sdmfr_t
 * @brief The type for real matrix with multi-precision float.
 * @warning This matrix is ROW-MAJOR.
 */
typedef struct {
  /*! the elements (row-major) */
  mpfr_t *elements;
  /*! the number of rows used */
  unsigned int rows;
  /*! the number of rows allocated */
  unsigned int rsize;
  /*! the number of cols used */
  unsigned int cols;
  /*! the number of cols allocated */
  unsigned int csize;
} sdmfr_struct;
typedef sdmfr_struct sdmfr_t[1];

/*!
 * @struct sdmfc_t
 * @brief The type for complex matrix with multi-precision float.
 * @warning This matrix is ROW-MAJOR.
 */
typedef struct {
  /*! the elements (row-major) */
  mpc_t *elements;
  /*! the number of rows used */
  unsigned int rows;
  /*! the number of rows allocated */
  unsigned int rsize;
  /*! the number of cols used */
  unsigned int cols;
  /*! the number of cols allocated */
  unsigned int csize;
} sdmfc_struct;
typedef sdmfc_struct sdmfc_t[1];

/* snap.c */

void snap_copyright();

/* snap_parser.c */
// IO functions.
typedef enum {SNAP_SP_ID, SNAP_SP_INT, SNAP_SP_REAL, SNAP_SP_ADD,
              SNAP_SP_SUB, SNAP_SP_MUL, SNAP_SP_DIV, SNAP_SP_POW, SNAP_SP_LPAR,
              SNAP_SP_RPAR, SNAP_SP_EOF, SNAP_SP_J
             } sp_lexical_token_type;
typedef struct sp_lexical_token {
  sp_lexical_token_type type;
  char *token;
  struct sp_lexical_token *next;
} sp_lexical_token_struct;
typedef sp_lexical_token_struct sp_lexical_token_t[1];
int sp_lexical_analysis_str(sp_lexical_token_t, char*);
void sp_lexical_token_free(sp_lexical_token_t);

/* snap_type_mpc.c */

void mpc_init_set(mpc_t, mpc_t, mpc_rnd_t);
void mpc_init_set_ui(mpc_t, unsigned int, mpc_rnd_t);
void mpc_init_set_si(mpc_t, signed int, mpc_rnd_t);
void mpc_sub_d(mpc_t, mpc_t, double, mpc_rnd_t);
void mpc_sub_dc(mpc_t, mpc_t, dcomplex, mpc_rnd_t);
void mpc_mul_d(mpc_t, mpc_t, double, mpc_rnd_t);
void mpc_mul_dc(mpc_t, mpc_t, dcomplex, mpc_rnd_t);
void mpc_div_d(mpc_t, mpc_t, double, mpc_rnd_t);
void mpc_div_dc(mpc_t, mpc_t, dcomplex, mpc_rnd_t);

/* snap_type_sup_base.c */

void supz_init(supz_t);
void supd_init(supd_t);
void supdc_init(supdc_t);
void supfr_init(supfr_t);
void supfc_init(supfc_t);

void supz_init2(supz_t, unsigned int);
void supd_init2(supd_t, unsigned int);
void supdc_init2(supdc_t, unsigned int);
void supfr_init2(supfr_t, unsigned int);
void supfc_init2(supfc_t, unsigned int);

void supz_init_set(supz_t, supz_t);
void supd_init_set(supd_t, supd_t);
void supdc_init_set(supdc_t, supdc_t);
void supfr_init_set(supfr_t, supfr_t);
void supfc_init_set(supfc_t, supfc_t);

void supz_clear(supz_t);
#define supd_clear(upoly) free(supd_coefs(upoly))
#define supdc_clear(upoly) free(supdc_coefs(upoly))
void supfr_clear(supfr_t);
void supfc_clear(supfc_t);

void supz_set(supz_t, supz_t);
void supd_set(supd_t, supd_t);
void supdc_set(supdc_t, supdc_t);
void supfr_set(supfr_t, supfr_t);
void supfc_set(supfc_t, supfc_t);

void supz_set_z(supz_t, mpz_t*, unsigned int);
void supd_set_z(supd_t, mpz_t*, unsigned int);
void supdc_set_z(supdc_t, mpz_t*, unsigned int);
void supfr_set_z(supfr_t, mpz_t*, unsigned int);
void supfc_set_z(supfc_t, mpz_t*, unsigned int);

void supz_set_d(supz_t, double*, unsigned int);
void supd_set_d(supd_t, double*, unsigned int);
void supdc_set_d(supdc_t, double*, unsigned int);
void supfr_set_d(supfr_t, double*, unsigned int);
void supfc_set_d(supfc_t, double*, unsigned int);

void supz_set_dc(supz_t, dcomplex*, unsigned int);
void supd_set_dc(supd_t, dcomplex*, unsigned int);
void supdc_set_dc(supdc_t, dcomplex*, unsigned int);
void supfr_set_dc(supfr_t, dcomplex*, unsigned int);
void supfc_set_dc(supfc_t, dcomplex*, unsigned int);

void supz_set_fr(supz_t, mpfr_t*, unsigned int);
void supd_set_fr(supd_t, mpfr_t*, unsigned int);
void supdc_set_fr(supdc_t, mpfr_t*, unsigned int);
void supfr_set_fr(supfr_t, mpfr_t*, unsigned int);
void supfc_set_fr(supfc_t, mpfr_t*, unsigned int);

void supz_set_fc(supz_t, mpc_t*, unsigned int);
void supd_set_fc(supd_t, mpc_t*, unsigned int);
void supdc_set_fc(supdc_t, mpc_t*, unsigned int);
void supfr_set_fc(supfr_t, mpc_t*, unsigned int);
void supfc_set_fc(supfc_t, mpc_t*, unsigned int);

void supz_set_rd(supz_t, double*, unsigned int);
void supd_set_rd(supd_t, double*, unsigned int);
void supdc_set_rd(supdc_t, double*, unsigned int);
void supfr_set_rd(supfr_t, double*, unsigned int);
void supfc_set_rd(supfc_t, double*, unsigned int);

void supz_set_rdc(supz_t, dcomplex*, unsigned int);
void supd_set_rdc(supd_t, dcomplex*, unsigned int);
void supdc_set_rdc(supdc_t, dcomplex*, unsigned int);
void supfr_set_rdc(supfr_t, dcomplex*, unsigned int);
void supfc_set_rdc(supfc_t, dcomplex*, unsigned int);

void supz_set_rfr(supz_t, mpfr_t*, unsigned int);
void supd_set_rfr(supd_t, mpfr_t*, unsigned int);
void supdc_set_rfr(supdc_t, mpfr_t*, unsigned int);
void supfr_set_rfr(supfr_t, mpfr_t*, unsigned int);
void supfc_set_rfr(supfc_t, mpfr_t*, unsigned int);

void supz_set_rfc(supz_t, mpc_t*, unsigned int);
void supd_set_rfc(supd_t, mpc_t*, unsigned int);
void supdc_set_rfc(supdc_t, mpc_t*, unsigned int);
void supfr_set_rfc(supfr_t, mpc_t*, unsigned int);
void supfc_set_rfc(supfc_t, mpc_t*, unsigned int);

void supz_set_coef(supz_t, unsigned int, mpz_t);
void supd_set_coef(supd_t, unsigned int, double);
void supdc_set_coef(supdc_t, unsigned int, dcomplex);
void supfr_set_coef(supfr_t, unsigned int, mpfr_t);
void supfc_set_coef(supfc_t, unsigned int, mpc_t);

void supz_set_coef_d(supz_t, unsigned int, double);
void supd_set_coef_d(supd_t, unsigned int, double);
void supdc_set_coef_d(supdc_t, unsigned int, double);
void supfr_set_coef_d(supfr_t, unsigned int, double);
void supfc_set_coef_d(supfc_t, unsigned int, double);

unsigned int supz_set_str(supz_t, char*);
unsigned int supd_set_str(supd_t, char*);
unsigned int supdc_set_str(supdc_t, char*);
unsigned int supfr_set_str(supfr_t, char*);
unsigned int supfc_set_str(supfc_t, char*);

#define supz_size(upoly) ((upoly)->size)
#define supd_size(upoly) ((upoly)->size)
#define supdc_size(upoly) ((upoly)->size)
#define supfr_size(upoly) ((upoly)->size)
#define supfc_size(upoly) ((upoly)->size)

#define supz_degree(upoly) ((upoly)->degree)
#define supd_degree(upoly) ((upoly)->degree)
#define supdc_degree(upoly) ((upoly)->degree)
#define supfr_degree(upoly) ((upoly)->degree)
#define supfc_degree(upoly) ((upoly)->degree)

#define supz_coefs(upoly) ((upoly)->coefficients)
#define supd_coefs(upoly) ((upoly)->coefficients)
#define supdc_coefs(upoly) ((upoly)->coefficients)
#define supfr_coefs(upoly) ((upoly)->coefficients)
#define supfc_coefs(upoly) ((upoly)->coefficients)

#define supz_coef(upoly,n) (*((upoly)->coefficients+(n)))
#define supd_coef(upoly,n) (*((upoly)->coefficients+(n)))
#define supdc_coef(upoly,n) (*((upoly)->coefficients+(n)))
#define supfr_coef(upoly,n) (*((upoly)->coefficients+(n)))
#define supfc_coef(upoly,n) (*((upoly)->coefficients+(n)))

#define supdc_coef_real(upoly,n) (creal(*((upoly)->coefficients+(n))))
#define supdc_coef_imag(upoly,n) (cimag(*((upoly)->coefficients+(n))))
#define supfc_coef_realref(upoly,n) (mpc_realref(*((upoly)->coefficients+(n))))
#define supfc_coef_imagref(upoly,n) (mpc_imagref(*((upoly)->coefficients+(n))))

#define supz_coefp(upoly,n) (((upoly)->coefficients+(n)))
#define supd_coefp(upoly,n) (((upoly)->coefficients+(n)))
#define supdc_coefp(upoly,n) (((upoly)->coefficients+(n)))
#define supfr_coefp(upoly,n) (((upoly)->coefficients+(n)))
#define supfc_coefp(upoly,n) (((upoly)->coefficients+(n)))

#define supz_lcoef(upoly) (*((upoly)->coefficients+(upoly)->degree))
#define supd_lcoef(upoly) (*((upoly)->coefficients+(upoly)->degree))
#define supdc_lcoef(upoly) (*((upoly)->coefficients+(upoly)->degree))
#define supfr_lcoef(upoly) (*((upoly)->coefficients+(upoly)->degree))
#define supfc_lcoef(upoly) (*((upoly)->coefficients+(upoly)->degree))

#define supdc_lcoef_real(upoly) (creal(*((upoly)->coefficients+(upoly)->degree)))
#define supdc_lcoef_imag(upoly) (cimag(*((upoly)->coefficients+(upoly)->degree)))
#define supfc_lcoef_realref(upoly) (mpc_realref(*((upoly)->coefficients+(upoly)->degree)))
#define supfc_lcoef_imagref(upoly) (mpc_imagref(*((upoly)->coefficients+(upoly)->degree)))

#define supz_lcoefp(upoly) (((upoly)->coefficients+(upoly)->degree))
#define supd_lcoefp(upoly) (((upoly)->coefficients+(upoly)->degree))
#define supdc_lcoefp(upoly) (((upoly)->coefficients+(upoly)->degree))
#define supfr_lcoefp(upoly) (((upoly)->coefficients+(upoly)->degree))
#define supfc_lcoefp(upoly) (((upoly)->coefficients+(upoly)->degree))

void supz_swap(supz_t, supz_t);
void supd_swap(supd_t, supd_t);
void supdc_swap(supdc_t, supdc_t);
void supfr_swap(supfr_t, supfr_t);
void supfc_swap(supfc_t, supfc_t);

void supz_c2norm(mpz_t, supz_t);
void supz_f2norm(mpz_t, supz_t);
double supd_2norm(supd_t);
double supd_inorm(supd_t);
double supdc_2norm(supdc_t);
double supdc_inorm(supdc_t);
void supfr_2norm(mpfr_t, supfr_t);
void supfc_2norm(mpfr_t, supfc_t);

void supz_self_neg(supz_t);
void supd_self_neg(supd_t);
void supdc_self_neg(supdc_t);
void supfr_self_neg(supfr_t);
void supfc_self_neg(supfc_t);

void supz_self_reversal(supz_t);
void supd_self_reversal(supd_t);
void supdc_self_reversal(supdc_t);
void supfr_self_reversal(supfr_t);
void supfc_self_reversal(supfc_t);

#define supz_mul supz_mul_karatsuba_block
#define supd_mul supd_mul_karatsuba_block
#define supdc_mul supdc_mul_karatsuba_block
#define supfr_mul supfr_mul_karatsuba_block
#define supfc_mul supfc_mul_karatsuba_block

void supz_mul_traditional(supz_t, supz_t, supz_t);
void supd_mul_traditional(supd_t, supd_t, supd_t);
void supdc_mul_traditional(supdc_t, supdc_t, supdc_t);
void supfr_mul_traditional(supfr_t, supfr_t, supfr_t);
void supfc_mul_traditional(supfc_t, supfc_t, supfc_t);

void supz_mul_karatsuba(supz_t, supz_t, supz_t);
void supd_mul_karatsuba(supd_t, supd_t, supd_t);
void supdc_mul_karatsuba(supdc_t, supdc_t, supdc_t);
void supfr_mul_karatsuba(supfr_t, supfr_t, supfr_t);
void supfc_mul_karatsuba(supfc_t, supfc_t, supfc_t);

void supz_mul_karatsuba_block(supz_t, supz_t, supz_t);
void supd_mul_karatsuba_block(supd_t, supd_t, supd_t);
void supdc_mul_karatsuba_block(supdc_t, supdc_t, supdc_t);
void supfr_mul_karatsuba_block(supfr_t, supfr_t, supfr_t);
void supfc_mul_karatsuba_block(supfc_t, supfc_t, supfc_t);

void supz_mul_si(supz_t, supz_t, signed int);
void supd_mul_si(supd_t, supd_t, signed int);
void supdc_mul_si(supdc_t, supdc_t, signed int);
void supfr_mul_si(supfr_t, supfr_t, signed int);
void supfc_mul_si(supfc_t, supfc_t, signed int);

void supz_mul_d(supz_t, supz_t, double);
void supd_mul_d(supd_t, supd_t, double);
void supdc_mul_d(supdc_t, supdc_t, double);
void supfr_mul_d(supfr_t, supfr_t, double);
void supfc_mul_d(supfc_t, supfc_t, double);

void supz_mul_dc(supz_t, supz_t, dcomplex);
void supd_mul_dc(supd_t, supd_t, dcomplex);
void supdc_mul_dc(supdc_t, supdc_t, dcomplex);
void supfr_mul_dc(supfr_t, supfr_t, dcomplex);
void supfc_mul_dc(supfc_t, supfc_t, dcomplex);

void supz_mul_fr(supz_t, supz_t, mpfr_t);
void supd_mul_fr(supd_t, supd_t, mpfr_t);
void supdc_mul_fr(supdc_t, supdc_t, mpfr_t);
void supfr_mul_fr(supfr_t, supfr_t, mpfr_t);
void supfc_mul_fr(supfc_t, supfc_t, mpfr_t);

void supz_add(supz_t, supz_t, supz_t);
void supd_add(supd_t, supd_t, supd_t);
void supdc_add(supdc_t, supdc_t, supdc_t);
void supfr_add(supfr_t, supfr_t, supfr_t);
void supfc_add(supfc_t, supfc_t, supfc_t);

void supz_sub(supz_t, supz_t, supz_t);
void supd_sub(supd_t, supd_t, supd_t);
void supdc_sub(supdc_t, supdc_t, supdc_t);
void supfr_sub(supfr_t, supfr_t, supfr_t);
void supfc_sub(supfc_t, supfc_t, supfc_t);

void supz_randpoly(supz_t, unsigned int, mpz_t, double, gmp_randstate_t);
void supd_randpoly(supd_t, unsigned int, double, double);
void supdc_randpoly(supdc_t, unsigned int, dcomplex, double);
void supfr_randpoly(supfr_t, unsigned int, mpfr_t, double, gmp_randstate_t);
void supfc_randpoly(supfc_t, unsigned int, mpc_t, double, gmp_randstate_t);

void supd_div_qr(supd_t, supd_t, supd_t, supd_t);
void supdc_div_qr(supdc_t, supdc_t, supdc_t, supdc_t);
void supfr_div_qr(supfr_t, supfr_t, supfr_t, supfr_t);
void supfc_div_qr(supfc_t, supfc_t, supfc_t, supfc_t);

void supd_lsdiv_q(supd_t, supd_t, supd_t);
void supdc_lsdiv_q(supdc_t, supdc_t, supdc_t);
void supfr_lsdiv_q(supfr_t, supfr_t, supfr_t);
void supfc_lsdiv_q(supfc_t, supfc_t, supfc_t);

void supd_fftdiv_q(supd_t, supd_t, supd_t);
void supdc_fftdiv_q(supdc_t, supdc_t, supdc_t);
void supfr_fftdiv_q(supfr_t, supfr_t, supfr_t);
void supfc_fftdiv_q(supfc_t, supfc_t, supfc_t);

void supz_div_d(supz_t, supz_t, double);
void supd_div_d(supd_t, supd_t, double);
void supdc_div_d(supdc_t, supdc_t, double);
void supfr_div_d(supfr_t rpoly, supfr_t poly, double divisor);
void supfc_div_d(supfc_t rpoly, supfc_t poly, double divisor);

void supz_div_dc(supz_t, supz_t, dcomplex);
void supd_div_dc(supd_t, supd_t, dcomplex);
void supdc_div_dc(supdc_t, supdc_t, dcomplex);
void supfr_div_dc(supfr_t rpoly, supfr_t poly, dcomplex divisor);
void supfc_div_dc(supfc_t rpoly, supfc_t poly, dcomplex divisor);

void supz_div_fr(supz_t, supz_t, mpfr_t);
void supd_div_fr(supd_t, supd_t, mpfr_t);
void supdc_div_fr(supdc_t, supdc_t, mpfr_t);
void supfr_div_fr(supfr_t rpoly, supfr_t poly, mpfr_t divisor);
void supfc_div_fr(supfc_t rpoly, supfc_t poly, mpfr_t divisor);

void supz_div_fc(supz_t, supz_t, mpc_t);
void supd_div_fc(supd_t, supd_t, mpc_t);
void supdc_div_fc(supdc_t, supdc_t, mpc_t);
void supfr_div_fc(supfr_t rpoly, supfr_t poly, mpc_t divisor);
void supfc_div_fc(supfc_t rpoly, supfc_t poly, mpc_t divisor);

/* snap_type_sdm_base.c */

#define sdmz_rows(matrix) ((matrix)->rows)
#define sdmd_rows(matrix) ((matrix)->rows)
#define sdmdc_rows(matrix) ((matrix)->rows)
#define sdmfr_rows(matrix) ((matrix)->rows)
#define sdmfc_rows(matrix) ((matrix)->rows)

#define sdmz_rsize(matrix) ((matrix)->rsize)
#define sdmd_rsize(matrix) ((matrix)->rsize)
#define sdmdc_rsize(matrix) ((matrix)->rsize)
#define sdmfr_rsize(matrix) ((matrix)->rsize)
#define sdmfc_rsize(matrix) ((matrix)->rsize)

#define sdmz_cols(matrix) ((matrix)->cols)
#define sdmd_cols(matrix) ((matrix)->cols)
#define sdmdc_cols(matrix) ((matrix)->cols)
#define sdmfr_cols(matrix) ((matrix)->cols)
#define sdmfc_cols(matrix) ((matrix)->cols)

#define sdmz_csize(matrix) ((matrix)->csize)
#define sdmd_csize(matrix) ((matrix)->csize)
#define sdmdc_csize(matrix) ((matrix)->csize)
#define sdmfr_csize(matrix) ((matrix)->csize)
#define sdmfc_csize(matrix) ((matrix)->csize)

#define sdmz_elems(matrix) ((matrix)->elements)
#define sdmd_elems(matrix) ((matrix)->elements)
#define sdmdc_elems(matrix) ((matrix)->elements)
#define sdmfr_elems(matrix) ((matrix)->elements)
#define sdmfc_elems(matrix) ((matrix)->elements)

#define sdmz_elem(matrix,i,j) (*((matrix)->elements+(i)*(sdmz_csize(matrix))+(j)))
#define sdmd_elem(matrix,i,j) (*((matrix)->elements+(i)*(sdmd_csize(matrix))+(j)))
#define sdmdc_elem(matrix,i,j) (*((matrix)->elements+(i)*(sdmdc_csize(matrix))+(j)))
#define sdmfr_elem(matrix,i,j) (*((matrix)->elements+(i)*(sdmfr_csize(matrix))+(j)))
#define sdmfc_elem(matrix,i,j) (*((matrix)->elements+(i)*(sdmfc_csize(matrix))+(j)))

#define sdmz_elemp(matrix,i,j) (((matrix)->elements+(i)*(sdmz_csize(matrix))+(j)))
#define sdmd_elemp(matrix,i,j) (((matrix)->elements+(i)*(sdmd_csize(matrix))+(j)))
#define sdmdc_elemp(matrix,i,j) (((matrix)->elements+(i)*(sdmdc_csize(matrix))+(j)))
#define sdmfr_elemp(matrix,i,j) (((matrix)->elements+(i)*(sdmfr_csize(matrix))+(j)))
#define sdmfc_elemp(matrix,i,j) (((matrix)->elements+(i)*(sdmfc_csize(matrix))+(j)))

#define sdmz_elem_column_major(matrix,i,j) (*((matrix)->elements+(j)*(sdmz_rsize(matrix))+(i)))
#define sdmd_elem_column_major(matrix,i,j) (*((matrix)->elements+(j)*(sdmd_rsize(matrix))+(i)))
#define sdmdc_elem_column_major(matrix,i,j) (*((matrix)->elements+(j)*(sdmdc_rsize(matrix))+(i)))
#define sdmfr_elem_column_major(matrix,i,j) (*((matrix)->elements+(j)*(sdmfr_rsize(matrix))+(i)))
#define sdmfc_elem_column_major(matrix,i,j) (*((matrix)->elements+(j)*(sdmfc_rsize(matrix))+(i)))

#define sdmz_elemp_column_major(matrix,i,j) (((matrix)->elements+(j)*(sdmz_rsize(matrix))+(i)))
#define sdmd_elemp_column_major(matrix,i,j) (((matrix)->elements+(j)*(sdmd_rsize(matrix))+(i)))
#define sdmdc_elemp_column_major(matrix,i,j) (((matrix)->elements+(j)*(sdmdc_rsize(matrix))+(i)))
#define sdmfr_elemp_column_major(matrix,i,j) (((matrix)->elements+(j)*(sdmfr_rsize(matrix))+(i)))
#define sdmfc_elemp_column_major(matrix,i,j) (((matrix)->elements+(j)*(sdmfc_rsize(matrix))+(i)))

void sdmz_init(sdmz_t);
void sdmd_init(sdmd_t);
void sdmdc_init(sdmdc_t);
void sdmfr_init(sdmfr_t);
void sdmfc_init(sdmfc_t);

void sdmz_init2(sdmz_t, unsigned int, unsigned int);
void sdmd_init2(sdmd_t, unsigned int, unsigned int);
void sdmdc_init2(sdmdc_t, unsigned int, unsigned int);
void sdmfr_init2(sdmfr_t, unsigned int, unsigned int);
void sdmfc_init2(sdmfc_t, unsigned int, unsigned int);

void sdmz_reinit2(sdmz_t, unsigned int, unsigned int);
void sdmd_reinit2(sdmd_t, unsigned int, unsigned int);
void sdmdc_reinit2(sdmdc_t, unsigned int, unsigned int);
void sdmfr_reinit2(sdmfr_t, unsigned int, unsigned int);
void sdmfc_reinit2(sdmfc_t, unsigned int, unsigned int);

void sdmz_reinit3(sdmz_t, unsigned int, unsigned int);
void sdmd_reinit3(sdmd_t, unsigned int, unsigned int);
void sdmdc_reinit3(sdmdc_t, unsigned int, unsigned int);
void sdmfr_reinit3(sdmfr_t, unsigned int, unsigned int);
void sdmfc_reinit3(sdmfc_t, unsigned int, unsigned int);

void sdmz_realloc(sdmz_t, unsigned int, unsigned int);
void sdmd_realloc(sdmd_t, unsigned int, unsigned int);
void sdmdc_realloc(sdmdc_t, unsigned int, unsigned int);
void sdmfr_realloc(sdmfr_t, unsigned int, unsigned int);
void sdmfc_realloc(sdmfc_t, unsigned int, unsigned int);

void sdmz_clear(sdmz_t);
void sdmd_clear(sdmd_t);
void sdmdc_clear(sdmdc_t);
void sdmfr_clear(sdmfr_t);
void sdmfc_clear(sdmfc_t);

void sdmz_swap(sdmz_t, sdmz_t);
void sdmd_swap(sdmd_t, sdmd_t);
void sdmdc_swap(sdmdc_t, sdmdc_t);
void sdmfr_swap(sdmfr_t, sdmfr_t);
void sdmfc_swap(sdmfc_t, sdmfc_t);

#define sdmz_sylvester_matrix_transpose(matrix, polyf, polyg) (sdmz_subresultant_mapping_matrix_transpose(matrix,polyf,polyg,0))
#define sdmd_sylvester_matrix_transpose(matrix, polyf, polyg) (sdmd_subresultant_mapping_matrix_transpose(matrix,polyf,polyg,0))
#define sdmdc_sylvester_matrix_transpose(matrix, polyf, polyg) (sdmdc_subresultant_mapping_matrix_transpose(matrix,polyf,polyg,0))
#define sdmfr_sylvester_matrix_transpose(matrix, polyf, polyg) (sdmfr_subresultant_mapping_matrix_transpose(matrix,polyf,polyg,0))
#define sdmfc_sylvester_matrix_transpose(matrix, polyf, polyg) (sdmfc_subresultant_mapping_matrix_transpose(matrix,polyf,polyg,0))

#define sdmz_sylvester_matrix_transpose_column_major(matrix, polyf, polyg) (sdmz_subresultant_mapping_matrix_transpose_column_major(matrix,polyf,polyg,0))
#define sdmd_sylvester_matrix_transpose_column_major(matrix, polyf, polyg) (sdmd_subresultant_mapping_matrix_transpose_column_major(matrix,polyf,polyg,0))
#define sdmdc_sylvester_matrix_transpose_column_major(matrix, polyf, polyg) (sdmdc_subresultant_mapping_matrix_transpose_column_major(matrix,polyf,polyg,0))
#define sdmfr_sylvester_matrix_transpose_column_major(matrix, polyf, polyg) (sdmfr_subresultant_mapping_matrix_transpose_column_major(matrix,polyf,polyg,0))
#define sdmfc_sylvester_matrix_transpose_column_major(matrix, polyf, polyg) (sdmfc_subresultant_mapping_matrix_transpose_column_major(matrix,polyf,polyg,0))

void sdmz_subresultant_mapping_matrix_asc(sdmz_t, supz_t, supz_t, unsigned int);
void sdmd_subresultant_mapping_matrix_asc(sdmd_t, supd_t, supd_t, unsigned int);
void sdmdc_subresultant_mapping_matrix_asc(sdmdc_t, supdc_t, supdc_t, unsigned int);
void sdmfr_subresultant_mapping_matrix_asc(sdmfr_t, supfr_t, supfr_t, unsigned int);
void sdmfc_subresultant_mapping_matrix_asc(sdmfc_t, supfc_t, supfc_t, unsigned int);

void sdmz_subresultant_mapping_matrix_dsc(sdmz_t, supz_t, supz_t, unsigned int);
void sdmd_subresultant_mapping_matrix_dsc(sdmd_t, supd_t, supd_t, unsigned int);
void sdmdc_subresultant_mapping_matrix_dsc(sdmdc_t, supdc_t, supdc_t, unsigned int);
void sdmfr_subresultant_mapping_matrix_dsc(sdmfr_t, supfr_t, supfr_t, unsigned int);
void sdmfc_subresultant_mapping_matrix_dsc(sdmfc_t, supfc_t, supfc_t, unsigned int);

void sdmz_subresultant_mapping_matrix_transpose(sdmz_t, supz_t, supz_t, unsigned int);
void sdmd_subresultant_mapping_matrix_transpose(sdmd_t, supd_t, supd_t, unsigned int);
void sdmdc_subresultant_mapping_matrix_transpose(sdmdc_t, supdc_t, supdc_t, unsigned int);
void sdmfr_subresultant_mapping_matrix_transpose(sdmfr_t, supfr_t, supfr_t, unsigned int);
void sdmfc_subresultant_mapping_matrix_transpose(sdmfc_t, supfc_t, supfc_t, unsigned int);

void sdmz_subresultant_mapping_matrix_transpose_column_major(sdmz_t, supz_t, supz_t, unsigned int);
void sdmd_subresultant_mapping_matrix_transpose_column_major(sdmd_t, supd_t, supd_t, unsigned int);
void sdmdc_subresultant_mapping_matrix_transpose_column_major(sdmdc_t, supdc_t, supdc_t, unsigned int);
void sdmfr_subresultant_mapping_matrix_transpose_column_major(sdmfr_t, supfr_t, supfr_t, unsigned int);
void sdmfc_subresultant_mapping_matrix_transpose_column_major(sdmfc_t, supfc_t, supfc_t, unsigned int);

void sdmz_convolution_matrix(sdmz_t, supz_t, unsigned int);
void sdmd_convolution_matrix(sdmd_t, supd_t, unsigned int);
void sdmdc_convolution_matrix(sdmdc_t, supdc_t, unsigned int);
void sdmdc_convolution_matrix_conj(sdmdc_t, supdc_t, unsigned int);
void sdmfr_convolution_matrix(sdmfr_t, supfr_t, unsigned int);
void sdmfc_convolution_matrix(sdmfc_t, supfc_t, unsigned int);
void sdmfc_convolution_matrix_conj(sdmfc_t, supfc_t, unsigned int);

void sdmz_convolution_matrix_column_major(sdmz_t, supz_t, unsigned int);
void sdmd_convolution_matrix_column_major(sdmd_t, supd_t, unsigned int);
void sdmdc_convolution_matrix_column_major(sdmdc_t, supdc_t, unsigned int);
void sdmfr_convolution_matrix_column_major(sdmfr_t, supfr_t, unsigned int);
void sdmfc_convolution_matrix_column_major(sdmfc_t, supfc_t, unsigned int);

void sdmz_convolution_matrix_zerofill(sdmz_t, supz_t, unsigned int);
void sdmd_convolution_matrix_zerofill(sdmd_t, supd_t, unsigned int);
void sdmdc_convolution_matrix_zerofill(sdmdc_t, supdc_t, unsigned int);
void sdmdc_convolution_matrix_zerofill_conj(sdmdc_t, supdc_t, unsigned int);
void sdmfr_convolution_matrix_zerofill(sdmfr_t, supfr_t, unsigned int);
void sdmfc_convolution_matrix_zerofill(sdmfc_t, supfc_t, unsigned int);
void sdmfc_convolution_matrix_zerofill_conj(sdmfc_t, supfc_t, unsigned int);

/* snap_type_mpz.c */
void mpz_inner(mpz_t, mpz_t*, mpz_t*, unsigned int);
void mpz_inner_matrix_transpose(mpz_t*, mpz_t*, mpz_t*, unsigned int, unsigned int, unsigned int);
void mpz_rdiv_q(mpz_t, mpz_t, mpz_t);
void mpz_mul_fr(mpz_t rop, mpz_t op1, mpfr_t op2, mpfr_rnd_t mode);
void mpz_mul_d(mpz_t rop, mpz_t op1, double op2);
void mpz_div_fr(mpz_t rop, mpz_t op1, mpfr_t op2, mpfr_rnd_t mode);
void mpz_div_d(mpz_t rop, mpz_t op1, double op2);

/* snap_type_mpq.c */

void mpq_sub_z(mpq_t, mpq_t, mpz_t);
void mpq_mul_z(mpq_t, mpq_t, mpz_t);
void mpq_get_z(mpz_t, mpq_t);
void mpq_inner(mpq_t, mpq_t*, mpq_t*, unsigned int);
void mpq_inner_z(mpq_t, mpq_t*, mpz_t*, unsigned int);

/* snap_misc_mblas_mpfr.c */

void snap_mblas_fcswap(const unsigned int len, mpc_t *vecX, const int incX, mpc_t *vecY, const int incY);
void snap_mblas_frdot(mpfr_t, const unsigned int, mpfr_t*, const int, mpfr_t*, const int);
void snap_mblas_fcdotu(mpc_t, const unsigned int, mpc_t*, const int, mpc_t*, const int);
void snap_mblas_fcdotc(mpc_t, const unsigned int, mpc_t*, const int, mpc_t*, const int);
void snap_mblas_fraxpy(const unsigned int, const mpfr_t, mpfr_t*, const int, mpfr_t*, const int);
void snap_mblas_fcaxpy(const unsigned int, const mpc_t, mpc_t*, const int, mpc_t*, const int);
void snap_mblas_frnrm2(mpfr_t, const unsigned int, mpfr_t*, const int);
void snap_mblas_fcnrm2(mpfr_t, const unsigned int, mpc_t*, const int);
void snap_mblas_frtrmv(const char, const char, const char, const unsigned int, mpfr_t*, const int, mpfr_t*, const int);
void snap_mblas_fctrmv(const char, const char, const char, const unsigned int, mpc_t*, const int, mpc_t*, const int);
void snap_mblas_frgemv(const char, const unsigned int, const unsigned int, mpfr_t, mpfr_t*, const int, mpfr_t*, const int, mpfr_t, mpfr_t*, const int);
void snap_mblas_fcgemv(const char, const unsigned int, const unsigned int, mpc_t, mpc_t*, const int, mpc_t*, const int, mpc_t, mpc_t*, const int);
void snap_mblas_fctrmm(const char, const char, const char, const char, const unsigned int, const unsigned int, mpc_t, mpc_t*, const int, mpc_t*, const int);

/* snap_misc_mlapack_mpfr.c */

int snap_mlapack_frgeqrf(const unsigned int, const unsigned int, mpfr_t*, const int, mpfr_t*);
int snap_mlapack_fcgeqrf(const unsigned int, const unsigned int, mpc_t*, const int, mpc_t*);
// int snap_mlapack_frgels(const unsigned int, const unsigned int, mpfr_t*, const int, mpfr_t*, const int);
int snap_mlapack_frgels(const char, const unsigned int, const unsigned int, const unsigned int, mpfr_t*, const int, mpfr_t*, const int);
int snap_mlapack_fcgels(const char, const unsigned int, const unsigned int, const unsigned int, mpc_t*, const int, mpc_t*, const int);
int snap_mlapack_frtrtrs(const char, const char, const char, const unsigned int, const unsigned int, mpfr_t*, const int, mpfr_t*, const int);
int snap_mlapack_fctrtrs(const char, const char, const char, const unsigned int, const unsigned int, mpc_t*, const int, mpc_t*, const int);
int snap_mlapack_frlarfx(const char, const unsigned int, const unsigned int, mpfr_t*, mpfr_t, mpfr_t*, const int, mpfr_t*);
int snap_mlapack_fclarfx(const char, const unsigned int, const unsigned int, mpc_t*, mpc_t, mpc_t*, const int, mpc_t*);
int snap_mlapack_frlarfg(const unsigned int, mpfr_t, mpfr_t*, const int, mpfr_t);
int snap_mlapack_fclarfg(const unsigned int, mpc_t, mpc_t*, const int, mpc_t);
int snap_mlapack_fcungqr(const unsigned int rows, const unsigned int cols, const unsigned int num, mpc_t *matA, const int ldA, mpc_t *tau);

/* snap_misc_lll.c */

#define lll_exact lll_exact_keepintegral
void lll_exact_mode(mpz_t*, unsigned int, unsigned int, unsigned int);
void lll_exact_original(mpz_t*, unsigned int, unsigned int);
void lll_exact_deepinsertion(mpz_t*, unsigned int, unsigned int);
void lll_exact_keepintegral(mpz_t*, unsigned int, unsigned int);
void lll_exact_reductionl3(mpz_t*, unsigned int, unsigned int);
void lll_exact_modreductionl3(mpz_t*, unsigned int, unsigned int);
void lll_exact_reductionl3_heuristics(mpz_t*, unsigned int, unsigned int);
void lll_exact_modreductionl3_heuristics(mpz_t*, unsigned int, unsigned int);

/* snap_misc_fft_mpc.c */
void snap_fcdft_forward(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS);
void snap_fcdft_forward2(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS);
void snap_fcdft_backward(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS);

/* snap_alg_agcd_integer.c */

#define agcd_over_integers(f,g,agcd,cf,cg,tol,eps,svt,lll_mode) (agcd_jsc2009(f,g,agcd,cf,cg,tol,1,10,0))
int agcd_jsc2009(supz_t, supz_t, supz_t, supz_t, supz_t, mpz_t, unsigned int, unsigned int, unsigned int);

/* snap_alg_agcd_exqrgcd_sup_base.c */
int supd_agcd_exqrgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol, int mode);
int supdc_agcd_exqrgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol, int mode);
int supfr_agcd_exqrgcd(supfr_t poly_f, supfr_t poly_g, supfr_t poly_d, supfr_t poly_f1, supfr_t poly_g1, const mpfr_t tol, int mode);
int supfc_agcd_exqrgcd(supfc_t poly_f, supfc_t poly_g, supfc_t poly_d, supfc_t poly_f1, supfc_t poly_g1, const mpfr_t tol, int mode);

/* snap_alg_agcd_exqrgcd2_sup_base.c */
int supd_agcd_exqrgcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const double tol, int mode, int mode_opt);
int supdc_agcd_exqrgcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const double tol, int mode, int mode_opt);

/* snap_alg_agcd_exqrgcdwc_sup_base.c */
void supdc_alg_qr_dgv2010refine_sylvester(supdc_t poly_inf, supdc_t poly_ing, sdmdc_t matQ, sdmdc_t matR);
int supdc_agcd_exqrgcdwc(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol, int mode);

/* snap_alg_agcd_qrgcd_sup_base.c */
int supd_agcd_qrgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol);
int supdc_agcd_qrgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol);
int supfr_agcd_qrgcd(supfr_t poly_f, supfr_t poly_g, supfr_t poly_d, supfr_t poly_f1, supfr_t poly_g1, const mpfr_t tol);
int supfc_agcd_qrgcd(supfc_t poly_f, supfc_t poly_g, supfc_t poly_d, supfc_t poly_f1, supfc_t poly_g1, const mpfr_t tol);

/* snap_alg_agcd_qrgcd2_sup_base.c */
int supd_agcd_qrgcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol, int mode_opt_in);
int supdc_agcd_qrgcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol, int mode_opt_in);

/* snap_alg_agcd_uvgcd_sup_base.c */
int supd_agcd_uvgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const double tol);
int supdc_agcd_uvgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const double tol);
int supfr_agcd_uvgcd(supfr_t poly_f, supfr_t poly_g, supfr_t poly_d, supfr_t poly_f1, supfr_t poly_g1, const const mpfr_t tol);
int supfc_agcd_uvgcd(supfc_t poly_f, supfc_t poly_g, supfc_t poly_d, supfc_t poly_f1, supfc_t poly_g1, const const mpfr_t tol);
int supd_agcd_uvgcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const double tol, const int mode_opt);
int supdc_agcd_uvgcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const double tol, const int mode_opt);
int supd_agcd_uvgcd_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol);
int supdc_agcd_uvgcd_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol);
int supd_agcd_uvgcdy_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol);
int supdc_agcd_uvgcdy_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol);

/* snap_alg_agcd_qrgcdwp_sup_base.c */
int supd_agcd_qrgcdwp(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol);
int supd_agcd_qrgcdfp(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol);
int supdc_agcd_qrgcdwp(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol);
int supdc_agcd_qrgcdfp(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol);

/* snap_alg_agcd_exqrgcdwp_sup_base.c */
int supd_agcd_exqrgcdwp(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol, int mode);
int supd_agcd_exqrgcdfp(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol, int mode);
int supdc_agcd_exqrgcdwp(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol, int mode);
int supdc_agcd_exqrgcdfp(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol, int mode);

/* snap_alg_agcd_uvgcdwp_sup_base.c */
int supd_agcd_uvgcdwp(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, double tol);
int supdc_agcd_uvgcdwp(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, double tol);

/* snap_alg_agcd_fastgcd_sup_base.c */
int supd_agcd_fastgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const double tol);
int supdc_agcd_fastgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const double tol);
int supd_agcd_fastgcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const double tol, const unsigned int mode);
int supdc_agcd_fastgcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const double tol, const unsigned int mode);

/* snap_alg_agcd_gpgcd_sup_base.c */
int supd_agcd_gpgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d);
int supdc_agcd_gpgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d);
int supd_agcd_gpgcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol);
int supdc_agcd_gpgcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol);
int supd_agcd_gpgcd_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supdc_agcd_gpgcd_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);

/* snap_alg_agcd_stlngcd_sup_base.c */
int supd_agcd_stlngcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol);
int supdc_agcd_stlngcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol);
int supd_agcd_stlngcd_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supdc_agcd_stlngcd_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supd_agcd_stlngcd2(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol);
int supdc_agcd_stlngcd2(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol);
int supd_agcd_stlngcd2_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supdc_agcd_stlngcd2_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supd_agcd_stlngcd2x_ex(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);
int supdc_agcd_stlngcd2x_ex(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double tol, const unsigned int max_iterations, const int mode);

/* snap_alg_agcd_ghlgcd_sup_base.c */
int supd_agcd_ghlgcd(supd_t poly_f, supd_t poly_g, supd_t poly_d, supd_t poly_f1, supd_t poly_g1, const unsigned int d, const double threshold);
int supdc_agcd_ghlgcd(supdc_t poly_f, supdc_t poly_g, supdc_t poly_d, supdc_t poly_f1, supdc_t poly_g1, const unsigned int d, const double threshold);




// int supfc_agcd_fastgcd(supfc_t poly_f, supfc_t poly_g, supfc_t poly_d, supfc_t poly_f1, supfc_t poly_g1, const mpfr_t tol);

#endif
