/*
 * Copyright (c) 2012-2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <snap.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <fftw3.h>
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

// define if shows elapsed time
// #define SNAPLIB_SHOW_ELAPSED_TIME

/* ================== START [agcd_exqrgcd2] ================== */
/* SEE ALSO: ExQRGCD algorithm introduced by K.Nagasaka and T.Masui. */
typedef enum {NORMAL, REVERSAL, SPLITALLOWED, SPLITDISALLOWED, BEZOUTONLY, NORMALONLY} SNAP_agcd_exqrgcd_qrstep_order;
// NORMAL: current order and first order in the main routine
// REVERSAL: current order and first order in the main routine
// SPLITALLOWED: split will be called in the case 3
// SPLITDISALLOWED: split won't be called even in the case 3
// BEZOUTONLY: only compute the bezout coefficients
// NORMALONLY: this effects only the debug message (e.g. non verbose mode)
// #define SNAPLIB_MPFR_RND MPFR_RNDN
// #define SNAPLIB_MPC_RND MPC_RNDNN
// ========== THE FOLLOWING MACROS SHOULD BE MODIFIED TO BE COMPATIBLE WITH MPFR AND MPC. ==========
#define SNAPLIB_MIN(x,y) ((x)<(y) ? (x):(y))
#define SNAPLIB_MAX(x,y) ((x)>(y) ? (x):(y))

typedef enum { SNAP_EXQRGCD2_REFINEMENT_NONE, SNAP_EXQRGCD2_REFINEMENT_UVGCD, SNAP_EXQRGCD2_REFINEMENT_FASTGCD, SNAP_EXQRGCD2_REFINEMENT_GPGCD } mode_type;

int SNAP_agcd_exqrgcd_split(SNAP_t poly_inside, SNAP_t poly, const SNAPBASEREAL_t tol);
void SNAP_agcd_exqrgcd_quicksort_uint_by_base(unsigned int *idx, SNAPBASEREAL_t *val, unsigned int len);
void SNAP_agcd_exqrgcd_Qkthcolumn(SNAPBASE_t *vect, SNAPMTX_t matrix_gqr, SNAPBASE_t *tau, const unsigned int k);
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
void SNAP_agcd_fastgcd_refine(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXCOMPLEX_t vecDt, SNAPMTXCOMPLEX_t vecDh, SNAPMTXCOMPLEX_t vecD1, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPMTXCOMPLEX_t matW, SNAPMTXCOMPLEX_t genTG, SNAPMTXCOMPLEX_t genTB, unsigned int *matP, unsigned int *matQ);
int SNAP_agcd_gpgcd_modified_newton2(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int max_iterations, const SNAPBASEREAL_t d_threshold, const SNAPBASEREAL_t tol);
void SNAP_agcd_gpgcd_construct_tildevectors(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g);
void SNAP_agcd_gpgcd_lsdiv(SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int m, const unsigned int n, const unsigned int d);

static unsigned int SNAP_agcd_exqrgcd_qrstep2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_cof, SNAP_t poly_cog, SNAP_t poly_d, SNAP_t poly_u, SNAP_t poly_v, const SNAPBASEREAL_t tol, const SNAPBASEREAL_t tol_temp, SNAPBASEREAL_t tol_res, SNAP_agcd_exqrgcd_qrstep_order qrstep_order, SNAP_agcd_exqrgcd_qrstep_order split_flag, mode_type mode_opt, SNAPMTXCOMPLEX_t vecDt, SNAPMTXCOMPLEX_t vecDh, SNAPMTXCOMPLEX_t vecD1, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPMTXCOMPLEX_t matW, SNAPMTXCOMPLEX_t genTG, SNAPMTXCOMPLEX_t genTB, unsigned int *matP, unsigned int *matQ, SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b) {
  // We'd like to try to split then force to use PRS.
  SNAPMTX_t matrix_syl;
  SNAP_t poly_r, poly_t, poly_ncf, poly_ncg, poly_prs1, poly_prs2;
  SNAPBASE_t *tau, *Qkth;
  SNAPBASEREAL_t prev_Rnorm, kth_Rnorm, sqrtcols, tol_f, tol_g, const_tmp, const_case1, *erval_list;
  unsigned int i,j,k,found_degree=UINT_MAX,case1flag=0,ergapk,ergapcount;
  unsigned int *eridx_list, mincdeg;
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
    printf("entering qrstep\n");
  }
#endif
  if ( SNAP_degree(poly_cof) == 0 || SNAP_degree(poly_cog) == 0 ) {
    return 0;
  }
  // Initialization and Sylvester matrix
  SNAPMTX_init(matrix_syl);
  SNAPMTX_sylvester_matrix_transpose(matrix_syl, poly_cof, poly_cog);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // QR decomposition
#ifdef SNAP_SUPD_T
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#endif
  ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: dgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#endif
  ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: zgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (unsigned int q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+f%+fI),", creal(SNAPMTX_elem(matrix_syl,q1,q2)), cimag(SNAPMTX_elem(matrix_syl,q1,q2)));
    }
    printf("\n");
  }
  printf("tau: ");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    printf("(%+f%+fI),", creal(*(tau+q1)), cimag(*(tau+q1)));
  }
  printf("\n");
#endif
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (unsigned int q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN), mpfr_get_ld(mpc_imagref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN));
    }
    printf("\n");
  }
  printf("tau: ");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(*(tau+q1)), MPFR_RNDN), mpfr_get_ld(mpc_imagref(*(tau+q1)), MPFR_RNDN));
  }
  printf("\n");
#endif
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (QR): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // Now we seek the relative smallest gap to recover a candidate approximate gcd.
  SNAPBASEREAL_init(sqrtcols);
  SNAPBASEREAL_sqrt_ui(sqrtcols, SNAPMTX_cols(matrix_syl));
  SNAPBASEREAL_init_set_ui(prev_Rnorm, 0);
  SNAPBASEREAL_init(kth_Rnorm);
  SNAPBASE_abs(kth_Rnorm, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  k = 0;
  SNAPBASEREAL_init(const_tmp);
  SNAPBASEREAL_init(const_case1);
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    mpfr_printf("%d-th Rnorm: %Re\n", k, kth_Rnorm);
#else
    printf("%d-th Rnorm: %e\n", k, kth_Rnorm);
#endif
  }
#endif
  SNAPBASEREAL_mul(const_case1, tol_temp, sqrtcols);
  if ( SNAPBASEREAL_cmp(kth_Rnorm, const_case1) > 0 ) { // CASE 1 (there is no chance to have agcd)
    case1flag = 1;
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      mpfr_printf("CASE 1 detected! (%Re > %Re=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#else
      printf("CASE 1 detected! (%e > %e=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#endif
    }
#endif
    // for CASE 1, we do nothing here.
  } else if ( qrstep_order != BEZOUTONLY ) { // CASES 2 and 3 (there is a possibility to have agcd)
    mincdeg = SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog));
    erval_list = (SNAPBASEREAL_t*) malloc(sizeof(SNAPBASEREAL_t)*mincdeg);
    for ( i=0; i<SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog)); i++ ) {
      SNAPBASEREAL_init(*(erval_list+i));
    }
    eridx_list = (unsigned int*) malloc(sizeof(unsigned int)*mincdeg);
    while ( ++k <= mincdeg ) {
      // F-norm can be computed by ...
      SNAPBASEREAL_mul(kth_Rnorm, kth_Rnorm, kth_Rnorm);
      SNAPBASEREAL_mul(prev_Rnorm, prev_Rnorm, prev_Rnorm);
      SNAPBASEREAL_add(prev_Rnorm, prev_Rnorm, kth_Rnorm);
      SNAPBASEREAL_sqrt(prev_Rnorm, prev_Rnorm);
      if ( SNAPBASEREAL_cmp(prev_Rnorm, const_case1) > 0 ) {
        break;  // seeking finished.
      }
      // Update kth_Rnorm
      SNAPMTX_BLAS_nrm2(kth_Rnorm, k+1, SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-k,SNAPMTX_rows(matrix_syl)-1-k),1);
      *(eridx_list+k-1) = k;
      SNAPBASEREAL_div(*(erval_list+k-1), prev_Rnorm, kth_Rnorm);
      if ( SNAPBASEREAL_sgn(prev_Rnorm) == 0 || SNAPBASEREAL_nan_p(*(erval_list+k-1)) ) {
        SNAPBASEREAL_set_ui(*(erval_list+k-1),1);
      }
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        mpfr_printf("%d-th Rnorm: %Re, e_r: %Re (Case1 threshold: %Re)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#else
        printf("%d-th Rnorm: %e, e_r: %e (Case1 threshold: %e)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#endif
      }
#endif
    }
    // sort erval_list and corresponding eridx_list by ascending order.
    SNAP_agcd_exqrgcd_quicksort_uint_by_base(eridx_list, erval_list, k-1);
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %e\n", *(eridx_list+i), *(erval_list+i));
      }
    }
#endif
    // Divisibility Check
    SNAP_init(poly_r);
    SNAP_init(poly_t);
    SNAP_init(poly_ncf);
    SNAP_init(poly_ncg);
    SNAPBASEREAL_init(tol_f);
    SNAPBASEREAL_init(tol_g);
    // for ( i=0; i<k-1; i++ ) { // find a candidate gcd until the given tolerance is satisfied.
    for ( i=0; i<SNAPLIB_MIN(3,k-1); i++ ) { // Top 3 Check
      // if ( split_flag == SPLITALLOWED && sqrt(tol_temp) <= (*(erval_list+i)) ) { break; }
      // poly_r <= the last $(k+1)$-th row of R
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i), SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i)), *(eridx_list+i)+1);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
      if ( SNAPBASE_sgn(SNAP_lcoef(poly_r)) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPDC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPFC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#endif
      // poly_t <= poly_d * poly_r
      SNAP_mul(poly_t, poly_d, poly_r);
      // poly_t <= poly_t/norm(poly_t) : normalization
      SNAP_2norm(const_tmp, poly_t);
      SNAP_div_basereal(poly_t, poly_t, const_tmp);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncf, poly_f, poly_t);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncg, poly_g, poly_t);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_UVGCD ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, &tol_f);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, tol_f);
#endif
      } else if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_FASTGCD ) {
        SNAP_agcd_fastgcd_refine(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ);
      } else if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_GPGCD ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
        SNAPMTXREAL_realloc(cofact_a, 1, 2*(SNAP_degree(poly_ncg)+1));
        SNAPMTXREAL_realloc(cofact_b, 1, 2*(SNAP_degree(poly_ncf)+1));
#else
        SNAPMTXREAL_realloc(cofact_a, 1, SNAP_degree(poly_ncg)+1);
        SNAPMTXREAL_realloc(cofact_b, 1, SNAP_degree(poly_ncf)+1);
#endif // defined
        SNAP_agcd_gpgcd_construct_tildevectors(poly_f, poly_g, input_f, input_g);
        SNAP_init(poly_prs1);
        SNAP_init(poly_prs2);
        SNAP_mul(poly_prs1, poly_t, poly_ncf);
        SNAP_mul(poly_prs2, poly_t, poly_ncg);
        SNAP_agcd_gpgcd_construct_tildevectors(poly_prs1, poly_prs2, tilde_f, tilde_g);
        SNAP_clear(poly_prs1);
        SNAP_clear(poly_prs2);
        SNAP_mul_si(poly_ncf, poly_ncf, -1);
        SNAP_agcd_gpgcd_construct_tildevectors(poly_ncf, poly_ncg, cofact_b, cofact_a);
        SNAP_agcd_gpgcd_modified_newton2(input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b, 100, 1.0e-8, tol);
        SNAP_agcd_gpgcd_lsdiv(poly_t, poly_ncf, poly_ncg, tilde_f, tilde_g, cofact_a, cofact_b, SNAP_degree(poly_f), SNAP_degree(poly_g), SNAP_degree(poly_t));
      }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      // check the tolerance
      SNAP_mul(poly_r, poly_ncf, poly_t);
      SNAP_sub(poly_r, poly_f, poly_r);
      SNAP_2norm(tol_f, poly_r);
      if ( SNAPBASEREAL_cmp(tol_f,tol) < 0 ) { // continue
        // check the tolerance
        SNAP_mul(poly_r, poly_ncg, poly_t);
        SNAP_sub(poly_r, poly_g, poly_r);
        SNAP_2norm(tol_g, poly_r);
        if ( SNAPBASEREAL_cmp(tol_g,tol) < 0 ) { // approximate gcd found!!
          // update several polynomials.
          SNAP_set(poly_cof, poly_ncf);
          SNAP_set(poly_cog, poly_ncg);
          SNAP_set(poly_d, poly_t);
          found_degree = *(eridx_list+i);
          if ( SNAPBASEREAL_cmp(tol_f, tol_g) < 0 ) {
            SNAPBASEREAL_set(tol_res, tol_g);
          } else {
            SNAPBASEREAL_set(tol_res, tol_f);
          }
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("Number of tries: %d (Degree: %d)\n", i+1, found_degree);
          }
#endif
          break;
        }
      }
    }
    SNAP_clear(poly_r);
    SNAP_clear(poly_t);
    SNAP_clear(poly_ncf);
    SNAP_clear(poly_ncg);
    SNAPBASEREAL_clear(tol_f);
    SNAPBASEREAL_clear(tol_g);
    // Check Difficult case or Not
    if ( found_degree == UINT_MAX ) {
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
        printf("Difficult CASE detected!\n");
      }
#endif
      if ( split_flag == SPLITALLOWED ) { // Call Split and qrstep recursively.
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is called...\n");
        }
#endif

        /* ========================== find gap beyond k ============================== */
        // compute erval from the k skipped.
        ergapk = k;
        ergapcount = 0;
        for ( i=k; i <= mincdeg; i++ ) {
          // Update kth_Rnorm
          SNAPMTX_BLAS_nrm2(kth_Rnorm, i+1,SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-i,SNAPMTX_rows(matrix_syl)-1-i),1);
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
            mpfr_printf("%d-th Rnorm: %Re (beyond k)\n", i, kth_Rnorm);
#else
            printf("%d-th Rnorm: %e (beyond k)\n", i, kth_Rnorm);
#endif
          }
#endif
          if ( SNAPBASEREAL_cmp_ui(kth_Rnorm, 1) >= 0 ) {
            ergapcount++;
            if ( ergapcount >= 2 ) {
              ergapk = i-1;
              break;
            }
          } else {
            ergapcount = 0;
          }
        }
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("Largest k=%d satisyfing small(1.0) gap beyond the case 1\n",ergapk);
        }
#endif
        /* ========================== find gap beyond k ============================== */

        SNAP_init(poly_ncf);
        SNAP_init(poly_ncg);
        SNAP_init(poly_prs1);
        SNAP_init(poly_prs2);
        for ( i=ergapk; i<=ergapk; i++ ) {
          SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-2-i), i+2);
          SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, SNAPMTX_rows(matrix_syl)-1-i), i+1);
          SNAP_2norm(kth_Rnorm, poly_prs1);
          SNAP_2norm(prev_Rnorm, poly_prs2);
          SNAP_div_basereal(poly_prs1, poly_prs1, kth_Rnorm);
          SNAP_div_basereal(poly_prs2, poly_prs2, prev_Rnorm);
#ifdef SNAPLIB_DEBUG
          printf("To be splitted (k=%d):\n", i);
#endif
          SNAPBASEREAL_sqrt_ui(const_tmp, SNAPLIB_MAX(SNAP_degree(poly_cof),SNAP_degree(poly_cog)));
          SNAPBASEREAL_mul_ui(const_tmp, const_tmp, 2);
          SNAPBASEREAL_mul(const_tmp, const_tmp, tol_temp);
          if ( SNAPBASEREAL_cmp(kth_Rnorm, prev_Rnorm) < 0 ) {
            SNAPBASEREAL_div(const_tmp, const_tmp, kth_Rnorm);
          } else {
            SNAPBASEREAL_div(const_tmp, const_tmp, prev_Rnorm);
          }
          if ( SNAP_agcd_exqrgcd_split(poly_ncf, poly_prs1, tol_temp) == 0 && SNAP_agcd_exqrgcd_split(poly_ncg, poly_prs2, tol_temp) == 0 ) {
            found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f, poly_g, poly_ncf, poly_ncg, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_ncf);
              SNAP_set(poly_cog, poly_ncg);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
              SNAPBASE_clear(*(tau+j));
            }
            free(tau);
            SNAPMTX_clear(matrix_syl);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
            return found_degree;
          } else {
#ifdef SNAPLIB_DEBUG
            if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
              printf("SPLIT failed...\n");
            }
#endif
            found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f, poly_g, poly_prs1, poly_prs2, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_prs1);
              SNAP_set(poly_cog, poly_prs2);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
              SNAPBASE_clear(*(tau+j));
            }
            free(tau);
            SNAPMTX_clear(matrix_syl);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
            return found_degree;
          }
        }
        SNAP_clear(poly_ncf);
        SNAP_clear(poly_ncg);
        SNAP_clear(poly_prs1);
        SNAP_clear(poly_prs2);
      } else {
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is not allowed. Skipped...\n");
        }
#endif
      }
    }
    for ( j=0; j<mincdeg; j++ ) {
      SNAPBASEREAL_clear(*(erval_list+j));
    }
    free(erval_list);
    free(eridx_list);
  }
  if ( found_degree == UINT_MAX ) {
    // poly_d <== 1.0 hence we do nothing for poly_d
    // SNAP_degree(poly_d) = 0; SNAP_set_coef(poly_d, 0, 1.0);
    if ( case1flag == 1 ) {
      found_degree = 0;
    }
    // poly_u and poly_v be the cofactors from the last row of Q.
    // NOTE: poly_u is the cofactor of poly_g and poly_v is of poly_f.
    // OLD if ( qrstep_order == NORMAL || qrstep_order == NORMALONLY || qrstep_order == BEZOUTONLY ) {
    if ( qrstep_order == BEZOUTONLY ) {
      Qkth = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_init(*(Qkth+j));
      }
      SNAP_agcd_exqrgcd_Qkthcolumn(Qkth, matrix_syl, tau, SNAPMTX_rows(matrix_syl)-1);
#if defined(SNAP_SUPDC_T) && defined(SNAPLIB_DEBUG)
      printf("AAAA\n");
      for (int q3=0; q3<SNAPMTX_rows(matrix_syl); q3++ ) {
        printf("(%+.5le%+.5leI),",creal(*(Qkth+q3)),cimag(*(Qkth+q3)));
      }
      printf("\n");
#endif
      SNAP_set_rbase(poly_u, Qkth, SNAP_degree(poly_cog));
      SNAP_set_rbase(poly_v, Qkth+SNAP_degree(poly_cog), SNAP_degree(poly_cof));
      /*
      			for ( i=0; i<SNAP_degree(poly_cog); i++ ) {
      				SNAP_set_coef(poly_u, i, *(Qkth+SNAP_degree(poly_cog)-1-i));
      			}
      			SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
      			for ( i=0; i<SNAP_degree(poly_cof); i++ ) {
      				SNAP_set_coef(poly_v, i, *(Qkth+SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i));
      			}
      			SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;
      */
      SNAP_div_base(poly_u, poly_u, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      SNAP_div_base(poly_v, poly_v, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_clear(*(Qkth+j));
      }
      free(Qkth);
    }
  }
  for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
    SNAPBASE_clear(*(tau+j));
  }
  free(tau);
  SNAPMTX_clear(matrix_syl);
  SNAPBASEREAL_clear(prev_Rnorm);
  SNAPBASEREAL_clear(kth_Rnorm);
  SNAPBASEREAL_clear(sqrtcols);
  SNAPBASEREAL_clear(const_tmp);
  SNAPBASEREAL_clear(const_case1);
  return found_degree;
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
// mode 0: automatic first order
// mode 1: normal side
// mode 2: reversal side
// mode_opt_in 0: none refinement
// mode_opt_in 1: UVGCD's refinement
// mode_opt_in 2: Fastgcd's refinement
// mode_opt_in 3: GPGCD's refinement
int SNAP_agcd_exqrgcd2(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, int mode, int mode_opt_in) {
  SNAPBASEREAL_t norm_f, norm_g, tol_temp, tol_res, norm_cf, norm_cg, lcf, lcg, scf, scg;
  SNAP_t poly_f, poly_g, poly_cof, poly_cog, poly_u, poly_v;
  SNAP_agcd_exqrgcd_qrstep_order qrstep_order, qrstep_firstorder;
  unsigned int found_degree, trials, coprimeresults;
  int result = -1;
  SNAPMTXCOMPLEX_t vecDt, vecDh, vecD1, genG, genB, genTG, genTB, matW;
  SNAPMTXREAL_t input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b;
  unsigned int *matP, *matQ;
  mode_type mode_opt;
  if ( mode_opt_in == 1 ) {
    mode_opt = SNAP_EXQRGCD2_REFINEMENT_UVGCD;
  } else if ( mode_opt_in == 2) {
    mode_opt = SNAP_EXQRGCD2_REFINEMENT_FASTGCD;
  } else if ( mode_opt_in == 3) {
    mode_opt = SNAP_EXQRGCD2_REFINEMENT_GPGCD;
  } else {
    mode_opt = SNAP_EXQRGCD2_REFINEMENT_NONE;
  }
  // initialization
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(tol_temp);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(norm_cf);
  SNAPBASEREAL_init(norm_cg);
  SNAPBASEREAL_init(lcf);
  SNAPBASEREAL_init(lcg);
  SNAPBASEREAL_init(scf);
  SNAPBASEREAL_init(scg);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
  SNAP_init_set(poly_cog, poly_g);
  SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
  SNAP_degree(poly_d) = 0;
  SNAPBASEREAL_set(tol_temp, tol);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  mpfr_printf("input tolerance: %Re\n", tol);
#else
  printf("input tolerance: %e\n", tol);
#endif
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_cof)) || SNAPBASE_zeroq(SNAP_lcoef(poly_cog)) ) {
    SNAP_clear(poly_f);
    SNAP_clear(poly_g);
    SNAP_clear(poly_cof);
    SNAP_clear(poly_cog);
    SNAPBASEREAL_clear(norm_f);
    SNAPBASEREAL_clear(norm_g);
    SNAPBASEREAL_clear(tol_temp);
    SNAPBASEREAL_clear(tol_res);
    SNAPBASEREAL_clear(norm_cf);
    SNAPBASEREAL_clear(norm_cg);
    SNAPBASEREAL_clear(lcf);
    SNAPBASEREAL_clear(lcg);
    SNAPBASEREAL_clear(scf);
    SNAPBASEREAL_clear(scg);
    fprintf(stderr, "EXQRGCD: polynomials must be non-zero!\n");
    return -1;
  }
  if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_FASTGCD ) {
    SNAPMTXCOMPLEX_init(vecDt);
    SNAPMTXCOMPLEX_init(vecDh);
    SNAPMTXCOMPLEX_init(vecD1);
    SNAPMTXCOMPLEX_init(genG);
    SNAPMTXCOMPLEX_init(genB);
    SNAPMTXCOMPLEX_init(genTG);
    SNAPMTXCOMPLEX_init(genTB);
    SNAPMTXCOMPLEX_init(matW);
    matP = (unsigned int*) malloc(sizeof(unsigned int)*(SNAP_degree(poly_f)+SNAP_degree(poly_g)+3)); // the size considering Jacobian
    matQ = (unsigned int*) malloc(sizeof(unsigned int)*(SNAP_degree(poly_f)+SNAP_degree(poly_g)+3)); // the size considering Jacobian
  }
  if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_GPGCD ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    SNAPMTXREAL_init2(input_f, 1, 2*(SNAP_degree(poly_f)+1));
    SNAPMTXREAL_init2(input_g, 1, 2*(SNAP_degree(poly_g)+1));
    SNAPMTXREAL_init2(tilde_f, 1, 2*(SNAP_degree(poly_f)+1));
    SNAPMTXREAL_init2(tilde_g, 1, 2*(SNAP_degree(poly_g)+1));
#else
    SNAPMTXREAL_init2(input_f, 1, SNAP_degree(poly_f)+1);
    SNAPMTXREAL_init2(input_g, 1, SNAP_degree(poly_g)+1);
    SNAPMTXREAL_init2(tilde_f, 1, SNAP_degree(poly_f)+1);
    SNAPMTXREAL_init2(tilde_g, 1, SNAP_degree(poly_g)+1);
#endif // defined
    SNAPMTXREAL_init(cofact_a);
    SNAPMTXREAL_init(cofact_b);
  }
  // depending on normal first or reversal first ?
  SNAPBASE_abs(lcf, SNAP_lcoef(poly_f));
  SNAPBASE_abs(lcg, SNAP_lcoef(poly_g));
  SNAPBASE_abs(scf, SNAP_coef(poly_f,0));
  SNAPBASE_abs(scg, SNAP_coef(poly_g,0));
  if ( SNAPBASEREAL_cmp(lcf,lcg) > 0 ) {
    SNAPBASEREAL_set(lcf, lcg);
  }
  if ( SNAPBASEREAL_cmp(scf,scg) > 0 ) {
    SNAPBASEREAL_set(scf, scg);
  }
  if ( mode == 1 || ( mode == 0 && SNAPBASEREAL_cmp(lcf,scf) >= 0 ) ) {
    qrstep_firstorder = NORMAL;
  } else {
    qrstep_firstorder = REVERSAL;
  }
#ifdef SNAPLIB_DEBUG
  printf("======= Order: %s =======\n", qrstep_firstorder == NORMAL ? "Normal":"Reversal" );
#endif
  // some agcd will be found then return, otherwise change qrstep_order.
  SNAPBASEREAL_set_ui(norm_cf, 1);
  SNAPBASEREAL_set_ui(norm_cg, 1);
  trials = 0;
  coprimeresults = 0;
  qrstep_order = qrstep_firstorder;
  while (1) {
    if ( trials > 0 || qrstep_firstorder == REVERSAL ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
    } else {
      found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      if ( trials == 1 ) {
        break;  // re-try from the other side.
      }
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  if ( result == -1 ) {
#ifdef SNAPLIB_DEBUG
    printf("======= Second Order: %s =======\n", qrstep_firstorder != NORMAL ? "Normal":"Reversal" );
#endif
    if ( qrstep_firstorder == qrstep_order ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
    }
    SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
    SNAP_init_set(poly_cog, poly_g);
    SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
    SNAP_degree(poly_d) = 0;
    SNAPBASEREAL_set(tol_temp, tol);
    trials = 0;
    coprimeresults = 0;
    qrstep_order = (qrstep_firstorder == NORMAL ? REVERSAL:NORMAL);
  }
  while ( result == -1 ) {
    if ( trials > 0 ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
    } else {
      found_degree = SNAP_agcd_exqrgcd_qrstep2(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED, mode_opt, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ, input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAP_clear(poly_cof);
  SNAP_clear(poly_cog);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(tol_temp);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(norm_cf);
  SNAPBASEREAL_clear(norm_cg);
  SNAPBASEREAL_clear(lcf);
  SNAPBASEREAL_clear(lcg);
  SNAPBASEREAL_clear(scf);
  SNAPBASEREAL_clear(scg);
  if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_FASTGCD ) {
    SNAPMTXCOMPLEX_clear(vecDt);
    SNAPMTXCOMPLEX_clear(vecDh);
    SNAPMTXCOMPLEX_clear(vecD1);
    SNAPMTXCOMPLEX_clear(genG);
    SNAPMTXCOMPLEX_clear(genB);
    SNAPMTXCOMPLEX_clear(genTG);
    SNAPMTXCOMPLEX_clear(genTB);
    SNAPMTXCOMPLEX_clear(matW);
    free(matP);
    free(matQ);
  }
  if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_GPGCD ) {
    SNAPMTXREAL_clear(input_f);
    SNAPMTXREAL_clear(input_g);
    SNAPMTXREAL_clear(tilde_f);
    SNAPMTXREAL_clear(tilde_g);
    SNAPMTXREAL_clear(cofact_a);
    SNAPMTXREAL_clear(cofact_b);
  }
#ifdef SNAPLIB_DEBUG
  printf("Finally the number of trials: %d\n", trials);
#endif
  return result;
}
/* ================== END[agcd_over_reals] ================== */
