/*
 * Copyright (c) 2012-2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// just for writing codes
// #define SNAP_SUPDC_T

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <snap.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <fftw3.h>
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// The following macro enables the use of FFTW for FFT
#define SNAP_QR_USE_FFTW

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

/* ================== START [agcd_exqrgcd] ================== */
/* SEE ALSO: ExQRGCD algorithm introduced by K.Nagasaka and T.Masui. */
/* THIS ROUTINE USES DGV2010 QR FACTORIZATION (REFINED VERSION) */
typedef enum {NORMAL, REVERSAL, SPLITALLOWED, SPLITDISALLOWED, BEZOUTONLY, NORMALONLY} SNAP_agcd_exqrgcdwc_qrstep_order;
// NORMAL: current order and first order in the main routine
// REVERSAL: current order and first order in the main routine
// SPLITALLOWED: split will be called in the case 3
// SPLITDISALLOWED: split won't be called even in the case 3
// BEZOUTONLY: only compute the bezout coefficients
// NORMALONLY: this effects only the debug message (e.g. non verbose mode)
// #define SNAPLIB_MPFR_RND MPFR_RNDN
// #define SNAPLIB_MPC_RND MPC_RNDNN
// ========== THE FOLLOWING MACROS SHOULD BE MODIFIED TO BE COMPATIBLE WITH MPFR AND MPC. ==========
#define SNAPLIB_MIN(x,y) ((x)<(y) ? (x):(y))
#define SNAPLIB_MAX(x,y) ((x)>(y) ? (x):(y))
static unsigned int SNAP_agcd_exqrgcdwc_qrstep(SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, const SNAPBASEREAL_t, const SNAPBASEREAL_t, SNAPBASEREAL_t, SNAP_agcd_exqrgcdwc_qrstep_order, SNAP_agcd_exqrgcdwc_qrstep_order);
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);

// Graeffe's Root Squaring (assume deg(poly_f) >= 2)
void SNAP_agcd_exqrgcdwc_rootsquaring(SNAP_t poly_r, SNAP_t poly_f) {
  unsigned int i;
  SNAP_t poly_t, poly_x;
  SNAP_init2(poly_x, 2);
  SNAPBASE_set_d(SNAP_coef(poly_x, 1), 1.0);
  SNAP_degree(poly_x) = 1;
  if ( SNAP_degree(poly_f) % 2 == 1 ) {
    i = (SNAP_degree(poly_f)+1)/2;
    SNAP_init2(poly_t, i);
    i--;
    SNAP_degree(poly_t) = i;
    do {
      SNAPBASE_set(SNAP_coef(poly_t, i), SNAP_coef(poly_f, 2*i+1));
    } while ( i-- != 0 );
    SNAP_mul(poly_r, poly_t, poly_t);
    SNAP_mul(poly_r, poly_r, poly_x);
    i = (SNAP_degree(poly_f)+1)/2;
    i--;
    do {
      SNAPBASE_set(SNAP_coef(poly_t, i), SNAP_coef(poly_f, 2*i));
    } while ( i-- != 0 );
    SNAP_mul(poly_t, poly_t, poly_t);
    SNAP_sub(poly_r, poly_r, poly_t);
  } else {
    i = SNAP_degree(poly_f)/2+1;
    SNAP_init2(poly_t, i);
    i--;
    SNAP_degree(poly_t) = i;
    do {
      SNAPBASE_set(SNAP_coef(poly_t, i), SNAP_coef(poly_f, 2*i));
    } while ( i-- != 0 );
    SNAP_mul(poly_r, poly_t, poly_t);
    i = SNAP_degree(poly_f)/2+1;
    i-=2;
    SNAP_degree(poly_t) = i;
    do {
      SNAPBASE_set(SNAP_coef(poly_t, i), SNAP_coef(poly_f, 2*i+1));
    } while ( i-- != 0 );
    SNAP_mul(poly_t, poly_t, poly_t);
    SNAP_mul(poly_t, poly_t, poly_x);
    SNAP_sub(poly_r, poly_r, poly_t);
  }
  SNAP_clear(poly_t);
  SNAP_clear(poly_x);
}

#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
// FFT: ans will be over written and it should be coefficients at the input.
// ans and unity are lists of length len. unity is with LDA incr.
// work is the list of temporary use of the same length len.
static void SNAP_agcd_exqrgcdwc_fft(SNAPBASECOMPLEX_t *list_ans, SNAPBASECOMPLEX_t *list_work, SNAPBASECOMPLEX_t *list_unity, const unsigned int len, const unsigned int incr) {
  unsigned int i, len_half;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3, *pt4;
  if ( len == 1 ) {
    return;
  }
  len_half = len/2;
  for( i=0,pt1=list_work,pt2=list_ans,pt3=list_ans+len_half,pt4=list_unity; i<len_half; i++,pt1++,pt2++,pt3++,pt4+=incr ) {
    SNAPBASECOMPLEX_add(*pt1, *pt2, *pt3);
    SNAPBASECOMPLEX_sub(*pt3, *pt2, *pt3);
    SNAPBASECOMPLEX_mul(*pt3, *pt3, *pt4);
  }
  // recursive calls
  SNAP_agcd_exqrgcdwc_fft(list_work, list_work+len_half, list_unity, len_half, incr*2);
  SNAP_agcd_exqrgcdwc_fft(list_ans+len_half, list_work+len_half, list_unity, len_half, incr*2);
  // re-order
  for( i=0,pt1=list_ans,pt2=list_work,pt3=list_ans+1,pt4=list_ans+len_half; i<len_half; i++,pt1+=2,pt2++,pt3+=2,pt4++ ) {
    SNAPBASECOMPLEX_set(*pt1, *pt2);
    SNAPBASECOMPLEX_set(*pt3, *pt4);
  }
}
#endif

// Rough Splitting by numerical integration and Newton's formula. see
// Optimal and Nearly Optimal Algorithms for Approximating Polynomial Zeros. V. Y. PAN
// A. Schonhage, The fundamental theorem of algebra in terms of computational complexity.
// We assume that the degree of poly_f < n
int SNAP_agcd_exqrgcdwc_roughsplit(SNAP_t poly_r, SNAP_t poly_f, const unsigned int n) {
  int res=0;
  unsigned int i, j, deg_inside;
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASE_prec_t dprec;
  SNAPBASECOMPLEX_t const_pin, *list_unity, *list_work;
#endif
  SNAPBASECOMPLEX_t *list_df, *list_f, const_tmp, *pt1, *pt2, *pt3;
  SNAPBASE_t *ptfr;
  // initialization for FFT
  list_f = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  list_df = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  list_unity = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  list_work = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
//        for(i=0,pt1=list_unity;i<n;i++,pt1++) { *pt1 = cexp(i*2*3.1415926535897932384626433832795029L*I/n); }
  dprec = SNAPBASE_get_default_prec();
  SNAPBASECOMPLEX_init2(const_pin, dprec);
  SNAPBASECOMPLEX_init2(const_tmp, dprec);
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), n);
  for(i=0,pt1=list_work; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_f; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_df; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_unity; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
    SNAPBASECOMPLEX_mul_ui(const_tmp, const_pin, i);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
#endif
  for(i=0,pt1=list_f,pt2=list_df,ptfr=SNAP_coefp(poly_f,0); i<SNAP_degree(poly_f); i++,pt1++,pt2++) {
    SNAPBASECOMPLEX_set_base(*pt1, *(ptfr++));
    SNAPBASECOMPLEX_set_base(*pt2, *ptfr);
    SNAPBASECOMPLEX_mul_ui(*pt2, *pt2, i+1);
  }
  SNAPBASECOMPLEX_set_base(*pt1, *ptfr);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  i++;
  pt1++;
  pt2++;
  for(; i<n; i++,pt1++,pt2++) {
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
  }
  // numerical integration by FFT
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_plan fftp;
  fftp = fftw_plan_dft_1d(n, list_f, list_f, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
  fftw_execute_dft(fftp, list_df, list_df);
#else
  SNAP_agcd_exqrgcdwc_fft(list_f, list_work, list_unity, n, 1);
  SNAP_agcd_exqrgcdwc_fft(list_df, list_work, list_unity, n, 1);
#endif
  for(i=0,pt1=list_f,pt2=list_df; i<n; i++,pt1++,pt2++) {
    SNAPBASECOMPLEX_div(*pt2, *pt2, *pt1);
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_execute_dft(fftp, list_df, list_df);
  fftw_destroy_plan(fftp);
#else
  SNAP_agcd_exqrgcdwc_fft(list_df, list_work, list_unity, n, 1);
#endif
  for(i=0,pt2=list_df; i<n; i++,pt2++) {
    SNAPBASECOMPLEX_div_ui(*pt2, *pt2, n);
  }
  // compute coefs by Newton's formula
#if defined(SNAP_SUPFC_T) || defined(SNAP_SUPFR_T)
  deg_inside = SNAPBASEREAL_get_ui_rnd(SNAPBASECOMPLEX_realref(*(list_df+1)));
#else
  deg_inside = SNAPBASEREAL_get_ui_rnd(SNAPBASECOMPLEX_real(*(list_df+1)));
#endif
#ifdef SNAPLIB_DEBUG
// mpfr_printf("deg_inside: %d (%Re)\n", deg_inside, SNAPBASECOMPLEX_realref(*(list_df+1)));
#endif
  if ( deg_inside <= 0 || deg_inside > SNAP_degree(poly_f) ) {
    res = -1;  // number of points is not enough to detect the degree.
  } else if ( deg_inside == SNAP_degree(poly_f) ) { // There is no need to split.
    SNAP_set(poly_r, poly_f);
  } else {
    SNAPBASECOMPLEX_set_ui(*(list_f+deg_inside), 1);
    SNAP_set_coef_d(poly_r, deg_inside, 1.0);
    SNAP_degree(poly_r) = deg_inside;
    for ( i=1,pt1=list_f+deg_inside-i; i<=deg_inside; i++,pt1-- ) { // by Newton's formula
      SNAPBASECOMPLEX_neg( *pt1, *(list_df+i+1));
      for ( j=1,pt2=list_f+deg_inside-j,pt3=list_df+i+1-j; j<i; j++,pt2--,pt3-- ) {
        SNAPBASECOMPLEX_mul( const_tmp, *pt2, *pt3);
        SNAPBASECOMPLEX_sub( *pt1, *pt1, const_tmp);
      }
      SNAPBASECOMPLEX_div_ui( *pt1, *pt1, i);
#if defined(SNAP_SUPFR_T)
      SNAP_set_coef(poly_r, deg_inside-i, SNAPBASECOMPLEX_realref(*pt1));
#elif defined(SNAP_SUPFC_T)
      SNAP_set_coef(poly_r, deg_inside-i, *pt1);
#elif defined(SNAP_SUPD_T)
      SNAP_set_coef(poly_r, deg_inside-i, SNAPBASECOMPLEX_real(*pt1));
#elif defined(SNAP_SUPDC_T)
      SNAP_set_coef(poly_r, deg_inside-i, *pt1);
#endif
    }
  }
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for(i=0,pt1=list_unity; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_clear(*pt1);
  }
  for(i=0,pt1=list_work; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_clear(*pt1);
  }
  free(list_unity);
  free(list_work);
#endif
  for(i=0,pt1=list_f; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_clear(*pt1);
  }
  for(i=0,pt1=list_df; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_clear(*pt1);
  }
  free(list_df);
  free(list_f);
  SNAPBASECOMPLEX_clear(const_pin);
  SNAPBASECOMPLEX_clear(const_tmp);
  return res;
}

// split
// Graeffe's Squaring's k range
// #define SNAPLIB_MIN_K 2
#define SNAPLIB_MIN_K 2
// #define SNAPLIB_MAX_K 3
#define SNAPLIB_MAX_K 4
// Minimum of log_2(n)
#define SNAPLIB_MIN_NL 2
// Sampling Points Range 2^(2*log_2(n)+minNP), 2^(2*log_2(n)+maxNP)
// #define SNAPLIB_MIN_NP -1
#define SNAPLIB_MIN_NP -1
// #define SNAPLIB_MAX_NP 2
#define SNAPLIB_MAX_NP 4
// Newton
#define SNAPLIB_MAX_NEWTONITR 2
// for supfr_t, we use the default prec.
// #define SNAPLIB_MIN_NEWTONTOL 1.0e-16
// Lifting
// for supfr_t, we use the half of the default prec.
// #define SNAPLIB_LIFTING_TOL 1.0e-8
int SNAP_agcd_exqrgcdwc_split(SNAP_t poly_inside, SNAP_t poly, const SNAPBASEREAL_t tol) {
  unsigned int i,k,oldk,nl,nitr,j,deg_in;
  signed int si;
  SNAP_t *poly_p;
  SNAP_t poly_outside,poly_u,poly_v,poly_d,poly_res,poly_t;
  SNAPBASEREAL_t norm_in, norm_out, tol_res, newtol_res, newton_mintol, lifting_tol;
  SNAPBASEREAL_init(norm_in);
  SNAPBASEREAL_init(norm_out);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(newtol_res);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_init_set_ui(newton_mintol, ((unsigned int) SNAPBASEREAL_get_default_prec()));
  SNAPBASEREAL_neg(newton_mintol, newton_mintol);
  SNAPBASEREAL_exp2(newton_mintol, newton_mintol);
  SNAPBASEREAL_init_set_ui(lifting_tol, ((unsigned int) SNAPBASEREAL_get_default_prec())/2);
  SNAPBASEREAL_neg(lifting_tol, lifting_tol);
  SNAPBASEREAL_exp2(lifting_tol, lifting_tol);
#else
  newton_mintol = 1.0e-16;
  lifting_tol = 1.0e-8;
#endif
  // OUTER LOOP of K: order of Graeffe's Squaring k=MIN_K,..,MAX_K
  oldk=0;
  SNAP_init(poly_outside);
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAP_init(poly_d);
  SNAP_init(poly_res);
  SNAP_init(poly_t);
  poly_p = (SNAP_t*) malloc(sizeof(SNAP_t)*(SNAPLIB_MAX_K+1));
  SNAP_init_set(*poly_p, poly);
  SNAP_2norm(norm_in, *poly_p);
  SNAP_div_basereal(*poly_p, *poly_p, norm_in);
  for ( k=SNAPLIB_MIN_K; k<=SNAPLIB_MAX_K; k++ ) {
#ifdef SNAPLIB_DEBUG
    printf("Number of squarings: %d\n", (unsigned int) k);
#endif
    // Graeffe's Squaring
    for( i=oldk+1; i<=k; i++ ) {
      SNAP_init(*(poly_p+i));
      oldk=i;
      SNAP_agcd_exqrgcdwc_rootsquaring(*(poly_p+i), *(poly_p+i-1));
      SNAP_2norm(norm_in, *(poly_p+i));
      SNAP_div_basereal(*(poly_p+i), *(poly_p+i), norm_in);
    }
    // Rough Splitting
    nl = SNAPLIB_MAX(SNAPLIB_MIN_NL, log2((double) SNAP_degree(*(poly_p+k))));
    for( si=SNAPLIB_MIN_NP; si<=SNAPLIB_MAX_NP; si++ ) {
#ifdef SNAPLIB_DEBUG
      printf("Number of divisions: %d\n", (unsigned int) pow(2,2*nl+si));
#endif
      if ( SNAP_agcd_exqrgcdwc_roughsplit(poly_inside, *(poly_p+k), (unsigned int) pow(2,2*nl+si)) != 0 ) {
        continue;
      }
      if ( SNAP_degree(poly_inside) == SNAP_degree(poly) ) {
        // if ( si!=SNAPLIB_MAX_NP && k!=SNAPLIB_MAX_K ) { continue; }
#ifdef SNAPLIB_DEBUG
        printf("All the roots are inside!\n");
#endif
        for( i=0; i<=oldk; i++ ) {
          SNAP_clear(*(poly_p+i));
        }
        free(poly_p);
        SNAP_clear(poly_outside);
        SNAP_clear(poly_u);
        SNAP_clear(poly_v);
        SNAP_clear(poly_d);
        SNAP_clear(poly_res);
        SNAP_clear(poly_t);
        SNAPBASEREAL_clear(norm_in);
        SNAPBASEREAL_clear(norm_out);
        SNAPBASEREAL_clear(tol_res);
        SNAPBASEREAL_clear(newtol_res);
        SNAPBASEREAL_clear(newton_mintol);
        SNAPBASEREAL_clear(lifting_tol);
        SNAP_set(poly_inside, poly);
        return 0;
      }
      SNAP_div_qr(poly_outside, poly_res, *(poly_p+k), poly_inside);
      SNAP_2norm(norm_in, poly_inside);
      SNAP_2norm(norm_out, poly_outside);
      SNAP_div_basereal(poly_inside, poly_inside, norm_in);
      SNAP_div_basereal(poly_outside, poly_outside, norm_out);
      SNAPBASEREAL_mul( norm_in, norm_in, norm_out);
      SNAP_div_basereal(*(poly_p+k), *(poly_p+k), norm_in);
      SNAP_div_basereal(poly_res, poly_res, norm_in);
      SNAP_2norm(norm_in, poly_res);
      SNAP_2norm(norm_out, *(poly_p+k));
      nitr = 0;
      SNAPBASEREAL_div(tol_res, norm_in, norm_out);
      // Newton Refinement
      while ( SNAPBASEREAL_cmp(tol_res, newton_mintol) > 0 && nitr++ < SNAPLIB_MAX_NEWTONITR ) {
        if ( SNAP_agcd_exqrgcdwc_qrstep(poly_inside, poly_outside, poly_inside, poly_outside, poly_d, poly_u, poly_v, newton_mintol, newton_mintol, tol_res, BEZOUTONLY, SPLITDISALLOWED) == UINT_MAX ) {
#ifdef SNAPLIB_DEBUG
          printf("Newton Refinement failed since bezout coefs were not computed.\n");
#endif
          break;
        }
        // rem((pk-In*Out)v, In)
        SNAP_mul(poly_t, poly_res, poly_v);
        SNAP_div_qr(poly_v , poly_u, poly_t, poly_inside);
        // div(pk-In*Out-Din*Out, In)
        SNAP_mul(poly_t, poly_u, poly_outside);
        SNAP_sub(poly_res, poly_res, poly_t);
        SNAP_lsdiv_q(poly_v, poly_res, poly_inside);
        // Update
        SNAP_add(poly_inside, poly_inside, poly_u);
        SNAP_add(poly_outside, poly_outside, poly_v);
        // Check
        SNAP_mul(poly_res, poly_inside, poly_outside);
        SNAP_sub(poly_res, *(poly_p+k), poly_res);
        // Normalization
        SNAP_2norm(norm_in, poly_inside);
        SNAP_2norm(norm_out, poly_outside);
        SNAP_div_basereal(poly_inside, poly_inside, norm_in);
        SNAP_div_basereal(poly_outside, poly_outside, norm_out);
        SNAPBASEREAL_mul(norm_in, norm_in, norm_out);
        SNAP_div_basereal(*(poly_p+k), *(poly_p+k), norm_in);
        SNAP_div_basereal(poly_res, poly_res, norm_in);
        SNAP_2norm(norm_in, poly_res);
        SNAP_2norm(norm_out, *(poly_p+k));
        SNAPBASEREAL_div(newtol_res, norm_in, norm_out);
#ifdef SNAPLIB_DEBUG
// mpfr_printf("residual(%d):=%Re < %Re (tol)\n", nitr, newtol_res, tol);
#endif
        if ( !(SNAPBASEREAL_cmp(newtol_res, tol_res) < 0) ) {
          SNAPBASEREAL_set(tol_res, newtol_res);
          break;
        } else {
          SNAPBASEREAL_set(tol_res, newtol_res);
        }
      }
      if ( SNAPBASEREAL_cmp(tol_res, tol) > 0 ) {
#ifdef SNAPLIB_DEBUG
// printf("Newton Method didn't converge!\n");
#endif
        continue;
      }
      // Lifting
      for( i=k; i>0; i-- ) {
        // inside <= gcd(*(poly_p+i-1), inside(x^2));
        deg_in = SNAP_degree(poly_inside);
        SNAP_set_coef_d(poly_inside, deg_in*2, 1.0);
        for( j = deg_in; j > 0; j-- ) {
          SNAPBASE_set(SNAP_coef(poly_inside, j*2), SNAP_coef(poly_inside, j));
          SNAPBASE_set_ui(SNAP_coef(poly_inside, j*2-1), 0);
        }
        SNAP_2norm(norm_in, *(poly_p+i-1));
        // SNAP_div_N(*(poly_p+i-1), *(poly_p+i-1), norm_in);
        SNAP_div_basereal(poly_res, *(poly_p+i-1), norm_in);
        SNAP_2norm(norm_in, poly_inside);
        SNAP_div_basereal(poly_outside, poly_inside, norm_in);
        SNAP_degree(poly_inside) = 0;
        SNAPBASE_set_ui(SNAP_coef(poly_inside,0), 1);
        if ( SNAP_agcd_exqrgcdwc_qrstep(poly_res, poly_outside, poly_res, poly_outside, poly_inside, poly_u, poly_v, lifting_tol, lifting_tol, tol_res, NORMALONLY, SPLITDISALLOWED) == UINT_MAX || SNAP_degree(poly_inside) == 0 ) {
          break;
        }
        // We don't need [outside <= polydiv(*(poly_p+i-1), inside);]
      }
      if ( i == 0 ) {
        break;
      }
    }
    if ( si<=SNAPLIB_MAX_NP ) {
      break;
    }
  }
  for( i=0; i<=oldk; i++ ) {
    SNAP_clear(*(poly_p+i));
  }
  free(poly_p);
  SNAP_clear(poly_outside);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAP_clear(poly_d);
  SNAP_clear(poly_res);
  SNAP_clear(poly_t);
  SNAPBASEREAL_clear(norm_in);
  SNAPBASEREAL_clear(norm_out);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(newtol_res);
  SNAPBASEREAL_clear(newton_mintol);
  SNAPBASEREAL_clear(lifting_tol);
  if ( k > SNAPLIB_MAX_K ) {
    return -1;
  } else {
    return 0;
  }
}

// sorting with index list by ascending order.
static void SNAP_agcd_exqrgcdwc_quicksort_uint_by_base_nc(unsigned int *idx, SNAPBASEREAL_t *val, unsigned int len, SNAPBASEREAL_t tval) {
  unsigned int i, j, pidx;
  if ( len == 2 ) {
    if ( SNAPBASEREAL_cmp(*val, *(val+1)) > 0 ) {
      // pval = *val; *val = *(val+1); *(val+1) = pval;
      SNAPBASEREAL_swap(*val, *(val+1));
      pidx = *idx;
      *idx = *(idx+1);
      *(idx+1) = pidx;
    }
  } else if ( len > 2 ) {
    i=0;
    j=len-1;
    pidx = i+(j-i)/2;
    if ( SNAPBASEREAL_cmp(*(val+i),*(val+pidx)) < 0 ) {
      if ( SNAPBASEREAL_cmp(*(val+pidx),*(val+j)) < 0 ) {
        SNAPBASEREAL_set(tval, *(val+pidx));
      } else if ( SNAPBASEREAL_cmp(*(val+j),*(val+i)) < 0 ) {
        SNAPBASEREAL_set(tval, *(val+i));
      } else {
        SNAPBASEREAL_set(tval,*(val+j));
      }
    } else {
      if ( SNAPBASEREAL_cmp(*(val+pidx),*(val+j)) > 0 ) {
        SNAPBASEREAL_set(tval,*(val+pidx));
      } else if ( SNAPBASEREAL_cmp(*(val+j),*(val+i)) > 0 ) {
        SNAPBASEREAL_set(tval,*(val+i));
      } else {
        SNAPBASEREAL_set(tval,*(val+j));
      }
    }
    while(1) {
      for(; i<len; i++) {
        if ( SNAPBASEREAL_cmp(tval, *(val+i)) <= 0 ) {
          break;
        }
      }
      // for(j=len-1;j>=0;j--) // NOTE j can not be < 0.
      do {
        if ( SNAPBASEREAL_cmp(tval, *(val+j)) >= 0 ) {
          break;
        }
      } while ( j-- != 0 );
      if ( i < j ) {
        // pval = *(val+i); *(val+i) = *(val+j); *(val+j) = pval;
        SNAPBASEREAL_swap(*(val+i), *(val+j));
        pidx = *(idx+i);
        *(idx+i) = *(idx+j);
        *(idx+j) = pidx;
        i++;
        j--;
      } else {
        break;
      }
    }
    //for (i=0;i<len;i++) {
    //	if ( SNAPBASE_cmp(tval, *(val+i)) == 0 ) { break; }
    //}
    //if ( i!=0 ) { SNAP_agcd_exqrgcdwc_quicksort_uint_by_N_nc(idx, val, i, tval); }
    SNAP_agcd_exqrgcdwc_quicksort_uint_by_base_nc(idx, val, i, tval);
    //if ( i!=len-1 ) { SNAP_agcd_exqrgcdwc_quicksort_uint_by_N_nc(idx+i+1, val+i+1,len-i-1, tval); }
    SNAP_agcd_exqrgcdwc_quicksort_uint_by_base_nc(idx+j+1, val+j+1,len-j-1, tval);
  }
}
void SNAP_agcd_exqrgcdwc_quicksort_uint_by_base(unsigned int *idx, SNAPBASEREAL_t *val, unsigned int len) {
  SNAPBASEREAL_t tval;
  SNAPBASEREAL_init(tval);
  SNAP_agcd_exqrgcdwc_quicksort_uint_by_base_nc(idx, val, len, tval);
  SNAPBASEREAL_clear(tval);
}

#define SNAPLIB_USE_QUICKSORT
// #undef SNAPLIB_USE_QUICKSORT
// for supfr_t, we use (unsigned int) 1.
//#define SNAPLIB_SPLIT_GAP 1.0
// Return the degree of newly found agcd
static unsigned int SNAP_agcd_exqrgcdwc_qrstep(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_cof, SNAP_t poly_cog, SNAP_t poly_d, SNAP_t poly_u, SNAP_t poly_v, const SNAPBASEREAL_t tol, const SNAPBASEREAL_t tol_temp, SNAPBASEREAL_t tol_res, SNAP_agcd_exqrgcdwc_qrstep_order qrstep_order, SNAP_agcd_exqrgcdwc_qrstep_order split_flag) {
  // We'd like to try to split then force to use PRS.
  SNAPMTX_t matrix_syl, matrix_Q;
  SNAP_t poly_r, poly_t, poly_ncf, poly_ncg, poly_prs1, poly_prs2;
  SNAPBASECOMPLEX_t sqrtn;
  SNAPBASEREAL_t prev_Rnorm, kth_Rnorm, sqrtcols, tol_f, tol_g, const_tmp, const_case1, *erval_list;
  unsigned int i,j,k,found_degree=UINT_MAX,case1flag=0,ergapk,ergapcount;
  unsigned int *eridx_list, mincdeg;
#ifdef SNAPLIB_DEBUG
  unsigned int q1,q2,q3;
#endif // SNAPLIB_DEBUG
#ifdef SNAP_QR_USE_FFTW
  fftw_plan fftp;
  int fftdim[1];
#endif // SNAP_QR_USE_FFTW
#ifndef SNAPLIB_USE_QUICKSORT
  unsigned int *ermin_list;
#endif
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
    printf("entering qrstep\n");
  }
#endif
  if ( SNAP_degree(poly_cof) == 0 || SNAP_degree(poly_cog) == 0 ) {
    return 0;
  }
  // Initialization and Sylvester matrix
  SNAPMTX_init2(matrix_syl, SNAP_degree(poly_cof)+SNAP_degree(poly_cog), SNAP_degree(poly_cof)+SNAP_degree(poly_cog));
  SNAPMTX_init2(matrix_Q, SNAP_degree(poly_cof)+SNAP_degree(poly_cog), SNAP_degree(poly_cof)+SNAP_degree(poly_cog));
  // QR decomposition
#ifdef SNAP_SUPD_T
  if ( 1 ) {
    SNAPMTX_clear(matrix_syl);
    SNAPMTX_clear(matrix_Q);
    printf("ERROR in agcd_exqrgcdwc: SNAP_SUPD_T is not supported.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPDC_T)
  SNAP_alg_qr_dgv2010refine_sylvester(poly_cof, poly_cog, matrix_Q, matrix_syl);
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+f%+fI),", creal(SNAPMTX_elem(matrix_syl,q1,q2)), cimag(SNAPMTX_elem(matrix_syl,q1,q2)));
    }
    printf("\n");
  }
#endif
#elif defined(SNAP_SUPFR_T)
  if ( 1 ) {
    SNAPMTX_clear(matrix_syl);
    SNAPMTX_clear(matrix_Q);
    printf("ERROR in agcd_exqrgcdwc: SNAP_SUPFR_T is not supported.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPFC_T)
  if ( 1 ) {
    SNAPMTX_clear(matrix_syl);
    SNAPMTX_clear(matrix_Q);
    printf("ERROR in agcd_exqrgcdwc: SNAP_SUPFC_T is not supported.\n");
    return found_degree;
  }
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (unsigned int q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN), mpfr_get_ld(mpc_imagref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN));
    }
    printf("\n");
  }
#endif
#endif
  // Now we seek the relative smallest gap to recover a candidate approximate gcd.
  SNAPBASEREAL_init(sqrtcols);
  SNAPBASEREAL_sqrt_ui(sqrtcols, SNAPMTX_cols(matrix_syl));
  SNAPBASEREAL_init_set_ui(prev_Rnorm, 0);
  SNAPBASEREAL_init(kth_Rnorm);
  SNAPBASE_abs(kth_Rnorm, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  k = 0;
  SNAPBASEREAL_init(const_tmp);
  SNAPBASEREAL_init(const_case1);
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    mpfr_printf("%d-th Rnorm: %Re\n", k, kth_Rnorm);
#else
    printf("%d-th Rnorm: %e\n", k, kth_Rnorm);
#endif
  }
#endif
  SNAPBASEREAL_mul(const_case1, tol_temp, sqrtcols);
  if ( SNAPBASEREAL_cmp(kth_Rnorm, const_case1) > 0 ) { // CASE 1 (there is no chance to have agcd)
    case1flag = 1;
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      mpfr_printf("CASE 1 detected! (%Re > %Re=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#else
      printf("CASE 1 detected! (%e > %e=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#endif
    }
#endif
    // for CASE 1, we do nothing here.
  } else if ( qrstep_order != BEZOUTONLY ) { // CASES 2 and 3 (there is a possibility to have agcd)
    mincdeg = SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog));
    erval_list = (SNAPBASEREAL_t*) malloc(sizeof(SNAPBASEREAL_t)*mincdeg);
    for ( i=0; i<SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog)); i++ ) {
      SNAPBASEREAL_init(*(erval_list+i));
    }
    eridx_list = (unsigned int*) malloc(sizeof(unsigned int)*mincdeg);
#ifndef SNAPLIB_USE_QUICKSORT
    ermin_list = (unsigned int*) malloc(sizeof(unsigned int)*3);
#endif
    while ( ++k <= mincdeg ) {
      // F-norm can be computed by ...
      SNAPBASEREAL_mul(kth_Rnorm, kth_Rnorm, kth_Rnorm);
      SNAPBASEREAL_mul(prev_Rnorm, prev_Rnorm, prev_Rnorm);
      SNAPBASEREAL_add(prev_Rnorm, prev_Rnorm, kth_Rnorm);
      SNAPBASEREAL_sqrt(prev_Rnorm, prev_Rnorm);
      if ( SNAPBASEREAL_cmp(prev_Rnorm, const_case1) > 0 ) {
        break;  // seeking finished.
      }
      // Update kth_Rnorm
      SNAPMTX_BLAS_nrm2(kth_Rnorm, k+1, SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-k,SNAPMTX_rows(matrix_syl)-1-k),1);
      *(eridx_list+k-1) = k;
      SNAPBASEREAL_div(*(erval_list+k-1), prev_Rnorm, kth_Rnorm);
      if ( SNAPBASEREAL_sgn(prev_Rnorm) == 0 || SNAPBASEREAL_nan_p(*(erval_list+k-1)) ) {
        SNAPBASEREAL_set_ui(*(erval_list+k-1),1);
      }
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        mpfr_printf("%d-th Rnorm: %Re, e_r: %Re (Case1 threshold: %Re)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#else
        printf("%d-th Rnorm: %e, e_r: %e (Case1 threshold: %e)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#endif
      }
#endif
    }
    // sort erval_list and corresponding eridx_list by ascending order.
#ifdef SNAPLIB_USE_QUICKSORT
    SNAP_agcd_exqrgcdwc_quicksort_uint_by_base(eridx_list, erval_list, k-1);
#else
    *(ermin_list) = 0;
    *(ermin_list+1) = 1;
    *(ermin_list+2) = 2;
    for ( i=1; i<SNAPLIB_MIN(3,k-1); i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list)), *(erval_list+*(ermin_list+i))) > 0 ) {
        j = *(ermin_list);
        *(ermin_list) = *(ermin_list+i);
        *(ermin_list+i) = j;
      }
    }
    for ( i=2; i<SNAPLIB_MIN(3,k-1); i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+1)), *(erval_list+*(ermin_list+i))) > 0 ) {
        j = *(ermin_list+1);
        *(ermin_list+1) = *(ermin_list+i);
        *(ermin_list+i) = j;
      }
    }
    for ( i=3; i<k-1; i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+2)), *(erval_list+i)) > 0 ) {
        if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+1)), *(erval_list+i)) > 0 ) {
          if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list)), *(erval_list+i)) > 0 ) {
            *(ermin_list+2) = *(ermin_list+1);
            *(ermin_list+1) = *(ermin_list);
            *(ermin_list) = i;
          } else {
            *(ermin_list+2) = *(ermin_list+1);
            *(ermin_list+1) = i;
          }
        } else {
          *(ermin_list+2) = i;
        }
      }
    }
#endif
    /* somehow this is slower than the full sorting. it should be a sorting algorithm for finding 3 minimums.
    for ( i=1; i<SNAPLIB_MIN(3,k-1); i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list), *(erval_list+i)) > 0 ) {
    		SNAPBASE_swap(*(erval_list), *(erval_list+i));
    		j = *(eridx_list); *(eridx_list) = *(eridx_list+i); *(eridx_list+i) = j;
    	}
    }
    for ( i=2; i<SNAPLIB_MIN(3,k-1); i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list+1), *(erval_list+i)) > 0 ) {
    		SNAPBASE_swap(*(erval_list+1), *(erval_list+i));
    		j = *(eridx_list+1); *(eridx_list+1) = *(eridx_list+i); *(eridx_list+i) = j;
    	}
    }
    for ( i=3; i<k-1; i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list+2), *(erval_list+i)) > 0 ) {
    		if ( SNAPBASE_cmp(*(erval_list+1), *(erval_list+i)) > 0 ) {
    			if ( SNAPBASE_cmp(*(erval_list), *(erval_list+i)) > 0 ) {
    				SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    				SNAPBASE_swap(*(erval_list+1), *(erval_list+2));
    				SNAPBASE_swap(*(erval_list), *(erval_list+1));
    				j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2);
    				*(eridx_list+2) = *(eridx_list+1); *(eridx_list+1) = *(eridx_list); *(eridx_list) = j;
    			} else {
    				SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    				SNAPBASE_swap(*(erval_list+1), *(erval_list+2));
    				j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2);
    				*(eridx_list+2) = *(eridx_list+1); *(eridx_list+1) = j;
    			}
    		} else {
    			SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    			j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2); *(eridx_list+2) = j;
    		}
    	}
    }
    */
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
#ifdef SNAPLIB_USE_QUICKSORT
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %Re\n", *(eridx_list+i), *(erval_list+i));
      }
#else
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %Re\n", *(eridx_list+*(ermin_list+i)), *(erval_list+*(ermin_list+i)));
      }
#endif
#else
#ifdef SNAPLIB_USE_QUICKSORT
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %e\n", *(eridx_list+i), *(erval_list+i));
      }
#else
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %e\n", *(eridx_list+*(ermin_list+i)), *(erval_list+*(ermin_list+i)));
      }
#endif
#endif
    }
#endif
    // Divisibility Check
    SNAP_init(poly_r);
    SNAP_init(poly_t);
    SNAP_init(poly_ncf);
    SNAP_init(poly_ncg);
    SNAPBASEREAL_init(tol_f);
    SNAPBASEREAL_init(tol_g);
    // for ( i=0; i<k-1; i++ ) { // find a candidate gcd until the given tolerance is satisfied.
    for ( i=0; i<SNAPLIB_MIN(3,k-1); i++ ) { // Top 3 Check
      // if ( split_flag == SPLITALLOWED && sqrt(tol_temp) <= (*(erval_list+i)) ) { break; }
      // poly_r <= the last $(k+1)$-th row of R
#ifndef SNAPLIB_USE_QUICKSORT
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-*(eridx_list+*(ermin_list+i)), SNAPMTX_rows(matrix_syl)-1-*(eridx_list+*(ermin_list+i))), *(eridx_list+*(ermin_list+i))+1);
#else
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i), SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i)), *(eridx_list+i)+1);
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
      if ( SNAPBASE_sgn(SNAP_lcoef(poly_r)) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPDC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPFC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#endif
      // poly_t <= poly_d * poly_r
      SNAP_mul(poly_t, poly_d, poly_r);
      // poly_t <= poly_t/norm(poly_t) : normalization
      SNAP_2norm(const_tmp, poly_t);
      SNAP_div_basereal(poly_t, poly_t, const_tmp);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncf, poly_f, poly_t);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncg, poly_g, poly_t);
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
/*
      if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_UVGCD ) {
*/
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, &tol_f);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, tol_f);
#endif
/*
      } else if ( mode_opt == SNAP_EXQRGCD2_REFINEMENT_FASTGCD ) {
        SNAP_agcd_fastgcd_refine(poly_f, poly_g, poly_t, poly_ncf, poly_ncg, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ);
      }
*/
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      // check the tolerance
      SNAP_mul(poly_r, poly_ncf, poly_t);
      SNAP_sub(poly_r, poly_f, poly_r);
      SNAP_2norm(tol_f, poly_r);
      if ( SNAPBASEREAL_cmp(tol_f,tol) < 0 ) { // continue
        // check the tolerance
        SNAP_mul(poly_r, poly_ncg, poly_t);
        SNAP_sub(poly_r, poly_g, poly_r);
        SNAP_2norm(tol_g, poly_r);
        if ( SNAPBASEREAL_cmp(tol_g,tol) < 0 ) { // approximate gcd found!!
          // update several polynomials.
          SNAP_set(poly_cof, poly_ncf);
          SNAP_set(poly_cog, poly_ncg);
          SNAP_set(poly_d, poly_t);
          found_degree = *(eridx_list+i);
          if ( SNAPBASEREAL_cmp(tol_f, tol_g) < 0 ) {
            SNAPBASEREAL_set(tol_res, tol_g);
          } else {
            SNAPBASEREAL_set(tol_res, tol_f);
          }
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("Number of tries: %d (Degree: %d)\n", i+1, found_degree);
          }
#endif
          break;
        }
      }
    }
    SNAP_clear(poly_r);
    SNAP_clear(poly_t);
    SNAP_clear(poly_ncf);
    SNAP_clear(poly_ncg);
    SNAPBASEREAL_clear(tol_f);
    SNAPBASEREAL_clear(tol_g);
    // Check Difficult case or Not
    if ( found_degree == UINT_MAX ) {
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
        printf("Difficult CASE detected!\n");
      }
#endif
      if ( split_flag == SPLITALLOWED ) { // Call Split and qrstep recursively.
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is called...\n");
        }
#endif

        /* ========================== find gap beyond k ============================== */
        // compute erval from the k skipped.
        ergapk = k;
        ergapcount = 0;
        for ( i=k; i <= mincdeg; i++ ) {
          // Update kth_Rnorm
          SNAPMTX_BLAS_nrm2(kth_Rnorm, i+1,SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-i,SNAPMTX_rows(matrix_syl)-1-i),1);
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
            mpfr_printf("%d-th Rnorm: %Re (beyond k)\n", i, kth_Rnorm);
#else
            printf("%d-th Rnorm: %e (beyond k)\n", i, kth_Rnorm);
#endif
          }
#endif
          if ( SNAPBASEREAL_cmp_ui(kth_Rnorm, 1) >= 0 ) {
            ergapcount++;
            if ( ergapcount >= 2 ) {
              ergapk = i-1;
              break;
            }
          } else {
            ergapcount = 0;
          }
        }
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("Largest k=%d satisyfing small(1.0) gap beyond the case 1\n",ergapk);
        }
#endif
        /* ========================== find gap beyond k ============================== */

        SNAP_init(poly_ncf);
        SNAP_init(poly_ncg);
        SNAP_init(poly_prs1);
        SNAP_init(poly_prs2);
        for ( i=ergapk; i<=ergapk; i++ ) {
          SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-2-i), i+2);
          SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, SNAPMTX_rows(matrix_syl)-1-i), i+1);
          SNAP_2norm(kth_Rnorm, poly_prs1);
          SNAP_2norm(prev_Rnorm, poly_prs2);
          SNAP_div_basereal(poly_prs1, poly_prs1, kth_Rnorm);
          SNAP_div_basereal(poly_prs2, poly_prs2, prev_Rnorm);
#ifdef SNAPLIB_DEBUG
          printf("To be splitted (k=%d):\n", i);
#endif
          SNAPBASEREAL_sqrt_ui(const_tmp, SNAPLIB_MAX(SNAP_degree(poly_cof),SNAP_degree(poly_cog)));
          SNAPBASEREAL_mul_ui(const_tmp, const_tmp, 2);
          SNAPBASEREAL_mul(const_tmp, const_tmp, tol_temp);
          if ( SNAPBASEREAL_cmp(kth_Rnorm, prev_Rnorm) < 0 ) {
            SNAPBASEREAL_div(const_tmp, const_tmp, kth_Rnorm);
          } else {
            SNAPBASEREAL_div(const_tmp, const_tmp, prev_Rnorm);
          }
          if ( SNAP_agcd_exqrgcdwc_split(poly_ncf, poly_prs1, tol_temp) == 0 && SNAP_agcd_exqrgcdwc_split(poly_ncg, poly_prs2, tol_temp) == 0 ) {
            found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f, poly_g, poly_ncf, poly_ncg, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
            free(ermin_list);
#endif
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_ncf);
              SNAP_set(poly_cog, poly_ncg);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            SNAPMTX_clear(matrix_syl);
            SNAPMTX_clear(matrix_Q);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
            return found_degree;
          } else {
#ifdef SNAPLIB_DEBUG
            if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
              printf("SPLIT failed...\n");
            }
#endif
            found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f, poly_g, poly_prs1, poly_prs2, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
            free(ermin_list);
#endif
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_prs1);
              SNAP_set(poly_cog, poly_prs2);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            SNAPMTX_clear(matrix_syl);
            SNAPMTX_clear(matrix_Q);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
            return found_degree;
          }
        }
        SNAP_clear(poly_ncf);
        SNAP_clear(poly_ncg);
        SNAP_clear(poly_prs1);
        SNAP_clear(poly_prs2);
      } else {
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is not allowed. Skipped...\n");
        }
#endif
      }
    }
    for ( j=0; j<mincdeg; j++ ) {
      SNAPBASEREAL_clear(*(erval_list+j));
    }
    free(erval_list);
    free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
    free(ermin_list);
#endif
  }
  if ( found_degree == UINT_MAX ) {
    // poly_d <== 1.0 hence we do nothing for poly_d
    // SNAP_degree(poly_d) = 0; SNAP_set_coef(poly_d, 0, 1.0);
    if ( case1flag == 1 ) {
      found_degree = 0;
    }
    // poly_u and poly_v be the cofactors from the last row of Q^h.
    // NOTE: poly_u is the cofactor of poly_g and poly_v is of poly_f.
    // OLD if ( qrstep_order == NORMAL || qrstep_order == NORMALONLY || qrstep_order == BEZOUTONLY ) {
    if ( qrstep_order == BEZOUTONLY ) {
      SNAPBASECOMPLEX_init(sqrtn);
      SNAPBASECOMPLEX_sqrt_ui(sqrtn, SNAPMTX_rows(matrix_Q));
#ifdef SNAP_QR_USE_FFTW
      // FFTW
      fftdim[0] = SNAPMTX_rows(matrix_Q);
      fftp = fftw_plan_many_dft(1, fftdim, 1, SNAPMTX_elemp(matrix_Q, 0, SNAPMTX_cols(matrix_Q)-1), fftdim, SNAPMTX_csize(matrix_Q), 1, SNAPMTX_elemp(matrix_Q, 0, SNAPMTX_cols(matrix_Q)-1), fftdim, SNAPMTX_csize(matrix_Q), 1, FFTW_BACKWARD, FFTW_ESTIMATE);
      fftw_execute(fftp);
      fftw_destroy_plan(fftp);
      for (i=0; i<sdmdc_rows(matrix_Q); i++ ) {
        SNAPBASE_div_basereal(sdmdc_elem(matrix_Q,i,SNAPMTX_cols(matrix_Q)-1),sdmdc_elem(matrix_Q,i,SNAPMTX_cols(matrix_Q)-1), sqrtn);
        SNAPBASE_conj(sdmdc_elem(matrix_Q,i,SNAPMTX_cols(matrix_Q)-1),sdmdc_elem(matrix_Q,i,SNAPMTX_cols(matrix_Q)-1));
      }
#else
      not_supported;
#endif
#if defined(SNAP_SUPDC_T) && defined(SNAPLIB_DEBUG)
      printf("AAAA\n");
      for (q3=0; q3<SNAPMTX_rows(matrix_Q); q3++ ) {
        printf("(%+.5le%+.5leI),",creal(sdmdc_elem(matrix_Q,q3,SNAPMTX_cols(matrix_Q)-1)),cimag(sdmdc_elem(matrix_Q,q3,SNAPMTX_cols(matrix_Q)-1)));
      }
      printf("\n");
#endif
      for ( i=0; i<SNAP_degree(poly_cog); i++ ) {
        SNAP_set_coef(poly_u, i, sdmdc_elem(matrix_Q,SNAP_degree(poly_cog)-1-i,SNAPMTX_cols(matrix_Q)-1));
      }
      SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
      for ( i=0; i<SNAP_degree(poly_cof); i++ ) {
        SNAP_set_coef(poly_v, i, sdmdc_elem(matrix_Q,SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i,SNAPMTX_cols(matrix_Q)-1));
      }
      SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;

      SNAP_div_base(poly_u, poly_u, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      SNAP_div_base(poly_v, poly_v, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      SNAPBASECOMPLEX_clear(sqrtn);
    }
  }
  SNAPMTX_clear(matrix_syl);
  SNAPMTX_clear(matrix_Q);
  SNAPBASEREAL_clear(prev_Rnorm);
  SNAPBASEREAL_clear(kth_Rnorm);
  SNAPBASEREAL_clear(sqrtcols);
  SNAPBASEREAL_clear(const_tmp);
  SNAPBASEREAL_clear(const_case1);
  return found_degree;
}
// mode 0: automatic first order
// mode 1: normal side
// mode 2: reversal side
int SNAP_agcd_exqrgcdwc(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, int mode) {
  SNAPBASEREAL_t norm_f, norm_g, tol_temp, tol_res, norm_cf, norm_cg, lcf, lcg, scf, scg;
  SNAP_t poly_f, poly_g, poly_cof, poly_cog, poly_u, poly_v;
  SNAP_agcd_exqrgcdwc_qrstep_order qrstep_order, qrstep_firstorder;
  unsigned int found_degree, trials, coprimeresults;
  int result = -1;
  // initialization
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(tol_temp);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(norm_cf);
  SNAPBASEREAL_init(norm_cg);
  SNAPBASEREAL_init(lcf);
  SNAPBASEREAL_init(lcg);
  SNAPBASEREAL_init(scf);
  SNAPBASEREAL_init(scg);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
  SNAP_init_set(poly_cog, poly_g);
  SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
  SNAP_degree(poly_d) = 0;
  SNAPBASEREAL_set(tol_temp, tol);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  mpfr_printf("input tolerance: %Re\n", tol);
#else
  printf("input tolerance: %e\n", tol);
#endif
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_cof)) || SNAPBASE_zeroq(SNAP_lcoef(poly_cog)) ) {
    SNAP_clear(poly_f);
    SNAP_clear(poly_g);
    SNAP_clear(poly_cof);
    SNAP_clear(poly_cog);
    SNAPBASEREAL_clear(norm_f);
    SNAPBASEREAL_clear(norm_g);
    SNAPBASEREAL_clear(tol_temp);
    SNAPBASEREAL_clear(tol_res);
    SNAPBASEREAL_clear(norm_cf);
    SNAPBASEREAL_clear(norm_cg);
    SNAPBASEREAL_clear(lcf);
    SNAPBASEREAL_clear(lcg);
    SNAPBASEREAL_clear(scf);
    SNAPBASEREAL_clear(scg);
    fprintf(stderr, "EXQRGCD: polynomials must be non-zero!\n");
    return -1;
  }
  // depending on normal first or reversal first ?
  SNAPBASE_abs(lcf, SNAP_lcoef(poly_f));
  SNAPBASE_abs(lcg, SNAP_lcoef(poly_g));
  SNAPBASE_abs(scf, SNAP_coef(poly_f,0));
  SNAPBASE_abs(scg, SNAP_coef(poly_g,0));
  if ( SNAPBASEREAL_cmp(lcf,lcg) > 0 ) {
    SNAPBASEREAL_set(lcf, lcg);
  }
  if ( SNAPBASEREAL_cmp(scf,scg) > 0 ) {
    SNAPBASEREAL_set(scf, scg);
  }
  if ( mode == 1 || ( mode == 0 && SNAPBASEREAL_cmp(lcf,scf) >= 0 ) ) {
    qrstep_firstorder = NORMAL;
  } else {
    qrstep_firstorder = REVERSAL;
  }
#ifdef SNAPLIB_DEBUG
  printf("======= Order: %s =======\n", qrstep_firstorder == NORMAL ? "Normal":"Reversal" );
#endif
  // some agcd will be found then return, otherwise change qrstep_order.
  SNAPBASEREAL_set_ui(norm_cf, 1);
  SNAPBASEREAL_set_ui(norm_cg, 1);
  trials = 0;
  coprimeresults = 0;
  qrstep_order = qrstep_firstorder;
  while (1) {
    if ( trials > 0 || qrstep_firstorder == REVERSAL ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      if ( trials == 1 ) {
        break;  // re-try from the other side.
      }
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  if ( result == -1 ) {
#ifdef SNAPLIB_DEBUG
    printf("======= Second Order: %s =======\n", qrstep_firstorder != NORMAL ? "Normal":"Reversal" );
#endif
    if ( qrstep_firstorder == qrstep_order ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
    }
    SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
    SNAP_init_set(poly_cog, poly_g);
    SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
    SNAP_degree(poly_d) = 0;
    SNAPBASEREAL_set(tol_temp, tol);
    trials = 0;
    coprimeresults = 0;
    qrstep_order = (qrstep_firstorder == NORMAL ? REVERSAL:NORMAL);
  }
  while ( result == -1 ) {
    if ( trials > 0 ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_exqrgcdwc_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAP_clear(poly_cof);
  SNAP_clear(poly_cog);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(tol_temp);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(norm_cf);
  SNAPBASEREAL_clear(norm_cg);
  SNAPBASEREAL_clear(lcf);
  SNAPBASEREAL_clear(lcg);
  SNAPBASEREAL_clear(scf);
  SNAPBASEREAL_clear(scg);
#ifdef SNAPLIB_DEBUG
  printf("Finally the number of trials: %d\n", trials);
#endif
  return result;
}
/* ================== END[agcd_over_reals] ================== */
