/*
 * Copyright (c) 2013-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <snap.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#define SNAPLIB_USE_BEZOUTONLY_SPLIT

#ifdef SNAP_SUPD_T
void supd_show(supd_t upoly) {
  int i;
  for(i=supd_degree(upoly); i>=0; i--) {
    if ( supd_coef(upoly,i) == 0.0 ) {
      continue;
    }
    switch(i) {
    case 0:
      printf("+(%+.15le)", supd_coef(upoly,i));
      break;
    case 1:
      printf("+(%+.15le)x", supd_coef(upoly,i));
      break;
    default:
      printf("+(%+.15le)x^%d", supd_coef(upoly,i), i);
    }
  }
  printf("\n");
}
#endif // SNAP_SUPD_T


/* ================== START [agcd_exqrgcd] ================== */
/* SEE ALSO: ExQRGCD algorithm introduced by K.Nagasaka and T.Masui. */
typedef enum {NORMAL, REVERSAL, SPLITALLOWED, SPLITDISALLOWED, BEZOUTONLY, NORMALONLY} SNAP_agcd_exqrgcd_qrstep_order;
// NORMAL: current order and first order in the main routine
// REVERSAL: current order and first order in the main routine
// SPLITALLOWED: split will be called in the case 3
// SPLITDISALLOWED: split won't be called even in the case 3
// BEZOUTONLY: only compute the bezout coefficients
// NORMALONLY: this effects only the debug message (e.g. non verbose mode)
// #define SNAPLIB_MPFR_RND MPFR_RNDN
// #define SNAPLIB_MPC_RND MPC_RNDNN
// ========== THE FOLLOWING MACROS SHOULD BE MODIFIED TO BE COMPATIBLE WITH MPFR AND MPC. ==========
#define SNAPLIB_MIN(x,y) ((x)<(y) ? (x):(y))
#define SNAPLIB_MAX(x,y) ((x)>(y) ? (x):(y))
extern void SNAP_agcd_exqrgcd_rootsquaring(SNAP_t poly_r, SNAP_t poly_f);
extern int SNAP_agcd_exqrgcd_roughsplit(SNAP_t poly_r, SNAP_t poly_f, const unsigned int n);
extern void SNAP_agcd_exqrgcd_Qkthcolumn(SNAPBASE_t *vect, SNAPMTX_t matrix_gqr, SNAPBASE_t *tau, const unsigned int k);
extern int SNAP_agcd_exqrgcd_split(SNAP_t poly_inside, SNAP_t poly, const SNAPBASEREAL_t tol);
extern void SNAP_agcd_exqrgcd_quicksort_uint_by_base(unsigned int *idx, SNAPBASEREAL_t *val, unsigned int len);
static unsigned int SNAP_agcd_exqrgcdwp_qrstep(SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, const SNAPBASEREAL_t, const SNAPBASEREAL_t, SNAPBASEREAL_t, SNAP_agcd_exqrgcd_qrstep_order, SNAP_agcd_exqrgcd_qrstep_order);

// this assumes that poly_cof and poly_cog are co-prime and return their bezout coefs.
static unsigned int SNAP_agcd_exqrgcdwp_qrstep_bezoutonly(SNAP_t poly_cof, SNAP_t poly_cog, SNAP_t poly_u, SNAP_t poly_v) {
  SNAPMTX_t matrix_syl;
  SNAPBASE_t *tau, *Qkth;
  unsigned int i,found_degree=UINT_MAX;
  unsigned int pivot_cols;
  lapack_int *jpvt;
  if ( SNAP_degree(poly_cof) == 0 || SNAP_degree(poly_cog) == 0 ) {
    return 0;
  }
  // Initialization and Sylvester matrix
  SNAPMTX_init(matrix_syl);
  SNAPMTX_sylvester_matrix_transpose(matrix_syl, poly_cof, poly_cog);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  pivot_cols = SNAPMTX_cols(matrix_syl)-1;
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*pivot_cols);
  // QR decomposition
#ifdef SNAP_SUPD_T
  // In this case, we partially use LAPACKE_dgeqp3 for partial QR dcomposition with column pivoting.
  for (i=0; i<pivot_cols; i++) {
    *(jpvt+i) = 0;
  }
  if ( LAPACKE_dgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt , tau ) != 0 ||
       LAPACKE_dormqr(LAPACK_ROW_MAJOR, 'L', 'T', SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl)-pivot_cols, pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau, SNAPMTX_elemp(matrix_syl,0,pivot_cols), SNAPMTX_csize(matrix_syl)) != 0 ||
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
       clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#else
       LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#endif
     ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    free(jpvt);
    printf("ERROR in agcd_exqrgcdwp: dgeqrf or related functions could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPDC_T)
  // In this case, we partially use LAPACKE_zgeqp3 for partial QR dcomposition with column pivoting.
  for (i=0; i<pivot_cols; i++) {
    *(jpvt+i) = 0;
  }
  if ( LAPACKE_zgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt , tau ) != 0 ||
       LAPACKE_zunmqr(LAPACK_ROW_MAJOR, 'L', 'C', SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl)-pivot_cols, pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau, SNAPMTX_elemp(matrix_syl,0,pivot_cols), SNAPMTX_csize(matrix_syl)) != 0 ||
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
       clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#else
       LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#endif
     ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    free(jpvt);
    printf("ERROR in agcd_exqrgcdwp: zgeqrf or related functions could not compute the QR decomposition.\n");
    return found_degree;
  }
#endif
  found_degree = 0;
  // poly_u and poly_v be the cofactors from the last row of Q.
  // NOTE: poly_u is the cofactor of poly_g and poly_v is of poly_f.
  Qkth = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_init(*(Qkth+i));
  }
  SNAP_agcd_exqrgcd_Qkthcolumn(Qkth, matrix_syl, tau, SNAPMTX_rows(matrix_syl)-1);
  SNAP_set_rbase(poly_u, Qkth, SNAP_degree(poly_cog));
  SNAP_set_rbase(poly_v, Qkth+SNAP_degree(poly_cog), SNAP_degree(poly_cof));
  /*
  	for ( i=0; i<SNAP_degree(poly_cog); i++ ) {
  		SNAP_set_coef(poly_u, i, *(Qkth+SNAP_degree(poly_cog)-1-i));
  	}
  	SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
  	for ( i=0; i<SNAP_degree(poly_cof); i++ ) {
  		SNAP_set_coef(poly_v, i, *(Qkth+SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i));
  	}
  	SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;
  */
  /*
  	for ( i=0; i < pivot_cols; i++ ) {
  		if ( (unsigned int) *(jpvt+i) < SNAP_degree(poly_cog) ) {
  			SNAP_set_coef(poly_u, SNAP_degree(poly_cog)-1-((unsigned int) *(jpvt+i)), *(Qkth+i));
  		} else {
  			SNAP_set_coef(poly_v, SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-((unsigned int) *(jpvt+i)), *(Qkth+i));
  		}
  	}
  	for ( ; i < SNAPMTX_cols(matrix_syl); i++ ) {
  		if ( i < SNAP_degree(poly_cog) ) {
  			SNAP_set_coef(poly_u, SNAP_degree(poly_cog)-1-i, *(Qkth+i));
  		} else {
  			SNAP_set_coef(poly_v, SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i, *(Qkth+i));
  		}
  	}
  	// the followings are not required since SNAP_set_coef already changes them.
  	// SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
  	// SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;
  */
  SNAP_div_base(poly_u, poly_u, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  SNAP_div_base(poly_v, poly_v, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_clear(*(Qkth+i));
  }
  free(Qkth);
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_clear(*(tau+i));
  }
  free(tau);
  SNAPMTX_clear(matrix_syl);
  free(jpvt);
  return found_degree;
}

// split
// Graeffe's Squaring's k range
// #define SNAPLIB_MIN_K 2
#define SNAPLIB_MIN_K 2
// #define SNAPLIB_MAX_K 3
#define SNAPLIB_MAX_K 4
// Minimum of log_2(n)
#define SNAPLIB_MIN_NL 2
// Sampling Points Range 2^(2*log_2(n)+minNP), 2^(2*log_2(n)+maxNP)
// #define SNAPLIB_MIN_NP -1
#define SNAPLIB_MIN_NP -1
// #define SNAPLIB_MAX_NP 2
#define SNAPLIB_MAX_NP 4
// Newton
#define SNAPLIB_MAX_NEWTONITR 2
// for supfr_t, we use the default prec.
// #define SNAPLIB_MIN_NEWTONTOL 1.0e-16
// Lifting
// for supfr_t, we use the half of the default prec.
// #define SNAPLIB_LIFTING_TOL 1.0e-8
int SNAP_agcd_exqrgcdwp_split(SNAP_t poly_inside, SNAP_t poly, const SNAPBASEREAL_t tol) {
  unsigned int i,k,oldk,nl,nitr,j,deg_in;
  signed int si;
  SNAP_t *poly_p;
  SNAP_t poly_outside,poly_u,poly_v,poly_d,poly_res,poly_t;
  SNAPBASEREAL_t norm_in, norm_out, tol_res, newtol_res, newton_mintol, lifting_tol;
  SNAPBASEREAL_init(norm_in);
  SNAPBASEREAL_init(norm_out);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(newtol_res);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_init_set_ui(newton_mintol, ((unsigned int) SNAPBASEREAL_get_default_prec()));
  SNAPBASEREAL_neg(newton_mintol, newton_mintol);
  SNAPBASEREAL_exp2(newton_mintol, newton_mintol);
  SNAPBASEREAL_init_set_ui(lifting_tol, ((unsigned int) SNAPBASEREAL_get_default_prec())/2);
  SNAPBASEREAL_neg(lifting_tol, lifting_tol);
  SNAPBASEREAL_exp2(lifting_tol, lifting_tol);
#else
  newton_mintol = 1.0e-16;
  lifting_tol = 1.0e-8;
#endif
  // OUTER LOOP of K: order of Graeffe's Squaring k=MIN_K,..,MAX_K
  oldk=0;
  SNAP_init(poly_outside);
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAP_init(poly_d);
  SNAP_init(poly_res);
  SNAP_init(poly_t);
  poly_p = (SNAP_t*) malloc(sizeof(SNAP_t)*(SNAPLIB_MAX_K+1));
  SNAP_init_set(*poly_p, poly);
#ifdef SNAPLIB_DEBUG
  printf("Split's initialization done.\n");
  fflush(NULL);
#endif
  SNAP_2norm(norm_in, *poly_p);
  SNAP_div_basereal(*poly_p, *poly_p, norm_in);
  for ( k=SNAPLIB_MIN_K; k<=SNAPLIB_MAX_K; k++ ) {
#ifdef SNAPLIB_DEBUG
    printf("Number of squarings: %d\n", (unsigned int) k);
    fflush(NULL);
#endif
    // Graeffe's Squaring
    for( i=oldk+1; i<=k; i++ ) {
      SNAP_init(*(poly_p+i));
      oldk=i;
      SNAP_agcd_exqrgcd_rootsquaring(*(poly_p+i), *(poly_p+i-1));
      SNAP_2norm(norm_in, *(poly_p+i));
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( norm_in == 0.0 ) {
        break;
      }
#endif // SNAP_SUPD_T
      SNAP_div_basereal(*(poly_p+i), *(poly_p+i), norm_in);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( norm_in == 0.0 ) {
      continue;
    }
#endif // SNAP_SUPD_T
    // Rough Splitting
    nl = SNAPLIB_MAX(SNAPLIB_MIN_NL, log2((double) SNAP_degree(*(poly_p+k))));
    for( si=SNAPLIB_MIN_NP; si<=SNAPLIB_MAX_NP; si++ ) {
#ifdef SNAPLIB_DEBUG
      printf("Number of divisions: %d\n", (unsigned int) pow(2,2*nl+si));
      fflush(NULL);
#endif
      if ( SNAP_agcd_exqrgcd_roughsplit(poly_inside, *(poly_p+k), (unsigned int) pow(2,2*nl+si)) != 0 ) {
        continue;
      }
      if ( SNAP_degree(poly_inside) == SNAP_degree(poly) ) {
        // if ( si!=SNAPLIB_MAX_NP && k!=SNAPLIB_MAX_K ) { continue; }
#ifdef SNAPLIB_DEBUG
        printf("All the roots are inside!\n");
        fflush(NULL);
#endif
        for( i=0; i<=oldk; i++ ) {
          SNAP_clear(*(poly_p+i));
        }
        free(poly_p);
        SNAP_clear(poly_outside);
        SNAP_clear(poly_u);
        SNAP_clear(poly_v);
        SNAP_clear(poly_d);
        SNAP_clear(poly_res);
        SNAP_clear(poly_t);
        SNAPBASEREAL_clear(norm_in);
        SNAPBASEREAL_clear(norm_out);
        SNAPBASEREAL_clear(tol_res);
        SNAPBASEREAL_clear(newtol_res);
        SNAPBASEREAL_clear(newton_mintol);
        SNAPBASEREAL_clear(lifting_tol);
        SNAP_set(poly_inside, poly);
        return 0;
      }
#ifdef SNAPLIB_DEBUG
      printf("Newton Refinement started.\n");
      fflush(NULL);
#endif
      SNAP_div_qr(poly_outside, poly_res, *(poly_p+k), poly_inside);
      SNAP_2norm(norm_in, poly_inside);
      SNAP_2norm(norm_out, poly_outside);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( norm_out == 0.0 ) {
        continue;
      }
#endif // SNAP_SUPD_T
      SNAP_div_basereal(poly_inside, poly_inside, norm_in);
      SNAP_div_basereal(poly_outside, poly_outside, norm_out);
      SNAPBASEREAL_mul( norm_in, norm_in, norm_out);
      SNAP_div_basereal(*(poly_p+k), *(poly_p+k), norm_in);
      SNAP_div_basereal(poly_res, poly_res, norm_in);
      SNAP_2norm(norm_in, poly_res);
      SNAP_2norm(norm_out, *(poly_p+k));
      nitr = 0;
      SNAPBASEREAL_div(tol_res, norm_in, norm_out);
      // Newton Refinement
      while ( SNAPBASEREAL_cmp(tol_res, newton_mintol) > 0 && nitr++ < SNAPLIB_MAX_NEWTONITR ) {
#ifdef SNAPLIB_DEBUG
        printf("Newton Refinement (%d).\n", nitr);
        fflush(NULL);
#endif
        if ( SNAP_agcd_exqrgcdwp_qrstep_bezoutonly(poly_inside, poly_outside, poly_u, poly_v) == UINT_MAX ) {
#ifdef SNAPLIB_DEBUG
          printf("Newton Refinement failed since bezout coefs were not computed.\n");
          fflush(NULL);
#endif
          break;
        }
        // rem((pk-In*Out)v, In)
        SNAP_mul(poly_t, poly_res, poly_v);
        SNAP_div_qr(poly_v , poly_u, poly_t, poly_inside);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        SNAP_inorm(newtol_res, poly_u);
        if ( newtol_res > 1.0e+100 ) {
          continue;
        }
#endif // SNAP_SUPD_T
        // div(pk-In*Out-Din*Out, In)
        SNAP_mul(poly_t, poly_u, poly_outside);
        SNAP_sub(poly_res, poly_res, poly_t);
        SNAP_lsdiv_q(poly_v, poly_res, poly_inside);
        // Update
        SNAP_add(poly_inside, poly_inside, poly_u);
        SNAP_add(poly_outside, poly_outside, poly_v);
        // Check
        SNAP_mul(poly_res, poly_inside, poly_outside);
        SNAP_sub(poly_res, *(poly_p+k), poly_res);
        // Normalization
        SNAP_2norm(norm_in, poly_inside);
        SNAP_2norm(norm_out, poly_outside);
        SNAP_div_basereal(poly_inside, poly_inside, norm_in);
        SNAP_div_basereal(poly_outside, poly_outside, norm_out);
        SNAPBASEREAL_mul(norm_in, norm_in, norm_out);
        SNAP_div_basereal(*(poly_p+k), *(poly_p+k), norm_in);
        SNAP_div_basereal(poly_res, poly_res, norm_in);
        SNAP_2norm(norm_in, poly_res);
        SNAP_2norm(norm_out, *(poly_p+k));
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        if ( norm_out == 0.0 ) {
          continue;
        }
#endif // SNAP_SUPD_T
        SNAPBASEREAL_div(newtol_res, norm_in, norm_out);
#ifdef SNAPLIB_DEBUG
        printf("residual(%d):=%e < %e (tol), %e (prev.)\n", nitr, newtol_res, tol, tol_res);
        fflush(NULL);
#endif
        if ( !(SNAPBASEREAL_cmp(newtol_res, tol_res) < 0) ) {
          SNAPBASEREAL_set(tol_res, newtol_res);
          break;
        } else {
          SNAPBASEREAL_set(tol_res, newtol_res);
        }
      }
      if ( SNAPBASEREAL_cmp(tol_res, tol) > 0 ) {
#ifdef SNAPLIB_DEBUG
        printf("Newton Method didn't converge!\n");
        fflush(NULL);
#endif
        continue;
      }
      // Lifting
      for( i=k; i>0; i-- ) {
        // inside <= gcd(*(poly_p+i-1), inside(x^2));
        deg_in = SNAP_degree(poly_inside);
        SNAP_set_coef_d(poly_inside, deg_in*2, 1.0);
        for( j = deg_in; j > 0; j-- ) {
          SNAPBASE_set(SNAP_coef(poly_inside, j*2), SNAP_coef(poly_inside, j));
          SNAPBASE_set_ui(SNAP_coef(poly_inside, j*2-1), 0);
        }
        SNAP_2norm(norm_in, *(poly_p+i-1));
        // SNAP_div_N(*(poly_p+i-1), *(poly_p+i-1), norm_in);
        SNAP_div_basereal(poly_res, *(poly_p+i-1), norm_in);
        SNAP_2norm(norm_in, poly_inside);
        SNAP_div_basereal(poly_outside, poly_inside, norm_in);
        SNAP_degree(poly_inside) = 0;
        SNAPBASE_set_ui(SNAP_coef(poly_inside,0), 1);
        if ( SNAP_agcd_exqrgcdwp_qrstep(poly_res, poly_outside, poly_res, poly_outside, poly_inside, poly_u, poly_v, lifting_tol, lifting_tol, tol_res, NORMALONLY, SPLITDISALLOWED) == UINT_MAX || SNAP_degree(poly_inside) == 0 ) {
          break;
        }
        // We don't need [outside <= polydiv(*(poly_p+i-1), inside);]
      }
      if ( i == 0 ) {
        break;
      }
    }
    if ( si<=SNAPLIB_MAX_NP ) {
      break;
    }
  }
  for( i=0; i<=oldk; i++ ) {
    SNAP_clear(*(poly_p+i));
  }
  free(poly_p);
  SNAP_clear(poly_outside);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAP_clear(poly_d);
  SNAP_clear(poly_res);
  SNAP_clear(poly_t);
  SNAPBASEREAL_clear(norm_in);
  SNAPBASEREAL_clear(norm_out);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(newtol_res);
  SNAPBASEREAL_clear(newton_mintol);
  SNAPBASEREAL_clear(lifting_tol);
  if ( k > SNAPLIB_MAX_K ) {
    return -1;
  } else {
    return 0;
  }
}

#define SNAPLIB_USE_QUICKSORT
// #undef SNAPLIB_USE_QUICKSORT
// for supfr_t, we use (unsigned int) 1.
//#define SNAPLIB_SPLIT_GAP 1.0
// Return the degree of newly found agcd
static unsigned int SNAP_agcd_exqrgcdwp_qrstep(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_cof, SNAP_t poly_cog, SNAP_t poly_d, SNAP_t poly_u, SNAP_t poly_v, const SNAPBASEREAL_t tol, const SNAPBASEREAL_t tol_temp, SNAPBASEREAL_t tol_res, SNAP_agcd_exqrgcd_qrstep_order qrstep_order, SNAP_agcd_exqrgcd_qrstep_order split_flag) {
  // We'd like to try to split then force to use PRS.
  SNAPMTX_t matrix_syl;
  SNAP_t poly_r, poly_t, poly_ncf, poly_ncg, poly_prs1, poly_prs2;
  SNAPBASE_t *tau, *Qkth;
  SNAPBASEREAL_t prev_Rnorm, kth_Rnorm, sqrtcols, tol_f, tol_g, const_tmp, const_case1, *erval_list;
  unsigned int i,j,k,found_degree=UINT_MAX,case1flag=0,ergapk,ergapcount;
  unsigned int *eridx_list, mincdeg;
#if defined(SNAPLIB_DEBUG) && defined(SNAP_SUPDC_T)
  unsigned int q1,q2;
  int q3;
#endif // SNAPLIB_DEBUG
#ifndef SNAPLIB_USE_QUICKSORT
  unsigned int *ermin_list;
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  unsigned int pivot_cols;
  lapack_int *jpvt;
#endif
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
    printf("entering qrstep\n");
  }
#endif
  if ( SNAP_degree(poly_cof) == 0 || SNAP_degree(poly_cog) == 0 ) {
    return 0;
  }
  // Initialization and Sylvester matrix
  SNAPMTX_init(matrix_syl);
  SNAPMTX_sylvester_matrix_transpose(matrix_syl, poly_cof, poly_cog);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  // QR decomposition
#ifdef SNAP_SUPD_T
  // In this case, we partially use LAPACKE_dgeqp3 for partial QR dcomposition with column pivoting.
  pivot_cols = SNAPLIB_MAX(SNAP_degree(poly_cof), SNAP_degree(poly_cog))-1;
  if ( pivot_cols != 0 ) {
    pivot_cols--;
  }
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*pivot_cols);
  for (i=0; i<pivot_cols; i++) {
    *(jpvt+i) = 0;
  }
  if ( LAPACKE_dgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt , tau ) != 0 ||
       LAPACKE_dormqr(LAPACK_ROW_MAJOR, 'L', 'T', SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl)-pivot_cols, pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau, SNAPMTX_elemp(matrix_syl,0,pivot_cols), SNAPMTX_csize(matrix_syl)) != 0 ||
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
       clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#else
       LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#endif
     ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    free(jpvt);
    printf("ERROR in agcd_exqrgcdwp: dgeqrf or related functions could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPDC_T)
  // In this case, we partially use LAPACKE_zgeqp3 for partial QR dcomposition with column pivoting.
  pivot_cols = SNAPLIB_MAX(SNAP_degree(poly_cof), SNAP_degree(poly_cog))-1;
  if ( pivot_cols != 0 ) {
    pivot_cols--;
  }
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*pivot_cols);
  for (i=0; i<pivot_cols; i++) {
    *(jpvt+i) = 0;
  }
  if ( LAPACKE_zgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt , tau ) != 0 ||
       LAPACKE_zunmqr(LAPACK_ROW_MAJOR, 'L', 'C', SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl)-pivot_cols, pivot_cols, SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau, SNAPMTX_elemp(matrix_syl,0,pivot_cols), SNAPMTX_csize(matrix_syl)) != 0 ||
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
       clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#else
       LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl)-pivot_cols, SNAPMTX_cols(matrix_syl)-pivot_cols, SNAPMTX_elemp(matrix_syl,pivot_cols,pivot_cols), SNAPMTX_csize(matrix_syl), tau+pivot_cols) != 0
#endif
     ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    free(jpvt);
    printf("ERROR in agcd_qrgcdwp: zgeqrf or related functions could not compute the QR decomposition.\n");
    return found_degree;
  }
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+f%+fI),", creal(SNAPMTX_elem(matrix_syl,q1,q2)), cimag(SNAPMTX_elem(matrix_syl,q1,q2)));
    }
    printf("\n");
  }
  printf("tau: ");
  for (q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    printf("(%+f%+fI),", creal(*(tau+q1)), cimag(*(tau+q1)));
  }
  printf("\n");
#endif
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_exqrgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#ifdef SNAPLIB_DEBUG
  printf("after\n");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    for (unsigned int q2=0; q2<SNAPMTX_cols(matrix_syl); q2++ ) {
      printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN), mpfr_get_ld(mpc_imagref(SNAPMTX_elem(matrix_syl,q1,q2)), MPFR_RNDN));
    }
    printf("\n");
  }
  printf("tau: ");
  for (unsigned int q1=0; q1<SNAPMTX_rows(matrix_syl); q1++ ) {
    printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(*(tau+q1)), MPFR_RNDN), mpfr_get_ld(mpc_imagref(*(tau+q1)), MPFR_RNDN));
  }
  printf("\n");
#endif
#endif
  // Now we seek the relative smallest gap to recover a candidate approximate gcd.
  SNAPBASEREAL_init(sqrtcols);
  SNAPBASEREAL_sqrt_ui(sqrtcols, SNAPMTX_cols(matrix_syl));
  SNAPBASEREAL_init_set_ui(prev_Rnorm, 0);
  SNAPBASEREAL_init(kth_Rnorm);
  SNAPBASE_abs(kth_Rnorm, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  k = 0;
  SNAPBASEREAL_init(const_tmp);
  SNAPBASEREAL_init(const_case1);
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    mpfr_printf("%d-th Rnorm: %Re\n", k, kth_Rnorm);
#else
    printf("%d-th Rnorm: %e\n", k, kth_Rnorm);
#endif
  }
#endif
  SNAPBASEREAL_mul(const_case1, tol_temp, sqrtcols);
  if ( SNAPBASEREAL_cmp(kth_Rnorm, const_case1) > 0 ) { // CASE 1 (there is no chance to have agcd)
    case1flag = 1;
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      mpfr_printf("CASE 1 detected! (%Re > %Re=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#else
      printf("CASE 1 detected! (%e > %e=tol*sqrtcols)\n", kth_Rnorm, const_case1);
#endif
    }
#endif
    // for CASE 1, we do nothing here.
  } else if ( qrstep_order != BEZOUTONLY ) { // CASES 2 and 3 (there is a possibility to have agcd)
    mincdeg = SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog));
    erval_list = (SNAPBASEREAL_t*) malloc(sizeof(SNAPBASEREAL_t)*mincdeg);
    for ( i=0; i<SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog)); i++ ) {
      SNAPBASEREAL_init(*(erval_list+i));
    }
    eridx_list = (unsigned int*) malloc(sizeof(unsigned int)*mincdeg);
#ifndef SNAPLIB_USE_QUICKSORT
    ermin_list = (unsigned int*) malloc(sizeof(unsigned int)*3);
#endif
    while ( ++k <= mincdeg ) {
      // F-norm can be computed by ...
      SNAPBASEREAL_mul(kth_Rnorm, kth_Rnorm, kth_Rnorm);
      SNAPBASEREAL_mul(prev_Rnorm, prev_Rnorm, prev_Rnorm);
      SNAPBASEREAL_add(prev_Rnorm, prev_Rnorm, kth_Rnorm);
      SNAPBASEREAL_sqrt(prev_Rnorm, prev_Rnorm);
      if ( SNAPBASEREAL_cmp(prev_Rnorm, const_case1) > 0 ) {
        break;  // seeking finished.
      }
      // Update kth_Rnorm
      SNAPMTX_BLAS_nrm2(kth_Rnorm, k+1, SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-k,SNAPMTX_rows(matrix_syl)-1-k),1);
      *(eridx_list+k-1) = k;
      SNAPBASEREAL_div(*(erval_list+k-1), prev_Rnorm, kth_Rnorm);
      if ( SNAPBASEREAL_sgn(prev_Rnorm) == 0 || SNAPBASEREAL_nan_p(*(erval_list+k-1)) ) {
        SNAPBASEREAL_set_ui(*(erval_list+k-1),1);
      }
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        mpfr_printf("%d-th Rnorm: %Re, e_r: %Re (Case1 threshold: %Re)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#else
        printf("%d-th Rnorm: %e, e_r: %e (Case1 threshold: %e)\n", k, kth_Rnorm, *(erval_list+k-1), const_case1);
#endif
      }
#endif
    }
    // sort erval_list and corresponding eridx_list by ascending order.
#ifdef SNAPLIB_USE_QUICKSORT
    SNAP_agcd_exqrgcd_quicksort_uint_by_base(eridx_list, erval_list, k-1);
#else
    *(ermin_list) = 0;
    *(ermin_list+1) = 1;
    *(ermin_list+2) = 2;
    for ( i=1; i<SNAPLIB_MIN(3,k-1); i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list)), *(erval_list+*(ermin_list+i))) > 0 ) {
        j = *(ermin_list);
        *(ermin_list) = *(ermin_list+i);
        *(ermin_list+i) = j;
      }
    }
    for ( i=2; i<SNAPLIB_MIN(3,k-1); i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+1)), *(erval_list+*(ermin_list+i))) > 0 ) {
        j = *(ermin_list+1);
        *(ermin_list+1) = *(ermin_list+i);
        *(ermin_list+i) = j;
      }
    }
    for ( i=3; i<k-1; i++ ) {
      if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+2)), *(erval_list+i)) > 0 ) {
        if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list+1)), *(erval_list+i)) > 0 ) {
          if ( SNAPBASEREAL_cmp(*(erval_list+*(ermin_list)), *(erval_list+i)) > 0 ) {
            *(ermin_list+2) = *(ermin_list+1);
            *(ermin_list+1) = *(ermin_list);
            *(ermin_list) = i;
          } else {
            *(ermin_list+2) = *(ermin_list+1);
            *(ermin_list+1) = i;
          }
        } else {
          *(ermin_list+2) = i;
        }
      }
    }
#endif
    /* somehow this is slower than the full sorting. it should be a sorting algorithm for finding 3 minimums.
    for ( i=1; i<SNAPLIB_MIN(3,k-1); i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list), *(erval_list+i)) > 0 ) {
    		SNAPBASE_swap(*(erval_list), *(erval_list+i));
    		j = *(eridx_list); *(eridx_list) = *(eridx_list+i); *(eridx_list+i) = j;
    	}
    }
    for ( i=2; i<SNAPLIB_MIN(3,k-1); i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list+1), *(erval_list+i)) > 0 ) {
    		SNAPBASE_swap(*(erval_list+1), *(erval_list+i));
    		j = *(eridx_list+1); *(eridx_list+1) = *(eridx_list+i); *(eridx_list+i) = j;
    	}
    }
    for ( i=3; i<k-1; i++ ) {
    	if ( SNAPBASE_cmp(*(erval_list+2), *(erval_list+i)) > 0 ) {
    		if ( SNAPBASE_cmp(*(erval_list+1), *(erval_list+i)) > 0 ) {
    			if ( SNAPBASE_cmp(*(erval_list), *(erval_list+i)) > 0 ) {
    				SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    				SNAPBASE_swap(*(erval_list+1), *(erval_list+2));
    				SNAPBASE_swap(*(erval_list), *(erval_list+1));
    				j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2);
    				*(eridx_list+2) = *(eridx_list+1); *(eridx_list+1) = *(eridx_list); *(eridx_list) = j;
    			} else {
    				SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    				SNAPBASE_swap(*(erval_list+1), *(erval_list+2));
    				j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2);
    				*(eridx_list+2) = *(eridx_list+1); *(eridx_list+1) = j;
    			}
    		} else {
    			SNAPBASE_swap(*(erval_list+2), *(erval_list+i));
    			j = *(eridx_list+i); *(eridx_list+i) = *(eridx_list+2); *(eridx_list+2) = j;
    		}
    	}
    }
    */
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
#ifdef SNAPLIB_USE_QUICKSORT
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %Re\n", *(eridx_list+i), *(erval_list+i));
      }
#else
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %Re\n", *(eridx_list+*(ermin_list+i)), *(erval_list+*(ermin_list+i)));
      }
#endif
#else
#ifdef SNAPLIB_USE_QUICKSORT
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %e\n", *(eridx_list+i), *(erval_list+i));
      }
#else
      for(i=0; i<SNAPLIB_MIN(3,k-1); i++) {
        mpfr_printf("Sorted e_r: (k=%d) %e\n", *(eridx_list+*(ermin_list+i)), *(erval_list+*(ermin_list+i)));
      }
#endif
#endif
    }
#endif
    // Divisibility Check
    SNAP_init(poly_r);
    SNAP_init(poly_t);
    SNAP_init(poly_ncf);
    SNAP_init(poly_ncg);
    SNAPBASEREAL_init(tol_f);
    SNAPBASEREAL_init(tol_g);
    // for ( i=0; i<k-1; i++ ) { // find a candidate gcd until the given tolerance is satisfied.
    for ( i=0; i<SNAPLIB_MIN(3,k-1); i++ ) { // Top 3 Check
      // if ( split_flag == SPLITALLOWED && sqrt(tol_temp) <= (*(erval_list+i)) ) { break; }
      // poly_r <= the last $(k+1)$-th row of R
#ifndef SNAPLIB_USE_QUICKSORT
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-*(eridx_list+*(ermin_list+i)), SNAPMTX_rows(matrix_syl)-1-*(eridx_list+*(ermin_list+i))), *(eridx_list+*(ermin_list+i))+1);
#else
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i), SNAPMTX_rows(matrix_syl)-1-*(eridx_list+i)), *(eridx_list+i)+1);
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
      if ( SNAPBASE_sgn(SNAP_lcoef(poly_r)) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPDC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPFC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#endif
      // poly_t <= poly_d * poly_r
      SNAP_mul(poly_t, poly_d, poly_r);
      // poly_t <= poly_t/norm(poly_t) : normalization
      SNAP_2norm(const_tmp, poly_t);
      SNAP_div_basereal(poly_t, poly_t, const_tmp);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncf, poly_f, poly_t);
      // check the tolerance
      SNAP_mul(poly_r, poly_ncf, poly_t);
      SNAP_sub(poly_r, poly_f, poly_r);
      SNAP_2norm(tol_f, poly_r);
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        mpfr_printf("tol_f: %Re\n", tol_f);
#else
        printf("tol_f: %e\n", tol_f);
#endif
      }
#endif
      if ( SNAPBASEREAL_cmp(tol_f,tol) < 0 ) { // continue
        // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
        SNAP_lsdiv_q(poly_ncg, poly_g, poly_t);
        // check the tolerance
        SNAP_mul(poly_r, poly_ncg, poly_t);
        SNAP_sub(poly_r, poly_g, poly_r);
        SNAP_2norm(tol_g, poly_r);
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
          mpfr_printf("tol_g: %Re\n", tol_g);
#else
          printf("tol_g: %e\n", tol_g);
#endif
        }
#endif
        if ( SNAPBASEREAL_cmp(tol_g,tol) < 0 ) { // approximate gcd found!!
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
            mpfr_printf("Found tol_f: %Re, tol_g: %Re\n", tol_f, tol_g);
#else
            printf("Found tol_f: %e, tol_g: %e\n", tol_f, tol_g);
#endif
          }
#endif
          // update several polynomials.
          SNAP_set(poly_cof, poly_ncf);
          SNAP_set(poly_cog, poly_ncg);
          SNAP_set(poly_d, poly_t);
#ifndef SNAPLIB_USE_QUICKSORT
          found_degree = *(eridx_list+*(ermin_list+i));
#else
          found_degree = *(eridx_list+i);
#endif
          if ( SNAPBASEREAL_cmp(tol_f, tol_g) < 0 ) {
            SNAPBASEREAL_set(tol_res, tol_g);
          } else {
            SNAPBASEREAL_set(tol_res, tol_f);
          }
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("Number of tries: %d (Degree: %d)\n", i+1, found_degree);
          }
#endif
          break;
        }
      }
    }
    SNAP_clear(poly_r);
    SNAP_clear(poly_t);
    SNAP_clear(poly_ncf);
    SNAP_clear(poly_ncg);
    SNAPBASEREAL_clear(tol_f);
    SNAPBASEREAL_clear(tol_g);
    // Check Difficult case or Not
    if ( found_degree == UINT_MAX ) {
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
        printf("Difficult CASE detected!\n");
      }
#endif
      if ( split_flag == SPLITALLOWED ) { // Call Split and qrstep recursively.
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is called...\n");
        }
#endif

        /* ========================== find gap beyond k ============================== */
        // compute erval from the k skipped.
        ergapk = k;
        ergapcount = 0;
        for ( i=k; i <= mincdeg; i++ ) {
          // Update kth_Rnorm
          SNAPMTX_BLAS_nrm2(kth_Rnorm, i+1,SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-i,SNAPMTX_rows(matrix_syl)-1-i),1);
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
            mpfr_printf("%d-th Rnorm: %Re (beyond k)\n", i, kth_Rnorm);
#else
            printf("%d-th Rnorm: %e (beyond k)\n", i, kth_Rnorm);
#endif
          }
#endif
          if ( SNAPBASEREAL_cmp_ui(kth_Rnorm, 1) >= 0 ) {
            ergapcount++;
            if ( ergapcount >= 2 ) {
              ergapk = i-1;
              break;
            }
          } else {
            ergapcount = 0;
          }
        }
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("Largest k=%d satisyfing small(1.0) gap beyond the case 1\n",ergapk);
        }
#endif
        /* ========================== find gap beyond k ============================== */

        SNAP_init(poly_ncf);
        SNAP_init(poly_ncg);
        SNAP_init(poly_prs1);
        SNAP_init(poly_prs2);
        for ( i=ergapk; i<=ergapk; i++ ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
          if ( i+2 > SNAPMTX_rows(matrix_syl) - pivot_cols ) {
            // safe for k+2 <= SNAPMTX_rows(matrix_syl) - pivot_cols
            SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, pivot_cols), SNAPMTX_rows(matrix_syl) - pivot_cols);
            SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, pivot_cols+1), SNAPMTX_rows(matrix_syl) - pivot_cols -1);
            // reverse of pivoting for k+2 > SNAPMTX_rows(matrix_syl) - pivot_cols
            for ( j = SNAPMTX_rows(matrix_syl) - pivot_cols +1; j < i+2; j++ ) {
              SNAP_set_coef(poly_prs1, SNAPMTX_rows(matrix_syl)-((unsigned int) *(jpvt+SNAPMTX_rows(matrix_syl)-j)-2), SNAPMTX_elem(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-j));
              SNAP_set_coef(poly_prs2, SNAPMTX_rows(matrix_syl)-((unsigned int) *(jpvt+SNAPMTX_rows(matrix_syl)-j)-2), SNAPMTX_elem(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, SNAPMTX_rows(matrix_syl)-j));
            }
            SNAP_set_coef(poly_prs1, SNAPMTX_rows(matrix_syl)-((unsigned int) *(jpvt+SNAPMTX_rows(matrix_syl)-j)-2), SNAPMTX_elem(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-j));
          } else {
            SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-2-i), i+2);
            SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, SNAPMTX_rows(matrix_syl)-1-i), i+1);
          }
#else
          SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-i, SNAPMTX_rows(matrix_syl)-2-i), i+2);
          SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-i, SNAPMTX_rows(matrix_syl)-1-i), i+1);
#endif
          SNAP_2norm(kth_Rnorm, poly_prs1);
          SNAP_2norm(prev_Rnorm, poly_prs2);
          SNAP_div_basereal(poly_prs1, poly_prs1, kth_Rnorm);
          SNAP_div_basereal(poly_prs2, poly_prs2, prev_Rnorm);
#ifdef SNAPLIB_DEBUG
          printf("To be splitted (k=%d):\n", i);
#endif
          SNAPBASEREAL_sqrt_ui(const_tmp, SNAPLIB_MAX(SNAP_degree(poly_cof),SNAP_degree(poly_cog)));
          SNAPBASEREAL_mul_ui(const_tmp, const_tmp, 2);
          SNAPBASEREAL_mul(const_tmp, const_tmp, tol_temp);
          if ( SNAPBASEREAL_cmp(kth_Rnorm, prev_Rnorm) < 0 ) {
            SNAPBASEREAL_div(const_tmp, const_tmp, kth_Rnorm);
          } else {
            SNAPBASEREAL_div(const_tmp, const_tmp, prev_Rnorm);
          }
          if (
#ifdef SNAPLIB_USE_BEZOUTONLY_SPLIT
            SNAP_agcd_exqrgcdwp_split(poly_ncf, poly_prs1, tol_temp) == 0 && SNAP_agcd_exqrgcdwp_split(poly_ncg, poly_prs2, tol_temp) == 0
#else
            SNAP_agcd_exqrgcd_split(poly_ncf, poly_prs1, tol_temp) == 0 && SNAP_agcd_exqrgcd_split(poly_ncg, poly_prs2, tol_temp) == 0
#endif
          ) {
            found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f, poly_g, poly_ncf, poly_ncg, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
            free(ermin_list);
#endif
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_ncf);
              SNAP_set(poly_cog, poly_ncg);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
              SNAPBASE_clear(*(tau+j));
            }
            free(tau);
            SNAPMTX_clear(matrix_syl);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
            free(jpvt);
#endif
            return found_degree;
          } else {
#ifdef SNAPLIB_DEBUG
            if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
              printf("SPLIT failed...\n");
            }
#endif
            found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f, poly_g, poly_prs1, poly_prs2, poly_d, poly_u, poly_v, tol, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED);
            for ( j=0; j<mincdeg; j++ ) {
              SNAPBASEREAL_clear(*(erval_list+j));
            }
            free(erval_list);
            free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
            free(ermin_list);
#endif
            if ( found_degree != UINT_MAX && found_degree != 0 ) {
              SNAP_set(poly_cof, poly_prs1);
              SNAP_set(poly_cog, poly_prs2);
            }
            SNAP_clear(poly_ncf);
            SNAP_clear(poly_ncg);
            SNAP_clear(poly_prs1);
            SNAP_clear(poly_prs2);
            for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
              SNAPBASE_clear(*(tau+j));
            }
            free(tau);
            SNAPMTX_clear(matrix_syl);
            SNAPBASEREAL_clear(prev_Rnorm);
            SNAPBASEREAL_clear(kth_Rnorm);
            SNAPBASEREAL_clear(sqrtcols);
            SNAPBASEREAL_clear(const_tmp);
            SNAPBASEREAL_clear(const_case1);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
            free(jpvt);
#endif
            return found_degree;
          }
        }
        SNAP_clear(poly_ncf);
        SNAP_clear(poly_ncg);
        SNAP_clear(poly_prs1);
        SNAP_clear(poly_prs2);
      } else {
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is not allowed. Skipped...\n");
        }
#endif
      }
    }
    for ( j=0; j<mincdeg; j++ ) {
      SNAPBASEREAL_clear(*(erval_list+j));
    }
    free(erval_list);
    free(eridx_list);
#ifndef SNAPLIB_USE_QUICKSORT
    free(ermin_list);
#endif
  }
  if ( found_degree == UINT_MAX ) {
    // poly_d <== 1.0 hence we do nothing for poly_d
    // SNAP_degree(poly_d) = 0; SNAP_set_coef(poly_d, 0, 1.0);
    if ( case1flag == 1 ) {
      found_degree = 0;
    }
    // poly_u and poly_v be the cofactors from the last row of Q.
    // NOTE: poly_u is the cofactor of poly_g and poly_v is of poly_f.
    // OLD if ( qrstep_order == NORMAL || qrstep_order == NORMALONLY || qrstep_order == BEZOUTONLY ) {
    if ( qrstep_order == BEZOUTONLY ) {
      Qkth = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_init(*(Qkth+j));
      }
      SNAP_agcd_exqrgcd_Qkthcolumn(Qkth, matrix_syl, tau, SNAPMTX_rows(matrix_syl)-1);
#if defined(SNAP_SUPDC_T) && defined(SNAPLIB_DEBUG)
      printf("AAAA\n");
      for (q3=0; q3<SNAPMTX_rows(matrix_syl); q3++ ) {
        printf("(%+.5le%+.5leI),",creal(*(Qkth+q3)),cimag(*(Qkth+q3)));
      }
      printf("\n");
#endif
      for ( i=0; i<SNAP_degree(poly_cog); i++ ) {
        SNAP_set_coef(poly_u, i, *(Qkth+SNAP_degree(poly_cog)-1-i));
      }
      SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
      for ( i=0; i<SNAP_degree(poly_cof); i++ ) {
        SNAP_set_coef(poly_v, i, *(Qkth+SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i));
      }
      SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;
      SNAP_div_base(poly_u, poly_u, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      SNAP_div_base(poly_v, poly_v, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_clear(*(Qkth+j));
      }
      free(Qkth);
    }
  }
  for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
    SNAPBASE_clear(*(tau+j));
  }
  free(tau);
  SNAPMTX_clear(matrix_syl);
  SNAPBASEREAL_clear(prev_Rnorm);
  SNAPBASEREAL_clear(kth_Rnorm);
  SNAPBASEREAL_clear(sqrtcols);
  SNAPBASEREAL_clear(const_tmp);
  SNAPBASEREAL_clear(const_case1);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  free(jpvt);
#endif
  return found_degree;
}

// mode 0: automatic first order
// mode 1: normal side
// mode 2: reversal side
int SNAP_agcd_exqrgcdwp(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, int mode) {
  SNAPBASEREAL_t norm_f, norm_g, tol_temp, tol_res, norm_cf, norm_cg, lcf, lcg, scf, scg;
  SNAP_t poly_f, poly_g, poly_cof, poly_cog, poly_u, poly_v;
  SNAP_agcd_exqrgcd_qrstep_order qrstep_order, qrstep_firstorder;
  unsigned int found_degree, trials, coprimeresults;
  int result = -1;
  // initialization
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(tol_temp);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(norm_cf);
  SNAPBASEREAL_init(norm_cg);
  SNAPBASEREAL_init(lcf);
  SNAPBASEREAL_init(lcg);
  SNAPBASEREAL_init(scf);
  SNAPBASEREAL_init(scg);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
  SNAP_init_set(poly_cog, poly_g);
  SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
  SNAP_degree(poly_d) = 0;
  SNAPBASEREAL_set(tol_temp, tol);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  mpfr_printf("input tolerance: %Re\n", tol);
#else
  printf("input tolerance: %e\n", tol);
#endif
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_cof)) || SNAPBASE_zeroq(SNAP_lcoef(poly_cog)) ) {
    SNAP_clear(poly_f);
    SNAP_clear(poly_g);
    SNAP_clear(poly_cof);
    SNAP_clear(poly_cog);
    SNAPBASEREAL_clear(norm_f);
    SNAPBASEREAL_clear(norm_g);
    SNAPBASEREAL_clear(tol_temp);
    SNAPBASEREAL_clear(tol_res);
    SNAPBASEREAL_clear(norm_cf);
    SNAPBASEREAL_clear(norm_cg);
    SNAPBASEREAL_clear(lcf);
    SNAPBASEREAL_clear(lcg);
    SNAPBASEREAL_clear(scf);
    SNAPBASEREAL_clear(scg);
    fprintf(stderr, "EXQRGCD: polynomials must be non-zero!\n");
    return -1;
  }
  // depending on normal first or reversal first ?
  SNAPBASE_abs(lcf, SNAP_lcoef(poly_f));
  SNAPBASE_abs(lcg, SNAP_lcoef(poly_g));
  SNAPBASE_abs(scf, SNAP_coef(poly_f,0));
  SNAPBASE_abs(scg, SNAP_coef(poly_g,0));
  if ( SNAPBASEREAL_cmp(lcf,lcg) > 0 ) {
    SNAPBASEREAL_set(lcf, lcg);
  }
  if ( SNAPBASEREAL_cmp(scf,scg) > 0 ) {
    SNAPBASEREAL_set(scf, scg);
  }
  if ( mode == 1 || ( mode == 0 && SNAPBASEREAL_cmp(lcf,scf) >= 0 ) ) {
    qrstep_firstorder = NORMAL;
  } else {
    qrstep_firstorder = REVERSAL;
  }
#ifdef SNAPLIB_DEBUG
  printf("======= Order: %s =======\n", qrstep_firstorder == NORMAL ? "Normal":"Reversal" );
#endif
  // some agcd will be found then return, otherwise change qrstep_order.
  SNAPBASEREAL_set_ui(norm_cf, 1);
  SNAPBASEREAL_set_ui(norm_cg, 1);
  trials = 0;
  coprimeresults = 0;
  qrstep_order = qrstep_firstorder;
  while (1) {
    if ( trials > 0 || qrstep_firstorder == REVERSAL ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      if ( trials == 1 ) {
        break;  // re-try from the other side.
      }
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  if ( result == -1 ) {
#ifdef SNAPLIB_DEBUG
    printf("======= Second Order: %s =======\n", qrstep_firstorder != NORMAL ? "Normal":"Reversal" );
#endif
    if ( qrstep_firstorder == qrstep_order ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
    }
    SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
    SNAP_init_set(poly_cog, poly_g);
    SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
    SNAP_degree(poly_d) = 0;
    SNAPBASEREAL_set(tol_temp, tol);
    trials = 0;
    coprimeresults = 0;
    qrstep_order = (qrstep_firstorder == NORMAL ? REVERSAL:NORMAL);
  }
  while ( result == -1 ) {
    if ( trials > 0 ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 2 ) {
      found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_exqrgcdwp_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_temp, tol_res,qrstep_order,SPLITDISALLOWED);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      coprimeresults++;
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
      SNAPBASEREAL_add(tol_temp, tol_temp, tol_res);
      coprimeresults = 0;
    }
    if ( coprimeresults >= 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      if ( qrstep_order == REVERSAL ) {
        SNAP_self_reversal(poly_d);
        SNAP_self_reversal(poly_f1);
        SNAP_self_reversal(poly_g1);
      }
      result = 0;
      break;
    }
    qrstep_order = (qrstep_order == NORMAL ? REVERSAL:NORMAL);
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAP_clear(poly_cof);
  SNAP_clear(poly_cog);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(tol_temp);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(norm_cf);
  SNAPBASEREAL_clear(norm_cg);
  SNAPBASEREAL_clear(lcf);
  SNAPBASEREAL_clear(lcg);
  SNAPBASEREAL_clear(scf);
  SNAPBASEREAL_clear(scg);
#ifdef SNAPLIB_DEBUG
  printf("Finally the number of trials: %d\n", trials);
#endif
  return result;
}

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
// With Full Pivoting (by Sorting)
int SNAP_agcd_exqrgcdfp(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, int mode) {
  SNAPBASEREAL_t norm_f, norm_g;
  SNAP_t poly_f, poly_g;
  int result = -1;
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAP_inorm(norm_f, poly_f);
  SNAP_inorm(norm_g, poly_g);
  if ( norm_f >= norm_g ) {
    result = SNAP_agcd_exqrgcdwp(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, tol, mode);
  } else {
    result = SNAP_agcd_exqrgcdwp(poly_ing, poly_inf, poly_d, poly_g1, poly_f1, tol, mode);
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  return result;
}
#endif
