/*
 * Copyright (c) 2013-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <fftw3.h>
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAPLIB_MAX(a,b) (((a)>(b)) ? (a):(b))
#define SNAPLIB_MIN(a,b) (((a)<(b)) ? (a):(b))

// #define SNAPLIB_QR_K_STEP 1000
// #define SNAPLIB_QR_K_STEP 10
#define SNAPLIB_QR_K_STEP 1

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

// This shows the degree information detected.
#define SNAPLIB_SHOW_STATUS

/* CURRENT SCHEME:
   1. First detection for tentative degree: Boito's (e.g. GKO with eps=0.0)
   2. Adjustment for tentative degree: by rank deficiency of subresultant matrix
   3. Null vector of subresultant: Boito's threshold.
   4. Max number of iterations is 20 and threshold = 1.0e-15 that are used in Boito's.
   5. QR in GKO is called everytime (K=1) as in Boito's.
 */

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
#define show_plist show_dlist
#define show_sdmp show_sdmd
#define show_sdmpc(a) {}
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.5le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPDC_T)
#define show_plist show_dclist
#define show_sdmp show_sdmdc
#define show_sdmpc show_sdmdc
void show_sdmdc(sdmdc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmdc_rows(mtx); i++) {
    for(j=0; j<sdmdc_cols(mtx); j++) {
      printf("(%+.8le%+.8le*I), ",creal(sdmdc_elem(mtx,i,j)),cimag(sdmdc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_dclist(dcomplex *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.16le%+.16le*I), ",creal(*(list+i)),cimag(*(list+i)));
  }
  printf("\n");
}
void show_dclist2(dcomplex *list, unsigned int rows, unsigned int cols, unsigned int rsize, unsigned int csize) {
  unsigned int i,j;
  for(i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      printf("(%+.16le%+.16le*I), ",creal(*(list+i*csize+j*rsize)),cimag(*(list+i*csize+j*rsize)));
    }
    printf("\n");
  }
}
#elif defined(SNAP_SUPFR_T)
#define show_plist show_frlist
#define show_sdmp show_sdmfr
#define show_sdmpc(a) {}
void show_sdmfr(sdmfr_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfr_rows(mtx); i++) {
    for(j=0; j<sdmfr_cols(mtx); j++) {
      mpfr_printf("%+.5Re ",sdmfr_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_frlist(mpfr_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("%+.5Re ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPFC_T)
#define show_plist show_fclist
#define show_sdmp show_sdmfc
#define show_sdmpc show_sdmfc
void show_sdmfc(sdmfc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfc_rows(mtx); i++) {
    for(j=0; j<sdmfc_cols(mtx); j++) {
      mpfr_printf("(%+.8Re%+.8Re*I), ",mpc_realref(sdmfc_elem(mtx,i,j)),mpc_imagref(sdmfc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_fclist(mpc_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("(%+.16Re%+.16Re*I), ",mpc_realref(*(list+i)),mpc_imagref(*(list+i)));
  }
  printf("\n");
}
#endif
#endif

// #define SNAP_SUPD_T

/*!
 * @brief The function computes a LU factorization of a Cauchy-like matrix C=PLUQ, given by generatos s.t. Omega C-C Lambda=AB.
 * @details See: M. Gu. Stable and efficient algorithms for structured systems of linear equations.
 *  SIAM J. Matrix Anal. Appl., 19(2):279–306, 1998, and M. Gu. New Fast Algorithms for Structured Linear Least Squares Problems.
 *  SIAM J. Matrix Anal. Appl., 20(1):244-269, 1999.
 * @warning All the matrices must be of the row major. Rank 1 update BLAS routines should be uses instead.
 * @param[in,out] opOm The 1d-array (of length rows) representing the displacement operator Omega on input and the resulting operator Omega on exit.
 * @param[in,out] opLm The 1d-array (of length cols) representing the displacement operator Lambda on input and the resulting operator Lambda on exit.
 * @param[in,out] genA The generator A on input and the resulting generator A on exit.
 * @param[in,out] genB The generator B on input and the resulting generator B on exit.
 * @param[in,out] matW The zero matrix LU (of size rows x cols) on input and the resulting lower/upper triangular matrices LU on exit.
 * @param[in,out] matP The 1d-array (of length rows) with row numbers on input and the resulting permutation on exit. The row i of the matrix was interchanged with row matP[i].
 * @param[in,out] matQ The 1d-array (of length cols) with col numbers on input and the resulting permutation on exit. The col i of the matrix was interchanged with col matQ[i].
 * @param[in] eps The threshold to terminate the GEPP by the modified GKO method.
 * @param[in] cmax The maximum number of columns should be reduced by the modified GKO method.
 * @return a positive integer (last pivot but not used, +1) if success or 0 if failure.
 */
unsigned int SNAP_agcd_fastgcd_modified_gko(SNAPMTXCOMPLEX_t opOm, SNAPMTXCOMPLEX_t opLm, SNAPMTXCOMPLEX_t genA, SNAPMTXCOMPLEX_t genB, SNAPMTXCOMPLEX_t matW, unsigned int *matP, unsigned int *matQ, SNAPBASEREAL_t eps, unsigned int cmax) {
  unsigned int n, m, r, k, i, j, jmax, kmax, err=0;
  SNAPMTXCOMPLEX_t tau;
  SNAPBASECOMPLEX_t tmp, zero, one, mone, *pt1, *pt4;
#if !defined(SNAP_SUPD_T) && !defined(SNAP_SUPDC_T)
  SNAPBASECOMPLEX_t *pt2, *pt3;
#endif // defined
  SNAPBASEREAL_t nrm, nrmmax;
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_t tmpr1, tmpr2;
#endif // SNAP_SUPFC_T
  // opOm:(1,m), opLm:(1,n), genA:(m,r), genB:(r,n), matW:(m,n), matP:(1,m), matQ:(1,n)
  m = SNAPMTXCOMPLEX_cols(opOm);
  n = SNAPMTXCOMPLEX_cols(opLm);
  r = SNAPMTXCOMPLEX_cols(genA);
#ifdef SNAPLIB_DEBUG
  printf("m: %d, n: %d, r: %d\n", m, n, r);
#endif
  SNAPMTXCOMPLEX_init2(tau, 1, m);
  SNAPBASECOMPLEX_init(tmp);
  SNAPBASECOMPLEX_init_set_ui(zero, 0);
  SNAPBASECOMPLEX_init_set_ui(one, 1);
  SNAPBASECOMPLEX_init_set_si(mone, -1);
  SNAPBASEREAL_init(nrm);
  SNAPBASEREAL_init(nrmmax);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_init(tmpr1);
  SNAPBASEREAL_init(tmpr2);
#endif // SNAP_SUPFC_T
  for ( k=0; k<cmax; k++ ) {
#ifdef SNAPLIB_DEBUG
    printf("loop top (genA)\n");
    show_sdmpc(genA);
    printf("loop top (genB)\n");
    show_sdmpc(genB);
    printf("loop top (matW)\n");
    show_sdmpc(matW);
#endif
    if ( (k+1) % SNAPLIB_QR_K_STEP == 0 && m-k >= r ) {
#ifdef SNAPLIB_DEBUG
      printf("==== QR called for preventing large generators ====\n");
      show_sdmpc(genA);
#endif
      // A_{k:m,}=Q.R
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
        clapack_zgeqrf(CblasRowMajor, m-k, r, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elems(tau)) != 0
#else
        LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, m-k, r, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elems(tau)) != 0
#endif
      ) {
        printf("ERROR in agcd_fastgcd2: zgeqrf could not compute the QR decomposition.\n");
        err = 1;
        goto post;
      }
#elif defined(SNAP_SUPFC_T)
      if ( snap_mlapack_fcgeqrf(m-k, r, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elems(tau)) != 0 ) {
        printf("ERROR in agcd_fastgcd2: zgeqrf could not compute the QR decomposition.\n");
        err = 1;
        goto post;
      }
#else
      ERR;
#endif
      // B_{,k:n}=R.B_{,k:n}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAPBASECOMPLEX_set_ui(tmp, 1);
      cblas_ztrmm(CblasRowMajor, CblasLeft, CblasUpper, CblasNoTrans, CblasNonUnit, r, n-k, &tmp, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB));
#elif defined(SNAP_SUPFC_T)
      SNAPBASECOMPLEX_set_ui(tmp, 1);
      snap_mblas_fctrmm('L', 'U', 'N', 'N', r, n-k, tmp, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB));
#else
      ERR;
#endif
      // A_{k:m,}=Q
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( LAPACKE_zungqr(LAPACK_ROW_MAJOR, m-k, r, r, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elems(tau)) != 0 ) {
        printf("ERROR in agcd_fastgcd2: LAPACKE_zungqr could not compute the matrix Q.\n");
        err = 1;
        goto post;
      }
#elif defined(SNAP_SUPFC_T)
      if ( snap_mlapack_fcungqr(m-k, r, r, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTXCOMPLEX_csize(genA), SNAPMTXCOMPLEX_elems(tau)) != 0 ) {
        printf("ERROR in agcd_fastgcd2: snap_mlapack_fcungqr could not compute the matrix Q.\n");
        err = 1;
        goto post;
      }
#else
      ERR;
#endif
      jmax = k;
      SNAPMTXCOMPLEX_BLAS_dotc(tmp, r, SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB), SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB));
      SNAPBASEREAL_set_basecomplex(nrmmax, tmp);
      for ( j=k+1,pt1=SNAPMTXCOMPLEX_elemp(genB,0,k+1); j<n; j++,pt1++ ) {
        SNAPMTXCOMPLEX_BLAS_dotc(tmp,r,pt1,SNAPMTXCOMPLEX_csize(genB),pt1,SNAPMTXCOMPLEX_csize(genB));
        SNAPBASEREAL_set_basecomplex(nrm, tmp);
        if ( SNAPBASEREAL_cmp(nrm, nrmmax) > 0 ) {
#ifdef SNAPLIB_DEBUG
          printf("[jmax ==> %d, nrm=%e, nrmmax=%e]\n", j, nrm, nrmmax);
#endif
          jmax = j;
          SNAPBASEREAL_set(nrmmax, nrm);
        }
      }
      if ( jmax > k ) {
        // swaptmp = *(matQ+k); *(matQ+k) = *(matQ+jmax); *(matQ+jmax) = swaptmp;
        *(matQ+k) = jmax+1;
        SNAPBASECOMPLEX_swap(SNAPMTXCOMPLEX_elem(opLm,0,k), SNAPMTXCOMPLEX_elem(opLm,0,jmax));
        SNAPMTXCOMPLEX_BLAS_swap(r, SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB), SNAPMTXCOMPLEX_elemp(genB,0,jmax), SNAPMTXCOMPLEX_csize(genB));
        SNAPMTXCOMPLEX_BLAS_swap(k, SNAPMTXCOMPLEX_elemp(matW,0,k), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elemp(matW,0,jmax), SNAPMTXCOMPLEX_csize(matW));
      }
#ifdef SNAPLIB_DEBUG
      printf("[jmax=%d, k=%d]\n", jmax, k);
#endif
#ifdef SNAPLIB_DEBUG
      printf("after QR (genA)\n");
      show_sdmpc(genA);
      printf("after QR (genB)\n");
      show_sdmpc(genB);
      printf("after QR (matW)\n");
      show_sdmpc(matW);
#endif
    }
    // L_{k:n,k}:=...
    // k_{max}:=argmax...
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_zgemv(CblasRowMajor, CblasNoTrans, m-k, r, &one, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTX_csize(genA), SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB), &zero, SNAPMTXCOMPLEX_elemp(matW,k,k), SNAPMTXCOMPLEX_csize(matW));
    kmax = k;
    pt4 = SNAPMTXCOMPLEX_elemp(matW,k,k);
    for ( i=k,pt1=pt4; i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
      SNAPBASECOMPLEX_sub(tmp,SNAPMTXCOMPLEX_elem(opOm,0,i),SNAPMTXCOMPLEX_elem(opLm,0,k));
      SNAPBASECOMPLEX_div(*pt1,*pt1,tmp);
#ifdef SNAPLIB_DEBUG
      printf("*pt1: %e%+eI, *pt4: %e%+eI\n", creal(*pt1), cimag(*pt1), creal(*pt4), cimag(*pt4));
      printf("[i=%d] cabs(*pt4): %e < eps: %e, fabs(eps): %e, cabs(*pt4)-fabs(eps): %e\n", i, cabs(*pt4), eps, fabs(eps), cabs(*pt4)-fabs(eps));
#endif
#ifdef SNAPLIB_DEBUG
      printf("|*pt1|=%.16e, |*pt4|=%.16e\n", cabs(*pt1), cabs(*pt4));
#endif // SNAPLIB_DEBUG
      if ( SNAPBASECOMPLEX_cmpabs(*pt1, *pt4) > 0 ) {
        kmax=i;
        pt4=pt1;
#ifdef SNAPLIB_DEBUG
        printf("kmax=%d\n", kmax);
#endif // SNAPLIB_DEBUG
      }
    }
#else
    snap_mblas_fcgemv('N', m-k, r, one, SNAPMTXCOMPLEX_elemp(genA,k,0), SNAPMTX_csize(genA), SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB), zero, SNAPMTXCOMPLEX_elemp(matW,k,k), SNAPMTXCOMPLEX_csize(matW));
    kmax = k;
    pt4 = SNAPMTXCOMPLEX_elemp(matW,k,k);
    for ( i=k,pt1=pt4; i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
      SNAPBASECOMPLEX_sub(tmp,SNAPMTXCOMPLEX_elem(opOm,0,i),SNAPMTXCOMPLEX_elem(opLm,0,k));
      SNAPBASECOMPLEX_div(*pt1,*pt1,tmp);
#ifdef SNAPLIB_DEBUG
      SNAPBASECOMPLEX_abs(tmpr1, *pt4);
      SNAPBASEREAL_abs(tmpr2, eps);
      SNAPBASEREAL_sub(tmpr2, tmpr1, tmpr2);
      mpfr_printf("*pt1: %Re%+ReI, *pt4: %Re%+ReI\n", mpc_realref(*pt1), mpc_imagref(*pt1), mpc_realref(*pt4), mpc_imagref(*pt4));
      mpfr_printf("[i=%d] cabs(*pt4): %Re < eps: %Re, fabs(eps): N/A, cabs(*pt4)-fabs(eps): %Re\n", i, tmpr1, eps, tmpr2);
#endif
      SNAPBASECOMPLEX_abs(tmpr1, *pt1);
      SNAPBASECOMPLEX_abs(tmpr2, *pt4);
#ifdef SNAPLIB_DEBUG
      mpfr_printf("|*pt1|=%.16Re, |*pt4|=%.16Re\n", tmpr1, tmpr2);
#endif // SNAPLIB_DEBUG
      if ( SNAPBASEREAL_cmp(tmpr1,tmpr2)>0 ) {
        kmax=i;
        pt4=pt1;
#ifdef SNAPLIB_DEBUG
        printf("kmax=%d\n", kmax);
#endif // SNAPLIB_DEBUG
      }
    }
#endif
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPDC_T
    printf("[kmax=%d, k=%d] cabs(*pt4): %e < eps: %e, fabs(eps): %e, cabs(*pt4)-fabs(eps): %e\n", kmax, k, cabs(*pt4), eps, fabs(eps), cabs(*pt4)-fabs(eps));
#endif
#ifdef SNAP_SUPFC_T
    SNAPBASECOMPLEX_abs(tmpr1, *pt4);
    SNAPBASEREAL_abs(tmpr2, eps);
    SNAPBASEREAL_sub(tmpr2, tmpr1, tmpr2);
    mpfr_printf("[kmax=%d, k=%d] cabs(*pt4): %Re < eps: %Re, fabs(eps): N/A, cabs(*pt4)-fabs(eps): %Re\n", kmax, k, tmpr1, eps, tmpr2);
#endif
#endif
    // rank deficient condition
#ifdef SNAP_SUPFC_T
    SNAPBASECOMPLEX_abs(tmpr1, *pt4);
    if ( SNAPBASEREAL_cmp(tmpr1, eps) <= 0 ) {
#else
    if ( SNAPBASECOMPLEX_cmpabs_basereal(*pt4,eps) <= 0 ) {
#endif // SNAP_SUPFC_T
      k++;
      goto post;
    }
    // if k_max > k then P:=, Omega:=, A:=, L_{:,1:k}:=... endif
    if ( kmax > k ) {
      // swaptmp = *(matP+k); *(matP+k) = *(matP+kmax); *(matP+kmax) = swaptmp;
      *(matP+k) = kmax+1;
      SNAPBASECOMPLEX_swap(SNAPMTXCOMPLEX_elem(opOm,0,k), SNAPMTXCOMPLEX_elem(opOm,0,kmax));
      SNAPMTXCOMPLEX_BLAS_swap(r, SNAPMTXCOMPLEX_elemp(genA,k,0), 1, SNAPMTXCOMPLEX_elemp(genA,kmax,0), 1);
      SNAPMTXCOMPLEX_BLAS_swap(k+1, SNAPMTXCOMPLEX_elemp(matW,k,0), 1, SNAPMTXCOMPLEX_elemp(matW,kmax,0), 1);
    }
    // U_{k,k}=L_{k,k}, L_{k,k}=1 <== do nothing
    // U_{k,k+1:n} = A_{k,}B_{,k+1:n}(w_k-Lm_{k+1:n})^{-1}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_zgemv(CblasRowMajor, CblasTrans, r, n-k-1, &one, SNAPMTXCOMPLEX_elemp(genB,0,k+1), SNAPMTX_csize(genB), SNAPMTXCOMPLEX_elemp(genA,k,0), 1, &zero, SNAPMTXCOMPLEX_elemp(matW,k,k+1), 1);
    for ( i=k+1,pt1=SNAPMTXCOMPLEX_elemp(matW,k,k+1); i<n; i++,pt1++ ) {
      SNAPBASECOMPLEX_sub(tmp,SNAPMTXCOMPLEX_elem(opOm,0,k),SNAPMTXCOMPLEX_elem(opLm,0,i));
      SNAPBASECOMPLEX_div(*pt1,*pt1,tmp);
    }
#else
    snap_mblas_fcgemv('T', r, n-k-1, one, SNAPMTXCOMPLEX_elemp(genB,0,k+1), SNAPMTX_csize(genB), SNAPMTXCOMPLEX_elemp(genA,k,0), 1, zero, SNAPMTXCOMPLEX_elemp(matW,k,k+1), 1);
    for ( i=k+1,pt1=SNAPMTXCOMPLEX_elemp(matW,k,k+1); i<n; i++,pt1++ ) {
      SNAPBASECOMPLEX_sub(tmp,SNAPMTXCOMPLEX_elem(opOm,0,k),SNAPMTXCOMPLEX_elem(opLm,0,i));
      SNAPBASECOMPLEX_div(*pt1,*pt1,tmp);
    }
#endif
    // L_{k+1:m,k} /= U_{k,k}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPBASECOMPLEX_rec(tmp, SNAPMTXCOMPLEX_elem(matW,k,k));
    cblas_zscal(m-k-1, &tmp, SNAPMTXCOMPLEX_elemp(matW,k+1,k), SNAPMTXCOMPLEX_csize(matW));
#else
    for ( i=k+1,pt1=SNAPMTXCOMPLEX_elemp(matW,k+1,k),pt2=SNAPMTXCOMPLEX_elemp(matW,k,k); i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
      SNAPBASECOMPLEX_div(*pt1,*pt1,*pt2);
    }
#endif // defined
    // A_{k+1:m,}:=A_{k+1:m,}-L_{k+1:m,k}*A_{k,}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_zgeru(CblasRowMajor, m-k-1, r, &mone, SNAPMTXCOMPLEX_elemp(matW,k+1,k), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elemp(genA,k,0), 1, SNAPMTXCOMPLEX_elemp(genA,k+1,0), SNAPMTXCOMPLEX_csize(genA));
#else
    for ( i=k+1; i<m; i++ ) {
      for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genA,i,0),pt2=SNAPMTXCOMPLEX_elemp(matW,i,k),pt3=SNAPMTXCOMPLEX_elemp(genA,k,0); j<r; j++,pt1++,pt3++ ) {
#ifdef SNAP_SUPFC_T
        SNAPBASECOMPLEX_mul(tmp, *pt2, *pt3);
        SNAPBASECOMPLEX_sub(*pt1, *pt1, tmp);
#else
        SNAPBASECOMPLEX_submul(*pt1,*pt2,*pt3);
#endif // SNAP_SUPFC_T
      }
    }
#endif // defined
    // B_{,k+1:n}:=B_{,k+1:n}-B_{,k}*U_{k,k+1:n}/U_{k,k}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPBASECOMPLEX_rec(tmp, SNAPMTXCOMPLEX_elem(matW,k,k));
    SNAPBASECOMPLEX_neg(tmp, tmp);
    cblas_zgeru(CblasRowMajor, r, n-k-1, &tmp, SNAPMTXCOMPLEX_elemp(genB,0,k), SNAPMTXCOMPLEX_csize(genB), SNAPMTXCOMPLEX_elemp(matW,k,k+1), 1, SNAPMTXCOMPLEX_elemp(genB,0,k+1), SNAPMTXCOMPLEX_csize(genB));
#else
    for ( i=k+1; i<n; i++ ) {
      for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genB,0,i),pt2=SNAPMTXCOMPLEX_elemp(matW,k,i),pt3=SNAPMTXCOMPLEX_elemp(genB,0,k),pt4=SNAPMTXCOMPLEX_elemp(matW,k,k); j<r; j++,pt1+=SNAPMTXCOMPLEX_csize(genB),pt3+=SNAPMTXCOMPLEX_csize(genB) ) {
        SNAPBASECOMPLEX_div(tmp, *pt2, *pt4);
#ifdef SNAP_SUPFC_T
        SNAPBASECOMPLEX_mul(tmp, tmp, *pt3);
        SNAPBASECOMPLEX_sub(*pt1, *pt1, tmp);
#else
        SNAPBASECOMPLEX_submul(*pt1,*pt3,tmp);
#endif // SNAP_SUPFC_T
      }
    }
#endif // defined
  }
  // loop ends without goto statement means k=n+1 since we return non-used pivot index "+1".
  k++;
post:
  SNAPBASECOMPLEX_clear(tmp);
  SNAPBASECOMPLEX_clear(zero);
  SNAPBASECOMPLEX_clear(one);
  SNAPBASECOMPLEX_clear(mone);
  SNAPBASEREAL_clear(nrm);
  SNAPBASEREAL_clear(nrmmax);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_clear(tmpr1);
  SNAPBASEREAL_clear(tmpr2);
#endif // SNAP_SUPFC_T
  SNAPMTXCOMPLEX_clear(tau);
  if ( err == 1 ) {
    return 0;
  } else {
    return k;
  }
}

/*!
 *  @brief This computes a right null vector of Cauchy-like matrix transformed from Toeplitz-like matrix.
 *  @details For the given C=PLUQ=FmMDhFn^h, this solves Mx=0 by x=DhFn^hQ^ty with PLUy=0 hence Uy=0.
 *  @param[out] vecX The complex vector of length m. A right null vector on exit.
 *  @param[in,out] matW The lower/upper triangular matrices LU as the result of GKO. The last column is overwritten on exit.
 *  @param[in] matQ The 1d-array (of length columns of matW) with col numbers, resulting permutation. matQ[i] denotes the original col number.
 *  @param[in] vecDh The complex vector of length m. i-th element is the conjugate of E^(Pi*I*(i-1)/(m*LCM(n,m))).
 *  @param[in] eps The threshold for the nullity.
 */
void SNAP_agcd_fastgcd_nullvector(SNAPMTXCOMPLEX_t vecX, SNAPMTXCOMPLEX_t matW, unsigned int *matQ, SNAPMTXCOMPLEX_t vecDh, SNAPBASEREAL_t eps) {
  unsigned int i, m, nullm, *ip;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_plan fftp;
  int fftdim[1];
  double sqrtofnm;
#endif // defined
  SNAPBASECOMPLEX_t *pt1,*pt2,*pt3;
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_t tmpr, sqrtofnm;
  SNAPBASEREAL_init(tmpr);
  SNAPBASEREAL_init(sqrtofnm);
#endif // SNAP_SUPFC_T
  m = SNAPMTXCOMPLEX_cols(matW);
  nullm = m;
#ifdef SNAP_SUPFC_T
  SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,nullm-2,nullm-2));
  while ( nullm > 2 && SNAPBASEREAL_cmp(tmpr, eps) <= 0 ) {
    nullm--;
    SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,nullm-2,nullm-2));
  }
#else
  while ( nullm > 2 && SNAPBASECOMPLEX_cmpabs_basereal(SNAPMTXCOMPLEX_elem(matW,nullm-2,nullm-2),eps) <= 0 ) {
    nullm--;
  }
#endif // SNAP_SUPFC_T
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
  mpfr_printf("nullity: %d (m=%d, nullm=%d) with eps: %Re\n", m-nullm+1, m, nullm, eps);
#else
  printf("nullity: %d (m=%d, nullm=%d) with eps: %e\n", m-nullm+1, m, nullm, eps);
#endif // SNAP_SUPFC_T
  printf("matW:\n");
  show_sdmpc(matW);
#endif
  // solve Uy=0
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', nullm-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elemp(matW,0,nullm-1), SNAPMTXCOMPLEX_csize(matW)) != 0 ) {
    printf("ERROR in ztrtrs hence we could not compute a null vector.\n");
    return;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fctrtrs('U', 'N', 'N', nullm-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elemp(matW,0,nullm-1), SNAPMTXCOMPLEX_csize(matW)) != 0 ) {
    printf("ERROR in fctrtrs hence we could not compute a null vector.\n");
    SNAPBASEREAL_clear(tmpr);
    SNAPBASEREAL_clear(sqrtofnm);
    return;
  }
#else
  ERR;
#endif
#ifdef SNAPLIB_DEBUG
  printf(" ==> matW:\n");
  show_sdmpc(matW);
#endif // SNAPLIB_DEBUG
  SNAPBASECOMPLEX_set_si(SNAPMTXCOMPLEX_elem(matW,nullm-1,nullm-1), -1);
  for ( i=nullm,pt1=SNAPMTXCOMPLEX_elemp(matW,nullm,nullm-1); i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
    SNAPBASECOMPLEX_set_ui(*pt1,0);
  }
  // x <= Qx
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(vecX,0,m-1),ip=matQ+m-1,pt2=SNAPMTXCOMPLEX_elemp(matW,m-1,nullm-1); i<m; i++,pt1--,ip--,pt2-=SNAPMTXCOMPLEX_csize(matW) ) {
    SNAPBASECOMPLEX_set(*pt1, *pt2);
    if ( *ip != 0 ) {
      SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(vecX,0,*ip-1));
    }
  }
#ifdef SNAPLIB_DEBUG
  printf("vecX(m=%d):\n",m);
  show_sdmpc(vecX);
#endif // SNAPLIB_DEBUG
  // x <= Q^t x
  //for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(vecX,0,0),ip=matQ,pt2=SNAPMTXCOMPLEX_elemp(matW,0,m-1); i<m; i++,pt1++,ip++,pt2+=SNAPMTXCOMPLEX_csize(matW) ) {
  //	SNAPBASECOMPLEX_set(*pt1, *pt2);
  //	if ( *ip != 0 ) { SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(vecX,0,*ip-1)); }
  //}
  // y <= Fn^h x
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  // FFTW
  fftdim[0] = m;
  fftp = fftw_plan_many_dft(1, fftdim, 1, SNAPMTXCOMPLEX_elems(vecX), fftdim, 1, 1, SNAPMTXCOMPLEX_elemp(matW,0,nullm-1), fftdim, SNAPMTXCOMPLEX_csize(matW), 1, FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
  fftw_destroy_plan(fftp);
  sqrtofnm = sqrt(m);
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(matW,0,nullm-1); i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
    SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASEREAL_sqrt_ui(sqrtofnm, m);
  snap_fcdft_forward(m, 1, SNAPMTXCOMPLEX_elemp(matW,0,nullm-1), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elems(vecX), 1);
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(matW,0,nullm-1); i<m; i++,pt1+=SNAPMTXCOMPLEX_csize(matW) ) {
    SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
  }
#else
  ERR;
#endif
#ifdef SNAPLIB_DEBUG
  printf("vecX:\n");
  show_sdmpc(vecX);
#endif // SNAPLIB_DEBUG
  // x <= Dh y
  for ( i=0,pt1=SNAPMTXCOMPLEX_elems(vecX),pt2=SNAPMTXCOMPLEX_elems(vecDh),pt3=SNAPMTXCOMPLEX_elemp(matW,0,nullm-1); i<m; i++,pt1++,pt2++,pt3+=SNAPMTXCOMPLEX_csize(matW) ) {
    SNAPBASECOMPLEX_mul(*pt1, *pt2, *pt3);
  }
#ifdef SNAPLIB_DEBUG
  printf("vecX:\n");
  show_sdmpc(vecX);
#endif // SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_clear(tmpr);
  SNAPBASEREAL_clear(sqrtofnm);
#endif // SNAP_SUPFC_T
}

/*!
 *  @brief This computes the generators of Cauchy-like matrix transformed from Toeplitz-like matrix.
 *  @param[in,out] genG G is the generator G of Sylvester matrix S s.t. Z_1 S - S Z_theta = GB on input, and is the generator G of Cauchy-like matrix s.t. D_1 C - C D_theta = GB on exit.
 *  @param[in,out] genB B is the generator B of Sylvester matrix S s.t. Z_1 S - S Z_theta = GB on input. and is the generator B of Cauchy-like matrix s.t. D_1 C - C D_theta = GB on exit.
 *  @param[in] genTG TG is a temporal array of rows/cols/rsize/csize dim(G) exactly.
 *  @param[in] genTB TB is a temporal array of rows/cols/rsize/csize dim(B) exactly.
 *  @param[in] vecDh The complex vector of length m. i-th element is the conjugate of E^(Pi*I*(i-1)/(m*LCM(n,m))).
 */
void SNAP_agcd_fastgcd_make_cauchy_generators(SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPMTXCOMPLEX_t genTG, SNAPMTXCOMPLEX_t genTB, SNAPMTXCOMPLEX_t vecDh) {
  unsigned int i, j;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3;
  // genG <= matFn genG
  SNAPMTXCOMPLEX_rows(genTG) = SNAPMTXCOMPLEX_rows(genG);
  SNAPMTXCOMPLEX_cols(genTG) = SNAPMTXCOMPLEX_cols(genG);
  SNAPMTXCOMPLEX_rsize(genTG) = SNAPMTXCOMPLEX_rsize(genG);
  SNAPMTXCOMPLEX_csize(genTG) = SNAPMTXCOMPLEX_csize(genG);
#ifdef SNAPLIB_DEBUG
  printf(" input genG: \n");
  show_sdmpc(genG);
  printf(" input genB: \n");
  show_sdmpc(genB);
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  // FFTW
  fftw_plan fftp;
  int fftdim[1];
  fftdim[0] = SNAPMTXCOMPLEX_rows(genG);
  fftp = fftw_plan_many_dft(1, fftdim, SNAPMTXCOMPLEX_cols(genG), SNAPMTXCOMPLEX_elems(genG), fftdim, SNAPMTXCOMPLEX_csize(genG), 1, SNAPMTXCOMPLEX_elems(genTG), fftdim, SNAPMTXCOMPLEX_csize(genTG), 1, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
  fftw_destroy_plan(fftp);
  double sqrtofnm;
  sqrtofnm = sqrt(SNAPMTXCOMPLEX_rows(genG));
  for ( i=0; i<SNAPMTXCOMPLEX_rows(genTG); i++ ) {
    for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genTG,i,0); j<SNAPMTXCOMPLEX_cols(genTG); j++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASEREAL_t sqrtofnm;
  SNAPBASEREAL_init(sqrtofnm);
  SNAPBASEREAL_sqrt_ui(sqrtofnm, SNAPMTXCOMPLEX_rows(genG));
  snap_fcdft_backward(SNAPMTXCOMPLEX_rows(genG), SNAPMTXCOMPLEX_cols(genG), SNAPMTXCOMPLEX_elems(genTG), SNAPMTXCOMPLEX_csize(genTG), SNAPMTXCOMPLEX_elems(genG), SNAPMTXCOMPLEX_csize(genG));
  for ( i=0; i<SNAPMTXCOMPLEX_rows(genTG); i++ ) {
    for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genTG,i,0); j<SNAPMTXCOMPLEX_cols(genTG); j++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
  }
#else
  ERR;
#endif
#ifdef SNAPLIB_DEBUG
  printf(" ==> genTG: \n");
  show_sdmpc(genTG);
#endif
  SNAPMTXCOMPLEX_swap(genG, genTG);
  // genB <= genB diagmatrix(vecD)^H matFm^H = (genB diagmatrix(vecDh)) matFm^H
  SNAPMTXCOMPLEX_rows(genTB) = SNAPMTXCOMPLEX_rows(genB);
  SNAPMTXCOMPLEX_cols(genTB) = SNAPMTXCOMPLEX_cols(genB);
  SNAPMTXCOMPLEX_rsize(genTB) = SNAPMTXCOMPLEX_rsize(genB);
  SNAPMTXCOMPLEX_csize(genTB) = SNAPMTXCOMPLEX_csize(genB);
  for ( i=0; i<SNAPMTXCOMPLEX_rows(genTB); i++ ) {
    for ( j=0, pt1=SNAPMTXCOMPLEX_elemp(genTB,i,0), pt2=SNAPMTXCOMPLEX_elems(vecDh), pt3=SNAPMTXCOMPLEX_elemp(genB,i,0); j<SNAPMTXCOMPLEX_cols(genB); j++,pt1++,pt2++,pt3++ ) {
      SNAPBASECOMPLEX_mul(*pt1, *pt2, *pt3);
    }
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  // FFTW
  fftdim[0] = SNAPMTXCOMPLEX_cols(genTB);
  fftp = fftw_plan_many_dft(1, fftdim, SNAPMTXCOMPLEX_rows(genTB), SNAPMTXCOMPLEX_elems(genTB), fftdim, 1, SNAPMTXCOMPLEX_csize(genTB), SNAPMTXCOMPLEX_elems(genB), fftdim, 1, SNAPMTXCOMPLEX_csize(genB), FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
  fftw_destroy_plan(fftp);
  sqrtofnm = sqrt(SNAPMTXCOMPLEX_cols(genTB));
  for ( i=0; i<SNAPMTXCOMPLEX_rows(genB); i++ ) {
    for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genB,i,0); j<SNAPMTXCOMPLEX_cols(genB); j++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
  }
#elif defined(SNAP_SUPFC_T)
  snap_fcdft_forward2(SNAPMTXCOMPLEX_cols(genTB), SNAPMTXCOMPLEX_rows(genTB), SNAPMTXCOMPLEX_elems(genB), SNAPMTXCOMPLEX_csize(genB), SNAPMTXCOMPLEX_elems(genTB), SNAPMTXCOMPLEX_csize(genTB));
  SNAPBASEREAL_sqrt_ui(sqrtofnm, SNAPMTXCOMPLEX_cols(genTB));
  for ( i=0; i<SNAPMTXCOMPLEX_rows(genB); i++ ) {
    for ( j=0,pt1=SNAPMTXCOMPLEX_elemp(genB,i,0); j<SNAPMTXCOMPLEX_cols(genB); j++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
  }
  SNAPBASEREAL_clear(sqrtofnm);
#else
  ERR;
#endif
#ifdef SNAPLIB_DEBUG
  printf(" ==> genB: \n");
  show_sdmpc(genB);
#endif
}

/*!
 *  @brief This computes the generators of subresultant matrix of f and g with the displacement operator Z_1 Sk - Sk Z_theta.
 *  @param[in] poly_inf The input polynomial f[x] of degree n.
 *  @param[in] poly_ing The input polynomial g[x] of degree m.
 *  @param[in] k The order of subresultant mapping (Sk^t is the Sylvester matrix if k=0).
 *  @param[out] genG G must have dim(n+m-k,2) and zero entries. The generator G of subresultant matrix Sk s.t. Z_1 Sk - Sk Z_theta = GB on exit.
 *  @param[out] genB B must have dim(2,n+m-2k) and zero entries. The generator B of subresultant matrix Sk s.t. Z_1 Sk - Sk Z_theta = GB on exit.
 *  @param[in] theta This defines the right displacement operator Z_theta s.t. Z_1 Sk - Sk Z_theta = GB.
 */
void SNAP_agcd_fastgcd_make_subresultant_generators(SNAP_t poly_inf, SNAP_t poly_ing, unsigned int k, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPBASECOMPLEX_t theta) {
  unsigned int i, n, m;
  SNAPBASECOMPLEX_t *pt1;
  SNAPBASE_t *pt2;

  n = SNAP_degree(poly_inf);
  m = SNAP_degree(poly_ing);
  /* |--- for genB, we just assign two 1. */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genB,0,m-k-1), 1);
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genB,1,n+m-2*k-1), 1);
  /* |--- for genG[0..n+m-k-1,0] */
  for ( i=0, pt1=SNAPMTXCOMPLEX_elems(genG), pt2=SNAP_coefp(poly_ing, m); i<=m; i++, pt1+=SNAPMTXCOMPLEX_csize(genG), pt2-- ) {
    SNAPBASECOMPLEX_neg_base(*pt1, *pt2);
  }
  for ( i=0, pt1=SNAPMTXCOMPLEX_elemp(genG,m-k,0), pt2=SNAP_coefp(poly_inf, n); i<n; i++, pt1+=SNAPMTXCOMPLEX_csize(genG), pt2-- ) {
    SNAPBASECOMPLEX_add_base(*pt1, *pt1, *pt2);
  }
  SNAPBASECOMPLEX_add_base(SNAPMTXCOMPLEX_elem(genG,0,0), SNAPMTXCOMPLEX_elem(genG,0,0), *pt2);
  /* |--- for genG[0..n+m-k-1,1] */
  for ( i=0, pt1=SNAPMTXCOMPLEX_elemp(genG,0,1), pt2=SNAP_coefp(poly_inf, n); i<=n; i++, pt1+=SNAPMTXCOMPLEX_csize(genG), pt2-- ) {
    SNAPBASECOMPLEX_neg_base(*pt1, *pt2);
    SNAPBASECOMPLEX_mul(*pt1, *pt1, theta);
  }
  for ( i=0, pt1=SNAPMTXCOMPLEX_elemp(genG,n-k,1), pt2=SNAP_coefp(poly_ing, m); i<m; i++, pt1+=SNAPMTXCOMPLEX_csize(genG), pt2-- ) {
    SNAPBASECOMPLEX_add_base(*pt1, *pt1, *pt2);
  }
  SNAPBASECOMPLEX_add_base(SNAPMTXCOMPLEX_elem(genG,0,1), SNAPMTXCOMPLEX_elem(genG,0,1), *pt2);
}

/*!
 *  @brief This computes the generators of sylvester matrix of f and g with the displacement operator Z_1 S - S Z_theta.
 *  @param[in] poly_inf The input polynomial f[x] of degree n.
 *  @param[in] poly_ing The input polynomial g[x] of degree m.
 *  @param[out] genG G must have dim(n+m,2) and zero entries. The generator G of Sylvester matrix S s.t. Z_1 S - S Z_theta = GB on exit.
 *  @param[out] genB B must have dim(2,n+m) and zero entries. The generator B of Sylvester matrix S s.t. Z_1 S - S Z_theta = GB on exit.
 *  @param[in] theta This defines the right displacement operator Z_theta s.t. Z_1 S - S Z_theta = GB.
 */
void SNAP_agcd_fastgcd_make_sylvester_generators(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPBASECOMPLEX_t theta) {
  unsigned int i, n, m;
  SNAPBASECOMPLEX_t *pt1, ctmp;

  n = SNAP_degree(poly_inf);
  m = SNAP_degree(poly_ing);
  SNAPBASECOMPLEX_init(ctmp);
  /* |--- for genG, we just assign two 1. */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genG,0,0), 1);
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genG,m,1), 1);
  /* |--- for genB[0,0..n+m-1] */
  for ( i=n-1, pt1=SNAPMTXCOMPLEX_elems(genB); i>0; i--,pt1++ ) {
    SNAPBASECOMPLEX_neg_base(*pt1, SNAP_coef(poly_inf, i));
  }
  SNAPBASECOMPLEX_neg_base(*pt1, SNAP_coef(poly_inf, 0));
  SNAPBASECOMPLEX_add_base(*pt1, *pt1, SNAP_coef(poly_ing, m));
  for ( i=m-1, pt1++; i>0; i--,pt1++ ) {
    SNAPBASECOMPLEX_set_base(*pt1, SNAP_coef(poly_ing, i));
  }
  SNAPBASECOMPLEX_set_base(*pt1, SNAP_coef(poly_ing, 0));
  SNAPBASECOMPLEX_mul_base(ctmp, theta, SNAP_coef(poly_inf, n));
  SNAPBASECOMPLEX_sub(*pt1, *pt1, ctmp);
  /* |--- for genB[1,0..n+m-1] */
  for ( i=m-1, pt1=SNAPMTXCOMPLEX_elemp(genB,1,0); i>0; i--,pt1++ ) {
    SNAPBASECOMPLEX_neg_base(*pt1, SNAP_coef(poly_ing, i));
  }
  SNAPBASECOMPLEX_neg_base(*pt1, SNAP_coef(poly_ing, 0));
  SNAPBASECOMPLEX_add_base(*pt1, *pt1, SNAP_coef(poly_inf, n));
  for ( i=n-1, pt1++; i>0; i--,pt1++ ) {
    SNAPBASECOMPLEX_set_base(*pt1, SNAP_coef(poly_inf, i));
  }
  SNAPBASECOMPLEX_set_base(*pt1, SNAP_coef(poly_inf, 0));
  SNAPBASECOMPLEX_mul_base(ctmp, theta, SNAP_coef(poly_ing, m));
  SNAPBASECOMPLEX_sub(*pt1, *pt1, ctmp);

  SNAPBASECOMPLEX_clear(ctmp);
}

/*!
 *  @brief LCM(n,m).
 *  @param[in] n The given unsigned integer n.
 *  @param[in] m The given unsigned integer m.
 *  @return LCM(n,m).
 */
unsigned int SNAP_agcd_fastgcd_lcm(unsigned int n, unsigned int m) {
  unsigned int nk, mk, rk;
  nk = 0;
  mk = n;
  rk = m;
  while ( rk != 0 ) {
    nk = mk;
    mk = rk;
    rk = nk % mk;
  }
  nk = n/mk;
  return nk*m;
}

/*!
 *  @brief Generate the normalized DFT matrices of order n and m, Dt, D and D1 for Toeplitz-like matrix.
 *  @param[in] n The number of rows in the target Toeplitz-like matrix.
 *  @param[in] m The number of columns in the target Toeplitz-like matrix.
 *  @param[out] vecDt The complex vector of length m. i-th element is E^(2Pi*I*(i-1)/m) * theta on exit.
 *  @param[out] vecDh The complex vector of length m. i-th element is the conjugate of theta^((i-1)/m) on exit.
 *  @param[out] vecD1 The complex vector of length n. i-th element is E^(2Pi*I*(i-1)/n) on exit.
 *  @param[in] theta The complex number: E^(Pi*I/LCM(n,m)).
 */
void SNAP_agcd_fastgcd_make_trans_matrices(unsigned int n, unsigned int m, SNAPMTXCOMPLEX_t vecDt, SNAPMTXCOMPLEX_t vecDh, SNAPMTXCOMPLEX_t vecD1, SNAPBASECOMPLEX_t theta) {
  unsigned int j;
  SNAPBASEREAL_t rsqn, rsqm;
  SNAPBASECOMPLEX_t *pt1, *pt2, thetatilde;

  SNAPBASECOMPLEX_init(thetatilde);
  // rsqn and rsqm <== 1/sqrt(n) and 1/sqrt(m)
  SNAPBASEREAL_init(rsqn);
  SNAPBASEREAL_sqrt_ui(rsqn, n);
  SNAPBASEREAL_rec(rsqn, rsqn);
  SNAPBASEREAL_init(rsqm);
  SNAPBASEREAL_sqrt_ui(rsqm, m);
  SNAPBASEREAL_rec(rsqm, rsqm);
  // D1 and matFn using D1
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  pt1 = SNAPMTXCOMPLEX_elems(vecD1);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  for ( j=1,pt1++; j<n; j++,pt1++ ) {
    *pt1 = cexp(2*j*3.1415926535897932384626433832795029L*I/n);
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASECOMPLEX_t const_pin, const_tmp;
  SNAPBASECOMPLEX_init(const_pin);
  SNAPBASECOMPLEX_init(const_tmp);
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), n);
  pt1 = SNAPMTXCOMPLEX_elems(vecD1);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  for ( j=1,pt1++; j<n; j++,pt1++ ) {
    SNAPBASECOMPLEX_mul_si(const_tmp, const_pin, j);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
#else
  pt1 = SNAPMTXCOMPLEX_elems(vecD1);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  ERR;
#endif
  // Dt and matFm using Dt
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  pt1 = SNAPMTXCOMPLEX_elems(vecDt);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  for ( j=1,pt1++; j<m; j++,pt1++ ) {
    *pt1 = cexp(2*j*3.1415926535897932384626433832795029L*I/m);
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), m);
  pt1 = SNAPMTXCOMPLEX_elems(vecDt);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  for ( j=1,pt1++; j<m; j++,pt1++ ) {
    SNAPBASECOMPLEX_mul_si(const_tmp, const_pin, j);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
#else
  pt1 = SNAPMTXCOMPLEX_elems(vecDt);
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  ERR;
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  thetatilde = cpow(theta, (complex double) 1/m);
  for ( j=0,pt1=SNAPMTXCOMPLEX_elems(vecDt); j<m; j++,pt1++ ) {
    *pt1 *= thetatilde;
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASECOMPLEX_set_ui(const_tmp, m);
  SNAPBASECOMPLEX_ui_div(const_tmp, 1, const_tmp);
  SNAPBASECOMPLEX_pow(thetatilde, theta, const_tmp);
  for ( j=0,pt1=SNAPMTXCOMPLEX_elems(vecDt); j<m; j++,pt1++ ) {
    SNAPBASECOMPLEX_mul(*pt1, *pt1, thetatilde);
  }
  SNAPBASECOMPLEX_clear(const_pin);
  SNAPBASECOMPLEX_clear(const_tmp);
#else
  ERR;
#endif
  // D_i : theta^((i-1)/m) => Dh_i = conj(D_i)
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  // ACTUAL
  SNAPMTXCOMPLEX_elem(vecDh,0,0) = 1.0;
  for ( j=1,pt1=SNAPMTXCOMPLEX_elemp(vecDh,0,1),pt2=SNAPMTXCOMPLEX_elemp(vecDh,0,0); j<m; j++,pt1++,pt2++ ) {
    *pt1 = (*pt2)*(thetatilde);
  }
  for ( j=1,pt1=SNAPMTXCOMPLEX_elemp(vecDh,0,1); j<m; j++,pt1++ ) {
    *pt1 = conj(*pt1);
  }
#elif defined(SNAP_SUPFC_T)
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(vecDh,0,0), 1);
  for ( j=1,pt1=SNAPMTXCOMPLEX_elemp(vecDh,0,1),pt2=SNAPMTXCOMPLEX_elemp(vecDh,0,0); j<m; j++,pt1++,pt2++ ) {
    SNAPBASECOMPLEX_mul(*pt1, *pt2, thetatilde);
  }
  for ( j=1,pt1=SNAPMTXCOMPLEX_elemp(vecDh,0,1); j<m; j++,pt1++ ) {
    SNAPBASECOMPLEX_conj(*pt1, *pt1);
  }
#else
  ERR;
#endif
  SNAPBASEREAL_clear(rsqn);
  SNAPBASEREAL_clear(rsqm);
  SNAPBASECOMPLEX_clear(thetatilde);
}

/*!
 *  @brief This computes the generators of Jacobian matrix J of d, f1 and g1 with the displacement operator Z_1 J - J Z_theta.
 *  @param[in] poly_d The input polynomial d[x] of degree k.
 *  @param[in] poly_f1 The input polynomial f1[x] of degree n-k.
 *  @param[in] poly_g1 The input polynomial g1[x] of degree m-k.
 *  @param[out] genG G must have dim(n+m+3,5) and zero entries. The generator G of Jacobian matrix J s.t. Z_1 J - J Z_theta = GB on exit.
 *  @param[out] genB B must have dim(5,n+m-k+3) and zero entries. The generator B of Jacobian matrix J s.t. Z_1 J - J Z_theta = GB on exit.
 *  @param[in] theta This defines the right displacement operator Z_theta s.t. Z_1 J - J Z_theta = GB.
 */
void SNAP_agcd_fastgcd_make_jacobian_generators(SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPBASECOMPLEX_t theta) {
  unsigned int i, n, m, k;
  SNAPBASECOMPLEX_t *pt1;
  SNAPBASE_t *pt2;
  k = SNAP_degree(poly_d);
  n = SNAP_degree(poly_f1)+k;
  m = SNAP_degree(poly_g1)+k;
  /* genG: -d,f1,0,g1 */
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genG,0,0),pt2=SNAP_coefp(poly_d,k); i<=k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  for ( i=0,pt2=SNAP_coefp(poly_f1,n-k); i<=n-k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  for ( i=0,pt1+=k*SNAPMTXCOMPLEX_csize(genG),pt2=SNAP_coefp(poly_g1,m-k); i<=m-k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  /* genG: 0(n-k+1),d(k:1),d(0)-d(k),-d(k-1:0),0(m-k+1) */
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genG,n-k+1,1),pt2=SNAP_coefp(poly_d,k); i<k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  pt2=SNAP_coefp(poly_d,k);
  SNAPBASECOMPLEX_sub_base(*pt1, *pt1, *pt2);
  for ( i=0,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2--; i<k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  /* genG: -t*f1,0(k),-t*g1,d */
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genG,0,2),pt2=SNAP_coefp(poly_f1,n-k); i<=n-k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_mul(*pt1, *pt2, theta);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genG,n+1,2),pt2=SNAP_coefp(poly_g1,m-k); i<=m-k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_mul(*pt1, *pt2, theta);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  for ( i=0,pt2=SNAP_coefp(poly_d,k); i<=k; i++,pt1+=SNAPMTXCOMPLEX_csize(genG),pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  /* genG: 1,0(n+m+2) */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genG,0,3), 1);
  /* genG: 0(m+m+2),1 */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genG,n+m+2,4), 1);
  /* genB: 0(k),1,0(n+m-2k+2) */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genB,0,k), 1);
  /* genB: 0(n+1),1,0(m-k+1) */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genB,1,n+1), 1);
  /* genB: 0(n+m-k+2),1 */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genB,2,n+m-k+2), 1);
  /* genB: d,f1,g1 ==> d^c,-f1^c,-g1^c */
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genB,3,0),pt2=SNAP_coefp(poly_d,k); i<=k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
  }
  for ( i=0,pt2=SNAP_coefp(poly_f1,n-k); i<=n-k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  for ( i=0,pt2=SNAP_coefp(poly_g1,m-k); i<=m-k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  /* genB: -d(k-1:0),-f1,-g1,-t*d(k) ==> -d(k-1:0)^c,f1^c,g1^c,-t*(d(k)^c) */
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(genB,4,0),pt2=SNAP_coefp(poly_d,k-1); i<k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
    SNAPBASECOMPLEX_neg(*pt1, *pt1);
  }
  for ( i=0,pt2=SNAP_coefp(poly_f1,n-k); i<=n-k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
  }
  for ( i=0,pt2=SNAP_coefp(poly_g1,m-k); i<=m-k; i++,pt1++,pt2-- ) {
    SNAPBASECOMPLEX_conj_base(*pt1, *pt2);
  }
  SNAPBASECOMPLEX_conj_base(*pt1, SNAP_coef(poly_d,k));
  SNAPBASECOMPLEX_mul(*pt1, *pt1, theta);
  SNAPBASECOMPLEX_neg(*pt1, *pt1);
}

/*
 * find a candidate of local minimum of
 *  func = |poly_fA - alpha*poly_fB - alpha^2 poly_fC| + |poly_gA - alpha*poly_gB - alpha^2 poly_gC|
 * by Newton-Raphson methd w.r.t. alpha.
 */
void SNAP_agcd_fastgcd_linesearch(SNAP_t poly_fA, SNAP_t poly_fB, SNAP_t poly_fC, SNAP_t poly_gA, SNAP_t poly_gB, SNAP_t poly_gC, SNAPBASEREAL_t *alpha) {
  // func ==> (fA-a*fB-a^2*fC).conj(fA-a*fB-a^2*fC) + (gA-a*gB-a^2*gC).conj(gA-a*gB-a^2*gC)
  //        = sum (fA_i-a*fB_i-a^2*fC_i)*conj(fA_i-a*fB_i-a^2*fC_i)
  //         +sum (gA_i-a*gB_i-a^2*gC_i)*conj(gA_i-a*gB_i-a^2*gC_i)
  //        = sum (fA_i-a*fB_i-a^2*fC_i)*conj(fA_i-a*fB_i-a^2*fC_i)
  //         +sum (gA_i-a*gB_i-a^2*gC_i)*conj(gA_i-a*gB_i-a^2*gC_i)
  //        = (fC.conj(fC)+gC.conj(gC)) a^4 + (fB.conj(fC)+fC.conj(fB)+gB.conj(gC)+gC.conj(gB)) a^3
  //         +(fB.conj(fB)-fA.conj(fC)-fC.conj(fA)+gB.conj(gB)-gA.conj(gC)-gC.conj(gA)) a^2
  //         +(-fA.conj(fB)-fB.conj(fA)-gA.conj(gB)-gB.conj(gA)) a + (fA.conj(fA)+gA.conj(gA))
  unsigned int n,m,iterations=0;
  unsigned int const maxiterations=3;
  SNAPBASE_t c0, c1, c2, c3, c4, dc2, dc3, dc4, ctmp, num, den;
  SNAPBASEREAL_t zero, five, newalpha;
  // initialization
  n = SNAP_degree(poly_fA)+1;
  m = SNAP_degree(poly_gA)+1;
  SNAPBASE_init(c0);
  SNAPBASE_init(c1);
  SNAPBASE_init(c2);
  SNAPBASE_init(c3);
  SNAPBASE_init(c4);
  SNAPBASE_init(dc2);
  SNAPBASE_init(dc3);
  SNAPBASE_init(dc4);
  SNAPBASE_init(ctmp);
  SNAPBASEREAL_init_set_ui(zero,0);
  SNAPBASEREAL_init_set_ui(five,5);
  SNAPBASE_init(num);
  SNAPBASE_init(den);
  SNAPBASEREAL_init_set(newalpha, *alpha);
  // fA.conj(fA)+gA.conj(gA)
  SNAPMTX_BLAS_dotc(c0,n,SNAP_coefs(poly_fA),1,SNAP_coefs(poly_fA),1);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gA),1,SNAP_coefs(poly_gA),1);
  SNAPBASE_add(c0, c0, ctmp);
  // -fA.conj(fB)-fB.conj(fA)-gA.conj(gB)-gB.conj(gA)
  SNAPMTX_BLAS_dotc(c1,n,SNAP_coefs(poly_fB),1,SNAP_coefs(poly_fA),1);
  SNAPMTX_BLAS_dotc(ctmp,n,SNAP_coefs(poly_fA),1,SNAP_coefs(poly_fB),1);
  SNAPBASE_add(c1, c1, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gB),1,SNAP_coefs(poly_gA),1);
  SNAPBASE_add(c1, c1, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gA),1,SNAP_coefs(poly_gB),1);
  SNAPBASE_add(c1, c1, ctmp);
  SNAPBASE_neg(c1, c1);
  // fB.conj(fB)-fA.conj(fC)-fC.conj(fA)+gB.conj(gB)-gA.conj(gC)-gC.conj(gA)
  SNAPMTX_BLAS_dotc(c2,n,SNAP_coefs(poly_fB),1,SNAP_coefs(poly_fB),1);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gB),1,SNAP_coefs(poly_gB),1);
  SNAPBASE_add(c2, c2, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,n,SNAP_coefs(poly_fC),1,SNAP_coefs(poly_fA),1);
  SNAPBASE_sub(c2, c2, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,n,SNAP_coefs(poly_fA),1,SNAP_coefs(poly_fC),1);
  SNAPBASE_sub(c2, c2, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gC),1,SNAP_coefs(poly_gA),1);
  SNAPBASE_sub(c2, c2, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gA),1,SNAP_coefs(poly_gC),1);
  SNAPBASE_sub(c2, c2, ctmp);
  // fB.conj(fC)+fC.conj(fB)+gB.conj(gC)+gC.conj(gB)
  SNAPMTX_BLAS_dotc(c3,n,SNAP_coefs(poly_fC),1,SNAP_coefs(poly_fB),1);
  SNAPMTX_BLAS_dotc(ctmp,n,SNAP_coefs(poly_fB),1,SNAP_coefs(poly_fC),1);
  SNAPBASE_add(c3, c3, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gC),1,SNAP_coefs(poly_gB),1);
  SNAPBASE_add(c3, c3, ctmp);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gB),1,SNAP_coefs(poly_gC),1);
  SNAPBASE_add(c3, c3, ctmp);
  // fC.conj(fC)+gC.conj(gC)
  SNAPMTX_BLAS_dotc(c4,n,SNAP_coefs(poly_fC),1,SNAP_coefs(poly_fC),1);
  SNAPMTX_BLAS_dotc(ctmp,m,SNAP_coefs(poly_gC),1,SNAP_coefs(poly_gC),1);
  SNAPBASE_add(c4, c4, ctmp);
  // differential
  SNAPBASE_mul_ui(dc4, c4, 4);
  SNAPBASE_mul_ui(dc3, c3, 3);
  SNAPBASE_mul_ui(dc2, c2, 2);
  // now we have : c4*a^4 + c3*a^3 + c2*a^2 + c1*a + c0
  // f(a) = (((c4*a + c3)*a + c2)*a + c1)*a + c0
  // f'(a) = ((4*c4*a + 3*c3)*a + 2*c2)*a + c1 = ((dc4*a + dc3)*a + dc2)*a + c1
  // a_{i+1} = a_i - f(a)/f'(a)
  while ( iterations++ < maxiterations ) {
    // f(a)
    SNAPBASE_mul_basereal(num, c4, *alpha);
    SNAPBASE_add(num, num, c3);
    SNAPBASE_mul_basereal(num, num, *alpha);
    SNAPBASE_add(num, num, c2);
    SNAPBASE_mul_basereal(num, num, *alpha);
    SNAPBASE_add(num, num, c1);
    SNAPBASE_mul_basereal(num, num, *alpha);
    SNAPBASE_add(num, num, c0);
    // f'(a)
    SNAPBASE_mul_basereal(den, dc4, *alpha);
    SNAPBASE_add(den, den, dc3);
    SNAPBASE_mul_basereal(den, den, *alpha);
    SNAPBASE_add(den, den, dc2);
    SNAPBASE_mul_basereal(den, den, *alpha);
    SNAPBASE_add(den, den, c1);
    if ( SNAPBASE_zeroq(den) ) {
      break;
    }
    // correctional term
    SNAPBASE_div(ctmp, num, den);
    SNAPBASE_set_basecomplex(newalpha, ctmp);
    SNAPBASE_sub(newalpha, *alpha, newalpha);
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
    mpfr_printf("alpha: %Re\n", newalpha);
#else
    printf("alpha: %e\n", newalpha);
#endif // SNAP_SUPFC_T
#endif
    if ( SNAPBASEREAL_cmp(zero,newalpha) > 0 || SNAPBASEREAL_cmp(newalpha,five) > 0 ) {
      break;
    } else {
      SNAPBASEREAL_set(*alpha, newalpha);
    }
  }
  SNAPBASE_clear(c0);
  SNAPBASE_clear(c1);
  SNAPBASE_clear(c2);
  SNAPBASE_clear(c3);
  SNAPBASE_clear(c4);
  SNAPBASE_clear(dc2);
  SNAPBASE_clear(dc3);
  SNAPBASE_clear(dc4);
  SNAPBASE_clear(ctmp);
  SNAPBASEREAL_clear(zero);
  SNAPBASEREAL_clear(five);
  SNAPBASEREAL_clear(newalpha);
  SNAPBASE_clear(num);
  SNAPBASE_clear(den);
}

/*!
 *  @brief refinement routine proposed in Bini and Boito.
 *  @param[in] poly_f The input polynomial f[x].
 *  @param[in] poly_g The input polynomial g[x].
 *  @param[in,out] poly_d The tentative approximate GCD on input, and the resulting refinment on exist.
 *  @param[in,out] poly_f1 The tentative approximate cofactor on input, and the resulting refinment on exist.
 *  @param[in,out] poly_g1 The tentative approximate cofactor on input, and the resulting refinment on exist.
 *  @param[in] vecDt The complex vector for temporal use.
 *  @param[in] vecDh The complex vector for temporal use.
 *  @param[in] vecD1 The complex vector for temporal use.
 *  @param[in] genG The complex vector for temporal use.
 *  @param[in] genB The complex vector for temporal use.
 *  @param[in] matW The complex array for temporal use.
 *  @param[in] genTG The complex vector for temporal use.
 *  @param[in] genTB The complex vector for temporal use.
 *  @param[in] matP The 1d-array for temporal use (must be allocated).
 *  @param[in] matQ The 1d-array for temporal use (must be allocated).
 */
void SNAP_agcd_fastgcd_refine(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXCOMPLEX_t vecDt, SNAPMTXCOMPLEX_t vecDh, SNAPMTXCOMPLEX_t vecD1, SNAPMTXCOMPLEX_t genG, SNAPMTXCOMPLEX_t genB, SNAPMTXCOMPLEX_t matW, SNAPMTXCOMPLEX_t genTG, SNAPMTXCOMPLEX_t genTB, unsigned int *matP, unsigned int *matQ) {
  unsigned int i, n, m, k, k0, iterations=0;
  unsigned int *pi1, *pi2;
  int *ip, *dummyP;
  // Boito's setting.
  unsigned int const maxiterations=20;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPD_T)
  fftw_plan fftp;
#endif // defined
  SNAPBASE_t *ptb;
  SNAPBASEREAL_t sqrtofnm, zero, newton_mintol, red, red_old, alpha;
  SNAPBASECOMPLEX_t tau, c_tmp, *pt1, *pt2, *pt3, theta;
  SNAPMTXCOMPLEX_t fz, fz_tmp;
  SNAP_t poly_newf1, poly_newg1, poly_newd, poly_tmp;
  SNAP_t poly_fA, poly_fB, poly_fC, poly_gA, poly_gB, poly_gC;
  // initialization
  n = SNAP_degree(poly_f);
  m = SNAP_degree(poly_g);
  k = SNAP_degree(poly_d);
  SNAPBASEREAL_init_set_ui(zero, 0);
  SNAPBASECOMPLEX_init(theta);
  SNAP_init_set(poly_newf1, poly_f1);
  SNAP_init_set(poly_newg1, poly_g1);
  SNAP_init_set(poly_newd, poly_d);
  SNAP_init(poly_tmp);
  SNAPMTXCOMPLEX_init2(fz,1,n+m+3);
  SNAPMTXCOMPLEX_init2(fz_tmp,1,n+m+3);
  SNAP_init(poly_fA);
  SNAP_init(poly_fB);
  SNAP_init(poly_fC);
  SNAP_init(poly_gA);
  SNAP_init(poly_gB);
  SNAP_init(poly_gC);
  SNAPBASEREAL_init(sqrtofnm);
  SNAPBASEREAL_init(alpha);
  SNAPBASECOMPLEX_init(tau);
  SNAPBASECOMPLEX_init(c_tmp);
  SNAPBASEREAL_init(red);
  SNAPBASEREAL_init(red_old);
  SNAPBASEREAL_init(newton_mintol);
  dummyP = (int*) malloc(sizeof(int)*(n+m+3));
  for ( i=0, ip=dummyP; i<n+m+3; i++,ip++ ) {
    *ip=0;
  }
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPD_T)
  newton_mintol = 1.0e-15;
#elif defined(SNAP_SUPFC_T)
  SNAPBASE_prec_t dprec;
  dprec = SNAPBASE_get_default_prec();
  SNAPBASEREAL_ui_pow_ui(newton_mintol, 2, (unsigned long int) dprec);
  SNAPBASEREAL_ui_div(newton_mintol, 1, newton_mintol);
#else
  ERR;
#endif
  // residual computation and fix tau
  SNAP_mul(poly_tmp, poly_f1, poly_d);
  SNAP_sub(poly_tmp, poly_tmp, poly_f);
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,0),ptb=SNAP_coefp(poly_tmp,n); i<=n; i++,pt1++,ptb-- ) {
    SNAPBASECOMPLEX_set_base(*pt1,*ptb);
  }
  SNAP_mul(poly_tmp, poly_g1, poly_d);
  SNAP_sub(poly_tmp, poly_tmp, poly_g);
  for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,n+1),ptb=SNAP_coefp(poly_tmp,m); i<=m; i++,pt1++,ptb-- ) {
    SNAPBASECOMPLEX_set_base(*pt1,*ptb);
  }
  SNAPMTX_BLAS_dotc(c_tmp, k+1, SNAP_coefs(poly_d), 1, SNAP_coefs(poly_d), 1);
  SNAPBASECOMPLEX_set(tau,c_tmp);
  SNAPMTX_BLAS_dotc(c_tmp, n-k+1, SNAP_coefs(poly_f1), 1, SNAP_coefs(poly_f1), 1);
  SNAPBASECOMPLEX_sub(tau,tau,c_tmp);
  SNAPMTX_BLAS_dotc(c_tmp, m-k+1, SNAP_coefs(poly_g1), 1, SNAP_coefs(poly_g1), 1);
  SNAPBASECOMPLEX_sub(tau,tau,c_tmp);
  SNAPMTXCOMPLEX_BLAS_nrm2(red, n+m+2, SNAPMTXCOMPLEX_elems(fz), 1);
  SNAPBASEREAL_set(red_old, red);
#ifdef SNAPLIB_DEBUG
  printf("len:%d(%d), %d(%d), %d(%d), %d(%d), %d(%d)\n",SNAP_degree(poly_f), SNAP_size(poly_f), SNAP_degree(poly_g), SNAP_size(poly_g),SNAP_degree(poly_d), SNAP_size(poly_d), SNAP_degree(poly_f1), SNAP_size(poly_f1), SNAP_degree(poly_g1), SNAP_size(poly_g1));
#ifdef SNAP_SUPFC_T
  mpfr_printf("iteration: %d, red: %Re\n", iterations, red);
#else
  printf("iteration: %d, red: %e\n", iterations, red);
#endif // SNAP_SUPFC_T
#endif
  // while ( residual not is tiny and the counter < the max )
  while ( SNAPBASEREAL_cmp(red, newton_mintol) > 0 && iterations++ < maxiterations ) {
    // matrix construction
    // matrix construction: Jacobian generators
    SNAPMTXCOMPLEX_reinit3(vecDt, 1, n+m-k+3);
    SNAPMTXCOMPLEX_reinit3(vecDh, 1, n+m-k+3);
    SNAPMTXCOMPLEX_reinit3(vecD1, 1, n+m+3);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    theta = cexp(3.1415926535897932384626433832795029L*I*(n+m-k+3)/SNAP_agcd_fastgcd_lcm(n+m+3,n+m-k+3));
    //theta = cexp(3.1415926535897932384626433832795029L*I/SNAP_agcd_fastgcd_lcm(n+m+3,n+m-k+3));
#elif defined(SNAP_SUPFC_T)
    SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(theta), 0);
    SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(theta));
    SNAPBASECOMPLEX_mul_si(theta, theta, (n+m-k+3)/SNAP_agcd_fastgcd_lcm(n+m+3,n+m-k+3));
    SNAPBASECOMPLEX_exp(theta, theta);
#else
    ERR;
#endif
    SNAP_agcd_fastgcd_make_trans_matrices(n+m+3, n+m-k+3, vecDt, vecDh, vecD1, theta);
    SNAPMTXCOMPLEX_reinit3(genG, n+m+3, 5);
    SNAPMTXCOMPLEX_reinit3(genTG, n+m+3, 5);
    SNAPMTXCOMPLEX_reinit3(genB, 5, n+m-k+3);
    SNAPMTXCOMPLEX_reinit3(genTB, 5, n+m-k+3);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    SNAP_agcd_fastgcd_make_jacobian_generators(poly_d, poly_f1, poly_g1, genG, genB, theta);
#else
    ERR;
#endif
    SNAP_agcd_fastgcd_make_cauchy_generators(genG, genB, genTG, genTB, vecDh);
    for ( i=0, pi1=matP, pi2=matQ; i<n+m+3; i++,pi1++,pi2++ ) {
      *pi1=0;
      *pi2=0;
    }
    // matrix construction: GKO
#ifdef SNAPLIB_DEBUG
    printf("modified_gko called with eps=%e\n", 0.0);
#endif
    SNAPMTXCOMPLEX_reinit3(matW, n+m+3, n+m-k+3);
    k0 = SNAP_agcd_fastgcd_modified_gko(vecD1, vecDt, genG, genB, matW, matP, matQ, zero, n+m-k+3);
    if ( k0 < n+m-k+3 ) {
      printf("ERROR in modified_gko hence we could not refine further.\n");
      goto post;
    }
#ifdef SNAPLIB_DEBUG
    printf("matW(k0=%d): \n",k0);
    show_sdmpc(matW);
#ifdef SNAP_SUPDC_T
//	for ( i=0, pi1=matP, pi2=matQ; i<n+m+3; i++,pi1++,pi2++ ) { if ( *pi1+*pi2>0 ) { printf("matP[%d]=%d, matQ[%d]=%d\n",i,*pi1,i,*pi2); } }
#endif
#endif
    // compute correction vector
    // compute correction vector: fz_tmp <== Fn fz
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    // FFTW
    fftp = fftw_plan_dft_1d(n+m+3, SNAPMTXCOMPLEX_elems(fz), SNAPMTXCOMPLEX_elems(fz_tmp), FFTW_BACKWARD, FFTW_ESTIMATE);
    fftw_execute(fftp);
    fftw_destroy_plan(fftp);
    sqrtofnm = sqrt(n+m+3);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elems(fz_tmp); i<n+m+3; i++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
#elif defined(SNAP_SUPFC_T)
    snap_fcdft_backward(n+m+3, 1, SNAPMTXCOMPLEX_elems(fz_tmp), 1, SNAPMTXCOMPLEX_elems(fz), 1);
    SNAPBASEREAL_sqrt_ui(sqrtofnm, n+m+3);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elems(fz_tmp); i<n+m+3; i++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
#else
    ERR;
#endif
    // compute correction vector: fz_tmp <== P fz_tmp
    //for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m+2),pi1=matP+n+m+2; i<n+m+3; i++,pt1--,pi1-- ) {
    //	if ( *pi1 != 0 ) { SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(fz_tmp,0,*pi1-1)); }
    //}
    // compute correction vector: fz_tmp <== P^-1 fz_tmp
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,0),pi1=matP; i<n+m+3; i++,pt1++,pi1++ ) {
      if ( *pi1 != 0 ) {
        SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(fz_tmp,0,(*pi1)-1));
      }
    }
    // compute correction vector: fz_tmp <== U^-1L^-1 fz_tmp
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'L', 'N', 'U', k0-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elems(fz_tmp), 1) != 0 ) {
      printf("ERROR in ztrtrs hence we could not refine further.\n");
      goto post;
    }
    if ( LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', k0-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elems(fz_tmp), 1) != 0 ) {
      printf("ERROR in ztrtrs hence we could not refine further.\n");
      goto post;
    }
#elif defined(SNAP_SUPFC_T)
    if ( snap_mlapack_fctrtrs('L', 'N', 'U', k0-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elems(fz_tmp), 1) != 0 ) {
      printf("ERROR in snap_mlapack_fctrtrs hence we could not refine further.\n");
      goto post;
    }
    if ( snap_mlapack_fctrtrs('U', 'N', 'N', k0-1, 1, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW), SNAPMTXCOMPLEX_elems(fz_tmp), 1) != 0 ) {
      printf("ERROR in snap_mlapack_fctrtrs hence we could not refine further.\n");
      goto post;
    }
#else
    ERR;
#endif
    // compute correction vector: fz_tmp(k0-1:n+m-k+2)=0
    for ( i=k0-1,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,k0-1); i<n+m-k+3; i++,pt1++ ) {
      SNAPBASECOMPLEX_set_ui(*pt1,0);
    }
    // compute correction vector: fz_tmp <== Q fz_tmp
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2),pi1=matQ+n+m-k+2; i<n+m-k+3; i++,pt1--,pi1-- ) {
      if ( *pi1 != 0 ) {
        SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(fz_tmp,0,*pi1-1));
      }
    }
    // compute correction vector: fz_tmp <== Q^t fz_tmp
    //for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,0),pi1=matQ; i<n+m-k+3; i++,pt1++,pi1++ ) {
    //	if ( *pi1 != 0 ) { SNAPBASECOMPLEX_swap(*pt1, SNAPMTXCOMPLEX_elem(fz_tmp,0,*pi1-1)); }
    //}
    // compute correction vector: fz <== Fm^H fz_tmp
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    // FFTW
    fftp = fftw_plan_dft_1d(n+m-k+3, SNAPMTXCOMPLEX_elems(fz_tmp), SNAPMTXCOMPLEX_elems(fz), FFTW_FORWARD, FFTW_ESTIMATE);
    fftw_execute(fftp);
    fftw_destroy_plan(fftp);
    sqrtofnm = sqrt(n+m-k+3);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elems(fz); i<n+m-k+3; i++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
#elif defined(SNAP_SUPFC_T)
    snap_fcdft_forward(n+m-k+3, 1, SNAPMTXCOMPLEX_elems(fz), 1, SNAPMTXCOMPLEX_elems(fz_tmp), 1);
    SNAPBASEREAL_sqrt_ui(sqrtofnm, n+m-k+3);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elems(fz); i<n+m-k+3; i++,pt1++ ) {
      SNAPBASECOMPLEX_div_basereal(*pt1, *pt1, sqrtofnm);
    }
#else
    ERR;
#endif
    // compute correction vector: fz_tmp <== Dh fz
    for ( i=0,pt1=SNAPMTXCOMPLEX_elems(fz),pt2=SNAPMTXCOMPLEX_elems(vecDh),pt3=SNAPMTXCOMPLEX_elems(fz_tmp); i<n+m-k+3; i++,pt1++,pt2++,pt3++ ) {
      SNAPBASECOMPLEX_mul(*pt3, *pt1, *pt2);
    }
    // if the Jacobian is rank deficient.
    if ( k0 == n+m-k+3 ) {
#ifdef SNAPLIB_DEBUG
      printf("Jacobian is rank deficient.\n");
#endif
      // compute correction vector: null vector
      SNAP_agcd_fastgcd_nullvector(fz, matW, matQ, vecDh, zero);
    } else {
      // compute correction vector: theoretical null vector
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,k),ptb=SNAP_coefs(poly_d); i<=k; i++,pt1--,ptb++ ) {
        SNAPBASECOMPLEX_set_base(*pt1, *ptb);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,n+1),ptb=SNAP_coefs(poly_f1); i<=n-k; i++,pt1--,ptb++ ) {
        SNAPBASECOMPLEX_neg_base(*pt1, *ptb);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,n+m-k+2),ptb=SNAP_coefs(poly_g1); i<=m-k; i++,pt1--,ptb++ ) {
        SNAPBASECOMPLEX_neg_base(*pt1, *ptb);
      }
    }
    // compute correction vector: projected vector and subtraction
    SNAPMTXCOMPLEX_BLAS_dotc(c_tmp, n+m-k+3,SNAPMTXCOMPLEX_elems(fz), 1, SNAPMTXCOMPLEX_elems(fz), 1);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(matW,n+m+2,0),pt2=SNAPMTXCOMPLEX_elems(fz); i<n+m-k+3; i++,pt1++,pt2++ ) {
      SNAPBASECOMPLEX_div(*pt1, *pt2, c_tmp);
    }
    SNAPMTXCOMPLEX_BLAS_dotc(c_tmp, n+m-k+3, SNAPMTXCOMPLEX_elems(fz), 1, SNAPMTXCOMPLEX_elems(fz_tmp), 1);
    SNAPBASECOMPLEX_neg(c_tmp, c_tmp);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_zaxpy(n+m-k+3, &c_tmp, SNAPMTXCOMPLEX_elemp(matW,n+m+2,0), 1, SNAPMTXCOMPLEX_elems(fz_tmp), 1);
#elif defined(SNAP_SUPFC_T)
    snap_mblas_fcaxpy(n+m-k+3, c_tmp, SNAPMTXCOMPLEX_elemp(matW,n+m+2,0), 1, SNAPMTXCOMPLEX_elems(fz_tmp), 1);
#else
    ERR;
#endif
#ifdef SNAPLIB_DEBUG
    printf("correction vector:\n");
    show_sdmpc(fz_tmp);
#endif
    // update the temporay new polys
    SNAPBASECOMPLEX_set_si(c_tmp, -1);
#if defined(SNAP_SUPDC_T)
    // BLAS_INCREMENT_STYLE
    // cblas_zaxpy(k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k), -1, SNAP_coefp(poly_newd,0), 1);
    cblas_zaxpy(k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,0), -1, SNAP_coefp(poly_newd,0), 1);
    // cblas_zaxpy(n-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1), -1, SNAP_coefp(poly_newf1,0), 1);
    cblas_zaxpy(n-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k+1), -1, SNAP_coefp(poly_newf1,0), 1);
    // cblas_zaxpy(m-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2), -1, SNAP_coefp(poly_newg1,0), 1);
    cblas_zaxpy(m-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+2), -1, SNAP_coefp(poly_newg1,0), 1);
#elif defined(SNAP_SUPD_T)
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,k),ptb=SNAP_coefs(poly_newd); i<=k; i++,pt1--,ptb++ ) {
      SNAPBASE_sub_basecomplex(*ptb,*ptb,*pt1);
    }
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1),ptb=SNAP_coefs(poly_newf1); i<=n-k; i++,pt1--,ptb++ ) {
      SNAPBASE_sub_basecomplex(*ptb,*ptb,*pt1);
    }
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2),ptb=SNAP_coefs(poly_newg1); i<=m-k; i++,pt1--,ptb++ ) {
      SNAPBASE_sub_basecomplex(*ptb,*ptb,*pt1);
    }
#elif defined(SNAP_SUPFC_T)
    snap_mblas_fcaxpy(k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k), -1, SNAP_coefp(poly_newd,0), 1);
    snap_mblas_fcaxpy(n-k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1), -1, SNAP_coefp(poly_newf1,0), 1);
    snap_mblas_fcaxpy(m-k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2), -1, SNAP_coefp(poly_newg1,0), 1);
#else
    ERR;
#endif
#ifdef SNAPLIB_DEBUG
    printf("poly_newd: \n");
    show_plist(SNAP_coefs(poly_newd), SNAP_degree(poly_newd)+1);
    printf("poly_newf1: \n");
    show_plist(SNAP_coefs(poly_newf1), SNAP_degree(poly_newf1)+1);
    printf("poly_newg1: \n");
    show_plist(SNAP_coefs(poly_newg1), SNAP_degree(poly_newg1)+1);
#endif
    // residual computation and fix tau
    SNAP_mul(poly_tmp, poly_newf1, poly_newd);
    SNAP_sub(poly_tmp, poly_tmp, poly_f);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,0),ptb=SNAP_coefp(poly_tmp,n); i<=n; i++,pt1++,ptb-- ) {
      SNAPBASECOMPLEX_set(*pt1,*ptb);
    }
    SNAP_mul(poly_tmp, poly_newg1, poly_newd);
    SNAP_sub(poly_tmp, poly_tmp, poly_g);
    for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,n+1),ptb=SNAP_coefp(poly_tmp,m); i<=m; i++,pt1++,ptb-- ) {
      SNAPBASECOMPLEX_set(*pt1,*ptb);
    }
    SNAPMTX_BLAS_dotc(c_tmp, k+1, SNAP_coefs(poly_newd), 1, SNAP_coefs(poly_newd), 1);
    SNAPBASECOMPLEX_sub(*pt1,c_tmp,tau);
    SNAPMTX_BLAS_dotc(c_tmp, n-k+1, SNAP_coefs(poly_newf1), 1, SNAP_coefs(poly_newf1), 1);
    SNAPBASECOMPLEX_sub(*pt1,*pt1,c_tmp);
    SNAPMTX_BLAS_dotc(c_tmp, m-k+1, SNAP_coefs(poly_newg1), 1, SNAP_coefs(poly_newg1), 1);
    SNAPBASECOMPLEX_sub(*pt1,*pt1,c_tmp);
    SNAPMTXCOMPLEX_BLAS_nrm2(red, n+m+2, SNAPMTXCOMPLEX_elems(fz), 1);
#ifdef SNAPLIB_DEBUG
    printf("residual vector:\n");
    show_sdmpc(fz);
#ifdef SNAP_SUPFC_T
    mpfr_printf("iteration: %d, red: %Re\n", iterations, red);
#else
    printf("iteration: %d, red: %e\n", iterations, red);
#endif // SNAP_SUPFC_T
#endif
    // if ( we have not better )
    if ( SNAPBASEREAL_cmp(red_old, red) < 0 ) {
      // in this case, we call the linesearch
      SNAP_mul(poly_tmp, poly_d, poly_f1);
      SNAP_sub(poly_fA, poly_f, poly_tmp);
      SNAP_mul(poly_tmp, poly_d, poly_g1);
      SNAP_sub(poly_gA, poly_g, poly_tmp);
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,k),ptb=SNAP_coefs(poly_newd); i<=k; i++,pt1--,ptb++ ) {
        SNAPBASE_set_basecomplex(*ptb,*pt1);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1),ptb=SNAP_coefs(poly_newf1); i<=n-k; i++,pt1--,ptb++ ) {
        SNAPBASE_set_basecomplex(*ptb,*pt1);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2),ptb=SNAP_coefs(poly_newg1); i<=m-k; i++,pt1--,ptb++ ) {
        SNAPBASE_set_basecomplex(*ptb,*pt1);
      }
      SNAP_mul(poly_fC, poly_newd, poly_newf1);
      SNAP_mul(poly_gC, poly_newd, poly_newg1);
      SNAP_mul(poly_fB, poly_d, poly_newf1);
      SNAP_mul(poly_tmp, poly_newd, poly_f1);
      SNAP_add(poly_fB, poly_fB, poly_tmp);
      SNAP_mul(poly_gB, poly_d, poly_newg1);
      SNAP_mul(poly_tmp, poly_newd, poly_g1);
      SNAP_add(poly_gB, poly_gB, poly_tmp);
      SNAPBASEREAL_set_d(alpha, 1);
      // poly_fA = poly_f - poly_d poly_f1
      // poly_gA = poly_g - poly_d poly_g1
      // poly_fB = poly_d poly_newf1 + poly_newd poly_f1
      // poly_gB = poly_d poly_newg1 + poly_newd poly_g1
      // poly_fC = poly_newd poly_newf1
      // poly_gC = poly_newd poly_newg1
      // min_alpha |poly_f - (poly_d + alpha*poly_newd)(poly_f1 + alpha*poly_newf1)|
      //          +|poly_g - (poly_d + alpha*poly_newd)(poly_g1 + alpha*poly_newg1)|
      //         = |poly_fA - alpha*poly_fB - alpha^2 poly_fC|
      //          +|poly_gA - alpha*poly_gB - alpha^2 poly_gC|
      SNAP_agcd_fastgcd_linesearch(poly_fA, poly_fB, poly_fC, poly_gA, poly_gB, poly_gC, &alpha);
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
      mpfr_printf("line search: %Re\n", alpha);
#else
      printf("line search: %e\n", alpha);
#endif // SNAP_SUPFC_T
#endif
#ifdef SNAP_SUPFC_T
      SNAPBASECOMPLEX_set_basereal(c_tmp, alpha);
      SNAPBASECOMPLEX_neg(c_tmp, c_tmp);
#else
      SNAPBASECOMPLEX_neg_basereal(c_tmp, alpha);
#endif // SNAP_SUPFC_T
      SNAP_set(poly_newd, poly_d);
      SNAP_set(poly_newf1, poly_f1);
      SNAP_set(poly_newg1, poly_g1);
#if defined(SNAP_SUPDC_T)
      // BLAS_INCREMENT_STYLE
      // cblas_zaxpy(k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k), -1, SNAP_coefp(poly_newd,0), 1);
      cblas_zaxpy(k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,0), -1, SNAP_coefp(poly_newd,0), 1);
      // cblas_zaxpy(n-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1), -1, SNAP_coefp(poly_newf1,0), 1);
      cblas_zaxpy(n-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k+1), -1, SNAP_coefp(poly_newf1,0), 1);
      // cblas_zaxpy(m-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2), -1, SNAP_coefp(poly_newg1,0), 1);
      cblas_zaxpy(m-k+1, &c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+2), -1, SNAP_coefp(poly_newg1,0), 1);
#elif defined(SNAP_SUPD_T)
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,k),ptb=SNAP_coefs(poly_newd); i<=k; i++,pt1--,ptb++ ) {
        SNAPBASE_submul_basecomplex_basecomplex(*ptb,*pt1,c_tmp);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1),ptb=SNAP_coefs(poly_newf1); i<=n-k; i++,pt1--,ptb++ ) {
        SNAPBASE_submul_basecomplex_basecomplex(*ptb,*pt1,c_tmp);
      }
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2),ptb=SNAP_coefs(poly_newg1); i<=m-k; i++,pt1--,ptb++ ) {
        SNAPBASE_submul_basecomplex_basecomplex(*ptb,*pt1,c_tmp);
      }
#elif defined(SNAP_SUPFC_T)
      snap_mblas_fcaxpy(k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,k), -1, SNAP_coefp(poly_newd,0), 1);
      snap_mblas_fcaxpy(n-k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+1), -1, SNAP_coefp(poly_newf1,0), 1);
      snap_mblas_fcaxpy(m-k+1, c_tmp, SNAPMTXCOMPLEX_elemp(fz_tmp,0,n+m-k+2), -1, SNAP_coefp(poly_newg1,0), 1);
#else
      ERR;
#endif
#ifdef SNAPLIB_DEBUG
      printf("poly_newd: \n");
      show_plist(SNAP_coefs(poly_newd), SNAP_degree(poly_newd)+1);
      printf("poly_newf1: \n");
      show_plist(SNAP_coefs(poly_newf1), SNAP_degree(poly_newf1)+1);
      printf("poly_newg1: \n");
      show_plist(SNAP_coefs(poly_newg1), SNAP_degree(poly_newg1)+1);
#endif
      // residual computation and fix tau
      SNAP_mul(poly_tmp, poly_newf1, poly_newd);
      SNAP_sub(poly_tmp, poly_tmp, poly_f);
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,0),ptb=SNAP_coefp(poly_tmp,n); i<=n; i++,pt1++,ptb-- ) {
        SNAPBASECOMPLEX_set_base(*pt1,*ptb);
      }
      SNAP_mul(poly_tmp, poly_newg1, poly_newd);
      SNAP_sub(poly_tmp, poly_tmp, poly_g);
      for ( i=0,pt1=SNAPMTXCOMPLEX_elemp(fz,0,n+1),ptb=SNAP_coefp(poly_tmp,m); i<=m; i++,pt1++,ptb-- ) {
        SNAPBASECOMPLEX_set_base(*pt1,*ptb);
      }
      SNAPMTX_BLAS_dotc(c_tmp, k+1, SNAP_coefs(poly_newd), 1, SNAP_coefs(poly_newd), 1);
      SNAPBASECOMPLEX_sub(*pt1,c_tmp,tau);
      SNAPMTX_BLAS_dotc(c_tmp, n-k+1, SNAP_coefs(poly_newf1), 1, SNAP_coefs(poly_newf1), 1);
      SNAPBASECOMPLEX_sub(*pt1,*pt1,c_tmp);
      SNAPMTX_BLAS_dotc(c_tmp, m-k+1, SNAP_coefs(poly_newg1), 1, SNAP_coefs(poly_newg1), 1);
      SNAPBASECOMPLEX_sub(*pt1,*pt1,c_tmp);
      SNAPMTXCOMPLEX_BLAS_nrm2(red, n+m+2, SNAPMTXCOMPLEX_elems(fz), 1);
#ifdef SNAPLIB_DEBUG
      printf("residual vector:\n");
      show_sdmpc(fz);
#ifdef SNAP_SUPFC_T
      mpfr_printf("iteration: %d, red: %Re\n", iterations, red);
#else
      printf("iteration: %d, red: %e\n", iterations, red);
#endif // SNAP_SUPFC_T
#endif
      if ( SNAPBASEREAL_cmp(red_old, red) <= 0 ) {
        // in this case, iterations finished.
        break;
      }
    }
    // update the polynomials
    SNAP_set(poly_d, poly_newd);
    SNAP_set(poly_f1, poly_newf1);
    SNAP_set(poly_g1, poly_newg1);
    SNAPBASEREAL_set(red_old, red);
  }
post:
  SNAPBASEREAL_clear(zero);
  SNAPBASECOMPLEX_clear(tau);
  free(dummyP);
  SNAPBASECOMPLEX_clear(c_tmp);
  SNAPBASEREAL_clear(red);
  SNAPBASEREAL_clear(red_old);
  SNAPBASEREAL_clear(newton_mintol);
  SNAPBASECOMPLEX_clear(theta);
  SNAP_clear(poly_newf1);
  SNAP_clear(poly_newg1);
  SNAP_clear(poly_newd);
  SNAP_clear(poly_tmp);
  SNAPMTXCOMPLEX_clear(fz);
  SNAPMTXCOMPLEX_clear(fz_tmp);
  SNAP_clear(poly_fA);
  SNAP_clear(poly_fB);
  SNAP_clear(poly_fC);
  SNAP_clear(poly_gA);
  SNAP_clear(poly_gB);
  SNAP_clear(poly_gC);
  SNAPBASEREAL_clear(alpha);
  SNAPBASEREAL_clear(sqrtofnm);
}

/*!
 *  @brief Fastgcd algorithm for two univariate polynomials.
 *  @details This is an implementation of the Fastgcd algorithm proposed by Bini and Boito
 *   in Structured Matrix Based Methods for Polynomial e-gcd: Analysis and Comparisons,
 *   ISSAC 2007. Note that we scaled the input polynomials at the beginning.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] tol The tolerance.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_fastgcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol) {
  int alreadyfound;
  unsigned int i, n, m, k, k0;
  SNAPBASEREAL_t eps, eps_boito, norm_f, norm_g, norm_d, zero;
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_t tmpr;
#endif // SNAP_SUPFC_T
  SNAPBASECOMPLEX_t theta;
  SNAPMTXCOMPLEX_t vecDt, vecDh, vecD1, genG, genB, genTG, genTB, matW, vecX;
  unsigned int *matP, *matQ, *pi1, *pi2;
  SNAP_t poly_f, poly_g, poly_newd, poly_newf1, poly_newg1, poly_tmp;
  // initialization
  SNAPMTXCOMPLEX_init(vecX);
  SNAPBASECOMPLEX_init(theta);
  SNAPBASEREAL_init_set_ui(zero,0);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(norm_d);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_init(tmpr);
#endif // SNAP_SUPFC_T
  SNAP_init(poly_newd);
  SNAP_init(poly_newf1);
  SNAP_init(poly_newg1);
  SNAP_init(poly_tmp);
  alreadyfound = 0;
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
#ifdef SNAPLIB_DEBUG
  printf("poly_inf: \n");
  show_plist(SNAP_coefs(poly_inf), SNAP_degree(poly_inf)+1);
  printf("poly_ing: \n");
  show_plist(SNAP_coefs(poly_ing), SNAP_degree(poly_ing)+1);
  printf("poly_f: \n");
  show_plist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("poly_g: \n");
  show_plist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#endif
  // generators
  n = SNAP_degree(poly_f);
  m = SNAP_degree(poly_g);
  SNAPMTXCOMPLEX_init2(vecDt, 1, n+m);
  SNAPMTXCOMPLEX_init2(vecDh, 1, n+m);
  SNAPMTXCOMPLEX_init2(vecD1, 1, n+m);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  theta = cexp(3.1415926535897932384626433832795029L*I*((n+m))/(n+m));
  //theta = cexp(3.1415926535897932384626433832795029L*I/(n+m));
#elif defined(SNAP_SUPFC_T)
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(theta), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(theta));
  //SNAPBASECOMPLEX_mul_si(theta, theta, ((n+m))/(n+m));
  SNAPBASECOMPLEX_exp(theta, theta);
#else
  ERR;
#endif
  SNAP_agcd_fastgcd_make_trans_matrices(n+m, n+m, vecDt, vecDh, vecD1, theta);
#ifdef SNAPLIB_DEBUG
  printf("vecDt: \n");
  show_sdmpc(vecDt);
  printf("vecDh: \n");
  show_sdmpc(vecDh);
  printf("vecD1: \n");
  show_sdmpc(vecD1);
#endif
  SNAPMTXCOMPLEX_init2(genG, n+m, 2);
  SNAPMTXCOMPLEX_init2(genTG, n+m, 2);
  SNAPMTXCOMPLEX_init2(genB, 2, n+m);
  SNAPMTXCOMPLEX_init2(genTB, 2, n+m);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // NOTE this is in Ming since the paper (Bini and Boito) is not correct.
  SNAP_agcd_fastgcd_make_sylvester_generators(poly_f, poly_g, genG, genB, theta);
#else
  ERR;
#endif
#ifdef SNAPLIB_DEBUG
  printf("genA: \n");
  show_sdmpc(genG);
  printf("genB: \n");
  show_sdmpc(genB);
#endif
  SNAP_agcd_fastgcd_make_cauchy_generators(genG, genB, genTG, genTB, vecDh);
#ifdef SNAPLIB_DEBUG
  printf("genA(C): \n");
  show_sdmpc(genG);
  printf("genB(C): \n");
  show_sdmpc(genB);
#endif
  /* compute the LU factorization */
  SNAPBASEREAL_init_set_ui(eps, n+m);
  SNAPBASEREAL_sqrt(eps, eps);
  SNAPBASEREAL_div(eps, tol, eps);
  /* Boito's relaxation */
  SNAPBASEREAL_init_set_ui(eps_boito, n+m);
  SNAPBASEREAL_sqrt(eps_boito, eps_boito);
  SNAPBASEREAL_mul(eps_boito, tol, eps_boito);
  /* matW should be zero matrices of size (n+m) */
  SNAPMTXCOMPLEX_init2(matW, n+m, n+m);
  /* matP and matQ should be indeces (0,,...,0) */
  matP = (unsigned int*) malloc(sizeof(unsigned int)*(n+m+3)); // the size considering Jacobian
  matQ = (unsigned int*) malloc(sizeof(unsigned int)*(n+m+3)); // the size considering Jacobian
  for ( i=0, pi1=matP, pi2=matQ; i<n+m; i++,pi1++,pi2++ ) {
    *pi1=0;
    *pi2=0;
  }
  // GKO
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  printf("modified_gko called with eps=%e\n", zero);
#else
  mpfr_printf("modified_gko called with eps=%Re\n", zero);
#endif
#endif
  /* note that k0 == 0 means that there is an error. */
  k0 = SNAP_agcd_fastgcd_modified_gko(vecD1, vecDt, genG, genB, matW, matP, matQ, zero, n+m);
#ifdef SNAPLIB_DEBUG
  printf("k0 = %d\n", k0);
  printf("genA: \n");
  show_sdmpc(genG);
  printf("genB: \n");
  show_sdmpc(genB);
  printf("matW: \n");
  show_sdmpc(matW);
#endif
  if ( k0 == 0 ) {
    goto post;  /* there was an error. */
  }
  /* This is more conservative way. */
  //if ( k0 > n+m ) { k0 = 0; } /* special treatment for the coprime case */
  //else { k0 = SNAPLIB_MIN(n+m-k0,SNAPLIB_MIN(n,m)-1); } /* k0 now represents the initial tentative degree of GCD. */
  /* This is the scheme Boito uses. */
  k0 = n+m;
#ifdef SNAP_SUPFC_T
  SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
  while ( k0 > 0 && SNAPBASEREAL_cmp(tmpr, eps) <= 0 ) {
    k0--;
    SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
  }
#else
#ifdef SNAPLIB_DEBUG
  printf("eps for tentative degree: %e (tolerance: %e)\n", eps_boito, tol);
#endif // SNAPLIB_DEBUG
  while ( k0 > 0 && SNAPBASECOMPLEX_cmpabs_basereal(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1), eps_boito) <= 0 ) {
#ifdef SNAPLIB_DEBUG
    printf("|matW(%d,%d)|: %e\n", k0-1,k0-1,cabs(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1)));
#endif // SNAPLIB_DEBUG
    k0--;
  }
#endif // SNAP_SUPFC_T
  if ( k0 == n+m ) {
    k0 = 0;  /* special treatment for the coprime case */
  } else {
    k0 = SNAPLIB_MIN(n+m-k0-1,SNAPLIB_MIN(n,m)-1);  /* k0 now represents the initial tentative degree of GCD. */
  }
#if defined(SNAPLIB_DEBUG) || defined(SNAPLIB_SHOW_STATUS)
  printf("The first tentative degree of e-divisor is %d\n", k0+1);
#endif
  /* look for a tentative e-divisor of degree k0+1 */
  /* NOTE THAT the degree (k+1) and the order (k) are different!! */
  k = k0;
  while (1) {
#ifdef SNAPLIB_DEBUG
    printf("The loop for e-gcd of degree: %d\n", k+1);
#endif
    // subresultant generators
    SNAPMTXCOMPLEX_reinit3(vecDt, 1, n+m-2*k);
    SNAPMTXCOMPLEX_reinit3(vecDh, 1, n+m-2*k);
    SNAPMTXCOMPLEX_reinit3(vecD1, 1, n+m-k);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    theta = cexp(3.1415926535897932384626433832795029L*I*(n+m-2*k)/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
    //theta = cexp(3.1415926535897932384626433832795029L*I/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
#elif defined(SNAP_SUPFC_T)
    SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(theta), 0);
    SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(theta));
    SNAPBASECOMPLEX_mul_si(theta, theta, (n+m-2*k)/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
    SNAPBASECOMPLEX_exp(theta, theta);
#else
    ERR;
#endif
    SNAP_agcd_fastgcd_make_trans_matrices(n+m-k, n+m-2*k, vecDt, vecDh, vecD1, theta);
    SNAPMTXCOMPLEX_reinit3(genG, n+m-k, 2);
    SNAPMTXCOMPLEX_reinit3(genTG, n+m-k, 2);
    SNAPMTXCOMPLEX_reinit3(genB, 2, n+m-2*k);
    SNAPMTXCOMPLEX_reinit3(genTB, 2, n+m-2*k);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    SNAP_agcd_fastgcd_make_subresultant_generators(poly_f, poly_g, k, genG, genB, theta);
#else
    ERR;
#endif
    SNAP_agcd_fastgcd_make_cauchy_generators(genG, genB, genTG, genTB, vecDh);
    for ( i=0, pi1=matP, pi2=matQ; i<n+m-k; i++,pi1++,pi2++ ) {
      *pi1=0;
      *pi2=0;
    }
    // GKO
#ifdef SNAPLIB_DEBUG
    printf("modified_gko called with eps=%e\n", 0.0);
#endif
    /* note that k0 == 0 means that there is an error. */
    SNAPMTXCOMPLEX_reinit3(matW, n+m-k, n+m-2*k);
    k0 = SNAP_agcd_fastgcd_modified_gko(vecD1, vecDt, genG, genB, matW, matP, matQ, zero, n+m-2*k);
    if ( alreadyfound == 0 ) {
      k0 = n+m-2*k-1;
      SNAPBASEREAL_set_ui(eps, n+m-2*k);
      SNAPBASEREAL_sqrt(eps, eps);
      SNAPBASEREAL_div(eps, tol, eps);
#ifdef SNAP_SUPFC_T
      SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
      while ( k0 > 0 && SNAPBASEREAL_cmp(tmpr, eps) <= 0 ) {
        k0--;
        SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
      }
#else
      while ( k0 > 0 && SNAPBASECOMPLEX_cmpabs_basereal(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1),eps) <= 0 ) {
        k0--;
      }
#endif // SNAP_SUPFC_T
      if ( k0 != n+m-2*k-1 ) {
        if ( k < SNAPLIB_MIN(n,m)-1 ) {
          k += (n+m-2*k-1)-k0;
          k = SNAPLIB_MIN(SNAPLIB_MIN(n,m)-1,k);
          continue;
        }
      }
    }
#ifdef SNAPLIB_DEBUG
    printf("k0 = %d, dm = %d\n", k0, n+m-2*k);
    printf("genA: \n");
    show_sdmpc(genG);
    printf("genB: \n");
    show_sdmpc(genB);
    printf("matW(k0=%d): \n",k0);
    show_sdmpc(matW);
//	for ( i=0, pi1=matP, pi2=matQ; i<n+m-k; i++,pi1++,pi2++ ) { if ( *pi1+*pi2>0 ) { printf("matP[%d]=%d, matQ[%d]=%d\n",i,*pi1,i,*pi2); } }
#endif
    // In Boito's implementation, she uses the fix threshold 1.0e-9 for subresultant matrix.
    // However, it may be better that we use some relative to the input tolerance, in theory.
    // SNAPBASEREAL_set_ui(eps, n+m-2*k); SNAPBASEREAL_sqrt(eps, eps); SNAPBASEREAL_div(eps, tol, eps); //SNAPBASEREAL_mul_ui(eps, eps, 100);
    SNAPBASEREAL_set_d(eps, 1.0e-9);
    SNAPMTXCOMPLEX_reinit3(vecX, 1, n+m-2*k);
    SNAP_agcd_fastgcd_nullvector(vecX, matW, matQ, vecDh, eps);
#ifdef SNAPLIB_DEBUG
    printf("vecX: \n");
    show_sdmpc(vecX);
#endif
    /* temporary, we set the vectors into poly arrays of cofactors */
    SNAP_set_rbasecomplex(poly_g1, SNAPMTXCOMPLEX_elemp(vecX,0,0), m-k);
    SNAP_set_rbasecomplex(poly_f1, SNAPMTXCOMPLEX_elemp(vecX,0,m-k), n-k);
    SNAP_self_neg(poly_g1);
    if ( n <= m ) {
      SNAP_fftdiv_q(poly_d, poly_f, poly_f1);
    } else {
      SNAP_fftdiv_q(poly_d, poly_g, poly_g1);
    }
#ifdef SNAPLIB_DEBUG
    printf("poly_d: \n");
    show_plist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
    printf("poly_f1: \n");
    show_plist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: \n");
    show_plist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    SNAP_mul(poly_tmp, poly_d, poly_f1);
    SNAP_sub(poly_tmp, poly_tmp, poly_f);
    SNAP_2norm(eps, poly_tmp);
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_f=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_f=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
    SNAP_mul(poly_tmp, poly_d, poly_g1);
    SNAP_sub(poly_tmp, poly_tmp, poly_g);
    SNAP_2norm(eps, poly_tmp);
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_g=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_g=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
#endif
    /* refinement */
    SNAP_agcd_fastgcd_refine(poly_f, poly_g, poly_d, poly_f1, poly_g1, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ);

    // check the condition for e-gcd with the given tolerance */
    SNAP_mul(poly_tmp, poly_d, poly_f1);
    SNAP_sub(poly_tmp, poly_tmp, poly_f);
    SNAP_2norm(eps, poly_tmp);
//#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_f=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_f=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
//#endif
    if ( SNAPBASEREAL_cmp(eps, tol) <= 0 ) {
      SNAP_mul(poly_tmp, poly_d, poly_g1);
      SNAP_sub(poly_tmp, poly_tmp, poly_g);
      SNAP_2norm(eps, poly_tmp);
//#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
      mpfr_printf("eps for poly_g=%Re (%Re)\n", eps, tol);
#else
      printf("eps for poly_g=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
//#endif
      if ( SNAPBASEREAL_cmp(eps, tol) <= 0 ) {
        SNAP_set(poly_newd, poly_d);
        SNAP_set(poly_newf1, poly_f1);
        SNAP_set(poly_newg1, poly_g1);
        if ( alreadyfound >= 0 && k < SNAPLIB_MIN(n,m)-1 ) {
          alreadyfound = 1;
          k++;
          continue;
        } else {
          alreadyfound = 1;
          break;
        }
      }
    }
    if ( alreadyfound <= 0 && k > 0 ) {
      alreadyfound = -1;
      k--;
      continue;
    } else if ( alreadyfound > 0 ) {
      break;
    } else {
      alreadyfound = 0;
      break;
    }
  }
  if ( alreadyfound == 0 ) {
    SNAP_set_str(poly_d, "1");
    SNAP_set(poly_f1, poly_f);
    SNAP_set(poly_g1, poly_g);
  } else {
    SNAP_2norm(norm_d, poly_newd);
    SNAP_div_basereal(poly_d, poly_newd, norm_d);
    SNAP_mul_basereal(poly_f1, poly_newf1, norm_d);
    SNAP_mul_basereal(poly_g1, poly_newg1, norm_d);
  }
  /* norm restore */
  SNAP_mul_basereal(poly_f1, poly_f1, norm_f);
  SNAP_mul_basereal(poly_g1, poly_g1, norm_g);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
  if ( SNAPBASE_sgn(SNAP_lcoef(poly_d)) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#elif defined(SNAP_SUPDC_T)
  if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_d))) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#elif defined(SNAP_SUPFC_T)
  if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_d))) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#endif
post:
  SNAPBASEREAL_clear(eps_boito);
  SNAPBASEREAL_clear(eps);
  SNAPBASECOMPLEX_clear(theta);
  SNAPBASEREAL_clear(zero);
  free(matP);
  free(matQ);
  SNAPMTXCOMPLEX_clear(matW);
  SNAPMTXCOMPLEX_clear(vecDt);
  SNAPMTXCOMPLEX_clear(vecDh);
  SNAPMTXCOMPLEX_clear(vecD1);
  SNAPMTXCOMPLEX_clear(genG);
  SNAPMTXCOMPLEX_clear(genTG);
  SNAPMTXCOMPLEX_clear(genB);
  SNAPMTXCOMPLEX_clear(genTB);
  SNAPMTXCOMPLEX_clear(vecX);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_d);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_init(tmpr);
#endif // SNAP_SUPFC_T
  SNAP_clear(poly_newd);
  SNAP_clear(poly_newf1);
  SNAP_clear(poly_newg1);
  SNAP_clear(poly_tmp);
  return 0;
}

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t *sigma_m1, gmp_randstate_t rstate);
int SNAP_agcd_uvgcd_solving_10_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r);
#endif // defined
typedef enum { SNAP_FASTGCD2_REFINEMENT_FASTGCD, SNAP_FASTGCD2_REFINEMENT_UVGCD,
               SNAP_FASTGCD2_COFACTORS_FASTGCD, SNAP_FASTGCD2_COFACTORS_UVGCD, SNAP_FASTGCD2_COFACTORS_UVGCDWP, SNAP_FASTGCD2_COFACTORS_UVGCDWP_ORG,
               SNAP_FASTGCD2_APPGCD_FFTDIV, SNAP_FASTGCD2_APPGCD_LSDIV,
               SNAP_FASTGCD2_DEGREE_FASTGCD, SNAP_FASTGCD2_DEGREE_PIVQR,
               SNAP_FASTGCD2_STOP_1, SNAP_FASTGCD2_STOP_2
             } mode_type;
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
/*!
 *  @brief Fastgcd algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the Fastgcd algorithm proposed by Bini and Boito
 *   in Structured Matrix Based Methods for Polynomial e-gcd: Analysis and Comparisons,
 *   ISSAC 2007. Note that we scaled the input polynomials at the beginning.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] tol The tolerance.
 *  @param[in] mode The control the behavior of the algorithm.
 *             (mode)%10: refinement algorithm: Fastgcd if 0, otherwise UVGCD.
 *             (mode/10)%10: finding candidate gcd: FFT (Fastgcd) if 0, otherwise least square (UVGCD).
 *             (mode/100)%10: finding candidate cofactors: Fastgcd if 0, UVGCD if 1, UVGCD with pivoting QR if 2, otherwise UVGCD with UVGCD's pivoting QR.
 *             (mode/1000)%10: tentative degree: Fastgcd if 0, otherwise PivQR with F-norm.
 *             (mode/10000)%10: normal if 0, otherwise stop2 criterion used.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_fastgcd2(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, const unsigned int mode) {
  int alreadyfound, stop2;
  unsigned int i, n, m, k, k0;
  SNAPBASEREAL_t sigma_m1, eps, eps_boito, norm_f, norm_g, norm_d, zero, norm_R1, norm_R2, norm_R3, sqrtnm;
  SNAPBASE_t *tau_syl, *tau, *zvect, *yvect, *pt;
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_t tmpr;
#endif // SNAP_SUPFC_T
  SNAPBASECOMPLEX_t theta;
  SNAPMTX_t Sn, matrix_syl;
  SNAPMTXCOMPLEX_t vecDt, vecDh, vecD1, genG, genB, genTG, genTB, matW, vecX;
  unsigned int *matP, *matQ, *pi1, *pi2;
  SNAP_t poly_f, poly_g, poly_newd, poly_newf1, poly_newg1, poly_tmp;
  mode_type mode_refinement, mode_cofactors, mode_appgcd, mode_degree, mode_stop2;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_syl;
#endif // defined
  if ( mode%10 == 0 ) {
    mode_refinement = SNAP_FASTGCD2_REFINEMENT_FASTGCD;
  } else {
    mode_refinement = SNAP_FASTGCD2_REFINEMENT_UVGCD;
  }
  if ( (mode/10)%10 == 0 ) {
    mode_appgcd = SNAP_FASTGCD2_APPGCD_FFTDIV;
  } else {
    mode_appgcd = SNAP_FASTGCD2_APPGCD_LSDIV;
  }
  if ( (mode/100)%10 == 0 ) {
    mode_cofactors = SNAP_FASTGCD2_COFACTORS_FASTGCD;
  } else if ( (mode/100)%10 == 1 ) {
    mode_cofactors = SNAP_FASTGCD2_COFACTORS_UVGCD;
  } else if ( (mode/100)%10 == 2 ) {
    mode_cofactors = SNAP_FASTGCD2_COFACTORS_UVGCDWP;
  } else {
    mode_cofactors = SNAP_FASTGCD2_COFACTORS_UVGCDWP_ORG;
  }
  if ( (mode/1000)%10 == 0 ) {
    mode_degree = SNAP_FASTGCD2_DEGREE_FASTGCD;
  } else {
    mode_degree = SNAP_FASTGCD2_DEGREE_PIVQR;
  }
  if ( (mode/10000)%10 == 0 ) {
    mode_stop2 = SNAP_FASTGCD2_STOP_1;
  } else {
    mode_stop2 = SNAP_FASTGCD2_STOP_2;
  }
  // initialization
  n = SNAP_degree(poly_inf);
  m = SNAP_degree(poly_ing);
  if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCD || mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP || mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP_ORG ) {
    SNAPMTX_init2(Sn, SNAP_degree(poly_inf)+SNAP_degree(poly_ing), SNAP_degree(poly_inf)+SNAP_degree(poly_ing));
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    gmp_randinit_mt(rstate);
    gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
    srand((unsigned int)time(NULL));
#endif
    tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
    zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
    yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
    for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
      SNAPBASE_init(*(tau+i));
    }
    for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
      SNAPBASE_init(*(zvect+i));
      SNAPBASE_init(*(yvect+i));
    }
    SNAPBASEREAL_init(sigma_m1);
  } else if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_FASTGCD ) {
    SNAPMTXCOMPLEX_init(vecX);
  }
  if ( mode_refinement == SNAP_FASTGCD2_REFINEMENT_FASTGCD || mode_cofactors == SNAP_FASTGCD2_COFACTORS_FASTGCD || mode_degree == SNAP_FASTGCD2_DEGREE_FASTGCD ) {
    SNAPBASECOMPLEX_init(theta);
    SNAPBASEREAL_init_set_ui(zero,0);
    SNAPMTXCOMPLEX_init(vecDt);
    SNAPMTXCOMPLEX_init(vecDh);
    SNAPMTXCOMPLEX_init(vecD1);
    SNAPMTXCOMPLEX_init(genG);
    SNAPMTXCOMPLEX_init(genB);
    SNAPMTXCOMPLEX_init(genTG);
    SNAPMTXCOMPLEX_init(genTB);
    SNAPMTXCOMPLEX_init(matW);
    matP = (unsigned int*) malloc(sizeof(unsigned int)*(n+m+3)); // the size considering Jacobian
    matQ = (unsigned int*) malloc(sizeof(unsigned int)*(n+m+3)); // the size considering Jacobian
  }
  if ( mode_degree == SNAP_FASTGCD2_DEGREE_PIVQR ) {
    SNAPMTX_init(matrix_syl);
    tau_syl = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(n+m));
    for ( i=0; i<n+m; i++ ) {
      SNAPBASE_init(*(tau_syl+i));
    }
  }
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(norm_d);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_init(tmpr);
#endif // SNAP_SUPFC_T
  SNAP_init(poly_newd);
  SNAP_init(poly_newf1);
  SNAP_init(poly_newg1);
  SNAP_init(poly_tmp);
  alreadyfound = 0;
  stop2 = 0;
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
#ifdef SNAPLIB_DEBUG
  printf("poly_inf: \n");
  show_plist(SNAP_coefs(poly_inf), SNAP_degree(poly_inf)+1);
  printf("poly_ing: \n");
  show_plist(SNAP_coefs(poly_ing), SNAP_degree(poly_ing)+1);
  printf("poly_f: \n");
  show_plist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("poly_g: \n");
  show_plist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#endif
  if ( mode_degree == SNAP_FASTGCD2_DEGREE_FASTGCD ) {

    // generators
    SNAPMTXCOMPLEX_reinit3(vecDt, 1, n+m);
    SNAPMTXCOMPLEX_reinit3(vecDh, 1, n+m);
    SNAPMTXCOMPLEX_reinit3(vecD1, 1, n+m);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    theta = cexp(3.1415926535897932384626433832795029L*I*((n+m))/(n+m));
    //theta = cexp(3.1415926535897932384626433832795029L*I/(n+m));
#elif defined(SNAP_SUPFC_T)
    SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(theta), 0);
    SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(theta));
    //SNAPBASECOMPLEX_mul_si(theta, theta, ((n+m))/(n+m));
    SNAPBASECOMPLEX_exp(theta, theta);
#else
    ERR;
#endif
    SNAP_agcd_fastgcd_make_trans_matrices(n+m, n+m, vecDt, vecDh, vecD1, theta);
#ifdef SNAPLIB_DEBUG
    printf("vecDt: \n");
    show_sdmpc(vecDt);
    printf("vecDh: \n");
    show_sdmpc(vecDh);
    printf("vecD1: \n");
    show_sdmpc(vecD1);
#endif
    SNAPMTXCOMPLEX_reinit3(genG, n+m, 2);
    SNAPMTXCOMPLEX_reinit3(genTG, n+m, 2);
    SNAPMTXCOMPLEX_reinit3(genB, 2, n+m);
    SNAPMTXCOMPLEX_reinit3(genTB, 2, n+m);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    // NOTE this is in Ming since the paper (Bini and Boito) is not correct.
    SNAP_agcd_fastgcd_make_sylvester_generators(poly_f, poly_g, genG, genB, theta);
#else
    ERR;
#endif
#ifdef SNAPLIB_DEBUG
    printf("genA: \n");
    show_sdmpc(genG);
    printf("genB: \n");
    show_sdmpc(genB);
#endif
    SNAP_agcd_fastgcd_make_cauchy_generators(genG, genB, genTG, genTB, vecDh);
#ifdef SNAPLIB_DEBUG
    printf("genA(C): \n");
    show_sdmpc(genG);
    printf("genB(C): \n");
    show_sdmpc(genB);
#endif
    /* compute the LU factorization */
    SNAPBASEREAL_init_set_ui(eps, n+m);
    SNAPBASEREAL_sqrt(eps, eps);
    SNAPBASEREAL_div(eps, tol, eps);
    /* Boito's relaxation */
    SNAPBASEREAL_init_set_ui(eps_boito, n+m);
    SNAPBASEREAL_sqrt(eps_boito, eps_boito);
    SNAPBASEREAL_mul(eps_boito, tol, eps_boito);
    /* matW should be zero matrices of size (n+m) */
    SNAPMTXCOMPLEX_reinit3(matW, n+m, n+m);
    /* matP and matQ should be indeces (0,,...,0) */
    for ( i=0, pi1=matP, pi2=matQ; i<n+m; i++,pi1++,pi2++ ) {
      *pi1=0;
      *pi2=0;
    }
    // GKO
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    printf("modified_gko called with eps=%e\n", zero);
#else
    mpfr_printf("modified_gko called with eps=%Re\n", zero);
#endif
#endif
    /* note that k0 == 0 means that there is an error. */
    k0 = SNAP_agcd_fastgcd_modified_gko(vecD1, vecDt, genG, genB, matW, matP, matQ, zero, n+m);
#ifdef SNAPLIB_DEBUG
    printf("k0 = %d\n", k0);
    printf("genA: \n");
    show_sdmpc(genG);
    printf("genB: \n");
    show_sdmpc(genB);
    printf("matW: \n");
    show_sdmpc(matW);
#endif
    if ( k0 == 0 ) {
      goto post;  /* there was an error. */
    }
    /* This is more conservative way. */
    //if ( k0 > n+m ) { k0 = 0; } /* special treatment for the coprime case */
    //else { k0 = SNAPLIB_MIN(n+m-k0,SNAPLIB_MIN(n,m)-1); } /* k0 now represents the initial tentative degree of GCD. */
    /* This is the scheme Boito uses. */
    k0 = n+m;
#ifdef SNAP_SUPFC_T
    SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
    while ( k0 > 0 && SNAPBASEREAL_cmp(tmpr, eps) <= 0 ) {
      k0--;
      SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
    }
#else
#ifdef SNAPLIB_DEBUG
    printf("eps for tentative degree: %e (tolerance: %e)\n", eps_boito, tol);
#endif // SNAPLIB_DEBUG
    while ( k0 > 0 && SNAPBASECOMPLEX_cmpabs_basereal(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1), eps_boito) <= 0 ) {
#ifdef SNAPLIB_DEBUG
      printf("|matW(%d,%d)|: %e\n", k0-1,k0-1,cabs(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1)));
#endif // SNAPLIB_DEBUG
      k0--;
    }
#endif // SNAP_SUPFC_T
    if ( k0 == n+m ) {
      k0 = 0;  /* special treatment for the coprime case */
    } else {
      k0 = SNAPLIB_MIN(n+m-k0-1,SNAPLIB_MIN(n,m)-1);  /* k0 now represents the initial tentative degree of GCD. */
    }
  } else {
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    printf("ERROR in agcd_fastgcd2: This mode is not supported.\n");
    goto post;
#else
    SNAPMTX_sylvester_matrix_transpose(matrix_syl, poly_f, poly_g);
    jpvt_syl = (lapack_int *) malloc(sizeof(lapack_int)*(n+m));
    for (i=0; i<n+m; i++) {
      *(jpvt_syl+i) = 0;
    }
    // QR decomposition
#if defined(SNAP_SUPD_T)
    if ( LAPACKE_dgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt_syl, tau_syl) != 0 ) {
      printf("ERROR in agcd_fastgcd2: dgeqrf could not compute the QR decomposition.\n");
      goto post;
    }
#elif defined(SNAP_SUPDC_T)
    if ( LAPACKE_zgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), jpvt_syl, tau_syl) != 0 ) {
      printf("ERROR in agcd_fastgcd2: dgeqrf could not compute the QR decomposition.\n");
      goto post;
    }
#endif
    free(jpvt_syl);
    // finding maximum k s.t. |R^_3|_2 <= e\sqrt{n+m}(1+|R^_1^{-1}|_2*|R^_2|_2)
    // 2norm is too much time-consuming hence we use F-norm as follows.
    // finding first k0 s.t. |R_3|_F <= e\sqrt{n+m}(1+|R_2|_F/|R_k0k0|)
    // where R_1 is the first block k0xk0 of R, and take k as n+m-k0-1 (k+1 is the degree)
    for ( k0 = 0; k0 < n+m; k0++ ) {
      for ( i = 0; i < k0; i++ ) {
        SNAPBASE_set_ui(SNAPMTX_elem(matrix_syl, k0, i), 0);
      }
    }
    SNAPBASEREAL_sqrt_ui(sqrtnm, n+m);
    SNAPBASEREAL_mul(sqrtnm, sqrtnm, tol);
    for ( k0 = 1; k0 < n+m; k0++ ) {
      SNAPBASE_abs(norm_R1, SNAPMTX_elem(matrix_syl, k0-1, k0-1));
      if ( SNAPBASE_zeroq(norm_R1) ) {
        break;
      }
#ifdef SNAP_SUPDC_T
      norm_R2 = LAPACKE_zlange(LAPACK_ROW_MAJOR, 'F', k0, n+m-k0, SNAPMTX_elemp(matrix_syl, 0, k0), SNAPMTX_csize(matrix_syl));
      norm_R3 = LAPACKE_zlange(LAPACK_ROW_MAJOR, 'F', n+m-k0, n+m-k0, SNAPMTX_elemp(matrix_syl, k0, k0), SNAPMTX_csize(matrix_syl));
#else
      norm_R2 = LAPACKE_dlange(LAPACK_ROW_MAJOR, 'F', k0, n+m-k0, SNAPMTX_elemp(matrix_syl, 0, k0), SNAPMTX_csize(matrix_syl));
      norm_R3 = LAPACKE_dlange(LAPACK_ROW_MAJOR, 'F', n+m-k0, n+m-k0, SNAPMTX_elemp(matrix_syl, k0, k0), SNAPMTX_csize(matrix_syl));
#endif // SNAP_SUPDC_T
      if ( norm_R3 <= sqrtnm*(1.0+norm_R2/norm_R1) ) {
        break;
      }
    }
    if ( k0 == n+m ) {
      k0 = 0;  /* special treatment for the coprime case */
    } else {
      k0 = SNAPLIB_MIN(n+m-k0-1,SNAPLIB_MIN(n,m)-1);  /* k0 now represents the initial tentative degree of GCD. */
    }
#endif // defined
  }
#if defined(SNAPLIB_DEBUG) || defined(SNAPLIB_SHOW_STATUS)
  printf("The first tentative degree of e-divisor is %d\n", k0+1);
#endif
  /* look for a tentative e-divisor of degree k0+1 */
  /* NOTE THAT the degree (k+1) and the order (k) are different!! */
  k = k0;
  while (1) {
#ifdef SNAPLIB_DEBUG
    printf("The loop for e-gcd of degree: %d\n", k+1);
#endif
    if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_FASTGCD ) {
      // subresultant generators
      SNAPMTXCOMPLEX_reinit3(vecDt, 1, n+m-2*k);
      SNAPMTXCOMPLEX_reinit3(vecDh, 1, n+m-2*k);
      SNAPMTXCOMPLEX_reinit3(vecD1, 1, n+m-k);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      theta = cexp(3.1415926535897932384626433832795029L*I*(n+m-2*k)/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
      //theta = cexp(3.1415926535897932384626433832795029L*I/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
#elif defined(SNAP_SUPFC_T)
      SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(theta), 0);
      SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(theta));
      SNAPBASECOMPLEX_mul_si(theta, theta, (n+m-2*k)/SNAP_agcd_fastgcd_lcm(n+m-k,n+m-2*k));
      SNAPBASECOMPLEX_exp(theta, theta);
#else
      ERR;
#endif
      SNAP_agcd_fastgcd_make_trans_matrices(n+m-k, n+m-2*k, vecDt, vecDh, vecD1, theta);
      SNAPMTXCOMPLEX_reinit3(genG, n+m-k, 2);
      SNAPMTXCOMPLEX_reinit3(genTG, n+m-k, 2);
      SNAPMTXCOMPLEX_reinit3(genB, 2, n+m-2*k);
      SNAPMTXCOMPLEX_reinit3(genTB, 2, n+m-2*k);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_fastgcd_make_subresultant_generators(poly_f, poly_g, k, genG, genB, theta);
#else
      ERR;
#endif
      SNAP_agcd_fastgcd_make_cauchy_generators(genG, genB, genTG, genTB, vecDh);
      for ( i=0, pi1=matP, pi2=matQ; i<n+m-k; i++,pi1++,pi2++ ) {
        *pi1=0;
        *pi2=0;
      }
      // GKO
#ifdef SNAPLIB_DEBUG
      printf("modified_gko called with eps=%e\n", 0.0);
#endif
      /* note that k0 == 0 means that there is an error. */
      SNAPMTXCOMPLEX_reinit3(matW, n+m-k, n+m-2*k);
      k0 = SNAP_agcd_fastgcd_modified_gko(vecD1, vecDt, genG, genB, matW, matP, matQ, zero, n+m-2*k);
      if ( alreadyfound == 0 ) {
        k0 = n+m-2*k-1;
        SNAPBASEREAL_set_ui(eps, n+m-2*k);
        SNAPBASEREAL_sqrt(eps, eps);
        SNAPBASEREAL_div(eps, tol, eps);
#ifdef SNAP_SUPFC_T
        SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
        while ( k0 > 0 && SNAPBASEREAL_cmp(tmpr, eps) <= 0 ) {
          k0--;
          SNAPBASECOMPLEX_abs(tmpr, SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1));
        }
#else
        while ( k0 > 0 && SNAPBASECOMPLEX_cmpabs_basereal(SNAPMTXCOMPLEX_elem(matW,k0-1,k0-1),eps) <= 0 ) {
          k0--;
        }
#endif // SNAP_SUPFC_T
        if ( k0 != n+m-2*k-1 ) {
          if ( k < SNAPLIB_MIN(n,m)-1 ) {
            k += (n+m-2*k-1)-k0;
            k = SNAPLIB_MIN(SNAPLIB_MIN(n,m)-1,k);
            continue;
          }
        }
      }
#ifdef SNAPLIB_DEBUG
      printf("k0 = %d, dm = %d\n", k0, n+m-2*k);
      printf("genA: \n");
      show_sdmpc(genG);
      printf("genB: \n");
      show_sdmpc(genB);
      printf("matW(k0=%d): \n",k0);
      show_sdmpc(matW);
//	for ( i=0, pi1=matP, pi2=matQ; i<n+m-k; i++,pi1++,pi2++ ) { if ( *pi1+*pi2>0 ) { printf("matP[%d]=%d, matQ[%d]=%d\n",i,*pi1,i,*pi2); } }
#endif
      // In Boito's implementation, she uses the fix threshold 1.0e-9 for subresultant matrix.
      // However, it may be better that we use some relative to the input tolerance, in theory.
      // SNAPBASEREAL_set_ui(eps, n+m-2*k); SNAPBASEREAL_sqrt(eps, eps); SNAPBASEREAL_div(eps, tol, eps); //SNAPBASEREAL_mul_ui(eps, eps, 100);
      SNAPBASEREAL_set_d(eps, 1.0e-9);
      SNAPMTXCOMPLEX_reinit3(vecX, 1, n+m-2*k);
      SNAP_agcd_fastgcd_nullvector(vecX, matW, matQ, vecDh, eps);
#ifdef SNAPLIB_DEBUG
      printf("vecX: \n");
      show_sdmpc(vecX);
#endif
      /* temporary, we set the vectors into poly arrays of cofactors */
      SNAP_set_rbasecomplex(poly_g1, SNAPMTXCOMPLEX_elemp(vecX,0,0), m-k);
      SNAP_set_rbasecomplex(poly_f1, SNAPMTXCOMPLEX_elemp(vecX,0,m-k), n-k);
      SNAP_self_neg(poly_g1);
    } else {
      // Construct the full subresultant matrix w.r.t. the current tentative degree.
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP_ORG ) {
        k0 = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
        SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_f, poly_g, k0);
        while ( k0 > k ) {
          for ( i=SNAP_degree(poly_g)-k0,pt=SNAP_coefs(poly_f); i<SNAP_degree(poly_g)+SNAP_degree(poly_f)-k0+1; i++,pt++ ) {
            SNAPBASE_set(SNAPMTX_elem(Sn, i, SNAPMTX_cols(Sn)), *pt);
          }
          for ( i=SNAP_degree(poly_f)-k0,pt=SNAP_coefs(poly_g); i<SNAP_degree(poly_g)+SNAP_degree(poly_f)-k0+1; i++,pt++ ) {
            SNAPBASE_set(SNAPMTX_elem(Sn, i, SNAPMTX_cols(Sn)+1), *pt);
          }
          SNAPMTX_rows(Sn)++;
          SNAPMTX_cols(Sn)+=2;
          k0--;
        }
      } else {
        SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_f, poly_g, k);
      }
#ifdef SNAPLIB_DEBUG
      printf("Sn\n");
      show_sdmpc(Sn);
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
        // In this case, we partially use LAPACKE_dgeqp3 for partial QR dcomposition with column pivoting.
        jpvt = (lapack_int *) malloc(sizeof(lapack_int)*SNAPMTX_csize(Sn));
        for (i=0; i<SNAPMTX_csize(Sn); i++) {
          *(jpvt+i) = 0;
        }
      }
#endif
      // QR decomposition
#if defined(SNAP_SUPD_T)
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
        if ( LAPACKE_dgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), jpvt, tau) != 0 ) {
          printf("ERROR in agcd_fastgcd2: dgeqrf could not compute the QR decomposition.\n");
          goto post;
        }
      } else {
        if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
          clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
          LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
        ) {
          printf("ERROR in agcd_fastgcd2: dgeqrf could not compute the QR decomposition.\n");
          goto post;
        }
      }
#elif defined(SNAP_SUPDC_T)
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
        if ( LAPACKE_zgeqp3(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), jpvt, tau) != 0 ) {
          printf("ERROR in agcd_fastgcd2: dgeqrf could not compute the QR decomposition.\n");
          goto post;
        }
      } else {
        if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
          clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
          LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
        ) {
          printf("ERROR in agcd_fastgcd2: zgeqrf could not compute the QR decomposition.\n");
          goto post;
        }
      }
#elif defined(SNAP_SUPFR_T)
      if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
        printf("ERROR in agcd_fastgcd2: mlapack_frgeqrf could not compute the QR decomposition.\n");
        goto post;
      }
#elif defined(SNAP_SUPFC_T)
      if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
        printf("ERROR in agcd_fastgcd2: mlapack_fcgeqrf could not compute the QR decomposition.\n");
        goto post;
      }
#endif
      // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, &sigma_m1, rstate) < 0 ) {
        printf("ERROR in agcd_fastgcd2: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, sigma_m1, rstate) < 0 ) {
        printf("ERROR in agcd_fastgcd2: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      printf("sigma_m1 = %e\n", sigma_m1);
#endif // defined
#endif // SNAPLIB_DEBUG
      // from 0 to deg(g)-k, deg(f)-k
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
        for ( i=0,pt=zvect; i < SNAPMTX_csize(Sn)-2*k; i++ ) {
          if ( (unsigned int) *(jpvt+i)-1 < SNAP_degree(poly_ing)-k ) {
            SNAP_set_coef(poly_g1, ((unsigned int) *(jpvt+i)-1), *pt++ );
          } else {
            SNAP_set_coef(poly_f1, ((unsigned int) *(jpvt+i)-1)-SNAP_degree(poly_ing)+k, *pt++ );
          }
        }
        SNAP_degree(poly_f1) = SNAP_degree(poly_inf)-k-1;
        SNAP_degree(poly_g1) = SNAP_degree(poly_ing)-k-1;
      } else {
#endif
        if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP_ORG ) {
          k0 = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
          // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
          SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-k0 );
          SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-k0, SNAP_degree(poly_inf)-k0 );
          // to deg(g)-init_r, deg(f)-init_r from deg(g)-r-1, deg(f)-r-1
          for ( i=k, pt=zvect+SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*k0; i<k0; i++ ) {
            SNAP_set_coef(poly_g1, SNAP_degree(poly_g)-(k0+k-i), *pt++ );
            SNAP_set_coef(poly_f1, SNAP_degree(poly_f)-(k0+k-i), *pt++ );
          }
        } else {
          SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-k );
          SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-k, SNAP_degree(poly_inf)-k);
        }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      }
#endif
      SNAP_self_neg(poly_f1);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
        free(jpvt);
      }
#endif // defined
    }
    if ( mode_appgcd == SNAP_FASTGCD2_APPGCD_LSDIV ) {
      SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, k);
    } else {
      if ( n <= m ) {
        SNAP_fftdiv_q(poly_d, poly_f, poly_f1);
      } else {
        SNAP_fftdiv_q(poly_d, poly_g, poly_g1);
      }
    }
#ifdef SNAPLIB_DEBUG
    printf("poly_d: \n");
    show_plist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
    printf("poly_f1: \n");
    show_plist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: \n");
    show_plist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    SNAP_mul(poly_tmp, poly_d, poly_f1);
    SNAP_sub(poly_tmp, poly_tmp, poly_f);
    SNAP_2norm(eps, poly_tmp);
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_f=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_f=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
    SNAP_mul(poly_tmp, poly_d, poly_g1);
    SNAP_sub(poly_tmp, poly_tmp, poly_g);
    SNAP_2norm(eps, poly_tmp);
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_g=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_g=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
#endif
    /* refinement */
    if ( mode_refinement == SNAP_FASTGCD2_REFINEMENT_FASTGCD ) {
      SNAP_agcd_fastgcd_refine(poly_f, poly_g, poly_d, poly_f1, poly_g1, vecDt, vecDh, vecD1, genG, genB, matW, genTG, genTB, matP, matQ);
    } else {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, &eps);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, eps);
#endif
    }
    // check the condition for e-gcd with the given tolerance */
    SNAP_mul(poly_tmp, poly_d, poly_f1);
    SNAP_sub(poly_tmp, poly_tmp, poly_f);
    SNAP_2norm(eps, poly_tmp);
//#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
    mpfr_printf("eps for poly_f=%Re (%Re)\n", eps, tol);
#else
    printf("eps for poly_f=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
//#endif
    if ( SNAPBASEREAL_cmp(eps, tol) <= 0 ) {
      SNAP_mul(poly_tmp, poly_d, poly_g1);
      SNAP_sub(poly_tmp, poly_tmp, poly_g);
      SNAP_2norm(eps, poly_tmp);
//#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFC_T
      mpfr_printf("eps for poly_g=%Re (%Re)\n", eps, tol);
#else
      printf("eps for poly_g=%e (%e)\n", eps, tol);
#endif // SNAP_SUPFC_T
//#endif
      if ( SNAPBASEREAL_cmp(eps, tol) <= 0 ) {
        SNAP_set(poly_newd, poly_d);
        SNAP_set(poly_newf1, poly_f1);
        SNAP_set(poly_newg1, poly_g1);
        if ( alreadyfound >= 0 && k < SNAPLIB_MIN(n,m)-1 ) {
          alreadyfound = 1;
          k++;
          continue;
        } else {
          alreadyfound = 1;
          break;
        }
      }
    }
    if ( alreadyfound <= 0 && k > 0 ) {
      alreadyfound = -1;
      k--;
      continue;
    } else if ( alreadyfound > 0 ) {
      if ( mode_stop2 == SNAP_FASTGCD2_STOP_2 && stop2 == 0 ) {
        stop2 = 1;
        k++;
        continue;
      }
      break;
    } else {
      alreadyfound = 0;
      break;
    }
  }
  if ( alreadyfound == 0 ) {
    SNAP_set_str(poly_d, "1");
    SNAP_set(poly_f1, poly_f);
    SNAP_set(poly_g1, poly_g);
  } else {
    SNAP_2norm(norm_d, poly_newd);
    SNAP_div_basereal(poly_d, poly_newd, norm_d);
    SNAP_mul_basereal(poly_f1, poly_newf1, norm_d);
    SNAP_mul_basereal(poly_g1, poly_newg1, norm_d);
  }
  /* norm restore */
  SNAP_mul_basereal(poly_f1, poly_f1, norm_f);
  SNAP_mul_basereal(poly_g1, poly_g1, norm_g);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
  if ( SNAPBASE_sgn(SNAP_lcoef(poly_d)) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#elif defined(SNAP_SUPDC_T)
  if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_d))) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#elif defined(SNAP_SUPFC_T)
  if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_d))) < 0 ) {
    SNAP_self_neg(poly_d);
    SNAP_self_neg(poly_f1);
    SNAP_self_neg(poly_g1);
  }
#endif
post:
  if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCD || mode_cofactors == SNAP_FASTGCD2_COFACTORS_UVGCDWP ) {
    SNAPMTX_clear(Sn);
    free(tau);
    free(zvect);
    free(yvect);
    SNAPBASEREAL_clear(sigma_m1);
  } else if ( mode_cofactors == SNAP_FASTGCD2_COFACTORS_FASTGCD ) {
    SNAPMTXCOMPLEX_clear(vecX);
  }
  SNAPBASEREAL_clear(eps_boito);
  SNAPBASEREAL_clear(eps);
  if ( mode_refinement == SNAP_FASTGCD2_REFINEMENT_FASTGCD || mode_cofactors == SNAP_FASTGCD2_COFACTORS_FASTGCD || mode_degree == SNAP_FASTGCD2_DEGREE_FASTGCD ) {
    SNAPBASECOMPLEX_clear(theta);
    SNAPBASEREAL_clear(zero);
    SNAPMTXCOMPLEX_clear(matW);
    SNAPMTXCOMPLEX_clear(vecDt);
    SNAPMTXCOMPLEX_clear(vecDh);
    SNAPMTXCOMPLEX_clear(vecD1);
    SNAPMTXCOMPLEX_clear(genG);
    SNAPMTXCOMPLEX_clear(genTG);
    SNAPMTXCOMPLEX_clear(genB);
    SNAPMTXCOMPLEX_clear(genTB);
    free(matP);
    free(matQ);
  }
  if ( mode_degree == SNAP_FASTGCD2_DEGREE_PIVQR ) {
    SNAPMTX_clear(matrix_syl);
    free(tau_syl);
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_d);
#ifdef SNAP_SUPFC_T
  SNAPBASEREAL_init(tmpr);
#endif // SNAP_SUPFC_T
  SNAP_clear(poly_newd);
  SNAP_clear(poly_newf1);
  SNAP_clear(poly_newg1);
  SNAP_clear(poly_tmp);
  return 0;
}
