/*
 * Copyright (c) 2018, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAP2str(a) #a
//#define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.5le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPDC_T)
void show_sdmdc(sdmdc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmdc_rows(mtx); i++) {
    for(j=0; j<sdmdc_cols(mtx); j++) {
      printf("(%+.5le%+.5le*I), ",creal(sdmdc_elem(mtx,i,j)),cimag(sdmdc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_dclist(dcomplex *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.5le%+.5le*I), ",creal(*(list+i)),cimag(*(list+i)));
  }
  printf("\n");
}
#endif // defined
#endif // SNAPLIB_DEBUG

/* ================== START [agcd_ghlgcd] ================== */

#if defined(SNAP_SUPD_T)
int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t *sigma_m1, gmp_randstate_t rstate);
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
int SNAP_agcd_uvgcd_solving_10_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r);
#endif // defined

/*!
 *  @brief revert the polynomials from Subresultant mapping matrix w.r.t. f and g of order r (Sylvester matrix is r=0).
 *  @param[in] matrix The sdm*_t variable which is the subresultant matrix.
 *  @param[in,out] poly_f This function returns poly_f from Syl_r(poly_f,poly_g).
 *  @param[in,out] poly_g This function returns poly_g from Syl_r(poly_f,poly_g).
 *  @param[in] r The order of Subresultant mapping.
 */
void SNAPMTX_revert_average_subresultant_mapping_matrix_asc(SNAPMTX_t matrix, SNAP_t poly_f, SNAP_t poly_g, unsigned int r) {
  unsigned int i, j;
  SNAPBASE_t *pt;
  for ( i=0,pt=SNAP_coefs(poly_f); i<=SNAP_degree(poly_f); i++,pt++ ) {
    SNAPBASE_set_ui(*pt, 0);
  }
  for ( i=0,pt=SNAP_coefs(poly_g); i<=SNAP_degree(poly_g); i++,pt++ ) {
    SNAPBASE_set_ui(*pt, 0);
  }
  // poly_f
  for ( i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_f); j++ ) {
      SNAPBASE_add(SNAP_coef(poly_f,j), SNAP_coef(poly_f,j), SNAPMTX_elem(matrix,j+i,i));
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_g); j++ ) {
      SNAPBASE_add(SNAP_coef(poly_g,j), SNAP_coef(poly_g,j), SNAPMTX_elem(matrix,j+i,i+SNAP_degree(poly_g)-r));
    }
  }
  for ( i=0,pt=SNAP_coefs(poly_f); i<=SNAP_degree(poly_f); i++,pt++ ) {
    SNAPBASE_div_ui(*pt, *pt, SNAP_degree(poly_g)-r);
  }
  for ( i=0,pt=SNAP_coefs(poly_g); i<=SNAP_degree(poly_g); i++,pt++ ) {
    SNAPBASE_div_ui(*pt, *pt, SNAP_degree(poly_f)-r);
  }
}

/*!
 *  @brief GHLGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the GHLGCD algorithm proposed by Kosaku Nagasaka,
 *   using the iterative algorithm by M.Giesbrecht, J.Haraldson and G.Labahn in 2017.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_ghlgcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold) {
  int result_code = 0;
  unsigned int deg_f, deg_g, deg_s, deg_t, r, i, j, jmin, jmax, loop_count, is_break;
  SNAPMTX_t matSr, matU, matVT, vecZ, matA, matL, vecL, matH, vecN;
  SNAPMTXREAL_t vecS;
  SNAPBASE_t *ptZ, *ptX, *ptL, *ptXF, *ptXG, *ptXS, *ptXT, *ptN;
  SNAPBASE_t *pt1, *pt2, *pt3, *pt4, *pt5;
  SNAPBASEREAL_t *ptr1, norm_f, norm_g, relative_tolerance, norm_z, residue_max, sigma_m1;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g;
#ifdef SNAP_SUPD_T
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
  srand((unsigned int)time(NULL));
#endif // SNAP_SUPD_T

  // initialize the degress and length
  r = d - 1;
  deg_f = SNAP_degree(poly_inf);
  deg_g = SNAP_degree(poly_ing);
  deg_s = deg_g - r - 1;
  deg_t = deg_f - r - 1;
  // initialize the matrices and vectors.
  SNAP_init_set(poly_f, poly_inf); // poly_f denotes F+dF
  SNAP_init_set(poly_g, poly_ing); // poly_g denotes G+dG
  SNAPMTX_init2(matSr, deg_f + deg_g - r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matU, deg_f + deg_g - r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matVT, deg_f + deg_g - 2*r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matH, 3*(deg_f + deg_g - r + 1), 3*(deg_f + deg_g - r + 1));
  SNAPMTXREAL_init2(vecS, 1, deg_f + deg_g - 2*r);
  SNAPMTX_init2(vecZ, 1, 4*(deg_f + deg_g - r + 1)); // len(lambda)=deg_f+deg_g-r+1 but 2(deg_f+deg_g-r+1) required for solving LS.
  ptZ = SNAPMTX_elems(vecZ);
  ptX = ptXF = ptZ;
  ptXG = ptXF + deg_f + 1;
  ptXS = ptXG + deg_g + 1;
  ptXT = ptXS + deg_g - r;
  ptL = ptZ + 2*(deg_f + deg_g - r + 1);
  SNAPMTX_init2(matA, 2*(deg_f + deg_g - r + 1), deg_f + deg_g - r + 1);
  SNAPMTX_init2(vecN, 1, 3*(deg_f + deg_g - r + 1));
  ptN = SNAPMTX_elems(vecN);
#ifdef SNAP_SUPD_T
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*3*(deg_f + deg_g - r + 1));
#endif // SNAP_SUPD_T
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAPBASEREAL_init(norm_z);
  SNAPBASEREAL_init(residue_max);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
  SNAPMTX_init2(matL, 3*(deg_f + deg_g - r + 1), 3*(deg_f + deg_g - r + 1));
  SNAPMTX_init2(vecL, 1, 3*(deg_f + deg_g - r + 1));

  // initialize \vec{x} by lift and project with average.
  // |=> construction of subresultant mapping matrix
  SNAPMTX_subresultant_mapping_matrix_asc(matSr, poly_inf, poly_ing, r);
#ifdef SNAP_SUPD_T
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'S', SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr),
                       SNAPMTXREAL_elems(vecS), SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
    printf("ERROR in agcd_ghlgcd: dgesdd could not compute the SVD.\n");
    result_code = -1;
    goto post;
  }
#else
  IMPLEMENT HERE
#endif
  // |=> store the singular vector in XS and overwrite them with 0
#ifdef SNAP_SUPD_T
  for ( i=0, pt1=SNAPMTX_elemp(matVT,SNAPMTX_cols(matVT)-1, 0), pt2=ptXS; i<SNAPMTX_cols(matVT); i++,pt1++,pt2++ ) {
    SNAPBASE_set(*pt2,*pt1);
    SNAPBASE_set_ui(*pt1, 0);
  }
#else
  IMPLEMENT HERE
#endif

#ifdef SNAPLIB_DEBUG
  printf("ptXS: ");
  show_dlist(ptXS, deg_f+deg_g-2*r);
#endif // SNAPLIB_DEBUG

  // |=> SIGMA*VT for computing residue
  for ( i=1, ptr1=SNAPMTXREAL_elems(vecS), pt1=SNAPMTX_elems(matVT); i < SNAPMTX_cols(matVT); i++,ptr1++,pt1+=SNAPMTX_csize(matVT) ) {
    for ( j=0, pt2=pt1; j < SNAPMTX_cols(matVT); j++,pt2++ ) {
      SNAPBASE_mul_basereal(*pt2, *pt2, *ptr1);
    }
  }
  // |=> U*SIGMA*VT for computing residue
#ifdef SNAP_SUPD_T
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_cols(matSr), 1.0, SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT), 0.0, SNAPMTX_elems(matSr), SNAPMTX_csize(matSr));
#else
  IMPLEMENT HERE
#endif

#ifdef SNAPLIB_DEBUG
  printf("matSr: \n");
  show_sdmd(matSr);
#endif // SNAPLIB_DEBUG

  // |=> initial perturbation by lift and project method
  SNAPMTX_revert_average_subresultant_mapping_matrix_asc(matSr, poly_f, poly_g, r);

#ifdef SNAPLIB_DEBUG
  printf("poly_f: ");
  show_dlist(SNAP_coefs(poly_f), deg_f+1);
  printf("poly_g: ");
  show_dlist(SNAP_coefs(poly_g), deg_g+1);
#endif // SNAPLIB_DEBUG

  for ( i=0, pt1=SNAP_coefs(poly_f), pt2=SNAP_coefs(poly_inf), pt3=ptX; i<=SNAP_degree(poly_f); i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_sub(*pt3, *pt1, *pt2);
  }
  for ( i=0, pt1=SNAP_coefs(poly_g), pt2=SNAP_coefs(poly_ing); i<=SNAP_degree(poly_g); i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_sub(*pt3, *pt1, *pt2);
  }

#ifdef SNAPLIB_DEBUG
  printf("vecXF: ");
  show_dlist(ptXF, deg_f+1);
  printf("vecXG: ");
  show_dlist(ptXG, deg_g+1);
#endif // SNAPLIB_DEBUG

  // initialize \vec{lambda} by linear least squares.
  // |=> coefficient matrix A
  for ( i=0, pt1=ptXS, pt2=SNAPMTX_elems(matA); i<=deg_s; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_f+1; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=ptXT, pt2=SNAPMTX_elemp(matA, deg_f+1, 0); i<=deg_t; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_g+1; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=ptXS, pt2=SNAPMTX_elemp(matA, deg_f+deg_g+2, (deg_f + deg_g - r)); i<=deg_s; i++,pt1++,pt2+=SNAPMTX_csize(matA) ) {
    SNAPBASE_mul_ui(*pt2, *pt1, 2);
  }
  for ( i=0, pt1=ptXT; i<=deg_t; i++,pt1++,pt2+=SNAPMTX_csize(matA) ) {
    SNAPBASE_mul_ui(*pt2, *pt1, 2);
  }
  for ( i=0, pt1=SNAP_coefs(poly_f), pt2=SNAPMTX_elemp(matA, deg_f+deg_g+2, 0); i<=deg_f; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_g-r; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=SNAP_coefs(poly_g), pt2=SNAPMTX_elemp(matA, deg_f+2*deg_g-r+2, 0); i<=deg_g; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_f-r; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  // |=> vector C
  for ( i=0, pt1=ptXF, pt2=ptL; i<=deg_f; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2*((signed int) deg_g));
  }
  for ( i=0, pt1=ptXG; i<=deg_g; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2*((signed int) deg_f));
  }
  for ( i=0, pt1=ptXS; i<=deg_s; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2);
  }
  for ( i=0, pt1=ptXT; i<=deg_t; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2);
  }

#ifdef SNAPLIB_DEBUG
  printf("matA: \n");
  show_sdmd(matA);
  printf("vecC: ");
  show_dlist(ptL, SNAPMTX_rows(matA));
#endif // SNAPLIB_DEBUG

  // |=> solve AL=C by least square
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matA); i++,jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matA), SNAPMTX_cols(matA), 1, SNAPMTX_elems(matA), SNAPMTX_csize(matA), ptL, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ghlgcd), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }

#ifdef SNAPLIB_DEBUG
  printf("ptL: ");
  show_dlist(ptL, deg_f+deg_g-r+1);
#endif // SNAPLIB_DEBUG

  // main loop (Newton method)
  loop_count = 1;
  while(1) {
    // |=> hessian ( + nabla^2 Lk ) in matH
    // ||-> zero clear
    for ( i=0, pt1=SNAPMTX_elems(matH); i < SNAPMTX_rows(matH); i++ ) {
      for ( j=0; j < SNAPMTX_cols(matH); j++, pt1++ ) {
        SNAPBASE_set_ui(*pt1, 0);
      }
    }
    // ||-> d^2Psi and d^2l/dsidsj  (diagonal elements)
    for ( i=0, pt1=SNAPMTX_elems(matH); i < deg_f + 1; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 2*deg_g);
    }
    for ( i=0; i < deg_g + 1; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 2*deg_f);
    }
    for ( i=0, pt2=ptL+deg_f+deg_g-r; i < deg_g - r; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 1);
      SNAPBASE_add(*pt1, *pt1, *pt2);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
    }
    for ( i=0; i < deg_f - r; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 1);
      SNAPBASE_add(*pt1, *pt1, *pt2);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
    }
    // ||-> d^2l/dsidfj, then d^2l/dsidgj
    for ( i=0; i < deg_f + 1; i++ ) {
      for ( j=0, pt1=SNAPMTX_elemp(matH, i, deg_f+deg_g+2), pt2=ptL+i; j < deg_g - r; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_g + 1; i++ ) {
      for ( j=0, pt1=SNAPMTX_elemp(matH, i + deg_f + 1, deg_f+2*deg_g-r+2), pt2=ptL+i; j < deg_f - r; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dfidlj, then d^2l/dgidlj
    for ( i=0; i < deg_f + 1; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i, 2*deg_f+2*deg_g+2-2*r+i), pt2=ptXS; j < i + deg_g - r; j++,pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_g + 1; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+1, 2*deg_f+2*deg_g+2-2*r+i), pt2=ptXT; j < i + deg_f - r; j++,pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dsidlj, then d^2l/dtidlj
    for ( i=0; i < deg_g - r; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+deg_g+2, 2*deg_f+2*deg_g+2-2*r+i), pt2=SNAP_coefs(poly_f); j < i + deg_f + 1; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_f - r; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+2*deg_g+2-r, 2*deg_f+2*deg_g+2-2*r+i), pt2=SNAP_coefs(poly_g); j < i + deg_g + 1; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dsidlx, then d^2l/dtidlx
    for ( i=0, pt1=SNAPMTX_elemp(matH, deg_f+deg_g+2, 3*deg_f+3*deg_g+2-3*r), pt2=ptXS; i < deg_g - r; i++, pt1+=SNAPMTX_csize(matH), pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2);
    }
    for ( i=0, pt2=ptXT; i < deg_f - r; i++, pt1+=SNAPMTX_csize(matH), pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2);
    }
    // ||-> copy the upper-right to the bottom-left. (by symmetry)
    for ( i=0; i < SNAPMTX_rows(matH); i++ ) {
      for ( j=i+1, pt1=SNAPMTX_elemp(matH, i, j), pt2=SNAPMTX_elemp(matH, j, i); j < SNAPMTX_cols(matH); j++, pt1++, pt2+=SNAPMTX_csize(matH) ) {
        SNAPBASE_set(*pt2, *pt1);
      }
    }
    // |=> gradient ( - nabla Lk ) in vecN (ptN)
    // ||-> zero clear
    for ( i=0, pt1=ptN; i < SNAPMTX_cols(vecN); i++, pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
    // ||-> dPsi/dfi + dl/dfi, then dPsi/dgi + dl/dgi
    for ( i=0, pt1=ptN, pt2=ptXF; i < deg_f + 1; i++, pt1++, pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2*deg_g);
      for ( j=0, pt3=ptXS, pt4=ptL+i; j < deg_g - r; j++, pt3++, pt4++ ) {
        SNAPBASE_addmul(*pt1, *pt3, *pt4);
      }
    }
    for ( i=0, pt2=ptXG; i < deg_g + 1; i++, pt1++, pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2*deg_f);
      for ( j=0, pt3=ptXT, pt4=ptL+i; j < deg_f - r; j++, pt3++, pt4++ ) {
        SNAPBASE_addmul(*pt1, *pt3, *pt4);
      }
    }
    // ||-> dPsi/dsi + dl/dsi, then dPsi/dti + dl/dti
    for ( i=0, pt2=ptL+deg_f+deg_g-r, pt3=ptXS; i < deg_g - r; i++, pt1++, pt3++ ) {
      SNAPBASE_add_ui(*pt1, *pt2, 1);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
      SNAPBASE_mul(*pt1, *pt1, *pt3);
      for ( j=0, pt4=SNAP_coefs(poly_f), pt5=ptL+i; j < deg_f + 1; j++, pt4++, pt5++ ) {
        SNAPBASE_addmul(*pt1, *pt4, *pt5);
      }
    }
    for ( i=0, pt2=ptL+deg_f+deg_g-r, pt3=ptXT; i < deg_f - r; i++, pt1++, pt3++ ) {
      SNAPBASE_add_ui(*pt1, *pt2, 1);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
      SNAPBASE_mul(*pt1, *pt1, *pt3);
      for ( j=0, pt4=SNAP_coefs(poly_g), pt5=ptL+i; j < deg_g + 1; j++, pt4++, pt5++ ) {
        SNAPBASE_addmul(*pt1, *pt4, *pt5);
      }
    }
    // ||-> dl/dli
    for ( i=0; i < deg_f + deg_g - r; i++, pt1++ ) {
      if ( i+r+1 > deg_g ) {
        jmin = i+r+1-deg_g;
      } else {
        jmin = 0;
      }
      if ( i > deg_f ) {
        jmax = deg_f;
      } else {
        jmax = i;
      }
      for ( j=jmin, pt2=SNAP_coefs(poly_f)+jmin, pt3=ptXS+i-jmin; j <= jmax; j++, pt2++, pt3-- ) {
        SNAPBASE_addmul(*pt1, *pt2, *pt3);
      }
      if ( i+r+1 > deg_f ) {
        jmin = i+r+1-deg_f;
      } else {
        jmin = 0;
      }
      if ( i > deg_g ) {
        jmax = deg_g;
      } else {
        jmax = i;
      }
      for ( j=jmin, pt2=SNAP_coefs(poly_g)+jmin, pt3=ptXT+i-jmin; j <= jmax; j++, pt2++, pt3-- ) {
        SNAPBASE_addmul(*pt1, *pt2, *pt3);
      }
    }
    // ||-> dl/dlx
    SNAPBASE_set_si(*pt1, -1);
    for ( i=0, pt2=ptXS; i < deg_g - r; i++, pt2++ ) {
      SNAPBASE_addmul(*pt1, *pt2, *pt2);
    }
    for ( i=0, pt2=ptXT; i < deg_f - r; i++, pt2++ ) {
      SNAPBASE_addmul(*pt1, *pt2, *pt2);
    }
    // ||-> *(-1)
    for ( i=0, pt1=ptN; i < SNAPMTX_rows(matH); i++, pt1++ ) {
      SNAPBASE_neg(*pt1, *pt1);
    }

#ifdef SNAPLIB_DEBUG
  printf("matH: \n");
  show_sdmd(matH);
  printf("ptN: ");
  show_dlist(ptN, SNAPMTX_rows(matH));
#endif // SNAPLIB_DEBUG

    // |=> solve (nabla^2 Lk) z = - nabla Lk by least square
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matH); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matH), SNAPMTX_cols(matH), 1, SNAPMTX_elems(matH), SNAPMTX_csize(matH), ptN, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ghlgcd), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }

#ifdef SNAPLIB_DEBUG
  printf("ptZ: ");
  show_dlist(ptN, SNAPMTX_cols(matH));
#endif // SNAPLIB_DEBUG

    // |=> update z_k ( x lambda )
    for ( i=0, pt1=ptZ, pt2=ptN; i < SNAPMTX_cols(matH); i++, pt1++, pt2++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }

#ifdef SNAPLIB_DEBUG
  printf("delta_F: ");
  show_dlist(ptXF, deg_f+1);
  printf("delta_G: ");
  show_dlist(ptXG, deg_g+1);
  printf("ptXS: ");
  show_dlist(ptXS, deg_g-r);
  printf("ptXT: ");
  show_dlist(ptXT, deg_f-r);
#endif // SNAPLIB_DEBUG

    // |=> convergent condition check
    SNAPMTX_BLAS_nrm2(norm_z, SNAPMTX_cols(matH), ptN, 1);

#ifdef SNAPLIB_DEBUG
    printf("NORM_Z (%d): %e\n", loop_count, norm_z);
#endif // SNAPLIB_DEBUG

    is_break = 0;
    if ( SNAPBASEREAL_cmp(norm_z, threshold) < 0 ) {
      is_break = 1;
    } else {
      if ( loop_count == 1 ) {
        SNAPBASEREAL_set(residue_max, norm_z);
      } else if ( SNAPBASEREAL_cmp(norm_z, residue_max) > 0 ) {
        is_break = 2;
      } else {
        SNAPBASEREAL_set(residue_max, norm_z);
      }
    }
    if ( is_break < 2 ) {
      // |=> update poly_f and poly_g
      for ( i=0, pt1=SNAP_coefs(poly_f), pt2=ptN; i <= deg_f; i++, pt1++, pt2++ ) {
        SNAPBASE_add(*pt1, *pt1, *pt2);
      }
      for ( i=0, pt1=SNAP_coefs(poly_g); i <= deg_g; i++, pt1++, pt2++ ) {
        SNAPBASE_add(*pt1, *pt1, *pt2);
      }
//      printf("poly_f: ");
//      show_dlist(SNAP_coefs(poly_f), deg_f+1);
//      printf("poly_g: ");
//      show_dlist(SNAP_coefs(poly_g), deg_g+1);
    }
    if ( is_break > 0 ) {
      break;
    }
    loop_count++;
  }
  printf("GHLGCD: #iterations: %u\n", loop_count);
  // NOTE: the resulting polynomials are stored in poly_f and poly_g!!!
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
  printf("F[x]: ");
  show_dlist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dlist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#else
  IMPLEMENT HERE!
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, pt1=SNAPMTX_elems(matSr); i < SNAPMTX_csize(matSr)*SNAPMTX_rsize(matSr); i++,pt1++ ) {
    SNAPBASE_set_d(*pt1, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSr, poly_f, poly_g, r);
  SNAPBASEREAL_init(sigma_m1);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  for ( i=0, pt1=tau, pt2=zvect, pt3=yvect; i < SNAPMTX_csize(matSr); i++, pt1++, pt2++, pt3++ ) {
    SNAPBASE_init(*pt1);
    SNAPBASE_init(*pt2);
    SNAPBASE_init(*pt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#else
  IMPLEMENT HERE!!!
#endif // defined

  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSr, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_ghlgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#else
  IMPLEMENT HERE!!!
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-r );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-r, SNAP_degree(poly_inf)-r );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, r) == 0 ) {
    //#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
    //#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
  } else {
    result_code = -1;
  }
post_uvgcd:
  SNAPBASEREAL_clear(sigma_m1);
  for ( i=0, pt1=tau, pt2=zvect, pt3=yvect; i < SNAPMTX_csize(matSr); i++, pt1++, pt2++, pt3++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
    SNAPBASE_clear(*pt3);
  }
  free(zvect);
  free(yvect);
  free(tau);
post:
  // free the memories.
  SNAPBASEREAL_clear(residue_max);
  SNAPBASEREAL_clear(norm_z);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPMTX_clear(vecL);
  SNAPMTX_clear(matL);
#ifdef SNAP_SUPD_T
  free(jpvt);
#endif // SNAP_SUPD_T
  SNAPMTX_clear(vecN);
  SNAPMTX_clear(matA);
  SNAPMTX_clear(vecZ);
  SNAPMTXREAL_clear(vecS);
  SNAPMTX_clear(matH);
  SNAPMTX_clear(matVT);
  SNAPMTX_clear(matU);
  SNAPMTX_clear(matSr);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

typedef enum { SNAP_GHLGCDEX_REFINEMENT_NONE, SNAP_GHLGCDEX_REFINEMENT_UVGCD } mode_type;
/*!
 *  @brief GHLGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the GHLGCD algorithm proposed by Kosaku Nagasaka,
 *   using the iterative algorithm by M.Giesbrecht, J.Haraldson and G.Labahn in 2017.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_ghlgcd_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const int mode_opt) {
  int result_code = 0;
  unsigned int deg_f, deg_g, deg_s, deg_t, r, i, j, jmin, jmax, loop_count, is_break;
  SNAPMTX_t matSr, matU, matVT, vecZ, matA, matL, vecL, matH, vecN;
  SNAPMTXREAL_t vecS;
  SNAPBASE_t *ptZ, *ptX, *ptL, *ptXF, *ptXG, *ptXS, *ptXT, *ptN;
  SNAPBASE_t *pt1, *pt2, *pt3, *pt4, *pt5;
  SNAPBASEREAL_t *ptr1, norm_f, norm_g, relative_tolerance, norm_z, residue_max, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g;
  mode_type mode_refinement;
#ifdef SNAP_SUPD_T
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
  srand((unsigned int)time(NULL));
#endif // SNAP_SUPD_T
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_GHLGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_GHLGCDEX_REFINEMENT_UVGCD;
  }

  // initialize the degress and length
  r = d - 1;
  deg_f = SNAP_degree(poly_inf);
  deg_g = SNAP_degree(poly_ing);
  deg_s = deg_g - r - 1;
  deg_t = deg_f - r - 1;
  // initialize the matrices and vectors.
  SNAP_init_set(poly_f, poly_inf); // poly_f denotes F+dF
  SNAP_init_set(poly_g, poly_ing); // poly_g denotes G+dG
  SNAPMTX_init2(matSr, deg_f + deg_g - r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matU, deg_f + deg_g - r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matVT, deg_f + deg_g - 2*r, deg_f + deg_g - 2*r);
  SNAPMTX_init2(matH, 3*(deg_f + deg_g - r + 1), 3*(deg_f + deg_g - r + 1));
  SNAPMTXREAL_init2(vecS, 1, deg_f + deg_g - 2*r);
  SNAPMTX_init2(vecZ, 1, 4*(deg_f + deg_g - r + 1)); // len(lambda)=deg_f+deg_g-r+1 but 2(deg_f+deg_g-r+1) required for solving LS.
  ptZ = SNAPMTX_elems(vecZ);
  ptX = ptXF = ptZ;
  ptXG = ptXF + deg_f + 1;
  ptXS = ptXG + deg_g + 1;
  ptXT = ptXS + deg_g - r;
  ptL = ptZ + 2*(deg_f + deg_g - r + 1);
  SNAPMTX_init2(matA, 2*(deg_f + deg_g - r + 1), deg_f + deg_g - r + 1);
  SNAPMTX_init2(vecN, 1, 3*(deg_f + deg_g - r + 1));
  ptN = SNAPMTX_elems(vecN);
#ifdef SNAP_SUPD_T
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*3*(deg_f + deg_g - r + 1));
#endif // SNAP_SUPD_T
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAPBASEREAL_init(norm_z);
  SNAPBASEREAL_init(residue_max);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
  SNAPMTX_init2(matL, 3*(deg_f + deg_g - r + 1), 3*(deg_f + deg_g - r + 1));
  SNAPMTX_init2(vecL, 1, 3*(deg_f + deg_g - r + 1));

  // initialize \vec{x} by lift and project with average.
  // |=> construction of subresultant mapping matrix
  SNAPMTX_subresultant_mapping_matrix_asc(matSr, poly_inf, poly_ing, r);
#ifdef SNAP_SUPD_T
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'S', SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr),
                       SNAPMTXREAL_elems(vecS), SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
    printf("ERROR in agcd_ghlgcd: dgesdd could not compute the SVD.\n");
    result_code = -1;
    goto post;
  }
#else
  IMPLEMENT HERE
#endif
  // |=> store the singular vector in XS and overwrite them with 0
#ifdef SNAP_SUPD_T
  for ( i=0, pt1=SNAPMTX_elemp(matVT,SNAPMTX_cols(matVT)-1, 0), pt2=ptXS; i<SNAPMTX_cols(matVT); i++,pt1++,pt2++ ) {
    SNAPBASE_set(*pt2,*pt1);
    SNAPBASE_set_ui(*pt1, 0);
  }
#else
  IMPLEMENT HERE
#endif

#ifdef SNAPLIB_DEBUG
  printf("ptXS: ");
  show_dlist(ptXS, deg_f+deg_g-2*r);
#endif // SNAPLIB_DEBUG

  // |=> SIGMA*VT for computing residue
  for ( i=1, ptr1=SNAPMTXREAL_elems(vecS), pt1=SNAPMTX_elems(matVT); i < SNAPMTX_cols(matVT); i++,ptr1++,pt1+=SNAPMTX_csize(matVT) ) {
    for ( j=0, pt2=pt1; j < SNAPMTX_cols(matVT); j++,pt2++ ) {
      SNAPBASE_mul_basereal(*pt2, *pt2, *ptr1);
    }
  }
  // |=> U*SIGMA*VT for computing residue
#ifdef SNAP_SUPD_T
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_cols(matSr), 1.0, SNAPMTX_elems(matU), SNAPMTX_csize(matU), SNAPMTX_elems(matVT), SNAPMTX_csize(matVT), 0.0, SNAPMTX_elems(matSr), SNAPMTX_csize(matSr));
#else
  IMPLEMENT HERE
#endif

#ifdef SNAPLIB_DEBUG
  printf("matSr: \n");
  show_sdmd(matSr);
#endif // SNAPLIB_DEBUG

  // |=> initial perturbation by lift and project method
  SNAPMTX_revert_average_subresultant_mapping_matrix_asc(matSr, poly_f, poly_g, r);

#ifdef SNAPLIB_DEBUG
  printf("poly_f: ");
  show_dlist(SNAP_coefs(poly_f), deg_f+1);
  printf("poly_g: ");
  show_dlist(SNAP_coefs(poly_g), deg_g+1);
#endif // SNAPLIB_DEBUG

  for ( i=0, pt1=SNAP_coefs(poly_f), pt2=SNAP_coefs(poly_inf), pt3=ptX; i<=SNAP_degree(poly_f); i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_sub(*pt3, *pt1, *pt2);
  }
  for ( i=0, pt1=SNAP_coefs(poly_g), pt2=SNAP_coefs(poly_ing); i<=SNAP_degree(poly_g); i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_sub(*pt3, *pt1, *pt2);
  }

#ifdef SNAPLIB_DEBUG
  printf("vecXF: ");
  show_dlist(ptXF, deg_f+1);
  printf("vecXG: ");
  show_dlist(ptXG, deg_g+1);
#endif // SNAPLIB_DEBUG

  // initialize \vec{lambda} by linear least squares.
  // |=> coefficient matrix A
  for ( i=0, pt1=ptXS, pt2=SNAPMTX_elems(matA); i<=deg_s; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_f+1; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=ptXT, pt2=SNAPMTX_elemp(matA, deg_f+1, 0); i<=deg_t; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_g+1; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=ptXS, pt2=SNAPMTX_elemp(matA, deg_f+deg_g+2, (deg_f + deg_g - r)); i<=deg_s; i++,pt1++,pt2+=SNAPMTX_csize(matA) ) {
    SNAPBASE_mul_ui(*pt2, *pt1, 2);
  }
  for ( i=0, pt1=ptXT; i<=deg_t; i++,pt1++,pt2+=SNAPMTX_csize(matA) ) {
    SNAPBASE_mul_ui(*pt2, *pt1, 2);
  }
  for ( i=0, pt1=SNAP_coefs(poly_f), pt2=SNAPMTX_elemp(matA, deg_f+deg_g+2, 0); i<=deg_f; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_g-r; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  for ( i=0, pt1=SNAP_coefs(poly_g), pt2=SNAPMTX_elemp(matA, deg_f+2*deg_g-r+2, 0); i<=deg_g; i++,pt1++,pt2++ ) {
    for ( j=0, pt3=pt2; j<deg_f-r; j++,pt3+=SNAPMTX_csize(matA)+1 ) {
      SNAPBASE_set(*pt3, *pt1);
    }
  }
  // |=> vector C
  for ( i=0, pt1=ptXF, pt2=ptL; i<=deg_f; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2*((signed int) deg_g));
  }
  for ( i=0, pt1=ptXG; i<=deg_g; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2*((signed int) deg_f));
  }
  for ( i=0, pt1=ptXS; i<=deg_s; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2);
  }
  for ( i=0, pt1=ptXT; i<=deg_t; i++,pt1++,pt2++ ) {
    SNAPBASE_mul_si(*pt2, *pt1, -2);
  }

#ifdef SNAPLIB_DEBUG
  printf("matA: \n");
  show_sdmd(matA);
  printf("vecC: ");
  show_dlist(ptL, SNAPMTX_rows(matA));
#endif // SNAPLIB_DEBUG

  // |=> solve AL=C by least square
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matA); i++,jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matA), SNAPMTX_cols(matA), 1, SNAPMTX_elems(matA), SNAPMTX_csize(matA), ptL, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ghlgcd), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }

#ifdef SNAPLIB_DEBUG
  printf("ptL: ");
  show_dlist(ptL, deg_f+deg_g-r+1);
#endif // SNAPLIB_DEBUG

  // main loop (Newton method)
  loop_count = 1;
  while(1) {
    // |=> hessian ( + nabla^2 Lk ) in matH
    // ||-> zero clear
    for ( i=0, pt1=SNAPMTX_elems(matH); i < SNAPMTX_rows(matH); i++ ) {
      for ( j=0; j < SNAPMTX_cols(matH); j++, pt1++ ) {
        SNAPBASE_set_ui(*pt1, 0);
      }
    }
    // ||-> d^2Psi and d^2l/dsidsj  (diagonal elements)
    for ( i=0, pt1=SNAPMTX_elems(matH); i < deg_f + 1; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 2*deg_g);
    }
    for ( i=0; i < deg_g + 1; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 2*deg_f);
    }
    for ( i=0, pt2=ptL+deg_f+deg_g-r; i < deg_g - r; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 1);
      SNAPBASE_add(*pt1, *pt1, *pt2);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
    }
    for ( i=0; i < deg_f - r; i++, pt1+=SNAPMTX_csize(matH)+1 ) {
      SNAPBASE_set_ui(*pt1, 1);
      SNAPBASE_add(*pt1, *pt1, *pt2);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
    }
    // ||-> d^2l/dsidfj, then d^2l/dsidgj
    for ( i=0; i < deg_f + 1; i++ ) {
      for ( j=0, pt1=SNAPMTX_elemp(matH, i, deg_f+deg_g+2), pt2=ptL+i; j < deg_g - r; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_g + 1; i++ ) {
      for ( j=0, pt1=SNAPMTX_elemp(matH, i + deg_f + 1, deg_f+2*deg_g-r+2), pt2=ptL+i; j < deg_f - r; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dfidlj, then d^2l/dgidlj
    for ( i=0; i < deg_f + 1; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i, 2*deg_f+2*deg_g+2-2*r+i), pt2=ptXS; j < i + deg_g - r; j++,pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_g + 1; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+1, 2*deg_f+2*deg_g+2-2*r+i), pt2=ptXT; j < i + deg_f - r; j++,pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dsidlj, then d^2l/dtidlj
    for ( i=0; i < deg_g - r; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+deg_g+2, 2*deg_f+2*deg_g+2-2*r+i), pt2=SNAP_coefs(poly_f); j < i + deg_f + 1; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    for ( i=0; i < deg_f - r; i++ ) {
      for ( j=i, pt1=SNAPMTX_elemp(matH, i+deg_f+2*deg_g+2-r, 2*deg_f+2*deg_g+2-2*r+i), pt2=SNAP_coefs(poly_g); j < i + deg_g + 1; j++, pt1++, pt2++ ) {
        SNAPBASE_set(*pt1, *pt2);
      }
    }
    // ||-> d^2l/dsidlx, then d^2l/dtidlx
    for ( i=0, pt1=SNAPMTX_elemp(matH, deg_f+deg_g+2, 3*deg_f+3*deg_g+2-3*r), pt2=ptXS; i < deg_g - r; i++, pt1+=SNAPMTX_csize(matH), pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2);
    }
    for ( i=0, pt2=ptXT; i < deg_f - r; i++, pt1+=SNAPMTX_csize(matH), pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2);
    }
    // ||-> copy the upper-right to the bottom-left. (by symmetry)
    for ( i=0; i < SNAPMTX_rows(matH); i++ ) {
      for ( j=i+1, pt1=SNAPMTX_elemp(matH, i, j), pt2=SNAPMTX_elemp(matH, j, i); j < SNAPMTX_cols(matH); j++, pt1++, pt2+=SNAPMTX_csize(matH) ) {
        SNAPBASE_set(*pt2, *pt1);
      }
    }
    // |=> gradient ( - nabla Lk ) in vecN (ptN)
    // ||-> zero clear
    for ( i=0, pt1=ptN; i < SNAPMTX_cols(vecN); i++, pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
    // ||-> dPsi/dfi + dl/dfi, then dPsi/dgi + dl/dgi
    for ( i=0, pt1=ptN, pt2=ptXF; i < deg_f + 1; i++, pt1++, pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2*deg_g);
      for ( j=0, pt3=ptXS, pt4=ptL+i; j < deg_g - r; j++, pt3++, pt4++ ) {
        SNAPBASE_addmul(*pt1, *pt3, *pt4);
      }
    }
    for ( i=0, pt2=ptXG; i < deg_g + 1; i++, pt1++, pt2++ ) {
      SNAPBASE_mul_ui(*pt1, *pt2, 2*deg_f);
      for ( j=0, pt3=ptXT, pt4=ptL+i; j < deg_f - r; j++, pt3++, pt4++ ) {
        SNAPBASE_addmul(*pt1, *pt3, *pt4);
      }
    }
    // ||-> dPsi/dsi + dl/dsi, then dPsi/dti + dl/dti
    for ( i=0, pt2=ptL+deg_f+deg_g-r, pt3=ptXS; i < deg_g - r; i++, pt1++, pt3++ ) {
      SNAPBASE_add_ui(*pt1, *pt2, 1);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
      SNAPBASE_mul(*pt1, *pt1, *pt3);
      for ( j=0, pt4=SNAP_coefs(poly_f), pt5=ptL+i; j < deg_f + 1; j++, pt4++, pt5++ ) {
        SNAPBASE_addmul(*pt1, *pt4, *pt5);
      }
    }
    for ( i=0, pt2=ptL+deg_f+deg_g-r, pt3=ptXT; i < deg_f - r; i++, pt1++, pt3++ ) {
      SNAPBASE_add_ui(*pt1, *pt2, 1);
      SNAPBASE_mul_ui(*pt1, *pt1, 2);
      SNAPBASE_mul(*pt1, *pt1, *pt3);
      for ( j=0, pt4=SNAP_coefs(poly_g), pt5=ptL+i; j < deg_g + 1; j++, pt4++, pt5++ ) {
        SNAPBASE_addmul(*pt1, *pt4, *pt5);
      }
    }
    // ||-> dl/dli
    for ( i=0; i < deg_f + deg_g - r; i++, pt1++ ) {
      if ( i+r+1 > deg_g ) {
        jmin = i+r+1-deg_g;
      } else {
        jmin = 0;
      }
      if ( i > deg_f ) {
        jmax = deg_f;
      } else {
        jmax = i;
      }
      for ( j=jmin, pt2=SNAP_coefs(poly_f)+jmin, pt3=ptXS+i-jmin; j <= jmax; j++, pt2++, pt3-- ) {
        SNAPBASE_addmul(*pt1, *pt2, *pt3);
      }
      if ( i+r+1 > deg_f ) {
        jmin = i+r+1-deg_f;
      } else {
        jmin = 0;
      }
      if ( i > deg_g ) {
        jmax = deg_g;
      } else {
        jmax = i;
      }
      for ( j=jmin, pt2=SNAP_coefs(poly_g)+jmin, pt3=ptXT+i-jmin; j <= jmax; j++, pt2++, pt3-- ) {
        SNAPBASE_addmul(*pt1, *pt2, *pt3);
      }
    }
    // ||-> dl/dlx
    SNAPBASE_set_si(*pt1, -1);
    for ( i=0, pt2=ptXS; i < deg_g - r; i++, pt2++ ) {
      SNAPBASE_addmul(*pt1, *pt2, *pt2);
    }
    for ( i=0, pt2=ptXT; i < deg_f - r; i++, pt2++ ) {
      SNAPBASE_addmul(*pt1, *pt2, *pt2);
    }
    // ||-> *(-1)
    for ( i=0, pt1=ptN; i < SNAPMTX_rows(matH); i++, pt1++ ) {
      SNAPBASE_neg(*pt1, *pt1);
    }

#ifdef SNAPLIB_DEBUG
  printf("matH: \n");
  show_sdmd(matH);
  printf("ptN: ");
  show_dlist(ptN, SNAPMTX_rows(matH));
#endif // SNAPLIB_DEBUG

    // |=> solve (nabla^2 Lk) z = - nabla Lk by least square
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matH); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matH), SNAPMTX_cols(matH), 1, SNAPMTX_elems(matH), SNAPMTX_csize(matH), ptN, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_ghlgcd), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }

#ifdef SNAPLIB_DEBUG
  printf("ptZ: ");
  show_dlist(ptN, SNAPMTX_cols(matH));
#endif // SNAPLIB_DEBUG

    // |=> update z_k ( x lambda )
    for ( i=0, pt1=ptZ, pt2=ptN; i < SNAPMTX_cols(matH); i++, pt1++, pt2++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }

#ifdef SNAPLIB_DEBUG
  printf("delta_F: ");
  show_dlist(ptXF, deg_f+1);
  printf("delta_G: ");
  show_dlist(ptXG, deg_g+1);
  printf("ptXS: ");
  show_dlist(ptXS, deg_g-r);
  printf("ptXT: ");
  show_dlist(ptXT, deg_f-r);
#endif // SNAPLIB_DEBUG

    // |=> convergent condition check
    SNAPMTX_BLAS_nrm2(norm_z, SNAPMTX_cols(matH), ptN, 1);

#ifdef SNAPLIB_DEBUG
    printf("NORM_Z (%d): %e\n", loop_count, norm_z);
#endif // SNAPLIB_DEBUG

    is_break = 0;
    if ( SNAPBASEREAL_cmp(norm_z, threshold) < 0 ) {
      is_break = 1;
    } else {
      if ( loop_count == 1 ) {
        SNAPBASEREAL_set(residue_max, norm_z);
      } else if ( SNAPBASEREAL_cmp(norm_z, residue_max) > 0 ) {
        is_break = 2;
      } else {
        SNAPBASEREAL_set(residue_max, norm_z);
      }
    }
    if ( is_break < 2 ) {
      // |=> update poly_f and poly_g
      for ( i=0, pt1=SNAP_coefs(poly_f), pt2=ptN; i <= deg_f; i++, pt1++, pt2++ ) {
        SNAPBASE_add(*pt1, *pt1, *pt2);
      }
      for ( i=0, pt1=SNAP_coefs(poly_g); i <= deg_g; i++, pt1++, pt2++ ) {
        SNAPBASE_add(*pt1, *pt1, *pt2);
      }
//      printf("poly_f: ");
//      show_dlist(SNAP_coefs(poly_f), deg_f+1);
//      printf("poly_g: ");
//      show_dlist(SNAP_coefs(poly_g), deg_g+1);
    }
    if ( is_break > 0 ) {
      break;
    }
    loop_count++;
  }
  printf("GHLGCD: #iterations: %u\n", loop_count);
  // NOTE: the resulting polynomials are stored in poly_f and poly_g!!!
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
  printf("F[x]: ");
  show_dlist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dlist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#else
  IMPLEMENT HERE!
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, pt1=SNAPMTX_elems(matSr); i < SNAPMTX_csize(matSr)*SNAPMTX_rsize(matSr); i++,pt1++ ) {
    SNAPBASE_set_d(*pt1, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSr, poly_f, poly_g, r);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_GHLGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSr));
  for ( i=0, pt1=tau, pt2=zvect, pt3=yvect; i < SNAPMTX_csize(matSr); i++, pt1++, pt2++, pt3++ ) {
    SNAPBASE_init(*pt1);
    SNAPBASE_init(*pt2);
    SNAPBASE_init(*pt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSr), SNAPMTX_cols(matSr), SNAPMTX_elems(matSr), SNAPMTX_csize(matSr), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#else
  IMPLEMENT HERE!!!
#endif // defined

  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSr, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_ghlgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#else
  IMPLEMENT HERE!!!
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-r );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-r, SNAP_degree(poly_inf)-r );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, r) == 0 ) {
    //#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
    //#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_GHLGCDEX_REFINEMENT_UVGCD ) {
      if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
      ) {
        result_code = -1;
      }
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  SNAPBASEREAL_clear(sigma_m1);
  if ( mode_refinement == SNAP_GHLGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  for ( i=0, pt1=tau, pt2=zvect, pt3=yvect; i < SNAPMTX_csize(matSr); i++, pt1++, pt2++, pt3++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
    SNAPBASE_clear(*pt3);
  }
  free(zvect);
  free(yvect);
  free(tau);
post:
  // free the memories.
  SNAPBASEREAL_clear(residue_max);
  SNAPBASEREAL_clear(norm_z);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPMTX_clear(vecL);
  SNAPMTX_clear(matL);
#ifdef SNAP_SUPD_T
  free(jpvt);
#endif // SNAP_SUPD_T
  SNAPMTX_clear(vecN);
  SNAPMTX_clear(matA);
  SNAPMTX_clear(vecZ);
  SNAPMTXREAL_clear(vecS);
  SNAPMTX_clear(matH);
  SNAPMTX_clear(matVT);
  SNAPMTX_clear(matU);
  SNAPMTX_clear(matSr);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}


/* ================== END[agcd_ghlgcd] ================== */

























