/*
 * Copyright (c) 2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

//#define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.5le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPDC_T)
void show_sdmdc(sdmdc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmdc_rows(mtx); i++) {
    for(j=0; j<sdmdc_cols(mtx); j++) {
      printf("(%+.5le%+.5le*I), ",creal(sdmdc_elem(mtx,i,j)),cimag(sdmdc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_dclist(dcomplex *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.5le%+.5le*I), ",creal(*(list+i)),cimag(*(list+i)));
  }
  printf("\n");
}
#endif // defined
#endif // SNAPLIB_DEBUG

/* ================== START [agcd_gpgcd] ================== */

// A.Terui uses the followings in his paper (100 or 200 with 1.0e-8)
#define SNAPLIB_GPGCD_MAX_ITERATIONS 100
#define SNAPLIB_GPGCD_THRESHOLD 1.0e-8

// dgesvd or dgesdd (maybe sdd is faster than svd)
// note: dgesvdx may be the best choice for the last singular value. however lapacke has a bug.
// #define SNAPLIB_GPGCD_USE_DGESVD

// define this if you do not want to use SVD to compute the pseudo-inverse of Jacobian matrix.
#define SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD

/*!
 *  @brief Subresultant mapping matrix w.r.t. f and g of order r (Sylvester matrix is r=0).
 *  @warning NOTE THAT THIS FUNCTION DOES NOT FILL ZEROS!!
 *  @param[in,out] matrix The sdm*_t variable which will be the subresultant matrix.
 *  @param[in] input_f The input polynomial f[x] as real vector.
 *  @param[in] input_g The input polynomial g[x] as real vector.
 *  @param[in] r The order of Subresultant mapping.
 */
void SNAP_agcd_gpgcd_subresultant_mapping_matrix_dsc(SNAPMTXREAL_t matrix, SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, unsigned int r) {
  unsigned int i, j, rows, cols;
  SNAPBASEREAL_t *pt1, *pt2;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  unsigned int hcols, hrows, fcols, gcols;
  SNAPBASEREAL_t *pt3, *pt4, *pt5;
#endif
  // the matrix size will be rows x cols.
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  fcols = SNAPMTXREAL_cols(input_f)/2;
  gcols = SNAPMTXREAL_cols(input_g)/2;
  hcols = fcols + gcols - 2*r - 2;
  hrows = hcols + r;
  cols = hcols*2;
  rows = hrows*2;
#else
  cols = SNAPMTXREAL_cols(input_f) + SNAPMTXREAL_cols(input_g) - 2*r - 2;
  rows = cols + r;
#endif
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTXREAL_rsize(matrix) || cols > SNAPMTXREAL_csize(matrix) ) {
    SNAPMTXREAL_realloc(matrix, rows, cols);
  } else {
    SNAPMTXREAL_rows(matrix) = rows;
    SNAPMTXREAL_cols(matrix) = cols;
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
  // poly_f
  for ( i=0; i<SNAPMTXREAL_cols(input_g)-r-1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matrix,i,i), pt2=SNAPMTXREAL_elems(input_f); j<SNAPMTXREAL_cols(input_f); j++,pt1+=SNAPMTXREAL_csize(matrix),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
  // poly_g
  for ( i=0; i<SNAPMTXREAL_cols(input_f)-r-1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matrix,i,i+SNAPMTXREAL_cols(input_g)-r-1), pt2=SNAPMTXREAL_elems(input_g); j<SNAPMTXREAL_cols(input_g); j++,pt1+=SNAPMTXREAL_csize(matrix),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
#else
  // poly_f
  for ( i=0; i<gcols-r-1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matrix,i,i), pt2=SNAPMTXREAL_elemp(matrix,i+hrows,i), pt3=SNAPMTXREAL_elemp(matrix,i,i+hcols), pt4=SNAPMTXREAL_elemp(matrix,i+hrows,i+hcols), pt5=SNAPMTXREAL_elems(input_f);
          j<fcols;
          j++,pt1+=SNAPMTXREAL_csize(matrix),pt2+=SNAPMTXREAL_csize(matrix),pt3+=SNAPMTXREAL_csize(matrix),pt4+=SNAPMTXREAL_csize(matrix),pt5+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt5);
      SNAPBASEREAL_set(*pt4, *pt5);
      SNAPBASEREAL_set(*pt2, *(pt5+1));
      SNAPBASEREAL_neg(*pt3, *(pt5+1));
    }
  }
  // poly_g
  for ( i=0; i<fcols-r-1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matrix,i,i+gcols-r-1), pt2=SNAPMTXREAL_elemp(matrix,i+hrows,i+gcols-r-1), pt3=SNAPMTXREAL_elemp(matrix,i,i+hcols+gcols-r-1), pt4=SNAPMTXREAL_elemp(matrix,i+hrows,i+hcols+gcols-r-1), pt5=SNAPMTXREAL_elems(input_g);
          j<gcols;
          j++,pt1+=SNAPMTXREAL_csize(matrix),pt2+=SNAPMTXREAL_csize(matrix),pt3+=SNAPMTXREAL_csize(matrix),pt4+=SNAPMTXREAL_csize(matrix),pt5+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt5);
      SNAPBASEREAL_set(*pt4, *pt5);
      SNAPBASEREAL_set(*pt2, *(pt5+1));
      SNAPBASEREAL_neg(*pt3, *(pt5+1));
    }
  }
#endif
}

/*!
 *  @brief finding initial cofactors in GPGCD algorithm for two univariate polynomials.
 *  @param[in] input_f The input polynomial f[x] as real vector.
 *  @param[in] input_g The input polynomial g[x] as real vector.
 *  @param[out] cofact_a The initial cofactor's real vector for poly_ing, and should be allocated in advance.
 *  @param[out] cofact_b The initial cofactor's real vector for poly_inf, and should be allocated in advance.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_initial_cofactors(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int d) {
  SNAPMTXREAL_t matN;
  SNAPBASEREAL_t *pt1, *pt2;
  unsigned int i;
  int result_code = 0;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  double *superb;
#endif // SNAPLIB_GPGCD_USE_DGESVD
  SNAPMTXREAL_t vecS;
  SNAPMTXREAL_t matVT;
#endif // defined
  SNAPMTXREAL_init(matN);
  // construct the subresultant matrix with descending order.
  SNAP_agcd_gpgcd_subresultant_mapping_matrix_dsc(matN, input_f, input_g, d-1);
#ifdef SNAPLIB_DEBUG
printf("----matN---->\n");
show_sdmd(matN);
printf("<----matN----\n");
#endif // SNAPLIB_DEBUG
  // compute SVD but in part by LAPACK
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  SNAPMTXREAL_init2(vecS, 1, SNAPMTXREAL_cols(matN));
  SNAPMTXREAL_init2(matVT, SNAPMTXREAL_cols(matN), SNAPMTXREAL_cols(matN));
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  superb = (double*) malloc(sizeof(double)*SNAPMTXREAL_cols(matN));
  if ( LAPACKE_dgesvd( LAPACK_ROW_MAJOR, 'N', 'A', SNAPMTXREAL_rows(matN), SNAPMTXREAL_cols(matN), SNAPMTXREAL_elems(matN), SNAPMTXREAL_csize(matN),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT), superb ) != 0 ) {
     printf("ERROR in agcd_gpgcd_initial_cofactors: dgesvd could not compute the SVD.\n");
     result_code = -1;
     goto dest;
  }
#else
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTXREAL_rows(matN), SNAPMTXREAL_cols(matN), SNAPMTXREAL_elems(matN), SNAPMTXREAL_csize(matN),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT) ) != 0 ) {
     printf("ERROR in agcd_gpgcd_initial_cofactors: dgesdd could not compute the SVD.\n");
     result_code = -1;
     goto dest;
  }
#endif // SNAPLIB_GPGCD_USE_DGESVD
#ifdef SNAPLIB_DEBUG
printf("----matVT---->\n");
show_sdmd(matVT);
printf("<----matVT----\n");
#endif // SNAPLIB_DEBUG

#ifdef SNAP_SUPD_T
  for ( i=0, pt1=SNAPMTXREAL_elemp(matVT,SNAPMTXREAL_cols(matN)-1, 0), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(matN); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
#else
  for ( i=0, pt1=SNAPMTXREAL_elemp(matVT,SNAPMTXREAL_cols(matN)-1, 0), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_a)+1; i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b)+1; i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
#endif
#else
  IMPLEMENTE HERE
#endif
dest:
  SNAPMTXREAL_clear(matN);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  free(superb);
#endif // SNAPLIB_GPGCD_USE_DGESVD
  SNAPMTXREAL_clear(vecS);
  SNAPMTXREAL_clear(matVT);
#endif // defined
  return result_code;
}

/*!
 *  @brief constructing initial real vectors in GPGCD algorithm for two univariate polynomials.
 *  @param[in] poly_inf The input polynomial f[x] (asc).
 *  @param[in] poly_ing The input polynomial g[x] (asc).
 *  @param[out] tilde_f The real desc. vector for poly_inf, and should be allocated in advance.
 *  @param[out] tilde_g The real desc. vector for poly_ing, and should be allocated in advance.
 *  @return 0 if success or -1 if failure.
 */
void SNAP_agcd_gpgcd_construct_tildevectors(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g) {
  unsigned int i;
  SNAPBASE_t *pt1;
  SNAPBASEREAL_t *pt2;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
  for ( i = 0, pt1 = SNAP_coefs(poly_inf), pt2 = SNAPMTX_elems(tilde_f) + SNAP_degree(poly_inf); i <= SNAP_degree(poly_inf); i++, pt1++, pt2-- ) {
    SNAPBASE_set(*pt2, *pt1);
  }
  for ( i = 0, pt1 = SNAP_coefs(poly_ing), pt2 = SNAPMTX_elems(tilde_g) + SNAP_degree(poly_ing); i <= SNAP_degree(poly_ing); i++, pt1++, pt2-- ) {
    SNAPBASE_set(*pt2, *pt1);
  }
#elif defined(SNAP_SUPDC_T)
  for ( i = 0, pt1 = SNAP_coefs(poly_inf), pt2 = SNAPMTX_elems(tilde_f) + 2*SNAP_degree(poly_inf); i <= SNAP_degree(poly_inf); i++, pt1++, pt2-=2 ) {
    SNAPBASEREAL_set(*pt2, SNAPBASE_real(*pt1));
    SNAPBASEREAL_set(*(pt2+1), SNAPBASE_imag(*pt1));
  }
  for ( i = 0, pt1 = SNAP_coefs(poly_ing), pt2 = SNAPMTX_elems(tilde_g) + 2*SNAP_degree(poly_ing); i <= SNAP_degree(poly_ing); i++, pt1++, pt2-=2 ) {
    SNAPBASEREAL_set(*pt2, SNAPBASE_real(*pt1));
    SNAPBASEREAL_set(*(pt2+1), SNAPBASE_imag(*pt1));
  }
#else
  for ( i = 0, pt1 = SNAP_coefs(poly_inf), pt2 = SNAPMTX_elems(tilde_f) + 2*SNAP_degree(poly_inf); i <= SNAP_degree(poly_inf); i++, pt1++, pt2-=2 ) {
    SNAPBASEREAL_set(*pt2, SNAPBASE_realref(*pt1));
    SNAPBASEREAL_set(*(pt2+1), SNAPBASE_imagref(*pt1));
  }
  for ( i = 0, pt1 = SNAP_coefs(poly_ing), pt2 = SNAPMTX_elems(tilde_g) + 2*SNAP_degree(poly_ing); i <= SNAP_degree(poly_ing); i++, pt1++, pt2-=2 ) {
    SNAPBASEREAL_set(*pt2, SNAPBASE_realref(*pt1));
    SNAPBASEREAL_set(*(pt2+1), SNAPBASE_imagref(*pt1));
  }
#endif
}

/*!
 *  @brief construct Jacobian matrix in GPGCD algorithm for two univariate polynomials.
 *  @param[out] matJ The Jacobian matrix (should be allocated in advance).
 *  @param[in] tilde_f The perturbed real vector for poly_inf.
 *  @param[in] tilde_g The perturbed real vector for poly_ing.
 *  @param[in] cofact_a The cofactor's real vector for poly_ing.
 *  @param[in] cofact_b The cofactor's real vector for poly_inf.
 *  @param[in] m The degree of poly_inf.
 *  @param[in] n The degree of poly_ing.
 *  @param[in] d The degree of approximate GCD.
 */
void SNAP_agcd_gpgcd_Jacobian(SNAPMTXREAL_t matJ, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int m, const unsigned int n, const unsigned int d) {
  unsigned int i, j;
  SNAPBASEREAL_t *pt1, *pt2;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_t *pt3;
  SNAPMTXREAL_reinit2(matJ, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  // A_1 (A)
  for ( i=0; i<=m; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i), pt2=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+m+n+2), pt3=SNAPMTXREAL_elems(cofact_a); j<SNAPMTXREAL_cols(cofact_a)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_set(*pt2, *pt3);
    }
  }
  // A_1 (B)
  for ( i=0; i<=n; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+m+1), pt2=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+2*m+n+3), pt3=SNAPMTXREAL_elems(cofact_b); j<SNAPMTXREAL_cols(cofact_b)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_set(*pt2, *pt3);
    }
  }
  // A_2 (A)
  for ( i=0; i<=m; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i), pt2=SNAPMTXREAL_elemp(matJ,i+1,i+m+n+2), pt3=SNAPMTXREAL_elems(cofact_a)+1; j<SNAPMTXREAL_cols(cofact_a)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_neg(*pt2, *pt3);
    }
  }
  // A_2 (B)
  for ( i=0; i<=n; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+m+1), pt2=SNAPMTXREAL_elemp(matJ,i+1,i+2*m+n+3), pt3=SNAPMTXREAL_elems(cofact_b)+1; j<SNAPMTXREAL_cols(cofact_b)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_neg(*pt2, *pt3);
    }
  }
  // N_1(F)
  for ( i=0; i<n-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+2*m+2*n+4), pt2=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+3*m+3*n-2*d+6), pt3=SNAPMTXREAL_elems(tilde_f); j<SNAPMTXREAL_cols(tilde_f)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_set(*pt2, *pt3);
    }
  }
  // N_1(G)
  for ( i=0; i<m-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+2*m+3*n-d+5), pt2=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+3*m+4*n-3*d+7), pt3=SNAPMTXREAL_elems(tilde_g); j<SNAPMTXREAL_cols(tilde_g)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_set(*pt2, *pt3);
    }
  }
  // N_2(F)
  for ( i=0; i<n-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+2*m+2*n+4), pt2=SNAPMTXREAL_elemp(matJ,i+1,i+3*m+3*n-2*d+6), pt3=SNAPMTXREAL_elems(tilde_f)+1; j<SNAPMTXREAL_cols(tilde_f)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_neg(*pt2, *pt3);
    }
  }
  // N_2(G)
  for ( i=0; i<m-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+m+n-d+2,i+2*m+3*n-d+5), pt2=SNAPMTXREAL_elemp(matJ,i+1,i+3*m+4*n-3*d+7), pt3=SNAPMTXREAL_elems(tilde_g)+1; j<SNAPMTXREAL_cols(tilde_g)/2;
          j++, pt1+=SNAPMTXREAL_csize(matJ), pt2+=SNAPMTXREAL_csize(matJ), pt3+=2 ) {
      SNAPBASEREAL_set(*pt1, *pt3);
      SNAPBASEREAL_neg(*pt2, *pt3);
    }
  }
  // 2v
  for ( i=0, pt1=SNAPMTXREAL_elemp(matJ, 0, 2*m+2*n+4), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_a)+1; i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b)+1; i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
#else
  SNAPMTXREAL_reinit2(matJ, m+n-d+2, 2*(m+n-d+2));
  // Cm(A)
  for ( i=0; i<=m; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i), pt2=SNAPMTXREAL_elems(cofact_a); j<SNAPMTXREAL_cols(cofact_a); j++,pt1+=SNAPMTXREAL_csize(matJ),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
  // Cn(B)
  for ( i=0; i<=n; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+m+1), pt2=SNAPMTXREAL_elems(cofact_b); j<SNAPMTXREAL_cols(cofact_b); j++,pt1+=SNAPMTXREAL_csize(matJ),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
  // N_{d-1}(F)
  for ( i=0; i<n-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+m+n+2), pt2=SNAPMTXREAL_elems(tilde_f); j<SNAPMTXREAL_cols(tilde_f); j++,pt1+=SNAPMTXREAL_csize(matJ),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
  // N_{d-1}(G)
  for ( i=0; i<m-d+1; i++ ) {
    for ( j=0, pt1=SNAPMTXREAL_elemp(matJ,i+1,i+m+2*n-d+3), pt2=SNAPMTXREAL_elems(tilde_g); j<SNAPMTXREAL_cols(tilde_g); j++,pt1+=SNAPMTXREAL_csize(matJ),pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
  }
  // 2v
  for ( i=0, pt1=SNAPMTXREAL_elemp(matJ, 0, m+n+2), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(cofact_b); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_mul_ui(*pt1, *pt2, 2);
  }
#endif // defined
}

/*!
 *  @brief construct Nabla vector (gradient) in GPGCD algorithm for two univariate polynomials.
 *  @warning This computes -\nabla f(v) where f=(sum_i (tilde_i-input_i)^2)/2.
 *  @param[out] vecN The Nabla vector (gradient) (should be allocated in advance).
 *  @param[in] input_f The real vector for poly_inf.
 *  @param[in] input_g The real vector for poly_ing.
 *  @param[in] tilde_f The perturbed real vector for poly_inf.
 *  @param[in] tilde_g The perturbed real vector for poly_ing.
 *  @param[in] m The degree of poly_inf.
 *  @param[in] n The degree of poly_ing.
 *  @param[in] d The degree of approximate GCD.
 */
void SNAP_agcd_gpgcd_nabla(SNAPMTXREAL_t vecN, SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, const unsigned int m, const unsigned int n, const unsigned int d) {
  unsigned int i;
  SNAPBASEREAL_t *pt1, *pt2, *pt3;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  for ( i=0, pt1=SNAPMTXREAL_elems(vecN), pt2=SNAPMTXREAL_elems(input_f), pt3=SNAPMTXREAL_elems(tilde_f); i<=m; i++,pt1++,pt2+=2,pt3+=2 ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(input_g), pt3=SNAPMTXREAL_elems(tilde_g); i<=n; i++,pt1++,pt2+=2,pt3+=2 ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(input_f)+1, pt3=SNAPMTXREAL_elems(tilde_f)+1; i<=m; i++,pt1++,pt2+=2,pt3+=2 ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(input_g)+1, pt3=SNAPMTXREAL_elems(tilde_g)+1; i<=n; i++,pt1++,pt2+=2,pt3+=2 ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0; i<2*n+2*m-4*d+4; i++, pt1++ ) {
    SNAPBASEREAL_set_ui(*pt1, 0);
  }
#else
  for ( i=0, pt1=SNAPMTXREAL_elems(vecN), pt2=SNAPMTXREAL_elems(input_f), pt3=SNAPMTXREAL_elems(tilde_f); i<=m; i++,pt1++,pt2++,pt3++ ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(input_g), pt3=SNAPMTXREAL_elems(tilde_g); i<=n; i++,pt1++,pt2++,pt3++ ) {
    SNAPBASEREAL_sub(*pt1, *pt2, *pt3);
  }
  for ( i=0; i<n+m-2*d+2; i++, pt1++ ) {
    SNAPBASEREAL_set_ui(*pt1, 0);
  }
#endif // defined
}

/*!
 *  @brief construct constraint vector in GPGCD algorithm for two univariate polynomials.
 *  @param[out] vecQ The constraint vector.
 *  @param[in] matJ The Jacobian matrix.
 *  @param[in] cofact_a The cofactor's real vector for poly_ing.
 *  @param[in] cofact_b The cofactor's real vector for poly_inf.
 *  @param[in] m The degree of poly_inf.
 *  @param[in] n The degree of poly_ing.
 *  @param[in] d The degree of approximate GCD.
 */
void SNAP_agcd_gpgcd_const(SNAPMTXREAL_t vecQ, SNAPMTXREAL_t matJ, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int m, const unsigned int n, const unsigned int d) {
  SNAPBASEREAL_t tmp;
  // q_0
  SNAPBASEREAL_init(tmp);
  SNAPMTXREAL_BLAS_dot(tmp, SNAPMTXREAL_cols(cofact_a), SNAPMTXREAL_elems(cofact_a), 1, SNAPMTXREAL_elems(cofact_a), 1);
  SNAPMTXREAL_BLAS_dot(SNAPMTXREAL_elem(vecQ,0,0), SNAPMTXREAL_cols(cofact_b), SNAPMTXREAL_elems(cofact_b), 1, SNAPMTXREAL_elems(cofact_b), 1);
  SNAPBASEREAL_add(SNAPMTXREAL_elem(vecQ,0,0), SNAPMTXREAL_elem(vecQ,0,0), tmp);
  SNAPBASEREAL_sub_ui(SNAPMTXREAL_elem(vecQ,0,0), SNAPMTXREAL_elem(vecQ,0,0), 1);
  // rest of q_i
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
#ifdef SNAP_SUPDC_T
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_a)/2, 1.0, SNAPMTXREAL_elemp(matJ, 1, 2*m+2*n+4),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_a), 2, 0.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_b)/2, 1.0, SNAPMTXREAL_elemp(matJ, 1, 2*m+3*n-d+5),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_b), 2, 1.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_a)/2, 1.0, SNAPMTXREAL_elemp(matJ, 1, 3*m+3*n-2*d+6),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_a)+1, 2, 1.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_b)/2, 1.0, SNAPMTXREAL_elemp(matJ, 1, 3*m+4*n-3*d+7),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_b)+1, 2, 1.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
#else
  IMPLEMENT HERE
#endif // SNAP_SUPDC_T
#else
#ifdef SNAP_SUPD_T
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_a), 1.0, SNAPMTXREAL_elemp(matJ, 1, m+n+2),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_a), 1, 0.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(vecQ)-1, SNAPMTXREAL_cols(cofact_b), 1.0, SNAPMTXREAL_elemp(matJ, 1, m+2*n-d+3),
              SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(cofact_b), 1, 1.0, SNAPMTXREAL_elemp(vecQ,0,1), 1);
#else
  IMPLEMENT HERE
#endif // SNAP_SUPD_T
#endif // defined
}

/*!
 *  @brief construct search direction vector in GPGCD algorithm for two univariate polynomials.
 *  @param[in,out] vecD The temporal vector.
 *  @param[in] matJ The Jacobian matrix..
 *  @param[in] matJP The pseudo-inverse of Jacobian matrix.
 *  @param[in,out] vecN The nabla (gradient) vector on input, and the search direction vector on exit.
 *  @param[in] vecQ The constraint vector.
 *  @param[in] m The degree of poly_inf.
 *  @param[in] n The degree of poly_ing.
 *  @param[in] d The degree of approximate GCD.
 */
void SNAP_agcd_gpgcd_direction(SNAPMTXREAL_t vecD, SNAPMTXREAL_t matJ, SNAPMTXREAL_t matJP, SNAPMTXREAL_t vecN, SNAPMTXREAL_t vecQ, const unsigned int m, const unsigned int n, const unsigned int d) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  // vecD <== matJ vecN
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matJ), SNAPMTXREAL_cols(matJ), 1.0, SNAPMTXREAL_elems(matJ), SNAPMTXREAL_csize(matJ), SNAPMTXREAL_elems(vecN), 1, 0.0, SNAPMTXREAL_elems(vecD), 1);
  // vecN <== -1.0 * matJP vecD + vecN
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matJP), SNAPMTXREAL_cols(matJP), -1.0, SNAPMTXREAL_elems(matJP), SNAPMTXREAL_csize(matJP), SNAPMTXREAL_elems(vecD), 1, 1.0, SNAPMTXREAL_elems(vecN), 1);
  // vecN <== -1.0 * matJP vecQ + vecN
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matJP), SNAPMTXREAL_cols(matJP), -1.0, SNAPMTXREAL_elems(matJP), SNAPMTXREAL_csize(matJP), SNAPMTXREAL_elems(vecQ), 1, 1.0, SNAPMTXREAL_elems(vecN), 1);
#else
  IMPLEMENT HERE
#endif
}

/*!
 *  @brief construct search direction vector in GPGCD algorithm for two univariate polynomials.
 *  @param[in,out] tilde_f The current perturbed real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in,out] tilde_g The current perturbed real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_a The current cofactor's real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_b The current cofactor's real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in] alpha The step size.
 *  @param[in] vecD The correction vector.
 */
void SNAP_agcd_gpgcd_update(SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, SNAPBASEREAL_t alpha, SNAPMTXREAL_t vecD) {
#ifdef SNAP_SUPD_T
  cblas_daxpy(SNAPMTXREAL_cols(tilde_f), alpha, SNAPMTXREAL_elemp(vecD, 0, 0), 1, SNAPMTXREAL_elems(tilde_f), 1);
  cblas_daxpy(SNAPMTXREAL_cols(tilde_g), alpha, SNAPMTXREAL_elemp(vecD, 0, SNAPMTXREAL_cols(tilde_f)), 1, SNAPMTXREAL_elems(tilde_g), 1);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_a), alpha, SNAPMTXREAL_elemp(vecD, 0, SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)), 1, SNAPMTXREAL_elems(cofact_a), 1);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_b), alpha, SNAPMTXREAL_elemp(vecD, 0, SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)+SNAPMTXREAL_cols(cofact_a)), 1, SNAPMTXREAL_elems(cofact_b), 1);
#elif defined(SNAP_SUPDC_T)
  cblas_daxpy(SNAPMTXREAL_cols(tilde_f)/2, alpha, SNAPMTXREAL_elemp(vecD,0,0), 1, SNAPMTXREAL_elemp(tilde_f,0,0), 2);
  cblas_daxpy(SNAPMTXREAL_cols(tilde_g)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)/2), 1, SNAPMTXREAL_elemp(tilde_g,0,0), 2);
  cblas_daxpy(SNAPMTXREAL_cols(tilde_f)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)/2+SNAPMTXREAL_cols(tilde_g)/2), 1, SNAPMTXREAL_elemp(tilde_f,0,1), 2);
  cblas_daxpy(SNAPMTXREAL_cols(tilde_g)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)/2), 1, SNAPMTXREAL_elemp(tilde_g,0,1), 2);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_a)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)), 1, SNAPMTXREAL_elemp(cofact_a,0,0), 2);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_b)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)+SNAPMTXREAL_cols(cofact_a)/2), 1, SNAPMTXREAL_elemp(cofact_b,0,0), 2);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_a)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)+SNAPMTXREAL_cols(cofact_a)/2+SNAPMTXREAL_cols(cofact_b)/2), 1, SNAPMTXREAL_elemp(cofact_a,0,1), 2);
  cblas_daxpy(SNAPMTXREAL_cols(cofact_b)/2, alpha, SNAPMTXREAL_elemp(vecD,0,SNAPMTXREAL_cols(tilde_f)+SNAPMTXREAL_cols(tilde_g)+SNAPMTXREAL_cols(cofact_a)+SNAPMTXREAL_cols(cofact_b)/2), 1, SNAPMTXREAL_elemp(cofact_b,0,1), 2);
#else
  IMPLEMENT HERE
#endif
}

/*!
 *  @brief pseudo-inverse in GPGCD algorithm for two univariate polynomials.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_pseudo_inverse_by_svd(SNAPMTXREAL_t matJP, SNAPMTXREAL_t matJC, SNAPMTXREAL_t vecSV, SNAPMTXREAL_t matU, SNAPMTXREAL_t matVT, double *superb) {
  SNAPBASEREAL_t tmp;
  unsigned int i;
  SNAPBASEREAL_t *pt1, *pt2;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  if ( LAPACKE_dgesvd(LAPACK_ROW_MAJOR, 'S', 'S', SNAPMTXREAL_rows(matJC), SNAPMTXREAL_cols(matJC), SNAPMTXREAL_elems(matJC), SNAPMTXREAL_csize(matJC),
                      SNAPMTXREAL_elems(vecSV), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT), superb) != 0 ) {
    printf("ERROR in agcd_gpgcd_pseudo_inverse_by_svd: dgesvd could not compute the SVD.\n");
    return -1;
  }
#else
  if ( LAPACKE_dgesdd(LAPACK_ROW_MAJOR, 'S', SNAPMTXREAL_rows(matJC), SNAPMTXREAL_cols(matJC), SNAPMTXREAL_elems(matJC), SNAPMTXREAL_csize(matJC),
                      SNAPMTXREAL_elems(vecSV), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT)) != 0 ) {
    printf("ERROR in agcd_gpgcd_pseudo_inverse_by_svd: dgesdd could not compute the SVD.\n");
    return -1;
  }
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  IMPLEMENT HERE
#endif
#ifdef SNAPLIB_DEBUG
printf("----matU---->\n");
show_sdmd(matU);
printf("----matVT---->\n");
show_sdmd(matVT);
#endif // SNAPLIB_DEBUG
  // form the pseudo-inverse of matJ
  SNAPBASEREAL_init(tmp);
  for ( i=0, pt1=SNAPMTXREAL_elems(matVT), pt2=SNAPMTXREAL_elems(vecSV); i<SNAPMTXREAL_rows(matVT); i++,pt1+=SNAPMTXREAL_csize(matVT),pt2++ ) {
    SNAPBASEREAL_rec(tmp, *pt2);
    SNAPMTXREAL_BLAS_scal(SNAPMTXREAL_cols(matVT), tmp, pt1, 1);
  }
  SNAPBASEREAL_clear(tmp);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  cblas_dgemm(CblasRowMajor, CblasTrans, CblasTrans, SNAPMTXREAL_rows(matJP), SNAPMTXREAL_cols(matJP), SNAPMTXREAL_cols(matJP), 1.0,
              SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), 0.0, SNAPMTXREAL_elems(matJP), SNAPMTXREAL_csize(matJP));
#else
  IMPLEMENT HERE
#endif // defined
#ifdef SNAPLIB_DEBUG
printf("----matJP---->\n");
show_sdmd(matJP);
#endif // SNAPLIB_DEBUG
  return 0;
}

/*!
 *  @brief pseudo-inverse in GPGCD algorithm for two univariate polynomials.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_pseudo_inverse_by_tri(SNAPMTXREAL_t matJP, SNAPMTXREAL_t matJC, SNAPMTXREAL_t matU, lapack_int *ipiv) {
  // JP = transpose(JC).inverse(JC.transpose(JC))
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, SNAPMTXREAL_rows(matJC), SNAPMTXREAL_rows(matJC), SNAPMTXREAL_cols(matJC), 1.0, SNAPMTXREAL_elems(matJC),
              SNAPMTXREAL_csize(matJC), SNAPMTXREAL_elems(matJC), SNAPMTXREAL_csize(matJC), 0.0, SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU));
  if ( LAPACKE_dgetrf(LAPACK_ROW_MAJOR, SNAPMTXREAL_rows(matU), SNAPMTXREAL_cols(matU), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), ipiv) != 0 ) {
    printf("ERROR in agcd_gpgcd_pseudo_inverse_by_tri: dgetrf could not compute the LU.\n");
    return -1;
  }
  if ( LAPACKE_dgetri(LAPACK_ROW_MAJOR, SNAPMTXREAL_rows(matU), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), ipiv) != 0 ) {
    printf("ERROR in agcd_gpgcd_pseudo_inverse_by_tri: dgetri could not compute the inverse.\n");
    return -1;
  }
  cblas_dgemm(CblasRowMajor, CblasTrans, CblasNoTrans, SNAPMTXREAL_cols(matJC), SNAPMTXREAL_cols(matU), SNAPMTXREAL_rows(matJC), 1.0, SNAPMTXREAL_elems(matJC),
              SNAPMTXREAL_csize(matJC), SNAPMTXREAL_elems(matU), SNAPMTXREAL_csize(matU), 0.0, SNAPMTXREAL_elems(matJP), SNAPMTXREAL_csize(matJP));
  return 0;
}

/*!
 *  @brief modified newton method in GPGCD algorithm for two univariate polynomials.
 *  @param[in] input_f The input polynomial f[x] as real vector.
 *  @param[in] input_g The input polynomial g[x] as real vector.
 *  @param[in,out] tilde_f The initial perturbed real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in,out] tilde_g The initial perturbed real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_a The initial cofactor's real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_b The initial cofactor's real vector for poly_inf on input, and the resulting vector on exit.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_modified_newton(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b) {
  int result_code = 0;
  unsigned int m, n, d, num_iterations;
  SNAPMTXREAL_t matJ, matJP, matU, vecN, vecQ, vecD;
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  unsigned int i;
  SNAPBASEREAL_t *pt1, *pt2;
  SNAPMTXREAL_t matJC, matVT, vecSV;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPBASEREAL_t tmp, alpha;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  double *superb;
#else
  lapack_int *ipiv;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  m = SNAPMTXREAL_cols(input_f)/2-1;
  n = SNAPMTXREAL_cols(input_g)/2-1;
  d = SNAPMTXREAL_cols(input_f)/2-SNAPMTXREAL_cols(cofact_b)/2;
#else
  m = SNAPMTXREAL_cols(input_f)-1;
  n = SNAPMTXREAL_cols(input_g)-1;
  d = SNAPMTXREAL_cols(input_f)-SNAPMTXREAL_cols(cofact_b);
#endif // defined
  SNAPBASEREAL_init(tmp);
  SNAPBASEREAL_init(alpha);
  SNAPMTXREAL_init(matJ);
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  SNAPMTXREAL_init2(matJP, 2*(m+n+2)+2*(m+n-2*d+2), 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(matU, 2*(m+n-d+1)+1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecQ, 1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n+2)+2*(m+n-2*d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(matVT, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecSV, 1, 2*(m+n-d+1)+1);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#else
  SNAPMTXREAL_init2(matJP, 2*(m+n-d+2), m+n-d+2);
  SNAPMTXREAL_init2(matU, m+n-d+2, m+n-d+2);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecQ, 1, m+n-d+2);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n-d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(matVT, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecSV, 1, m+n-d+2);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  superb = (double*) malloc(sizeof(double)*(m+n-d+2));
#else
  superb = NULL;
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  ipiv = (lapack_int*) malloc(sizeof(lapack_int)*SNAPMTXREAL_rows(matU));
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

  // do iterations
  num_iterations = 0;
  while (1) {
    // Construct the Jacobian matrix with the current vector
    SNAP_agcd_gpgcd_Jacobian(matJ, tilde_f, tilde_g, cofact_a, cofact_b, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----matJ---->\n");
show_sdmd(matJ);
#endif // SNAPLIB_DEBUG
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
    for ( i=0, pt1=SNAPMTXREAL_elems(matJC), pt2=SNAPMTXREAL_elems(matJ); i<SNAPMTXREAL_csize(matJ)*SNAPMTXREAL_rsize(matJ); i++,pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_svd(matJP, matJC, vecSV, matU, matVT, superb) != 0 ) {
      result_code = -1;
      goto dest;
    }
#else
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_tri(matJP, matJ, matU, ipiv) != 0 ) {
      result_code = -1;
      goto dest;
    }
#endif
    // form the nabla and constraint
    SNAP_agcd_gpgcd_nabla(vecN, input_f, input_g, tilde_f, tilde_g, m, n, d);
    SNAP_agcd_gpgcd_const(vecQ, matJ, cofact_a, cofact_b, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----vecN---->\n");
show_sdmd(vecN);
printf("----vecQ---->\n");
show_sdmd(vecQ);
#endif // SNAPLIB_DEBUG
    // search direction (NOTE: The result is in vecN !!)
    SNAP_agcd_gpgcd_direction(vecD, matJ, matJP, vecN, vecQ, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----vecD---->\n");
show_sdmd(vecN);
#endif // SNAPLIB_DEBUG
    // check the size of the search direction
    SNAPMTXREAL_BLAS_nrm2(tmp, SNAPMTXREAL_cols(vecN), SNAPMTXREAL_elems(vecN), 1);
#ifdef SNAPLIB_DEBUG
printf("the size of search direction: %e\n", tmp);
#endif // SNAPLIB_DEBUG
    if ( SNAPBASEREAL_cmp(tmp, SNAPLIB_GPGCD_THRESHOLD) < 0 ) {
      printf("INFO in agcd_gpgcd_modified_newton: converged within %d iterations.\n", num_iterations);
      break;
    }
    // step size (at first, we follow the step size in the paper)
    SNAPBASEREAL_set_ui(alpha, 1);
    // update
    SNAP_agcd_gpgcd_update(tilde_f, tilde_g, cofact_a, cofact_b, alpha, vecN);
#ifdef SNAPLIB_DEBUG
printf("----tilde_f---->\n");
show_sdmd(tilde_f);
printf("----tilde_g---->\n");
show_sdmd(tilde_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
    // check the number of iterations
    if ( ++num_iterations >= SNAPLIB_GPGCD_MAX_ITERATIONS ) {
      printf("INFO in agcd_gpgcd_modified_newton: didn't converge within %d iterations.\n", SNAPLIB_GPGCD_MAX_ITERATIONS);
      break;
    }
  }
dest:
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  free(superb);
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  free(ipiv);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
  SNAPBASEREAL_clear(tmp);
  SNAPBASEREAL_clear(alpha);
  SNAPMTXREAL_clear(matJ);
  SNAPMTXREAL_clear(matJP);
  SNAPMTXREAL_clear(matU);
  SNAPMTXREAL_clear(vecN);
  SNAPMTXREAL_clear(vecQ);
  SNAPMTXREAL_clear(vecD);
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_clear(matJC);
  SNAPMTXREAL_clear(matVT);
  SNAPMTXREAL_clear(vecSV);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  return result_code;
}

/*!
 *  @brief calculates approximate GCD by LS in GPGCD algorithm for two univariate polynomials.
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] tilde_f The final perturbed real vector for poly_inf.
 *  @param[in] tilde_g The final perturbed real vector for poly_ing.
 *  @param[in] cofact_a The final cofactor's real vector for poly_ing.
 *  @param[in] cofact_b The final cofactor's real vector for poly_inf.
 *  @param[in] m The degree of poly_inf.
 *  @param[in] n The degree of poly_ing.
 *  @param[in] d The degree of approximate GCD.
 */
void SNAP_agcd_gpgcd_lsdiv(SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int m, const unsigned int n, const unsigned int d) {
  unsigned int i;
  SNAP_t poly_h1, poly_h2, poly_f, poly_g, poly_tmp;
  SNAPBASE_t *ptb;
//  SNAPBASE_t one;
  SNAPBASEREAL_t *ptr;
  SNAPBASEREAL_t nrmt, nrm1, nrm2;

//  SNAPBASE_init_set_ui(one ,1);
  SNAPBASEREAL_init(nrmt);
  SNAPBASEREAL_init(nrm1);
  SNAPBASEREAL_init(nrm2);
  SNAP_init(poly_tmp);
  SNAP_init(poly_h1);
  SNAP_init(poly_h2);
  SNAP_init2(poly_f, m+1);
  SNAP_degree(poly_f) = m;
  SNAP_init2(poly_g, n+1);
  SNAP_degree(poly_g) = n;
  SNAP_clear(poly_f1);
  SNAP_init2(poly_f1, m-d+1);
  SNAP_degree(poly_f1) = m-d;
  SNAP_clear(poly_g1);
  SNAP_init2(poly_g1, n-d+1);
  SNAP_degree(poly_g1) = n-d;
  // restore the coefficients as polys
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPFR_T)
  for ( i=0, ptb=SNAP_coefs(poly_f), ptr=SNAPMTXREAL_elemp(tilde_f,0,m); i<=m; i++,ptb++,ptr-- ) {
    SNAPBASE_set_basereal(*ptb, *ptr);
  }
  for ( i=0, ptb=SNAP_coefs(poly_g), ptr=SNAPMTXREAL_elemp(tilde_g,0,n); i<=n; i++,ptb++,ptr-- ) {
    SNAPBASE_set_basereal(*ptb, *ptr);
  }
  for ( i=0, ptb=SNAP_coefs(poly_g1), ptr=SNAPMTXREAL_elemp(cofact_a,0,n-d); i<=n-d; i++,ptb++,ptr-- ) {
    SNAPBASE_set_basereal(*ptb, *ptr);
  }
  for ( i=0, ptb=SNAP_coefs(poly_f1), ptr=SNAPMTXREAL_elemp(cofact_b,0,m-d); i<=m-d; i++,ptb++,ptr-- ) {
    SNAPBASE_neg_basereal(*ptb, *ptr);
  }
#else
  for ( i=0, ptb=SNAP_coefs(poly_f), ptr=SNAPMTXREAL_elemp(tilde_f,0,2*m); i<=m; i++,ptb++,ptr-=2 ) {
    SNAPBASE_set_basereal2(*ptb, *ptr, *(ptr+1));
  }
  for ( i=0, ptb=SNAP_coefs(poly_g), ptr=SNAPMTXREAL_elemp(tilde_g,0,2*n); i<=n; i++,ptb++,ptr-=2 ) {
    SNAPBASE_set_basereal2(*ptb, *ptr, *(ptr+1));
  }
  for ( i=0, ptb=SNAP_coefs(poly_g1), ptr=SNAPMTXREAL_elemp(cofact_a,0,2*(n-d)); i<=n-d; i++,ptb++,ptr-=2 ) {
    SNAPBASE_set_basereal2(*ptb, *ptr, *(ptr+1));
  }
  for ( i=0, ptb=SNAP_coefs(poly_f1), ptr=SNAPMTXREAL_elemp(cofact_b,0,2*(m-d)); i<=m-d; i++,ptb++,ptr-=2 ) {
    SNAPBASE_set_basereal2(*ptb, *ptr, *(ptr+1));
    SNAPBASE_neg(*ptb, *ptb);
  }
#endif
  // compute approximate GCDs by lsdiv
  SNAP_lsdiv_q(poly_h1, poly_f, poly_f1);
  SNAP_lsdiv_q(poly_h2, poly_g, poly_g1);
  // check the condition
  SNAP_mul(poly_tmp, poly_h1, poly_f1);
  SNAP_sub(poly_tmp, poly_f, poly_tmp);
  SNAP_2norm(nrmt, poly_tmp);
  SNAPBASEREAL_mul(nrm1, nrmt, nrmt);
  SNAP_mul(poly_tmp, poly_h1, poly_g1);
  SNAP_sub(poly_tmp, poly_g, poly_tmp);
  SNAP_2norm(nrmt, poly_tmp);
  SNAPBASEREAL_mul(nrmt, nrmt, nrmt);
  SNAPBASEREAL_add(nrm1, nrm1, nrmt);
  SNAP_mul(poly_tmp, poly_h2, poly_f1);
  SNAP_sub(poly_tmp, poly_f, poly_tmp);
  SNAP_2norm(nrmt, poly_tmp);
  SNAPBASEREAL_mul(nrm2, nrmt, nrmt);
  SNAP_mul(poly_tmp, poly_h2, poly_g1);
  SNAP_sub(poly_tmp, poly_g, poly_tmp);
  SNAP_2norm(nrmt, poly_tmp);
  SNAPBASEREAL_mul(nrmt, nrmt, nrmt);
  SNAPBASEREAL_add(nrm2, nrm2, nrmt);
  if ( SNAPBASEREAL_cmp(nrm1, nrm2) < 0 ) {
    SNAP_set(poly_d, poly_h1);
  } else {
    SNAP_set(poly_d, poly_h2);
  }
//  SNAPBASE_clear(one);
  SNAPBASEREAL_clear(nrmt);
  SNAPBASEREAL_clear(nrm1);
  SNAPBASEREAL_clear(nrm2);
  SNAP_clear(poly_tmp);
  SNAP_clear(poly_h1);
  SNAP_clear(poly_h2);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
}

/*!
 *  @brief GPGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the GPGCD algorithm proposed by Akira Terui
 *   in ISSAC 2009, ASCM-MACIS 2009 and so on. (see http://arxiv.org/abs/1207.0630v2)
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d) {
  SNAPMTXREAL_t input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b;
  unsigned int len_f, len_g, len_a, len_b;
  int result_code = 0;

  // the length of real vectors
  // our implementation preserves unknown vector x in tilde_f, tilde_g, cofact_a and cofact_b separetedly.
  len_f = SNAP_degree(poly_inf)+1;
  len_g = SNAP_degree(poly_ing)+1;
  len_a = SNAP_degree(poly_ing)-d+1;
  len_b = SNAP_degree(poly_inf)-d+1;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  len_f *= 2;
  len_g *= 2;
  len_a *= 2;
  len_b *= 2;
#endif // defined
  SNAPMTXREAL_init2(input_f, 1, len_f);
  SNAPMTXREAL_init2(input_g, 1, len_g);
  SNAPMTXREAL_init2(tilde_f, 1, len_f);
  SNAPMTXREAL_init2(tilde_g, 1, len_g);
  SNAPMTXREAL_init2(cofact_a, 1, len_a);
  SNAPMTXREAL_init2(cofact_b, 1, len_b);
  // copy input polynomials
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, input_f, input_g);
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, tilde_f, tilde_g);
  // initial cofactors
  if ( SNAP_agcd_gpgcd_initial_cofactors(input_f, input_g, cofact_a, cofact_b, d) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not compute initial cofactors.\n");
    result_code = -1;
    goto dest;
  }
#ifdef SNAPLIB_DEBUG
printf("----input_f---->\n");
show_sdmd(input_f);
printf("----input_g---->\n");
show_sdmd(input_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
  // iterations
  if ( SNAP_agcd_gpgcd_modified_newton(input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not proceed the modified newton iterations.\n");
    result_code = -1;
    goto dest;
  }
  // form the approximate GCD
  SNAP_agcd_gpgcd_lsdiv(poly_d, poly_f1, poly_g1, tilde_f, tilde_g, cofact_a, cofact_b, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d);
#ifdef SNAPLIB_DEBUG
show_dlist(SNAP_coefs(poly_d), d+1);
printf("=== END ===\n");
#endif // SNAPLIB_DEBUG
dest:
  SNAPMTXREAL_clear(input_f);
  SNAPMTXREAL_clear(input_g);
  SNAPMTXREAL_clear(tilde_f);
  SNAPMTXREAL_clear(tilde_g);
  SNAPMTXREAL_clear(cofact_a);
  SNAPMTXREAL_clear(cofact_b);

  return result_code;
}

/*!
 *  @brief finding initial cofactors in GPGCD algorithm for two univariate polynomials.
 *  @param[in,out] tilde_f The input polynomial f[x] as real vector on input, and the initial vector on exit.
 *  @param[in,out] tilde_g The input polynomial g[x] as real vector on input, and the initial vector on exit.
 *  @param[out] cofact_a The initial cofactor's real vector for poly_ing, and should be allocated in advance.
 *  @param[out] cofact_b The initial cofactor's real vector for poly_inf, and should be allocated in advance.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_initial_vector(SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int d) {
  SNAPMTXREAL_t matN;
  SNAPBASEREAL_t *pt1, *pt2;
  unsigned int i, j;
  int result_code = 0;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  double *superb;
#endif // SNAPLIB_GPGCD_USE_DGESVD
  SNAPMTXREAL_t vecS;
  SNAPMTXREAL_t matVT;
#endif // defined
  SNAPMTXREAL_init(matN);
  // construct the subresultant matrix with descending order.
  SNAP_agcd_gpgcd_subresultant_mapping_matrix_dsc(matN, tilde_f, tilde_g, d-1);
#ifdef SNAPLIB_DEBUG
printf("----matN---->\n");
show_sdmd(matN);
printf("<----matN----\n");
#endif // SNAPLIB_DEBUG
  // compute SVD but in part by LAPACK
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  SNAPMTXREAL_init2(vecS, 1, SNAPMTXREAL_cols(matN));
  SNAPMTXREAL_init2(matVT, SNAPMTXREAL_cols(matN), SNAPMTXREAL_cols(matN));
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  superb = (double*) malloc(sizeof(double)*SNAPMTXREAL_cols(matN));
  if ( LAPACKE_dgesvd( LAPACK_ROW_MAJOR, 'O', 'A', SNAPMTXREAL_rows(matN), SNAPMTXREAL_cols(matN), SNAPMTXREAL_elems(matN), SNAPMTXREAL_csize(matN),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT), superb ) != 0 ) {
     printf("ERROR in agcd_gpgcd_initial_cofactors: dgesvd could not compute the SVD.\n");
     result_code = -1;
     goto dest;
  }
#else
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTXREAL_rows(matN), SNAPMTXREAL_cols(matN), SNAPMTXREAL_elems(matN), SNAPMTXREAL_csize(matN),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTXREAL_elems(matVT), SNAPMTXREAL_csize(matVT) ) != 0 ) {
     printf("ERROR in agcd_gpgcd_initial_cofactors: dgesdd could not compute the SVD.\n");
     result_code = -1;
     goto dest;
  }
#endif // SNAPLIB_GPGCD_USE_DGESVD
#ifdef SNAPLIB_DEBUG
printf("----matVT---->\n");
show_sdmd(matVT);
printf("<----matVT----\n");
#endif // SNAPLIB_DEBUG

#ifdef SNAP_SUPD_T
  for ( i=0, pt1=SNAPMTXREAL_elemp(matVT,SNAPMTXREAL_cols(matN)-1, 0), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(matN); i++,pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
#else
  for ( i=0, pt1=SNAPMTXREAL_elemp(matVT,SNAPMTXREAL_cols(matN)-1, 0), pt2=SNAPMTXREAL_elems(cofact_a); i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b); i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_a)+1; i<SNAPMTXREAL_cols(cofact_a)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
  for ( i=0, pt2=SNAPMTXREAL_elems(cofact_b)+1; i<SNAPMTXREAL_cols(cofact_b)/2; i++,pt1++,pt2+=2 ) {
    SNAPBASEREAL_set(*pt2,*pt1);
  }
#endif
#else
  IMPLEMENTE HERE
#endif
  for ( i=0, pt1=SNAPMTXREAL_elems(vecS); i<SNAPMTXREAL_cols(matN)-1; i++,pt1++ ) {
    for ( j=0, pt2=SNAPMTXREAL_elemp(matVT,i,0); j<SNAPMTXREAL_cols(matN); j++,pt2++ ) {
      SNAPBASEREAL_mul(*pt2, *pt2, *pt1);
    }
  }
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_f)/2, SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,0,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,0), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_f), 2);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_g)/2, SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,0,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,SNAPMTXREAL_cols(tilde_g)/2-d), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_g), 2);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_f)/2, SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,SNAPMTXREAL_cols(tilde_f)/2+SNAPMTXREAL_cols(tilde_g)/2-d-1,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,0), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_f)+1, 2);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_g)/2, SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,SNAPMTXREAL_cols(tilde_f)/2+SNAPMTXREAL_cols(tilde_g)/2-d-1,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,SNAPMTXREAL_cols(tilde_g)/2-d), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_g)+1, 2);
#else
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_f), SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,0,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,0), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_f), 1);
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_cols(tilde_g), SNAPMTXREAL_cols(matN)-1, 1.0, SNAPMTXREAL_elemp(matN,0,0), SNAPMTXREAL_csize(matN),
              SNAPMTXREAL_elemp(matVT,0,SNAPMTXREAL_cols(tilde_g)-d), SNAPMTXREAL_csize(matVT), 0.0, SNAPMTXREAL_elems(tilde_g), 1);
#endif // defined
dest:
  SNAPMTXREAL_clear(matN);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  free(superb);
#endif // SNAPLIB_GPGCD_USE_DGESVD
  SNAPMTXREAL_clear(vecS);
  SNAPMTXREAL_clear(matVT);
#endif // defined
  return result_code;
}

/*!
 *  @brief modified newton method in GPGCD algorithm for two univariate polynomials, with some modifications.
 *  @param[in] input_f The input polynomial f[x] as real vector.
 *  @param[in] input_g The input polynomial g[x] as real vector.
 *  @param[in,out] tilde_f The initial perturbed real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in,out] tilde_g The initial perturbed real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_a The initial cofactor's real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_b The initial cofactor's real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in] max_iterations The iteration stops if the number of iterations reaches this value.
 *  @param[in] d_threshold The iteration stops if and only if the 2-norm of constraint vector Q is less than or equal to d_threshold.
 *  @param[in] tol The iteration stops if and only if the 2-norms of perturbations of f and g are less than or equal to tol.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_modified_newton2(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int max_iterations, const SNAPBASEREAL_t d_threshold, const SNAPBASEREAL_t tol) {
  int result_code = 0;
  unsigned int m, n, d, num_iterations;
  SNAPMTXREAL_t matJ, matJP, matU, vecN, vecQ, vecD;
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  unsigned int i;
  SNAPBASEREAL_t *pt1, *pt2;
  SNAPMTXREAL_t matJC, matVT, vecSV;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPBASEREAL_t tmp, alpha, nrm_f, nrm_g;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  double *superb;
#else
  lapack_int *ipiv;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  m = SNAPMTXREAL_cols(input_f)/2-1;
  n = SNAPMTXREAL_cols(input_g)/2-1;
  d = SNAPMTXREAL_cols(input_f)/2-SNAPMTXREAL_cols(cofact_b)/2;
#else
  m = SNAPMTXREAL_cols(input_f)-1;
  n = SNAPMTXREAL_cols(input_g)-1;
  d = SNAPMTXREAL_cols(input_f)-SNAPMTXREAL_cols(cofact_b);
#endif // defined
  SNAPBASEREAL_init(tmp);
  SNAPBASEREAL_init(alpha);
  SNAPBASEREAL_init(nrm_f);
  SNAPBASEREAL_init(nrm_g);
  SNAPMTXREAL_init(matJ);
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  SNAPMTXREAL_init2(matJP, 2*(m+n+2)+2*(m+n-2*d+2), 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(matU, 2*(m+n-d+1)+1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecQ, 1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n+2)+2*(m+n-2*d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(matVT, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecSV, 1, 2*(m+n-d+1)+1);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#else
  SNAPMTXREAL_init2(matJP, 2*(m+n-d+2), m+n-d+2);
  SNAPMTXREAL_init2(matU, m+n-d+2, m+n-d+2);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecQ, 1, m+n-d+2);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n-d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(matVT, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecSV, 1, m+n-d+2);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  superb = (double*) malloc(sizeof(double)*(m+n-d+2));
#else
  superb = NULL;
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  ipiv = (lapack_int*) malloc(sizeof(lapack_int)*SNAPMTXREAL_rows(matU));
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

  // do iterations
  num_iterations = 0;
  while (1) {
    // Construct the Jacobian matrix with the current vector
    SNAP_agcd_gpgcd_Jacobian(matJ, tilde_f, tilde_g, cofact_a, cofact_b, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----matJ---->\n");
show_sdmd(matJ);
#endif // SNAPLIB_DEBUG
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
    for ( i=0, pt1=SNAPMTXREAL_elems(matJC), pt2=SNAPMTXREAL_elems(matJ); i<SNAPMTXREAL_csize(matJ)*SNAPMTXREAL_rsize(matJ); i++,pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_svd(matJP, matJC, vecSV, matU, matVT, superb) != 0 ) {
      result_code = -1;
      goto dest;
    }
#else
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_tri(matJP, matJ, matU, ipiv) != 0 ) {
      result_code = -1;
      goto dest;
    }
#endif
    // form the nabla
    SNAP_agcd_gpgcd_nabla(vecN, input_f, input_g, tilde_f, tilde_g, m, n, d);
    // form the constraint
    SNAP_agcd_gpgcd_const(vecQ, matJ, cofact_a, cofact_b, m, n, d);
    // compute the norms of perturbations
    SNAPMTXREAL_BLAS_nrm2(nrm_f, SNAPMTXREAL_cols(input_f), SNAPMTXREAL_elemp(vecN,0,0), 1);
    SNAPMTXREAL_BLAS_nrm2(nrm_g, SNAPMTXREAL_cols(input_g), SNAPMTXREAL_elemp(vecN,0,SNAPMTXREAL_cols(input_f)), 1);
    SNAPMTXREAL_BLAS_nrm2(tmp, SNAPMTXREAL_cols(vecQ), SNAPMTXREAL_elemp(vecQ,0,0), 1);
    if ( SNAPBASEREAL_cmp(nrm_f, tol) <= 0 && SNAPBASEREAL_cmp(nrm_g, tol) <= 0 && SNAPBASEREAL_cmp(tmp, d_threshold) <= 0 ) {
      printf("INFO in agcd_gpgcd_modified_newton: converged (#%d) w.r.t. tolerance: (%e,%e) <= %e, %e <= %e.\n", num_iterations, nrm_f, nrm_g, tol, tmp, d_threshold);
      break;
    }
#ifdef SNAPLIB_DEBUG
printf("----vecN---->\n");
show_sdmd(vecN);
printf("----vecQ---->\n");
show_sdmd(vecQ);
#endif // SNAPLIB_DEBUG
    // search direction (NOTE: The result is in vecN !!)
    SNAP_agcd_gpgcd_direction(vecD, matJ, matJP, vecN, vecQ, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----vecD---->\n");
show_sdmd(vecN);
#endif // SNAPLIB_DEBUG
/*
    // check the size of the search direction
    SNAPMTXREAL_BLAS_nrm2(tmp, SNAPMTXREAL_cols(vecN), SNAPMTXREAL_elems(vecN), 1);
#ifdef SNAPLIB_DEBUG
printf("the size of search direction: %e\n", tmp);
#endif // SNAPLIB_DEBUG
    if ( SNAPBASEREAL_cmp(tmp, d_threshold) < 0 ) {
      printf("INFO in agcd_gpgcd_modified_newton: converged within %d iterations.\n", num_iterations);
      break;
    }
*/
    // step size (at first, we follow the step size in the paper)
    SNAPBASEREAL_set_ui(alpha, 1);
    // update
    SNAP_agcd_gpgcd_update(tilde_f, tilde_g, cofact_a, cofact_b, alpha, vecN);
#ifdef SNAPLIB_DEBUG
printf("----tilde_f---->\n");
show_sdmd(tilde_f);
printf("----tilde_g---->\n");
show_sdmd(tilde_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
    // check the number of iterations
    if ( ++num_iterations >= max_iterations ) {
      printf("INFO in agcd_gpgcd_modified_newton: didn't converge within %d iterations.\n", max_iterations);
      break;
    }
  }
dest:
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  free(superb);
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  free(ipiv);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
  SNAPBASEREAL_clear(tmp);
  SNAPBASEREAL_clear(alpha);
  SNAPBASEREAL_clear(nrm_f);
  SNAPBASEREAL_clear(nrm_g);
  SNAPMTXREAL_clear(matJ);
  SNAPMTXREAL_clear(matJP);
  SNAPMTXREAL_clear(matU);
  SNAPMTXREAL_clear(vecN);
  SNAPMTXREAL_clear(vecQ);
  SNAPMTXREAL_clear(vecD);
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_clear(matJC);
  SNAPMTXREAL_clear(matVT);
  SNAPMTXREAL_clear(vecSV);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  return result_code;
}

/*!
 *  @brief GPGCD algorithm for two univariate polynomials with some modifications.
 *  @details This is an implementation of the GPGCD algorithm proposed by Akira Terui
 *   in ISSAC 2009, ASCM-MACIS 2009 and so on. (see http://arxiv.org/abs/1207.0630v2).
 *   Moreover, tilde_f and tilde_g are computed from matU of SVD, and
 *   the modified newton method iterates until the tolerance is satisfied.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] tol The tolerance.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd2(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t tol) {
  SNAPMTXREAL_t input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b;
  unsigned int len_f, len_g, len_a, len_b;
  int result_code = 0;

  // the length of real vectors
  // our implementation preserves unknown vector x in tilde_f, tilde_g, cofact_a and cofact_b separetedly.
  len_f = SNAP_degree(poly_inf)+1;
  len_g = SNAP_degree(poly_ing)+1;
  len_a = SNAP_degree(poly_ing)-d+1;
  len_b = SNAP_degree(poly_inf)-d+1;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  len_f *= 2;
  len_g *= 2;
  len_a *= 2;
  len_b *= 2;
#endif // defined
  SNAPMTXREAL_init2(input_f, 1, len_f);
  SNAPMTXREAL_init2(input_g, 1, len_g);
  SNAPMTXREAL_init2(tilde_f, 1, len_f);
  SNAPMTXREAL_init2(tilde_g, 1, len_g);
  SNAPMTXREAL_init2(cofact_a, 1, len_a);
  SNAPMTXREAL_init2(cofact_b, 1, len_b);
  // copy input polynomials
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, input_f, input_g);
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, tilde_f, tilde_g);
  // initial cofactors
  if ( SNAP_agcd_gpgcd_initial_vector(tilde_f, tilde_g, cofact_a, cofact_b, d) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not compute the initial vector.\n");
    result_code = -1;
    goto dest;
  }
#ifdef SNAPLIB_DEBUG
printf("----input_f---->\n");
show_sdmd(input_f);
printf("----input_g---->\n");
show_sdmd(input_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
  // iterations
  if ( SNAP_agcd_gpgcd_modified_newton2(input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b, 8912, 1.0e-14, tol) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not proceed the modified newton iterations.\n");
    result_code = -1;
    goto dest;
  }
  // form the approximate GCD
  SNAP_agcd_gpgcd_lsdiv(poly_d, poly_f1, poly_g1, tilde_f, tilde_g, cofact_a, cofact_b, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d);
#ifdef SNAPLIB_DEBUG
show_dlist(SNAP_coefs(poly_d), d+1);
printf("=== END ===\n");
#endif // SNAPLIB_DEBUG
dest:
  SNAPMTXREAL_clear(input_f);
  SNAPMTXREAL_clear(input_g);
  SNAPMTXREAL_clear(tilde_f);
  SNAPMTXREAL_clear(tilde_g);
  SNAPMTXREAL_clear(cofact_a);
  SNAPMTXREAL_clear(cofact_b);

  return result_code;
}

/*!
 *  @brief modified newton method in GPGCD algorithm for two univariate polynomials with two bounds.
 *  @param[in] input_f The input polynomial f[x] as real vector.
 *  @param[in] input_g The input polynomial g[x] as real vector.
 *  @param[in,out] tilde_f The initial perturbed real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in,out] tilde_g The initial perturbed real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_a The initial cofactor's real vector for poly_ing on input, and the resulting vector on exit.
 *  @param[in,out] cofact_b The initial cofactor's real vector for poly_inf on input, and the resulting vector on exit.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_modified_newton_ex(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, SNAPBASEREAL_t threshold, unsigned int max_iterations) {
  int result_code = 0;
  unsigned int m, n, d, num_iterations;
  SNAPMTXREAL_t matJ, matJP, matU, vecN, vecQ, vecD;
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  unsigned int i;
  SNAPBASEREAL_t *pt1, *pt2;
  SNAPMTXREAL_t matJC, matVT, vecSV;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPBASEREAL_t tmp, alpha;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  double *superb;
#else
  lapack_int *ipiv;
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  m = SNAPMTXREAL_cols(input_f)/2-1;
  n = SNAPMTXREAL_cols(input_g)/2-1;
  d = SNAPMTXREAL_cols(input_f)/2-SNAPMTXREAL_cols(cofact_b)/2;
#else
  m = SNAPMTXREAL_cols(input_f)-1;
  n = SNAPMTXREAL_cols(input_g)-1;
  d = SNAPMTXREAL_cols(input_f)-SNAPMTXREAL_cols(cofact_b);
#endif // defined
  SNAPBASEREAL_init(tmp);
  SNAPBASEREAL_init(alpha);
  SNAPMTXREAL_init(matJ);
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  SNAPMTXREAL_init2(matJP, 2*(m+n+2)+2*(m+n-2*d+2), 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(matU, 2*(m+n-d+1)+1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecQ, 1, 2*(m+n-d+1)+1);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n+2)+2*(m+n-2*d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(matVT, 2*(m+n-d+1)+1, 2*(m+n+2)+2*(m+n-2*d+2));
  SNAPMTXREAL_init2(vecSV, 1, 2*(m+n-d+1)+1);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#else
  SNAPMTXREAL_init2(matJP, 2*(m+n-d+2), m+n-d+2);
  SNAPMTXREAL_init2(matU, m+n-d+2, m+n-d+2);
  SNAPMTXREAL_init2(vecN, 1, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecQ, 1, m+n-d+2);
  SNAPMTXREAL_init2(vecD, 1, 2*(m+n-d+2));
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_init2(matJC, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(matVT, m+n-d+2, 2*(m+n-d+2));
  SNAPMTXREAL_init2(vecSV, 1, m+n-d+2);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  superb = (double*) malloc(sizeof(double)*(m+n-d+2));
#else
  superb = NULL;
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  ipiv = (lapack_int*) malloc(sizeof(lapack_int)*SNAPMTXREAL_rows(matU));
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined

  // do iterations
  num_iterations = 0;
  while (1) {
    // Construct the Jacobian matrix with the current vector
    SNAP_agcd_gpgcd_Jacobian(matJ, tilde_f, tilde_g, cofact_a, cofact_b, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----matJ---->\n");
show_sdmd(matJ);
#endif // SNAPLIB_DEBUG
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
    for ( i=0, pt1=SNAPMTXREAL_elems(matJC), pt2=SNAPMTXREAL_elems(matJ); i<SNAPMTXREAL_csize(matJ)*SNAPMTXREAL_rsize(matJ); i++,pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_svd(matJP, matJC, vecSV, matU, matVT, superb) != 0 ) {
      result_code = -1;
      goto dest;
    }
#else
    if ( SNAP_agcd_gpgcd_pseudo_inverse_by_tri(matJP, matJ, matU, ipiv) != 0 ) {
      result_code = -1;
      goto dest;
    }
#endif
    // form the nabla and constraint
    SNAP_agcd_gpgcd_nabla(vecN, input_f, input_g, tilde_f, tilde_g, m, n, d);
    SNAP_agcd_gpgcd_const(vecQ, matJ, cofact_a, cofact_b, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----vecN---->\n");
show_sdmd(vecN);
printf("----vecQ---->\n");
show_sdmd(vecQ);
#endif // SNAPLIB_DEBUG
    // search direction (NOTE: The result is in vecN !!)
    SNAP_agcd_gpgcd_direction(vecD, matJ, matJP, vecN, vecQ, m, n, d);
#ifdef SNAPLIB_DEBUG
printf("----vecD---->\n");
show_sdmd(vecN);
#endif // SNAPLIB_DEBUG
    // check the size of the search direction
    SNAPMTXREAL_BLAS_nrm2(tmp, SNAPMTXREAL_cols(vecN), SNAPMTXREAL_elems(vecN), 1);
#ifdef SNAPLIB_DEBUG
printf("the size of search direction: %e\n", tmp);
#endif // SNAPLIB_DEBUG
    if ( SNAPBASEREAL_cmp(tmp, threshold) < 0 ) {
      printf("INFO in agcd_gpgcd_modified_newton: converged within %d iterations.\n", num_iterations);
      break;
    }
    // step size (at first, we follow the step size in the paper)
    SNAPBASEREAL_set_ui(alpha, 1);
    // update
    SNAP_agcd_gpgcd_update(tilde_f, tilde_g, cofact_a, cofact_b, alpha, vecN);
#ifdef SNAPLIB_DEBUG
printf("----tilde_f---->\n");
show_sdmd(tilde_f);
printf("----tilde_g---->\n");
show_sdmd(tilde_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
    // check the number of iterations
    if ( ++num_iterations >= max_iterations ) {
      printf("INFO in agcd_gpgcd_modified_newton: didn't converge within %d iterations.\n", max_iterations);
      break;
    }
  }
dest:
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#ifdef SNAPLIB_GPGCD_USE_DGESVD
  free(superb);
#endif // SNAPLIB_GPGCD_USE_DGESVD
#else
  free(ipiv);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
#endif // defined
  SNAPBASEREAL_clear(tmp);
  SNAPBASEREAL_clear(alpha);
  SNAPMTXREAL_clear(matJ);
  SNAPMTXREAL_clear(matJP);
  SNAPMTXREAL_clear(matU);
  SNAPMTXREAL_clear(vecN);
  SNAPMTXREAL_clear(vecQ);
  SNAPMTXREAL_clear(vecD);
#ifndef SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  SNAPMTXREAL_clear(matJC);
  SNAPMTXREAL_clear(matVT);
  SNAPMTXREAL_clear(vecSV);
#endif // SNAPLIB_GPGCD_PSEUDO_INVERSE_WO_SVD
  return result_code;
}

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
#endif // defined

/*!
 *  @brief GPGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the GPGCD algorithm proposed by Akira Terui
 *   in ISSAC 2009, ASCM-MACIS 2009 and so on. (see http://arxiv.org/abs/1207.0630v2)
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_gpgcd_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  SNAPMTXREAL_t input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b;
  SNAPBASEREAL_t rho;
  unsigned int len_f, len_g, len_a, len_b;
  int result_code = 0;

  // the length of real vectors
  // our implementation preserves unknown vector x in tilde_f, tilde_g, cofact_a and cofact_b separetedly.
  len_f = SNAP_degree(poly_inf)+1;
  len_g = SNAP_degree(poly_ing)+1;
  len_a = SNAP_degree(poly_ing)-d+1;
  len_b = SNAP_degree(poly_inf)-d+1;
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  len_f *= 2;
  len_g *= 2;
  len_a *= 2;
  len_b *= 2;
#endif // defined
  SNAPMTXREAL_init2(input_f, 1, len_f);
  SNAPMTXREAL_init2(input_g, 1, len_g);
  SNAPMTXREAL_init2(tilde_f, 1, len_f);
  SNAPMTXREAL_init2(tilde_g, 1, len_g);
  SNAPMTXREAL_init2(cofact_a, 1, len_a);
  SNAPMTXREAL_init2(cofact_b, 1, len_b);
  // copy input polynomials
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, input_f, input_g);
  SNAP_agcd_gpgcd_construct_tildevectors(poly_inf, poly_ing, tilde_f, tilde_g);
  // initial cofactors
  if ( SNAP_agcd_gpgcd_initial_cofactors(input_f, input_g, cofact_a, cofact_b, d) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not compute initial cofactors.\n");
    result_code = -1;
    goto dest;
  }
#ifdef SNAPLIB_DEBUG
printf("----input_f---->\n");
show_sdmd(input_f);
printf("----input_g---->\n");
show_sdmd(input_g);
printf("----cofact_a---->\n");
show_sdmd(cofact_a);
printf("----cofact_b---->\n");
show_sdmd(cofact_b);
#endif // SNAPLIB_DEBUG
  // iterations
  if ( SNAP_agcd_gpgcd_modified_newton_ex(input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b, threshold, max_iterations) != 0 ) {
    printf("ERROR in agcd_gpgcd: could not proceed the modified newton iterations.\n");
    result_code = -1;
    goto dest;
  }
  // form the approximate GCD
  SNAP_agcd_gpgcd_lsdiv(poly_d, poly_f1, poly_g1, tilde_f, tilde_g, cofact_a, cofact_b, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d);
  // getting a better approximate GCD.
  if ( mode_opt != 0 ) {
    SNAPBASEREAL_init(rho);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho);
#endif
    SNAPBASEREAL_clear(rho);
  }
#ifdef SNAPLIB_DEBUG
show_dlist(SNAP_coefs(poly_d), d+1);
printf("=== END ===\n");
#endif // SNAPLIB_DEBUG
dest:
  SNAPMTXREAL_clear(input_f);
  SNAPMTXREAL_clear(input_g);
  SNAPMTXREAL_clear(tilde_f);
  SNAPMTXREAL_clear(tilde_g);
  SNAPMTXREAL_clear(cofact_a);
  SNAPMTXREAL_clear(cofact_b);

  return result_code;
}





























/* ================== END[agcd_gpgcd] ================== */






