/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <snap.h>

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG


#ifdef SNAPLIB_DEBUG
void show_mtx(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  for(i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      mpz_out_str(stdout, 10, *(mtx+i*cols+j));
      printf(" ");
    }
    printf("\n");
  }
}
#endif

/* ================== START[agcd_over_integers] ================== */
void agcd_jsc2009_get_cH(mpz_t cH, supz_t f, supz_t g, unsigned int rowsH) {
  unsigned int i;
  mpz_t tmp1,tmp2;
  mpz_init(tmp1);
  mpz_init(tmp2);
  for(i=0; i<=supz_degree(f); i++) {
    mpz_addmul(tmp1, supz_coef(f,i), supz_coef(f,i));
  }
  for(i=0; i<=supz_degree(g); i++) {
    mpz_addmul(tmp2, supz_coef(g,i), supz_coef(g,i));
  }
  mpz_mul_2exp(tmp1, tmp1, 2*supz_degree(f));
  mpz_mul_2exp(tmp2, tmp2, 2*supz_degree(g));
  if ( mpz_cmp(tmp1, tmp2) < 0 ) {
    mpz_add_ui(cH, tmp1, 1);
  } else {
    mpz_add_ui(cH, tmp2, 1);
  }
  mpz_sqrtrem(cH,tmp1,cH);
  if ( mpz_sgn(tmp1) > 0 ) {
    mpz_add_ui(cH,cH,1);
  }
  mpz_mul_2exp(cH, cH, ((unsigned int) rowsH/2));
  mpz_clear(tmp1);
  mpz_clear(tmp2);
}

void agcd_jsc2009_get_cB(mpz_t cB, supz_t f, supz_t g, unsigned int rows) {
  unsigned int i,d;
  mpz_t tmp1,tmp2;
  mpz_init(tmp1);
  mpz_init(tmp2);
  // Landau Mignotte Bounds
  for(i=0; i<=supz_degree(f); i++) {
    mpz_addmul(tmp1, supz_coef(f,i), supz_coef(f,i));
  }
  for(i=0; i<=supz_degree(g); i++) {
    mpz_addmul(tmp2, supz_coef(g,i), supz_coef(g,i));
  }
  mpz_sqrtrem(tmp1,cB,tmp1);
  if ( mpz_sgn(cB) > 0 ) {
    mpz_add_ui(tmp1,tmp1,1);
  }
  mpz_sqrtrem(tmp2,cB,tmp2);
  if ( mpz_sgn(cB) > 0 ) {
    mpz_add_ui(tmp2,tmp2,1);
  }
  d = supz_degree(f)/2;
  mpz_bin_uiui(cB, d-1, ((unsigned int) d/2));
  mpz_mul(tmp1, tmp1, cB);
  mpz_bin_uiui(cB, d-1, ((unsigned int) d/2)-1);
  mpz_add(tmp1, tmp1, cB);
  d = supz_degree(g)/2;
  mpz_bin_uiui(cB, d-1, ((unsigned int) d/2));
  mpz_mul(tmp2, cB, tmp2);
  mpz_bin_uiui(cB, d-1, ((unsigned int) d/2)-1);
  mpz_add(tmp2, tmp2, cB);
  if ( mpz_cmp( tmp1, tmp2) < 0 ) {
    mpz_set(tmp1, tmp2);
  }
#ifdef SNAPLIB_DEBUG
  printf("Landau Mignotte Bound: %s\n", mpz_get_str(NULL,10,tmp1));
#endif
  // Compute c for matL
  mpz_set_ui(tmp2, rows);
  mpz_sqrtrem(cB,tmp2,tmp2);
  if ( mpz_sgn(tmp2) > 0 ) {
    mpz_add_ui(cB,cB,1);
  }
  mpz_mul(cB, cB, tmp1);
  mpz_mul_2exp(cB, cB, ((unsigned int) rows/2));
  mpz_clear(tmp1);
  mpz_clear(tmp2);
}

void agcd_jsc2009_quicksort_uint_by_mpz_t(unsigned int *cc, mpz_t *norm, unsigned int len) {
  unsigned int i, j, p;
  mpz_t tmp;
  if ( len == 2 ) {
    if ( mpz_cmp(*norm, *(norm+1)) > 0 ) {
      mpz_swap(*norm,*(norm+1));
      p = *cc;
      *cc = *(cc+1);
      *(cc+1) = p;
    }
  } else if ( len > 2 ) {
    i=0;
    j=len-1;
    p = i+(j-i)/2;
    mpz_init(tmp);
    if ( mpz_cmp(*(norm+i),*(norm+p)) < 0 ) {
      if ( mpz_cmp(*(norm+p),*(norm+j)) < 0 ) {
        mpz_set(tmp,*(norm+p));
      } else if ( mpz_cmp(*(norm+j),*(norm+i)) < 0 ) {
        mpz_set(tmp,*(norm+i));
      } else {
        mpz_set(tmp,*(norm+j));
      }
    } else {
      if ( mpz_cmp(*(norm+p),*(norm+j)) > 0 ) {
        mpz_set(tmp,*(norm+p));
      } else if ( mpz_cmp(*(norm+j),*(norm+i)) > 0 ) {
        mpz_set(tmp,*(norm+i));
      } else {
        mpz_set(tmp,*(norm+j));
      }
    }
    while(1) {
      for(; i<len; i++) {
        if ( mpz_cmp(tmp, *(norm+i)) <= 0 ) {
          break;
        }
      }
      // for(j=len-1;j>=0;j--) // NOTE j can not be < 0.
      do {
        if ( mpz_cmp(tmp, *(norm+j)) >= 0 ) {
          break;
        }
      } while ( j-- != 0 );
      if ( i < j ) {
        mpz_swap(*(norm+i), *(norm+j));
        p = *(cc+i);
        *(cc+i) = *(cc+j);
        *(cc+j) = p;
        i++;
        j--;
      } else {
        break;
      }
    }
    //for (i=0;i<len;i++) {
    //	if ( mpz_cmp(tmp, *(norm+i)) == 0 ) { break; }
    //}
    //if ( i!=0 ) { agcd_jsc2009_quicksort_uint_by_mpz_t(cc, norm, i); }
    agcd_jsc2009_quicksort_uint_by_mpz_t(cc, norm, i);
    //if ( i!=len-1 ) { agcd_jsc2009_quicksort_uint_by_mpz_t(cc+i+1, norm+i+1,len-i-1); }
    agcd_jsc2009_quicksort_uint_by_mpz_t(cc+j+1, norm+j+1,len-j-1);
    mpz_clear(tmp);
  }
}

unsigned int agcd_jsc2009_get_candidate_gcds(unsigned int *cgcd, mpz_t *mtxH, unsigned int rows, unsigned int cols) {
  // pick candidate vectors up s.t. its first element = \pm1.
  unsigned int i, rs, len_rs=0;
  mpz_t *norm;
  for(rs=0; rs<rows; rs++) {
    if( mpz_cmpabs_ui(*(mtxH+rs*cols),1) == 0 ) {
      // rs-th row is now a candidate vector.
      *(cgcd+len_rs) = rs;
      len_rs++;
    }
  }
  // sort the candidate vectors.
  norm = (mpz_t*) malloc(len_rs*sizeof(mpz_t));
  for(i=0; i<len_rs; i++) {
    mpz_init(*(norm+i));
    mpz_inner(*(norm+i),mtxH+cols*(*(cgcd+i))+rows, mtxH+cols*(*(cgcd+i))+rows,cols-rows);
  }
  agcd_jsc2009_quicksort_uint_by_mpz_t(cgcd, norm, len_rs);
  for(i=0; i<len_rs; i++) {
    mpz_clear(*(norm+i));
  }
  free(norm);
  return len_rs;
}

unsigned int agcd_jsc2009_get_candidate_cofactors(unsigned int *cc, mpz_t *mtxL, unsigned int rows, unsigned int cols, supz_t f, supz_t g, unsigned int r, mpz_t eps, unsigned int svt_ui) {
  // pick candidate vectors up by the criteria.
  // cc[i] => j if j-th row is one of candidate cofactors.
  // The return value is the number of candidates.
  unsigned int i, rs, len_rs=0;
  unsigned int *absf, *absg;
  unsigned int deg_s, deg_t;
  unsigned short int notfound;
  mpz_t abs1, abs2, den, rhs, lhs;
  mpz_t *norm;

  mpz_init(abs1);
  mpz_init(abs2);
  mpz_init(den);
  mpz_init(rhs);
  mpz_init(lhs);
  absf = (unsigned int*) malloc(sizeof(unsigned int)*(supz_degree(f)+1));
  absg = (unsigned int*) malloc(sizeof(unsigned int)*(supz_degree(g)+1));
  for(i=0; i<=supz_degree(f); i++) {
    *(absf+i) = ( mpz_cmpabs(supz_coef(f,i), eps)>0 ? SNAPLIB_TRUE : SNAPLIB_FALSE);
  }
  for(i=0; i<=supz_degree(g); i++) {
    *(absg+i) = ( mpz_cmpabs(supz_coef(g,i), eps)>0 ? SNAPLIB_TRUE : SNAPLIB_FALSE);
  }
  for(rs=0; rs<rows; rs++) {
#ifdef SNAPLIB_DEBUG
    printf("rs=%d\n", rs);
#endif
    // default: not a candidate vector.
    notfound = SNAPLIB_TRUE;
    // for( deg_s=supz_degree(g)-r-1; deg_s>=0; deg_s-- )
    deg_s = supz_degree(g)-r-1;
    do {
      if ( mpz_sgn(*(mtxL+rs*cols+deg_s))!=0 ) {
        notfound = SNAPLIB_FALSE;
        break;
      }
    } while ( deg_s-- != 0 );
    // zero cofactor is not related to any app-gcd.
    if ( notfound == SNAPLIB_TRUE ) continue;
    notfound = SNAPLIB_TRUE;
    // for( deg_t=supz_degree(f)-r-1; deg_t>=0; deg_t-- )
    deg_t = supz_degree(f)-r-1;
    do {
      if ( mpz_sgn(*(mtxL+rs*cols+supz_degree(g)-r+deg_t))!=0 ) {
        notfound = SNAPLIB_FALSE;
        break;
      }
    } while ( deg_t-- != 0 );
    // zero cofactor is not related to any app-gcd.
    if ( notfound == SNAPLIB_TRUE ) continue;
#ifdef SNAPLIB_DEBUG
    printf("Entering Criterion 1 Check.\n");
#endif
    // Criterion 1
    if( *absf == SNAPLIB_TRUE && mpz_sgn(*(mtxL+rs*cols))==0 ) continue;
    if( *absg == SNAPLIB_TRUE && mpz_sgn(*(mtxL+rs*cols+supz_degree(g)-r))==0 ) continue;
#ifdef SNAPLIB_DEBUG
    printf("Entering Criterion 2 Check.\n");
#endif
    // Criterion 2
    if( *(absf+supz_degree(f)) == SNAPLIB_TRUE && supz_degree(f)-deg_t>supz_degree(g)-deg_s ) continue;
    if( *(absg+supz_degree(g)) == SNAPLIB_TRUE && supz_degree(f)-deg_t<supz_degree(g)-deg_s ) continue;
#ifdef SNAPLIB_DEBUG
    printf("Entering Criterion 3 Check.\n");
#endif
    // Criterion 3 (the second ineqn of this criterion included in the above)
    if ( supz_degree(f)-deg_t==supz_degree(g)-deg_s ) {
      mpz_abs(abs1, *(mtxL+rs*cols+deg_s)); // s
      mpz_abs(abs2, *(mtxL+rs*cols+supz_degree(g)-r+deg_t)); // t
      mpz_add(den, abs1, abs2);
      if ( mpz_sgn(den)!=0 ) {
        mpz_mul(rhs, den, eps);
        mpz_mul(abs1, *(mtxL+rs*cols+supz_degree(g)-r+deg_t), supz_lcoef(g));
        mpz_abs(abs1, abs1);
        mpz_mul(abs2, *(mtxL+rs*cols+deg_s), supz_lcoef(f));
        mpz_abs(abs2, abs2);
        mpz_sub(lhs, abs1, abs2);
        mpz_abs(lhs, lhs);
        if ( mpz_cmp( lhs, rhs) > 0 ) continue;
        mpz_abs(lhs, supz_lcoef(g));
        mpz_abs(abs1, supz_lcoef(f));
        mpz_add(abs1, abs1, lhs);
        mpz_mul(lhs, lhs, den);
        mpz_abs(abs2, *(mtxL+rs*cols+deg_s));
        mpz_mul(abs2, abs2, abs1);
        mpz_sub(lhs, lhs, abs2);
        mpz_abs(lhs, lhs);
        if ( mpz_cmp( lhs, rhs) > 0 ) continue;
      }
    }
#ifdef SNAPLIB_DEBUG
    printf("Now this is a candidate: %d\n", rs);
#endif
    // rs-th row is now a candidate vector.
    *(cc+len_rs) = rs;
    len_rs++;
  }
  free(absf);
  free(absg);
  mpz_clear(abs1);
  mpz_clear(abs2);
  mpz_clear(rhs);
  mpz_clear(lhs);
  mpz_clear(den);
  if ( len_rs == 0 ) {
    return 0;
  }
  // sort the candidate vectors.
  norm = (mpz_t*) malloc(len_rs*sizeof(mpz_t));
  for(i=0; i<len_rs; i++) {
    mpz_init(*(norm+i));
    mpz_inner(*(norm+i),mtxL+cols*(*(cc+i))+rows, mtxL+cols*(*(cc+i))+rows,cols-rows);
  }
#ifdef SNAPLIB_DEBUG
  printf("==== norm (0..%d) ====\n", len_rs-1);
  show_mtx(norm,1,len_rs);
#endif
  agcd_jsc2009_quicksort_uint_by_mpz_t(cc, norm, len_rs);
  mpz_mul_ui(*norm, *norm, svt_ui);
  rs = len_rs;
  for(i=1; i<len_rs; i++) {
    if( mpz_cmp(*norm, *(norm+i)) < 0 ) {
      rs = i;
      break;
    }
  }
  for(i=0; i<len_rs; i++) {
    mpz_clear(*(norm+i));
  }
  free(norm);
  return rs;
}

void agcd_jsc2009_make_mtxL(mpz_t *mtxL, unsigned int rows, unsigned int cols, supz_t f, supz_t g, unsigned int r, mpz_t c) {
  // mtxL must be zero matrix.
  // normal order
  unsigned int s, i;
  for(i=0; i<rows; i++) {
    mpz_set_ui(*(mtxL+i*cols+i),1);
  }
  for(s=0; s<supz_degree(g)-r; s++) {
    for(i=0; i<=supz_degree(f); i++) {
      mpz_mul(*(mtxL+s*cols+rows+s+i),supz_coef(f,i),c);
    }
  }
  for(s=0; s<supz_degree(f)-r; s++)      {
    for(i=0; i<=supz_degree(g); i++) {
      mpz_mul(*(mtxL+(s+supz_degree(g)-r)*cols+rows+s+i),supz_coef(g,i),c);
    }
  }
}

void agcd_jsc2009_make_mtxH(mpz_t *mtxH, unsigned int rows, unsigned int cols, supz_t f, supz_t g, supz_t t, supz_t s, unsigned int r, mpz_t c) {
  // mtxH must be zero matrix.
  unsigned int st, i;
  for(i=0; i<rows; i++) {
    mpz_set_ui(*(mtxH+i*cols+i),1);
  }
  for(i=0; i<=supz_degree(f); i++) {
    mpz_mul(*(mtxH+rows+i),supz_coef(f,i),c);
  }
  for(i=0; i<=supz_degree(g); i++) {
    mpz_mul(*(mtxH+rows+supz_degree(f)+1+i),supz_coef(g,i),c);
  }
  for(st=0; st<r+2; st++) { // This is OK. r+2<=size
    for(i=0; i<=supz_degree(t); i++) {
      mpz_mul(*(mtxH+st*cols+cols+rows+st+i),supz_coef(t,i),c);
    }
  }
  for(st=0; st<r+2; st++) { // This is OK. r+2<=size
    for(i=0; i<=supz_degree(s); i++) {
      mpz_mul(*(mtxH+st*cols+cols+rows+supz_degree(f)+1+st+i),supz_coef(s,i),c);
    }
  }
}

int agcd_jsc2009(supz_t f, supz_t g, supz_t agcd, supz_t cf, supz_t cg, mpz_t tol, unsigned int eps_ui, unsigned int svt_ui, unsigned int lll_mode) {
  unsigned int deg_f, deg_g, r;
  unsigned short int wegotagcd = SNAPLIB_FALSE;
  unsigned int rows, cols, rowsH, colsH, i, len_cc, idx_cc, len_cgcd, idx_cgcd;
  unsigned int *cc, *cgcd;
  mpz_t eps, norm_f, norm_g, max_eps, c, cH, max_c, max_cH, eps_tmp1, eps_tmp2;
  mpz_t *mtxL, *mtxH;
  supz_t s, t, gcd, rpoly;
  // initializations
  deg_f = supz_degree(f);
  deg_g = supz_degree(g);
  supz_init(s);
  supz_init(t);
  supz_init(gcd);
  supz_init(rpoly);
  mpz_init_set_ui(eps,eps_ui);
  mpz_init(norm_f);
  mpz_init(norm_g);
  supz_f2norm(norm_f, f);
  supz_f2norm(norm_g, g);
  mpz_init(max_eps);
  mpz_init(c);
  mpz_init(cH);
  mpz_init(max_c);
  mpz_init(max_cH);
  if ( mpz_cmp(norm_f, norm_g) < 0 ) {
    mpz_set(max_eps, norm_f);
  } else {
    mpz_set(max_eps, norm_g);
  }
  supz_c2norm(norm_f, f);
  supz_c2norm(norm_g, g);
  mpz_init(eps_tmp1);
  mpz_init(eps_tmp2);
  // eps's WHILE loop.
  while( mpz_cmp(eps, max_eps) < 0 || mpz_cmp_ui(eps, 1) == 0 ) {
    if ( deg_f < deg_g ) {
      r = deg_f-1;
    } else {
      r = deg_g-1;
    }
    // r's DO-WHILE loop.
    do {
      if ( mpz_cmp(norm_f, norm_g) > 0 ) {
        mpz_set(c, norm_f);
      } else {
        mpz_set(c, norm_g);
      }
      // size of mtxL.
      rows = deg_f+deg_g-2*r;
      cols = deg_f+deg_g-r + rows;
      // size of mtxH
      rowsH = r+3;
      colsH = supz_degree(f)+supz_degree(g)+2+rowsH;
      // compute the bound c_B and c_H
      agcd_jsc2009_get_cB(max_c, f, g, rows);
      agcd_jsc2009_get_cH(max_cH, f, g, rowsH);
      // mpz_set(max_c,c); mpz_set(max_cH, cH); <== mostly this is ok.
      // c's WHILE loop.
      while ( 1 ) {
#ifdef SNAPLIB_DEBUG
        printf("Current Tolerance Goal: %s\n", mpz_get_str(NULL, 10, eps));
        printf("Current Subresultant Degree: %d (degF=%d, degG=%d)\n", r, deg_f, deg_g);
        printf("Current Scaling Factor: %s (%s)\n", mpz_get_str(NULL, 10, c), mpz_get_str(NULL,10,max_c));
#endif
        cc = (unsigned int*) malloc(rows*sizeof(unsigned int));
        mtxL = (mpz_t*) malloc(rows*cols*sizeof(mpz_t));
        for(i=0; i<rows*cols; i++) {
          mpz_init(*(mtxL+i));
        }
        agcd_jsc2009_make_mtxL(mtxL, rows, cols, f, g, r, c);
        lll_exact_mode(mtxL, rows, cols, lll_mode);
        len_cc = agcd_jsc2009_get_candidate_cofactors(cc ,mtxL, rows, cols, f, g, r, eps, svt_ui);
#ifdef SNAPLIB_DEBUG
        printf("Candidate cofactors found: %d\n", len_cc);
#endif
        // candidate_cofactor's FOR loop.
        for(idx_cc=0; idx_cc<len_cc; idx_cc++) {
          supz_set_z(s, mtxL+cols*(*(cc+idx_cc)), supz_degree(g)-r);
          supz_set_z(t, mtxL+cols*(*(cc+idx_cc))+supz_degree(g)-r, supz_degree(f)-r);
          supz_self_neg(t);
#ifdef SNAPLIB_DEBUG
          printf("CoFactor [s]\n");
          show_mtx(supz_coefs(s), 1, supz_size(s));
          printf("CoFactor [t]\n");
          show_mtx(supz_coefs(t), 1, supz_size(t));
#endif
          if ( mpz_cmp(norm_f, norm_g) > 0 ) {
            mpz_set(cH, norm_f);
          } else {
            mpz_set(cH, norm_g);
          }
          // cH's WHILE loop.
          while ( 1 ) {
            cgcd = (unsigned int*) malloc(rowsH*sizeof(unsigned int));
            mtxH = (mpz_t*) malloc(rowsH*colsH*sizeof(mpz_t));
            for(i=0; i<rowsH*colsH; i++) {
              mpz_init(*(mtxH+i));
            }
            agcd_jsc2009_make_mtxH(mtxH, rowsH, colsH, f, g, t, s, r, cH);
            lll_exact_mode(mtxH, rowsH, colsH, lll_mode);
#ifdef SNAPLIB_DEBUG
            printf("Lattice Reduce for Cofactors.\n");
            show_mtx(mtxH,rowsH,colsH);
#endif
            len_cgcd = agcd_jsc2009_get_candidate_gcds(cgcd, mtxH, rowsH, colsH);
            // candidate_gcd's FOR loop.
            for(idx_cgcd=0; idx_cgcd<len_cgcd; idx_cgcd++) {
              supz_set_z(gcd, mtxH+colsH*(*(cgcd+idx_cgcd))+1, r+2);
              if ( supz_degree(gcd) == 0 && mpz_sgn(supz_coef(gcd,0)) == 0 ) {
                continue;
              }
              if( mpz_sgn(*(mtxH+colsH*(*(cgcd+idx_cgcd)))) > 0 ) {
                supz_self_neg(gcd);
              }
              supz_mul(rpoly, t, gcd);
              supz_sub(rpoly, rpoly, f);
              supz_c2norm(eps_tmp1, rpoly);
              supz_mul(rpoly, s, gcd);
              supz_sub(rpoly, rpoly, g);
              supz_c2norm(eps_tmp2, rpoly);
#ifdef SNAPLIB_DEBUG
              printf("eps for f: %s\n", mpz_get_str(NULL,10,eps_tmp1));
              printf("eps for g: %s\n", mpz_get_str(NULL,10,eps_tmp2));
#endif
              if ( mpz_cmp(eps_tmp1, eps_tmp2) > 0 ) {
                mpz_set(tol, eps_tmp1);
              } else {
                mpz_set(tol, eps_tmp2);
              }
              mpz_add(eps_tmp1, eps_tmp1, eps_tmp2);
              mpz_mul_ui(eps_tmp2,eps,2);
              if( mpz_cmp(eps_tmp1,eps_tmp2)<=0 ) {
                if( mpz_sgn(supz_lcoef(gcd))<0 ) {
                  supz_self_neg(gcd);
                  supz_self_neg(s);
                  supz_self_neg(t);
                }
                supz_set(agcd, gcd);
                supz_set(cf, t);
                supz_set(cg, s);
                wegotagcd = SNAPLIB_TRUE;
                break;
              }
            } // end of candidate_gcds's FOR loop.
            for(i=0; i<rowsH*colsH; i++) {
              mpz_clear(*(mtxH+i));
            }
            free(mtxH);
            free(cgcd);
            if ( wegotagcd == SNAPLIB_TRUE ) {
              break;
            }
            if ( mpz_cmp(cH, max_cH) >= 0 ) {
              break;
            }
            if ( mpz_cmp(norm_f, norm_g) > 0 ) {
              mpz_mul(cH, cH, norm_f);
            } else {
              mpz_mul(cH, cH, norm_g);
            }
            if ( mpz_cmp(cH, max_cH) > 0 ) {
              mpz_set(cH ,max_cH);
            }
          } // end of cH's WHILE loop.
          if ( wegotagcd == SNAPLIB_TRUE ) {
            break;
          }
        } // end of candidate_cofactors's FOR loop.
        for(i=0; i<rows*cols; i++) {
          mpz_clear(*(mtxL+i));
        }
        free(mtxL);
        free(cc);
        if ( wegotagcd == SNAPLIB_TRUE ) {
          break;
        }
        if ( mpz_cmp(c, max_c) >= 0 ) {
          break;
        }
        if ( mpz_cmp(norm_f, norm_g) > 0 ) {
          mpz_mul(c, c, norm_f);
        } else {
          mpz_mul(c, c, norm_g);
        }
        if ( mpz_cmp(c, max_c) > 0 ) {
          mpz_set(c, max_c);
        }
      } // end of c's WHILE loop.
      if ( wegotagcd == SNAPLIB_TRUE ) {
        break;
      }
    } while ( r-- != 0 ); // end of r's WHILE loop.
    if ( wegotagcd == SNAPLIB_TRUE ) {
      break;
    }
    mpz_mul_ui(eps,eps,10);
  } // end of eps's WHILE loop.
  // DO NOT FORGET TO CLEAR AND FREE!!
  mpz_clear(eps);
  mpz_clear(norm_f);
  mpz_clear(norm_g);
  mpz_clear(max_eps);
  mpz_clear(c);
  mpz_clear(eps_tmp1);
  mpz_clear(eps_tmp2);
  supz_clear(s);
  supz_clear(t);
  supz_clear(gcd);
  supz_clear(rpoly);
  if ( wegotagcd == SNAPLIB_TRUE ) {
    return 0;
  }
  return -1;
}
/* ================== END[agcd_over_integers] ================== */
