/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <snap.h>
#if defined(SNAP_SUPD_T)||defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

/* ================== START [agcd_qrgcd] ================== */
/* SEE ALSO: QR Factoring to Compute the GCD of Univariate Approximate Polynomials.
   Robert M. Corless, Stephen M. Watt, and Lihong Zhi, 2004 */
typedef enum {NORMAL, REVERSAL, SPLITALLOWED, SPLITDISALLOWED, BEZOUTONLY, NORMALONLY} SNAP_agcd_qrgcd_qrstep_order;
// NORMAL: current order and first order in the main routine
// REVERSAL: current order and first order in the main routine
// SPLITALLOWED: split will be called in the case 3
// SPLITDISALLOWED: split won't be called even in the case 3
// BEZOUTONLY: only compute the bezout coefficients
// NORMALONLY: this effects only the debug message (e.g. non verbose mode)
//
// #define SNAPLIB_MPFR_RND MPFR_RNDN
// #define SNAPLIB_MPC_RND MPC_RNDNN
// ========== THE FOLLOWING MACROS SHOULD BE MODIFIED TO BE COMPATIBLE WITH MPFR AND MPC. ==========
#define SNAPLIB_MIN(x,y) ((x)<(y) ? (x):(y))
#define SNAPLIB_MAX(x,y) ((x)>(y) ? (x):(y))
static unsigned int SNAP_agcd_qrgcd_qrstep(SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, SNAP_t, const SNAPBASEREAL_t, SNAPBASEREAL_t, SNAP_agcd_qrgcd_qrstep_order, SNAP_agcd_qrgcd_qrstep_order);
extern int SNAP_agcd_exqrgcd_split(SNAP_t poly_inside, SNAP_t poly, const SNAPBASEREAL_t tol);
extern void SNAP_agcd_exqrgcd_Qkthcolumn(SNAPBASE_t *vect, SNAPMTX_t matrix_gqr, SNAPBASE_t *tau, const unsigned int k);

// for supfr_t, we use (unsigned int) 1.
//#define SNAPLIB_SPLIT_GAP 1.0
// Return the degree of newly found agcd
static unsigned int SNAP_agcd_qrgcd_qrstep(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_cof, SNAP_t poly_cog, SNAP_t poly_d, SNAP_t poly_u, SNAP_t poly_v, const SNAPBASEREAL_t tol, SNAPBASEREAL_t tol_res, SNAP_agcd_qrgcd_qrstep_order qrstep_order, SNAP_agcd_qrgcd_qrstep_order split_flag) {
  // We'd like to try to split then force to use PRS.
  SNAPMTX_t matrix_syl;
  SNAP_t poly_r, poly_t, poly_ncf, poly_ncg, poly_prs1, poly_prs2;
  SNAPBASE_t *tau, *Qkth;
  SNAPBASEREAL_t prev_Snorm, kth_Snorm, kth_Rnorm, tol_f, tol_g, const_tmp;
  unsigned int i,j,k,found_degree=UINT_MAX,case1flag=0,case4flag=0;
  unsigned int mincdeg, k4case3;
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
    printf("entering qrstep\n");
  }
#endif
  if ( SNAP_degree(poly_cof) == 0 || SNAP_degree(poly_cog) == 0 ) {
    return 0;
  }
  // Initialization and Sylvester matrix
  SNAPMTX_init(matrix_syl);
  SNAPMTX_sylvester_matrix_transpose(matrix_syl, poly_cof, poly_cog);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
  for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  // QR decomposition
#ifdef SNAP_SUPD_T
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#endif
  ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_qrgcd: dgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0
#endif
  ) {
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_qrgcd: zgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_qrgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matrix_syl), SNAPMTX_cols(matrix_syl), SNAPMTX_elems(matrix_syl), SNAPMTX_csize(matrix_syl), tau) != 0 ) {
    for ( i=0; i<SNAPMTX_rows(matrix_syl); i++ ) {
      SNAPBASE_clear(*(tau+i));
    }
    free(tau);
    SNAPMTX_clear(matrix_syl);
    printf("ERROR in agcd_qrgcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    return found_degree;
  }
#endif
  // Now we seek the relative smallest gap to recover a candidate approximate gcd.
  SNAPBASEREAL_init(kth_Rnorm);
  SNAPBASEREAL_init(kth_Snorm);
  SNAPBASE_abs(kth_Snorm,SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
  SNAPBASEREAL_init_set(prev_Snorm, kth_Snorm);
  k = 0;
  k4case3 = 0;
  SNAPBASEREAL_init(const_tmp);
#ifdef SNAPLIB_DEBUG
  if ( qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
    mpfr_printf("%d-th Rnorm: %Re\n", k, kth_Snorm);
#else
    printf("%d-th Rnorm: %e\n", k, kth_Snorm);
#endif
  }
#endif
//	SNAPBASE_mul(const_case1, tol_temp, sqrtcols);
  if ( SNAPBASEREAL_cmp(kth_Snorm, tol) > 0 ) { // CASE 1 (there is no chance to have agcd)
    case1flag = 1;
#ifdef SNAPLIB_DEBUG
    if ( qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
      mpfr_printf("CASE 1 detected! (%Re > %Re=tol*sqrtcols)\n", kth_Snorm, tol);
#else
      printf("CASE 1 detected! (%e > %e=tol*sqrtcols)\n", kth_Snorm, tol);
#endif
    }
#endif
    // for CASE 1, we do nothing here.
  } else if ( qrstep_order != BEZOUTONLY ) { // CASES 2, 3 and 4 (there is a possibility to have agcd)
    mincdeg = SNAPLIB_MIN(SNAP_degree(poly_cof),SNAP_degree(poly_cog));
    while ( ++k <= mincdeg ) {
      // Update kth_Rnorm
      SNAPMTX_BLAS_nrm2(kth_Rnorm, k+1, SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-k,SNAPMTX_rows(matrix_syl)-1-k),1);
      // S-norm can be computed by ...
      SNAPBASEREAL_add(kth_Snorm, kth_Snorm, kth_Rnorm);
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
        mpfr_printf("%d-th Rnorm: %Re, Snorm: %Re (Case1 threshold: %Re)\n", k, kth_Rnorm, kth_Snorm, tol);
#else
        printf("%d-th Rnorm: %e, Snorm: %e (Case1 threshold: %e)\n", k, kth_Rnorm, kth_Snorm, tol);
#endif
      }
#endif
      if ( SNAPBASEREAL_cmp(kth_Snorm, tol) < 0 ) {
        // 1/10/tol < NR/eps => eps < 10*NR*tol
        SNAPBASEREAL_mul(const_tmp, kth_Snorm, tol);
        SNAPBASEREAL_mul_ui(const_tmp, const_tmp, 10);
        if ( SNAPBASEREAL_cmp(prev_Snorm, const_tmp) < 0 ) { // one of case 3
          k4case3 = k;
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("Big Jump! (k=%d)\n", k4case3);
          }
#endif
        }
        SNAPBASEREAL_set(prev_Snorm, kth_Snorm);
      } else {
        // 1/10/tol < NR/eps => eps < 10*NR*tol
        SNAPBASEREAL_mul(const_tmp, kth_Snorm, tol);
        SNAPBASEREAL_mul_ui(const_tmp, const_tmp, 10);
        if ( SNAPBASEREAL_cmp(prev_Snorm, const_tmp) < 0 ) { // case 2
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("CASE 2 Detected! (k=%d)\n", k);
          }
#endif
          break;
        } else if ( k4case3 != 0 ) { // case 3
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("CASE 3 Detected! (k=%d)\n", k4case3);
          }
#endif
          k = k4case3;
          break;
        } else { // case 4
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
            printf("CASE 4 Detected! (k=%d)\n", k);
          }
#endif
          case4flag = 1;
          break;
        }
      }
    } // END OF WHILE LOOP for k
    if ( mincdeg < k ) {
      case1flag = 1;  // this means no gcd found.
    }
    if ( case1flag == 0 && case4flag == 0 ) {
      // Construct approximate GCD and cofactors
      SNAP_init(poly_r);
      SNAP_init(poly_t);
      SNAP_init(poly_ncf);
      SNAP_init(poly_ncg);
      SNAPBASEREAL_init(tol_f);
      SNAPBASEREAL_init(tol_g);
      // poly_r <= the last $(k)$-th row of R
      SNAP_set_rbase(poly_r, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-k, SNAPMTX_rows(matrix_syl)-1-k), k+1);
#ifdef SNAP_SUPDC_T
      if ( SNAPBASEREAL_sgn(SNAPBASE_real(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#elif defined(SNAP_SUPFC_T)
      if ( SNAPBASEREAL_sgn(SNAPBASE_realref(SNAP_lcoef(poly_r))) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#else
      if ( SNAPBASE_sgn(SNAP_lcoef(poly_r)) < 0 ) {
        SNAP_self_neg(poly_r);
      }
#endif
      // poly_t <= poly_d * poly_r
      SNAP_mul(poly_t, poly_d, poly_r);
      // poly_t <= poly_t/norm(poly_t) : normalization
      SNAP_2norm(const_tmp, poly_t);
      SNAP_div_basereal(poly_t, poly_t, const_tmp);
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncf, poly_f, poly_t);
      // check the tolerance
      SNAP_mul(poly_r, poly_ncf, poly_t);
      SNAP_sub(poly_r, poly_f, poly_r);
      SNAP_2norm(tol_f, poly_r);
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
        mpfr_printf("tol_f: %Re\n", tol_f);
#else
        printf("tol_f: %e\n", tol_f);
#endif
      }
#endif
      // poly_ncf <= cofactor of poly_f by poly_t (update candidate cofactor)
      SNAP_lsdiv_q(poly_ncg, poly_g, poly_t);
      // check the tolerance
      SNAP_mul(poly_r, poly_ncg, poly_t);
      SNAP_sub(poly_r, poly_g, poly_r);
      SNAP_2norm(tol_g, poly_r);
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
        mpfr_printf("tol_g: %Re\n", tol_g);
#else
        printf("tol_g: %e\n", tol_g);
#endif
      }
#endif
      SNAPBASEREAL_add(const_tmp, tol_f, tol_g);
      SNAPBASEREAL_div_ui(const_tmp, const_tmp, 10);
      if ( SNAPBASEREAL_cmp(const_tmp, tol) < 0 ) { // approximate gcd found!!
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
          mpfr_printf("Found tol_f: %Re, tol_g: %Re\n", tol_f, tol_g);
#else
          printf("Found tol_f: %e, tol_g: %e\n", tol_f, tol_g);
#endif
        }
#endif
        // update several polynomials.
        SNAP_set(poly_cof, poly_ncf);
        SNAP_set(poly_cog, poly_ncg);
        SNAP_set(poly_d, poly_t);
        found_degree = k;
        if ( SNAPBASEREAL_cmp(tol_f, tol_g) < 0 ) {
          SNAPBASEREAL_set(tol_res, tol_g);
        } else {
          SNAPBASEREAL_set(tol_res, tol_f);
        }
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("Found Degree: %d\n", found_degree);
        }
#endif
      }
      SNAP_clear(poly_r);
      SNAP_clear(poly_t);
      SNAP_clear(poly_ncf);
      SNAP_clear(poly_ncg);
      SNAPBASEREAL_clear(tol_f);
      SNAPBASEREAL_clear(tol_g);
    }
    // Check Difficult case or Not
    if ( case4flag == 1 ) {
#ifdef SNAPLIB_DEBUG
      if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
        printf("Difficult CASE detected!\n");
      }
#endif
      if ( split_flag == SPLITALLOWED ) { // Call Split and qrstep recursively.
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is called...\n");
        }
#endif
        /* ========================== find Rnorm > 5000*tol ============================== */
        // at the above step kth_Snorm > tol and (k-1)th_Snorm <= tol
        // we re-use kth_Snorm as 5000*tol
        SNAPBASEREAL_mul_ui(kth_Snorm, tol, 5000);
        while ( k < SNAPMTX_rows(matrix_syl) ) {
          // Update kth_Rnorm
          SNAPMTX_BLAS_nrm2(kth_Rnorm, k+1, SNAPMTX_elemp(matrix_syl,SNAPMTX_rows(matrix_syl)-1-k,SNAPMTX_rows(matrix_syl)-1-k),1);
#ifdef SNAPLIB_DEBUG
          if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
#ifdef SNAP_SUPFR_T
            mpfr_printf("%d-th Rnorm: %Re (from k)\n", k, kth_Rnorm);
#else
            printf("%d-th Rnorm: %e (from k)\n", k, kth_Rnorm);
#endif
          }
#endif
          // Rnorm > 5000*tol
          if ( SNAPBASEREAL_cmp(kth_Rnorm, kth_Snorm) > 0 ) {
#ifdef SNAPLIB_DEBUG
            if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
              printf("Splitting at k=%d satisyfing Rnorm > 5000*tol.\n",k);
            }
#endif
            SNAP_init(poly_ncf);
            SNAP_init(poly_ncg);
            SNAP_init(poly_prs1);
            SNAP_init(poly_prs2);
            SNAP_set_rbase(poly_prs1, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-2-k, SNAPMTX_rows(matrix_syl)-2-k), k+2);
            SNAP_set_rbase(poly_prs2, SNAPMTX_elemp(matrix_syl, SNAPMTX_rows(matrix_syl)-1-k, SNAPMTX_rows(matrix_syl)-1-k), k+1);
            SNAP_2norm(kth_Rnorm, poly_prs1);
            SNAP_2norm(kth_Snorm, poly_prs2);
            SNAP_div_basereal(poly_prs1, poly_prs1, kth_Rnorm);
            SNAP_div_basereal(poly_prs2, poly_prs2, kth_Snorm);
            SNAPBASEREAL_div_ui(const_tmp, tol, 100);
            if ( SNAP_agcd_exqrgcd_split(poly_ncf, poly_prs1, const_tmp) == 0 && SNAP_agcd_exqrgcd_split(poly_ncg, poly_prs2, const_tmp) == 0 ) {
              found_degree = SNAP_agcd_qrgcd_qrstep(poly_f, poly_g, poly_ncf, poly_ncg, poly_d, poly_u, poly_v, const_tmp, tol_res, qrstep_order, SPLITDISALLOWED);
              if ( found_degree != UINT_MAX && found_degree != 0 ) {
                SNAP_set(poly_cof, poly_ncf);
                SNAP_set(poly_cog, poly_ncg);
              }
              SNAP_clear(poly_ncf);
              SNAP_clear(poly_ncg);
              SNAP_clear(poly_prs1);
              SNAP_clear(poly_prs2);
              for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
                SNAPBASE_clear(*(tau+j));
              }
              free(tau);
              SNAPMTX_clear(matrix_syl);
              SNAPBASEREAL_clear(kth_Snorm);
              SNAPBASEREAL_clear(kth_Rnorm);
              SNAPBASEREAL_clear(const_tmp);
              SNAPBASEREAL_clear(prev_Snorm);
              return found_degree;
            } else {
#ifdef SNAPLIB_DEBUG
              if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
                printf("SPLIT failed...\n");
              }
#endif
              SNAP_clear(poly_ncf);
              SNAP_clear(poly_ncg);
              SNAP_clear(poly_prs1);
              SNAP_clear(poly_prs2);
              for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
                SNAPBASE_clear(*(tau+j));
              }
              free(tau);
              SNAPMTX_clear(matrix_syl);
              SNAPBASEREAL_clear(kth_Snorm);
              SNAPBASEREAL_clear(kth_Rnorm);
              SNAPBASEREAL_clear(const_tmp);
              SNAPBASEREAL_clear(prev_Snorm);
              return found_degree;
            }
          }
          k++;
        }
      } else {
#ifdef SNAPLIB_DEBUG
        if ( qrstep_order != BEZOUTONLY && qrstep_order != NORMALONLY ) {
          printf("SPLIT is not allowed. Skipped...\n");
        }
#endif
      }
    }
  }
  if ( found_degree == UINT_MAX ) {
    // poly_d <== 1.0 hence we do nothing for poly_d
    if ( case1flag == 1 ) {
      found_degree = 0;
    }
    // poly_u and poly_v be the cofactors from the last row of Q.
    // NOTE: poly_u is the cofactor of poly_g and poly_v is of poly_f.
    // if ( qrstep_order == NORMAL || qrstep_order == NORMALONLY || qrstep_order == BEZOUTONLY ) {
    if ( qrstep_order == BEZOUTONLY ) {
      Qkth = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_rows(matrix_syl));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_init(*(Qkth+j));
      }
      SNAP_agcd_exqrgcd_Qkthcolumn(Qkth, matrix_syl, tau, SNAPMTX_rows(matrix_syl)-1);
      for ( i=0; i<SNAP_degree(poly_cog); i++ ) {
        SNAP_set_coef(poly_u, i, *(Qkth+SNAP_degree(poly_cog)-1-i));
      }
      SNAP_degree(poly_u) = SNAP_degree(poly_cog)-1;
      for ( i=0; i<SNAP_degree(poly_cof); i++ ) {
        SNAP_set_coef(poly_v, i, *(Qkth+SNAP_degree(poly_cog)+SNAP_degree(poly_cof)-1-i));
      }
      SNAP_degree(poly_v) = SNAP_degree(poly_cof)-1;
      SNAP_div_base(poly_u, poly_u, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      SNAP_div_base(poly_v, poly_v, SNAPMTX_elem(matrix_syl,SNAPMTX_rows(matrix_syl)-1,SNAPMTX_rows(matrix_syl)-1));
      for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
        SNAPBASE_clear(*(Qkth+j));
      }
      free(Qkth);
    }
  }
  for ( j=0; j<SNAPMTX_rows(matrix_syl); j++ ) {
    SNAPBASE_clear(*(tau+j));
  }
  free(tau);
  SNAPMTX_clear(matrix_syl);
  SNAPBASEREAL_clear(kth_Snorm);
  SNAPBASEREAL_clear(kth_Rnorm);
  SNAPBASEREAL_clear(const_tmp);
  SNAPBASEREAL_clear(prev_Snorm);
  return found_degree;
}
int SNAP_agcd_qrgcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol) {
  SNAPBASEREAL_t norm_f, norm_g, tol_res, norm_cf, norm_cg;
  SNAP_t poly_f, poly_g, poly_cof, poly_cog, poly_u, poly_v;
  unsigned int found_degree, trials;
  int result = -1;
  // initialization
  SNAP_init(poly_u);
  SNAP_init(poly_v);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(tol_res);
  SNAPBASEREAL_init(norm_cf);
  SNAPBASEREAL_init(norm_cg);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
  SNAP_init_set(poly_cog, poly_g);
  SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
  SNAP_degree(poly_d) = 0;
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPFR_T
  mpfr_printf("input tolerance: %Re\n", tol);
#else
  printf("input tolerance: %e\n", tol);
#endif
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_cof)) || SNAPBASE_zeroq(SNAP_lcoef(poly_cog)) ) {
    SNAP_clear(poly_f);
    SNAP_clear(poly_g);
    SNAP_clear(poly_cof);
    SNAP_clear(poly_cog);
    SNAPBASEREAL_clear(norm_f);
    SNAPBASEREAL_clear(norm_g);
    SNAPBASEREAL_clear(tol_res);
    SNAPBASEREAL_clear(norm_cf);
    SNAPBASEREAL_clear(norm_cg);
    fprintf(stderr, "qrgcd: polynomials must be non-zero!\n");
    return -1;
  }
#ifdef SNAPLIB_DEBUG
  printf("======= Order: Normal =======\n");
#endif
  // some agcd will be found then return, otherwise change qrstep_order.
  SNAPBASEREAL_set_ui(norm_cf, 1);
  SNAPBASEREAL_set_ui(norm_cg, 1);
  trials = 0;
  while (1) {
    if ( trials > 0 ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    // SNAP implementation calls split only once.
    if ( trials++ < 1 ) {
      found_degree = SNAP_agcd_qrgcd_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_res,NORMAL,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_qrgcd_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_res,REVERSAL,SPLITDISALLOWED);
    }
    if ( found_degree == UINT_MAX || found_degree == 0 ) {
      if ( trials == 1 ) {
        break;  // re-try from the other side.
      }
    } else {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
    }
    if ( trials == 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      SNAP_self_reversal(poly_d);
      SNAP_self_reversal(poly_f1);
      SNAP_self_reversal(poly_g1);
      result = 0;
      break;
    }
  }
  if ( result == -1 ) {
#ifdef SNAPLIB_DEBUG
    printf("======= Second Order: Reversal =======\n");
#endif
    SNAP_self_reversal(poly_f);
    SNAP_self_reversal(poly_g);
    SNAP_init_set(poly_cof, poly_f); // always qrstep does for cof and cog.
    SNAP_init_set(poly_cog, poly_g);
    SNAP_set_coef_d(poly_d, 0, 1.0); // always qrstep does keep poly_d being agcd.
    SNAP_degree(poly_d) = 0;
    trials = 0;
  }
  while ( result == -1 ) {
    if ( trials > 0 ) {
      SNAP_self_reversal(poly_f);
      SNAP_self_reversal(poly_g);
      SNAP_self_reversal(poly_cof);
      SNAP_self_reversal(poly_cog);
      SNAP_self_reversal(poly_d);
    }
    if ( trials++ < 1 ) {
      found_degree = SNAP_agcd_qrgcd_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_res,REVERSAL,SPLITALLOWED);
    } else {
      found_degree = SNAP_agcd_qrgcd_qrstep(poly_f,poly_g,poly_cof,poly_cog,poly_d,poly_u,poly_v, tol, tol_res,NORMAL,SPLITDISALLOWED);
    }
    if ( found_degree != UINT_MAX && found_degree != 0 ) {
      SNAP_2norm(norm_cf, poly_cof);
      SNAP_2norm(norm_cg, poly_cog);
      SNAP_div_basereal(poly_cof, poly_cof, norm_cf);
      SNAP_div_basereal(poly_cog, poly_cog, norm_cg);
    }
    if ( trials == 2 ) {
      SNAPBASEREAL_mul(norm_cf, norm_f, norm_cf);
      SNAPBASEREAL_mul(norm_cg, norm_g, norm_cg);
      SNAP_mul_basereal(poly_f1, poly_cof, norm_cf);
      SNAP_mul_basereal(poly_g1, poly_cog, norm_cg);
      result = 0;
      break;
    }
  }
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAP_clear(poly_cof);
  SNAP_clear(poly_cog);
  SNAP_clear(poly_u);
  SNAP_clear(poly_v);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(tol_res);
  SNAPBASEREAL_clear(norm_cf);
  SNAPBASEREAL_clear(norm_cg);
  return result;
}
/* ================== END[agcd_over_reals] ================== */
