/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPjoin(a,b) a##b
#define SNAP2str(a) #a

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

//#define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

// define if shows elapsed time
// #define SNAPLIB_SHOW_ELAPSED_TIME

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.5le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPDC_T)
void show_sdmdc(sdmdc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmdc_rows(mtx); i++) {
    for(j=0; j<sdmdc_cols(mtx); j++) {
      printf("(%+.5le%+.5le*I), ",creal(sdmdc_elem(mtx,i,j)),cimag(sdmdc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_dclist(dcomplex *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.5le%+.5le*I), ",creal(*(list+i)),cimag(*(list+i)));
  }
  printf("\n");
}
#elif defined(SNAP_SUPFR_T)
void show_sdmfr(sdmfr_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfr_rows(mtx); i++) {
    for(j=0; j<sdmfr_cols(mtx); j++) {
      mpfr_printf("%+.5Re ",sdmfr_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_frlist(mpfr_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("%+.5Re ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPFC_T)
void show_sdmfc(sdmfc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfc_rows(mtx); i++) {
    for(j=0; j<sdmfc_cols(mtx); j++) {
      mpfr_printf("(%+.5Re%+.5Re*I), ",mpc_realref(sdmfc_elem(mtx,i,j)),mpc_imagref(sdmfc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_fclist(mpc_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("(%+.5Re%+.5Re*I), ",mpc_realref(*(list+i)),mpc_imagref(*(list+i)));
  }
  printf("\n");
}
#endif
#endif

/* ================== START [agcd_uvgcd] ================== */

// #define SNAPLIB_MPFR_RND MPFR_RNDN
// #define SNAPLIB_MPC_RND MPC_RNDNN
// ========== THE FOLLOWING MACROS SHOULD BE MODIFIED TO BE COMPATIBLE WITH MPFR AND MPC. ==========
#define SNAPLIB_MIN(x,y) ((x)<(y) ? (x):(y))
#define SNAPLIB_MAX(x,y) ((x)>(y) ? (x):(y))

/*!
 * @brief The smallest singular value and corresponding singular vector.
 * @param[in] R The upper triangular (m,n) matrix of QR decomposition.
 * @param[out] zvect The corresponding singular vector. size >= n.
 * @param[in] yvect A temporary array for the vector used in the calculation. size >= n.
 * @param[out] sigma_m1 The smallest singular value.
 * @return The number of iterations.
 */
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t *sigma_m1, gmp_randstate_t rstate)
#else
static int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t sigma_m1, gmp_randstate_t rstate)
#endif
{
  unsigned int i, iteration=0, max_iteration;
  SNAPBASE_t *zpt,*ypt;
  SNAPBASEREAL_t abs_threshold;
  // initialize threshold
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  max_iteration = ((unsigned int) SNAPBASEREAL_get_default_prec());
  SNAPBASEREAL_init_set_ui(abs_threshold, max_iteration-4);
  SNAPBASEREAL_neg(abs_threshold, abs_threshold);
  SNAPBASEREAL_exp2(abs_threshold, abs_threshold);
  max_iteration = (unsigned int) ((int) (-1)*log10(SNAPBASEREAL_get_d(abs_threshold)))+1;
#elif defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  abs_threshold = 1.0e-12;
  max_iteration = 3;
#endif
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we multiply the max_iteration by 2 for complex coefficients.
  // max_iteration *= 2;
#endif
  // initialize yvect
  for ( i=0,ypt=yvect; i<SNAPMTX_cols(R); i++,ypt++ ) {
    SNAPBASE_urandom(*ypt, rstate);
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    SNAPBASE_sub_dc(*ypt, *ypt, 0.5+0.5*_Complex_I);
    // for a debug purpose
    // SNAPBASE_set_dc(*ypt, i+i*_Complex_I);
#else
    SNAPBASE_sub_d(*ypt, *ypt, 0.5);
#endif
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), yvect,1);
#else
  SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), yvect,1);
#endif
  for ( i=0,ypt=yvect; i<SNAPMTX_cols(R); i++,ypt++ ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPBASE_div_basereal(*ypt, *ypt, *sigma_m1);
#else
    SNAPBASE_div_basereal(*ypt, *ypt, sigma_m1);
#endif
  }
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  printf("norm: %e (#iteration = %d (%d))\n", *sigma_m1, iteration, max_iteration);
#else
  mpfr_printf("norm: %Re (#iteration = %d (%d))\n", sigma_m1, iteration, max_iteration);
#endif
#endif
  while ( iteration++ < max_iteration ) {
    // clone yvect to zvect
    for ( i=0,ypt=yvect,zpt=zvect; i<SNAPMTX_cols(R); i++,ypt++,zpt++ ) {
      SNAPBASE_set(*zpt, *ypt);
    }
#if defined(SNAP_SUPD_T)
    if ( LAPACKE_dtrtrs(LAPACK_ROW_MAJOR, 'U', 'T', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": LAPACKE_dtrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
#elif defined(SNAP_SUPDC_T)
    if ( LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'U', 'C', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": LAPACKE_ztrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
#elif defined(SNAP_SUPFR_T)
    if ( snap_mlapack_frtrtrs('U', 'T', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": snap_mlapack_frtrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
#elif defined(SNAP_SUPFC_T)
    if ( snap_mlapack_fctrtrs('U', 'C', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": snap_mlapack_fctrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
    printf("yvect: ");
    show_dlist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPDC_T)
    printf("yvect: ");
    show_dclist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPFR_T)
    printf("yvect: ");
    show_frlist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPFC_T)
    printf("yvect: ");
    show_fclist(yvect, SNAPMTX_cols(R));
#endif
#endif
#if defined(SNAP_SUPD_T)
    if ( LAPACKE_dtrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": LAPACKE_dtrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
    SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPDC_T)
    if ( LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": LAPACKE_ztrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
    SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPFR_T)
    // WRITE HERE FOR SUPFR_T
    if ( snap_mlapack_frtrtrs('U', 'N', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": snap_mlapack_frtrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
    SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPFC_T)
    // WRITE HERE FOR SUPFC_T
    if ( snap_mlapack_fctrtrs('U', 'N', 'N', SNAPMTX_cols(R), 1, SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_iteration_7_3), ": snap_mlapack_fctrtrs could not compute the linear equation solution.\n");
      SNAPBASEREAL_clear(abs_threshold);
      return -1;
    }
    SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), yvect,1);
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
    printf("yvect: ");
    show_dlist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPDC_T)
    printf("yvect: ");
    show_dclist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPFR_T)
    printf("yvect: ");
    show_frlist(yvect, SNAPMTX_cols(R));
#elif defined(SNAP_SUPFC_T)
    printf("yvect: ");
    show_fclist(yvect, SNAPMTX_cols(R));
#endif
#endif
    for ( i=0,ypt=yvect; i<SNAPMTX_cols(R); i++,ypt++ ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAPBASE_div_basereal(*ypt, *ypt, *sigma_m1);
#else
      SNAPBASE_div_basereal(*ypt, *ypt, sigma_m1);
#endif
    }
    // convergence check
    for ( i=0,ypt=yvect,zpt=zvect; i<SNAPMTX_cols(R); i++,ypt++,zpt++ ) {
      SNAPBASE_sub(*zpt, *zpt, *ypt);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), zvect,1);
#ifdef SNAPLIB_DEBUG
    printf("residual-norm: %e (#iteration = %d)\n", *sigma_m1, iteration);
#endif
    if ( SNAPBASEREAL_cmp(*sigma_m1, abs_threshold) <= 0 ) {
      break;
    }
#else
    SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), zvect,1);
#ifdef SNAPLIB_DEBUG
    mpfr_printf("residual-norm: %Re (#iteration = %d)\n", sigma_m1, iteration);
#endif
    if ( SNAPBASEREAL_cmp(sigma_m1, abs_threshold) <= 0 ) {
      break;
    }
#endif
  }
  // clone yvect to zvect
  for ( i=0,ypt=yvect,zpt=zvect; i<SNAPMTX_cols(R); i++,ypt++,zpt++ ) {
    SNAPBASE_set(*zpt, *ypt);
  }
#if defined(SNAP_SUPD_T)
  cblas_dtrmv(CblasRowMajor, CblasUpper, CblasNoTrans, CblasNonUnit, SNAPMTX_cols(R), SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1);
  SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPDC_T)
  cblas_ztrmv(CblasRowMajor, CblasUpper, CblasNoTrans, CblasNonUnit, SNAPMTX_cols(R), SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1);
  SNAPMTX_BLAS_nrm2(*sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPFR_T)
  // WRITE HERE FOR SUPFR_T
  snap_mblas_frtrmv('U', 'N', 'N', SNAPMTX_cols(R), SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1);
  SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), yvect,1);
#elif defined(SNAP_SUPFC_T)
  // WRITE HERE FOR SUPFC_T
  snap_mblas_fctrmv('U', 'N', 'N', SNAPMTX_cols(R), SNAPMTX_elems(R), SNAPMTX_csize(R), yvect, 1);
  SNAPMTX_BLAS_nrm2(sigma_m1, SNAPMTX_cols(R), yvect,1);
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("corresponding singular vector\n");
  show_dlist(zvect, SNAPMTX_cols(R));
  printf("smallest singular value: %e (#iteration = %d)\n", *sigma_m1, iteration);
#elif defined(SNAP_SUPDC_T)
  printf("corresponding singular vector\n");
  show_dclist(zvect, SNAPMTX_cols(R));
  printf("smallest singular value: %e (#iteration = %d)\n", *sigma_m1, iteration);
#elif defined(SNAP_SUPFR_T)
  printf("corresponding singular vector\n");
  show_frlist(zvect, SNAPMTX_cols(R));
  mpfr_printf("smallest singular value: %Re (#iteration = %d)\n", sigma_m1, iteration);
#elif defined(SNAP_SUPFC_T)
  printf("corresponding singular vector\n");
  show_fclist(zvect, SNAPMTX_cols(R));
  mpfr_printf("smallest singular value: %Re (#iteration = %d)\n", sigma_m1, iteration);
#endif
#endif
  SNAPBASEREAL_clear(abs_threshold);
  printf("SV: #iterations: %d\n", iteration);
  return iteration;
}

/*!
 *  @brief Update subresultant matrix from the order j to j-1.
 *  @param[in,out] Sn The subresultant matrix array.
 *  @param[in] poly_f The input polynomial f.
 *  @param[in] poly_g The input polynomial g.
 *  @param[in] vect1 A temporary vector. size >= current_r+1.
 *  @param[in] vect2 A temporary vector. size >= cols(Sn).
 *  @param[in] current_r The current order j of the subresultant mapping.
 *  @return 0 if success or -1 if failure.
 */
static int SNAP_agcd_uvgcd_update(SNAPMTX_t Sn, SNAP_t poly_f, SNAP_t poly_g, SNAPBASE_t *vect1, SNAPBASE_t *vect2, SNAPBASE_t *tau, unsigned int current_r) {
  unsigned int i,j,c,maxi;
  SNAPBASE_t *pt,*pt2;
  // Insert the additional columns into the last two columns and householder reflector target vectors.
  c = SNAPMTX_cols(Sn);
  maxi = SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r+1;
  for ( i=SNAP_degree(poly_g)-current_r,pt=SNAP_coefs(poly_f); i<maxi; i++,pt++ ) {
    SNAPBASE_set(SNAPMTX_elem(Sn, i, c), *pt);
  }
  c = SNAPMTX_cols(Sn)+1;
  for ( i=SNAP_degree(poly_f)-current_r,pt=SNAP_coefs(poly_g); i<maxi; i++,pt++ ) {
    SNAPBASE_set(SNAPMTX_elem(Sn, i, c), *pt);
  }
  SNAPMTX_rows(Sn)++;
  SNAPMTX_cols(Sn)+=2;
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("Sn (inserted,%d,%d)\n",sdmd_rows(Sn),sdmd_cols(Sn));
  show_sdmd(Sn);
#elif defined(SNAP_SUPDC_T)
  printf("Sn (inserted,%d,%d)\n",sdmdc_rows(Sn),sdmdc_cols(Sn));
  show_sdmdc(Sn);
#elif defined(SNAP_SUPFR_T)
  printf("Sn (inserted,%d,%d)\n",sdmfr_rows(Sn),sdmfr_cols(Sn));
  show_sdmfr(Sn);
#elif defined(SNAP_SUPFC_T)
  printf("Sn (inserted,%d,%d)\n",sdmfc_rows(Sn),sdmfc_cols(Sn));
  show_sdmfc(Sn);
#endif
#endif
  maxi = SNAPMTX_cols(Sn)-2;
  for ( i=0; i<maxi; i++ ) {
    SNAPBASE_set_ui(*vect1, 1);
    for ( j=i+1,pt=vect1+1,pt2=SNAPMTX_elemp(Sn,i+1,i); j<SNAPMTX_rows(Sn)-1; j++,pt++,pt2+=SNAPMTX_csize(Sn)) {
      SNAPBASE_set(*pt,*pt2);
    }
    SNAPBASE_set_ui(*pt,0);
    // apply the househoulder reflector.
#if defined(SNAP_SUPD_T)
    if ( LAPACKE_dlarfx( LAPACK_ROW_MAJOR, 'L', SNAPMTX_rows(Sn)-i, 2, vect1, *(tau+i), SNAPMTX_elemp(Sn,i,maxi), SNAPMTX_csize(Sn), vect2) < 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_update), ": LAPACKE_dlarfx could not apply the past householder reflection.\n");
      return -1;
    }
#elif defined(SNAP_SUPDC_T)
    if ( LAPACKE_zlarfx( LAPACK_ROW_MAJOR, 'L', SNAPMTX_rows(Sn)-i, 2, vect1, *(tau+i), SNAPMTX_elemp(Sn,i,maxi), SNAPMTX_csize(Sn), vect2) < 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_update), ": LAPACKE_zlarfx could not apply the past householder reflection.\n");
      return -1;
    }
#elif defined(SNAP_SUPFR_T)
    if ( snap_mlapack_frlarfx( 'L', SNAPMTX_rows(Sn)-i, 2, vect1, *(tau+i), SNAPMTX_elemp(Sn,i,maxi), SNAPMTX_csize(Sn), vect2) < 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_update), ": snap_mlapack_frlarfx could not apply the past householder reflection.\n");
      return -1;
    }
#elif defined(SNAP_SUPFC_T)
    if ( snap_mlapack_fclarfx( 'L', SNAPMTX_rows(Sn)-i, 2, vect1, *(tau+i), SNAPMTX_elemp(Sn,i,maxi), SNAPMTX_csize(Sn), vect2) < 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_update), ": snap_mlapack_fclarfx could not apply the past householder reflection.\n");
      return -1;
    }
#endif
  }
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("Sn (applied,%d,%d)\n",sdmd_rows(Sn),sdmd_cols(Sn));
  show_sdmd(Sn);
#elif defined(SNAP_SUPDC_T)
  printf("Sn (applied,%d,%d)\n",sdmdc_rows(Sn),sdmdc_cols(Sn));
  show_sdmdc(Sn);
#elif defined(SNAP_SUPFR_T)
  printf("Sn (applied,%d,%d)\n",sdmfr_rows(Sn),sdmfr_cols(Sn));
  show_sdmfr(Sn);
#elif defined(SNAP_SUPFC_T)
  printf("Sn (applied,%d,%d)\n",sdmfc_rows(Sn),sdmfc_cols(Sn));
  show_sdmfc(Sn);
#endif
#endif
  // QR decomposition for the added last 2 columns.
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0
#endif
  ) {
    printf("ERROR in agcd_uvgcd_update: dgeqrf could not compute the QR decomposition.\n");
    return -1;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0
#endif
  ) {
    printf("ERROR in agcd_uvgcd_update: zgeqrf could not compute the QR decomposition.\n");
    return -1;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0 ) {
    printf("ERROR in agcd_uvgcd_update: mlapack_frgeqrf could not compute the QR decomposition.\n");
    return -1;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAP_degree(poly_g)+SNAP_degree(poly_f)-current_r-SNAPMTX_cols(Sn)+3, 2, SNAPMTX_elemp(Sn, SNAPMTX_cols(Sn)-2, SNAPMTX_cols(Sn)-2), SNAPMTX_csize(Sn), tau+SNAPMTX_cols(Sn)-2) != 0 ) {
    printf("ERROR in agcd_uvgcd_update: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    return -1;
  }
#endif
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we keep the conjugated tau for larfx.
  SNAPBASE_conj(*(tau+SNAPMTX_cols(Sn)-2),*(tau+SNAPMTX_cols(Sn)-2));
  SNAPBASE_conj(*(tau+SNAPMTX_cols(Sn)-1),*(tau+SNAPMTX_cols(Sn)-1));
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("Sn (updated,%d,%d)\n",sdmd_rows(Sn),sdmd_cols(Sn));
  show_sdmd(Sn);
#elif defined(SNAP_SUPDC_T)
  printf("Sn (updated,%d,%d)\n",sdmdc_rows(Sn),sdmdc_cols(Sn));
  show_sdmdc(Sn);
#elif defined(SNAP_SUPFR_T)
  printf("Sn (updated,%d,%d)\n",sdmfr_rows(Sn),sdmfr_cols(Sn));
  show_sdmfr(Sn);
#elif defined(SNAP_SUPFC_T)
  printf("Sn (updated,%d,%d)\n",sdmfc_rows(Sn),sdmfc_cols(Sn));
  show_sdmfc(Sn);
#endif
#endif
  return 0;
}

/*!
 * @brief An initial approximate GCD by Least Square.
 * @warning poly_X must be all different.
 * @param[in] poly_f The input polynomial f.
 * @param[in] poly_g The input polynomial g.
 * @param[in] poly_f1 An initial cofactor of f.
 * @param[in] poly_g1 An initial cofactor of g.
 * @param[out] poly_d An initial approximate GCD.
 * @param[in] current_r The current order j of the subresultant mapping.
 * @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcd_solving_10_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r) {
  SNAPMTX_t matrix_conv, matrix_dummy;
  unsigned int i,deg_d;
  // initialization
  deg_d = current_r+1;
  // initialize the matrix as the whole sized matrix.
  SNAPMTX_init2(matrix_conv, SNAP_degree(poly_f)+SNAP_degree(poly_g)+2, deg_d+1);
  // set up the upper half as C(f1)
#if defined(SNAP_SUPDC_T)
  SNAPMTX_convolution_matrix_conj(matrix_conv, poly_f1, deg_d);
#else
  SNAPMTX_convolution_matrix(matrix_conv, poly_f1, deg_d);
#endif
  SNAPMTX_rows(matrix_conv) = SNAP_degree(poly_f)+SNAP_degree(poly_g)+2;
  // set up the lower half as C(g1)
  SNAPMTX_csize(matrix_dummy) = SNAPMTX_csize(matrix_conv);
  SNAPMTX_rsize(matrix_dummy) = SNAP_degree(poly_g)+1;
  SNAPMTX_cols(matrix_dummy) = SNAPMTX_cols(matrix_conv);
  SNAPMTX_rows(matrix_dummy) = SNAP_degree(poly_g)+1;
  SNAPMTX_elems(matrix_dummy) = SNAPMTX_elemp(matrix_conv, SNAP_degree(poly_f)+1, 0);
#if defined(SNAP_SUPDC_T)
  SNAPMTX_convolution_matrix_conj(matrix_dummy, poly_g1, deg_d);
#else
  SNAPMTX_convolution_matrix(matrix_dummy, poly_g1, deg_d);
#endif
  // *B should be the coefficient of poly_d but the clone of poly_f and poly_g.
  SNAP_set(poly_d, poly_f);
  for ( i=0; i<=SNAP_degree(poly_g); i++ ) {
    SNAP_set_coef(poly_d, SNAP_degree(poly_f)+1+i, SNAP_coef(poly_g,i));
  }
#ifdef SNAPLIB_DEBUG
//printf("now calling the corresponding _gels routine.\n");
#if defined(SNAP_SUPDC_T)
  printf("-poly_s: ");
  show_dclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
  printf("poly_t: ");
  show_dclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
  printf("poly_d: ");
  show_dclist(SNAP_coefs(poly_d), SNAPMTX_rows(matrix_conv));
  printf("matrix_conv:\n");
  show_sdmdc(matrix_conv);
#elif defined(SNAP_SUPD_T)
  printf("-poly_s: ");
  show_dlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
  printf("poly_t: ");
  show_dlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
  printf("poly_d: ");
  show_dlist(SNAP_coefs(poly_d), SNAPMTX_rows(matrix_conv));
  printf("matrix_conv:\n");
  show_sdmd(matrix_conv);
#elif defined(SNAP_SUPFC_T)
  printf("-poly_s: ");
  show_fclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
  printf("poly_t: ");
  show_fclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
  printf("poly_d: ");
  show_fclist(SNAP_coefs(poly_d), SNAPMTX_rows(matrix_conv));
  printf("matrix_conv:\n");
  show_sdmfc(matrix_conv);
#endif
#endif
#if defined(SNAP_SUPD_T)
  if ( clapack_dgels(CblasColMajor, CblasTrans, SNAPMTX_cols(matrix_conv), SNAPMTX_rows(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), SNAPMTX_rows(matrix_conv) ) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#elif defined(SNAP_SUPDC_T)
  if ( clapack_zgels(CblasColMajor, CblasConjTrans, SNAPMTX_cols(matrix_conv), SNAPMTX_rows(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), SNAPMTX_rows(matrix_conv) ) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#elif defined(SNAP_SUPFR_T)
//	if ( snap_mlapack_frgels(SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), 1) != 0 ) {
  if ( snap_mlapack_frgels('N', SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), 1) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgels('N', SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), 1) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#endif
#ifdef SNAPLIB_DEBUG
//printf("now returned from the corresponding _gels routine.\n");
#if defined(SNAP_SUPDC_T)
  printf("=>poly_d: ");
  show_dclist(SNAP_coefs(poly_d), SNAPMTX_cols(matrix_conv));
#elif defined(SNAP_SUPD_T)
  printf("=>poly_d: ");
  show_dlist(SNAP_coefs(poly_d), SNAPMTX_cols(matrix_conv));
#elif defined(SNAP_SUPFC_T)
  printf("=>poly_d: ");
  show_fclist(SNAP_coefs(poly_d), SNAPMTX_cols(matrix_conv));
#endif
#endif
  SNAP_degree(poly_d) = deg_d;
  SNAPMTX_clear(matrix_conv);
  return 0;
}

/*!
 * @brief The Gauss-Newton iteration for getting better approximate GCD.  Stop condition = 3 times not decreasing.
 * @param[in] poly_inf The input polynomial f[x].
 * @param[in] poly_ing The input polynomial g[x].
 * @param[in,out] poly_d The approximate GCD found.
 * @param[in,out] poly_f1 The approximate cofactor found for poly_inf.
 * @param[in,out] poly_g1 The approximate cofactor found for poly_ing.
 * @param[out] rho The residual.
 * @return The number of iterations.
 */
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho)
#else
static int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t rho)
#endif
{
  unsigned int num_descreasing = 0;
  SNAPMTX_t matrix_Jh, matrix_C;
  SNAPBASE_t *fhvect, *hvect, *pt;
  SNAPBASEREAL_t delta_prev, delta, norm_d;
#ifdef SNAP_SUPDC_T
  SNAPBASE_t tmpC;
  SNAPBASEREAL_t tmpR;
#endif // SNAP_SUPDC_T
#if defined (SNAP_SUPFC_T) || defined (SNAP_SUPFR_T) || defined(SNAP_SUPDC_T)
  SNAPBASE_t one, zero;
#endif
  unsigned int i,j, iteration = 0, len_fh, len_Jhc, len_Jhr;
  // initialization
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASE_init_set_ui(one, 1);
  SNAPBASE_init_set_ui(zero, 0);
#endif
#ifdef SNAP_SUPDC_T
  SNAPBASE_init(tmpC);
  SNAPBASEREAL_init(tmpR);
#endif // SNAP_SUPDC_T
  SNAPBASEREAL_init(delta);
  SNAPBASEREAL_init_set_d(delta_prev, DBL_MAX);
  SNAPBASEREAL_init(norm_d);
  len_fh = SNAP_degree(poly_f)+SNAP_degree(poly_g)+3;
  len_Jhc = SNAP_degree(poly_f)+SNAP_degree(poly_g)-SNAP_degree(poly_d)+3;
  len_Jhr = SNAP_degree(poly_f)+SNAP_degree(poly_g)+3;
  SNAPMTX_init2(matrix_Jh, len_Jhr, len_Jhc);
  SNAPMTX_csize(matrix_C) = SNAPMTX_csize(matrix_Jh);
  SNAPMTX_rsize(matrix_C) = SNAPMTX_rsize(matrix_Jh);
  hvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(SNAP_degree(poly_d)+1));
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<=SNAP_degree(poly_d); i++ ) {
    SNAPBASE_init(*(hvect+i));
  }
#endif
  fhvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*len_fh);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<len_fh; i++ ) {
    SNAPBASE_init(*(fhvect+i));
  }
#endif
  // iteration
  while(1) {
#if defined(SNAPLIB_DEBUG) && defined(SNAP_SUPDC_T)
    printf("before: poly_f: %p, norm_d: %p, poly_d: %p, poly_f1: %p, poly_g1: %p\n", SNAP_coefs(poly_f), &norm_d, SNAP_coefs(poly_d), SNAP_coefs(poly_f1), SNAP_coefs(poly_g1));
#endif // SNAPLIB_DEBUG
    iteration++;
    // form the vector hvect which is the scaled coefficient vectors of poly_d.
    // beta is 1 as noted in the algorithm, and hvect be a scalor multiple of poly_d.
#ifdef SNAP_SUPFC_T
    SNAPMTX_BLAS_dotc(one, SNAP_degree(poly_d)+1, SNAP_coefs(poly_d), 1, SNAP_coefs(poly_d), 1);
    mpfr_set(norm_d, mpc_realref(one), SNAPLIB_MPFR_RND);
    mpc_set_ui(one, 1, SNAPLIB_MPC_RND);
#elif defined(SNAP_SUPDC_T)
    // calling ATLAS modifies the pointer of poly_f in some case. we do by ourself.
    SNAPBASE_set_d(norm_d, 0.0);
    for ( i = 0; i <= SNAP_degree(poly_d); i++ ) {
      SNAPBASE_conj(tmpC, SNAP_coef(poly_d, i));
      SNAPBASE_mul(tmpC, tmpC, SNAP_coef(poly_d, i));
      SNAPBASE_abs(tmpR, tmpC);
      SNAPBASE_add(norm_d, norm_d, tmpR);
    }
#else
    SNAPMTX_BLAS_dotc(norm_d, SNAP_degree(poly_d)+1, SNAP_coefs(poly_d), 1, SNAP_coefs(poly_d), 1);
#endif
#if defined(SNAPLIB_DEBUG) && defined(SNAP_SUPDC_T)
    printf(" after: poly_f: %p, norm_d: %p, poly_d: %p, poly_f1: %p, poly_g1: %p\n", SNAP_coefs(poly_f), &norm_d, SNAP_coefs(poly_d), SNAP_coefs(poly_f1), SNAP_coefs(poly_g1));
#endif // SNAPLIB_DEBUG
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
      SNAPBASE_div_basereal(*pt, SNAP_coef(poly_d,i), norm_d);
    }
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
    printf("hvect: \n");
    show_dlist(hvect,SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPDC_T)
    printf("hvect: \n");
    show_dclist(hvect,SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFR_T)
    printf("hvect: \n");
    show_frlist(hvect,SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFC_T)
    printf("hvect: \n");
    show_fclist(hvect,SNAP_degree(poly_d)+1);
#endif
#endif
    // constrcut the first row of J_h
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
      SNAPBASE_conj(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#else
      SNAPBASE_set(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#endif
    }
    for ( ; i<len_Jhc; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,0,i), 0);
    }
    // constrcut the first column block of J_h other than the first row
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_f1, SNAP_degree(poly_d));
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_g1, SNAP_degree(poly_d));
    // construct f_h (note that f_h is not depending on the prev f_h).
#ifdef SNAPLIB_DEBUG
    printf("now calling the corresponding _gemv routine.\n");
#endif
#if defined(SNAP_SUPD_T)
    cblas_dgemv(CblasRowMajor, CblasNoTrans, len_Jhr, SNAP_degree(poly_d)+1, 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, 0, fhvect, 1);
#elif defined(SNAP_SUPDC_T)
    one = 1.0;
    zero = 0.0;
    cblas_zgemv(CblasRowMajor, CblasNoTrans, len_Jhr, SNAP_degree(poly_d)+1, &one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, &zero, fhvect, 1);
#elif defined(SNAP_SUPFR_T)
    snap_mblas_frgemv( 'N', len_Jhr, SNAP_degree(poly_d)+1, one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, zero, fhvect, 1);
#elif defined(SNAP_SUPFC_T)
    snap_mblas_fcgemv( 'N', len_Jhr, SNAP_degree(poly_d)+1, one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, zero, fhvect, 1);
#endif
#ifdef SNAPLIB_DEBUG
    printf("now returned from the corresponding _gemv routine.\n");
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
    printf("fhvect: \n");
    show_dlist(fhvect,len_fh);
#elif defined(SNAP_SUPDC_T)
    printf("fhvect: \n");
    show_dclist(fhvect,len_fh);
#elif defined(SNAP_SUPFR_T)
    printf("fhvect: \n");
    show_frlist(fhvect,len_fh);
#elif defined(SNAP_SUPFC_T)
    printf("fhvect: \n");
    show_fclist(fhvect,len_fh);
#endif
#endif
    // recursion (one substract) and note that beta is 1.0.
    SNAPBASE_sub_d(*fhvect, *fhvect, 1.0);
    for ( i=0,pt=fhvect+1; i<=SNAP_degree(poly_f); i++,pt++) {
      SNAPBASE_sub(*pt, *pt, SNAP_coef(poly_f, i));
    }
    for ( i=0; i<=SNAP_degree(poly_g); i++,pt++) {
      SNAPBASE_sub(*pt, *pt, SNAP_coef(poly_g, i));
    }
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
    printf("fhvect: \n");
    show_dlist(fhvect,len_fh);
#elif defined(SNAP_SUPDC_T)
    printf("fhvect: \n");
    show_dclist(fhvect,len_fh);
#elif defined(SNAP_SUPFR_T)
    printf("fhvect: \n");
    show_frlist(fhvect,len_fh);
#elif defined(SNAP_SUPFC_T)
    printf("fhvect: \n");
    show_fclist(fhvect,len_fh);
#endif
#endif
    // compute residual
    SNAPMTX_BLAS_nrm2(delta, len_fh, fhvect, 1);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    printf("delta: %e\n", delta);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    mpfr_printf("delta: %Re\n", delta);
#endif
#endif
    if ( SNAPBASEREAL_cmp(delta_prev, delta) <= 0 ) {
      if ( num_descreasing >= 3-1 ) {
        SNAPBASEREAL_set(delta, delta_prev);
        break;
      } else {
        num_descreasing++;
        SNAPBASEREAL_set(delta_prev, delta);
      }
    } else {
      SNAPBASEREAL_set(delta_prev, delta);
    }
    // constrcut the rest column blocks of J_h
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, SNAP_degree(poly_d)+1);
#if defined(SNAP_SUPDC_T)
    SNAPMTX_convolution_matrix_zerofill_conj(matrix_C, poly_d, SNAP_degree(poly_f1));
#else
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_d, SNAP_degree(poly_f1));
#endif
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, SNAP_degree(poly_f)+2);
#if defined(SNAP_SUPDC_T)
    SNAPMTX_convolution_matrix_zerofill_conj(matrix_C, poly_d, SNAP_degree(poly_g1));
#else
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_d, SNAP_degree(poly_g1));
#endif
    // construct the zero blocks of J_h
    for ( i=1; i<SNAP_degree(poly_f)+2; i++ ) {
      for ( j=SNAP_degree(poly_f)+2; j<len_Jhc; j++ ) {
        SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,i,j), 0);
      }
    }
    for ( ; i<len_Jhr; i++ ) {
      for ( j=SNAP_degree(poly_d)+1; j<SNAP_degree(poly_f)+2; j++ ) {
        SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,i,j), 0);
      }
    }
    // take conjugates the first col blocks to call zgels with CblasConjTrans.
    //  == constrcut the first row of J_h
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
#if defined(SNAP_SUPFC_T)
      SNAPBASE_conj(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#else
      SNAPBASE_set(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#endif
    }
    //  == constrcut the first column block of J_h other than the first row
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, 0);
#if defined(SNAP_SUPDC_T)
    SNAPMTX_convolution_matrix_zerofill_conj(matrix_C, poly_f1, SNAP_degree(poly_d));
#else
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_f1, SNAP_degree(poly_d));
#endif
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, 0);
#if defined(SNAP_SUPDC_T)
    SNAPMTX_convolution_matrix_zerofill_conj(matrix_C, poly_g1, SNAP_degree(poly_d));
#else
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_g1, SNAP_degree(poly_d));
#endif
#ifdef SNAPLIB_DEBUG
//	printf("fh: "); show_dlist(fhvect, len_fh);
//	printf("Jh:\n"); show_sdmd(matrix_Jh);
#endif
#if defined(SNAP_SUPD_T)
    if ( clapack_dgels(CblasColMajor, CblasTrans, SNAPMTX_cols(matrix_Jh), SNAPMTX_rows(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, SNAPMTX_rows(matrix_Jh) ) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#elif defined(SNAP_SUPDC_T)
    if ( clapack_zgels(CblasColMajor, CblasConjTrans, SNAPMTX_cols(matrix_Jh), SNAPMTX_rows(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, SNAPMTX_rows(matrix_Jh) ) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#elif defined(SNAP_SUPFR_T)
//		if ( snap_mlapack_frgels(SNAPMTX_rows(matrix_Jh), SNAPMTX_cols(matrix_Jh), SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, 1) != 0 )
    if ( snap_mlapack_frgels('N', SNAPMTX_rows(matrix_Jh), SNAPMTX_cols(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#elif defined(SNAP_SUPFC_T)
    if ( snap_mlapack_fcgels('N', SNAPMTX_rows(matrix_Jh), SNAPMTX_cols(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, 1) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#endif
#ifdef SNAPLIB_DEBUG
//	printf("fh(sol): "); show_dlist(fhvect, len_fh);
#endif
    for ( i=0,pt=fhvect; i<=SNAP_degree(poly_d); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_d,i), SNAP_coef(poly_d,i), *pt);
    }
    for ( i=0; i<=SNAP_degree(poly_f1); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_f1,i), SNAP_coef(poly_f1,i), *pt);
    }
    for ( i=0; i<=SNAP_degree(poly_g1); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_g1,i), SNAP_coef(poly_g1,i), *pt);
    }
#ifdef SNAPLIB_DEBUG
    printf("iteration: %d\n", iteration);
#if defined(SNAP_SUPD_T)
    printf("poly_f1: ");
    show_dlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: ");
    show_dlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    printf("poly_d: ");
    show_dlist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPDC_T)
    printf("poly_f1: ");
    show_dclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: ");
    show_dclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    printf("poly_d: ");
    show_dclist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFR_T)
    printf("poly_f1: ");
    show_frlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: ");
    show_frlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    printf("poly_d: ");
    show_frlist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFC_T)
    printf("poly_f1: ");
    show_fclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
    printf("poly_g1: ");
    show_fclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
    printf("poly_d: ");
    show_fclist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#endif
#endif
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  SNAPBASEREAL_set(*rho, delta);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_set(rho, delta);
#endif
  SNAPMTX_clear(matrix_Jh);
#ifdef SNAP_SUPDC_T
  SNAPBASE_clear(tmpC);
  SNAPBASEREAL_clear(tmpR);
#endif // SNAP_SUPDC_T
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_clear(delta);
  SNAPBASEREAL_clear(delta_prev);
  SNAPBASEREAL_clear(norm_d);
  SNAPBASE_clear(one);
  SNAPBASE_clear(zero);
  for ( i=0; i<SNAP_degree(poly_d)+1; i++ ) {
    SNAPBASE_clear(*(hvect+i));
  }
  for ( i=0; i<len_fh; i++ ) {
    SNAPBASE_clear(*(fhvect+i));
  }
#endif
  free(fhvect);
  free(hvect);
  return iteration;
}

/*!
 *  @brief UVGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the UVGCD algorithm proposed by Zhonggang Zeng
 *   in The Numerical Greatest Common Divisor of Univariate Polynomials, Contemporary Mathematics,
 *   Volume 556, 2011. Note that we scaled the input polynomials at the beginning to make its
 *   behavior being the same as QRGCD and ExQRGCD. This preconditioning is not required by
 *   the original algorithm. Please take care of this fact for your theoretical analysis.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] tol The tolerance.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol) {
  SNAPBASE_t *pt, *tau, *zvect, *yvect;
  SNAPBASEREAL_t norm_f, norm_g, sigma_m1, tol_work, rho;
  SNAP_t poly_f, poly_g, poly_tmp;
  int result = 0;
  unsigned int i,r,init_r;
  SNAPMTX_t Sn;
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_inf)) || SNAPBASE_zeroq(SNAP_lcoef(poly_ing)) ) {
    fprintf(stderr, "uvgcd: polynomials must be non-zero!\n");
    return -1;
  }
  // initialization
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(sigma_m1);
  SNAPBASEREAL_init(tol_work);
  SNAPBASEREAL_init(rho);
  SNAP_init(poly_tmp);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAPMTX_init2(Sn, SNAP_degree(poly_inf)+SNAP_degree(poly_ing), SNAP_degree(poly_inf)+SNAP_degree(poly_ing));
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  r = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
  init_r = r;
  SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_f, poly_g, r);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("Sn (init)\n");
  show_sdmd(Sn);
#elif defined(SNAP_SUPDC_T)
  printf("Sn (init)\n");
  show_sdmdc(Sn);
#elif defined(SNAP_SUPFR_T)
  printf("Sn (init)\n");
  show_sdmfr(Sn);
#elif defined(SNAP_SUPFC_T)
  printf("Sn (init)\n");
  show_sdmfc(Sn);
#endif
#endif
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(zvect+i));
    SNAPBASE_init(*(yvect+i));
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: dgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: zgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
  }
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (QR): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we keep the conjugated tau for larfx.
  for ( i=0,pt=tau; i<SNAPMTX_cols(Sn); i++,pt++ ) {
    SNAPBASE_conj(*pt,*pt);
  }
#endif
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
  printf("Sn (QR)\n");
  show_sdmd(Sn);
#elif defined(SNAP_SUPDC_T)
  printf("Sn (QR)\n");
  show_sdmdc(Sn);
#elif defined(SNAP_SUPFR_T)
  printf("Sn (QR)\n");
  show_sdmfr(Sn);
#elif defined(SNAP_SUPFC_T)
  printf("Sn (QR)\n");
  show_sdmfc(Sn);
#endif
#endif
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  while ( result != -1 ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // NOTE THAT we don't use the permutation matrix Pj. Instead, we use another way to reflext it.
    // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, &sigma_m1, rstate) < 0 ) {
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      break;
    }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, sigma_m1, rstate) < 0 ) {
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      break;
    }
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    printf("elapsed_time (SV): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    SNAPBASEREAL_set_ui(tol_work, SNAPLIB_MAX(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-r);
    SNAPBASEREAL_sqrt(tol_work, tol_work);
    SNAPBASEREAL_mul(tol_work, tol_work, tol);
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    printf("sigma_m1 = %e, tol_work = %e\n", sigma_m1, tol_work);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    mpfr_printf("sigma_m1 = %Re, tol_work = %Re\n", sigma_m1, tol_work);
#endif
#endif
    if ( SNAPBASEREAL_cmp(sigma_m1, tol_work) < 0 ) { // aGCD may exist!!
#ifdef SNAPLIB_DEBUG
      printf("aGCD may exist!!\n");
#endif
      // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
      SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-init_r );
      SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-init_r, SNAP_degree(poly_inf)-init_r );
      // to deg(g)-init_r, deg(f)-init_r from deg(g)-r-1, deg(f)-r-1
      for ( i=r, pt=zvect+SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*init_r; i<init_r; i++ ) {
        SNAP_set_coef(poly_g1, SNAP_degree(poly_g)-(init_r+r-i), *pt++ );
        SNAP_set_coef(poly_f1, SNAP_degree(poly_f)-(init_r+r-i), *pt++ );
      }
      SNAP_self_neg(poly_f1);
      if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, r) == 0 ) {
#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
        printf("poly_f1: ");
        show_dlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
        printf("poly_g1: ");
        show_dlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
        printf("poly_d: ");
        show_dlist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPDC_T)
        printf("poly_f (%p): ", SNAP_coefs(poly_f));
        show_dclist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
        printf("poly_f1 (%p): ", SNAP_coefs(poly_f1));
        show_dclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
        printf("poly_g1 (%p): ", SNAP_coefs(poly_g1));
        show_dclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
        printf("poly_d (%p): ", SNAP_coefs(poly_d));
        show_dclist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFR_T)
        printf("poly_f1: ");
        show_frlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
        printf("poly_g1: ");
        show_frlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
        printf("poly_d: ");
        show_frlist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPFC_T)
        printf("poly_f1: ");
        show_fclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
        printf("poly_g1: ");
        show_fclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
        printf("poly_d: ");
        show_fclist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#endif
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
        elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
        // getting a better approximate GCD.
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        if ( SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, &rho) >= 0 )
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        if ( SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, rho) >= 0 )
#endif
        {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
          printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
          // (ORIGINAL VERSION) check the threshold
          // if ( SNAPBASEREAL_cmp(rho, tol) < 0 ) {
          //   break;
          // }
          // (OUR DEFINITION) check the threshold
          SNAP_mul(poly_tmp, poly_d, poly_f1);
          SNAP_sub(poly_tmp, poly_tmp, poly_f);
          SNAP_2norm(tol_work, poly_tmp);
          if ( SNAPBASEREAL_cmp(tol_work, tol) <= 0 ) {
            SNAP_mul(poly_tmp, poly_d, poly_g1);
            SNAP_sub(poly_tmp, poly_tmp, poly_g);
            SNAP_2norm(tol_work, poly_tmp);
            if ( SNAPBASEREAL_cmp(tol_work, tol) <= 0 ) {
              break;
            }
          }
        }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
        printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      }
    }
    // r=0 means that we could not find any aGCD.
    if ( r == 0 ) {
      result = -1;
      break;
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // otherwise update!
    if ( SNAP_agcd_uvgcd_update(Sn, poly_f, poly_g, zvect, yvect, tau, r) < 0 ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
      printf("elapsed_time (UPDATE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_update could not work correctly.\n");
      break;
    } else {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
      printf("elapsed_time (UPDATE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      r--;
    }
  }
// NOTE: f1 and g1 should be correctly made up!!
  if ( result == 0 ) { // approximate GCD found!!
    SNAP_2norm(tol_work, poly_d);
    SNAP_div_basereal(poly_d, poly_d, tol_work);
    SNAPBASEREAL_mul(norm_f, norm_f, tol_work);
    SNAP_mul_basereal(poly_f1, poly_f1, norm_f);
    SNAPBASEREAL_mul(norm_g, norm_g, tol_work);
    SNAP_mul_basereal(poly_g1, poly_g1, norm_g);
  } else { // couldn't find any approximate GCD!!
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d,0,1);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_clear(*(tau+i));
    SNAPBASE_clear(*(zvect+i));
    SNAPBASE_clear(*(yvect+i));
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
  SNAP_clear(poly_tmp);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(sigma_m1);
  SNAPBASEREAL_clear(tol_work);
  SNAPBASEREAL_clear(rho);
  SNAPMTX_clear(Sn);
  return result;
}

#if !defined(SNAP_SUPFR_T) && !defined(SNAP_SUPFC_T)
typedef enum { SNAP_UVGCD2_REFINEMENT_NONE, SNAP_UVGCD2_REFINEMENT_UVGCD, SNAP_UVGCD2_REFINEMENT_GPGCD } mode_type;
int SNAP_agcd_gpgcd_modified_newton2(SNAPMTXREAL_t input_f, SNAPMTXREAL_t input_g, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int max_iterations, const SNAPBASEREAL_t d_threshold, const SNAPBASEREAL_t tol);
void SNAP_agcd_gpgcd_construct_tildevectors(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g);
void SNAP_agcd_gpgcd_lsdiv(SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPMTXREAL_t tilde_f, SNAPMTXREAL_t tilde_g, SNAPMTXREAL_t cofact_a, SNAPMTXREAL_t cofact_b, const unsigned int m, const unsigned int n, const unsigned int d);
/*!
 *  @brief UVGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the UVGCD algorithm proposed by Zhonggang Zeng
 *   in The Numerical Greatest Common Divisor of Univariate Polynomials, Contemporary Mathematics,
 *   Volume 556, 2011. Note that we scaled the input polynomials at the beginning to make its
 *   behavior being the same as QRGCD and ExQRGCD. This preconditioning is not required by
 *   the original algorithm. Please take care of this fact for your theoretical analysis.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] tol The tolerance.
 *  @param[in] mode_opt UVGCD's refinement if 0, and GPGCD's refinement if 1.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcd2(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const SNAPBASEREAL_t tol, const int mode_opt_in) {
  SNAPBASE_t *pt, *tau, *zvect, *yvect;
  SNAPBASEREAL_t norm_f, norm_g, sigma_m1, tol_work, rho;
  SNAP_t poly_f, poly_g, poly_tmp, poly_tmp2;
  int iteration_result, result = 0;
  unsigned int i,r,init_r;
  SNAPMTX_t Sn;
  SNAPMTXREAL_t input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b;
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  mode_type mode_opt;
  if ( mode_opt_in == 0 ) {
    mode_opt = SNAP_UVGCD2_REFINEMENT_UVGCD;
  } else if ( mode_opt_in == 1) {
    mode_opt = SNAP_UVGCD2_REFINEMENT_GPGCD;
  } else {
    mode_opt = SNAP_UVGCD2_REFINEMENT_NONE;
  }
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_inf)) || SNAPBASE_zeroq(SNAP_lcoef(poly_ing)) ) {
    fprintf(stderr, "uvgcd: polynomials must be non-zero!\n");
    return -1;
  }
  // initialization
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init(sigma_m1);
  SNAPBASEREAL_init(tol_work);
  SNAPBASEREAL_init(rho);
  SNAP_init(poly_tmp);
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  SNAP_div_basereal(poly_f, poly_f, norm_f);
  SNAP_div_basereal(poly_g, poly_g, norm_g);
  SNAPMTX_init2(Sn, SNAP_degree(poly_inf)+SNAP_degree(poly_ing), SNAP_degree(poly_inf)+SNAP_degree(poly_ing));
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  r = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
  init_r = r;
  SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_f, poly_g, r);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(zvect+i));
    SNAPBASE_init(*(yvect+i));
  }
  if ( mode_opt == SNAP_UVGCD2_REFINEMENT_GPGCD ) {
    SNAP_init(poly_tmp2);
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    SNAPMTXREAL_init2(input_f, 1, 2*(SNAP_degree(poly_f)+1));
    SNAPMTXREAL_init2(input_g, 1, 2*(SNAP_degree(poly_g)+1));
    SNAPMTXREAL_init2(tilde_f, 1, 2*(SNAP_degree(poly_f)+1));
    SNAPMTXREAL_init2(tilde_g, 1, 2*(SNAP_degree(poly_g)+1));
#else
    SNAPMTXREAL_init2(input_f, 1, SNAP_degree(poly_f)+1);
    SNAPMTXREAL_init2(input_g, 1, SNAP_degree(poly_g)+1);
    SNAPMTXREAL_init2(tilde_f, 1, SNAP_degree(poly_f)+1);
    SNAPMTXREAL_init2(tilde_g, 1, SNAP_degree(poly_g)+1);
#endif // defined
    SNAPMTXREAL_init(cofact_a);
    SNAPMTXREAL_init(cofact_b);
  }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  double elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: dgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: zgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
  }
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
  printf("elapsed_time (QR): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we keep the conjugated tau for larfx.
  for ( i=0,pt=tau; i<SNAPMTX_cols(Sn); i++,pt++ ) {
    SNAPBASE_conj(*pt,*pt);
  }
#endif
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  while ( result != -1 ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // NOTE THAT we don't use the permutation matrix Pj. Instead, we use another way to reflext it.
    // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, &sigma_m1, rstate) < 0 ) {
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      break;
    }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
    if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, sigma_m1, rstate) < 0 ) {
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
      break;
    }
#endif
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    printf("elapsed_time (SV): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    SNAPBASEREAL_set_ui(tol_work, SNAPLIB_MAX(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-r);
    SNAPBASEREAL_sqrt(tol_work, tol_work);
    SNAPBASEREAL_mul(tol_work, tol_work, tol);
    if ( SNAPBASEREAL_cmp(sigma_m1, tol_work) < 0 ) { // aGCD may exist!!
      // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
      SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-init_r );
      SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-init_r, SNAP_degree(poly_inf)-init_r );
      // to deg(g)-init_r, deg(f)-init_r from deg(g)-r-1, deg(f)-r-1
      for ( i=r, pt=zvect+SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*init_r; i<init_r; i++ ) {
        SNAP_set_coef(poly_g1, SNAP_degree(poly_g)-(init_r+r-i), *pt++ );
        SNAP_set_coef(poly_f1, SNAP_degree(poly_f)-(init_r+r-i), *pt++ );
      }
      SNAP_self_neg(poly_f1);
      if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, r) == 0 ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
        elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
        // getting a better approximate GCD.
        if ( mode_opt == SNAP_UVGCD2_REFINEMENT_UVGCD ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
          iteration_result = SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, &rho);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
          iteration_result = SNAP_agcd_uvgcd_iteration_9_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, rho);
#endif
        } else if ( mode_opt == SNAP_UVGCD2_REFINEMENT_GPGCD ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
          SNAPMTXREAL_realloc(cofact_a, 1, 2*(SNAP_degree(poly_g1)+1));
          SNAPMTXREAL_realloc(cofact_b, 1, 2*(SNAP_degree(poly_f1)+1));
#else
          SNAPMTXREAL_realloc(cofact_a, 1, SNAP_degree(poly_g1)+1);
          SNAPMTXREAL_realloc(cofact_b, 1, SNAP_degree(poly_f1)+1);
#endif // defined
          SNAP_agcd_gpgcd_construct_tildevectors(poly_f, poly_g, input_f, input_g);
          SNAP_mul(poly_tmp, poly_d, poly_f1);
          SNAP_mul(poly_tmp2, poly_d, poly_g1);
          SNAP_agcd_gpgcd_construct_tildevectors(poly_tmp, poly_tmp2, tilde_f, tilde_g);
          SNAP_mul_si(poly_f1, poly_f1, -1);
          SNAP_agcd_gpgcd_construct_tildevectors(poly_f1, poly_g1, cofact_b, cofact_a);
          iteration_result = SNAP_agcd_gpgcd_modified_newton2(input_f, input_g, tilde_f, tilde_g, cofact_a, cofact_b, 8192, 1.0e-8, tol);
          SNAP_agcd_gpgcd_lsdiv(poly_d, poly_f1, poly_g1, tilde_f, tilde_g, cofact_a, cofact_b, SNAP_degree(poly_f), SNAP_degree(poly_g), SNAP_degree(poly_d));
        } else {
          iteration_result = 0;
        }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
        printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
        if ( iteration_result >= 0 ) {
          // (ORIGINAL VERSION) check the threshold
          // if ( SNAPBASEREAL_cmp(rho, tol) < 0 ) {
          //   break;
          // }
          // (OUR DEFINITION) check the threshold
          SNAP_mul(poly_tmp, poly_d, poly_f1);
          SNAP_sub(poly_tmp, poly_tmp, poly_f);
          SNAP_2norm(tol_work, poly_tmp);
          if ( SNAPBASEREAL_cmp(tol_work, tol) <= 0 ) {
            SNAP_mul(poly_tmp, poly_d, poly_g1);
            SNAP_sub(poly_tmp, poly_tmp, poly_g);
            SNAP_2norm(tol_work, poly_tmp);
            if ( SNAPBASEREAL_cmp(tol_work, tol) <= 0 ) {
              break;
            }
          }
        }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
        printf("elapsed_time (REFINE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      }
    }
    // r=0 means that we could not find any aGCD.
    if ( r == 0 ) {
      result = -1;
      break;
    }
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
    elapsed_time = snap_getrusageinsec();
#endif // SNAPLIB_SHOW_ELAPSED_TIME
    // otherwise update!
    if ( SNAP_agcd_uvgcd_update(Sn, poly_f, poly_g, zvect, yvect, tau, r) < 0 ) {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
      printf("elapsed_time (UPDATE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      result = -1;
      printf("ERROR in agcd_uvgcd: agcd_uvgcd_update could not work correctly.\n");
      break;
    } else {
#ifdef SNAPLIB_SHOW_ELAPSED_TIME
      printf("elapsed_time (UPDATE): %.18e\n", snap_getrusageinsec()-elapsed_time);
#endif // SNAPLIB_SHOW_ELAPSED_TIME
      r--;
    }
  }
// NOTE: f1 and g1 should be correctly made up!!
  if ( result == 0 ) { // approximate GCD found!!
    SNAP_2norm(tol_work, poly_d);
    SNAP_div_basereal(poly_d, poly_d, tol_work);
    SNAPBASEREAL_mul(norm_f, norm_f, tol_work);
    SNAP_mul_basereal(poly_f1, poly_f1, norm_f);
    SNAPBASEREAL_mul(norm_g, norm_g, tol_work);
    SNAP_mul_basereal(poly_g1, poly_g1, norm_g);
  } else { // couldn't find any approximate GCD!!
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d,0,1);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_clear(*(tau+i));
    SNAPBASE_clear(*(zvect+i));
    SNAPBASE_clear(*(yvect+i));
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
  SNAP_clear(poly_tmp);
  SNAP_clear(poly_f);
  SNAP_clear(poly_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(sigma_m1);
  SNAPBASEREAL_clear(tol_work);
  SNAPBASEREAL_clear(rho);
  SNAPMTX_clear(Sn);
  if ( mode_opt == SNAP_UVGCD2_REFINEMENT_GPGCD ) {
    SNAP_clear(poly_tmp2);
    SNAPMTXREAL_clear(input_f);
    SNAPMTXREAL_clear(input_g);
    SNAPMTXREAL_clear(tilde_f);
    SNAPMTXREAL_clear(tilde_g);
    SNAPMTXREAL_clear(cofact_a);
    SNAPMTXREAL_clear(cofact_b);
  }
  return result;
}
#endif

/*!
 *  @brief UVGCD algorithm for two univariate polynomials with degree specified and paper original iteration threshold.
 *  @details This is an implementation of the UVGCD algorithm proposed by Zhonggang Zeng
 *   in The Numerical Greatest Common Divisor of Univariate Polynomials, Contemporary Mathematics,
 *   Volume 556, 2011.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcd_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold) {
  SNAPBASE_t *pt, *tau, *zvect, *yvect;
  SNAPBASEREAL_t sigma_m1, rho;
  int result = 0, iterations;
  unsigned int i,r,init_r;
  SNAPMTX_t Sn;
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_inf)) || SNAPBASE_zeroq(SNAP_lcoef(poly_ing)) ) {
    fprintf(stderr, "uvgcd: polynomials must be non-zero!\n");
    return -1;
  }
  // initialization
  SNAPBASEREAL_init(sigma_m1);
  SNAPBASEREAL_init(rho);
  SNAPMTX_init2(Sn, SNAP_degree(poly_inf)+SNAP_degree(poly_ing), SNAP_degree(poly_inf)+SNAP_degree(poly_ing));
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  r = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
  init_r = r;
  SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_inf, poly_ing, r);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(zvect+i));
    SNAPBASE_init(*(yvect+i));
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: dgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: zgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
  }
#endif
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we keep the conjugated tau for larfx.
  for ( i=0,pt=tau; i<SNAPMTX_cols(Sn); i++,pt++ ) {
    SNAPBASE_conj(*pt,*pt);
  }
#endif
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  while ( result != -1 ) {
    if ( r+1 == d ) {
      // NOTE THAT we don't use the permutation matrix Pj. Instead, we use another way to reflext it.
      // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, &sigma_m1, rstate) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, sigma_m1, rstate) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#endif
      // we assume that aGCD may exist!!
      // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
      SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-init_r );
      SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-init_r, SNAP_degree(poly_inf)-init_r );
      // to deg(g)-init_r, deg(f)-init_r from deg(g)-r-1, deg(f)-r-1
      for ( i=r, pt=zvect+SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*init_r; i<init_r; i++ ) {
        SNAP_set_coef(poly_g1, SNAP_degree(poly_ing)-(init_r+r-i), *pt++ );
        SNAP_set_coef(poly_f1, SNAP_degree(poly_inf)-(init_r+r-i), *pt++ );
      }
      SNAP_self_neg(poly_f1);
      if ( SNAP_agcd_uvgcd_solving_10_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, r) == 0 ) {
        // getting a better approximate GCD.
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        iterations = SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        iterations = SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho);
#endif
        if ( iterations >= 0 ) {
          printf("INFO in agcd_uvgcd_iteration_9_2: converged within %d iterations.\n", iterations);
          // (ORIGINAL VERSION) check the threshold
          //if ( SNAPBASEREAL_cmp(rho, threshold) < 0 ) {
            break;
          //}
        } else {
          printf("INFO in agcd_uvgcd_iteration_9_2: didn't converge.\n");
        }
      }
      // r=0 means that we could not find any aGCD.
    } else if ( r == 0 ) {
      result = -1;
      break;
    } else {
      // otherwise update!
      if ( SNAP_agcd_uvgcd_update(Sn, poly_inf, poly_ing, zvect, yvect, tau, r) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_update could not work correctly.\n");
        break;
      } else {
        r--;
      }
    }
  }
// NOTE: f1 and g1 should be correctly made up!!
  if ( result != 0 ) { // couldn't find any approximate GCD!!
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d,0,1);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_clear(*(tau+i));
    SNAPBASE_clear(*(zvect+i));
    SNAPBASE_clear(*(yvect+i));
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
  SNAPBASEREAL_clear(sigma_m1);
  SNAPBASEREAL_clear(rho);
  SNAPMTX_clear(Sn);
  return result;
}

// =========== gesly version =============
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
/*!
 * @brief An initial approximate GCD by Least Square.
 * @warning poly_X must be all different.
 * @warning This uses gelsy instead of gels.
 * @param[in] poly_f The input polynomial f.
 * @param[in] poly_g The input polynomial g.
 * @param[in] poly_f1 An initial cofactor of f.
 * @param[in] poly_g1 An initial cofactor of g.
 * @param[out] poly_d An initial approximate GCD.
 * @param[in] current_r The current order j of the subresultant mapping.
 * @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcd_solving_10_2_lsy(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r) {
  SNAPMTX_t matrix_conv, matrix_dummy;
  unsigned int i,deg_d;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance;
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#endif // SNAP_SUPD_T
  // initialization
  deg_d = current_r+1;
  // initialize the matrix as the whole sized matrix.
  SNAPMTX_init2(matrix_conv, SNAP_degree(poly_f)+SNAP_degree(poly_g)+2, deg_d+1);
  // set up the upper half as C(f1)
  SNAPMTX_convolution_matrix(matrix_conv, poly_f1, deg_d);
  SNAPMTX_rows(matrix_conv) = SNAP_degree(poly_f)+SNAP_degree(poly_g)+2;
  // set up the lower half as C(g1)
  SNAPMTX_csize(matrix_dummy) = SNAPMTX_csize(matrix_conv);
  SNAPMTX_rsize(matrix_dummy) = SNAP_degree(poly_g)+1;
  SNAPMTX_cols(matrix_dummy) = SNAPMTX_cols(matrix_conv);
  SNAPMTX_rows(matrix_dummy) = SNAP_degree(poly_g)+1;
  SNAPMTX_elems(matrix_dummy) = SNAPMTX_elemp(matrix_conv, SNAP_degree(poly_f)+1, 0);
  SNAPMTX_convolution_matrix(matrix_dummy, poly_g1, deg_d);
  // *B should be the coefficient of poly_d but the clone of poly_f and poly_g.
  SNAP_set(poly_d, poly_f);
  for ( i=0; i<=SNAP_degree(poly_g); i++ ) {
    SNAP_set_coef(poly_d, SNAP_degree(poly_f)+1+i, SNAP_coef(poly_g,i));
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*SNAPMTX_cols(matrix_conv));
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matrix_conv); i++,jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
#endif // defined
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), 1, jpvt, relative_tolerance, &rank) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_d), 1, jpvt, relative_tolerance, &rank) != 0 ) {
    SNAPMTX_clear(matrix_conv);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_10_2), ": could not compute the least square solution.\n");
    return -1;
  }
#elif defined(SNAP_SUPFR_T)
  IMPLEMENT HERE!
#elif defined(SNAP_SUPFC_T)
  IMPLEMENT HERE!
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  free(jpvt);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
#endif // defined
  SNAP_degree(poly_d) = deg_d;
  SNAPMTX_clear(matrix_conv);
  return 0;
}

/*!
 * @brief The Gauss-Newton iteration for getting better approximate GCD.  Stop condition = 3 times not decreasing.
 * @warning This uses gelsy instead of gels.
 * @param[in] poly_inf The input polynomial f[x].
 * @param[in] poly_ing The input polynomial g[x].
 * @param[in,out] poly_d The approximate GCD found.
 * @param[in,out] poly_f1 The approximate cofactor found for poly_inf.
 * @param[in,out] poly_g1 The approximate cofactor found for poly_ing.
 * @param[out] rho The residual.
 * @return The number of iterations.
 */
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2_lsy(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho)
#else
static int SNAP_agcd_uvgcd_iteration_9_2_lsy(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t rho)
#endif
{
  unsigned int num_descreasing = 0;
  SNAPMTX_t matrix_Jh, matrix_C;
  SNAPBASE_t *fhvect, *hvect, *pt;
  SNAPBASEREAL_t delta_prev, delta, norm_d;
#ifdef SNAP_SUPDC_T
  SNAPBASE_t tmpC;
  SNAPBASEREAL_t tmpR;
#endif // SNAP_SUPDC_T
#if defined (SNAP_SUPFC_T) || defined (SNAP_SUPFR_T) || defined(SNAP_SUPDC_T)
  SNAPBASE_t one, zero;
#endif
  unsigned int i,j, iteration = 0, len_fh, len_Jhc, len_Jhr;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance;
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAP_2norm(norm_f, poly_f);
  SNAP_2norm(norm_g, poly_g);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#endif // SNAP_SUPD_T
  // initialization
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASE_init_set_ui(one, 1);
  SNAPBASE_init_set_ui(zero, 0);
#endif
#ifdef SNAP_SUPDC_T
  SNAPBASE_init(tmpC);
  SNAPBASEREAL_init(tmpR);
#endif // SNAP_SUPDC_T
  SNAPBASEREAL_init(delta);
  SNAPBASEREAL_init_set_d(delta_prev, DBL_MAX);
  SNAPBASEREAL_init(norm_d);
  len_fh = SNAP_degree(poly_f)+SNAP_degree(poly_g)+3;
  len_Jhc = SNAP_degree(poly_f)+SNAP_degree(poly_g)-SNAP_degree(poly_d)+3;
  len_Jhr = SNAP_degree(poly_f)+SNAP_degree(poly_g)+3;
  SNAPMTX_init2(matrix_Jh, len_Jhr, len_Jhc);
  SNAPMTX_csize(matrix_C) = SNAPMTX_csize(matrix_Jh);
  SNAPMTX_rsize(matrix_C) = SNAPMTX_rsize(matrix_Jh);
  hvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(SNAP_degree(poly_d)+1));
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<=SNAP_degree(poly_d); i++ ) {
    SNAPBASE_init(*(hvect+i));
  }
#endif
  fhvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*len_fh);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<len_fh; i++ ) {
    SNAPBASE_init(*(fhvect+i));
  }
#endif
  // iteration
  while(1) {
    iteration++;
    // form the vector hvect which is the scaled coefficient vectors of poly_d.
    // beta is 1 as noted in the algorithm, and hvect be a scalor multiple of poly_d.
#ifdef SNAP_SUPFC_T
    SNAPMTX_BLAS_dotc(one, SNAP_degree(poly_d)+1, SNAP_coefs(poly_d), 1, SNAP_coefs(poly_d), 1);
    mpfr_set(norm_d, mpc_realref(one), SNAPLIB_MPFR_RND);
    mpc_set_ui(one, 1, SNAPLIB_MPC_RND);
#elif defined(SNAP_SUPDC_T)
    // calling ATLAS modifies the pointer of poly_f in some case. we do by ourself.
    SNAPBASE_set_d(norm_d, 0.0);
    for ( i = 0; i <= SNAP_degree(poly_d); i++ ) {
      SNAPBASE_conj(tmpC, SNAP_coef(poly_d, i));
      SNAPBASE_mul(tmpC, tmpC, SNAP_coef(poly_d, i));
      SNAPBASE_abs(tmpR, tmpC);
      SNAPBASE_add(norm_d, norm_d, tmpR);
    }
#else
    SNAPMTX_BLAS_dotc(norm_d, SNAP_degree(poly_d)+1, SNAP_coefs(poly_d), 1, SNAP_coefs(poly_d), 1);
#endif
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
      SNAPBASE_div_basereal(*pt, SNAP_coef(poly_d,i), norm_d);
    }
    // constrcut the first row of J_h
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
      SNAPBASE_conj(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#else
      SNAPBASE_set(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#endif
    }
    for ( ; i<len_Jhc; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,0,i), 0);
    }
    // constrcut the first column block of J_h other than the first row
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_f1, SNAP_degree(poly_d));
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_g1, SNAP_degree(poly_d));
    // construct f_h (note that f_h is not depending on the prev f_h).
#if defined(SNAP_SUPD_T)
    cblas_dgemv(CblasRowMajor, CblasNoTrans, len_Jhr, SNAP_degree(poly_d)+1, 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, 0, fhvect, 1);
#elif defined(SNAP_SUPDC_T)
    one = 1.0;
    zero = 0.0;
    cblas_zgemv(CblasRowMajor, CblasNoTrans, len_Jhr, SNAP_degree(poly_d)+1, &one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, &zero, fhvect, 1);
#elif defined(SNAP_SUPFR_T)
    snap_mblas_frgemv( 'N', len_Jhr, SNAP_degree(poly_d)+1, one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, zero, fhvect, 1);
#elif defined(SNAP_SUPFC_T)
    snap_mblas_fcgemv( 'N', len_Jhr, SNAP_degree(poly_d)+1, one, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), SNAP_coefs(poly_d), 1, zero, fhvect, 1);
#endif
    // recursion (one substract) and note that beta is 1.0.
    SNAPBASE_sub_d(*fhvect, *fhvect, 1.0);
    for ( i=0,pt=fhvect+1; i<=SNAP_degree(poly_f); i++,pt++) {
      SNAPBASE_sub(*pt, *pt, SNAP_coef(poly_f, i));
    }
    for ( i=0; i<=SNAP_degree(poly_g); i++,pt++) {
      SNAPBASE_sub(*pt, *pt, SNAP_coef(poly_g, i));
    }
    // compute residual
    SNAPMTX_BLAS_nrm2(delta, len_fh, fhvect, 1);
    if ( SNAPBASEREAL_cmp(delta_prev, delta) <= 0 ) {
      if ( num_descreasing >= 3-1 ) {
        SNAPBASEREAL_set(delta, delta_prev);
        break;
      } else {
        num_descreasing++;
        SNAPBASEREAL_set(delta_prev, delta);
      }
    } else {
      SNAPBASEREAL_set(delta_prev, delta);
    }
    // constrcut the rest column blocks of J_h
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, SNAP_degree(poly_d)+1);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_d, SNAP_degree(poly_f1));
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, SNAP_degree(poly_f)+2);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_d, SNAP_degree(poly_g1));
    // construct the zero blocks of J_h
    for ( i=1; i<SNAP_degree(poly_f)+2; i++ ) {
      for ( j=SNAP_degree(poly_f)+2; j<len_Jhc; j++ ) {
        SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,i,j), 0);
      }
    }
    for ( ; i<len_Jhr; i++ ) {
      for ( j=SNAP_degree(poly_d)+1; j<SNAP_degree(poly_f)+2; j++ ) {
        SNAPBASE_set_ui(SNAPMTX_elem(matrix_Jh,i,j), 0);
      }
    }
    // take conjugates the first col blocks to call zgels with CblasConjTrans.
    //  == constrcut the first row of J_h
    for ( i=0,pt=hvect; i<=SNAP_degree(poly_d); i++,pt++ ) {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
      SNAPBASE_conj(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#else
      SNAPBASE_set(SNAPMTX_elem(matrix_Jh,0,i), *pt);
#endif
    }
    //  == constrcut the first column block of J_h other than the first row
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, 1, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_f1, SNAP_degree(poly_d));
    SNAPMTX_elems(matrix_C) = SNAPMTX_elemp(matrix_Jh, SNAP_degree(poly_f)+2, 0);
    SNAPMTX_convolution_matrix_zerofill(matrix_C, poly_g1, SNAP_degree(poly_d));
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*SNAPMTX_cols(matrix_Jh));
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_cols(matrix_Jh); i++,jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
#endif // defined
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_Jh), SNAPMTX_cols(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matrix_Jh), SNAPMTX_cols(matrix_Jh), 1, SNAPMTX_elems(matrix_Jh), SNAPMTX_csize(matrix_Jh), fhvect, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_uvgcd_solving_9_2), ": could not compute the least square solution.\n");
      iteration = -1;
      break;
    }
#elif defined(SNAP_SUPFR_T)
  IMPLEMENT HERE!
#elif defined(SNAP_SUPFC_T)
  IMPLEMENT HERE!
#endif
    for ( i=0,pt=fhvect; i<=SNAP_degree(poly_d); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_d,i), SNAP_coef(poly_d,i), *pt);
    }
    for ( i=0; i<=SNAP_degree(poly_f1); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_f1,i), SNAP_coef(poly_f1,i), *pt);
    }
    for ( i=0; i<=SNAP_degree(poly_g1); i++,pt++) {
      SNAPBASE_sub(SNAP_coef(poly_g1,i), SNAP_coef(poly_g1,i), *pt);
    }
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  SNAPBASEREAL_set(*rho, delta);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_set(rho, delta);
#endif
  SNAPMTX_clear(matrix_Jh);
#ifdef SNAP_SUPDC_T
  SNAPBASE_clear(tmpC);
  SNAPBASEREAL_clear(tmpR);
#endif // SNAP_SUPDC_T
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  SNAPBASEREAL_clear(delta);
  SNAPBASEREAL_clear(delta_prev);
  SNAPBASEREAL_clear(norm_d);
  SNAPBASE_clear(one);
  SNAPBASE_clear(zero);
  for ( i=0; i<SNAP_degree(poly_d)+1; i++ ) {
    SNAPBASE_clear(*(hvect+i));
  }
  for ( i=0; i<len_fh; i++ ) {
    SNAPBASE_clear(*(fhvect+i));
  }
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  free(jpvt);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
#endif // defined
  free(fhvect);
  free(hvect);
  return iteration;
}

/*!
 *  @brief UVGCD algorithm for two univariate polynomials with degree specified and paper original iteration threshold.
 *  @details This is an implementation of the UVGCD algorithm proposed by Zhonggang Zeng
 *   in The Numerical Greatest Common Divisor of Univariate Polynomials, Contemporary Mathematics,
 *   Volume 556, 2011.
 * @warning This uses gelsy instead of gels.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_uvgcdy_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold) {
  SNAPBASE_t *pt, *tau, *zvect, *yvect;
  SNAPBASEREAL_t sigma_m1, rho;
  int result = 0, iterations;
  unsigned int i,r,init_r;
  SNAPMTX_t Sn;
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // check non-zero
  if ( SNAPBASE_zeroq(SNAP_lcoef(poly_inf)) || SNAPBASE_zeroq(SNAP_lcoef(poly_ing)) ) {
    fprintf(stderr, "uvgcd: polynomials must be non-zero!\n");
    return -1;
  }
  // initialization
  SNAPBASEREAL_init(sigma_m1);
  SNAPBASEREAL_init(rho);
  SNAPMTX_init2(Sn, SNAP_degree(poly_inf)+SNAP_degree(poly_ing), SNAP_degree(poly_inf)+SNAP_degree(poly_ing));
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  r = SNAPLIB_MIN(SNAP_degree(poly_inf),SNAP_degree(poly_ing))-1;
  init_r = r;
  SNAPMTX_subresultant_mapping_matrix_asc(Sn, poly_inf, poly_ing, r);
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sn));
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(tau+i));
  }
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_init(*(zvect+i));
    SNAPBASE_init(*(yvect+i));
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: dgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0
#endif
  ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: zgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sn), SNAPMTX_cols(Sn), SNAPMTX_elems(Sn), SNAPMTX_csize(Sn), tau) != 0 ) {
    result = -1;
    printf("ERROR in agcd_uvgcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
  }
#endif
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
  // we keep the conjugated tau for larfx.
  for ( i=0,pt=tau; i<SNAPMTX_cols(Sn); i++,pt++ ) {
    SNAPBASE_conj(*pt,*pt);
  }
#endif
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  // we use "r" instead of "j" in the paper and note that "r=j-1".
  while ( result != -1 ) {
    if ( r+1 == d ) {
      // NOTE THAT we don't use the permutation matrix Pj. Instead, we use another way to reflext it.
      // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, &sigma_m1, rstate) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      if ( SNAP_agcd_uvgcd_iteration_7_3(Sn, zvect, yvect, sigma_m1, rstate) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
        break;
      }
#endif
      // we assume that aGCD may exist!!
      // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
      SNAP_set_base(poly_g1, zvect, SNAP_degree(poly_ing)-init_r );
      SNAP_set_base(poly_f1, zvect+SNAP_degree(poly_ing)-init_r, SNAP_degree(poly_inf)-init_r );
      // to deg(g)-init_r, deg(f)-init_r from deg(g)-r-1, deg(f)-r-1
      for ( i=r, pt=zvect+SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*init_r; i<init_r; i++ ) {
        SNAP_set_coef(poly_g1, SNAP_degree(poly_ing)-(init_r+r-i), *pt++ );
        SNAP_set_coef(poly_f1, SNAP_degree(poly_inf)-(init_r+r-i), *pt++ );
      }
      SNAP_self_neg(poly_f1);
      if ( SNAP_agcd_uvgcd_solving_10_2_lsy(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, r) == 0 ) {
        // getting a better approximate GCD.
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        iterations = SNAP_agcd_uvgcd_iteration_9_2_lsy(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho);
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        IMPLEMENT HERE!
#endif
        if ( iterations >= 0 ) {
          printf("INFO in agcd_uvgcd_iteration_9_2: converged within %d iterations.\n", iterations);
          // (ORIGINAL VERSION) check the threshold
          //if ( SNAPBASEREAL_cmp(rho, threshold) < 0 ) {
            break;
          //}
        } else {
          printf("INFO in agcd_uvgcd_iteration_9_2: didn't converge.\n");
        }
      }
      // r=0 means that we could not find any aGCD.
    } else if ( r == 0 ) {
      result = -1;
      break;
    } else {
      // otherwise update!
      if ( SNAP_agcd_uvgcd_update(Sn, poly_inf, poly_ing, zvect, yvect, tau, r) < 0 ) {
        result = -1;
        printf("ERROR in agcd_uvgcd: agcd_uvgcd_update could not work correctly.\n");
        break;
      } else {
        r--;
      }
    }
  }
// NOTE: f1 and g1 should be correctly made up!!
  if ( result != 0 ) { // couldn't find any approximate GCD!!
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d,0,1);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0; i<SNAPMTX_csize(Sn); i++ ) {
    SNAPBASE_clear(*(tau+i));
    SNAPBASE_clear(*(zvect+i));
    SNAPBASE_clear(*(yvect+i));
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
  SNAPBASEREAL_clear(sigma_m1);
  SNAPBASEREAL_clear(rho);
  SNAPMTX_clear(Sn);
  return result;
}
#endif // defined






























/* ================== END[agcd_over_reals] ================== */
