/*
 * Copyright (c) 2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <snap.h>
#include <math.h>
#include "snap_type_supfc_conv.h"

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

#define SNAPLIB_DEBUG
// #undef SNAPLIB_DEBUG

/*!
 * @brief The recursive call for FFT.
 * @warning The size of the vectors must be the power of 2.
 */
void snap_fcdft_internal(const unsigned int len, mpc_t *target, const unsigned int incrT, mpc_t *list_work, const unsigned int incrW, mpc_t *list_unity, const unsigned int incrU) {
  unsigned int i, len_half;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3, *pt4;
  if ( len == 1 ) {
    return;
  }
  len_half = len/2;
  for( i=0,pt1=list_work,pt2=target,pt3=target+len_half*incrT,pt4=list_unity; i<len_half; i++,pt1+=incrW,pt2+=incrT,pt3+=incrT,pt4+=incrU ) {
    SNAPBASECOMPLEX_add(*pt1, *pt2, *pt3);
    SNAPBASECOMPLEX_sub(*pt3, *pt2, *pt3);
    SNAPBASECOMPLEX_mul(*pt3, *pt3, *pt4);
  }
  // recursive calls
  snap_fcdft_internal(len_half, list_work, incrW, list_work+len_half, incrW, list_unity, incrU*2);
  snap_fcdft_internal(len_half, target+len_half*incrT, incrT, list_work+len_half, incrW, list_unity, incrU*2);
  // re-order
  for( i=0,pt1=target,pt2=list_work,pt3=target+incrT,pt4=target+len_half*incrT; i<len_half; i++,pt1+=2*incrT,pt2+=incrW,pt3+=2*incrT,pt4+=incrT ) {
    SNAPBASECOMPLEX_set(*pt1, *pt2);
    SNAPBASECOMPLEX_set(*pt3, *pt4);
  }
}

/*!
 * @brief The recursive call for FFT.
 * @warning The size of the vectors must be the power of 2.
 */
void snap_fcdft_forward(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS) {
  unsigned int i, j;
  SNAPBASE_prec_t dprec;
  SNAPBASECOMPLEX_t const_pin, const_tmp, *list_unity, *list_work, *pt1, *pt2;
#ifdef SNAPLIB_DEBUG
  if ( n != pow(2, (unsigned int) log2(n)) ) {
    printf("ERROR: fcdft requires that the size of vector is the power of 2.\n");
    return;
  }
#endif // SNAPLIB_DEBUG
  // initialization for FFT
  list_unity = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  list_work = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  dprec = SNAPBASE_get_default_prec();
  SNAPBASECOMPLEX_init2(const_pin, dprec);
  SNAPBASECOMPLEX_init2(const_tmp, dprec);
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), n);
  for(i=0,pt1=list_work; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_unity; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
    SNAPBASECOMPLEX_mul_si(const_tmp, const_pin, -i);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
  for ( i=0; i<num; i++ ) {
    if ( target != source ) {
      for ( j=0, pt1 = target+i, pt2 = source+i; j<n; j++,pt1+=incrT, pt2+=incrS ) {
        SNAPBASECOMPLEX_set(*pt1, *pt2);
      }
    }
    snap_fcdft_internal(n, target+i, incrT, list_work, 1, list_unity, 1);
  }
}

/*!
 * @brief The recursive call for FFT.
 * @warning The size of the vectors must be the power of 2.
 */
void snap_fcdft_forward2(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS) {
  unsigned int i, j;
  SNAPBASE_prec_t dprec;
  SNAPBASECOMPLEX_t const_pin, const_tmp, *list_unity, *list_work, *pt1, *pt2;
#ifdef SNAPLIB_DEBUG
  if ( n != pow(2, (unsigned int) log2(n)) ) {
    printf("ERROR: fcdft requires that the size of vector is the power of 2.\n");
    return;
  }
#endif // SNAPLIB_DEBUG
  // initialization for FFT
  list_unity = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  list_work = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  dprec = SNAPBASE_get_default_prec();
  SNAPBASECOMPLEX_init2(const_pin, dprec);
  SNAPBASECOMPLEX_init2(const_tmp, dprec);
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), n);
  for(i=0,pt1=list_work; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_unity; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
    SNAPBASECOMPLEX_mul_si(const_tmp, const_pin, -i);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
  for ( i=0; i<num; i++ ) {
    if ( target != source ) {
      for ( j=0, pt1 = target+i*incrT, pt2 = source+i*incrS; j<n; j++,pt1++, pt2++ ) {
        SNAPBASECOMPLEX_set(*pt1, *pt2);
      }
    }
    snap_fcdft_internal(n, target+i*incrT, 1, list_work, 1, list_unity, 1);
  }
}

/*!
 * @brief The recursive call for FFT.
 * @warning The size of the vectors must be the power of 2.
 */
void snap_fcdft_backward(const unsigned int n, const unsigned int num, mpc_t *target, const unsigned int incrT, mpc_t *source, const unsigned int incrS) {
  unsigned int i, j;
  SNAPBASE_prec_t dprec;
  SNAPBASECOMPLEX_t const_pin, const_tmp, *list_unity, *list_work, *pt1, *pt2;
#ifdef SNAPLIB_DEBUG
  if ( n != pow(2, (unsigned int) log2(n)) ) {
    printf("ERROR: fcdft requires that the size of vector is the power of 2.\n");
    return;
  }
#endif // SNAPLIB_DEBUG
  // initialization for FFT
  list_unity = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  list_work = (SNAPBASECOMPLEX_t*) malloc(sizeof(SNAPBASECOMPLEX_t)*n);
  dprec = SNAPBASE_get_default_prec();
  SNAPBASECOMPLEX_init2(const_pin, dprec);
  SNAPBASECOMPLEX_init2(const_tmp, dprec);
  SNAPBASEREAL_set_ui(SNAPBASECOMPLEX_realref(const_pin), 0);
  SNAPBASEREAL_const_pi(SNAPBASECOMPLEX_imagref(const_pin));
  SNAPBASEREAL_mul_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), 2);
  SNAPBASEREAL_div_si(SNAPBASECOMPLEX_imagref(const_pin), SNAPBASECOMPLEX_imagref(const_pin), n);
  for(i=0,pt1=list_work; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
  }
  for(i=0,pt1=list_unity; i<n; i++,pt1++) {
    SNAPBASECOMPLEX_init2(*pt1, dprec);
    SNAPBASECOMPLEX_mul_si(const_tmp, const_pin, i);
    SNAPBASECOMPLEX_exp(*pt1, const_tmp);
  }
  for ( i=0; i<num; i++ ) {
    if ( target != source ) {
      for ( j=0, pt1 = target+i, pt2 = source+i; j<n; j++,pt1+=incrT, pt2+=incrS ) {
        SNAPBASECOMPLEX_set(*pt1, *pt2);
      }
    }
    snap_fcdft_internal(n, target+i, incrT, list_work, 1, list_unity, 1);
  }
}

