/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <snap.h>

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
void show_mtx_d(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  for(i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      printf("%e ", mpz_get_d(*(mtx+i*cols+j)));
    }
    printf("\n");
  }
}
void show_mtx_d2(mpq_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  for(i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      printf("%e ", mpq_get_d(*(mtx+i*cols+j)));
    }
    printf("\n");
  }
}
void show_mtx_m(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  printf("{");
  for(i=0; i<rows; i++) {
    printf("{");
    for(j=0; j<cols-1; j++) {
      printf("%s,", mpz_get_str(NULL,10,*(mtx+i*cols+j)));
    }
    if ( i < rows -1 ) {
      printf("%s},\n", mpz_get_str(NULL,10,*(mtx+i*cols+j)));
    } else {
      printf("%s}", mpz_get_str(NULL,10,*(mtx+i*cols+j)));
    }
  }
  printf("}\n");
}
void show_mtx_m2(mpq_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  printf("{");
  for(i=0; i<rows; i++) {
    printf("{");
    for(j=0; j<cols-1; j++) {
      printf("%s,", mpq_get_str(NULL,10,*(mtx+i*cols+j)));
    }
    if ( i < rows -1 ) {
      printf("%s},\n", mpq_get_str(NULL,10,*(mtx+i*cols+j)));
    } else {
      printf("%s}", mpq_get_str(NULL,10,*(mtx+i*cols+j)));
    }
  }
  printf("}\n");
}
#endif

void lll_exact_mode(mpz_t *mtx, unsigned int rows, unsigned int cols, unsigned int lll_mode) {
#ifdef SNAPLIB_DEBUG
  printf("INPUT\n");
  show_mtx_m(mtx,rows,cols);
#endif
  switch(lll_mode) {
  case 0: /* lll_mode=0 : lll_exact_original */
    lll_exact_original(mtx, rows, cols);
    break;
  case 1: /* lll_mode=1 : lll_exact_deepinsertion */
    lll_exact_deepinsertion(mtx, rows, cols);
    break;
  case 2: /* lll_mode=2 : lll_exact_keepintegral */
    lll_exact_keepintegral(mtx, rows, cols);
    break;
  case 3: /* lll_mode=3 : storjohann, 1996, reductionl3 */
    lll_exact_reductionl3(mtx, rows, cols);
    break;
  case 4: /* lll_mode=4 : storjohann, 1996, modreductionl3 */
    lll_exact_modreductionl3(mtx, rows, cols);
    break;
  case 5: /* lll_mode=5 : storjohann, 1996, reductionl3 with heuristics */
    lll_exact_reductionl3_heuristics(mtx, rows, cols);
    break;
  case 6: /* lll_mode=6 : storjohann, 1996, modreductionl3 with heuristics */
    lll_exact_modreductionl3_heuristics(mtx, rows, cols);
    break;
  default: /* default : lll_exact_keepintegral */
    lll_exact_keepintegral(mtx, rows, cols);
  }
#ifdef SNAPLIB_DEBUG
  printf("OUTPUT\n");
  show_mtx_m(mtx,rows,cols);
#endif
}

/* === lll_exact_original === */

#define INDEX_OF_MTX(i,j) ((i)*cols+(j))
#define INDEX_OF_MU(i,j) ((i)*((i)+1)/2+(j))

void lll_exact_original_gramschmidt_z(mpq_t *gsm, mpq_t *mu, mpz_t *mtx, unsigned int rows, unsigned int cols, unsigned int from, unsigned int to) {
  unsigned int r, i, s;
  mpq_t tmp;
  mpq_init(tmp);
  for(r=from; r<=to&&r<rows; r++) {
    for(i=0; i<cols; i++) {
      mpq_set_z(*(gsm+INDEX_OF_MTX(r,i)),*(mtx+INDEX_OF_MTX(r,i)));
    }
    for(s=0; s<r; s++) {
      mpq_inner_z(*(mu+INDEX_OF_MU(r,s)),gsm+INDEX_OF_MTX(s,0),mtx+INDEX_OF_MTX(r,0),cols);
      mpq_div(*(mu+INDEX_OF_MU(r,s)), *(mu+INDEX_OF_MU(r,s)), *(mu+INDEX_OF_MU(s,s)));
      for(i=0; i<cols; i++) {
        mpq_mul(tmp, *(mu+INDEX_OF_MU(r,s)), *(gsm+INDEX_OF_MTX(s,i)));
        mpq_sub(*(gsm+INDEX_OF_MTX(r,i)), *(gsm+INDEX_OF_MTX(r,i)), tmp);
      }
    }
    mpq_inner(*(mu+INDEX_OF_MU(r,r)), gsm+INDEX_OF_MTX(r,0), gsm+INDEX_OF_MTX(r,0), cols);
  }
  mpq_clear(tmp);
}

void lll_exact_original(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i, j, elements, loop = 1;
  mpq_t *gsm, *mu;
  mpz_t rmu;
  mpq_t qtmp, delta;
  // Matrix Size Check
  elements = rows*cols;
  i = rows*(1+rows);
  if ( rows != elements/cols || rows == UINT_MAX || rows+1 != i/rows ) {
    printf("ERROR in lll_exact_original: Matrix Size Overflow.\n");
    return;
  }
  gsm = (mpq_t*) malloc(elements*sizeof(mpq_t));
  for(i=0; i<elements; i++) mpq_init(*(gsm+i));
  mu = (mpq_t*) malloc(sizeof(mpq_t)*((rows*(1+rows))/2));
  for(i=0; i<(rows*(1+rows))/2; i++) mpq_init(*(mu+i));
  mpz_init(rmu);
  mpq_init(qtmp);
  mpq_init(delta);
  mpq_set_ui(delta, 3, 4);
  // STEP 1.
  lll_exact_original_gramschmidt_z(gsm, mu, mtx, rows, cols, 0, 0);
  while(loop<rows) {
    // STEP 2 and 3.
    lll_exact_original_gramschmidt_z(gsm, mu, mtx, rows, cols, loop, loop);
    // STEP 4 and 5.
    i=loop-1;
    while(1) {
      mpq_get_z(rmu, *(mu+INDEX_OF_MU(loop,i)));
      if(mpz_sgn(rmu) != 0) {
        for(j=0; j<cols; j++) {
          mpz_submul(*(mtx+INDEX_OF_MTX(loop,j)), *(mtx+INDEX_OF_MTX(i,j)), rmu);
        }
        mpq_sub_z(*(mu+INDEX_OF_MU(loop,i)), *(mu+INDEX_OF_MU(loop,i)), rmu);
      }
      if(i==0) {
        break;
      } else {
        i--;
      }
    }
    // STEP 6 and 7.
#ifdef SNAPLIB_DEBUG
    printf("k=%d\n", loop);
    show_mtx_m(mtx,rows,cols);
    printf("gsm:\n");
    show_mtx_m2(gsm,rows,cols);
    printf("mu:\n");
    show_mtx_m2(mu,1,(rows*(1+rows))/2);
#endif
    mpq_mul(qtmp, *(mu+INDEX_OF_MU(loop,loop-1)), *(mu+INDEX_OF_MU(loop,loop-1)));
    mpq_sub(qtmp, delta, qtmp);
    mpq_mul(qtmp, qtmp, *(mu+INDEX_OF_MU(loop-1,loop-1)));
    mpq_sub(qtmp, *(mu+INDEX_OF_MU(loop,loop)), qtmp);
#ifdef SNAPLIB_DEBUG
    printf("k: %d, qtmp: %s\n", loop, mpq_get_str(NULL,10,qtmp));
#endif
    if(mpq_sgn(qtmp)>=0) {
      loop++;
    } else {
      for(i=0; i<cols; i++) {
        mpz_swap(*(mtx+INDEX_OF_MTX(loop-1,i)),*(mtx+INDEX_OF_MTX(loop,i)));
      }
      loop--;
      if(loop==0) {
        lll_exact_original_gramschmidt_z(gsm, mu, mtx, rows, cols, 0, 0);
        loop++;
      }
    }
  }
  for(i=0; i<elements; i++) mpq_clear(*(gsm+i));
  free(gsm);
  for(i=0; i<(rows*(1+rows))/2; i++) mpq_clear(*(mu+i));
  free(mu);
  mpz_clear(rmu);
  mpq_clear(qtmp);
  mpq_clear(delta);
}

/* === lll_exact_deepinsertion === */

#undef INDEX_OF_MTX
#undef INDEX_OF_MU
#define INDEX_OF_MTX(i,j) ((i)*cols+(j))
#define INDEX_OF_MU(i,j) ((i)*((i)+1)/2+(j))

void lll_exact_deepinsertion_red(mpz_t *mtx, unsigned int cols, mpq_t *mu, unsigned int k, unsigned int l) {
  unsigned int i;
  mpz_t q;
  mpq_t qtmp;
  if ( mpq_cmp_si(*(mu+INDEX_OF_MU(k,l)), 1,2) <= 0 &&
       mpq_cmp_si(*(mu+INDEX_OF_MU(k,l)),-1,2) >= 0 ) {
    return;
  }
  mpz_init(q);
  mpq_init(qtmp);
  // q is the nearest integer of mu_{k,l}
  mpq_get_z(q, *(mu+INDEX_OF_MU(k,l)));
  for ( i=0; i<cols; i++ ) {
    // b_k <- b_k - q b_l
    mpz_submul(*(mtx+INDEX_OF_MTX(k,i)), q, *(mtx+INDEX_OF_MTX(l,i)));
  }
  // mu_{k,l} <- mu_{k,l} - q
  mpq_sub_z(*(mu+INDEX_OF_MU(k,l)), *(mu+INDEX_OF_MU(k,l)), q);
  for ( i=0; i<l; i++ ) {
    // mu_{k,i} <- mu_{k,i} - q mu_{l,i}
    mpq_mul_z(qtmp, *(mu+INDEX_OF_MU(l,i)), q);
    mpq_sub(*(mu+INDEX_OF_MU(k,i)), *(mu+INDEX_OF_MU(k,i)), qtmp);
  }
  mpz_clear(q);
  mpq_clear(qtmp);
}

void lll_exact_deepinsertion_insert(mpz_t *mtx, unsigned int cols, unsigned int k, unsigned int idx) {
  unsigned int i,j;
  mpz_t *mtxtmp;
  mtxtmp = (mpz_t*) malloc(cols*sizeof(mpz_t));
  for(i=0; i<cols; i++) mpz_init_set(*(mtxtmp+i), *(mtx+INDEX_OF_MTX(k,i)));
  for ( j=k; j>idx; j-- ) {
    for(i=0; i<cols; i++) mpz_set(*(mtx+INDEX_OF_MTX(j,i)), *(mtx+INDEX_OF_MTX(j-1,i)));
  }
  for(i=0; i<cols; i++) mpz_set(*(mtx+INDEX_OF_MTX(idx,i)), *(mtxtmp+i));
  for(i=0; i<cols; i++) mpz_clear(*(mtxtmp+i));
  free(mtxtmp);
}

void lll_exact_deepinsertion(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i, j, k, elements;
  mpq_t *gsm, *mu;
  mpq_t delta, qtmp, btmp;
  mpz_t ztmp;
  // Matrix Size Check
  elements = rows*cols;
  i = rows*(1+rows);
  if ( rows != elements/cols || rows == UINT_MAX || rows+1 != i/rows ) {
    printf("ERROR in lll_exact_deepinsertion: Matrix Size Overflow.\n");
    return;
  }
  // Allocate all the GMP objects.
  gsm = (mpq_t*) malloc(elements*sizeof(mpq_t));
  for(i=0; i<elements; i++) mpq_init(*(gsm+i));
  mu = (mpq_t*) malloc(sizeof(mpq_t)*((rows*(1+rows))/2));
  for(i=0; i<(rows*(1+rows))/2; i++) mpq_init(*(mu+i));
  mpz_init(ztmp);
  mpq_init(qtmp);
  mpq_init(btmp);
  mpq_init(delta);
  mpq_set_ui(delta, 3, 4);
  // STEP 1.
  k = 0;
  while ( k < rows ) {
    // STEP 2.
    for ( i=0; i<cols; i++ ) {
      mpq_set_z(*(gsm+INDEX_OF_MTX(k,i)),*(mtx+INDEX_OF_MTX(k,i)));
    }
    for ( j=0; j<k; j++ ) {
      mpq_inner_z(qtmp, gsm+INDEX_OF_MTX(j,0), mtx+INDEX_OF_MTX(k,0), cols);
      mpq_div(*(mu+INDEX_OF_MU(k,j)), qtmp, *(mu+INDEX_OF_MU(j,j)));
      for ( i=0; i<cols; i++ ) {
        mpq_mul(qtmp, *(mu+INDEX_OF_MU(k,j)), *(gsm+INDEX_OF_MTX(j,i)));
        mpq_sub(*(gsm+INDEX_OF_MTX(k,i)), *(gsm+INDEX_OF_MTX(k,i)), qtmp);
      }
    }
    mpq_inner(*(mu+INDEX_OF_MU(k,k)), gsm+INDEX_OF_MTX(k,0), gsm+INDEX_OF_MTX(k,0), cols);
    if ( k == 0 ) {
      k++;
    } else {
      // STEP 3.
      //for ( j=k-1; j>=0; j-- ) {
      j = k - 1;
      do {
        lll_exact_deepinsertion_red(mtx, cols, mu, k, j);
      } while ( j-- != 0 );
      mpz_inner(ztmp, mtx+INDEX_OF_MTX(k,0), mtx+INDEX_OF_MTX(k,0), cols);
      mpq_set_z(btmp, ztmp);
      i = 0;
      // STEP 4.
      while (1) {
        if ( i == k ) {
          k++;
          break;
        } else {
          mpq_mul(qtmp, delta, *(mu+INDEX_OF_MU(i,i)));
          if ( mpq_cmp(qtmp, btmp) <= 0 ) {
            mpq_mul(qtmp, *(mu+INDEX_OF_MU(k,i)), *(mu+INDEX_OF_MU(k,i)));
            mpq_mul(qtmp, qtmp, *(mu+INDEX_OF_MU(i,i)));
            mpq_sub(btmp, btmp, qtmp);
            i++;
            continue;
          } else {
            // INSERT(k,i)
            lll_exact_deepinsertion_insert(mtx, cols, k, i);
            if ( i > 0 ) {
              k=i-1;
              mpz_inner(ztmp, mtx+INDEX_OF_MTX(k,0), mtx+INDEX_OF_MTX(k,0), cols);
              mpq_set_z(btmp, ztmp);
              i = 0;
              continue;
            }
            if ( i == 0 ) {
              k = 0;
              break;
            }
          }
        }
      }
    }
  } // End of while <=> STEP 5.
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpq_clear(*(gsm+i));
  free(gsm);
  for(i=0; i<(rows*(1+rows))/2; i++) mpq_clear(*(mu+i));
  free(mu);
  mpz_clear(ztmp);
  mpq_clear(qtmp);
  mpq_clear(btmp);
  mpq_clear(delta);
}

/* === lll_exact_keepintegral === */

#undef INDEX_OF_MTX
#undef INDEX_OF_MU
#define INDEX_OF_MTX(i,j) ((i)*cols+(j))
#define INDEX_OF_MU(i,j) ((i)*((i)+1)/2+(j))
#define INDEX_OF_LAMBDA(i,j) ((i)*((i)-1)/2+(j))

void lll_exact_keepintegral_swapi(mpz_t *mtx, unsigned int cols, mpz_t *lambda, mpz_t *d, unsigned int k, unsigned int kmax) {
  unsigned int i;
  mpz_t ltmp, btmp, ttmp;
  mpz_init(ltmp);
  mpz_init(btmp);
  mpz_init(ttmp);
  // exchange vectors.
  for ( i=0; i<cols; i++ ) {
    mpz_swap(*(mtx+INDEX_OF_MTX(k,i)), *(mtx+INDEX_OF_MTX(k-1,i)));
  }
  if ( k > 1 ) {
    for ( i=0; i<k-1; i++ ) {
      mpz_swap(*(lambda+INDEX_OF_LAMBDA(k,i)), *(lambda+INDEX_OF_LAMBDA(k-1,i)));
    }
  }
  // set lambda and B.
  mpz_set(ltmp, *(lambda+INDEX_OF_LAMBDA(k,k-1)));
  mpz_mul(btmp, *(d+k-1), *(d+k+1));
  mpz_addmul(btmp, ltmp, ltmp);
  mpz_divexact(btmp, btmp, *(d+k));
  for ( i=k+1; i<=kmax; i++ ) { // This is OK since kmax < UINT_MAX
    mpz_set(ttmp, *(lambda+INDEX_OF_LAMBDA(i,k)));
    mpz_mul(*(lambda+INDEX_OF_LAMBDA(i,k)), *(d+k+1), *(lambda+INDEX_OF_LAMBDA(i,k-1)));
    mpz_submul(*(lambda+INDEX_OF_LAMBDA(i,k)), ltmp, ttmp);
    mpz_divexact(*(lambda+INDEX_OF_LAMBDA(i,k)), *(lambda+INDEX_OF_LAMBDA(i,k)), *(d+k));
    mpz_mul(*(lambda+INDEX_OF_LAMBDA(i,k-1)), btmp, ttmp);
    mpz_addmul(*(lambda+INDEX_OF_LAMBDA(i,k-1)), ltmp, *(lambda+INDEX_OF_LAMBDA(i,k)));
    mpz_divexact(*(lambda+INDEX_OF_LAMBDA(i,k-1)), *(lambda+INDEX_OF_LAMBDA(i,k-1)), *(d+k+1));
  }
  mpz_set(*(d+k), btmp);
  mpz_clear(ltmp);
  mpz_clear(btmp);
  mpz_clear(ttmp);
}

void lll_exact_keepintegral_redi(mpz_t *mtx, unsigned int cols, mpz_t *lambda, mpz_t *d, unsigned int k, unsigned int l) {
  unsigned int i;
  mpz_t ztmp;
  mpz_init(ztmp);
  mpz_mul_ui(ztmp, *(lambda+INDEX_OF_LAMBDA(k,l)), 2);
  mpz_abs(ztmp, ztmp);
  if ( mpz_cmp(ztmp, *(d+l+1)) <= 0 ) {
    return;
  }
  // ztmp = q
  mpz_rdiv_q(ztmp, *(lambda+INDEX_OF_LAMBDA(k,l)), *(d+l+1));
  for ( i=0; i<cols; i++ ) {
    mpz_submul(*(mtx+INDEX_OF_MTX(k,i)), ztmp, *(mtx+INDEX_OF_MTX(l,i)));
  }
  mpz_submul(*(lambda+INDEX_OF_LAMBDA(k,l)), ztmp, *(d+l+1));
  for ( i=0; i<l; i++ ) {
    mpz_submul(*(lambda+INDEX_OF_LAMBDA(k,i)), ztmp, *(lambda+INDEX_OF_LAMBDA(l,i)));
  }
  mpz_clear(ztmp);
}

void lll_exact_keepintegral(mpz_t *mtx, unsigned int rows, unsigned int cols) {
  unsigned int i, j, k, kmax, elements;
  static const unsigned int numerator = 3;
  static const unsigned int denominator = 4;
  mpz_t *lambda, *d, u, ztmp, lhs, rhs;
  // Matrix Size Check
  elements = rows*cols;
  if ( rows != elements/cols || rows >= UINT_MAX -1 ) {
    printf("ERROR in lll_exact_keepintegral: Matrix Size Overflow.\n");
    return;
  }
  elements = rows*(rows-1);
  if ( elements/rows != rows-1 ) {
    printf("ERROR in lll_exact_keepintegral: Matrix Size Overflow.\n");
    return;
  } else {
    elements /= 2;
  }
  // Allocate all the GMP objects.
  lambda = (mpz_t*) malloc(elements*sizeof(mpz_t));
  for(i=0; i<elements; i++) mpz_init(*(lambda+i));
  d = (mpz_t*) malloc(sizeof(mpz_t)*(rows+1));
  for(i=0; i<=rows; i++) mpz_init(*(d+i));
  mpz_init(u);
  mpz_init(ztmp);
  mpz_init(lhs);
  mpz_init(rhs);
  // STEP 1.
  k = 1;
  kmax = 0;
  mpz_set_ui(*d,1);
  mpz_inner(*(d+1), mtx, mtx, cols);
  while (  k < rows ) {
    // STEP 2.
    if ( k > kmax ) {
      kmax = k;
      for ( j=0; j<=k; j++ ) { // This is OK since k<=rows<UINT_MAX.
        mpz_inner(u, mtx+INDEX_OF_MTX(k,0), mtx+INDEX_OF_MTX(j,0), cols);
        for ( i=0; i<j; i++ ) {
          // u <- (d_i u - lamda_{k,i} lambda_{j,i})/d_{i-1}
          mpz_mul(u, u, *(d+i+1));
          mpz_mul(ztmp, *(lambda+INDEX_OF_LAMBDA(k,i)), *(lambda+INDEX_OF_LAMBDA(j,i)));
          mpz_sub(u, u, ztmp);
          mpz_divexact(u, u, *(d+i));
        }
        if ( j<k ) { // lambda_{k,j} <- u
          mpz_set(*(lambda+INDEX_OF_LAMBDA(k,j)), u);
        } else { // d_k <- u
          mpz_set(*(d+k+1), u);
        }
      }
    }
    // STEP 3.
    while (1) {
      // call REDI(k,k-1)
      lll_exact_keepintegral_redi(mtx, cols, lambda, d, k, k-1);
      // compute den*d_k*d_{k-2} < num*d_{k-1}^2-den*lambda_{k,k-1}^2
      mpz_mul(lhs, *(d+k+1), *(d+k-1));
      mpz_mul_ui(lhs, lhs, denominator);
      mpz_mul(rhs, *(d+k), *(d+k));
      mpz_mul_ui(rhs, rhs, numerator);
      mpz_mul(ztmp, *(lambda+INDEX_OF_LAMBDA(k,k-1)), *(lambda+INDEX_OF_LAMBDA(k,k-1)));
      mpz_mul_ui(ztmp, ztmp, denominator);
      mpz_sub(rhs, rhs ,ztmp);
      if ( mpz_cmp(lhs, rhs) < 0 ) {
        // call SWAPI(k)
        lll_exact_keepintegral_swapi(mtx, cols, lambda, d, k, kmax);
        if ( k > 1 ) {
          k--;
        }
        continue;
      } else {
        if ( k > 1 ) {
          i = k - 2;
          do {
            lll_exact_keepintegral_redi(mtx, cols, lambda, d, k, i);
          } while ( i-- != 0 );
        }
        k++;
        break;
      }
    }
  } // End of While <=> STEP 4.
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpz_clear(*(lambda+i));
  free(lambda);
  for(i=0; i<=rows; i++) mpz_clear(*(d+i));
  free(d);
  mpz_clear(u);
  mpz_clear(ztmp);
  mpz_clear(lhs);
  mpz_clear(rhs);
}

/* === lll_exact_modreductionl3 === */

#undef INDEX_OF_MTX
#undef INDEX_OF_MU
#define INDEX_OF_MTX(i,j) ((i)*cols+(j))
#define INDEX_OF_MU(i,j) ((i)*((i)+1)/2+(j))
#define INDEX_OF_MTXA(i,j) ((i)*cols+(j))
#define INDEX_OF_MTXT(i,j) ((i)*rows+(j))

void lll_exact_modreductionl3_ffgewop(mpz_t *mtx, unsigned int cols) {
  // Fraction-Free Gaussian Elimination without Row Pivoting.
  // See Algorithm 9.1, Algorithms for Computer Algebra, Kluwer Academic Publishers, 1992.
  // We don't check a_{kk}!=0. Hence if so then zero-division exception.
  unsigned int i,j,k;
  mpz_t divisor;
  mpz_init_set_ui(divisor, 1);
  for ( k=0; k<cols; k++ ) {
    // piviot is always a_{kk}.
    for ( i=k+1; i<cols; i++ ) {
      for ( j=k+1; j<cols; j++ ) {
        // a_ij <- (a_kk a_ij - a_kj a_ik) / divisor
        mpz_mul(*(mtx+INDEX_OF_MTX(i,j)), *(mtx+INDEX_OF_MTX(i,j)), *(mtx+INDEX_OF_MTX(k,k)));
        mpz_submul(*(mtx+INDEX_OF_MTX(i,j)), *(mtx+INDEX_OF_MTX(k,j)), *(mtx+INDEX_OF_MTX(i,k)));
        mpz_divexact(*(mtx+INDEX_OF_MTX(i,j)), *(mtx+INDEX_OF_MTX(i,j)), divisor);
      }
      // a_ik <- 0
      mpz_set_ui(*(mtx+INDEX_OF_MTX(i,k)), 0);
    }
    // divisor <- a_kk
    mpz_set(divisor, *(mtx+INDEX_OF_MTX(k,k)));
  }
  mpz_clear(divisor);
}

void lll_exact_modreductionl3_subtractrow(mpz_t *mtxA, unsigned int rows, unsigned int cols, mpz_t *mtxT, unsigned int k, unsigned int r, mpz_t q) {
  unsigned int i;
  for ( i=0; i<cols; i++ ) {
    mpz_submul(*(mtxA+INDEX_OF_MTXA(k,i)), *(mtxA+INDEX_OF_MTXA(r,i)), q);
  }
  for ( i=0; i<=r; i++ ) {
    mpz_submul(*(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,r)), q);
  }
}

void lll_exact_modreductionl3_modsubtractrow(mpz_t *mtxA, unsigned int rows, unsigned int cols, mpz_t *mtxT, mpz_t valM, unsigned int k, unsigned int r, mpz_t q) {
  unsigned int i;
  mpz_t ztmp;
  mpz_init(ztmp);
  lll_exact_modreductionl3_subtractrow(mtxA, rows, cols, mtxT, k, r, q);
  // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
  mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(0,0)), valM);
  mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(0,k)), *(mtxT+INDEX_OF_MTXT(0,k)), ztmp);
  for ( i=1; i<k; i++ ) {
    // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
    mpz_mul(ztmp, ztmp, valM);
    mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,k)), ztmp);
  }
  for ( i=0; i<cols; i++ ) {
    mpz_tdiv_r(*(mtxA+INDEX_OF_MTXA(k,i)), *(mtxA+INDEX_OF_MTXA(k,i)), valM);
  }
  mpz_clear(ztmp);
}

void lll_exact_modreductionl3_switchrow(mpz_t *mtxA, unsigned int rows, unsigned int cols, mpz_t *mtxT, unsigned int k) {
  unsigned int i;
  // switch rows k-1 and k in A;
  for ( i=0; i<cols; i++ ) {
    mpz_swap(*(mtxA+INDEX_OF_MTXA(k,i)), *(mtxA+INDEX_OF_MTXA(k-1,i)));
  }
  // update row(T,k)
  for ( i=k-1; i<rows; i++ ) {
    if ( k > 1 ) {
      mpz_mul(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
    }
    mpz_addmul(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,i)));
    mpz_divexact(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
  }
  // switch rows k-1 and k in T;
  for ( i=k-1; i<rows; i++ ) {
    mpz_swap(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-1,i)));
  }
  // switch columns k-1 and k in T;
  for ( i=0; i<=k; i++ ) { // This is OK since k < UINT_MAX.
    mpz_swap(*(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,k-1)));
  }
  // update row(T,k)
  for ( i=k-1; i<rows; i++ ) {
    mpz_mul(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    mpz_submul(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,i)));
    if ( k > 1 ) {
      mpz_divexact(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
    }
  }
}

void lll_exact_modreductionl3_modswitchrow(mpz_t *mtxA, unsigned int rows, unsigned int cols, mpz_t *mtxT, mpz_t valM, unsigned int k) {
  unsigned int i;
  mpz_t ztmp;
  mpz_init(ztmp);
  lll_exact_modreductionl3_switchrow(mtxA, rows, cols, mtxT, k);
  // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
  mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(0,0)), valM);
  mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(0,k-1)), *(mtxT+INDEX_OF_MTXT(0,k-1)), ztmp);
  for ( i=1; i<k-1; i++ ) {
    // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
    mpz_mul(ztmp, ztmp, valM);
    mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(i,k-1)), *(mtxT+INDEX_OF_MTXT(i,k-1)), ztmp);
  }
  // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
  mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(0,0)), valM);
  mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(0,k)), *(mtxT+INDEX_OF_MTXT(0,k)), ztmp);
  for ( i=1; i<k; i++ ) {
    // d_i d_{i-1} M, d_i = *(mtxT+INDEX_OF_MTXT(i,i))
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
    mpz_mul(ztmp, ztmp, valM);
    mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,k)), ztmp);
  }
  // for rows.
  if ( k >1 ) {
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
    mpz_mul(ztmp, ztmp, valM);
  } else {
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), valM);
  }
  for ( i=k; i<rows; i++ ) {
    mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(k-1,i)), *(mtxT+INDEX_OF_MTXT(k-1,i)), ztmp);
  }
  mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
  mpz_mul(ztmp, ztmp, valM);
  for ( i=k+1; i<rows; i++ ) {
    mpz_tdiv_r(*(mtxT+INDEX_OF_MTXT(k,i)), *(mtxT+INDEX_OF_MTXT(k,i)), ztmp);
  }
  mpz_clear(ztmp);
}

void lll_exact_modreductionl3(mpz_t *mtxA, unsigned int rows, unsigned int cols) {
  unsigned int i, k, elements;
  mpz_t *mtxT, valM, ztmp, ztmp2;
  // Matrix Size Check
  elements = rows*rows;
  if ( rows != elements/rows || rows >= UINT_MAX -1 ) {
    printf("ERROR in lll_exact_modreductionl3: Matrix Size Overflow.\n");
    return;
  }
  // Allocate all the GMP objects.
  mtxT = (mpz_t*) malloc(elements*sizeof(mpz_t));
  for(i=0; i<elements; i++) mpz_init(*(mtxT+i));
  mpz_init(valM);
  mpz_init(ztmp);
  mpz_init(ztmp2);
  // STEP 1.
  // Compute T.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 1...\n");
#endif
  mpz_inner_matrix_transpose(mtxT, mtxA, mtxA, rows, cols, rows);
  lll_exact_modreductionl3_ffgewop(mtxT, rows);
  // STEP 2.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 2...\n");
#endif
  // compute e.
  mpz_set(valM, *(mtxT+INDEX_OF_MTXT(0,0)));
  for ( i=1; i<rows; i++ ) {
    mpz_divexact(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
    if ( mpz_cmp(valM, ztmp) < 0 ) {
      mpz_set(valM, ztmp);
    }
  }
  mpz_mul_ui(valM, valM, rows);
  mpz_sqrtrem(valM, ztmp, valM);
  if ( mpz_sgn(ztmp) > 0 ) {
    mpz_add_ui(valM, valM, 1);
  }
  mpz_mul_ui(valM, valM, 2);
  mpz_add_ui(valM, valM, 1);
  // k <- 2
  k = 1;
  // do
  while (1) {
    // ModSubtractRow(A,T,M,k,k-1,[mu_{k,k-1}])
    mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    lll_exact_modreductionl3_modsubtractrow(mtxA, rows, cols, mtxT, valM, k, k-1, ztmp);
    // if B_k < (3/4-mu_{k,k-1}^2)B_{k-1}
    // if 4B_k < (3-4mu_{k,k-1}^2)B_{k-1}
    // if 4 d_k/d_{k-1} < (3-4\tilde{mu}_{k,k-1}^2/d_{k-1}^2) d_{k-1}/d_{k-2}
    // if 4 d_k d_{k-2} < (3-4\tilde{mu}_{k,k-1}^2/d_{k-1}^2) d_{k-1}^2
    // if 4 d_k d_{k-2} < (3d_{k-1}^2-4\tilde{mu}_{k,k-1}^2)
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k)));
    mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    mpz_mul_ui(ztmp2, ztmp2, 3);
    mpz_submul_ui(ztmp2, ztmp, 4);
    if ( k > 1 ) {
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      mpz_mul_ui(ztmp, ztmp, 4);
    } else {
      mpz_mul_ui(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), 4);
    }
#ifdef SNAPLIB_DEBUG
    printf("k=%d\n", k);
    show_mtx_m(mtxA,rows,cols);
    show_mtx_m(mtxT,rows,rows);
#endif
    if ( mpz_cmp(ztmp, ztmp2) < 0 ) {
      // ModSwitchRow(A,T,M,k)
      lll_exact_modreductionl3_modswitchrow(mtxA, rows, cols, mtxT, valM, k);
      if ( k>1 ) {
        k--;
      }
    } else {
      if ( k == rows - 1 ) {
        break;
      } else {
        k++;
      }
    }
  }
  // STEP 3.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 3...\n");
#endif
  for ( k=1; k<rows; k++ ) {
    // for ( i=k-1; i>=0; i-- ) {
    i = k - 1;
    do {
      // ModSubtractRow(A,T,M,k,j,[mu_{k,j}])
      mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,i)));
      lll_exact_modreductionl3_modsubtractrow(mtxA, rows, cols, mtxT, valM, k, i, ztmp);
    } while ( i-- != 0 );
  }
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpz_clear(*(mtxT+i));
  free(mtxT);
  mpz_clear(valM);
  mpz_clear(ztmp);
  mpz_clear(ztmp2);
}

void lll_exact_modreductionl3_heuristics(mpz_t *mtxA, unsigned int rows, unsigned int cols) {
  unsigned int i, k, e, elements;
  mpz_t *mtxT, valM, ztmp, ztmp2;
  // Matrix Size Check
  elements = rows*rows;
  if ( rows != elements/rows || rows >= UINT_MAX -1 ) {
    printf("ERROR in lll_exact_modreductionl3_heuristics: Matrix Size Overflow.\n");
    return;
  }
  // Allocate all the GMP objects.
  mtxT = (mpz_t*) malloc(elements*sizeof(mpz_t));
  for(i=0; i<elements; i++) mpz_init(*(mtxT+i));
  mpz_init(valM);
  mpz_init(ztmp);
  mpz_init(ztmp2);
  // STEP 1.
  // Compute T.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 1...\n");
#endif
  mpz_inner_matrix_transpose(mtxT, mtxA, mtxA, rows, cols, rows);
  lll_exact_modreductionl3_ffgewop(mtxT, rows);
  // STEP 2 with three heuristics.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 2...\n");
#endif
  // do
  while (1) {
    // compute e.
    mpz_set(valM, *(mtxT+INDEX_OF_MTXT(0,0)));
    for ( i=1; i<rows; i++ ) {
      mpz_divexact(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
      if ( mpz_cmp(valM, ztmp) < 0 ) {
        mpz_set(valM, ztmp);
      }
    }
    mpz_mul_ui(valM, valM, rows);
    i = mpz_sizeinbase(valM, 10);
    e = i/2;
    if ( 2*e != i ) {
      e = e + 2;
    } else {
      e++;
    }
    // compute M.
    mpz_ui_pow_ui(valM, 10, e);
    // k <- 2.
    k = 1;
    // for i from 3 to n do
    for ( i=2; i<rows; i++ ) {
      // B_i = d_i/d_{i-1}, B_k = d_k/d_{k-1}, d_{-1} = 1
      // d_i = *(mtxT+INDEX_OF_MTXT(i,i))
      // if B_i/B_{i-1} < B_k/B_{k-1}
      // if (d_i/d_{i-1})/(d_{i-1}/d_{i-2}) < (d_k/d_{k-1})/(d_{k-1}/d_{k-2})
      // if d_i d_{i-2}/d_{i-1}^2 < d_k d_{k-2}/d_{k-1}^2
      // if d_i d_{i-2} d_{k-1}^2 < d_k d_{k-2} d_{i-1}^2
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
      mpz_mul(ztmp, ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
      mpz_mul(ztmp, ztmp, *(mtxT+INDEX_OF_MTXT(i-2,i-2)));
      mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
      mpz_mul(ztmp2, ztmp2, *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
      if ( k>1 ) {
        mpz_mul(ztmp2, ztmp2, *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      }
      if ( mpz_cmp(ztmp, ztmp2) < 0 ) {
        k = i;
      }
    }
    // if 2*B_k >= B_{k-1} then break fi
    // if 2*d_k/d_{k-1} >= d_{k-1}/d_{k-2}
    // if 2*d_k*d_{k-2} >= d_{k-1}^2
    if (k>1) {
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      mpz_mul_ui(ztmp, ztmp, 2);
    } else {
      mpz_mul_ui(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), 2);
    }
    mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
#ifdef SNAPLIB_DEBUG
    printf("k=%d\n", k);
    show_mtx_m(mtxA,rows,cols);
    show_mtx_m(mtxT,rows,rows);
#endif
    if ( mpz_cmp(ztmp, ztmp2) >= 0 ) {
      break;
    }
    // ModSubtractRow(A,T,M,k,k-1,[mu_{k,k-1}])
    mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    lll_exact_modreductionl3_modsubtractrow(mtxA, rows, cols, mtxT, valM, k, k-1, ztmp);
    // ModSwitchRow(A,T,M,k)
    lll_exact_modreductionl3_modswitchrow(mtxA, rows, cols, mtxT, valM, k);
  }
  // STEP 3.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 3...\n");
#endif
  for ( k=1; k<rows; k++ ) {
    // for ( i=k-1; i>=0; i-- ) {
    i = k - 1;
    do {
      // ModSubtractRow(A,T,M,k,j,[mu_{k,j}])
      mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,i)));
      lll_exact_modreductionl3_modsubtractrow(mtxA, rows, cols, mtxT, valM, k, i, ztmp);
    } while ( i-- != 0 );
  }
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpz_clear(*(mtxT+i));
  free(mtxT);
  mpz_clear(valM);
  mpz_clear(ztmp);
  mpz_clear(ztmp2);
}

void lll_exact_reductionl3(mpz_t *mtxA, unsigned int rows, unsigned int cols) {
  unsigned int i, k, elements;
  mpz_t *mtxT, ztmp, ztmp2;
  // Matrix Size Check
  elements = rows*rows;
  if ( rows != elements/rows || rows >= UINT_MAX -1 ) {
    printf("ERROR in lll_exact_reductionl3: Matrix Size Overflow.\n");
    return;
  }
  // Allocate all the GMP objects.
  mtxT = (mpz_t*) malloc(elements*sizeof(mpz_t));
  for(i=0; i<elements; i++) mpz_init(*(mtxT+i));
  mpz_init(ztmp);
  mpz_init(ztmp2);
  // STEP 1.
  // Compute T.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 1...\n");
#endif
  mpz_inner_matrix_transpose(mtxT, mtxA, mtxA, rows, cols, rows);
  lll_exact_modreductionl3_ffgewop(mtxT, rows);
#ifdef SNAPLIB_DEBUG
  printf("mtxT\n");
  show_mtx_m(mtxT,rows,rows);
#endif
  // STEP 2 with three heuristics.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 2...\n");
#endif
  // do
  k = 1;
  while (1) {
#ifdef SNAPLIB_DEBUG
    printf("k=%d\n", k);
#endif
    // SubtractRow(A,T,k,k-1,[mu_{k,k-1}]),
    // mu_{k,k-1} = \tilde{mu}_{k,k-1}/d_{k-1}.
    mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
#ifdef SNAPLIB_DEBUG
    printf("Top: \\tilde{mu}_{k,k-1}=%s\n     d_{k-1}=%s\n     q=%s\n", mpz_get_str(NULL,10,*(mtxT+INDEX_OF_MTXT(k-1,k))), mpz_get_str(NULL,10,*(mtxT+INDEX_OF_MTXT(k-1,k-1))), mpz_get_str(NULL,10,ztmp));
#endif
    lll_exact_modreductionl3_subtractrow(mtxA, rows, cols, mtxT, k, k-1, ztmp);
#ifdef SNAPLIB_DEBUG
    printf("Substract row:%d by row:%d\n",k,k-1);
    show_mtx_m(mtxA,rows,cols);
    printf("mtxT\n");
    show_mtx_m(mtxT,rows,rows);
#endif
    // if B_k < (3/4-mu_{k,k-1}^2)B_{k-1}
    // if 4B_k < (3-4mu_{k,k-1}^2)B_{k-1}
    // if 4 d_k/d_{k-1} < (3-4\tilde{mu}_{k,k-1}^2/d_{k-1}^2) d_{k-1}/d_{k-2}
    // if 4 d_k d_{k-2} < (3-4\tilde{mu}_{k,k-1}^2/d_{k-1}^2) d_{k-1}^2
    // if 4 d_k d_{k-2} < (3d_{k-1}^2-4\tilde{mu}_{k,k-1}^2)
    mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k)));
    mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    mpz_mul_ui(ztmp2, ztmp2, 3);
    mpz_submul_ui(ztmp2, ztmp, 4);
    if ( k > 1 ) {
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      mpz_mul_ui(ztmp, ztmp, 4);
    } else {
      mpz_mul_ui(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), 4);
    }
#ifdef SNAPLIB_DEBUG
    printf("k: %d, lhs: %s, rhs: %s\n", k, mpz_get_str(NULL,10,ztmp), mpz_get_str(NULL,10,ztmp2));
#endif
    if ( mpz_cmp(ztmp, ztmp2) < 0 ) { // Case 1
      // SwitchRow(A,T,k)
      lll_exact_modreductionl3_switchrow(mtxA, rows, cols, mtxT, k);
#ifdef SNAPLIB_DEBUG
      printf("Switch\n");
      show_mtx_m(mtxA,rows,cols);
#endif
      if ( k > 1 ) {
        k--;
      }
    } else { // Case 2
      // for ( i=k-2; i>=0; i-- ) {
      if ( k > 1 ) {
        i = k - 2;
        do {
          // SubtractRow(A,T,k,i,[mu_{k,i}])
          // mu_{k,i} = \tilde{mu}_{k,i}/d_{i}.
          mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,i)));
#ifdef SNAPLIB_DEBUG
          printf("Last: \\tilde{mu}_{k,i}=%s\n     d_{i}=%s\n     q=%s\n", mpz_get_str(NULL,10,*(mtxT+INDEX_OF_MTXT(i,k))), mpz_get_str(NULL,10,*(mtxT+INDEX_OF_MTXT(i,i))), mpz_get_str(NULL,10,ztmp));
#endif
          lll_exact_modreductionl3_subtractrow(mtxA, rows, cols, mtxT, k, i, ztmp);
        } while ( i-- != 0 );
      }
      if ( k == rows - 1 ) {
        break;
      } else {
        k++;
      }
    }
#ifdef SNAPLIB_DEBUG
    show_mtx_m(mtxA,rows,cols);
#endif
  }
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpz_clear(*(mtxT+i));
  free(mtxT);
  mpz_clear(ztmp);
  mpz_clear(ztmp2);
}

void lll_exact_reductionl3_heuristics(mpz_t *mtxA, unsigned int rows, unsigned int cols) {
  unsigned int i, k, elements;
  mpz_t *mtxT, ztmp, ztmp2;
  // Matrix Size Check
  elements = rows*rows;
  if ( rows != elements/rows || rows >= UINT_MAX -1 ) {
    printf("ERROR in lll_exact_reductionl3_heuristics: Matrix Size Overflow.\n");
    return;
  }
  // Allocate all the GMP objects.
  mtxT = (mpz_t*) malloc(elements*sizeof(mpz_t));
  for(i=0; i<elements; i++) mpz_init(*(mtxT+i));
  mpz_init(ztmp);
  mpz_init(ztmp2);
  // STEP 1.
  // Compute T.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 1...\n");
#endif
  mpz_inner_matrix_transpose(mtxT, mtxA, mtxA, rows, cols, rows);
  lll_exact_modreductionl3_ffgewop(mtxT, rows);
#ifdef SNAPLIB_DEBUG
  printf("mtxT\n");
  show_mtx_m(mtxT,rows,rows);
#endif
  // STEP 2 with three heuristics.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 2...\n");
#endif
  // do
  while (1) {
    // k <- 2.
    k = 1;
    // for i from 3 to n do
    for ( i=2; i<rows; i++ ) {
      // B_i = d_i/d_{i-1}, B_k = d_k/d_{k-1}, d_{-1} = 1
      // d_i = *(mtxT+INDEX_OF_MTXT(i,i))
      // if B_i/B_{i-1} < B_k/B_{k-1}
      // if (d_i/d_{i-1})/(d_{i-1}/d_{i-2}) < (d_k/d_{k-1})/(d_{k-1}/d_{k-2})
      // if d_i d_{i-2}/d_{i-1}^2 < d_k d_{k-2}/d_{k-1}^2
      // if d_i d_{i-2} d_{k-1}^2 < d_k d_{k-2} d_{i-1}^2
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(i,i)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
      mpz_mul(ztmp, ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
      mpz_mul(ztmp, ztmp, *(mtxT+INDEX_OF_MTXT(i-2,i-2)));
      mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
      mpz_mul(ztmp2, ztmp2, *(mtxT+INDEX_OF_MTXT(i-1,i-1)));
      if ( k>1 ) {
        mpz_mul(ztmp2, ztmp2, *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      }
      if ( mpz_cmp(ztmp, ztmp2) < 0 ) {
        k = i;
      }
    }
    // if 2*B_k >= B_{k-1} then break fi
    // if 2*d_k/d_{k-1} >= d_{k-1}/d_{k-2}
    // if 2*d_k*d_{k-2} >= d_{k-1}^2
    if (k>1) {
      mpz_mul(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), *(mtxT+INDEX_OF_MTXT(k-2,k-2)));
      mpz_mul_ui(ztmp, ztmp, 2);
    } else {
      mpz_mul_ui(ztmp, *(mtxT+INDEX_OF_MTXT(k,k)), 2);
    }
    mpz_mul(ztmp2, *(mtxT+INDEX_OF_MTXT(k-1,k-1)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
#ifdef SNAPLIB_DEBUG
    printf("k=%d\n", k);
    show_mtx_m(mtxA,rows,cols);
    show_mtx_m(mtxT,rows,rows);
#endif
    if ( mpz_cmp(ztmp, ztmp2) >= 0 ) {
      break;
    }
    // ModSubtractRow(A,T,M,k,k-1,[mu_{k,k-1}])
    mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(k-1,k)), *(mtxT+INDEX_OF_MTXT(k-1,k-1)));
    lll_exact_modreductionl3_subtractrow(mtxA, rows, cols, mtxT, k, k-1, ztmp);
    // ModSwitchRow(A,T,M,k)
    lll_exact_modreductionl3_switchrow(mtxA, rows, cols, mtxT, k);
  }
  // STEP 3.
#ifdef SNAPLIB_DEBUG
  printf("Entering STEP 3...\n");
#endif
  for ( k=1; k<rows; k++ ) {
    // for ( i=k-1; i>=0; i-- ) {
    i = k - 1;
    do {
      // ModSubtractRow(A,T,M,k,j,[mu_{k,j}])
      mpz_rdiv_q(ztmp, *(mtxT+INDEX_OF_MTXT(i,k)), *(mtxT+INDEX_OF_MTXT(i,i)));
      lll_exact_modreductionl3_subtractrow(mtxA, rows, cols, mtxT, k, i, ztmp);
    } while ( i-- != 0 );
  }
  // Clear all the GMP objects.
  for(i=0; i<elements; i++) mpz_clear(*(mtxT+i));
  free(mtxT);
  mpz_clear(ztmp);
  mpz_clear(ztmp2);
}

