/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <snap.h>
#include <math.h>

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

/*!
 * @brief swap two vectors.
 * @param[in] N The length of the vectors.
 * @param[in,out] *X A vector of N length.
 * @param[in] incX The leading dimension of the vector X.
 * @param[in,out] *Y A vector of N length.
 * @param[in] incY The leading dimension of the vector Y.
 */
void snap_mblas_fcswap(const unsigned int len, mpc_t *vecX, const int incX, mpc_t *vecY, const int incY) {
  unsigned int i;
  mpc_t *ptX, *ptY;
  if ( vecX != vecY ) {
    for (i=0,ptX=vecX,ptY=vecY; i<len; i++, ptX+=incX, ptY+=incY) {
      mpc_swap(*ptX, *ptY);
    }
  }
}

/*!
 * @brief The dot product of two vectors.
 * @warning This function is not safe if dotp is an (not the first) element of X or Y.
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 * @param[out] dotp The dot product.
 * @param[in] N The length of the vectors.
 * @param[in] *X A vector of N length.
 * @param[in] incX The leading dimension of the vector X.
 * @param[in] *Y A vector of N length.
 * @param[in] incY The leading dimension of the vector Y.
 */
void snap_mblas_frdot(mpfr_t dotp, const unsigned int len, mpfr_t *vecX, const int incX, mpfr_t *vecY, const int incY) {
  unsigned int i;
  mpfr_t elem2;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_frdot called: len(%d), incX(%d), incY(%d).\n",len,incX,incY);
#endif
  if ( len == 0 ) {
    mpfr_set_ui(dotp,0,SNAPLIB_MPFR_RND);
    return;
  }
  mpfr_init(elem2);
  if ( vecX == vecY ) {
    mpfr_mul(dotp, *vecX, *vecX, SNAPLIB_MPFR_RND);
    for (i=1,vecX+=incX; i<len; i++, vecX+=incX) {
      mpfr_mul(elem2, *vecX, *vecX, SNAPLIB_MPFR_RND);
      mpfr_add(dotp, dotp, elem2, SNAPLIB_MPFR_RND);
    }
  } else {
    mpfr_mul(dotp, *vecX, *vecY, SNAPLIB_MPFR_RND);
    for (i=1,vecX+=incX,vecY+=incY; i<len; i++, vecX+=incX, vecY+=incY) {
      mpfr_mul(elem2, *vecX, *vecY, SNAPLIB_MPFR_RND);
      mpfr_add(dotp, dotp, elem2, SNAPLIB_MPFR_RND);
    }
  }
  mpfr_clear(elem2);
}

/*!
 * @brief The dot product of two vectors.
 * @warning This function is not safe if dotp is an (not the first) element of X or Y.
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 * @param[out] dotp The dot product.
 * @param[in] N The length of the vectors.
 * @param[in] *X A vector of N length.
 * @param[in] incX The leading dimension of the vector X.
 * @param[in] *Y A vector of N length.
 * @param[in] incY The leading dimension of the vector Y.
 */
void snap_mblas_fcdotu(mpc_t dotp, const unsigned int len, mpc_t *vecX, const int incX, mpc_t *vecY, const int incY) {
  unsigned int i;
  mpc_t elem2;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_fcdotu called: len(%d), incX(%d), incY(%d).\n",len,incX,incY);
#endif
  if ( len == 0 ) {
    mpc_set_ui(dotp,0,SNAPLIB_MPC_RND);
    return;
  }
  mpc_init2(elem2,mpfr_get_default_prec());
  if ( vecX == vecY ) {
    mpc_mul(dotp, *vecX, *vecX, SNAPLIB_MPC_RND);
    for (i=1,vecX+=incX; i<len; i++, vecX+=incX) {
      mpc_mul(elem2, *vecX, *vecX, SNAPLIB_MPC_RND);
      mpc_add(dotp, dotp, elem2, SNAPLIB_MPC_RND);
    }
  } else {
    mpc_mul(dotp, *vecX, *vecY, SNAPLIB_MPC_RND);
    for (i=1,vecX+=incX,vecY+=incY; i<len; i++, vecX+=incX, vecY+=incY) {
      mpc_mul(elem2, *vecX, *vecY, SNAPLIB_MPC_RND);
      mpc_add(dotp, dotp, elem2, SNAPLIB_MPC_RND);
    }
  }
  mpc_clear(elem2);
}

/*!
 * @brief The dot product conj(X).Y of two vectors X and Y.
 * @warning This function is not safe if dotp is an (not the first) element of X or Y.
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 * @param[out] dotp The dot product.
 * @param[in] N The length of the vectors.
 * @param[in] *X A vector of N length.
 * @param[in] incX The leading dimension of the vector X.
 * @param[in] *Y A vector of N length.
 * @param[in] incY The leading dimension of the vector Y.
 */
void snap_mblas_fcdotc(mpc_t dotp, const unsigned int len, mpc_t *vecX, const int incX, mpc_t *vecY, const int incY) {
  unsigned int i;
  mpc_t elem2;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_fcdotc called: len(%d), incX(%d), incY(%d).\n",len,incX,incY);
#endif
#ifdef SNAPLIB_DEBUG
  mpfr_printf("dotp1: %+5Re%+5ReI\n",mpc_realref(dotp),mpc_imagref(dotp));
#endif
  if ( len == 0 ) {
    mpc_set_ui(dotp,0,SNAPLIB_MPC_RND);
    return;
  }
  mpc_init_set_ui(elem2, 0, SNAPLIB_MPC_RND);
#ifdef SNAPLIB_DEBUG
  mpfr_printf("dotp2: %+5Re%+5ReI\n",mpc_realref(dotp),mpc_imagref(dotp));
#endif
  if ( vecX == vecY ) {
    mpc_set_ui(dotp,0,SNAPLIB_MPC_RND);
    mpc_norm(mpc_realref(dotp), *vecX, MPC_RND_RE(SNAPLIB_MPC_RND));
    for (i=1,vecX+=incX; i<len; i++, vecX+=incX) {
      mpc_norm(mpc_realref(elem2), *vecX, MPC_RND_RE(SNAPLIB_MPC_RND));
      mpc_add(dotp, dotp, elem2, SNAPLIB_MPC_RND);
    }
  } else {
    mpc_conj(dotp, *vecX, SNAPLIB_MPC_RND);
    mpc_mul(dotp, dotp, *vecY, SNAPLIB_MPC_RND);
    for (i=1,vecX+=incX,vecY+=incY; i<len; i++, vecX+=incX, vecY+=incY) {
      mpc_conj(elem2, *vecX, SNAPLIB_MPC_RND);
      mpc_mul(elem2, elem2, *vecY, SNAPLIB_MPC_RND);
      mpc_add(dotp, dotp, elem2, SNAPLIB_MPC_RND);
    }
  }
#ifdef SNAPLIB_DEBUG
  mpfr_printf("dotp3: %+5Re%+5ReI\n",mpc_realref(dotp),mpc_imagref(dotp));
#endif
  mpc_clear(elem2);
}

/*!
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 */
void snap_mblas_fraxpy(const unsigned int len, const mpfr_t alpha, mpfr_t *vecX, const int incX, mpfr_t *vecY, const int incY) {
  unsigned int i;
  mpfr_t elem2;
  mpfr_init(elem2);
  for (i=0; i<len; i++, vecX+=incX, vecY+=incY) {
    mpfr_mul(elem2, alpha, *vecX, SNAPLIB_MPFR_RND);
    mpfr_add(*vecY, *vecY, elem2, SNAPLIB_MPFR_RND);
  }
  mpfr_clear(elem2);
}

/*!
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 */
void snap_mblas_fcaxpy(const unsigned int len, const mpc_t alpha, mpc_t *vecX, const int incX, mpc_t *vecY, const int incY) {
  unsigned int i;
  mpc_t elem2;
  mpc_init2(elem2, mpfr_get_default_prec());
  for (i=0; i<len; i++, vecX+=incX, vecY+=incY) {
    mpc_mul(elem2, alpha, *vecX, SNAPLIB_MPC_RND);
    mpc_add(*vecY, *vecY, elem2, SNAPLIB_MPC_RND);
  }
  mpc_clear(elem2);
}

/*!
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 */
void snap_mblas_frnrm2(mpfr_t norm, const unsigned int len, mpfr_t *vect, const int inc) {
  unsigned int i;
  mpfr_t elem2;
  mpfr_init(elem2);
  mpfr_set_ui(norm, 0, SNAPLIB_MPFR_RND);
  for (i=0; i<len; i++, vect+=inc) {
    mpfr_mul(elem2, *vect, *vect, SNAPLIB_MPFR_RND);
    mpfr_add(norm, norm, elem2, SNAPLIB_MPFR_RND);
  }
  mpfr_sqrt(norm, norm, SNAPLIB_MPFR_RND);
  mpfr_clear(elem2);
}

/*!
 * @warning The increments are treated as C style (not Fortran style) different from BLAS.
 */
 void snap_mblas_fcnrm2(mpfr_t norm, const unsigned int len, mpc_t *vect, const int inc) {
  unsigned int i;
  mpfr_t elem2;
  mpfr_init(elem2);
  mpfr_set_ui(norm, 0, SNAPLIB_MPFR_RND);
  for (i=0; i<len; i++, vect+=inc) {
    mpc_norm(elem2, *vect, MPC_RND_RE(SNAPLIB_MPC_RND));
    mpfr_add(norm, norm, elem2, SNAPLIB_MPFR_RND);
  }
  mpfr_sqrt(norm, norm, SNAPLIB_MPFR_RND);
  mpfr_clear(elem2);
}

/*!
 * @brief This function performs one of the matrix-vector operations: Ax or A^Tx.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for trans!='C' and diag=='N'.
 * @param[in] uplo 'U' if A is upper triangular, or 'L' if A is lower triangular.
 * @param[in] trans 'N' if AX=B, 'T' if A^TX=B, or 'C' if A^HX=B.
 * @param[in] diag 'N' if A is non-unit triangular, or 'U' if A is unit triangular.
 * @param[in] rows The number of rows in A.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] incB The leading dimension of B (column size of B).
 */
void snap_mblas_frtrmv(const char uplo, const char trans, const char diag, const unsigned int rows, mpfr_t *matA, const int ldA, mpfr_t *vecB, const int incB) {
  mpfr_t *ptA,*ptB;
  int inca, incb, lda;
  unsigned int i;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_frtrmv called: rows(%d), ldA(%d), incB(%d).\n",rows,ldA,incB);
#endif
  if ( rows == 0 ) {
    return;
  }
  // only implemented for few cases.
  if ( trans == 'C' || diag != 'N' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mblas_frtrmv met an unimplemented combination of parameters.\n");
#endif
    return;
  }
  // initialization
  if ( uplo == 'L' ) {
    if ( trans == 'N' ) { // Lower and normal
      lda = -ldA;
      inca = -1;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    } else { // Lower and transpose
      lda = 1;
      inca = ldA;
      incb = incB;
    }
  } else { // uplo == 'U'
    if ( trans == 'N' ) { // Upper and normal
      lda = ldA;
      inca = 1;
      incb = incB;
    } else { // Upper and transpose
      lda = -1;
      inca = -ldA;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    }
  }
  // matA is always the first element of the corresponding row.
  for ( i=rows,ptB=vecB,ptA=matA; i>0; i--,ptB+=incb,ptA+=lda+inca ) {
    snap_mblas_frdot(*ptB, i, ptA, inca, ptB, incb);
  }
}

/*!
 * @brief This function performs one of the matrix-vector operations: Ax, A^Tx or A^Hx.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for diag=='N'.
 * @param[in] uplo 'U' if A is upper triangular, or 'L' if A is lower triangular.
 * @param[in] trans 'N' if AX=B, 'T' if A^TX=B, or 'C' if A^HX=B.
 * @param[in] diag 'N' if A is non-unit triangular, or 'U' if A is unit triangular.
 * @param[in] rows The number of rows in A.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] incB The leading dimension of B (column size of B).
 */
void snap_mblas_fctrmv(const char uplo, const char trans, const char diag, const unsigned int rows, mpc_t *matA, const int ldA, mpc_t *vecB, const int incB) {
  mpc_t *ptA,*ptB;
  int inca, incb, lda;
  unsigned int i;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_fctrmv called: rows(%d), ldA(%d), incB(%d).\n",rows,ldA,incB);
#endif
  if ( rows == 0 ) {
    return;
  }
  // only implemented for few cases.
  if ( diag != 'N' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mblas_fctrmv met an unimplemented combination of parameters.\n");
#endif
    return;
  }
  // initialization
  if ( uplo == 'L' ) {
    if ( trans == 'N' ) { // Lower and normal
      lda = -ldA;
      inca = -1;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    } else { // Lower and transpose
      lda = 1;
      inca = ldA;
      incb = incB;
    }
  } else { // uplo == 'U'
    if ( trans == 'N' ) { // Upper and normal
      lda = ldA;
      inca = 1;
      incb = incB;
    } else { // Upper and transpose
      lda = -1;
      inca = -ldA;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    }
  }
  // matA is always the first element of the corresponding row.
  if ( trans != 'C' ) {
    for ( i=rows,ptB=vecB,ptA=matA; i>0; i--,ptB+=incb,ptA+=lda+inca ) {
      snap_mblas_fcdotu(*ptB, i, ptA, inca, ptB, incb);
    }
  } else {
    for ( i=rows,ptB=vecB,ptA=matA; i>0; i--,ptB+=incb,ptA+=lda+inca ) {
      snap_mblas_fcdotc(*ptB, i, ptA, inca, ptB, incb);
    }
  }
}

/*!
 * @brief This function performs one of the matrix-vector operations: alpha*A*x+beta*y or alpha*A^T*x+beta*y.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for trans='N' or trans='T'.
 * @param[in] trans 'N' if Ax, 'T' if A^Tx, or 'C' if A^Hx.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of cols in A.
 * @param[in] alpha A scalar.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in] vecX The array X.
 * @param[in] incX The leading dimension of X (column size of X).
 * @param[in] beta A scalar.
 * @param[in,out] vecY The array Y.
 * @param[in] incY The leading dimension of Y (column size of Y).
 */
void snap_mblas_frgemv(const char trans, const unsigned int rows, const unsigned int cols, mpfr_t alpha, mpfr_t *matA, const int ldA, mpfr_t *vecX, const int incX, mpfr_t beta, mpfr_t *vecY, const int incY) {
  mpfr_t tmp,*ptY,*ptA;
  int inca,lda;
  unsigned int i,maxi,len;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_frgemv called: rows(%d), cols(%d), ldA(%d), incX(%d), incY(%d).\n",rows,cols,ldA,incX,incY);
#endif
  if ( rows == 0 ) {
    return;
  }
  // only implemented for few cases.
  if ( trans == 'C' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mblas_frgemv met an unimplemented combination of parameters.\n");
#endif
    return;
  }
  // initialization
  if ( trans == 'N' ) {
    inca = 1;
    lda = ldA;
    maxi=rows;
    len=cols;
  } else {
    inca = ldA;
    lda = 1;
    maxi=cols;
    len=rows;
  }
  if ( mpfr_sgn(beta) == 0 ) {
    if ( mpfr_cmp_ui(alpha, 1) == 0 ) {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_frdot(*ptY, len, ptA, inca, vecX, incX);
      }
    } else {
      mpfr_init(tmp);
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_frdot(tmp, len, ptA, inca, vecX, incX);
        mpfr_mul(*ptY, tmp, alpha, SNAPLIB_MPFR_RND);
      }
      mpfr_clear(tmp);
    }
  } else {
    mpfr_init(tmp);
    if ( mpfr_cmp_ui(alpha, 1) == 0 ) {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_frdot(tmp, len, ptA, inca, vecX, incX);
        mpfr_mul(*ptY, *ptY, beta, SNAPLIB_MPFR_RND);
        mpfr_add(*ptY, *ptY, tmp, SNAPLIB_MPFR_RND);
      }
    } else {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_frdot(tmp, len, ptA, inca, vecX, incX);
        mpfr_mul(tmp, tmp, alpha, SNAPLIB_MPFR_RND);
        mpfr_mul(*ptY, *ptY, beta, SNAPLIB_MPFR_RND);
        mpfr_add(*ptY, *ptY, tmp, SNAPLIB_MPFR_RND);
      }
    }
    mpfr_clear(tmp);
  }
}

/*!
 * @brief This function performs one of the matrix-vector operations: alpha*A*x+beta*y, alpha*A^T*x+beta*y or alpha*A^H*x+beta*y.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for trans='N' or trans='T'.
 * @param[in] trans 'N' if Ax, 'T' if A^Tx, or 'C' if A^Hx.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of cols in A.
 * @param[in] alpha A scalar.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in] vecX The array X.
 * @param[in] incX The leading dimension of X (column size of X).
 * @param[in] beta A scalar.
 * @param[in,out] vecY The array Y.
 * @param[in] incY The leading dimension of Y (column size of Y).
 */
void snap_mblas_fcgemv(const char trans, const unsigned int rows, const unsigned int cols, mpc_t alpha, mpc_t *matA, const int ldA, mpc_t *vecX, const int incX, mpc_t beta, mpc_t *vecY, const int incY) {
  mpc_t tmp,*ptY,*ptA;
  int inca,lda;
  unsigned int i,maxi,len;
#ifdef SNAPLIB_DEBUG
  printf("snap_mblas_fcgemv called: rows(%d), cols(%d), ldA(%d), incX(%d), incY(%d).\n",rows,cols,ldA,incX,incY);
#endif
  if ( rows == 0 ) {
    return;
  }
  // only implemented for few cases.
  if ( trans == 'C' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mblas_fcgemv met an unimplemented combination of parameters.\n");
#endif
    return;
  }
  // initialization
  if ( trans == 'N' ) {
    inca = 1;
    lda = ldA;
    maxi=rows;
    len=cols;
  } else {
    inca = ldA;
    lda = 1;
    maxi=cols;
    len=rows;
  }
  if ( mpfr_sgn(mpc_realref(beta)) == 0 && mpfr_sgn(mpc_imagref(beta)) == 0 ) {
    if ( mpc_cmp_si_si(alpha, 1, 0) == 0 ) {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_fcdotu(*ptY, len, ptA, inca, vecX, incX);
      }
    } else {
      mpc_init2(tmp, mpfr_get_default_prec());
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_fcdotu(tmp, len, ptA, inca, vecX, incX);
        mpc_mul(*ptY, tmp, alpha, SNAPLIB_MPC_RND);
      }
      mpc_clear(tmp);
    }
  } else {
    mpc_init2(tmp, mpfr_get_default_prec());
    if ( mpc_cmp_si_si(alpha, 1, 0) == 0 ) {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_fcdotu(tmp, len, ptA, inca, vecX, incX);
        mpc_mul(*ptY, *ptY, beta, SNAPLIB_MPC_RND);
        mpc_add(*ptY, *ptY, tmp, SNAPLIB_MPC_RND);
      }
    } else {
      for ( i=0,ptA=matA,ptY=vecY; i<maxi; i++,ptA+=lda,ptY+=incY ) {
        snap_mblas_fcdotu(tmp, len, ptA, inca, vecX, incX);
        mpc_mul(tmp, tmp, alpha, SNAPLIB_MPC_RND);
        mpc_mul(*ptY, *ptY, beta, SNAPLIB_MPC_RND);
        mpc_add(*ptY, *ptY, tmp, SNAPLIB_MPC_RND);
      }
    }
    mpc_clear(tmp);
  }
}

/*!
 * @brief This function performs one of the matrix-matrix operations: alpha*op(A)B or alpha*Bop(A).
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for side='L', uplo='U', trans='N' and diag=='N'.
 * @param[in] side 'L' if op(A) B, or 'R' if B op(A).
 * @param[in] uplo 'U' if A is upper triangular, or 'L' if A is lower triangular.
 * @param[in] trans 'N' if AX=B, 'T' if A^TX=B, or 'C' if A^HX=B.
 * @param[in] diag 'N' if A is non-unit triangular, or 'U' if A is unit triangular.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of columns in A.
 * @param[in] alpha The scalar multiple.
 * @param[in] matA The triangular array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in,out] matB The array B.
 * @param[in] ldB The leading dimension of B (column size of B).
 */
void snap_mblas_fctrmm(const char side, const char uplo, const char trans, const char diag, const unsigned int rows, const unsigned int cols, mpc_t alpha, mpc_t *matA, const int ldA, mpc_t *matB, const int ldB) {
  mpc_t *ptB;
  unsigned int i;
  if ( rows == 0 || cols == 0 ) {
    return;
  }
  // only implemented for few cases.
  if ( side != 'L' || uplo != 'U' || trans != 'N' || diag != 'N' || mpc_cmp_si(alpha, 1) != 0 ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mblas_fctrmm met an unimplemented combination of parameters.\n");
#endif
    return;
  }
  for ( i = 0, ptB=matB; i < cols; i++, ptB++ ) {
    snap_mblas_fctrmv(uplo, trans, diag, rows, matA, ldA, ptB, ldB);
  }
}


