/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <snap.h>
#include <math.h>
#include <cblas.h>

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAPLIB_MAX(a,b) (((a)>(b)) ? (a):(b))
#define SNAPLIB_MIN(a,b) (((a)<(b)) ? (a):(b))

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
void show_mpfrarray(mpfr_t *mtx,const unsigned int rows,const unsigned int cols,const int lda,const int incA) {
  unsigned int i,j;
  mpfr_t *pt;
  for(i=0; i<rows; i++) {
    for(j=0,pt=mtx+i*lda; j<cols; j++,pt+=incA) {
      mpfr_printf("%+5Re ",*pt);
    }
    printf("\n");
  }
}
void show_mpfrarray1d(mpfr_t *mtx,const unsigned int rows,const int incA) {
  unsigned int j;
  mpfr_t *pt;
  for(j=0,pt=mtx; j<rows; j++,pt+=incA) {
    mpfr_printf("%+5Re, ",*pt);
  }
  printf("\n");
}
void show_mpcarray(mpc_t *mtx,const unsigned int rows,const unsigned int cols,const int lda,const int incA) {
  unsigned int i,j;
  mpc_t *pt;
  for(i=0; i<rows; i++) {
    for(j=0,pt=mtx+i*lda; j<cols; j++,pt+=incA) {
      // mpfr_printf("%+5Re%+5ReI ",mpc_realref(*pt),mpc_imagref(*pt));
      printf("(%+Lf%+LfI),", mpfr_get_ld(mpc_realref(*pt), MPFR_RNDN), mpfr_get_ld(mpc_imagref(*pt), MPFR_RNDN));
    }
    printf("\n");
  }
}
void show_mpcarray1d(mpc_t *mtx,const unsigned int rows,const int incA) {
  unsigned int j;
  mpc_t *pt;
  for(j=0,pt=mtx; j<rows; j++,pt+=incA) {
    mpfr_printf("%+5Re%+5ReI, ",mpc_realref(*pt),mpc_imagref(*pt));
  }
  printf("\n");
}
#endif

/*!
 * @brief The function computes a QR factorization of a real M-by-N matrix A: A=QR.
 * @warning A must be a matrix of the row major.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of rows in A.
 * @param[in,out] matA A matrix A on input and the resulting matrices QR on exit.
 * @param[in] lda The leading dimension of A.
 * @param[out] tau An array for values tau.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_frgeqrf(const unsigned int rows, const unsigned int cols, mpfr_t *matA, const int lda, mpfr_t *tau) {
  unsigned int i,j,sz;
  int result = 0;
  // we need vecV and vecW since matA is a row-major matrix.
  mpfr_t *ptT,*vecV,*ptV,beta,*ptA,*ptX,*vecW,*ptW;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_frgeqrf called: rows(%d), cols(%d), matA(pt=%p), lda(%d), tau(pt=%p).\n", rows,cols,matA,lda,tau);
#endif
  // initialization
  sz = SNAPLIB_MIN(rows,cols);
  mpfr_init(beta);
  vecV = (mpfr_t*) malloc(sizeof(mpfr_t)*rows);
  mpfr_init_set_ui(*vecV,1,SNAPLIB_MPFR_RND); // vecV[0] is always 1.
  for ( i=1,ptV=vecV+1; i<rows; i++,ptV++ ) {
    mpfr_init(*ptV);
  }
  vecW = (mpfr_t*) malloc(sizeof(mpfr_t)*(cols-1));
  for ( i=1,ptW=vecW; i<cols; i++,ptW++ ) {
    mpfr_init(*ptW);
  }
  // main loop
  for ( i=0,ptT=tau,ptA=matA; i<sz; i++,ptT++,ptA+=lda+1 ) {
    // generate reflector
    mpfr_set(beta, *ptA, SNAPLIB_MPFR_RND);
    for ( j=i+1,ptV=vecV+1,ptX=ptA+lda; j<rows; j++,ptV++,ptX+=lda ) {
      mpfr_set(*ptV, *ptX, SNAPLIB_MPFR_RND);
    }
    if ( snap_mlapack_frlarfg( rows-i, beta, vecV+1, 1, *ptT) < 0 ) {
      printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frlarfg could not compute the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
#ifdef SNAPLIB_DEBUG
    mpfr_printf("beta: %+5Re\n",beta);
#endif
    if ( mpfr_sgn(beta) == 0 ) {
      continue;
    }
    // apply reflector
    if ( snap_mlapack_frlarfx( 'L', rows-i, cols-i-1, vecV, *ptT, ptA+1, lda, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frlarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    // overwrite A with v
    mpfr_set(*ptA, beta, SNAPLIB_MPFR_RND);
    for ( j=i+1,ptX=ptA+lda,ptV=vecV+1; j<rows; j++,ptX+=lda,ptV++ ) {
      mpfr_set(*ptX, *ptV, SNAPLIB_MPFR_RND);
    }
#ifdef SNAPLIB_DEBUG
    printf("====\n");
    show_mpfrarray(matA,rows,cols,lda,1);
#endif
  }
MEMFREE:
  // memory free
  mpfr_clear(beta);
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    mpfr_clear(*ptV);
  }
  for ( i=1,ptW=vecW; i<cols; i++,ptW++ ) {
    mpfr_clear(*ptW);
  }
  free(vecV);
  free(vecW);
  return result;
}

/*!
 * @brief The function computes a QR factorization of a real M-by-N complex matrix A: A=QR.
 * @warning A must be a complex matrix of the row major.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of rows in A.
 * @param[in,out] matA A complex matrix A on input and the resulting matrices QR on exit.
 * @param[in] lda The leading dimension of A.
 * @param[out] tau An complex array for values tau.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fcgeqrf(const unsigned int rows, const unsigned int cols, mpc_t *matA, const int lda, mpc_t *tau) {
  unsigned int i,j,sz;
  int result = 0;
  // we need vecV and vecW since matA is a row-major matrix.
  mpc_t *ptT,*vecV,*ptV,beta,*ptA,*ptX,*vecW,*ptW;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_fcgeqrf called: rows(%d), cols(%d), matA(pt=%p), lda(%d), tau(pt=%p).\n", rows,cols,matA,lda,tau);
#endif
  // initialization
  sz = SNAPLIB_MIN(rows,cols);
  mpc_init2(beta, mpfr_get_default_prec());
  vecV = (mpc_t*) malloc(sizeof(mpc_t)*rows);
  mpc_init_set_ui(*vecV,1,SNAPLIB_MPC_RND); // vecV[0] is always 1.
  for ( i=1,ptV=vecV+1; i<rows; i++,ptV++ ) {
    mpc_init2(*ptV, mpfr_get_default_prec());
  }
  vecW = (mpc_t*) malloc(sizeof(mpc_t)*(cols-1));
  for ( i=1,ptW=vecW; i<cols; i++,ptW++ ) {
    mpc_init2(*ptW, mpfr_get_default_prec());
  }
  // main loop
  for ( i=0,ptT=tau,ptA=matA; i<sz; i++,ptT++,ptA+=lda+1 ) {
    // generate reflector
    mpc_set(beta, *ptA, SNAPLIB_MPC_RND);
#ifdef SNAPLIB_DEBUG
    mpfr_printf("beta(alpha): %+5Re%+5ReI\n",mpc_realref(beta),mpc_imagref(beta));
#endif
    for ( j=i+1,ptV=vecV+1,ptX=ptA+lda; j<rows; j++,ptV++,ptX+=lda ) {
      mpc_set(*ptV, *ptX, SNAPLIB_MPC_RND);
    }
    if ( snap_mlapack_fclarfg( rows-i, beta, vecV+1, 1, *ptT) < 0 ) {
      printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fclarfg could not compute the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
#ifdef SNAPLIB_DEBUG
    mpfr_printf("beta: %+5Re%+5ReI\n",mpc_realref(beta),mpc_imagref(beta));
    mpfr_printf("tau: %+5Re%+5ReI\n",mpc_realref(*ptT),mpc_imagref(*ptT));
#endif
    if ( mpfr_sgn(mpc_realref(beta)) == 0 && mpfr_sgn(mpc_imagref(beta)) == 0 ) {
      continue;
    }
    // apply reflector
    // just for larfx we take the conjugate of tau temporary.
    mpc_conj(*ptT, *ptT, SNAPLIB_MPC_RND);
    if ( snap_mlapack_fclarfx( 'L', rows-i, cols-i-1, vecV, *ptT, ptA+1, lda, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fclarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    mpc_conj(*ptT, *ptT, SNAPLIB_MPC_RND);
    // overwrite A with v
    mpc_set(*ptA, beta, SNAPLIB_MPC_RND);
    for ( j=i+1,ptX=ptA+lda,ptV=vecV+1; j<rows; j++,ptX+=lda,ptV++ ) {
      mpc_set(*ptX, *ptV, SNAPLIB_MPC_RND);
    }
#ifdef SNAPLIB_DEBUG
    printf("==== in fcgeqrf ====\n");
    show_mpcarray(matA,rows,cols,lda,1);
#endif
  }
MEMFREE:
  // memory free
  mpc_clear(beta);
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    mpc_clear(*ptV);
  }
  for ( i=1,ptW=vecW; i<cols; i++,ptW++ ) {
    mpc_clear(*ptW);
  }
  free(vecV);
  free(vecW);
  return result;
}

/*!
 * @brief The function solves overdetermined or underdetermined systems for GE matrices.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for trans='N', NRHS=1 and overdetermined system.
 * @param[in] trans 'N' if AX=B or 'T' if A^TX=B.
 * @param[in] rowsA The number of rows in A.
 * @param[in] colsA The number of rows in A.
 * @param[in,out] matA A matrix A on input and the resulting matrices QR on exit.
 * @param[in] ldA The leading dimension of A.
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] ldB The leading dimension of B (column size of B).
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_frgels(const char trans, const unsigned int rowsA, const unsigned int colsA, const unsigned int nrhs, mpfr_t *matA, const int ldA, mpfr_t *vecB, const int ldB) {
  unsigned int i,j,rows,cols,sz,lenW;
  int lda,inca,incb,result = 0;
  char uplo;
  // we need vecV and vecW since matA is a row-major matrix.
  mpfr_t *vecV,*ptV,beta,*ptA,*ptX,*vecW,*ptW,tau;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_frgels called: trans(%c), rowsA(%d), colsA(%d), nrhs(%d), matA(pt=%p), ldA(%d), vecB(pt=%p), ldB(%d).\n",trans,rowsA,colsA,nrhs,matA,ldA,vecB,ldB);
#endif
  if ( trans != 'N' || nrhs != 1 ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_frgels met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
  // initialization
  rows = rowsA;
  cols = colsA;
  lda = ldA;
  inca = 1;
  incb = ldB;
  uplo = 'U';
  sz = SNAPLIB_MIN(rows,cols);
  mpfr_init(beta);
  mpfr_init(tau);
  vecV = (mpfr_t*) malloc(sizeof(mpfr_t)*rows);
  mpfr_init_set_ui(*vecV,1,SNAPLIB_MPFR_RND); // vecV[0] is always 1.
  for ( i=1,ptV=vecV+1; i<rows; i++,ptV++ ) {
    mpfr_init(*ptV);
  }
  lenW = SNAPLIB_MAX(1,cols-1);
  vecW = (mpfr_t*) malloc(sizeof(mpfr_t)*(lenW));
  for ( i=0,ptW=vecW; i<lenW; i++,ptW++ ) {
    mpfr_init(*ptW);
  }
  // main loop
  for ( i=0,ptA=matA; i<sz; i++,ptA+=lda+inca ) {
#ifdef SNAPLIB_DEBUG
//printf("matA (%d)\n",i); show_mpfrarray(matA,rowsA,colsA,ldA,1);
#endif
    // generate reflector
    mpfr_set(beta, *ptA, SNAPLIB_MPFR_RND);
    for ( j=i+1,ptV=vecV+1,ptX=ptA+lda; j<rows; j++,ptV++,ptX+=lda ) {
      mpfr_set(*ptV, *ptX, SNAPLIB_MPFR_RND);
    }
    if ( snap_mlapack_frlarfg( rows-i, beta, vecV+1, 1, tau) < 0 ) {
      printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frlarfg could not compute the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    if ( mpfr_sgn(beta) == 0 ) {
      continue;
    }
    // apply reflector to A
    if ( snap_mlapack_frlarfx( 'L', rows-i, cols-i-1, vecV, tau, ptA+inca, lda, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frlarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    // apply reflector to b
    if ( snap_mlapack_frlarfx( 'L', rows-i, 1, vecV, tau, vecB+i*incb, incb, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frlarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    // overwrite A with v
    mpfr_set(*ptA, beta, SNAPLIB_MPFR_RND);
    for ( j=i+1,ptX=ptA+lda,ptV=vecV+1; j<rows; j++,ptX+=lda,ptV++ ) {
      mpfr_set(*ptX, *ptV, SNAPLIB_MPFR_RND);
    }
  }
  // Solve R x = b by back substitution.
  if ( snap_mlapack_frtrtrs(uplo, trans, 'N', sz, 1, matA, lda, vecB, incb) < 0 ) {
    printf("ERROR in snap_mlapack_frgeqrf: snap_mlapack_frtrtrs could not solve the rectangular system.\n");
    result = -1;
    goto MEMFREE;
  }
MEMFREE:
  // memory free
  mpfr_clear(beta);
  mpfr_clear(tau);
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    mpfr_clear(*ptV);
  }
  for ( i=0,ptW=vecW; i<lenW; i++,ptW++ ) {
    mpfr_clear(*ptW);
  }
  free(vecV);
  free(vecW);
  return result;
}

/*!
 * @brief The function solves overdetermined or underdetermined systems for GE matrices.
 * @warning A must be a complex matrix of the row major.
 * @warning Only implemented for trans='N', NRHS=1 and overdetermined system.
 * @param[in] trans 'N' if AX=B or 'T' if A^TX=B.
 * @param[in] rowsA The number of rows in A.
 * @param[in] colsA The number of rows in A.
 * @param[in,out] matA A matrix A on input and the resulting matrices QR on exit.
 * @param[in] ldA The leading dimension of A.
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] ldB The leading dimension of B (column size of B).
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fcgels(const char trans, const unsigned int rowsA, const unsigned int colsA, const unsigned int nrhs, mpc_t *matA, const int ldA, mpc_t *vecB, const int ldB) {
  unsigned int i,j,rows,cols,sz,lenW;
  int lda,inca,incb,result = 0;
  char uplo;
  // we need vecV and vecW since matA is a row-major matrix.
  mpc_t *vecV,*ptV,beta,*ptA,*ptX,*vecW,*ptW,tau;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_fcgels called: trans(%c), rowsA(%d), colsA(%d), nrhs(%d), matA(pt=%p), ldA(%d), vecB(pt=%p), ldB(%d).\n",trans,rowsA,colsA,nrhs,matA,ldA,vecB,ldB);
#endif
  if ( trans != 'N' || nrhs != 1 ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_fcgels met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
  // initialization
  rows = rowsA;
  cols = colsA;
  lda = ldA;
  inca = 1;
  incb = ldB;
  uplo = 'U';
  sz = SNAPLIB_MIN(rows,cols);
  mpc_init2(beta,mpfr_get_default_prec());
  mpc_init2(tau,mpfr_get_default_prec());
  vecV = (mpc_t*) malloc(sizeof(mpc_t)*rows);
  mpc_init_set_ui(*vecV,1,SNAPLIB_MPC_RND); // vecV[0] is always 1.
  for ( i=1,ptV=vecV+1; i<rows; i++,ptV++ ) {
    mpc_init2(*ptV,mpfr_get_default_prec());
  }
  lenW = SNAPLIB_MAX(1,cols-1);
  vecW = (mpc_t*) malloc(sizeof(mpc_t)*(lenW));
  for ( i=0,ptW=vecW; i<lenW; i++,ptW++ ) {
    mpc_init2(*ptW,mpfr_get_default_prec());
  }
  // main loop
  for ( i=0,ptA=matA; i<sz; i++,ptA+=lda+inca ) {
#ifdef SNAPLIB_DEBUG
//printf("matA (%d)\n",i); show_mpcarray(matA,rowsA,colsA,ldA,1);
#endif
    // generate reflector
    mpc_set(beta, *ptA, SNAPLIB_MPC_RND);
    for ( j=i+1,ptV=vecV+1,ptX=ptA+lda; j<rows; j++,ptV++,ptX+=lda ) {
      mpc_set(*ptV, *ptX, SNAPLIB_MPC_RND);
    }
    if ( snap_mlapack_fclarfg( rows-i, beta, vecV+1, 1, tau) < 0 ) {
      printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fclarfg could not compute the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    if ( mpfr_sgn(mpc_realref(beta)) == 0 && mpfr_sgn(mpc_imagref(beta)) == 0 ) {
      continue;
    }
    // apply reflector to A
    // for larfx, we need to take the conjugate of tau.
    mpc_conj(tau, tau, SNAPLIB_MPC_RND);
    if ( snap_mlapack_fclarfx( 'L', rows-i, cols-i-1, vecV, tau, ptA+inca, lda, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fclarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    // apply reflector to b
    if ( snap_mlapack_fclarfx( 'L', rows-i, 1, vecV, tau, vecB+i*incb, incb, vecW) < 0 ) {
      printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fclarfx could not apply the householder reflection.\n");
      result = -1;
      goto MEMFREE;
    }
    // overwrite A with v
    mpc_set(*ptA, beta, SNAPLIB_MPC_RND);
    for ( j=i+1,ptX=ptA+lda,ptV=vecV+1; j<rows; j++,ptX+=lda,ptV++ ) {
      mpc_set(*ptX, *ptV, SNAPLIB_MPC_RND);
    }
  }
  // Solve R x = b by back substitution.
  if ( snap_mlapack_fctrtrs(uplo, trans, 'N', sz, 1, matA, lda, vecB, incb) < 0 ) {
    printf("ERROR in snap_mlapack_fcgeqrf: snap_mlapack_fctrtrs could not solve the rectangular system.\n");
    result = -1;
    goto MEMFREE;
  }
MEMFREE:
  // memory free
  mpc_clear(beta);
  mpc_clear(tau);
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    mpc_clear(*ptV);
  }
  for ( i=0,ptW=vecW; i<lenW; i++,ptW++ ) {
    mpc_clear(*ptW);
  }
  free(vecV);
  free(vecW);
  return result;
}

/*!
 * @brief The function solves a triangular system of the form AX=B or A^TX=B.
 * @warning A must be a matrix of the row major.
 * @warning B must be a vector (i.e. must be one dimensional matrix).
 * @warning Only implemented for trans!='C', diag=='N' and NRHS=1.
 * @param[in] uplo 'U' if A is upper triangular, or 'L' if A is lower triangular.
 * @param[in] trans 'N' if AX=B, 'T' if A^TX=B, or 'C' if A^HX=B.
 * @param[in] diag 'N' if A is non-unit triangular, or 'U' if A is unit triangular.
 * @param[in] rows The number of rows in A.
 * @param[in] nrhs The number of the right hand side vector B.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] incB The leading dimension of B (column size of B).
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_frtrtrs(const char uplo, const char trans, const char diag, const unsigned int rows, const unsigned int nrhs, mpfr_t *matA, const int ldA, mpfr_t *vecB, const int incB) {
  mpfr_t dotp, *ptA,*ptB;
  int inca, incb, lda;
  unsigned int i;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_frtrtrs called: uplo(%c), trans(%c), diag(%c), rows(%d), nrhs(%d), matA(pt=%p), ldA(%d), vecB(pt=%p), incB(%d).\n",uplo,trans,diag,rows,nrhs,matA,ldA,vecB,incB);
#endif
  // only implemented for few cases.
  if ( trans == 'C' || diag != 'N' || nrhs != 1 ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_frtrtrs met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
#ifdef SNAPLIB_DEBUG
  printf("matA (%d)\n",rows);
  show_mpfrarray(matA,rows,rows,ldA,1);
  printf("vecB (%d)\n",rows);
  show_mpfrarray1d(vecB,rows,incB);
#endif
  // initialization
  mpfr_init(dotp);
  if ( uplo == 'L' ) {
    if ( trans == 'N' ) { // Lower and normal
      lda = ldA;
      inca = 1;
      incb = incB;
    } else { // Lower and transpose
      lda = -1;
      inca = -ldA;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    }
  } else { // uplo == 'U'
    if ( trans == 'N' ) { // Upper and normal
      lda = -ldA;
      inca = -1;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    } else { // Upper and transpose
      lda = 1;
      inca = ldA;
      incb = incB;
    }
  }
#ifdef SNAPLIB_DEBUG
  mpfr_printf("==> rows(%d), inca(%d), matA(pt=%p), lda(%d), vecB(pt=%p), incb(%d).\n",rows,inca,matA,lda,vecB,incb);
#endif
  // solve by simple substitutions
  // matA is always the first element of the corresponding row.
  for ( i=0,ptB=vecB,ptA=matA; i<rows; i++,ptB+=incb,ptA+=lda+inca,matA+=lda ) {
    snap_mblas_frdot(dotp, i, matA, inca, vecB, incb);
#ifdef SNAPLIB_DEBUG
    mpfr_printf("dotp(%d): %Re, ptA: %Re, ptB: %Re, ", i, dotp, *ptA, *ptB);
#endif
    mpfr_sub(*ptB, *ptB, dotp, SNAPLIB_MPFR_RND);
    mpfr_div(*ptB, *ptB, *ptA, SNAPLIB_MPFR_RND);
  }
#ifdef SNAPLIB_DEBUG
  printf("result of vecB (%p)\n",vecB);
  show_mpfrarray1d(vecB,rows,incb);
#endif
  mpfr_clear(dotp);
  return 0;
}

/*!
 * @brief The function solves a triangular system of the form AX=B or A^TX=B.
 * @warning A must be a complex matrix of the row major.
 * @warning B must be a complex vector (i.e. must be one dimensional matrix).
 * @warning Only implemented for NRHS=1.
 * @param[in] uplo 'U' if A is upper triangular, or 'L' if A is lower triangular.
 * @param[in] trans 'N' if AX=B, 'T' if A^TX=B, or 'C' if A^HX=B.
 * @param[in] diag 'N' if A is non-unit triangular, or 'U' if A is unit triangular.
 * @param[in] rows The number of rows in A.
 * @param[in] nrhs The number of the right hand side vector B.
 * @param[in] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in,out] vecB The right hand side vectors.
 * @param[in] incB The leading dimension of B (column size of B).
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fctrtrs(const char uplo, const char trans, const char diag, const unsigned int rows, const unsigned int nrhs, mpc_t *matA, const int ldA, mpc_t *vecB, const int incB) {
  mpc_t dotp, *ptA,*ptB;
  int inca, incb, lda;
  unsigned int i;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_fctrtrs called: uplo(%c), trans(%c), diag(%c), rows(%d), nrhs(%d), matA(pt=%p), ldA(%d), vecB(pt=%p), incB(%d).\n",uplo,trans,diag,rows,nrhs,matA,ldA,vecB,incB);
#endif
  // only implemented for few cases.
  if ( nrhs != 1 ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_fctrtrs met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
#ifdef SNAPLIB_DEBUG
  printf("matA (%d)\n",rows);
  show_mpcarray(matA,rows,rows,ldA,1);
  printf("vecB (%d)\n",rows);
  show_mpcarray1d(vecB,rows,incB);
#endif
  // initialization
  mpc_init2(dotp,mpfr_get_default_prec());
  if ( uplo == 'L' ) {
    if ( trans == 'N' ) { // Lower and normal
      lda = ldA;
      inca = 1;
      incb = incB;
    } else { // Lower and transpose
      lda = -1;
      inca = -ldA;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    }
  } else { // uplo == 'U'
    if ( trans == 'N' ) { // Upper and normal
      lda = -ldA;
      inca = -1;
      incb = -incB;
      matA += (rows-1)*(ldA+1);
      vecB += (rows-1)*incB;
    } else { // Upper and transpose
      lda = 1;
      inca = ldA;
      incb = incB;
    }
  }
#ifdef SNAPLIB_DEBUG
  mpfr_printf("==> rows(%d), inca(%d), matA(pt=%p), lda(%d), vecB(pt=%p), incb(%d).\n",rows,inca,matA,lda,vecB,incb);
#endif
  // solve by simple substitutions
  // matA is always the first element of the corresponding row.
  if ( diag == 'N' ) {
    if ( trans != 'C' ) {
      for ( i=0,ptB=vecB,ptA=matA; i<rows; i++,ptB+=incb,ptA+=lda+inca,matA+=lda ) {
        snap_mblas_fcdotu(dotp, i, matA, inca, vecB, incb);
#ifdef SNAPLIB_DEBUG
        mpfr_printf("dotp(%d): (%Re,%ReI), ptA: (%Re,%ReI), ptB: (%Re,%ReI), ", i, mpc_realref(dotp), mpc_imagref(dotp), mpc_realref(*ptA), mpc_imagref(*ptA), mpc_realref(*ptB), mpc_imagref(*ptB));
#endif
        mpc_sub(*ptB, *ptB, dotp, SNAPLIB_MPC_RND);
        mpc_div(*ptB, *ptB, *ptA, SNAPLIB_MPC_RND);
      }
    } else {
      for ( i=0,ptB=vecB,ptA=matA; i<rows; i++,ptB+=incb,ptA+=lda+inca,matA+=lda ) {
        snap_mblas_fcdotc(dotp, i, matA, inca, vecB, incb);
#ifdef SNAPLIB_DEBUG
        mpfr_printf("dotp(%d): (%Re,%ReI), ptA: (%Re,%ReI), ptB: (%Re,%ReI), ", i, mpc_realref(dotp), mpc_imagref(dotp), mpc_realref(*ptA), mpc_imagref(*ptA), mpc_realref(*ptB), mpc_imagref(*ptB));
#endif
        mpc_sub(*ptB, *ptB, dotp, SNAPLIB_MPC_RND);
        mpc_conj(*ptB, *ptB, SNAPLIB_MPC_RND);
        mpc_div(*ptB, *ptB, *ptA, SNAPLIB_MPC_RND);
        mpc_conj(*ptB, *ptB, SNAPLIB_MPC_RND);
      }
    }
  } else {
    if ( trans != 'C' ) {
      for ( i=0,ptB=vecB,ptA=matA; i<rows; i++,ptB+=incb,ptA+=lda+inca,matA+=lda ) {
        snap_mblas_fcdotu(dotp, i, matA, inca, vecB, incb);
#ifdef SNAPLIB_DEBUG
        mpfr_printf("dotp(%d): (%Re,%ReI), ptA: (%Re,%ReI), ptB: (%Re,%ReI), ", i, mpc_realref(dotp), mpc_imagref(dotp), mpc_realref(*ptA), mpc_imagref(*ptA), mpc_realref(*ptB), mpc_imagref(*ptB));
#endif
        mpc_sub(*ptB, *ptB, dotp, SNAPLIB_MPC_RND);
      }
    } else {
      for ( i=0,ptB=vecB,ptA=matA; i<rows; i++,ptB+=incb,ptA+=lda+inca,matA+=lda ) {
        snap_mblas_fcdotc(dotp, i, matA, inca, vecB, incb);
#ifdef SNAPLIB_DEBUG
        mpfr_printf("dotp(%d): (%Re,%ReI), ptA: (%Re,%ReI), ptB: (%Re,%ReI), ", i, mpc_realref(dotp), mpc_imagref(dotp), mpc_realref(*ptA), mpc_imagref(*ptA), mpc_realref(*ptB), mpc_imagref(*ptB));
#endif
        mpc_sub(*ptB, *ptB, dotp, SNAPLIB_MPC_RND);
      }
    }
  }
#ifdef SNAPLIB_DEBUG
  printf("result of vecB (%p)\n",vecB);
  show_mpcarray1d(vecB,rows,incb);
#endif
  mpc_clear(dotp);
  return 0;
}

/*!
 * @brief The function applies an elementary reflector to a general rectangular matrix.
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for side='L'.
 * @param[in] side 'L' if H*C, or 'R' if C*H.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of cols in A.
 * @param[in] vecV The vector v in the representation of H.
 * @param[in] tau The value tau in the representation of H.
 * @param[in,out] matA The array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in] work The allocated array of length cols if side='L', rows if side='R'.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_frlarfx(const char side, const unsigned int rows, const unsigned int cols, mpfr_t *vecV, mpfr_t tau, mpfr_t *matA, const int ldA, mpfr_t *work) {
  unsigned int i,j;
  mpfr_t *ptV, *ptW, *ptA, tmp;
#ifdef SNAPLIB_DEBUG
  printf("snap_mlapack_frlarfx called: side(%c), rows(%d), cols(%d), ldA(%d).\n",side,rows,cols,ldA);
  mpfr_printf("==> vecV(pt=%p),tau(%Re),matA(pt=%p),work(pt=%p).\n", vecV,tau,matA,work);
#endif
  if ( cols == 0 || rows == 0 ) {
    return 0;
  }
  // only implemented for few cases.
  if ( side != 'L' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_frlarfx met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
  // initialization
  mpfr_init(tmp);
  // tau*vecV*matA ==> work
  for ( i=0,ptW=work; i<cols; i++,ptW++ ) {
    snap_mblas_frdot(*ptW, rows, vecV, 1, matA+i, ldA);
    mpfr_mul(*ptW, *ptW, tau, SNAPLIB_MPFR_RND);
  }
  // matA-vecV*work ==> matA
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    for ( j=0,ptA=matA+i*ldA,ptW=work; j<cols; j++,ptA++,ptW++ ) {
      mpfr_mul(tmp, *ptV, *ptW, SNAPLIB_MPFR_RND);
      mpfr_sub(*ptA, *ptA, tmp, SNAPLIB_MPFR_RND);
    }
  }
  mpfr_clear(tmp);
  return 0;
}

/*!
 * @brief The function applies an elementary reflector to a general rectangular matrix.
 * @warning A must be a complex matrix of the row major.
 * @warning Only implemented for side='L'.
 * @param[in] side 'L' if H*C, or 'R' if C*H.
 * @param[in] rows The number of rows in A.
 * @param[in] cols The number of cols in A.
 * @param[in] vecV The vector v in the representation of H.
 * @param[in] tau The value tau in the representation of H.
 * @param[in,out] matA The complex array A.
 * @param[in] ldA The leading dimension of A (column size of A).
 * @param[in] work The allocated array of length cols if side='L', rows if side='R'.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fclarfx(const char side, const unsigned int rows, const unsigned int cols, mpc_t *vecV, mpc_t tau, mpc_t *matA, const int ldA, mpc_t *work) {
  unsigned int i,j;
  mpc_t *ptV, *ptW, *ptA, tmp;
#ifdef SNAPLIB_DEBUG
  printf("snap_mlapack_fclarfx called: side(%c), rows(%d), cols(%d), ldA(%d).\n",side,rows,cols,ldA);
  mpfr_printf("==> vecV(pt=%p),tau(%Re),matA(pt=%p),work(pt=%p).\n", vecV,mpc_realref(tau),matA,work);
#endif
  if ( cols == 0 || rows == 0 ) {
    return 0;
  }
  // only implemented for few cases.
  if ( side != 'L' ) {
#ifdef SNAPLIB_DEBUG
    printf("snap_mlapack_fclarfx met an unimplemented combination of parameters.\n");
#endif
    return -1;
  }
  // initialization
  mpc_init2(tmp, mpfr_get_default_prec());
  // tau*vecV*matA ==> work
  for ( i=0,ptW=work; i<cols; i++,ptW++ ) {
    snap_mblas_fcdotc(*ptW, rows, vecV, 1, matA+i, ldA);
    mpc_mul(*ptW, *ptW, tau, SNAPLIB_MPC_RND);
  }
  // matA-vecV*work ==> matA
  for ( i=0,ptV=vecV; i<rows; i++,ptV++ ) {
    for ( j=0,ptA=matA+i*ldA,ptW=work; j<cols; j++,ptA++,ptW++ ) {
      mpc_mul(tmp, *ptV, *ptW, SNAPLIB_MPC_RND);
      mpc_sub(*ptA, *ptA, tmp, SNAPLIB_MPC_RND);
    }
  }
  mpc_clear(tmp);
  return 0;
}

/*!
 * @brief The function generates an elementary reflector (Householder matrix).
 * @warning A must be a matrix of the row major.
 * @warning Only implemented for side='L'.
 * @param[in] len The order of the elementary reflector.
 * @param[in,out] alpha The first element of X on input and beta on exit.
 * @param[in,out] vecX The vector X(except the first element) on input and reflector on exit.
 * @param[in] incX The increment of vector X.
 * @param[out] tau The value tau on exit.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_frlarfg(const unsigned int len, mpfr_t alpha, mpfr_t *vecX, const int incX, mpfr_t tau) {
  unsigned int i;
  mpfr_t *ptX,tmp;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_frlarfg called: len(%d), alpha(%Re), vecX(pt=%p), incX(%d), tau(%Re).\n",len,alpha,vecX,incX,tau);
#endif
  snap_mblas_frdot(tau, len-1, vecX, incX, vecX, incX);
  if ( mpfr_sgn(tau) == 0 ) {
    mpfr_set_ui(alpha, 0, SNAPLIB_MPFR_RND);
  } else {
    mpfr_init(tmp);
    // mu
    mpfr_mul(tmp, alpha, alpha, SNAPLIB_MPFR_RND);
    mpfr_add(tmp, tmp, tau, SNAPLIB_MPFR_RND);
    mpfr_sqrt(tmp, tmp, SNAPLIB_MPFR_RND);
    if ( mpfr_sgn(alpha) <= 0 ) {
      mpfr_sub(alpha, alpha, tmp, SNAPLIB_MPFR_RND);
    } else {
      mpfr_add(alpha, alpha, tmp, SNAPLIB_MPFR_RND);
      mpfr_div(alpha, tau, alpha, SNAPLIB_MPFR_RND);
      mpfr_neg(alpha, alpha, SNAPLIB_MPFR_RND);
    }
    for ( i=1,ptX=vecX; i<len; i++,ptX++ ) {
      mpfr_div(*ptX, *ptX, alpha, SNAPLIB_MPFR_RND);
    }
    mpfr_swap(tmp, alpha);
    mpfr_mul(tmp, tmp, tmp, SNAPLIB_MPFR_RND);
    mpfr_add(tau, tau, tmp, SNAPLIB_MPFR_RND);
    mpfr_div(tau, tmp, tau, SNAPLIB_MPFR_RND);
    mpfr_mul_ui(tau, tau, 2, SNAPLIB_MPFR_RND);
    mpfr_clear(tmp);
  }
  return 0;
}

/*!
 * @brief The function generates an elementary reflector (Householder matrix) based on the algorithm used in ATLAS.
 * @warning A must be a complex matrix of the row major.
 * @warning Only implemented for side='L'.
 * @param[in] len The order of the elementary reflector.
 * @param[in,out] alpha The first element of X on input and beta on exit.
 * @param[in,out] vecX The vector X(except the first element) on input and reflector on exit.
 * @param[in] incX The increment of vector X.
 * @param[out] tau The value tau on exit.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fclarfg(const unsigned int len, mpc_t alpha, mpc_t *vecX, const int incX, mpc_t tau) {
  unsigned int i;
  mpc_t *ptX,tmp;
#ifdef SNAPLIB_DEBUG
  mpfr_printf("snap_mlapack_fclarfg called: len(%d), alpha(%Re%+ReI), vecX(pt=%p), incX(%d), tau(%Re%+ReI).\n",len,mpc_realref(alpha),mpc_imagref(alpha),vecX,incX,mpc_realref(tau),mpc_imagref(tau));
#endif
  snap_mblas_fcdotc(tau, len-1, vecX, incX, vecX, incX);
  if ( mpfr_sgn(mpc_realref(tau)) == 0 && mpfr_sgn(mpc_imagref(tau)) == 0 && mpfr_sgn(mpc_imagref(alpha)) == 0 ) {
    mpc_set_ui(alpha, 0, SNAPLIB_MPC_RND);
  } else {
    // tau <= (x_{2,...,n})^H (x_{2,...n})
    mpc_init2(tmp, mpfr_get_default_prec());
    // tmp <= x_1^H x_1 = alpha^H alpha
    mpc_set_ui(tmp, 0, SNAPLIB_MPC_RND);
    mpc_norm(mpc_realref(tmp), alpha, MPC_RND_RE(SNAPLIB_MPC_RND));
    // tmp <= tmp + tau = x^H x
    mpc_add(tmp, tmp, tau, SNAPLIB_MPC_RND);
    // tmp <= sqrt(tmp) = norm(x)
    mpc_sqrt(tmp, tmp, SNAPLIB_MPC_RND);
    if ( mpfr_sgn(mpc_realref(alpha)) > 0 ) {
      mpc_neg(tmp, tmp, SNAPLIB_MPC_RND);
    }
    mpc_sub(tau, tmp, alpha, SNAPLIB_MPC_RND);
    mpc_div(tau, tau, tmp, SNAPLIB_MPC_RND);
    // zlarfg in ATLAS takes the conjugate of tau at the output. (NEED TO VERIFY) <= should be a buf of ATLAS.
    // mpc_conj(tau, tau, SNAPLIB_MPC_RND); // this is not required.
    mpc_sub(alpha, alpha, tmp, SNAPLIB_MPC_RND);
    // div and mul, or just div
    // mpc_ui_div(alpha, 1, alpha, SNAPLIB_MPC_RND);
    for ( i=1,ptX=vecX; i<len; i++,ptX++ ) {
      mpc_div(*ptX, *ptX, alpha, SNAPLIB_MPC_RND);
    }
    mpc_set(alpha, tmp, SNAPLIB_MPC_RND);
    mpc_clear(tmp);
  }
  return 0;
}

/*!
 * @brief The function generates an M-by-N complex matrix Q with orthonormal columns.
 * @warning A must be a complex matrix of the row major.
 * @param[in] rows The number of rows of A.
 * @param[in] cols The number of colums of A.
 * @param[in] num The number of elementary reflectors.
 * @param[in,out] matA The matrix A on input and matrix Q on exit.
 * @param[in] ldA The leading dimension of matrix A.
 * @param[in] tau The value tau.
 * @return 0 if success or -1 if failure.
 */
int snap_mlapack_fcungqr(const unsigned int rows, const unsigned int cols, const unsigned int num, mpc_t *matA, const int ldA, mpc_t *tau) {
  // see eq. 5.1.5, Matrix Computations
  unsigned int i, j, k;
  mpc_t *pt1, *vec, *pt2, tmp;
  mpc_init_set_ui(tmp, 0, SNAPLIB_MPC_RND);
  vec = (mpc_t *) malloc(sizeof(mpc_t)*rows);
  for ( i = 0, pt1 = vec; i < rows; i++, pt1++ ) {
    mpc_init_set_ui(*pt1, 0, SNAPLIB_MPC_RND);
  }
  j = cols - 1;
  while ( j >= num ) {
    for ( i = 0, pt1 = matA+j; i < rows; i++, pt1+=ldA ) {
      if ( i == j ) {
        mpc_set_si(*pt1, 1, SNAPLIB_MPC_RND);
      } else {
        mpc_set_si(*pt1, 0, SNAPLIB_MPC_RND);
      }
    }
  }
  do {
    // take v(j:m)
    mpc_set_si(*(vec+j), 1, SNAPLIB_MPC_RND);
    for ( i = j+1, pt1 = vec+j+1, pt2= matA+j+(j+1)*ldA; i < rows; i++, pt1++, pt2+=ldA ) {
      mpc_set(*pt1, *pt2, SNAPLIB_MPC_RND);
    }
    // init identity vector A(:,j)
    for ( i = 0, pt1 = matA+j; i < rows; i++, pt1+=ldA ) {
      if ( i == j ) {
        mpc_set_si(*pt1, 1, SNAPLIB_MPC_RND);
      } else {
        mpc_set_si(*pt1, 0, SNAPLIB_MPC_RND);
      }
    }
    // update
    for ( k = j; k < num; k++ ) {
      snap_mblas_fcdotc(tmp, rows-j, vec+j, 1, matA+k+j*ldA, ldA);
      mpc_mul(tmp, tmp, *(tau+j), SNAPLIB_MPC_RND);
      mpc_neg(tmp, tmp, SNAPLIB_MPC_RND);
      snap_mblas_fcaxpy(rows-j, tmp, vec+j, 1, matA+k+j*ldA, ldA);
    }
  } while ( j-- > 0 );
  for ( i = 0, pt1 = vec; i < rows; i++, pt1++ ) {
    mpc_clear(*pt1);
  }
  mpc_clear(tmp);
  return 0;
}
