/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <snap.h>

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

// =========================================
// ================ SDMZ_T =================
// =========================================

/*!
 *  @brief Initialize sdm*_t variable (allocate its elements array).
 *  @param[in,out] matrix The sdm*_t variable which shoule be initialized.
 */
void SNAPMTX_init(SNAPMTX_t matrix) {
  SNAPMTX_rows(matrix) = 1;
  SNAPMTX_rsize(matrix) = 1;
  SNAPMTX_cols(matrix) = 1;
  SNAPMTX_csize(matrix) = 1;
  SNAPMTX_elems(matrix) = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t));
  SNAPBASE_init_set_ui(SNAPMTX_elem(matrix,0,0),0);
}

/*!
 *  @brief Initialize sdm*_t variable (allocate its specified elements array and initialize).
 *  @param[in,out] matrix The sdm*_t variable which shoule be initialized.
 *  @param[in] rows The number of rows that should be allocated.
 *  @param[in] cols The number of cols that should be allocated.
 */
void SNAPMTX_init2(SNAPMTX_t matrix, unsigned int rows, unsigned int cols) {
  unsigned int i,j;
  SNAPMTX_rows(matrix) = rows;
  SNAPMTX_rsize(matrix) = rows;
  SNAPMTX_cols(matrix) = cols;
  SNAPMTX_csize(matrix) = cols;
  SNAPMTX_elems(matrix) = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*rows*cols);
  for ( i=0; i<rows; i++ ) {
    for ( j=0; j<cols; j++ ) {
      SNAPBASE_init_set_ui(SNAPMTX_elem(matrix,i,j),0);
    }
  }
}

/*!
 *  @brief Re-initialize sdm*_t variable (allocate its specified elements array and initialize all).
 *  @param[in,out] matrix The sdm*_t variable which should be re-allocated.
 *  @param[in] rows The number of rows that should be re-allocated.
 *  @param[in] cols The number of cols that should be re-allocated.
 */
void SNAPMTX_reinit2(SNAPMTX_t matrix, unsigned int rows, unsigned int cols) {
  unsigned int i, old_rows, old_cols;
  old_rows = SNAPMTX_rsize(matrix);
  old_cols = SNAPMTX_csize(matrix);
  if ( rows*cols <= old_rows*old_cols ) {
    for ( i=0; i<rows*cols; i++ ) {
      SNAPBASE_set_ui(*(SNAPMTX_elems(matrix)+i), 0);
    }
    for ( i=rows*cols; i<old_rows*old_cols; i++ ) {
      SNAPBASE_clear(*(SNAPMTX_elems(matrix)+i));
    }
  }
  SNAPMTX_rows(matrix) = rows;
  SNAPMTX_rsize(matrix) = rows;
  SNAPMTX_cols(matrix) = cols;
  SNAPMTX_csize(matrix) = cols;
  if ( rows*cols == old_cols*old_rows ) return;
  SNAPMTX_elems(matrix) = (SNAPBASE_t*) realloc(SNAPMTX_elems(matrix), sizeof(SNAPBASE_t)*rows*cols);
  if ( rows*cols > old_rows*old_cols ) {
    for ( i=0; i<old_rows*old_cols; i++ ) {
      SNAPBASE_set_ui(*(SNAPMTX_elems(matrix)+i),0);
    }
    for ( i=old_rows*old_cols; i<rows*cols; i++ ) {
      SNAPBASE_init_set_ui(*(SNAPMTX_elems(matrix)+i),0);
    }
  }
}

/*!
 *  @brief Re-initialize sdm*_t variable (allocate its specified elements array and initialize all).
 *  @warn This functions does not shrink the memory.
 *  @param[in,out] matrix The sdm*_t variable which should be re-allocated.
 *  @param[in] rows The number of rows that should be re-allocated.
 *  @param[in] cols The number of cols that should be re-allocated.
 */
void SNAPMTX_reinit3(SNAPMTX_t matrix, unsigned int rows, unsigned int cols) {
  unsigned int i, j, old_rows, old_cols;
  SNAPBASE_t *pt;
  old_rows = SNAPMTX_rsize(matrix);
  old_cols = SNAPMTX_csize(matrix);
  if ( rows <= old_rows && cols <= old_cols ) {
    for ( i=0; i<rows; i++ ) {
      for ( j=0,pt=SNAPMTX_elemp(matrix,i,0); j<cols; j++,pt++ ) {
        SNAPBASE_set_ui(*pt, 0);
      }
    }
  }
  SNAPMTX_rows(matrix) = rows;
  SNAPMTX_cols(matrix) = cols;
  if ( rows <= old_rows && cols <= old_cols ) return;
  SNAPMTX_elems(matrix) = (SNAPBASE_t*) realloc(SNAPMTX_elems(matrix), sizeof(SNAPBASE_t)*rows*cols);
  SNAPMTX_rsize(matrix) = rows;
  SNAPMTX_csize(matrix) = cols;
  if ( rows*cols > old_rows*old_cols ) {
    for ( i=0,pt=SNAPMTX_elems(matrix); i<old_rows*old_cols; i++,pt++ ) {
      SNAPBASE_set_ui(*pt,0);
    }
    for ( ; i<rows*cols; i++,pt++ ) {
      SNAPBASE_init_set_ui(*pt,0);
    }
  } else {
    for ( i=0,pt=SNAPMTX_elems(matrix); i<rows*cols; i++,pt++ ) {
      SNAPBASE_set_ui(*pt,0);
    }
  }
}

/*!
 *  @brief Re-allocate sdm*_t variable (allocate its specified elements array and initialize only the allocated part).
 *  @warning NOTE THAT THIS FUNCTION DOES NOT RE-ORDER ANY ELEMENT!!
 *  @param[in,out] matrix The sdm*_t variable which should be re-allocated.
 *  @param[in] rows The number of rows that should be re-allocated.
 *  @param[in] cols The number of cols that should be re-allocated.
 */
void SNAPMTX_realloc(SNAPMTX_t matrix, unsigned int rows, unsigned int cols) {
  unsigned int i;
  if ( rows*cols == SNAPMTX_rsize(matrix)*SNAPMTX_csize(matrix) ) {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_rsize(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
    SNAPMTX_csize(matrix) = cols;
    return;
  }
  if ( rows*cols < SNAPMTX_rsize(matrix)*SNAPMTX_csize(matrix) ) {
    for ( i=rows*cols; i<SNAPMTX_rsize(matrix)*SNAPMTX_csize(matrix); i++ ) {
      SNAPBASE_clear(*(SNAPMTX_elems(matrix)+i));
    }
  }
  SNAPMTX_elems(matrix) = (SNAPBASE_t*) realloc(SNAPMTX_elems(matrix), sizeof(SNAPBASE_t)*rows*cols);
  if ( rows*cols > SNAPMTX_rsize(matrix)*SNAPMTX_csize(matrix) ) {
    for ( i=SNAPMTX_rsize(matrix)*SNAPMTX_csize(matrix); i<rows*cols; i++ ) {
      SNAPBASE_init_set_ui(*(SNAPMTX_elems(matrix)+i),0);
    }
  }
  SNAPMTX_rows(matrix) = rows;
  SNAPMTX_rsize(matrix) = rows;
  SNAPMTX_cols(matrix) = cols;
  SNAPMTX_csize(matrix) = cols;
}

/*!
 *  @brief Clear sdm*_t variable (free its elements array).
 *  @param[in,out] matrix The sdm*_t variable which should be cleared.
 */
void SNAPMTX_clear(SNAPMTX_t matrix) {
  unsigned int i,j;
  for ( i=0; i<SNAPMTX_rsize(matrix); i++ ) {
    for ( j=0; j<SNAPMTX_csize(matrix); j++ ) {
      SNAPBASE_clear(SNAPMTX_elem(matrix,i,j));
    }
  }
  free(SNAPMTX_elems(matrix));
}

/*!
 *  @brief Swap sdm*_t variable (just replace members).
 *  @param[in,out] matrixA The sdm*_t variable which should be swapped.
 *  @param[in,out] matrixB The sdm*_t variable which should be swapped.
 */
void SNAPMTX_swap(SNAPMTX_t matrixA, SNAPMTX_t matrixB) {
  SNAPBASE_t *pt;
  unsigned int tmp;
  pt = SNAPMTX_elems(matrixB);
  SNAPMTX_elems(matrixB) = SNAPMTX_elems(matrixA);
  SNAPMTX_elems(matrixA) = pt;
  tmp = SNAPMTX_rows(matrixB);
  SNAPMTX_rows(matrixB) = SNAPMTX_rows(matrixA);
  SNAPMTX_rows(matrixA) = tmp;
  tmp = SNAPMTX_cols(matrixB);
  SNAPMTX_cols(matrixB) = SNAPMTX_cols(matrixA);
  SNAPMTX_cols(matrixA) = tmp;
  tmp = SNAPMTX_rsize(matrixB);
  SNAPMTX_rsize(matrixB) = SNAPMTX_rsize(matrixA);
  SNAPMTX_rsize(matrixA) = tmp;
  tmp = SNAPMTX_csize(matrixB);
  SNAPMTX_csize(matrixB) = SNAPMTX_csize(matrixA);
  SNAPMTX_csize(matrixA) = tmp;
}

/*!
 *  @brief Subresultant mapping matrix w.r.t. f and g of order r (Sylvester matrix is r=0).
 *  @warning NOTE THAT*1 each col represents a coefficient list of a given poly.
 *  @warning NOTE THAT*2 coefficient vectors are ascending order!! (supX_t is ascending order!!)
 *  @warning NOTE THAT*3 this function re-allocate the variable matrix if required.
 *  @warning NOTE THAT*4 this function does NOT zero-fill the elements other than coefficients.
 *  @param[in,out] matrix The sdm*_t variable which will be the subresultant matrix.
 *  @param[in] poly_f This function returns Syl_r(poly_f,poly_g).
 *  @param[in] poly_g This function returns Syl_r(poly_f,poly_g).
 *  @param[in] r The order of Subresultant mapping.
 *  @param[in] mode The rounding mode for multi-precision floating numbers.
 */
void SNAPMTX_subresultant_mapping_matrix_asc(SNAPMTX_t matrix, SNAP_t poly_f, SNAP_t poly_g, unsigned int r) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  cols = SNAP_degree(poly_f) + SNAP_degree(poly_g) - 2*r;
  rows = cols + r;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly_f
  for ( i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_f); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,j+i,i), SNAP_coef(poly_f,j));
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_g); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,j+i,i+SNAP_degree(poly_g)-r), SNAP_coef(poly_g,j));
    }
  }
}
void SNAPMTX_subresultant_mapping_matrix_dsc(SNAPMTX_t matrix, SNAP_t poly_f, SNAP_t poly_g, unsigned int r) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  cols = SNAP_degree(poly_f) + SNAP_degree(poly_g) - 2*r;
  rows = cols + r;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly_f
  for ( i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_f); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,j+i,i), SNAP_coef(poly_f,SNAP_degree(poly_f)-j));
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_g); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,j+i,i+SNAP_degree(poly_g)-r), SNAP_coef(poly_g,SNAP_degree(poly_g)-j));
    }
  }
}

/*!
 *  @brief Subresultant mapping matrix^T w.r.t. f and g of order r (Sylvester matrix is r=0).
 *  @warning NOTE THAT*1 each row represents a coefficient list of a given poly.
 *  @warning NOTE THAT*2 coefficient vectors are descending order!! (supX_t is ascending order!!)
 *  @warning NOTE THAT*3 this function re-allocate the variable matrix if required.
 *  @warning NOTE THAT*4 this function does NOT zero-fill the elements other than coefficients.
 *  @param[in,out] matrix The sdm*_t variable which will be the subresultant matrix.
 *  @param[in] poly_f This function returns Syl_r(poly_f,poly_g).
 *  @param[in] poly_g This function returns Syl_r(poly_f,poly_g).
 *  @param[in] r The order of Subresultant mapping.
 *  @param[in] mode The rounding mode for multi-precision floating numbers.
 */
void SNAPMTX_subresultant_mapping_matrix_transpose(SNAPMTX_t matrix, SNAP_t poly_f, SNAP_t poly_g, unsigned int r) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly_f) + SNAP_degree(poly_g) - 2*r;
  cols = rows + r;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly_f
  for ( i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_f); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,i,j+i), SNAP_coef(poly_f,SNAP_degree(poly_f)-j));
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_g); j++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,i+SNAP_degree(poly_g)-r,j+i), SNAP_coef(poly_g,SNAP_degree(poly_g)-j));
    }
  }
}
void SNAPMTX_subresultant_mapping_matrix_transpose_column_major(SNAPMTX_t matrix, SNAP_t poly_f, SNAP_t poly_g, unsigned int r) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly_f) + SNAP_degree(poly_g) - 2*r;
  cols = rows + r;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly_f
  for ( i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_f); j++ ) {
      SNAPBASE_set(SNAPMTX_elem_column_major(matrix,i,j+i), SNAP_coef(poly_f,SNAP_degree(poly_f)-j));
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    for ( j=0; j<=SNAP_degree(poly_g); j++ ) {
      SNAPBASE_set(SNAPMTX_elem_column_major(matrix,i+SNAP_degree(poly_g)-r,j+i), SNAP_coef(poly_g,SNAP_degree(poly_g)-j));
    }
  }
}

/*!
 *  @brief Convolution matrix C(f) s.t. C(f)\vec{g} = \vec{f*g} where \deg{g}=m.
 *  @warning NOTE THAT*1 this function re-allocate the variable matrix if required.
 *  @warning NOTE THAT*2 this function does NOT zero-fill the elements other than coefficients.
 *  @param[in,out] matrix The sdm*_t variable which will be the convolution matrix.
 *  @param[in] poly This function returns C(poly).
 *  @param[in] m The order of convolution matrix.
 *  @param[in] mode The rounding mode for multi-precision floating numbers.
 */
void SNAPMTX_convolution_matrix(SNAPMTX_t matrix, SNAP_t poly, unsigned int m) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly) + m + 1;
  cols = m + 1;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly
  for ( j=0; j<=m; j++ ) {
    for ( i=0; i<=SNAP_degree(poly); i++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,i+j,j), SNAP_coef(poly,i));
    }
  }
}
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
void SNAPMTX_convolution_matrix_conj(SNAPMTX_t matrix, SNAP_t poly, unsigned int m) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly) + m + 1;
  cols = m + 1;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly
  for ( j=0; j<=m; j++ ) {
    for ( i=0; i<=SNAP_degree(poly); i++ ) {
      SNAPBASE_conj(SNAPMTX_elem(matrix,i+j,j), SNAP_coef(poly,i));
    }
  }
}
#endif
void SNAPMTX_convolution_matrix_column_major(SNAPMTX_t matrix, SNAP_t poly, unsigned int m) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly) + m + 1;
  cols = m + 1;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly
  for ( j=0; j<=m; j++ ) {
    for ( i=0; i<=SNAP_degree(poly); i++ ) {
      SNAPBASE_set(SNAPMTX_elem_column_major(matrix,i+j,j), SNAP_coef(poly,i));
    }
  }
}
/*!
 *  @brief Convolution matrix C(f) s.t. C(f)\vec{g} = \vec{f*g} where \deg{g}=m.
 *  @warning NOTE THAT*1 this function re-allocate the variable matrix if required.
 *  @warning NOTE THAT*2 this function DOES zero-fill the elements other than coefficients.
 *  @param[in,out] matrix The sdm*_t variable which will be the convolution matrix.
 *  @param[in] poly This function returns C(poly).
 *  @param[in] m The order of convolution matrix.
 *  @param[in] mode The rounding mode for multi-precision floating numbers.
 */
void SNAPMTX_convolution_matrix_zerofill(SNAPMTX_t matrix, SNAP_t poly, unsigned int m) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly) + m + 1;
  cols = m + 1;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly
  for ( j=0; j<=m; j++ ) {
    for ( i=0; i<j; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix,i,j), 0);
    }
    for ( i=0; i<=SNAP_degree(poly); i++ ) {
      SNAPBASE_set(SNAPMTX_elem(matrix,i+j,j), SNAP_coef(poly,i));
    }
    for ( i=SNAP_degree(poly)+j+1; i<rows; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix,i,j), 0);
    }
  }
}
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
void SNAPMTX_convolution_matrix_zerofill_conj(SNAPMTX_t matrix, SNAP_t poly, unsigned int m) {
  unsigned int i, j, rows, cols;
  // the matrix size will be rows x cols.
  rows = SNAP_degree(poly) + m + 1;
  cols = m + 1;
  // re-initialize the matrix will be covered by the coefficients.
  if ( rows > SNAPMTX_rsize(matrix) || cols > SNAPMTX_csize(matrix) ) {
    SNAPMTX_realloc(matrix, rows, cols);
  } else {
    SNAPMTX_rows(matrix) = rows;
    SNAPMTX_cols(matrix) = cols;
  }
  // poly
  for ( j=0; j<=m; j++ ) {
    for ( i=0; i<j; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix,i,j), 0);
    }
    for ( i=0; i<=SNAP_degree(poly); i++ ) {
      SNAPBASE_conj(SNAPMTX_elem(matrix,i+j,j), SNAP_coef(poly,i));
    }
    for ( i=SNAP_degree(poly)+j+1; i<rows; i++ ) {
      SNAPBASE_set_ui(SNAPMTX_elem(matrix,i,j), 0);
    }
  }
}
#endif
