/*
 * Copyright (c) 2012-2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <snap.h>
#include <math.h>
#if (defined SNAP_SUPD_T || defined SNAP_SUPDC_T)
#include <fftw3.h>
#include <cblas.h>
#include <clapack.h>
#endif

#define SNAPjoin(a,b) a##b
#define SNAP2str(a) #a

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAPLIB_MAX(a,b) (((a)>(b)) ? (a):(b))
#define SNAPLIB_MIN(a,b) (((a)<(b)) ? (a):(b))

// #define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

/**************** HERE TO MODIFY with DOXYGEN *********************/

// initialize SNAP_t variable (allocate its coefficient array)
void SNAP_init(SNAP_t upoly) {
  SNAP_coefs(upoly) = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t));
  SNAPBASE_init_set_ui(SNAP_coef(upoly,0),0);
  SNAP_size(upoly) = 1;
  SNAP_degree(upoly) = 0;
}

// initialize SNAP_t variable (allocate its specified coefficient array)
void SNAP_init2(SNAP_t upoly, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *pt;
  SNAP_coefs(upoly) = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*len);
  for ( i=0, pt=SNAP_coefp(upoly,i); i<len; i++, pt++ ) {
    SNAPBASE_init_set_ui(*pt,0);
  }
  SNAP_size(upoly) = len;
  SNAP_degree(upoly) = 0;
}

// initialize SNAP_t variable rpoly and clone poly
void SNAP_init_set(SNAP_t rpoly, SNAP_t poly) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  SNAP_size(rpoly) = SNAP_degree(poly)+1;
  SNAP_degree(rpoly) = SNAP_degree(poly);
  SNAP_coefs(rpoly) = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(SNAP_size(rpoly)));
  for ( i=0, ptr=SNAP_coefp(rpoly,i), pt=SNAP_coefp(poly,i); i<SNAP_size(rpoly); i++, ptr++, pt++ ) {
    SNAPBASE_init_set(*ptr,*pt);
  }
}

// clear SNAP_t variable (free its coefficient array)
#if !(defined SNAP_SUPD_T || defined SNAP_SUPDC_T)
void SNAP_clear(SNAP_t upoly) {
  unsigned int i;
  SNAPBASE_t *pt;
  for ( i=0, pt=SNAP_coefp(upoly,i); i<SNAP_size(upoly); i++, pt++ ) {
    SNAPBASE_clear(*pt);
  }
  free(SNAP_coefs(upoly));
}
#endif

// override upoly1 with upoly2 (w/o size shrinking)
void SNAP_set(SNAP_t upoly1, SNAP_t upoly2) {
  unsigned int i;
  SNAPBASE_t *pt1, *pt2;
  if ( SNAP_size(upoly1) <= SNAP_degree(upoly2) ) {
    SNAP_coefs(upoly1) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly1), sizeof(SNAPBASE_t)*(SNAP_degree(upoly2)+1));
    for(i=SNAP_size(upoly1), pt1=SNAP_coefp(upoly1,i); i<=SNAP_degree(upoly2); i++,pt1++) {
      SNAPBASE_init(*pt1);
    }
    SNAP_size(upoly1) = SNAP_degree(upoly2)+1;
  }
  SNAP_degree(upoly1) = SNAP_degree(upoly2);
  for(i=0,pt1=SNAP_coefp(upoly1,i),pt2=SNAP_coefp(upoly2,i); i<=SNAP_degree(upoly2); i++,pt1++,pt2++) {
    SNAPBASE_set(*pt1, *pt2);
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking)
void SNAP_set_z(SNAP_t upoly, mpz_t *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  mpz_t *ptz;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptz=vec; i<len; i++,ptu++,ptz++) {
      SNAPBASE_set_z(*ptu, *ptz);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking)
void SNAP_set_d(SNAP_t upoly, double *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  double *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec; i<len; i++,ptu++,ptv++) {
      SNAPBASE_set_d(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking)
void SNAP_set_dc(SNAP_t upoly, dcomplex *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  dcomplex *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec; i<len; i++,ptu++,ptv++) {
      SNAPBASE_set_dc(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

void SNAP_set_fr(SNAP_t upoly, mpfr_t *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  mpfr_t *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec; i<len; i++,ptu++,ptv++) {
      SNAPBASE_set_fr(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

void SNAP_set_fc(SNAP_t upoly, mpc_t *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  mpc_t *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec; i<len; i++,ptu++,ptv++) {
      SNAPBASE_set_fc(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking) with reversal order.
void SNAP_set_rd(SNAP_t upoly, double *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  double *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec+len-i-1; i<len; i++,ptu++,ptv--) {
      SNAPBASE_set_d(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking) with reversal order.
void SNAP_set_rdc(SNAP_t upoly, dcomplex *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  dcomplex *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec+len-i-1; i<len; i++,ptu++,ptv--) {
      SNAPBASE_set_dc(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking) with reversal order.
void SNAP_set_rfr(SNAP_t upoly, mpfr_t *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  mpfr_t *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec+len-i-1; i<len; i++,ptu++,ptv--) {
      SNAPBASE_set_fr(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override upoly with SNAPBASE_t vector of dim()=len (w/o size shrinking) with reversal order.
void SNAP_set_rfc(SNAP_t upoly, mpc_t *vec, unsigned int len) {
  unsigned int i;
  SNAPBASE_t *ptu;
  mpc_t *ptv;
  if ( SNAP_size(upoly) < len ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*len);
    for(i=SNAP_size(upoly),ptu=SNAP_coefp(upoly,i); i<len; i++,ptu++) {
      SNAPBASE_init(*ptu);
    }
    SNAP_size(upoly) = len;
  }
  SNAP_degree(upoly) = 0;
  if ( len == 0 ) {
    SNAPBASE_set_ui(SNAP_coef(upoly,0),0);
  } else {
    for(i=0,ptu=SNAP_coefp(upoly,i),ptv=vec+len-i-1; i<len; i++,ptu++,ptv--) {
      SNAPBASE_set_fc(*ptu, *ptv);
      if ( !SNAPBASE_zeroq(*ptu) ) {
        SNAP_degree(upoly) = i;
      }
    }
  }
}

// override the coefficient of upoly with the given
void SNAP_set_coef(SNAP_t upoly, unsigned int n, SNAPBASE_t c) {
  unsigned int i;
  SNAPBASE_t *pt;
  if ( n == UINT_MAX ) {
    return;
  }
  if ( SNAP_size(upoly) <= n ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*(n+1));
    for(i=SNAP_size(upoly),pt=SNAP_coefp(upoly,i); i<=n; i++,pt++) {
      SNAPBASE_init(*pt);
    }
    SNAP_size(upoly) = n+1;
  }
  if ( SNAP_degree(upoly) < n ) {
    for(i=SNAP_degree(upoly)+1,pt=SNAP_coefp(upoly,i); i<n; i++,pt++) {
      SNAPBASE_set_ui(*pt,0);
    }
    SNAP_degree(upoly) = n;
  }
  SNAPBASE_set(SNAP_coef(upoly,n), c);
  if ( SNAP_degree(upoly) == n && n != 0 ) {
    for(i=0,pt=SNAP_coefp(upoly,n); i<=n; i++,pt--) {
      if ( !SNAPBASE_zeroq(*pt) ) {
        break;
      }
      SNAP_degree(upoly)--;
    }
  }
}

// override the coefficient of upoly with the given
void SNAP_set_coef_d(SNAP_t upoly, unsigned int n, double c) {
  unsigned int i;
  SNAPBASE_t *pt;
  if ( n == UINT_MAX ) {
    return;
  }
  if ( SNAP_size(upoly) <= n ) {
    SNAP_coefs(upoly) = (SNAPBASE_t*) realloc(SNAP_coefs(upoly), sizeof(SNAPBASE_t)*(n+1));
    for(i=SNAP_size(upoly),pt=SNAP_coefp(upoly,i); i<=n; i++,pt++) {
      SNAPBASE_init(*pt);
    }
    SNAP_size(upoly) = n+1;
  }
  if ( SNAP_degree(upoly) < n ) {
    for(i=SNAP_degree(upoly)+1,pt=SNAP_coefp(upoly,i); i<n; i++,pt++) {
      SNAPBASE_set_ui(*pt,0);
    }
    SNAP_degree(upoly) = n;
  }
  SNAPBASE_set_d(SNAP_coef(upoly,n), c);
  if ( SNAP_degree(upoly) == n && n != 0 ) {
    for(i=0,pt=SNAP_coefp(upoly,n); i<=n; i++,pt--) {
      if ( !SNAPBASE_zeroq(*pt) ) {
        break;
      }
      SNAP_degree(upoly)--;
    }
  }
}

// the following functions are for SNAP_set_str
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
static int SNAP_set_str_real(SNAPBASE_t *num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#else
static int SNAP_set_str_real(SNAPBASE_t num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#endif
{
  SNAPBASE_t local_num;
#ifdef SNAPLIB_DEBUG
  printf("str_real (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_REAL || (*current)->type == SNAP_SP_INT ) {
#if !(defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T))
    SNAPBASE_init(local_num);
#endif
    SNAPBASE_set_str(local_num, (*current)->token, 10);
    if ( sign == SNAP_SP_SUB ) {
      SNAPBASE_neg(local_num, local_num);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPBASE_add(*num, *num, local_num);
#else
    SNAPBASE_add(num, num, local_num);
    SNAPBASE_clear(local_num);
#endif
    (*current) = (*current)->next;
    return 0;
  } else {
    return -1;
  }
}
#if defined(SNAP_SUPFC_T)
static int SNAP_set_str_imag(SNAPBASE_t num, sp_lexical_token_struct **current, sp_lexical_token_type sign) {
  SNAPBASE_t local_num;
  SNAPBASE_init_set_ui(local_num,0);
#ifdef SNAPLIB_DEBUG
  printf("str_imag (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_J ) {
    if ( (*current)->next->type == SNAP_SP_MUL ) {
      (*current) = (*current)->next->next;
      if ( SNAP_set_str_real(local_num,current,sign) < 0 ) {
        goto err;
      }
    } else {
      (*current) = (*current)->next;
      if ( sign == SNAP_SP_ADD ) {
        SNAPBASE_set_ui(local_num,1);
      } else {
        SNAPBASE_set_si(local_num,-1);
      }
    }
    SNAPBASEREAL_swap(SNAPBASE_realref(local_num),SNAPBASE_imagref(local_num));
    SNAPBASE_add(num, num, local_num);
  } else {
    if ( SNAP_set_str_real(local_num,current,sign) < 0 ) {
      goto err;
    }
    if ( (*current)->type == SNAP_SP_J ) {
      (*current) = (*current)->next;
    } else if ( (*current)->type == SNAP_SP_MUL && (*current)->next->type == SNAP_SP_J ) {
      (*current) = (*current)->next->next;
    } else {
      goto err;
    }
    SNAPBASEREAL_swap(SNAPBASE_realref(local_num),SNAPBASE_imagref(local_num));
    SNAPBASE_add(num, num, local_num);
  }
  SNAPBASE_clear(local_num);
  return 0;
err:
  SNAPBASE_clear(local_num);
  return -1;
}
#elif defined(SNAP_SUPDC_T)
static int SNAP_set_str_imag(SNAPBASE_t *num, sp_lexical_token_struct **current, sp_lexical_token_type sign) {
  SNAPBASE_t local_num;
  SNAPBASE_init_set_ui(local_num,0);
#ifdef SNAPLIB_DEBUG
  printf("str_imag (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_J ) {
    if ( (*current)->next->type == SNAP_SP_MUL ) {
      (*current) = (*current)->next->next;
      if ( SNAP_set_str_real(&local_num,current,sign) < 0 ) {
        return -1;
      }
      *num += local_num*_Complex_I;
    } else {
      (*current) = (*current)->next;
      if ( sign == SNAP_SP_ADD ) {
        *num += _Complex_I;
      } else {
        *num -= _Complex_I;
      }
    }
  } else {
    if ( SNAP_set_str_real(&local_num,current,sign) < 0 ) {
      return -1;
    }
    if ( (*current)->type == SNAP_SP_J ) {
      (*current) = (*current)->next;
    } else if ( (*current)->type == SNAP_SP_MUL && (*current)->next->type == SNAP_SP_J ) {
      (*current) = (*current)->next->next;
    } else {
      return -1;
    }
    *num += local_num*_Complex_I;
  }
  return 0;
}
#endif
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
static int SNAP_set_str_ppnumber(SNAPBASE_t *num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#else
static int SNAP_set_str_ppnumber(SNAPBASE_t num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#endif
{
#ifdef SNAPLIB_DEBUG
  printf("str_ppnumber (%d)\n", (*current)->type);
#endif
  if ( ((*current)->type == SNAP_SP_REAL || (*current)->type == SNAP_SP_INT ) &&
       ((*current)->next->type == SNAP_SP_EOF || ((*current)->next->type != SNAP_SP_J && !((*current)->next->type == SNAP_SP_MUL && (*current)->next->next->type == SNAP_SP_J))) ) {
    if ( SNAP_set_str_real(num,current,sign) < 0 ) {
      return -1;
    }
  } else {
#if defined(SNAP_SUPDC_T) || defined(SNAP_SUPFC_T)
    if ( SNAP_set_str_imag(num,current,sign) < 0 ) {
      return -1;
    }
#else
    return -1;
#endif
  }
  return 0;
}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
static int SNAP_set_str_number(SNAPBASE_t *num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#else
static int SNAP_set_str_number(SNAPBASE_t num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#endif
{
  sp_lexical_token_type local_sign;
#ifdef SNAPLIB_DEBUG
  printf("str_number (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_ADD || (*current)->type == SNAP_SP_SUB ) {
    local_sign = (*current)->type;
    (*current) = (*current)->next;
  } else {
    local_sign = SNAP_SP_ADD;
  }
  if ( SNAP_set_str_ppnumber(num,current,local_sign) < 0 ) {
    return -1;
  }
  if ( (*current)->type == SNAP_SP_ADD || (*current)->type == SNAP_SP_SUB ) {
    local_sign = (*current)->type;
    (*current) = (*current)->next;
    if ( SNAP_set_str_ppnumber(num,current,local_sign) < 0 ) {
      return -1;
    }
  }
  if ( sign == SNAP_SP_SUB ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    SNAPBASE_neg(*num,*num);
#else
    SNAPBASE_neg(num,num);
#endif
  }
  return 0;
}
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
static int SNAP_set_str_num(SNAPBASE_t *num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#else
static int SNAP_set_str_num(SNAPBASE_t num, sp_lexical_token_struct **current, sp_lexical_token_type sign)
#endif
{
#ifdef SNAPLIB_DEBUG
  printf("str_num (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_LPAR ) {
    (*current) = (*current)->next;
    if ( SNAP_set_str_number(num,current,sign) < 0 ) {
      return -1;
    }
    (*current) = (*current)->next;
  } else {
    if ( SNAP_set_str_ppnumber(num,current,sign) < 0 ) {
      return -1;
    }
  }
  return 0;
}
static int SNAP_set_str_term(unsigned int *degree, sp_lexical_token_struct **current) {
#ifdef SNAPLIB_DEBUG
  printf("str_term (%d)\n", (*current)->type);
#endif
  if ( (*current)->next->type == SNAP_SP_POW ) {
    (*current) = (*current)->next;
    if ( (*current)->next->type == SNAP_SP_INT ) {
      (*current) = (*current)->next;
      *degree = strtoul((*current)->token, NULL, 10);
      (*current) = (*current)->next;
    } else {
      return -1;
    }
  } else {
    *degree = 1;
    (*current) = (*current)->next;
  }
  return 0;
}
static int SNAP_set_str_ppmonomial(SNAP_t upoly, sp_lexical_token_struct **current, sp_lexical_token_type sign) {
  unsigned int degree=0;
  SNAPBASE_t num;
  SNAPBASE_init_set_ui(num,0);
#ifdef SNAPLIB_DEBUG
  printf("str_ppmonomial (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_ID ) {
    if ( SNAP_set_str_term(&degree,current) < 0 ) {
      goto err;
    }
    if ( (*current)->type == SNAP_SP_MUL ) {
      (*current) = (*current)->next;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      if ( SNAP_set_str_num(&num,current,sign) < 0 )
#else
      if ( SNAP_set_str_num(num,current,sign) < 0 )
#endif
      {
        goto err;
      }
      SNAP_set_coef(upoly,degree,num);
    } else {
      if ( sign == SNAP_SP_ADD ) {
        SNAP_set_coef_d(upoly,degree,1);
      } else {
        SNAP_set_coef_d(upoly,degree,-1);
      }
    }
  } else {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( SNAP_set_str_num(&num,current,sign) < 0 )
#else
    if ( SNAP_set_str_num(num,current,sign) < 0 )
#endif
    {
      goto err;
    }
    if ( (*current)->type == SNAP_SP_ID ) {
      if ( SNAP_set_str_term(&degree,current) < 0 ) {
        goto err;
      }
      SNAP_set_coef(upoly,degree,num);
    } else if ( (*current)->type == SNAP_SP_MUL ) {
      (*current) = (*current)->next;
      if ( SNAP_set_str_term(&degree,current) < 0 ) {
        goto err;
      }
      SNAP_set_coef(upoly,degree,num);
    } else {
      SNAP_set_coef(upoly,0,num);
    }
  }
#if !(defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T))
  SNAPBASE_clear(num);
#endif
  return 0;
err:
#if !(defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T))
  SNAPBASE_clear(num);
#endif
  return -1;
}
static int SNAP_set_str_monomial(SNAP_t upoly, sp_lexical_token_struct **current) {
  sp_lexical_token_type sign;
#ifdef SNAPLIB_DEBUG
  printf("str_monomial (%d)\n", (*current)->type);
#endif
  sign = (*current)->type;
  if ( sign == SNAP_SP_ADD || sign == SNAP_SP_SUB ) {
    (*current) = (*current)->next;
    return SNAP_set_str_ppmonomial(upoly,current,sign);
  } else {
    return -1;
  }
}
static int SNAP_set_str_poly(SNAP_t upoly, sp_lexical_token_struct **current) {
#ifdef SNAPLIB_DEBUG
  printf("str_poly (%d)\n", (*current)->type);
#endif
  if ( SNAP_set_str_monomial(upoly,current) < 0 ) {
    return -1;
  } else {
    if ( (*current)->type == SNAP_SP_EOF ) {
      return 0;
    } else {
      return SNAP_set_str_poly(upoly,current);
    }
  }
}
static int SNAP_set_str_polyall(SNAP_t upoly, sp_lexical_token_struct **current) {
#ifdef SNAPLIB_DEBUG
  printf("str_polyall (%d)\n", (*current)->type);
#endif
  if ( (*current)->type == SNAP_SP_ADD || (*current)->type == SNAP_SP_SUB ) {
    return SNAP_set_str_poly(upoly,current);
  } else {
    if ( SNAP_set_str_ppmonomial(upoly,current,SNAP_SP_ADD) < 0 ) {
      return -1;
    } else {
      if ( (*current)->type == SNAP_SP_EOF ) {
        return 0;
      } else {
        return SNAP_set_str_poly(upoly,current);
      }
    }
  }
}

/*!
 * @brief override upoly with the given string (w/o size shrinking).
 * @warning The string must be in the expanded form (no duplicate terms).
 * @warning Any parenthesis can be used only for complex numbers.
 * @param[out] upoly The accepted polynomial.
 * @param[in] str The string will be analyzed.
 * @return The degree of accepted polynomial.
 */
unsigned int SNAP_set_str(SNAP_t upoly, char *str) {
  // polyall   : ppmonomial | ppmonomial poly | poly
  // poly      : monomial | monomial poly
  // monomial  : (SNAP_SP_ADD|SNAP_SP_SUB) ppmonomial
  // ppmonomial: num | num term | num SNAP_SP_MUL term | term | term SNAP_SP_MUL num
  // term      : SNAP_SP_ID | SNAP_SP_ID SNAP_SP_POW SNAP_SP_INT
  // num       : ppnumber | SNAP_SP_LPAR number SNAP_SP_RPAR
  // ppnumber  : real | imag
  // number    : ppnumber | (SNAP_SP_ADD|SNAP_SP_SUB) ppnumber |
  //             (SNAP_SP_ADD|SNAP_SP_SUB|none) imag (SNAP_SP_ADD|SNAP_SP_SUB) real | (SNAP_SP_ADD|SNAP_SP_SUB|none) real (SNAP_SP_ADD|SNAP_SP_SUB) imag
  // real      : SNAP_SP_REAL | SNAP_SP_INT
  // imag      : SNAP_SP_J | real SNAP_SP_J | real SNAP_SP_MUL SNAP_SP_J | SNAP_SP_J SNAP_SP_MUL real

  sp_lexical_token_t head;
  sp_lexical_token_struct *current;
  head->next = NULL;
  head->token = NULL;
  SNAP_set_coef_d(upoly,0,0);
  SNAP_degree(upoly) = 0;
  if ( sp_lexical_analysis_str(head, str) == 0 ) {
    current = head;
    if ( SNAP_set_str_polyall(upoly, &current) < 0 ) {
      SNAP_set_coef_d(upoly,0,0);
      SNAP_degree(upoly) = 0;
    }
  }
  sp_lexical_token_free(head);
  return SNAP_degree(upoly);
}

// swap the arguments
void SNAP_swap(SNAP_t p1, SNAP_t p2) {
  SNAPBASE_t *coef;
  unsigned int num;
  if ( p1 == p2 ) {
    return;
  }
  coef = SNAP_coefs(p1);
  SNAP_coefs(p1) = SNAP_coefs(p2);
  SNAP_coefs(p2) = coef;
  num = SNAP_degree(p1);
  SNAP_degree(p1) = SNAP_degree(p2);
  SNAP_degree(p2) = num;
  num = SNAP_size(p1);
  SNAP_size(p1) = SNAP_size(p2);
  SNAP_size(p2) = num;
}

// compute the i-norm and 2-norm of the given upoly.
#ifdef SNAP_SUPD_T
double supd_2norm(SNAP_t upoly) {
  double norm=0.0;
  unsigned int i;
  double *pt;
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    norm += (*pt)*(*pt);
  }
  return sqrt(norm);
}
double supd_inorm(SNAP_t upoly) {
  double norm=0.0;
  unsigned int i;
  double *pt;
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    if ( norm < fabs(*pt) ) {
      norm = fabs(*pt);
    }
  }
  return norm;
}
#endif
#ifdef SNAP_SUPDC_T
double supdc_2norm(SNAP_t upoly) {
  double norm=0.0;
  unsigned int i;
  dcomplex *pt;
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    norm += creal(*pt)*creal(*pt)+cimag(*pt)*cimag(*pt);
  }
  return sqrt(norm);
}
double supdc_inorm(SNAP_t upoly) {
  double norm=0.0;
  unsigned int i;
  dcomplex *pt;
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    if ( norm < cabs(*pt) ) {
      norm = cabs(*pt);
    }
  }
  return norm;
}
#endif
#ifdef SNAP_SUPFR_T
void supfr_2norm(SNAPBASE_t norm, SNAP_t upoly) {
  unsigned int i;
  SNAPBASE_t tmp, *pt;
  SNAPBASE_init(tmp);
  SNAPBASE_set_ui(norm, 0);
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    SNAPBASE_mul(tmp, *pt, *pt);
    SNAPBASE_add(norm, norm, tmp);
  }
  SNAPBASE_sqrt(norm, norm);
  SNAPBASE_clear(tmp);
}
#endif
#ifdef SNAP_SUPFC_T
void supfc_2norm(SNAPBASEREAL_t norm, SNAP_t upoly) {
  unsigned int i;
  SNAPBASEREAL_t tmp;
  SNAPBASE_t *pt;
  SNAPBASEREAL_init(tmp);
  SNAPBASEREAL_set_ui(norm, 0);
  for(i=0, pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    SNAPBASEREAL_mul(tmp, SNAPBASE_realref(*pt), SNAPBASE_realref(*pt));
    SNAPBASEREAL_add(norm, norm, tmp);
    SNAPBASEREAL_mul(tmp, SNAPBASE_imagref(*pt), SNAPBASE_imagref(*pt));
    SNAPBASEREAL_add(norm, norm, tmp);
  }
  SNAPBASEREAL_sqrt(norm, norm);
  SNAPBASEREAL_clear(tmp);
}
#endif
#ifdef SNAP_SUPZ_T
// compute the ceil of the 2-norm of the given upoly.
void supz_c2norm(mpz_t norm, supz_t upoly) {
  unsigned int i;
  mpz_t tmp;
  mpz_init(tmp);
  mpz_set_ui(norm,0);
  for(i=0; i<=supz_degree(upoly); i++) {
    mpz_addmul(norm, supz_coef(upoly,i), supz_coef(upoly,i));
  }
  mpz_sqrtrem(norm,tmp,norm);
  if ( mpz_sgn(tmp) > 0 ) {
    mpz_add_ui(norm,norm,1);
  }
  mpz_clear(tmp);
}
// compute the floor of the 2-norm of the given upoly.
void supz_f2norm(mpz_t norm, supz_t upoly) {
  unsigned int i;
  mpz_set_ui(norm,0);
  for(i=0; i<=supz_degree(upoly); i++) {
    mpz_addmul(norm, supz_coef(upoly,i), supz_coef(upoly,i));
  }
  mpz_sqrt(norm,norm);
}
#endif

// sign inversion.
void SNAP_self_neg(SNAP_t upoly) {
  unsigned int i;
  SNAPBASE_t *pt;
  for(i=0,pt=SNAP_coefs(upoly); i<=SNAP_degree(upoly); i++,pt++) {
    SNAPBASE_neg(*pt, *pt);
  }
}

// self reversal.
void SNAP_self_reversal(SNAP_t upoly) {
  unsigned int i;
  SNAPBASE_t *ptn, *ptr;
  for(i=0,ptn=SNAP_coefp(upoly,0),ptr=SNAP_lcoefp(upoly); 2*i<SNAP_degree(upoly); i++,ptn++,ptr--) {
    SNAPBASE_swap(*ptn, *ptr);
  }

}

// Multiply SNAP_t by SNAP_t and store it in the first arg.
static void SNAP_mul_traditional_nc(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  unsigned int i,j,k;
  SNAPBASE_t tmp, *ptr, *pt1, *pt2;
  // Degree Range Check
  k = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  if ( k < SNAP_degree(upoly1) || k == UINT_MAX ) {
    SNAP_clear(rpoly);
    SNAP_init(rpoly);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_mul), ": Degree Range Overflow.\n");
    return;
  }
  SNAPBASE_init(tmp);
  // Initialize the resulting polynomial store
  if ( SNAP_size(rpoly) <= k ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(k+1));
    for ( i = 0, ptr=SNAP_coefp(rpoly,0); i < SNAP_size(rpoly); i++,ptr++ ) {
      SNAPBASE_set_ui(*ptr, 0);
    }
    for( i = SNAP_size(rpoly), ptr=SNAP_coefp(rpoly,i); i<=k ; i++,ptr++ ) {
      SNAPBASE_init_set_ui(*ptr,0);
    }
    SNAP_size(rpoly) = k+1;
    SNAP_degree(rpoly) = k;
  } else {
    for ( i = 0,ptr=SNAP_coefp(rpoly,0); i <= k; i++,ptr++ ) {
      SNAPBASE_set_ui(*ptr, 0);
    }
    SNAP_degree(rpoly) = k;
  }
  for(i=0,pt1=SNAP_coefp(upoly1,0); i<=SNAP_degree(upoly1); i++,pt1++) {
    for(j=0,pt2=SNAP_coefp(upoly2,0),ptr=SNAP_coefp(rpoly,i+j); j<=SNAP_degree(upoly2); j++,pt2++,ptr++) {
      // mpfr_addmul(SNAP_coef(rpoly,i+j),SNAP_coef(upoly1,i),SNAP_coef(upoly2,j));
      SNAPBASE_mul(tmp, *pt1, *pt2);
      SNAPBASE_add(*ptr, *ptr, tmp);
    }
  }
  SNAPBASE_clear(tmp);
}

void SNAP_mul_traditional(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  SNAP_t tpoly;
  if ( rpoly == upoly1 || rpoly == upoly2 ) {
    SNAP_init(tpoly);
    SNAP_mul_traditional_nc(tpoly, upoly1, upoly2);
    SNAP_swap(tpoly, rpoly);
    SNAP_clear(tpoly);
  } else {
    SNAP_mul_traditional_nc(rpoly, upoly1, upoly2);
  }
}

// Multiply SNAP_t by SNAP_t and store it in the first arg by the Karatsuba Algorithm.
static void SNAP_mul_karatsuba_main(SNAPBASE_t *rcoefs, unsigned int fromr, SNAPBASE_t *tcoefs, unsigned int fromt, SNAPBASE_t *pcoefs, unsigned int fromp, SNAPBASE_t *ucoefs1, unsigned int from1, unsigned int to1, SNAPBASE_t *ucoefs2, unsigned int from2, unsigned int to2, unsigned int len, SNAPBASE_t addmultmp) {
  // upoly{1,2} may not have zero-coefs on terms of degrees between from{1,2} and to{1,2}.
  unsigned int newlen, i;
  SNAPBASE_t *ptr, *ptrp, *ptt, *pttp, *ptp, *ptpp, *pt1, *pt2, *pt1p, *pt2p;
#ifdef SNAPLIB_DEBUG
  printf("(%d), %d:%d,%d:%d\n", len, from1, to1, from2, to2);
//	printf("u1: ");
//	for(i=from1;i<=to1;i++) printf("%s, ", SNAPBASE_get_str(NULL,10,*(ucoefs1+i)));
//	printf("\n");
//	printf("u2: ");
//	for(i=from2;i<=to2;i++) printf("%s, ", SNAPBASE_get_str(NULL,10,*(ucoefs2+i)));
//	printf("\n");
#endif
  if ( from1 > to1 || from2 > to2 ) {
    return;
  }
  if ( len == 1 ) { // End of the recursions.
    // if ( from1 <= to1 && from2 <= to2 ) {
    // mpfr_addmul(*(rcoefs+fromr), *(ucoefs1+from1), *(ucoefs2+from2));
    SNAPBASE_mul(addmultmp, *(ucoefs1+from1), *(ucoefs2+from2));
    SNAPBASE_add(*(rcoefs+fromr), *(rcoefs+fromr), addmultmp);
    // }
  } else {
    // upoly1 = F1 x^newlen + F_0, upoly2 = G1 x^newlen + G0
    newlen = len/2;
    // F1 G1 x^len
    if ( from1+newlen <= to1 && from2+newlen <= to2 ) {
      SNAP_mul_karatsuba_main(rcoefs, fromr+len, tcoefs, fromt+len, pcoefs, fromp+len, ucoefs1, from1+newlen, to1, ucoefs2, from2+newlen, to2, newlen, addmultmp);
    }
    // F0 G0
    SNAP_mul_karatsuba_main(rcoefs, fromr, tcoefs, fromt+len, pcoefs, fromp+len, ucoefs1, from1, to1, ucoefs2, from2, to2, newlen, addmultmp);
    // Copy the lower half of F1G1 and the upper half of F0G0
    // the lower half of F1G1 => ppoly[fromp+newlen]..ppoly[fromp+len-1]
    // the upper half of F0G0 => ppoly[fromp]..ppoly[fromp+newlen-1]
    for ( i=0, ptp=pcoefs+fromp+newlen, ptpp=pcoefs+fromp, ptr=rcoefs+fromr+len, ptrp=rcoefs+fromr+newlen; i<newlen; i++,ptp++,ptpp++,ptr++,ptrp++ ) {
      SNAPBASE_set(*ptp, *ptr);
      SNAPBASE_set_ui(*ptr, 0);
      SNAPBASE_set(*ptpp, *ptrp);
      SNAPBASE_set_ui(*ptrp, 0);
    }
    // (F1 + F0)(G1 + G0) - F1 G1 - F0 G0
    // (F1+F0) => tpoly[fromt]..tpoly[fromt+newlen-1]
    // (G1+G0) => tpoly[fromt+newlen]..tpoly[fromt+len-1]
    for ( i=0,ptt=tcoefs+fromt,pt1=ucoefs1+from1,pt1p=ucoefs1+from1+newlen,pttp=tcoefs+fromt+newlen,pt2=ucoefs2+from2,pt2p=ucoefs2+from2+newlen; i<newlen; i++,ptt++,pt1++,pt1p++,pttp++,pt2++,pt2p++ ) {
      if ( from1+newlen+i <= to1 ) {
        SNAPBASE_add(*ptt, *pt1, *pt1p);
      } else if ( from1+i <= to1 ) {
        SNAPBASE_set(*ptt, *pt1);
      }
      if ( from2+newlen+i <= to2 ) {
        SNAPBASE_add(*pttp, *pt2, *pt2p);
      } else if ( from2+i <= to2 ) {
        SNAPBASE_set(*pttp, *pt2);
      }
    }
    // (F1+F0)(G1+G0) => rpoly[fromr+newlen]..rpoly[fromr+newlen+len-1]
    SNAP_mul_karatsuba_main(rcoefs, fromr+newlen, tcoefs, fromt+len, pcoefs, fromp+len, tcoefs, fromt, fromt+SNAPLIB_MIN(newlen-1,to1-from1), tcoefs, fromt+newlen, fromt+newlen+SNAPLIB_MIN(newlen-1,to2-from2), newlen, addmultmp);
    //  - F1 G1 - F0 G0 => rpoly[fromr+newlen]..rpoly[fromr+newlen+len-1]
    for ( i=0,ptr=rcoefs+fromr+newlen,ptrp=rcoefs+fromr+len,ptp=pcoefs+fromp,ptpp=pcoefs+fromp+newlen,pt1=rcoefs+fromr,pt2=rcoefs+fromr+newlen+len; i<newlen; i++,ptr++,ptrp++,ptp++,ptpp++,pt1++,pt2++ ) {
      SNAPBASE_add(*ptr, *ptr, *ptp);
      SNAPBASE_sub(*ptr, *ptr, *pt1);
      SNAPBASE_sub(*ptr, *ptr, *ptpp);
      SNAPBASE_add(*ptrp, *ptrp, *ptpp);
      SNAPBASE_sub(*ptrp, *ptrp, *pt2);
      SNAPBASE_sub(*ptrp, *ptrp, *ptp);
    }
  }
#ifdef SNAPLIB_DEBUG
//	printf("r: ");
//	for(i=0;i<len;i++) printf("%s, ", SNAPBASE_get_str(NULL,10,*(rcoefs+fromr+i)));
//	printf("\n");
#endif
}

static void SNAP_mul_karatsuba_nc(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2, SNAPBASE_t addmultmp) {
  unsigned int i,k,maxsize;
  SNAPBASE_t *tcoefs, *pcoefs, *pt1, *pt2;
  // Degree Range Check
  if ( SNAP_degree(upoly1) > SNAP_degree(upoly2) ) {
    maxsize = SNAP_degree(upoly1) + 1;
  } else {
    maxsize = SNAP_degree(upoly2) + 1;
  }
  i = maxsize;
  k = 2;
  while ( i != 0 ) {
    i >>= 1;
    k *= 2;
  }
  if ( k < maxsize || k == UINT_MAX ) {
    SNAP_clear(rpoly);
    SNAP_init(rpoly);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_mul_karatsuba), ": Degree Range Overflow.\n");
    return;
  }
  // Initialize the resulting polynomial store
  if ( SNAP_size(rpoly) < k ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(k));
    for ( i = 0, pt1=SNAP_coefp(rpoly,0); i < SNAP_size(rpoly); i++,pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
    for( i = SNAP_size(rpoly), pt1=SNAP_coefp(rpoly,i); i<k ; i++,pt1++ ) {
      SNAPBASE_init_set_ui(*pt1,0);
    }
    SNAP_size(rpoly) = k;
  } else {
    for ( i = 0,pt1=SNAP_coefp(rpoly,0); i < k; i++,pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
  }
  // Initialize the temporary polynomial store
  tcoefs = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(k));
  pcoefs = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(k));
  for( i = 0,pt1=tcoefs,pt2=pcoefs; i<k ; i++,pt1++,pt2++ ) {
    SNAPBASE_init_set_ui(*pt1,0);
    SNAPBASE_init_set_ui(*pt2,0);
  }
  // Call the Karatsuba algorithm recursively.
  SNAP_mul_karatsuba_main(SNAP_coefs(rpoly), 0, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), 0, SNAP_degree(upoly1), SNAP_coefs(upoly2), 0, SNAP_degree(upoly2), k/2, addmultmp);
  SNAP_degree(rpoly) = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  // Clear the temporary polynomial store
  for( i = 0,pt1=tcoefs,pt2=pcoefs; i<k ; i++,pt1++,pt2++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
  }
  free(tcoefs);
  free(pcoefs);
}

void SNAP_mul_karatsuba(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  SNAP_t tpoly;
  SNAPBASE_t addmultmp;
  SNAPBASE_init(addmultmp);
  if ( rpoly == upoly1 || rpoly == upoly2 ) {
    SNAP_init(tpoly);
    SNAP_mul_karatsuba_nc(tpoly, upoly1, upoly2, addmultmp);
    SNAP_swap(tpoly, rpoly);
    SNAP_clear(tpoly);
  } else {
    SNAP_mul_karatsuba_nc(rpoly, upoly1, upoly2, addmultmp);
  }
  SNAPBASE_clear(addmultmp);
}

// Multiply SNAP_t by SNAP_t and store it in the first arg by the Karatsuba Algorithm.
//#define KARATSUBA_BLOCK_THRESHOLD 1024
#if (defined SNAP_SUPD_T || defined SNAP_SUPDC_T)
#define KARATSUBA_BLOCK_THRESHOLD 2048
#elif (defined SNAP_SUPFR_T || defined SNAP_SUPFC_T)
#define KARATSUBA_BLOCK_THRESHOLD 64
#elif defined(SNAP_SUPZ_T)
#define KARATSUBA_BLOCK_THRESHOLD 1024
#endif
// #define KARATSUBA_BLOCK_THRESHOLD 4
static void SNAP_mul_karatsuba_block_nc(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  unsigned int i,j,k,bs,ibs,numbs,done;
  SNAPBASE_t *tcoefs, *pcoefs, *ccoefs, *pt1, *pt2, *pt3;
  SNAPBASE_t addmultmp;
  // Degree Range Check
  if ( SNAP_degree(upoly1) < SNAP_degree(upoly2) ) {
    SNAP_mul_karatsuba_block_nc(rpoly, upoly2, upoly1);
    return;
  }
  k = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  if ( k < SNAP_degree(upoly1) || k == UINT_MAX ) {
    SNAP_clear(rpoly);
    SNAP_init(rpoly);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_mul_karatsuba), ": Degree Range Overflow.\n");
    return;
  }
  // Below, we suppose that deg(p1)>=deg(p2).
  i = SNAP_degree(upoly2)+1;
  bs = 1;
  while ( i > 1 ) {
    i >>= 1;
    bs *= 2;
  }
  // Count the number of blocks in upoly1.
  numbs = (SNAP_degree(upoly1)+1)/bs;
  if ( numbs*bs < SNAP_degree(upoly1)+1 ) {
    numbs++;
  }
#ifdef SNAPLIB_DEBUG
  printf("block size=%d, number of blocks=%d\n",bs,numbs);
#endif
  // We prefer to use the traditional routine for smaller block sizes.
  if ( bs < KARATSUBA_BLOCK_THRESHOLD ) {
    SNAP_mul_traditional_nc(rpoly, upoly1, upoly2);
    return;
  }
  SNAPBASE_init(addmultmp);
  // Initialize the resulting polynomial store
  if ( SNAP_size(rpoly) < 2*numbs*bs ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(2*numbs*bs));
    for ( i = 0, pt1=SNAP_coefp(rpoly,i); i < SNAP_size(rpoly); i++,pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
    for( i = SNAP_size(rpoly),pt1=SNAP_coefp(rpoly,i); i<2*numbs*bs ; i++,pt1++ ) {
      SNAPBASE_init_set_ui(*pt1,0);
    }
    SNAP_size(rpoly) = 2*numbs*bs;
    SNAP_degree(rpoly) = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  } else {
    for ( i = 0, pt1=SNAP_coefp(rpoly,0); i < 2*numbs*bs; i++,pt1++ ) {
      SNAPBASE_set_ui(*pt1, 0);
    }
    SNAP_degree(rpoly) = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  }
  // Initialize the temporary polynomial store
  ibs = 2*bs;
  tcoefs = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(ibs));
  pcoefs = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(ibs));
  ccoefs = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*(ibs));
  for( i = 0,pt1=tcoefs,pt2=pcoefs,pt3=ccoefs; i<ibs ; i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_init_set_ui(*pt1,0);
    SNAPBASE_init_set_ui(*pt2,0);
    SNAPBASE_init_set_ui(*pt3,0);
  }
  // Low(upoly2)*Even Parts of Blocks(upoly1)
  for ( i=0; i<numbs; i+=2 ) {
    // supfr_mul_karatsuba_block_main(SNAP_coefs(rpoly), i*bs, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), 0, SNAP_degree(upoly2), bs);
    SNAP_mul_karatsuba_main(SNAP_coefs(rpoly), i*bs, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), 0, SNAP_degree(upoly2), bs, addmultmp);
  }
  // Low(upoly2)*Odd Parts of Blocks(upoly1)
  for ( i=1; i<numbs; i+=2 ) {
    for ( j=0,pt1=ccoefs,pt2=SNAP_coefp(rpoly,i*bs); j<2*bs&&i*bs+j<=SNAP_degree(rpoly); j++,pt1++,pt2++ ) {
      SNAPBASE_set_ui(*pt1, 0);
      SNAPBASE_swap(*pt1, *pt2);
    }
    // supfr_mul_karatsuba_block_main(SNAP_coefs(rpoly), i*bs, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), 0, SNAP_degree(upoly2), bs);
    SNAP_mul_karatsuba_main(SNAP_coefs(rpoly), i*bs, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), 0, SNAP_degree(upoly2), bs, addmultmp);
    for ( j=0,pt1=ccoefs,pt2=SNAP_coefp(rpoly,i*bs); j<2*bs&&i*bs+j<=SNAP_degree(rpoly); j++,pt1++,pt2++ ) {
      SNAPBASE_add(*pt2, *pt2, *pt1);
    }
  }
  done = bs;
  k = SNAP_degree(upoly2)+1-done;
  // Decrease the block size recusively
  while ( k >= KARATSUBA_BLOCK_THRESHOLD ) {
    i = k;
    bs = 1;
    while ( i > 1 ) {
      i >>= 1;
      bs *= 2;
    }
    // Count the number of blocks in upoly1.
    numbs = (SNAP_degree(upoly1)+1)/bs;
    if ( numbs*bs < SNAP_degree(upoly1)+1 ) {
      numbs++;
    }
#ifdef SNAPLIB_DEBUG
    printf("block size=%d, number of blocks=%d\n",bs,numbs);
#endif
    // Block(upoly2)*All Parts of Blocks(upoly1)
    for ( i=0; i<numbs; i++ ) {
      for ( j=0,pt1=ccoefs,pt2=SNAP_coefp(rpoly,i*bs+done); j<2*bs&&i*bs+j<=SNAP_degree(rpoly); j++,pt1++,pt2++ ) {
        SNAPBASE_set_ui(*pt1, 0);
        SNAPBASE_swap(*pt1, *pt2);
      }
      // supfr_mul_karatsuba_block_main(SNAP_coefs(rpoly), i*bs+done, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), done, SNAP_degree(upoly2), bs);
      SNAP_mul_karatsuba_main(SNAP_coefs(rpoly), i*bs+done, tcoefs, 0, pcoefs, 0, SNAP_coefs(upoly1), i*bs, SNAP_degree(upoly1), SNAP_coefs(upoly2), done, SNAP_degree(upoly2), bs, addmultmp);
      for ( j=0,pt1=ccoefs,pt2=SNAP_coefp(rpoly,i*bs+done); j<2*bs&&i*bs+j<=SNAP_degree(rpoly); j++,pt1++,pt2++ ) {
        SNAPBASE_add(*pt2, *pt2, *pt1);
      }
    }
    done += bs;
    k = SNAP_degree(upoly2)+1-done;
  }
  // Call the traditional routine for the rest small block.
  if ( k > 0 ) {
    for(i=0,pt1=SNAP_coefp(upoly1,i); i<=SNAP_degree(upoly1); i++,pt1++) {
      for(j=0,pt2=SNAP_coefp(upoly2,done),pt3=SNAP_coefp(rpoly,i+done); j<k; j++,pt2++,pt3++) {
        // mpfr_addmul(SNAP_coef(rpoly,i+j+done),SNAP_coef(upoly1,i),SNAP_coef(upoly2,j+done));
        SNAPBASE_mul(addmultmp, *pt1, *pt2);
        SNAPBASE_add(*pt3, *pt3, addmultmp);
      }
    }
  }
  // Set the correct degree
  // SNAP_degree(rpoly) = SNAP_degree(upoly1) + SNAP_degree(upoly2);
  // Clear the temporary polynomial store
  for( i = 0,pt1=tcoefs,pt2=pcoefs,pt3=ccoefs; i<ibs ; i++,pt1++,pt2++,pt3++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
    SNAPBASE_clear(*pt3);
  }
  free(tcoefs);
  free(pcoefs);
  free(ccoefs);
  SNAPBASE_clear(addmultmp);
}

void SNAP_mul_karatsuba_block(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  SNAP_t tpoly;
  if ( rpoly == upoly1 || rpoly == upoly2 ) {
    SNAP_init(tpoly);
    SNAP_mul_karatsuba_block_nc(tpoly, upoly1, upoly2);
    SNAP_swap(tpoly, rpoly);
    SNAP_clear(tpoly);
  } else {
    SNAP_mul_karatsuba_block_nc(rpoly, upoly1, upoly2);
  }
}

// rpoly = poly * multiplier. e.g. scalar multiplication.
void SNAP_mul_si(SNAP_t rpoly, SNAP_t poly, signed int multiplier) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_mul_si( *ptr, *pt, multiplier);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

// rpoly = poly * multiplier. e.g. scalar multiplication.
void SNAP_mul_d(SNAP_t rpoly, SNAP_t poly, double multiplier) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_mul_d( *ptr, *pt, multiplier);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

// rpoly = poly * multiplier. e.g. scalar multiplication.
void SNAP_mul_dc(SNAP_t rpoly, SNAP_t poly, dcomplex multiplier) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_mul_dc( *ptr, *pt, multiplier);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

void SNAP_mul_fr(SNAP_t rpoly, SNAP_t poly, mpfr_t multiplier) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_mul_fr( *ptr, *pt, multiplier);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

// Add the third SNAP_t from the second SNAP_t and store it in the first.
void SNAP_add(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  unsigned int i,deg;
  SNAPBASE_t *ptr, *pt1, *pt2;
  if ( SNAP_degree(upoly1) < SNAP_degree(upoly2) ) {
    if ( SNAP_size(rpoly) < SNAP_degree(upoly2) + 1 ) {
      SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(upoly2) + 1));
      for(i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(upoly2); i++,ptr++) {
        SNAPBASE_init(*ptr);
      }
      SNAP_size(rpoly) = SNAP_degree(upoly2) + 1;
    }
    for(i=0,ptr=SNAP_coefp(rpoly,0),pt1=SNAP_coefp(upoly1,0),pt2=SNAP_coefp(upoly2,0); i<=SNAP_degree(upoly1); i++,ptr++,pt1++,pt2++) {
      SNAPBASE_add(*ptr, *pt1, *pt2);
    }
    for(i=SNAP_degree(upoly1)+1,ptr=SNAP_coefp(rpoly,i),pt2=SNAP_coefp(upoly2,i); i<=SNAP_degree(upoly2); i++,ptr++,pt2++) {
      SNAPBASE_set(*ptr, *pt2);
    }
    deg = SNAP_degree(upoly2);
    for(i=0; i<=deg; i++) {
      SNAP_degree(rpoly) = deg-i;
      if ( !SNAPBASE_zeroq(SNAP_lcoef(rpoly)) ) {
        break;
      }
    }
  } else {
    if ( SNAP_size(rpoly) < SNAP_degree(upoly1) + 1 ) {
      SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(upoly1) + 1));
      for(i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(upoly1); i++,ptr++) {
        SNAPBASE_init(*ptr);
      }
      SNAP_size(rpoly) = SNAP_degree(upoly1) + 1;
    }
    for(i=0,ptr=SNAP_coefp(rpoly,0),pt1=SNAP_coefp(upoly1,0),pt2=SNAP_coefp(upoly2,0); i<=SNAP_degree(upoly2); i++,ptr++,pt1++,pt2++) {
      SNAPBASE_add(*ptr, *pt1, *pt2);
    }
    for(i=SNAP_degree(upoly2)+1,ptr=SNAP_coefp(rpoly,i),pt1=SNAP_coefp(upoly1,i); i<=SNAP_degree(upoly1); i++,ptr++,pt1++) {
      SNAPBASE_set(*ptr, *pt1);
    }
    deg = SNAP_degree(upoly1);
    for(i=0; i<=deg; i++) {
      SNAP_degree(rpoly) = deg-i;
      if ( !SNAPBASE_zeroq(SNAP_lcoef(rpoly)) ) {
        break;
      }
    }
  }
}

// Subtract the third SNAP_t from the second SNAP_t and store it in the first.
void SNAP_sub(SNAP_t rpoly, SNAP_t upoly1, SNAP_t upoly2) {
  unsigned int i,deg;
  SNAPBASE_t *ptr, *pt1, *pt2;
  if ( SNAP_degree(upoly1) < SNAP_degree(upoly2) ) {
    if ( SNAP_size(rpoly) < SNAP_degree(upoly2) + 1 ) {
      SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(upoly2) + 1));
      for(i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(upoly2); i++,ptr++) {
        SNAPBASE_init(*ptr);
      }
      SNAP_size(rpoly) = SNAP_degree(upoly2) + 1;
    }
    for(i=0,ptr=SNAP_coefp(rpoly,0),pt1=SNAP_coefp(upoly1,0),pt2=SNAP_coefp(upoly2,0); i<=SNAP_degree(upoly1); i++,ptr++,pt1++,pt2++) {
      SNAPBASE_sub(*ptr, *pt1, *pt2);
    }
    for(i=SNAP_degree(upoly1)+1,ptr=SNAP_coefp(rpoly,i),pt2=SNAP_coefp(upoly2,i); i<=SNAP_degree(upoly2); i++,ptr++,pt2++) {
      SNAPBASE_neg(*ptr, *pt2);
    }
    deg = SNAP_degree(upoly2);
    for(i=0; i<=deg; i++) {
      SNAP_degree(rpoly) = deg-i;
      if ( !SNAPBASE_zeroq(SNAP_lcoef(rpoly)) ) {
        break;
      }
    }
  } else {
    if ( SNAP_size(rpoly) < SNAP_degree(upoly1) + 1 ) {
      SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(upoly1) + 1));
      for(i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(upoly1); i++,ptr++) {
        SNAPBASE_init(*ptr);
      }
      SNAP_size(rpoly) = SNAP_degree(upoly1) + 1;
    }
    for(i=0,ptr=SNAP_coefp(rpoly,0),pt1=SNAP_coefp(upoly1,0),pt2=SNAP_coefp(upoly2,0); i<=SNAP_degree(upoly2); i++,ptr++,pt1++,pt2++) {
      SNAPBASE_sub(*ptr, *pt1, *pt2);
    }
    for(i=SNAP_degree(upoly2)+1,ptr=SNAP_coefp(rpoly,i),pt1=SNAP_coefp(upoly1,i); i<=SNAP_degree(upoly1); i++,ptr++,pt1++) {
      SNAPBASE_set(*ptr, *pt1);
    }
    deg = SNAP_degree(upoly1);
    for(i=0; i<=deg; i++) {
      SNAP_degree(rpoly) = deg-i;
      if ( !SNAPBASE_zeroq(SNAP_lcoef(rpoly)) ) {
        break;
      }
    }
  }
}

// Generate a random polynomial
void SNAP_randpoly(SNAP_t poly, unsigned int degree, SNAPBASE_t max_abs_coef, double nonzero_ratio, gmp_randstate_t rstate) {
  unsigned int i;
  SNAPBASE_t randmax;
  SNAPBASE_init(randmax);
  SNAPBASE_mul_ui(randmax, max_abs_coef, 2);
#ifdef SNAP_SUPZ_T
  SNAPBASE_add_ui(randmax, randmax, 1);
#endif
  // Initialize the resulting polynomial store
  if ( SNAP_size(poly) < degree + 1) {
    SNAP_coefs(poly) = (SNAPBASE_t*) realloc(SNAP_coefs(poly), sizeof(SNAPBASE_t)*(degree+1));
    for ( i = 0; i < SNAP_size(poly); i++ ) {
      SNAPBASE_set_ui(SNAP_coef(poly,i), 0);
    }
    for( i = SNAP_size(poly); i<degree+1 ; i++ ) {
      SNAPBASE_init_set_ui(SNAP_coef(poly,i),0);
    }
    SNAP_size(poly) = degree+1;
  } else {
    for ( i = 0; i <= degree; i++ ) {
      SNAPBASE_set_ui(SNAP_coef(poly,i), 0);
    }
  }
  SNAP_degree(poly) = degree;
  do {
#ifdef SNAP_SUPZ_T
    SNAPBASE_urandom(SNAP_lcoef(poly), rstate, randmax);
#else
    SNAPBASE_urandom(SNAP_lcoef(poly), rstate);
    SNAPBASE_mul(SNAP_lcoef(poly), SNAP_lcoef(poly), randmax);
#endif
    SNAPBASE_sub(SNAP_lcoef(poly), SNAP_lcoef(poly), max_abs_coef);
  } while ( SNAPBASE_zeroq(SNAP_lcoef(poly)) );
  for ( i = 0; i < degree; i++ ) {
    if ( ((double)rand()/((double)RAND_MAX+1)) <= nonzero_ratio ) {
      do {
#ifdef SNAP_SUPZ_T
        SNAPBASE_urandom(SNAP_lcoef(poly), rstate, randmax);
#else
        SNAPBASE_urandom(SNAP_coef(poly,i), rstate);
        SNAPBASE_mul(SNAP_coef(poly,i), SNAP_coef(poly,i), randmax);
#endif
        SNAPBASE_sub(SNAP_coef(poly,i), SNAP_coef(poly,i), max_abs_coef);
      } while ( SNAPBASE_zeroq(SNAP_coef(poly,i)) );
    }
  }
  SNAPBASE_clear(randmax);
}

#ifndef SNAP_SUPZ_T
// q is the quotient of f by g, by the traditional division algorithm.
// We can assume that deg(poly_g) > 0.
static void SNAP_div_qr_nc(SNAP_t poly_q, SNAP_t poly_r, SNAP_t poly_g) {
  SNAPBASE_t rp, irp, *ptr, *ptg;
  unsigned int i, deg_m;
  SNAPBASE_init(rp);
  SNAPBASE_init(irp);
  SNAPBASE_set_ui(SNAP_coef(poly_q,0), 0);
  SNAP_degree(poly_q) = 0;
  while ( SNAP_degree(poly_r) >= SNAP_degree(poly_g) ) {
    SNAPBASE_div(rp, SNAP_lcoef(poly_r), SNAP_lcoef(poly_g));
    deg_m = SNAP_degree(poly_r) - SNAP_degree(poly_g);
    SNAP_set_coef(poly_q, deg_m, rp);
    for ( i=0,ptg=SNAP_coefp(poly_g,0),ptr=SNAP_coefp(poly_r,deg_m); i<SNAP_degree(poly_g); i++,ptg++,ptr++ ) {
      SNAPBASE_mul(irp, rp, *ptg);
      SNAPBASE_sub(*ptr, *ptr, irp);
    }
    SNAP_degree(poly_r)--;
    while( SNAP_degree(poly_r) > 0 && SNAPBASE_zeroq(SNAP_lcoef(poly_r)) ) {
      SNAP_degree(poly_r)--;
    }
  }
  SNAPBASE_clear(rp);
  SNAPBASE_clear(irp);
}
void SNAP_div_qr(SNAP_t poly_q, SNAP_t poly_r, SNAP_t poly_f, SNAP_t poly_g) {
  SNAP_t tpoly;
  if ( SNAP_degree(poly_g) == 0 ) {
    SNAP_div_base(poly_q, poly_f, SNAP_coef(poly_g,0));
    SNAP_degree(poly_r) = 0;
    SNAP_set_coef_d(poly_r, 0, 0);
    return;
  }
  if ( poly_q == poly_g ) {
    SNAP_init(tpoly);
    if ( poly_r != poly_f ) {
      SNAP_set(poly_r, poly_f);
    }
    SNAP_div_qr_nc(tpoly, poly_r, poly_g);
    SNAP_swap(tpoly, poly_q);
    SNAP_clear(tpoly);
  } else if ( poly_r == poly_g ) {
    SNAP_init_set(tpoly, poly_f);
    SNAP_div_qr_nc(poly_q, tpoly, poly_g);
    SNAP_swap(tpoly, poly_r);
    SNAP_clear(tpoly);
  } else {
    if ( poly_r != poly_f ) {
      SNAP_set(poly_r, poly_f);
    }
    SNAP_div_qr_nc(poly_q, poly_r, poly_g);
  }
}

// q is the least square solution w.r.t. conv(g)vec(q)=vec(f).
// This means that |f - q*g|_2 and |q| are minimized.
void SNAP_lsdiv_q(SNAP_t poly_q, SNAP_t poly_f, SNAP_t poly_g) {
  SNAPMTX_t matrix_conv;
  unsigned int deg_fg;
  if ( SNAP_degree(poly_f) < SNAP_degree(poly_g) ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_lsdiv_q), ": Degree of Disivor is Larger than that of Dividend.\n");
    return;
  }
  deg_fg = SNAP_degree(poly_f)-SNAP_degree(poly_g);
  SNAPMTX_init(matrix_conv);
#ifdef SNAP_SUPDC_T
  SNAPMTX_convolution_matrix_conj(matrix_conv, poly_g, deg_fg);
#else
  SNAPMTX_convolution_matrix(matrix_conv, poly_g, deg_fg);
#endif
  // *B should be the coefficient of poly_q but the clone of poly_f.
  if ( poly_q != poly_f ) {
    SNAP_set(poly_q, poly_f);
  }
  // the return value must be 0 if success
#ifdef SNAP_SUPD_T
  if ( clapack_dgels(CblasColMajor, CblasTrans, SNAPMTX_cols(matrix_conv), SNAPMTX_rows(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_q), SNAP_degree(poly_f)+1 ) != 0 ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_lsdiv_q), ": clapack_dgels could not compute the least square solution.\n");
    return;
  }
#elif defined(SNAP_SUPDC_T)
  if ( clapack_zgels(CblasColMajor, CblasConjTrans, SNAPMTX_cols(matrix_conv), SNAPMTX_rows(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_q), SNAP_degree(poly_f)+1 ) != 0 ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_lsdiv_q), ": clapack_zgels could not compute the least square solution.\n");
    return;
  }
#elif defined(SNAP_SUPFR_T)
//	if ( snap_mlapack_frgels(SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_q), 1) != 0 )
  if ( snap_mlapack_frgels('N',SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_q), 1) != 0 ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_lsdiv_q), ": snap_mlapack_frgels could not compute the least square solution.\n");
    return;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgels('N',SNAPMTX_rows(matrix_conv), SNAPMTX_cols(matrix_conv), 1, SNAPMTX_elems(matrix_conv), SNAPMTX_csize(matrix_conv), SNAP_coefs(poly_q), 1) != 0 ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_lsdiv_q), ": snap_mlapack_fcgels could not compute the least square solution.\n");
    return;
  }
#endif
  SNAP_degree(poly_q) = deg_fg;
  SNAPMTX_clear(matrix_conv);
}
/*!
 *  @brief Polynomial division by FFT and Inverse FFT.
 *  @details q[x] = IFFT(Sum_i(FFT(f_i[x]))/Sum_i(FFT(g_i[x]))) where the length of FFT/IFFT is 2^k,
 *   2^k > deg(q[x]),deg(f_i[x]),deg(g_i[x]), f[x]=Sum_i(f_i[x]*x^(ik)) and g[x]=Sum_i(g_i[x]*x^(ik)).
 *  @warning This method fails if g[w] is zero for some n-th root of unity w. For the case, this routine calls lsdiv_q instead.
 *   Moreover, this implementation fails if f[w] is zero for some power of primitive n-th root of unity w. For the case f[w] is zero, this routine calls lsdiv_q instead.
 *  @param[out] poly_q The approximate quotient of f[x] by g[x].
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 */
void SNAP_fftdiv_q(SNAP_t poly_q, SNAP_t poly_f, SNAP_t poly_g) {
  unsigned int i, j, n, call_lsdiv=0;
  SNAPMTXCOMPLEX_t vecF, vecG, vecQ;
  SNAPBASE_t *ptb1;
  SNAPBASECOMPLEX_t *ptc2, *ptc3, *ptc4;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_plan fftp = NULL;
#else
  // MPFR and MPC
#endif
  if ( SNAP_degree(poly_f) < SNAP_degree(poly_g) ) {
    SNAP_clear(poly_q);
    SNAP_init(poly_q);
    return;
  }
  // n = (unsigned int) log2((double) SNAP_degree(poly_f)-SNAP_degree(poly_g)+1)+1;
  // n = (unsigned int) pow(2.0,(double) n);
  i = SNAP_degree(poly_f)-SNAP_degree(poly_g)+1;
  i >>= 1;
  n = 1;
  while ( i != 0 ) {
    i >>= 1;
    n *= 2;
  }
  if ( n<=SNAP_degree(poly_f)-SNAP_degree(poly_g) ) {
    n *= 2;
  }
  SNAPMTXCOMPLEX_init2(vecF,1,n);
  SNAPMTXCOMPLEX_init2(vecG,1,n);
  SNAPMTXCOMPLEX_init2(vecQ,1,n);
  // Sum_i(FFT(f_i[x]))
  for ( i=0,ptb1=SNAP_coefs(poly_f); i<=SNAP_degree(poly_f); i+=n ) {
    for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ); j<n&&i+j<=SNAP_degree(poly_f); j++,ptb1++,ptc2++ ) {
      SNAPBASECOMPLEX_set_base(*ptc2, *ptb1);
    }
    for ( ; j<n; j++,ptc2++ ) {
      SNAPBASECOMPLEX_set_ui(*ptc2, 0);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    if ( fftp == NULL ) {
      fftp = fftw_plan_dft_1d(n, SNAPMTXCOMPLEX_elems(vecQ), SNAPMTXCOMPLEX_elems(vecQ), FFTW_BACKWARD, FFTW_ESTIMATE);
      fftw_execute(fftp);
    } else {
      fftw_execute_dft(fftp, SNAPMTXCOMPLEX_elems(vecQ), SNAPMTXCOMPLEX_elems(vecQ));
    }
#else
    // MPFR and MPC
    snap_fcdft_backward(n, 1, SNAPMTXCOMPLEX_elems(vecQ), 1, SNAPMTXCOMPLEX_elems(vecQ), 1);
#endif
    for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ),ptc3=SNAPMTXCOMPLEX_elems(vecF); j<n; j++,ptc2++,ptc3++ ) {
      SNAPBASECOMPLEX_add(*ptc3, *ptc3, *ptc2);
    }
  }
  // Sum_i(FFT(g_i[x]))

  for ( i=0,ptb1=SNAP_coefs(poly_g); i<=SNAP_degree(poly_g); i+=n ) {
    for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ); j<n&&i+j<=SNAP_degree(poly_g); j++,ptb1++,ptc2++ ) {
      SNAPBASECOMPLEX_set_base(*ptc2, *ptb1);
    }
    for ( ; j<n; j++,ptc2++ ) {
      SNAPBASECOMPLEX_set_ui(*ptc2, 0);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    fftw_execute_dft(fftp, SNAPMTXCOMPLEX_elems(vecQ), SNAPMTXCOMPLEX_elems(vecQ));
#else
    // MPFR and MPC
    snap_fcdft_backward(n, 1, SNAPMTXCOMPLEX_elems(vecQ), 1, SNAPMTXCOMPLEX_elems(vecQ), 1);
#endif
    for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ),ptc3=SNAPMTXCOMPLEX_elems(vecG); j<n; j++,ptc2++,ptc3++ ) {
      SNAPBASECOMPLEX_add(*ptc3, *ptc3, *ptc2);
    }
  }
  // Sum_i(FFT(f_i[x]))/Sum_i(FFT(g_i[x]))
  for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ),ptc3=SNAPMTXCOMPLEX_elems(vecF),ptc4=SNAPMTXCOMPLEX_elems(vecG); j<n; j++,ptc2++,ptc3++,ptc4++ ) {
    if ( SNAPBASECOMPLEX_zeroq(*ptc3) || SNAPBASECOMPLEX_zeroq(*ptc4) ) {
      call_lsdiv = 1;
      goto post;
    }
    SNAPBASECOMPLEX_div(*ptc2, *ptc3, *ptc4);
  }
  // IFFT(Sum_i(FFT(f_i[x]))/Sum_i(FFT(g_i[x])))
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_destroy_plan(fftp);
  fftp = fftw_plan_dft_1d(n, SNAPMTXCOMPLEX_elems(vecQ), SNAPMTXCOMPLEX_elems(vecQ), FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
#else
  // MPFR and MPC
  snap_fcdft_forward(n, 1, SNAPMTXCOMPLEX_elems(vecQ), 1, SNAPMTXCOMPLEX_elems(vecQ), 1);
#endif
  // rest work
  for ( j=0,ptc2=SNAPMTXCOMPLEX_elems(vecQ); j<n; j++,ptc2++ ) {
    SNAPBASECOMPLEX_div_ui(*ptc2,*ptc2,n);
  }
  SNAP_set_basecomplex(poly_q, SNAPMTXCOMPLEX_elems(vecQ), SNAP_degree(poly_f)-SNAP_degree(poly_g)+1);
  SNAP_degree(poly_q) = SNAP_degree(poly_f)-SNAP_degree(poly_g);
post:
  SNAPMTXCOMPLEX_clear(vecF);
  SNAPMTXCOMPLEX_clear(vecG);
  SNAPMTXCOMPLEX_clear(vecQ);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  fftw_destroy_plan(fftp);
#else
  // MPFR and MPC
#endif
  if ( call_lsdiv == 1 ) {
    SNAP_lsdiv_q(poly_q, poly_f, poly_g);
  }
}
#endif
// rpoly = poly / divisor. e.g. scalar division.
void SNAP_div_fr(SNAP_t rpoly, SNAP_t poly, mpfr_t divisor) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_div_fr( *ptr, *pt, divisor);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

void SNAP_div_fc(SNAP_t rpoly, SNAP_t poly, mpc_t divisor) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_div_fc( *ptr, *pt, divisor);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

void SNAP_div_d(SNAP_t rpoly, SNAP_t poly, double divisor) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_div_d( *ptr, *pt, divisor);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}

void SNAP_div_dc(SNAP_t rpoly, SNAP_t poly, dcomplex divisor) {
  unsigned int i;
  SNAPBASE_t *ptr, *pt;
  // size check
  if ( rpoly != poly && SNAP_size(rpoly) <= SNAP_degree(poly) ) {
    SNAP_coefs(rpoly) = (SNAPBASE_t*) realloc(SNAP_coefs(rpoly), sizeof(SNAPBASE_t)*(SNAP_degree(poly)+1));
    for ( i=SNAP_size(rpoly),ptr=SNAP_coefp(rpoly,i); i<=SNAP_degree(poly); i++,ptr++ ) {
      SNAPBASE_init(*ptr);
    }
    SNAP_size(rpoly) = SNAP_degree(poly)+1;
  }
  for ( i=0,ptr=SNAP_coefp(rpoly,0),pt=SNAP_coefp(poly,0); i<=SNAP_degree(poly); i++,ptr++,pt++ ) {
    SNAPBASE_div_dc( *ptr, *pt, divisor);
  }
  SNAP_degree(rpoly) = SNAP_degree(poly);
}


