/*
 * Copyright (c) 2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "snap.h"

double getrusageinsec() {
  struct rusage t;
  getrusage(RUSAGE_SELF, &t);
  return t.ru_utime.tv_sec + (double)t.ru_utime.tv_usec*1e-6;
}

double getrtimeinsec() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec + (double)t.tv_usec*1e-6;
}

void supfc_show(supfc_t upoly) {
  int i;
  for(i=supfc_degree(upoly); i>=0; i--) {
    if ( mpfr_sgn(mpc_realref(supfc_coef(upoly,i))) == 0 ) {
      if ( mpfr_sgn(mpc_imagref(supfc_coef(upoly,i))) == 0 ) {
        continue;
      }
      switch(i) {
      case 0:
        printf("%+LfI", mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN));
        break;
      case 1:
        printf("%+LfI x", mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN));
        break;
      default:
        printf("%+LfI x^%d", mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN), i);
      }
    } else if ( mpfr_sgn(mpc_imagref(supfc_coef(upoly,i))) != 0 ) {
      switch(i) {
      case 0:
        printf("%+Lf%+LfI", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN), mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN));
        break;
      case 1:
        printf("+(%+Lf%+LfI)x", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN), mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN));
        break;
      default:
        printf("+(%+Lf%+LfI)x^%d", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN), mpfr_get_ld(supfc_coef_imagref(upoly,i), MPFR_RNDN), i);
      }

    } else {
      switch(i) {
      case 0:
        printf("%+Lf", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN));
        break;
      case 1:
        printf("%+Lf x", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN));
        break;
      default:
        printf("%+Lf x^%d", mpfr_get_ld(supfc_coef_realref(upoly,i), MPFR_RNDN), i);
      }
    }
  }
  printf("\n");
}

void interrupted(int sig) {
  printf("Program Terminated due to Signal %d.\n", sig);
  printf("Please do not enter CTRL-D at the beginning of inputs.\n");
  exit(1);
}

int main() {
  char *buf;
  supfc_t f_fc,g_fc,agcd_fc,utmp1_fc,utmp2_fc;
  gmp_randstate_t rstate;
  unsigned int deg_f, deg_g;
  double start_time, end_time, start_rtime, end_rtime;
  mpfr_t tol_fc, norm1, norm2, norm3;
  signal(SIGSEGV, interrupted);
  using_history();
  read_history(".fesnap_history");
  snap_copyright();
  printf("fastgcdMC begins.\n");
  buf = readline("Input PREC (base 2): ");
  mpfr_set_default_prec(strtoul(buf,NULL,10));
  mpfr_init(tol_fc);
  mpfr_init(norm1);
  mpfr_init(norm2);
  mpfr_init(norm3);
  supfc_init(utmp1_fc);
  supfc_init(utmp2_fc);
  supfc_init(f_fc);
  supfc_init(g_fc);
  supfc_init(agcd_fc);
  deg_f = UINT_MAX;
  deg_g = UINT_MAX;
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
  srand((unsigned int)time(NULL));
  buf = readline("Tolerance: ");
  mpfr_set_str(tol_fc, buf, 10, MPFR_RNDN);
  free(buf);
  while(1) {
    if ( deg_f==UINT_MAX ) {
      buf = readline("Input F[x]: ");
    } else {
      buf = readline("Input G[x]: ");
    }
    add_history(buf);
    if( strstr( buf, "quit" )!=NULL ) {
      free(buf);
      break;
    }
    if ( deg_f==UINT_MAX ) deg_f=supfc_set_str(f_fc, buf);
    else deg_g=supfc_set_str(g_fc, buf);
    free(buf);
    if ( deg_f != UINT_MAX && deg_g != UINT_MAX ) {
      printf("===== fastgcdMC =====\n");

      printf("F[x]: ");
      supfc_show(f_fc);
      printf("G[x]: ");
      supfc_show(g_fc);
      start_rtime = getrtimeinsec();
      start_time = getrusageinsec();
      if ( supfc_agcd_fastgcd(f_fc,g_fc,agcd_fc,utmp1_fc,utmp2_fc,tol_fc) != 0 ) {
        printf("Could not reach the case 1. Interrupted!\n");
      }
      end_time = getrusageinsec();
      end_rtime = getrtimeinsec();
      printf("%e sec. (cpu: %e sec.)\n", end_rtime - start_rtime, end_time - start_time);
      printf("Degree of D[X]: %d\n", supfc_degree(agcd_fc));
      printf("D[x]: ");
      supfc_show(agcd_fc);
      printf("F1[x]: ");
      supfc_show(utmp1_fc);
      printf("G1[x]: ");
      supfc_show(utmp2_fc);

      supfc_mul(utmp1_fc, utmp1_fc, agcd_fc);
      supfc_mul(utmp2_fc, utmp2_fc, agcd_fc);
      supfc_sub(utmp1_fc, f_fc, utmp1_fc);
      supfc_sub(utmp2_fc, g_fc, utmp2_fc);
      supfc_2norm(norm1, utmp1_fc);
      supfc_2norm(norm2, utmp2_fc);
      mpfr_printf("Residual of F[X]: %Re\n", norm1);
      mpfr_printf("Residual of G[X]: %Re\n", norm2);
      mpfr_mul(norm1, norm1, norm1, MPFR_RNDN);
      mpfr_mul(norm2, norm2, norm2, MPFR_RNDN);
      mpfr_add(norm3, norm1, norm2, MPFR_RNDN);
      mpfr_sqrt(norm3, norm3, MPFR_RNDN);
      mpfr_printf("Residual of Both: %Re\n", norm3);

      deg_f=UINT_MAX;
      deg_g=UINT_MAX;
    }
  }
  mpfr_clear(tol_fc);
  mpfr_clear(norm1);
  mpfr_clear(norm2);
  mpfr_clear(norm3);
  supfc_clear(utmp1_fc);
  supfc_clear(utmp2_fc);
  supfc_clear(f_fc);
  supfc_clear(g_fc);
  supfc_clear(agcd_fc);
  gmp_randclear(rstate);
  write_history(".fesnap_history");
  return (0);
}
