/*
 * Copyright (c) 2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <clapack.h>

#include "snap.h"

double getrusageinsec() {
  struct rusage t;
  getrusage(RUSAGE_SELF, &t);
  return t.ru_utime.tv_sec + (double)t.ru_utime.tv_usec*1e-6;
}

void supdc_show(supdc_t upoly) {
  int i;
  for(i=supdc_degree(upoly); i>=0; i--) {
    if ( supdc_coef(upoly,i) == 0.0 ) {
      continue;
    }
    switch(i) {
    case 0:
      printf("+(%+.15le+%+.15leI)", creal(supdc_coef(upoly,i)), cimag(supdc_coef(upoly,i)));
      break;
    case 1:
      printf("+(%+.15le+%+.15leI)x", creal(supdc_coef(upoly,i)), cimag(supdc_coef(upoly,i)));
      break;
    default:
      printf("+(%+.15le+%+.15leI)x^%d", creal(supdc_coef(upoly,i)), cimag(supdc_coef(upoly,i)), i);
    }
  }
  printf("\n");
}

void interrupted(int sig) {
  printf("Program Terminated due to Signal %d.\n", sig);
  printf("Please do not enter CTRL-D at the beginning of inputs.\n");
  exit(1);
}

int supdc_qr_lapack(supdc_t poly_inf, supdc_t poly_ing) {
  unsigned int i, q1, q2;
  sdmdc_t matrix_syl;
  dcomplex *tau;

  sdmdc_init(matrix_syl);
  sdmdc_sylvester_matrix_transpose(matrix_syl, poly_inf, poly_ing);

  printf("Sylvester Matrix:\n");
  for (q1=0; q1<sdmdc_rows(matrix_syl); q1++ ) {
    for (q2=0; q2<sdmdc_cols(matrix_syl); q2++ ) {
      printf("(%+f%+fI),", creal(sdmdc_elem(matrix_syl,q1,q2)), cimag(sdmdc_elem(matrix_syl,q1,q2)));
    }
    printf("\n");
  }

  tau = (dcomplex*) malloc(sizeof(dcomplex)*sdmdc_rows(matrix_syl));
  for ( i=0; i<sdmdc_rows(matrix_syl); i++ ) {
    *(tau+i) = 0.0;
  }
  clapack_zgeqrf(CblasRowMajor, sdmdc_rows(matrix_syl), sdmdc_cols(matrix_syl), sdmdc_elems(matrix_syl), sdmdc_csize(matrix_syl), tau);

  printf("QR decomposition by clapack:\n");
  for (q1=0; q1<sdmdc_rows(matrix_syl); q1++ ) {
    for (q2=0; q2<sdmdc_cols(matrix_syl); q2++ ) {
      printf("(%+f%+fI),", creal(sdmdc_elem(matrix_syl,q1,q2)), cimag(sdmdc_elem(matrix_syl,q1,q2)));
    }
    printf("\n");
  }
  printf("tau: ");
  for (q1=0; q1<sdmdc_rows(matrix_syl); q1++ ) {
    printf("(%+f%+fI),", creal(*(tau+q1)), cimag(*(tau+q1)));
  }
  printf("\n");

  sdmdc_clear(matrix_syl);
  free(tau);
  return 0;
}

void supdc_alg_qr_dgv2010refine_make_sylvester_generators(supdc_t poly_inf, supdc_t poly_ing, sdmdc_t genV, sdmdc_t genH);
void supdc_alg_qr_dgv2010refine_sylvester(supdc_t poly_inf, supdc_t poly_ing, sdmdc_t matQ, sdmdc_t matR);
void supdc_alg_qr_dgv2010refine_givens_R(dcomplex *ca, dcomplex *cb, dcomplex *c, dcomplex *s, dcomplex *ctmp);
void supdc_alg_qr_dgv2010refine_givens_C(dcomplex *ca, dcomplex *cb, dcomplex *c, dcomplex *s, dcomplex *ctmp);
void supdc_alg_qr_dgv2010refine_apply_S(const unsigned int s, sdmdc_t leftS, sdmdc_t rightS, sdmdc_t scalarS, sdmdc_t matR, const unsigned int k);
void supdc_alg_qr_dgv2010refine_pulldownthrough(dcomplex *g0, dcomplex *g1, dcomplex *g2, dcomplex *w);
void supdc_alg_qr_dgv2010refine_pulldownthroughright(const unsigned int r, dcomplex *g, dcomplex *scalarS1, dcomplex *leftS1, dcomplex *scalarS2, dcomplex *leftS2, const unsigned int num_op, dcomplex *w);
void supdc_alg_qr_dgv2010refine_pullupthrough(dcomplex *g0, dcomplex *g1, dcomplex *g2, dcomplex *w);
void supdc_alg_qr_dgv2010refine_pulldownthrough_left(dcomplex *g0, dcomplex *g1, dcomplex *g2, dcomplex *w);


#define PRINT_LEFT_S {printf("leftS:\n");for(q1=0;q1<sdmdc_rows(leftS);q1++){for(q2=0;q2<sdmdc_cols(leftS);q2++){printf("(%+f%+fI),",creal(sdmdc_elem(leftS,q1,q2)),cimag(sdmdc_elem(leftS,q1,q2)));}printf("\n");}}
#define PRINT_RIGHT_S {printf("rightS:\n");for(q1=0;q1<sdmdc_rows(rightS);q1++){for(q2=0;q2<sdmdc_cols(rightS);q2++){printf("(%+f%+fI),",creal(sdmdc_elem(rightS,q1,q2)),cimag(sdmdc_elem(rightS,q1,q2)));}printf("\n");}}
#define PRINT_SCALAR_S {printf("scalarS:\n");for(q1=0;q1<sdmdc_rows(scalarS);q1++){for(q2=0;q2<sdmdc_cols(scalarS);q2++){printf("(%+f%+fI),",creal(sdmdc_elem(scalarS,q1,q2)),cimag(sdmdc_elem(scalarS,q1,q2)));}printf("\n");}}

void check_pulldownthrough_left() {
  dcomplex *g0, *g1, *g2, *w;
  g0 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g1 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g2 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  w = (dcomplex *) malloc(sizeof(dcomplex)*11);
  *g0 = 0.534142 + 0.767933*_Complex_I;
  *(g0+1) = 0.2718 - 0.226045*_Complex_I;
  *g1 = 0.30031 - 0.35612*_Complex_I;
  *(g1+1) = -0.633374 - 0.617924*_Complex_I;
  *g2 = -0.955158 - 0.141521*_Complex_I;
  *(g2+1) = 0.157342 - 0.207094*_Complex_I;
  supdc_alg_qr_dgv2010refine_pulldownthrough_left(g0, g1, g2, w);
  printf("g1: (%+e%+eI),(%+e%+eI)\n", creal(*g1), cimag(*g1), creal(*(g1+1)), cimag(*(g1+1)));
  printf("g2: (%+e%+eI),(%+e%+eI)\n", creal(*g2), cimag(*g2), creal(*(g2+1)), cimag(*(g2+1)));
  printf("g0: (%+e%+eI),(%+e%+eI)\n", creal(*g0), cimag(*g0), creal(*(g0+1)), cimag(*(g0+1)));
}

void check_pullupthrough() {
  dcomplex *g0, *g1, *g2, *w;
  g0 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g1 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g2 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  w = (dcomplex *) malloc(sizeof(dcomplex)*11);
  *g0 = 0.267385 + 0.157036*_Complex_I;
  *(g0+1) = -0.843213 - 0.439131*_Complex_I;
  *g1 = 0.22236 - 0.511015*_Complex_I;
  *(g1+1) = 0.0736922 + 0.827036*_Complex_I;
  *g2 = -0.194933 - 0.377646*_Complex_I;
  *(g2+1) = -0.273565 - 0.862871*_Complex_I;
  supdc_alg_qr_dgv2010refine_pullupthrough(g0, g1, g2, w);
  printf("g1: (%+e%+eI),(%+e%+eI)\n", creal(*g1), cimag(*g1), creal(*(g1+1)), cimag(*(g1+1)));
  printf("g2: (%+e%+eI),(%+e%+eI)\n", creal(*g2), cimag(*g2), creal(*(g2+1)), cimag(*(g2+1)));
  printf("g0: (%+e%+eI),(%+e%+eI)\n", creal(*g0), cimag(*g0), creal(*(g0+1)), cimag(*(g0+1)));
}

void check_pulldownthroughright() {
  unsigned int i, r=3, num_op=3;
  dcomplex *g, *scalarS1, *leftS1, *scalarS2, *leftS2, *w;
  g = (dcomplex *) malloc(sizeof(dcomplex)*2);
  scalarS1 = (dcomplex *) malloc(sizeof(dcomplex)*1);
  scalarS2 = (dcomplex *) malloc(sizeof(dcomplex)*1);
  leftS1 = (dcomplex *) malloc(sizeof(dcomplex)*2*3);
  leftS2 = (dcomplex *) malloc(sizeof(dcomplex)*2*3);
  w = (dcomplex *) malloc(sizeof(dcomplex)*11);
  *g = -0.391019 + 0.146769*_Complex_I;
  *(g+1) = 0.87136 - 0.257477*_Complex_I;
  *scalarS1 = 0.592816 + 0.805338*_Complex_I;
  *scalarS2 = -0.982295 - 0.187339*_Complex_I;
  *leftS1 = 0.0217075 + 0.665021*_Complex_I;
  *(leftS1+1) = 0.665289 - 0.338624*_Complex_I;
  *(leftS1+2) = -0.278464 + 0.464365*_Complex_I;
  *(leftS1+3) = 0.466247 + 0.699597*_Complex_I;
  *(leftS1+4) = 0.60202 - 0.627915*_Complex_I;
  *(leftS1+5) = 0.301447 + 0.390416*_Complex_I;
  *leftS2 = 0.452777 + 0.886996*_Complex_I;
  *(leftS2+1) = 0.0896392 - 0.0140008*_Complex_I;
  *(leftS2+2) = 0.104096 - 0.412408*_Complex_I;
  *(leftS2+3) = -0.305605 -0.851874*_Complex_I;
  *(leftS2+4) = -0.0650806 - 0.845536*_Complex_I;
  *(leftS2+5) = -0.138848 - 0.511425*_Complex_I;
  supdc_alg_qr_dgv2010refine_pulldownthroughright(r, g, scalarS1, leftS1, scalarS2, leftS2, num_op, w);
  printf("leftS1: "); for (i=0;i<6;i++) printf(",(%+e%+eI)", creal(*(leftS1+i)), cimag(*(leftS1+i)));
  printf("\nleftS2: "); for (i=0;i<6;i++) printf(",(%+e%+eI)", creal(*(leftS2+i)), cimag(*(leftS2+i)));
  printf("\ng: (%+e%+eI),(%+e%+eI)\n", creal(*g), cimag(*g), creal(*(g+1)), cimag(*(g+1)));
}

void check_pulldownthrough() {
  dcomplex *g0, *g1, *g2, *w;
  g0 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g1 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  g2 = (dcomplex *) malloc(sizeof(dcomplex)*2);
  w = (dcomplex *) malloc(sizeof(dcomplex)*11);
  *g0 = 0.477925 + 0.438103*_Complex_I;
  *(g0+1) = 0.612736 - 0.451893*_Complex_I;
  *g1 = -0.190296 - 0.185684*_Complex_I;
  *(g1+1) = 0.740085 + 0.617725*_Complex_I;
  *g2 = -0.32635 - 0.0379438*_Complex_I;
  *(g2+1) = 0.352955 + 0.876059*_Complex_I;
  supdc_alg_qr_dgv2010refine_pulldownthrough(g0, g1, g2, w);
  printf("g1: (%+e%+eI),(%+e%+eI)\n", creal(*g1), cimag(*g1), creal(*(g1+1)), cimag(*(g1+1)));
  printf("g2: (%+e%+eI),(%+e%+eI)\n", creal(*g2), cimag(*g2), creal(*(g2+1)), cimag(*(g2+1)));
  printf("g0: (%+e%+eI),(%+e%+eI)\n", creal(*g0), cimag(*g0), creal(*(g0+1)), cimag(*(g0+1)));
}

void check_applyS() {
  sdmdc_t leftS, rightS, scalarS, matR;
  const unsigned int n = 7;
  const unsigned int r = 3;
  unsigned int i,j,q1,q2;
  sdmdc_init2(matR, n, n);
  sdmdc_init2(leftS, n-1, 2*r);                      // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))^h
  sdmdc_init2(rightS, n-r-1, 2*(r-1));               // accesor: i,2*j S(i,2*j) => (c,s)
  sdmdc_init2(scalarS, 1, n-1);                      // accesor: i
  for (i=0; i<sdmdc_rows(matR); i++ ) {
    for (j=0; j<sdmdc_cols(matR); j++ ) {
      sdmdc_elem(matR,i,j) = 2*i + 3*j;
    }
  }
  for ( i=0; i<sdmdc_cols(scalarS); i++ ) {
    sdmdc_elem(scalarS, 0, i) = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i)/(n));
  }               // accesor: i
  for (i=0; i<sdmdc_rows(leftS); i++ ) {
    for (j=0; j<sdmdc_cols(leftS); j+=2 ) {
      sdmdc_elem(leftS,i,j) = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i*j)/(n))/sqrt(2);
      sdmdc_elem(leftS,i,j+1) = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i*(j+1))/(n))/sqrt(2);
    }
  }
  for (i=0; i<sdmdc_rows(rightS); i++ ) {
    for (j=0; j<sdmdc_cols(rightS); j+=2 ) {
      sdmdc_elem(rightS,i,j) = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i*j)/(n))/sqrt(2);
      sdmdc_elem(rightS,i,j+1) = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i*(j+1))/(n))/sqrt(2);
    }
  }
  for (i=0; i<sdmdc_rows(matR); i++ ) {
    for (j=0; j<sdmdc_cols(matR); j++ ) {
      printf("(%+e%+eI),", creal(sdmdc_elem(matR,i,j)), cimag(sdmdc_elem(matR,i,j)));
    }
    printf("\n");
  }
  PRINT_SCALAR_S
  PRINT_LEFT_S
  PRINT_RIGHT_S
  supdc_alg_qr_dgv2010refine_apply_S(0, leftS, rightS, scalarS, matR, 3);
  for (i=0; i<sdmdc_rows(matR); i++ ) {
    for (j=0; j<sdmdc_cols(matR); j++ ) {
      printf("(%+e%+eI),", creal(sdmdc_elem(matR,i,j)), cimag(sdmdc_elem(matR,i,j)));
    }
    printf("\n");
  }
  sdmdc_clear(matR);
  sdmdc_clear(leftS);
  sdmdc_clear(rightS);
  sdmdc_clear(scalarS);
}

void check_givens() {
  dcomplex ca, cb, c, s, ctmp;
  ca = -0.773516 - 0.647476*_Complex_I;
  cb =  0.284862 + 0.436704*_Complex_I;
  printf("(%+e%+eI),(%+e%+eI) => \n", creal(ca), cimag(ca), creal(cb), cimag(cb));
  supdc_alg_qr_dgv2010refine_givens_R(&ca, &cb, &c, &s, &ctmp);
  printf("(%+e%+eI),(%+e%+eI),(%+e%+eI)\n", creal(c), cimag(c), creal(s), cimag(s), creal(ca), cimag(ca));
  printf("mme: {-0.681199 + 0.570202 I, 0.250864 - 0.384584 I}, 1.13552\n");
  ca = -0.773516 - 0.647476*_Complex_I;
  cb =  0.284862 + 0.436704*_Complex_I;
  printf("(%+e%+eI),(%+e%+eI) => \n", creal(ca), cimag(ca), creal(cb), cimag(cb));
  supdc_alg_qr_dgv2010refine_givens_C(&ca, &cb, &c, &s, &ctmp);
  printf("(%+e%+eI),(%+e%+eI),(%+e%+eI)\n", creal(c), cimag(c), creal(s), cimag(s), creal(ca), cimag(ca));
  printf("mme: {-0.681199 - 0.570202 I, 0.422735 + 0.179256 I}, 0.199866 + 1.11779 I\n");
}

int supdc_qr_dgv2010refine(supdc_t poly_inf, supdc_t poly_ing) {
  unsigned int i, q1, q2, n, m;
  sdmdc_t genV, genH, matQ, matR;

  n = supdc_degree(poly_inf);
  m = supdc_degree(poly_ing);

  sdmdc_init2(genV, n+m, 2);
  sdmdc_init2(genH, 2, n+m);
  supdc_alg_qr_dgv2010refine_make_sylvester_generators(poly_inf, poly_ing, genV, genH);

  printf("Generators of Sylvester Matrix: V:\n");
  for (q1=0; q1<sdmdc_rows(genV); q1++ ) {
    for (q2=0; q2<sdmdc_cols(genV); q2++ ) {
      printf("(%+f%+fI),", creal(sdmdc_elem(genV,q1,q2)), cimag(sdmdc_elem(genV,q1,q2)));
    }
    printf("\n");
  }
  printf("Generators of Sylvester Matrix: H:\n");
  for (q1=0; q1<sdmdc_rows(genH); q1++ ) {
    for (q2=0; q2<sdmdc_cols(genH); q2++ ) {
      printf("(%+f%+fI),", creal(sdmdc_elem(genH,q1,q2)), cimag(sdmdc_elem(genH,q1,q2)));
    }
    printf("\n");
  }

  sdmdc_clear(genV);
  sdmdc_clear(genH);

  sdmdc_init2(matQ, n+m, n+m);
  sdmdc_init2(matR, n+m, n+m);
  supdc_alg_qr_dgv2010refine_sylvester(poly_inf, poly_ing, matQ, matR);
  sdmdc_clear(matQ);
  sdmdc_clear(matR);

  return 0;
}

int main() {
  char *buf;
  supdc_t f_d,g_d;
  unsigned int deg_f, deg_g;
  double start_time, end_time;
  signal(SIGSEGV, interrupted);
  using_history();
  read_history(".fesnap_history");
  snap_copyright();
  supdc_init(f_d);
  supdc_init(g_d);
  printf("supdc_agcd_exqrgcd begins.\n");
  deg_f = UINT_MAX;
  deg_g = UINT_MAX;
  // check_givens();
  // check_applyS();
  // check_pulldownthrough();
  // check_pulldownthroughright();
  // check_pullupthrough();
  check_pulldownthrough_left();
  while(1) {
    if ( deg_f==UINT_MAX ) {
      buf = readline("Input F[x]: ");
    } else {
      buf = readline("Input G[x]: ");
    }
    add_history(buf);
    if( strstr( buf, "quit" )!=NULL ) {
      free(buf);
      break;
    }
    if ( deg_f==UINT_MAX ) deg_f=supdc_set_str(f_d, buf);
    else deg_g=supdc_set_str(g_d, buf);
    free(buf);
    if ( deg_f != UINT_MAX && deg_g != UINT_MAX ) {
      // supdc_mul(f_d, f_d, f_d); // out_f2
      // supdc_mul(g_d, g_d, g_d); // out_g2
      printf("F[x]: ");
      supdc_show(f_d);
      printf("G[x]: ");
      supdc_show(g_d);
      start_time = getrusageinsec();
      if ( supdc_qr_lapack(f_d,g_d) != 0 ) {
        printf("Could not reach the case 1. Interrupted!\n");
      }
      end_time = getrusageinsec();
      printf("%f sec.\n", end_time - start_time);
      start_time = getrusageinsec();
      if ( supdc_qr_dgv2010refine(f_d,g_d) != 0 ) {
        printf("Could not reach the case 1. Interrupted!\n");
      }
      end_time = getrusageinsec();
      printf("%f sec.\n", end_time - start_time);

      deg_f=UINT_MAX;
      deg_g=UINT_MAX;
    }
  }
  supdc_clear(f_d);
  supdc_clear(g_d);
  write_history(".fesnap_history");
  return (0);
}
