/*
 * Copyright (c) 2016, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "snap.h"

double getrusageinsec() {
  struct rusage t;
  getrusage(RUSAGE_SELF, &t);
  return t.ru_utime.tv_sec + (double)t.ru_utime.tv_usec*1e-6;
}

double getrtimeinsec() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return t.tv_sec + (double)t.tv_usec*1e-6;
}

void supd_show(supd_t upoly) {
  int i;
  for(i=supd_degree(upoly); i>=0; i--) {
    if ( supd_coef(upoly,i) == 0.0 ) {
      continue;
    }
    switch(i) {
    case 0:
      printf("+(%+.15le)", supd_coef(upoly,i));
      break;
    case 1:
      printf("+(%+.15le)x", supd_coef(upoly,i));
      break;
    default:
      printf("+(%+.15le)x^%d", supd_coef(upoly,i), i);
    }
  }
  printf("\n");
}

void interrupted(int sig) {
  printf("Program Terminated due to Signal %d.\n", sig);
  printf("Please do not enter CTRL-D at the beginning of inputs.\n");
  exit(1);
}

int main(int argc,char *argv[]) {
  char *buf;
  supd_t f_d,g_d,agcd_d,utmp1_d,utmp2_d;
  gmp_randstate_t rstate;
  unsigned int deg_f, deg_g, mode;
  double start_time, end_time, start_rtime, end_rtime, tol;
  signal(SIGSEGV, interrupted);
  using_history();
  read_history(".fesnap_history");
  snap_copyright();
  supd_init(utmp1_d);
  supd_init(utmp2_d);
  supd_init(f_d);
  supd_init(g_d);
  supd_init(agcd_d);
  printf("uvgcd2 begins.\n");
  deg_f = UINT_MAX;
  deg_g = UINT_MAX;
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
  srand((unsigned int)time(NULL));
  if ( argc > 1 ) {
    mode = strtoul(argv[1],NULL,10);
    printf("Mode: %d\n", mode);
  } else {
    buf = readline("Mode: ");
    mode = strtoul(buf,NULL,10);
    free(buf);
  }
  buf = readline("Tolerance: ");
  tol = strtod(buf,NULL);
  free(buf);
  while(1) {
    if ( deg_f==UINT_MAX ) {
      buf = readline("Input F[x]: ");
    } else {
      buf = readline("Input G[x]: ");
    }
    add_history(buf);
    if( strstr( buf, "quit" )!=NULL ) {
      free(buf);
      break;
    }
    if ( deg_f==UINT_MAX ) deg_f=supd_set_str(f_d, buf);
    else deg_g=supd_set_str(g_d, buf);
    free(buf);
    if ( deg_f != UINT_MAX && deg_g != UINT_MAX ) {
      printf("===== uvgcd2 =====\n");

      printf("F[x]: ");
      supd_show(f_d);
      printf("G[x]: ");
      supd_show(g_d);
      start_rtime = getrtimeinsec();
      start_time = getrusageinsec();
      if ( supd_agcd_uvgcd2(f_d,g_d,agcd_d,utmp1_d,utmp2_d,tol,mode) != 0 ) {
        printf("Could not compute the aGCD. Interrupted!\n");
      }
      end_time = getrusageinsec();
      end_rtime = getrtimeinsec();
      printf("%e sec. (cpu: %e sec.)\n", end_rtime - start_rtime, end_time - start_time);
      printf("Degree of D[X]: %d\n", supd_degree(agcd_d));
      printf("D[x]: ");
      supd_show(agcd_d);
      printf("F1[x]: ");
      supd_show(utmp1_d);
      printf("G1[x]: ");
      supd_show(utmp2_d);

      supd_mul(utmp1_d, utmp1_d, agcd_d);
      supd_mul(utmp2_d, utmp2_d, agcd_d);
      supd_sub(utmp1_d, f_d, utmp1_d);
      supd_sub(utmp2_d, g_d, utmp2_d);
      printf("Residual of F[X]: %le\n", supd_2norm(utmp1_d));
      printf("Residual of G[X]: %le\n", supd_2norm(utmp2_d));
      printf("Residual of Both: %le\n", sqrt(supd_2norm(utmp1_d)*supd_2norm(utmp1_d)+supd_2norm(utmp2_d)*supd_2norm(utmp2_d)));

      deg_f=UINT_MAX;
      deg_g=UINT_MAX;
    }
  }
  supd_clear(utmp1_d);
  supd_clear(utmp2_d);
  supd_clear(f_d);
  supd_clear(g_d);
  supd_clear(agcd_d);
  gmp_randclear(rstate);
  write_history(".fesnap_history");
  return (0);
}
