/*
 * Copyright (c) 2017, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_TRUE 1
#define SNAPLIB_FALSE 0
#define SNAP2str(a) #a
//#define SNAPLIB_DEBUG
#undef SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
#if defined(SNAP_SUPD_T)
void show_sdmd(sdmd_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmd_rows(mtx); i++) {
    for(j=0; j<sdmd_cols(mtx); j++) {
      printf("%+.1le ",sdmd_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_dlist(double *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
  }
  printf("\n");
}
void show_dlist2(double *list, unsigned int len, unsigned int cols) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("%+.5le ",*(list+i));
    if ( (i+1)%cols == 0 ) {
      printf("\n");
    }
  }
  printf("\n");
}
#elif defined(SNAP_SUPDC_T)
void show_sdmdc(sdmdc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmdc_rows(mtx); i++) {
    for(j=0; j<sdmdc_cols(mtx); j++) {
      printf("(%+.5le%+.5le*I), ",creal(sdmdc_elem(mtx,i,j)),cimag(sdmdc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_dclist(dcomplex *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.5le%+.5le*I), ",creal(*(list+i)),cimag(*(list+i)));
  }
  printf("\n");
}
void show_dclist2(dcomplex *list, unsigned int len, unsigned int cols) {
  unsigned int i;
  for(i=0; i<len; i++) {
    printf("(%+.5le%+.5le*I), ",creal(*(list+i)),cimag(*(list+i)));
    if ( (i+1)%cols == 0 ) {
      printf("\n");
    }
  }
  printf("\n");
}
#elif defined(SNAP_SUPFR_T)
void show_sdmfr(sdmfr_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfr_rows(mtx); i++) {
    for(j=0; j<sdmfr_cols(mtx); j++) {
      mpfr_printf("%+.5Re ",sdmfr_elem(mtx,i,j));
    }
    printf("\n");
  }
}
void show_frlist(mpfr_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("%+.5Re ",*(list+i));
  }
  printf("\n");
}
#elif defined(SNAP_SUPFC_T)
void show_sdmfc(sdmfc_t mtx) {
  unsigned int i,j;
  for(i=0; i<sdmfc_rows(mtx); i++) {
    for(j=0; j<sdmfc_cols(mtx); j++) {
      mpfr_printf("(%+.5Re%+.5Re*I), ",mpc_realref(sdmfc_elem(mtx,i,j)),mpc_imagref(sdmfc_elem(mtx,i,j)));
    }
    printf("\n");
  }
}
void show_fclist(mpc_t *list, unsigned int len) {
  unsigned int i;
  for(i=0; i<len; i++) {
    mpfr_printf("(%+.5Re%+.5Re*I), ",mpc_realref(*(list+i)),mpc_imagref(*(list+i)));
  }
  printf("\n");
}
#endif
#endif

/* ================== START [agcd_stlngcd] ================== */

// in Kaltofen, Yang, Zhi (2007), between 10^8 and 10^10.
#define SNAPLIB_STLN_PENALTY_VALUE 1.0e9
// there is no maximum loop in Kaltofen, Yang, Zhi (2007).
#define SNAPLIB_STLN_MAXIMUM_LOOP 50
// if use snap_dgelsy by givens instead. This is applicable only for stln NOT stln2
//#define SNAPLIB_STLN_USE_GIVENS
// if use the original additional condition. This is applicable only for stln NOT stln2
//#define SNAPLIB_STLN_RES_COND
// if use refinement of UVGCD
//#define SNAPLIB_STLN_USE_REFINEMENT

#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
int SNAP_agcd_uvgcd_iteration_9_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, SNAPBASEREAL_t *rho);
int SNAP_agcd_uvgcd_iteration_7_3(SNAPMTX_t R, SNAPBASE_t *zvect, SNAPBASE_t *yvect, SNAPBASEREAL_t *sigma_m1, gmp_randstate_t rstate);
int SNAP_agcd_uvgcd_solving_10_2(SNAP_t poly_f, SNAP_t poly_g, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, unsigned int current_r);
extern void dlaic1_(int *job, int *j, double *x, double *sest, double *w, double *gamma, double *sestpr, double *s, double *c);
extern void zlaic1_(int *job, int *j, double complex *x, double *sest, double complex *w, double complex *gamma, double *sestpr, double complex *s, double complex *c);
#endif // defined

/*!
 *  @brief LS by Givens QR factorization.
 *  @param[in] rows The number of rows of A.
 *  @param[in] cols The number of columns of A.
 *  @param[in] nrhs The number of r.h.s. (we only support 1)
 *  @param[in,out] matA The pointer of A.
 *  @param[in] ldA The leading dimension of A.
 *  @param[in,out] vecB The r.h.s. on input and the solution on exit.
 *  @param[in] ldB The leading dimension of B.
 *  @param[in,out] jpvt The column pivotting.  It should be zero vector on input.
 *  @param[in] relative_tolerance The threshold for numerical rank. ignore!!
 *  @param[out] rank The numerical rank.
 */
int SNAP_gelsy(unsigned int rows, unsigned int cols, unsigned int nrhs, SNAPBASE_t *matA, unsigned int ldA, SNAPBASE_t *vecB, unsigned int ldB, lapack_int *jpvt, SNAPBASEREAL_t relative_tolerance, lapack_int *rank) {
  SNAPMTX_t vec_s;
  SNAPMTXREAL_t vecC, vec_c;
  SNAPBASE_t s;
  SNAPBASE_t tmpb;
  SNAPBASEREAL_t c,tmpr;
  SNAPBASEREAL_t *ptr1;
  SNAPBASE_t *pt1, *pt2, *pt5, *pt6;
  lapack_int *ptj;
#ifdef SNAP_SUPDC_T
  SNAPBASE_t *pt3, *pt4;
#endif // SNAP_SUPDC_T
  unsigned int i,j,k,r,cs_length;
  // init
  SNAPBASEREAL_init(tmpr);
  SNAPBASEREAL_init(c);
  SNAPBASE_init(tmpb);
  SNAPBASE_init(s);
  SNAPMTXREAL_init2(vecC, 1, cols);
  // first QR by givens.
  for ( i = 0, ptr1 = SNAPMTXREAL_elems(vecC); i < cols; i++, ptr1++ ) {
    SNAPBASEREAL_set_d(*ptr1, 0.0);
    for ( j = 0, pt1 = matA + i; j < rows; j++, pt1 += ldA ) {
      SNAPBASE_normsquare(tmpr, *pt1);
      SNAPBASEREAL_add(*ptr1, *ptr1, tmpr);
    }
  }
  // outer loop for columns.
  for ( r = 0; r < cols; r++ ) {
    // find the column of maximum norm.
    SNAPBASEREAL_set(tmpr, SNAPMTXREAL_elem(vecC, 0, r));
    k = r;
    for ( i = r + 1, ptr1 = SNAPMTXREAL_elemp(vecC, 0, r+1); i < cols; i++, ptr1++ ) {
      if ( SNAPBASEREAL_cmp(tmpr, *ptr1) < 0 ) {
        SNAPBASEREAL_set(tmpr, *ptr1);
        k = i;
      }
    }
    // swap the k-th and r-th columns.
    *(jpvt + r) = k + 1;
    SNAPBASEREAL_swap(SNAPMTXREAL_elem(vecC, 0, r), SNAPMTXREAL_elem(vecC, 0, k));
#if defined(SNAP_SUPD_T)
    cblas_dswap(rows, matA + r, ldA, matA + k, ldA);
#elif defined(SNAP_SUPDC_T)
    cblas_zswap(rows, matA + r, ldA, matA + k, ldA);
#else
    IMPLEMENT HERE.
#endif // defined
    // inner loop for rows.
    for ( i = r + 1, pt1 = matA + (rows - 2)*ldA + r, pt2 = matA + (rows - 1)*ldA + r, pt5 = vecB + (rows - 2)*ldB, pt6 = vecB + (rows - 1)*ldB; i < rows; i++, pt1-=ldA, pt2-=ldA, pt5-=ldB, pt6-=ldB ) {
#if defined(SNAP_SUPD_T)
      cblas_drotg(pt1, pt2, &c, &s);
      // apply the givens rotation to row vectors.
      cblas_drot(cols - r - 1, pt1 + 1, 1, pt2 + 1, 1, c, s);
      cblas_drot(1, pt5, 1, pt6, 1, c, s);
#elif defined(SNAP_SUPDC_T)
      cblas_zrotg(pt1, pt2, &c, &s);
      // apply the givens rotation to row vectors.
      for ( j = r + 1, pt3 = pt1 + 1, pt4 = pt2 + 1; j < cols; j++, pt3++, pt4++) {
        tmpb = c*(*pt3) + s*(*pt4);
        *pt4 = c*(*pt4) - s*(*pt3);
        *pt3 = tmpb;
      }
      tmpb = c*(*pt5) + s*(*pt6);
      *pt6 = c*(*pt6) - s*(*pt5);
      *pt5 = tmpb;
#else
      IMPLEMENT HERE.
#endif // defined
      SNAPBASE_set_d(*pt2, 0.0);
    }
    // column norm correction
    for ( i = r + 1, ptr1 = SNAPMTXREAL_elemp(vecC, 0, r+1), pt1 = matA + r*ldA + r + 1; i < cols; i++, ptr1++, pt1++ ) {
      SNAPBASE_normsquare(tmpr, *pt1);
      SNAPBASEREAL_sub(*ptr1, *ptr1, tmpr);
    }
  }
  /*
  #ifdef SNAPLIB_DEBUG
  #ifdef SNAP_SUPD_T
      printf("Q^t matA:\n");
      show_dlist2(matA, cols*rows, cols);
  #elif defined(SNAP_SUPDC_T)
      printf("Q^t matA:\n");
      show_dclist2(matA, cols*rows, cols);
  #endif // SNAP_SUPD_T
  #endif // SNAPLIB_DEBUG
  */
  // now "r" denotes the rank of matA!!!
  int fr_job, fr_j;
  unsigned int min_rows_cols;
  SNAPMTX_t minwork, maxwork, matAcol;
  SNAPBASE_t s1, s2, c1, c2;
  SNAPBASEREAL_t smin, smax, sminpr, smaxpr;
  if ( rows < cols ) {
    min_rows_cols = rows;
  } else {
    min_rows_cols = cols;
  }
  SNAPBASE_init(s1);
  SNAPBASE_init(s2);
  SNAPBASE_init(c1);
  SNAPBASE_init(c2);
  SNAPBASEREAL_init(smin);
  SNAPBASEREAL_init(smax);
  SNAPBASEREAL_init(sminpr);
  SNAPBASEREAL_init(smaxpr);
  SNAPMTX_init2(minwork, 1, min_rows_cols);
  SNAPMTX_init2(maxwork, 1, min_rows_cols);
  SNAPMTX_init2(matAcol, 1, min_rows_cols);
  SNAPBASE_set_ui(SNAPMTX_elem(minwork, 0, 0), 1);
  SNAPBASE_set_ui(SNAPMTX_elem(maxwork, 0, 0), 1);
  SNAPBASE_abs(smax, *matA);
  SNAPBASEREAL_set(smin, smax);
  if ( SNAPBASEREAL_cmp_ui(smax, 0) == 0 ) {
    r = 0;
  } else {
    r = 1;
    while ( r < min_rows_cols ) {
      for ( i = 0, pt1 = matA+r, pt2 = SNAPMTX_elems(matAcol); i < min_rows_cols; i++, pt1+=ldA, pt2++ ) {
        SNAPBASE_set(*pt2, *pt1);
      }
      fr_j = r;
      fr_job = 1;
#if defined(SNAP_SUPD_T)
      dlaic1_(&fr_job, &fr_j, SNAPMTX_elems(minwork), &smin, SNAPMTX_elems(matAcol), SNAPMTX_elemp(matAcol, 0, r), &sminpr, &s1, &c1);
#elif defined(SNAP_SUPDC_T)
      zlaic1_(&fr_job, &fr_j, SNAPMTX_elems(minwork), &smin, SNAPMTX_elems(matAcol), SNAPMTX_elemp(matAcol, 0, r), &sminpr, &s1, &c1);
#else
      IMPLEMENT HERE.
#endif // defined
      fr_job = 2;
#if defined(SNAP_SUPD_T)
      dlaic1_(&fr_job, &fr_j, SNAPMTX_elems(maxwork), &smin, SNAPMTX_elems(matAcol), SNAPMTX_elemp(matAcol, 0, r), &smaxpr, &s2, &c2);
#elif defined(SNAP_SUPDC_T)
      zlaic1_(&fr_job, &fr_j, SNAPMTX_elems(maxwork), &smin, SNAPMTX_elems(matAcol), SNAPMTX_elemp(matAcol, 0, r), &smaxpr, &s2, &c2);
#else
      IMPLEMENT HERE.
#endif // defined
      SNAPBASEREAL_mul(tmpr, smaxpr, relative_tolerance);
      if ( SNAPBASEREAL_cmp(tmpr, sminpr) < 0 ) {
        for ( i = 0, pt1 = SNAPMTX_elems(minwork), pt2 = SNAPMTX_elems(maxwork); i < r; i++, pt1++, pt2++ ) {
          SNAPBASE_mul( *pt1, s1, *pt1);
          SNAPBASE_mul( *pt2, s2, *pt2);
        }
        SNAPBASE_set(*pt1, c1);
        SNAPBASE_set(*pt2, c2);
        SNAPBASEREAL_set(smin, sminpr);
        SNAPBASEREAL_set(smax, smaxpr);
        r++;
      } else {
        break;
      }
    }
  }
  SNAPBASE_clear(s1);
  SNAPBASE_clear(s2);
  SNAPBASE_clear(c1);
  SNAPBASE_clear(c2);
  SNAPBASEREAL_clear(smin);
  SNAPBASEREAL_clear(smax);
  SNAPBASEREAL_clear(sminpr);
  SNAPBASEREAL_clear(smaxpr);
  SNAPMTX_clear(minwork);
  SNAPMTX_clear(maxwork);
  SNAPMTX_clear(matAcol);
  // second QR to the columns
  if ( r < cols ) {
    cs_length = r*(r+1);
    cs_length /= 2;
    cs_length = r*cols - cs_length;
    SNAPMTXREAL_init2(vec_c, 1, cs_length);
    SNAPMTX_init2(vec_s, 1, cs_length);
    // outer loop for rows.
    for ( k = 0, ptr1 = SNAPMTXREAL_elems(vec_c), pt5 = SNAPMTX_elems(vec_s); k < r; k++ ) {
      // inner loop for columns.
      for ( i = k + 1, pt1 = matA + k*ldA + (cols - 2), pt2 = matA + k*ldA + (cols - 1); i < cols; i++, pt1--, pt2--, ptr1++, pt5++ ) {
#if defined(SNAP_SUPD_T)
        cblas_drotg(pt1, pt2, ptr1, pt5);
        // apply the givens rotation to row vectors.
        cblas_drot(r - k - 1, pt1 + ldA, ldA, pt2 + ldA, ldA, *ptr1, *pt5);
#elif defined(SNAP_SUPDC_T)
        cblas_zrotg(pt1, pt2, ptr1, pt5);
        // apply the givens rotation to row vectors.
        for ( j = k + 1, pt3 = pt1 + ldA, pt4 = pt2 + ldA; j < r; j++, pt3+=ldA, pt4+=ldA) {
          tmpb = (*ptr1)*(*pt3) + (*pt5)*(*pt4);
          *pt4 = (*ptr1)*(*pt4) - (*pt5)*(*pt3);
          *pt3 = tmpb;
        }
#else
        IMPLEMENT HERE.
#endif // defined
        SNAPBASE_set_d(*pt2, 0.0);
      }
    }
    // now we have T_11 as the top-left block of matA!!
    // update vecB as T_11^(-1) vecB <== but (r+1)-th ... rows-th are overwritten by zero.
#if defined(SNAP_SUPD_T)
    LAPACKE_dtrtrs(LAPACK_ROW_MAJOR, 'L', 'N', 'N', r, 1, matA, ldA, vecB, 1);
#elif defined(SNAP_SUPDC_T)
    LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'L', 'N', 'N', r, 1, matA, ldA, vecB, 1);
#else
    IMPLEMENT HERE.
#endif // defined
  } else {
#if defined(SNAP_SUPD_T)
    LAPACKE_dtrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', r, 1, matA, ldA, vecB, 1);
#elif defined(SNAP_SUPDC_T)
    LAPACKE_ztrtrs(LAPACK_ROW_MAJOR, 'U', 'N', 'N', r, 1, matA, ldA, vecB, 1);
#else
    IMPLEMENT HERE.
#endif // defined
  }
  for ( i = r, pt1 = vecB + r; i < cols; i++, pt1++ ) {
    SNAPBASE_set_d(*pt1, 0.0);
  }
  if ( r < cols ) {
    // outer loop for rows.
    for ( k = 0, ptr1 = SNAPMTXREAL_elemp(vec_c, 0, cs_length-1), pt5 = SNAPMTX_elemp(vec_s, 0, cs_length-1); k < r; k++ ) {
      // inner loop for columns.
      for ( i = 0, pt1 = vecB + (cols-2)*ldB, pt2 = vecB + (cols-1)*ldB, ptr1 = SNAPMTXREAL_elems(vec_c), pt5 = SNAPMTX_elems(vec_s); i < cols - r + k; i++, pt1-=ldB, pt2-=ldB, ptr1--, pt5-- ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        tmpb = (*ptr1)*(*pt1) - (*pt5)*(*pt2);
        *pt2 = (*ptr1)*(*pt2) + (*pt5)*(*pt1);
        *pt1 = tmpb;
#else
        IMPLEMENT HERE.
#endif // defined
      }
    }
  }
  // permute
  for ( i = 0, ptj = jpvt + r -1, pt1 = vecB + (r - 1)*ldB; i < r; i++, ptj--, pt1-=ldB ) {
    if ( *ptj != 0 ) {
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      pt2 = vecB + ((*ptj)-1)*ldB;
      tmpb = *pt1;
      *pt1 = *pt2;
      *pt2 = tmpb;
#else
      IMPLEMENT HERE.
#endif // defined
    }
  }
  *rank = r;
  if ( r < cols ) {
    SNAPMTX_clear(vec_s);
    SNAPMTXREAL_clear(vec_c);
  }
  SNAPMTXREAL_clear(vecC);
  SNAPBASE_clear(s);
  SNAPBASE_clear(tmpb);
  SNAPBASEREAL_clear(c);
  SNAPBASEREAL_clear(tmpr);
  return 0;
}

/*!
 *  @brief STLNGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the STLNGCD algorithm proposed by Lihong et.al.
 *   in 2004 and so on. (see MM Research Preprints, 375–387, No. 24, December 2004)
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning In the original paper, the modified GKO method is used for LS.
 *   However, we use LAPACK routine for LS since our fastgcd routine is not better than
 *   uvgcd in computational time and also there is no significant difference for not
 *   large matrices in the original paper on solving LS by the modified GKO method.
 *  @warning In this implementation, complex numbers are treated as is!!
 *   However, in the ISSAC paper, it not. In the 2004 and 2007 paper, it is.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] tol The tolerance.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_stlngcd(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t tol) {
  SNAPMTX_t Sk, Sk_tmp, rz, zx, YPk, vtmp;
  SNAPBASE_t *Ak, *bk, *vec_rz, *vec_r, *vec_tmp_z, *vec_zx, *vec_z, *vec_x;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance, w, mw, sigma_m1;
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_t rho;
#endif // SNAPLIB_STLN_USE_REFINEMENT;
  SNAPBASE_t one, mone;
  SNAPBASE_t *pt1, *pt2, *pt3;
  SNAP_t poly_f, poly_g;
  unsigned int ldAk, ldbk, cAk, rAk, rbk;
  unsigned int lenR, lenZ, lenX;
  unsigned int ui, uj, uk, umax;
  unsigned int loop_condition, realonly_case, loop_count;
  int result_code = 0;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // defined
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // initializations
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAPBASE_init_set_d(one, 1.0);
  SNAPBASE_init_set_d(mone, -1.0);
  SNAPBASEREAL_init_set_d(w, SNAPLIB_STLN_PENALTY_VALUE);
  SNAPBASEREAL_init_set_d(mw, -SNAPLIB_STLN_PENALTY_VALUE);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#if defined(SNAP_SUPDC_T)
  realonly_case = SNAPLIB_TRUE;
  for ( pt1 = SNAP_coefs(poly_f), ui = 0; realonly_case && ui <= SNAP_degree(poly_f); pt1++, ui++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*pt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  for ( pt1 = SNAP_coefs(poly_g), ui = 0; realonly_case && ui <= SNAP_degree(poly_g); pt1++, ui++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*pt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  if ( realonly_case ) {
    printf("realonly_case!!!\n");
    supdc_randpoly(poly_f, SNAP_degree(poly_f), tol, 1.0);
    for ( pt1 = SNAP_coefs(poly_f), pt2 = SNAP_coefs(poly_inf), ui = 0; ui <= SNAP_degree(poly_f); pt1++, pt2++, ui++ ) {
      *pt1 = cimag(*pt1)*_Complex_I + *pt2;
    }
    supdc_randpoly(poly_g, SNAP_degree(poly_g), tol, 1.0);
    for ( pt1 = SNAP_coefs(poly_g), pt2 = SNAP_coefs(poly_ing), ui = 0; ui <= SNAP_degree(poly_g); pt1++, pt2++, ui++ ) {
      *pt1 = cimag(*pt1)*_Complex_I + *pt2;
    }
  }
#else
  realonly_case = SNAPLIB_FALSE;
#endif // defined
  SNAPMTX_init2(Sk, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+2);
  SNAPMTX_init2(Sk_tmp, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+2);
  // r: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, z: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2
  SNAPMTX_init2(rz, 1, 2*SNAP_degree(poly_inf)+2*SNAP_degree(poly_ing)-d+3);
  lenR = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
  vec_rz = vec_r = SNAPMTX_elems(rz);
  vec_tmp_z = vec_rz + lenR;
  // z: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2, x: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1
  // we enlarge the memory space of x up to taht of r
  SNAPMTX_init2(zx, 1, 2*SNAP_degree(poly_inf)+2*SNAP_degree(poly_ing)-d+3);
  lenZ = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenX = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  vec_zx = vec_z = SNAPMTX_elems(zx);
  vec_x = vec_zx + lenZ;
  SNAPMTX_init2(vtmp, 1, lenX + lenZ);
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*(lenX + lenZ));
#endif // defined
  SNAPMTX_init2(YPk, lenR+lenZ, lenZ+lenX);
  // STEP1: [b_k A_k] := S_k(f,g), E_k = 0, h_k = 0
  // note: in STLNGCD, S_1 is Sylvester Matrix hence r = k - 1 for calling subresultant_mapping_matrix_dsc.
  SNAPMTX_subresultant_mapping_matrix_dsc(Sk, poly_f, poly_g, d-1);
  SNAPMTX_subresultant_mapping_matrix_dsc(Sk_tmp, poly_f, poly_g, d-1);
  ldAk = ldbk = SNAPMTX_csize(Sk);
  cAk = SNAPMTX_cols(Sk) - 1;
  rAk = rbk = SNAPMTX_rows(Sk);
  bk = SNAPMTX_elems(Sk);
  Ak = SNAPMTX_elemp(Sk, 0, 1);
#ifdef SNAPLIB_DEBUG
  printf("Sk(f,g):\n");
#ifdef SNAP_SUPD_T
  show_sdmd(Sk);
#elif defined(SNAP_SUPDC_T)
  show_sdmdc(Sk);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // STEP2: min||A_k x - b_k||_2, r = b_k - A_k x, Form P_k and Y_k
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  for ( jpvt_pt = jpvt, ui = 0; ui < lenX; jpvt_pt++, ui++ ) {
    *jpvt_pt = 0;
  }
  for ( pt1 = vec_x, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
#ifdef SNAPLIB_DEBUG
  printf("vec_x:\n");
#ifdef SNAP_SUPD_T
  show_dlist(vec_x, lenX);
#elif defined(SNAP_SUPDC_T)
  show_dclist(vec_x, lenX);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // solve min||A_k x - b_k||_2 and vec_x := x!!!
#ifdef SNAP_SUPD_T
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, rAk, cAk, 1, SNAPMTX_elemp(Sk_tmp, 0, 1), ldAk, vec_x, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, rAk, cAk, 1, SNAPMTX_elemp(Sk_tmp, 0, 1), ldAk, vec_x, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_zgelsy could not compute the solution.\n");
    goto post;
  }
#endif // SNAP_SUPD_T
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  SNAPMTX_clear(Sk_tmp);
#ifdef SNAPLIB_DEBUG
  printf("vec_x:\n");
#ifdef SNAP_SUPD_T
  show_dlist(vec_x, lenX);
#elif defined(SNAP_SUPDC_T)
  show_dclist(vec_x, lenX);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // r := bk - Ak x = bk - Ak.vec_x
  for ( pt1 = vec_r, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
#ifdef SNAP_SUPD_T
  cblas_dgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, mone, Ak, ldAk, vec_x, 1, one, vec_r, 1);
#elif defined(SNAP_SUPDC_T)
  cblas_zgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, &mone, Ak, ldAk, vec_x, 1, &one, vec_r, 1);
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
#ifdef SNAPLIB_DEBUG
  printf("vec_r:\n");
#ifdef SNAP_SUPD_T
  show_dlist(vec_r, lenR);
  printf("norm(vec_r)=%e\n", cblas_dnrm2(lenR, vec_r, 1));
#elif defined(SNAP_SUPDC_T)
  show_dclist(vec_r, lenR);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  if ( realonly_case ) {
    SNAPMTX_subresultant_mapping_matrix_dsc(Sk, poly_inf, poly_ing, d-1);
  }
#ifdef SNAPLIB_DEBUG
  printf("Sk(f,g):\n");
#ifdef SNAP_SUPD_T
  show_sdmd(Sk);
#elif defined(SNAP_SUPDC_T)
  show_sdmdc(Sk);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // STEP3: Repeat
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < SNAPLIB_STLN_MAXIMUM_LOOP ) {
#ifdef SNAPLIB_DEBUG
    printf(" ========== loop start (%d) ========== \n", loop_count);
#endif // SNAPLIB_DEBUG
    // form Pk and Yk
    // reset YPk
    umax = SNAPMTX_csize(YPk)*SNAPMTX_rsize(YPk);
    for ( pt1 = SNAPMTX_elems(YPk), ui = 0; ui < umax; pt1++, ui++ ) {
      SNAPBASE_set_d(*pt1, 0.0);
    }
    // the left-bottom block is I_{m+n+2}
    for ( pt1 = SNAPMTX_elemp(YPk, lenR, 0), ui = 0; ui < lenZ; ui++, pt1 += SNAPMTX_csize(YPk)+1 ) {
      SNAPBASE_set_d(*pt1, 1.0);
    }
    // the upper-right block is w(Ak+Ek)
    for ( uj = 0; uj < SNAP_degree(poly_ing)-d; uj++ ) {
//      for ( pt1 = SNAPMTX_elemp(YPk, uj + 1, lenZ + uj), pt2 = Ak + ldAk, pt3 = vec_z, ui = 0; ui <= SNAP_degree(poly_inf); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk, pt3++ ) {
//        SNAPBASE_add(*pt1, *pt2, *pt3);
//        SNAPBASE_mul_basereal(*pt1, *pt1, w);
      for ( pt1 = SNAPMTX_elemp(YPk, uj + 1, lenZ + uj), pt2 = Ak + ldAk, ui = 0; ui <= SNAP_degree(poly_inf); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    for ( ; uj < lenX; uj++ ) {
//      for ( pt1 = SNAPMTX_elemp(YPk, uj - SNAP_degree(poly_ing) + d, lenZ + uj), pt2 = Ak + SNAP_degree(poly_inf) - d, pt3 = vec_z + SNAP_degree(poly_inf) + 1, ui = 0; ui <= SNAP_degree(poly_ing); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk, pt3++ ) {
//        SNAPBASE_add(*pt1, *pt2, *pt3);
//        SNAPBASE_mul_basereal(*pt1, *pt1, w);
      for ( pt1 = SNAPMTX_elemp(YPk, uj - SNAP_degree(poly_ing) + d, lenZ + uj), pt2 = Ak + SNAP_degree(poly_inf) - d, ui = 0; ui <= SNAP_degree(poly_ing); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    // the upper-left block is w(Yk-Pk)
    for ( uj = 0; uj <= SNAP_degree(poly_inf); uj++ ) {
      pt1 = SNAPMTX_elemp(YPk, uj, uj);
      SNAPBASE_set_basereal(*pt1, mw);
      for ( pt1 += SNAPMTX_csize(YPk), pt2 = vec_x, ui = 0; ui < SNAP_degree(poly_ing) - d; pt1 += SNAPMTX_csize(YPk), pt2++, ui++ ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    for ( uj = 0; uj <= SNAP_degree(poly_ing); uj++ ) {
      for ( pt1 = SNAPMTX_elemp(YPk, uj, uj + SNAP_degree(poly_inf) + 1), pt2 = vec_x + SNAP_degree(poly_ing) - d, ui = 0; ui <= SNAP_degree(poly_inf) - d; pt1 += SNAPMTX_csize(YPk), pt2++, ui++ ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
#ifdef SNAPLIB_DEBUG
    printf("YPk:\n");
#ifdef SNAP_SUPD_T
    show_sdmd(YPk);
#elif defined(SNAP_SUPDC_T)
    show_sdmdc(YPk);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
    // form "-" [-wr, z]
    for ( pt1 = vec_r, ui = 0; ui < lenR; ui++, pt1++ ) {
      SNAPBASE_mul(*pt1, *pt1, w);
    }
    for ( pt1 = vec_tmp_z, pt2 = vec_z, ui = 0; ui < lenZ; ui++, pt1++, pt2++ ) {
      SNAPBASE_mul(*pt1, mone, *pt2);
    }
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
    printf(" - [-wr, z]:\n");
    show_dlist(vec_rz, lenR + lenZ);
#elif defined(SNAP_SUPDC_T)
    printf(" - [-wr, z]:\n");
    show_dclist(vec_rz, lenR + lenZ);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
    // STEP3-a: min|| .... ||_2
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    for ( jpvt_pt = jpvt, ui = 0; ui < lenX + lenZ; jpvt_pt++, ui++ ) {
      *jpvt_pt = 0;
    }
    // solve min|| .... ||_2 and vec_rz := [del_z, del_x]!!!
#ifdef SNAPLIB_STLN_USE_GIVENS
    if ( SNAP_gelsy(SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }
#else
#ifdef SNAP_SUPD_T
    /*
            if ( LAPACKE_dgels(LAPACK_ROW_MAJOR, 'N', SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1) != 0 ) {
              printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_dgels could not compute the solution.\n");
              goto post;
            }
    */
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }
#elif defined(SNAP_SUPDC_T)
    if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd), ": LAPACKE_zgelsy could not compute the solution.\n");
      goto post;
    }
#endif // SNAP_SUPD_T
#endif // SNAPLIB_STLN_USE_GIVENS
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3-b: x = x + del_x, z = z + del_z
    for ( pt1 = vec_zx, pt2 = vec_rz, ui = 0; ui < lenX + lenZ; pt1++, pt2++, ui++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
    printf("vec_delz vec_delx:\n");
    show_dlist(vec_rz, lenX + lenZ);
    printf("vec_z vec_x:\n");
    show_dlist(vec_zx, lenX + lenZ);
#elif defined(SNAP_SUPDC_T)
    printf("vec_delz vec_delx:\n");
    show_dclist(vec_rz, lenX + lenZ);
    printf("vec_z vec_x:\n");
    show_dclist(vec_zx, lenX + lenZ);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
    // STEP3-c: E_k, h_k from z, Y_k from x, A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    // actual: loop condition check and A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    SNAPMTX_BLAS_nrm2(norm_g, lenZ, vec_rz, 1);
    SNAPMTX_BLAS_nrm2(norm_f, lenX, vec_rz + lenZ, 1);
#ifdef SNAPLIB_DEBUG
    printf("norm_delz: %e, norm_delx: %e, tol: %e\n", norm_g, norm_f, tol);
#endif // SNAPLIB_DEBUG
    if ( SNAPBASEREAL_cmp(norm_f, tol) <= 0 && SNAPBASEREAL_cmp(norm_g, tol) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // A_k = A_k + E_k
    for ( uj = 0; uj < SNAP_degree(poly_ing)-d; uj++ ) {
      for ( pt2 = Ak + (uj + 1)*ldAk + uj, pt3 = vec_rz, uk = 0; uk <= SNAP_degree(poly_inf); uk++, pt2 += ldAk, pt3++ ) {
        SNAPBASE_add(*pt2, *pt2, *pt3);
      }
    }
    for ( ; uj < lenX; uj++ ) {
      for ( pt2 = Ak + (uj - SNAP_degree(poly_ing) + d)*ldAk + uj, pt3 = vec_rz + SNAP_degree(poly_inf) + 1, uk = 0; uk <= SNAP_degree(poly_ing); uk++, pt2 += ldAk, pt3++ ) {
        SNAPBASE_add(*pt2, *pt2, *pt3);
      }
    }
    // b_k = b_k + h_k
    for ( pt1 = bk, pt2 = vec_rz, ui = 0; ui <= SNAP_degree(poly_inf); pt1 += ldbk, pt2++, ui++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }
    //
    // r := bk - Ak x = bk - Ak.vec_x
    for ( pt1 = vec_r, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
      SNAPBASE_set(*pt1, *pt2);
    }
#ifdef SNAP_SUPD_T
    cblas_dgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, mone, Ak, ldAk, vec_x, 1, one, vec_r, 1);
#elif defined(SNAP_SUPDC_T)
    cblas_zgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, &mone, Ak, ldAk, vec_x, 1, &one, vec_r, 1);
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    for ( pt1 = SNAP_coefp(poly_inf, SNAP_degree(poly_inf)), pt2 = bk, pt3 = SNAPMTX_elems(vtmp), ui = 0; ui <= SNAP_degree(poly_inf); pt1--, pt2 += ldbk, pt3++, ui++ ) {
      SNAPBASE_sub(*pt3, *pt1, *pt2);
    }
    for ( pt1 = SNAP_coefp(poly_ing, SNAP_degree(poly_ing)), pt2 = Ak + SNAP_degree(poly_ing) - d, ui = 0; ui <= SNAP_degree(poly_ing); pt1--, pt2 += ldbk, pt3++, ui++ ) {
      SNAPBASE_sub(*pt3, *pt1, *pt2);
    }
    SNAPMTX_BLAS_nrm2(norm_g, SNAP_degree(poly_inf) + SNAP_degree(poly_ing) + 2, SNAPMTX_elems(vtmp), 1);
    SNAPMTX_BLAS_nrm2(norm_f, lenR, vec_r, 1);
#ifdef SNAPLIB_DEBUG
    printf("norm_diff: %e, norm_resi: %e, tol: %e\n", norm_g, norm_f, tol);
#endif // SNAPLIB_DEBUG
#ifdef SNAPLIB_STLN_RES_COND
    if ( SNAPBASEREAL_cmp(norm_f, tol) <= 0 && SNAPBASEREAL_cmp(norm_g, tol) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    } else {
      loop_condition = SNAPLIB_TRUE;
    }
#endif // SNAPLIB_STLN_RES_COND
#ifdef SNAPLIB_DEBUG
    printf("vec_r:\n");
#ifdef SNAP_SUPD_T
    show_dlist(vec_r, lenR);
    printf("norm(vec_r)=%e\n", cblas_dnrm2(lenR, vec_r, 1));
#elif defined(SNAP_SUPDC_T)
    show_dclist(vec_r, lenR);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
#ifdef SNAPLIB_DEBUG
    printf("Sk(f,g):\n");
#ifdef SNAP_SUPD_T
    show_sdmd(Sk);
#elif defined(SNAP_SUPDC_T)
    show_sdmdc(Sk);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
    // STEP3: until ||del_x||_2 <= tol and ||del_z||_2 <= tol
  }
  printf("STLN: #iterations: %u\n", loop_count);
#ifdef SNAPLIB_DEBUG
  if ( loop_count >= SNAPLIB_STLN_MAXIMUM_LOOP ) {
    printf("maximum loop reached!\n");
  }
#endif // SNAPLIB_DEBUG
  // STEP4: output polys formed by b_k and A_k
  for ( pt1 = SNAP_coefp(poly_f, SNAP_degree(poly_inf)), pt2 = bk, ui = 0; ui <= SNAP_degree(poly_inf); pt1--, pt2 += ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
  for ( pt1 = SNAP_coefp(poly_g, SNAP_degree(poly_ing)), pt2 = Ak + SNAP_degree(poly_ing) - d, ui = 0; ui <= SNAP_degree(poly_ing); pt1--, pt2 += ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
  printf("F[x]: ");
  show_dlist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dlist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#elif defined(SNAP_SUPDC_T)
  printf("F[x]: ");
  show_dclist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dclist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  SNAPBASEREAL_init(sigma_m1);
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_init(rho);
#endif // SNAPLIB_STLN_USE_REFINEMENT
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  for ( pt1 = tau, pt2 = zvect, pt3 = yvect, ui=0; ui<SNAPMTX_csize(Sk); pt1++, pt2++, pt3++, ui++ ) {
    SNAPBASE_init(*pt1);
    SNAPBASE_init(*pt2);
    SNAPBASE_init(*pt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: zgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: mlapack_frgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#endif // defined
  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(Sk, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(Sk, zvect, yvect, sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
  SNAP_self_neg(poly_f1);
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
  printf("zvect: ");
  show_dlist(zvect, SNAPMTX_csize(Sk));
  printf("F1[x]: ");
  show_dlist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
  printf("G1[x]: ");
  show_dlist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
#elif defined(SNAP_SUPDC_T)
  printf("zvect: ");
  show_dclist(zvect, SNAPMTX_csize(Sk));
  printf("F1[x]: ");
  show_dclist(SNAP_coefs(poly_f1), SNAP_degree(poly_f1)+1);
  printf("G1[x]: ");
  show_dclist(SNAP_coefs(poly_g1), SNAP_degree(poly_g1)+1);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
    printf("D[x]: ");
    show_dlist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#elif defined(SNAP_SUPDC_T)
    printf("D[x]: ");
    show_dclist(SNAP_coefs(poly_d), SNAP_degree(poly_d)+1);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG

//#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
//#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
#ifdef SNAPLIB_STLN_USE_REFINEMENT
    if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
    ) {
      result_code = -1;
    }
#endif // SNAPLIB_STLN_USE_REFINEMENT
  } else {
    result_code = -1;
  }
post_uvgcd:
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_clear(rho);
#endif // SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_clear(sigma_m1);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( pt1 = tau, pt2 = zvect, pt3 = yvect, ui=0; ui<SNAPMTX_csize(Sn); pt1++, pt2++, pt3++, ui++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
    SNAPBASE_clear(*pt3);
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(vtmp);
  SNAPMTX_clear(YPk);
  free(jpvt);
  SNAPMTX_clear(zx);
  SNAPMTX_clear(rz);
  SNAPMTX_clear(Sk);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(mw);
  SNAPBASEREAL_clear(w);
  SNAPBASE_clear(mone);
  SNAPBASE_clear(one);
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef(poly_d, 0, one);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

typedef enum { SNAP_STLNGCDEX_REFINEMENT_NONE, SNAP_STLNGCDEX_REFINEMENT_UVGCD, SNAP_STLNGCDEX_GIVENS, SNAP_STLNGCDEX_HOUSEHOLDER } mode_type;
/*!
 *  @brief STLNGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the STLNGCD algorithm proposed by Lihong et.al.
 *   in 2004 and so on. (see MM Research Preprints, 375–387, No. 24, December 2004)
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning In the original paper, the modified GKO method is used for LS.
 *   However, we use LAPACK routine for LS since our fastgcd routine is not better than
 *   uvgcd in computational time and also there is no significant difference for not
 *   large matrices in the original paper on solving LS by the modified GKO method.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0. Use givens if (mode_opt/10) is 1.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_stlngcd_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  SNAPMTX_t Sk, Sk_tmp, rz, zx, YPk;
  SNAPBASE_t *Ak, *bk, *vec_rz, *vec_r, *vec_tmp_z, *vec_zx, *vec_z, *vec_x;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance, w, mw, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t one, mone;
  SNAPBASE_t *pt1, *pt2, *pt3;
  SNAP_t poly_f, poly_g;
  unsigned int ldAk, ldbk, cAk, rAk, rbk;
  unsigned int lenR, lenZ, lenX;
  unsigned int ui, uj, uk, umax;
  unsigned int loop_condition, realonly_case, loop_count;
  int result_code = 0;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // defined
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // initializations
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  mode_type mode_refinement, mode_transform;
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_UVGCD;
  }
  if ( mode_opt / 10 == 0 ) {
    mode_transform = SNAP_STLNGCDEX_HOUSEHOLDER;
  } else {
    mode_transform = SNAP_STLNGCDEX_GIVENS;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
  SNAPBASE_init_set_d(one, 1.0);
  SNAPBASE_init_set_d(mone, -1.0);
  SNAPBASEREAL_init_set_d(w, SNAPLIB_STLN_PENALTY_VALUE);
  SNAPBASEREAL_init_set_d(mw, -SNAPLIB_STLN_PENALTY_VALUE);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#if defined(SNAP_SUPDC_T)
  realonly_case = SNAPLIB_TRUE;
  for ( pt1 = SNAP_coefs(poly_f), ui = 0; realonly_case && ui <= SNAP_degree(poly_f); pt1++, ui++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*pt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  for ( pt1 = SNAP_coefs(poly_g), ui = 0; realonly_case && ui <= SNAP_degree(poly_g); pt1++, ui++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*pt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  if ( realonly_case ) {
    printf("realonly_case!!!\n");
    supdc_randpoly(poly_f, SNAP_degree(poly_f), threshold, 1.0);
    for ( pt1 = SNAP_coefs(poly_f), pt2 = SNAP_coefs(poly_inf), ui = 0; ui <= SNAP_degree(poly_f); pt1++, pt2++, ui++ ) {
      *pt1 = cimag(*pt1)*_Complex_I + *pt2;
    }
    supdc_randpoly(poly_g, SNAP_degree(poly_g), threshold, 1.0);
    for ( pt1 = SNAP_coefs(poly_g), pt2 = SNAP_coefs(poly_ing), ui = 0; ui <= SNAP_degree(poly_g); pt1++, pt2++, ui++ ) {
      *pt1 = cimag(*pt1)*_Complex_I + *pt2;
    }
  }
#else
  realonly_case = SNAPLIB_FALSE;
#endif // defined
  SNAPMTX_init2(Sk, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+2);
  SNAPMTX_init2(Sk_tmp, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+2);
  // r: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1, z: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2
  SNAPMTX_init2(rz, 1, 2*SNAP_degree(poly_inf)+2*SNAP_degree(poly_ing)-d+3);
  lenR = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
  vec_rz = vec_r = SNAPMTX_elems(rz);
  vec_tmp_z = vec_rz + lenR;
  // z: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2, x: SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1
  // we enlarge the memory space of x up to taht of r
  SNAPMTX_init2(zx, 1, 2*SNAP_degree(poly_inf)+2*SNAP_degree(poly_ing)-d+3);
  lenZ = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenX = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  vec_zx = vec_z = SNAPMTX_elems(zx);
  vec_x = vec_zx + lenZ;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*(lenX + lenZ));
#endif // defined
  SNAPMTX_init2(YPk, lenR+lenZ, lenZ+lenX);
  // STEP1: [b_k A_k] := S_k(f,g), E_k = 0, h_k = 0
  // note: in STLNGCD, S_1 is Sylvester Matrix hence r = k - 1 for calling subresultant_mapping_matrix_dsc.
  SNAPMTX_subresultant_mapping_matrix_dsc(Sk, poly_f, poly_g, d-1);
  SNAPMTX_subresultant_mapping_matrix_dsc(Sk_tmp, poly_f, poly_g, d-1);
  ldAk = ldbk = SNAPMTX_csize(Sk);
  cAk = SNAPMTX_cols(Sk) - 1;
  rAk = rbk = SNAPMTX_rows(Sk);
  bk = SNAPMTX_elems(Sk);
  Ak = SNAPMTX_elemp(Sk, 0, 1);
  // STEP2: min||A_k x - b_k||_2, r = b_k - A_k x, Form P_k and Y_k
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  for ( jpvt_pt = jpvt, ui = 0; ui < lenX; jpvt_pt++, ui++ ) {
    *jpvt_pt = 0;
  }
  for ( pt1 = vec_x, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
  // solve min||A_k x - b_k||_2 and vec_x := x!!!
#ifdef SNAP_SUPD_T
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, rAk, cAk, 1, SNAPMTX_elemp(Sk_tmp, 0, 1), ldAk, vec_x, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, rAk, cAk, 1, SNAPMTX_elemp(Sk_tmp, 0, 1), ldAk, vec_x, 1, jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd_ex), ": LAPACKE_zgelsy could not compute the solution.\n");
    goto post;
  }
#endif // SNAP_SUPD_T
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  SNAPMTX_clear(Sk_tmp);
  // r := bk - Ak x = bk - Ak.vec_x
  for ( pt1 = vec_r, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
#ifdef SNAP_SUPD_T
  cblas_dgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, mone, Ak, ldAk, vec_x, 1, one, vec_r, 1);
#elif defined(SNAP_SUPDC_T)
  cblas_zgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, &mone, Ak, ldAk, vec_x, 1, &one, vec_r, 1);
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  if ( realonly_case ) {
    SNAPMTX_subresultant_mapping_matrix_dsc(Sk, poly_inf, poly_ing, d-1);
  }
  // STEP3: Repeat
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < max_iterations ) {
    // form Pk and Yk
    // reset YPk
    umax = SNAPMTX_csize(YPk)*SNAPMTX_rsize(YPk);
    for ( pt1 = SNAPMTX_elems(YPk), ui = 0; ui < umax; pt1++, ui++ ) {
      SNAPBASE_set_d(*pt1, 0.0);
    }
    // the left-bottom block is I_{m+n+2}
    for ( pt1 = SNAPMTX_elemp(YPk, lenR, 0), ui = 0; ui < lenZ; ui++, pt1 += SNAPMTX_csize(YPk)+1 ) {
      SNAPBASE_set_d(*pt1, 1.0);
    }
    // the upper-right block is w(Ak+Ek)
    for ( uj = 0; uj < SNAP_degree(poly_ing)-d; uj++ ) {
      for ( pt1 = SNAPMTX_elemp(YPk, uj + 1, lenZ + uj), pt2 = Ak + ldAk, ui = 0; ui <= SNAP_degree(poly_inf); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    for ( ; uj < lenX; uj++ ) {
      for ( pt1 = SNAPMTX_elemp(YPk, uj - SNAP_degree(poly_ing) + d, lenZ + uj), pt2 = Ak + SNAP_degree(poly_inf) - d, ui = 0; ui <= SNAP_degree(poly_ing); ui++, pt1 += SNAPMTX_csize(YPk), pt2 += ldAk ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    // the upper-left block is w(Yk-Pk)
    for ( uj = 0; uj <= SNAP_degree(poly_inf); uj++ ) {
      pt1 = SNAPMTX_elemp(YPk, uj, uj);
      SNAPBASE_set_basereal(*pt1, mw);
      for ( pt1 += SNAPMTX_csize(YPk), pt2 = vec_x, ui = 0; ui < SNAP_degree(poly_ing) - d; pt1 += SNAPMTX_csize(YPk), pt2++, ui++ ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    for ( uj = 0; uj <= SNAP_degree(poly_ing); uj++ ) {
      for ( pt1 = SNAPMTX_elemp(YPk, uj, uj + SNAP_degree(poly_inf) + 1), pt2 = vec_x + SNAP_degree(poly_ing) - d, ui = 0; ui <= SNAP_degree(poly_inf) - d; pt1 += SNAPMTX_csize(YPk), pt2++, ui++ ) {
        SNAPBASE_mul_basereal(*pt1, *pt2, w);
      }
    }
    // form "-" [-wr, z]
    for ( pt1 = vec_r, ui = 0; ui < lenR; ui++, pt1++ ) {
      SNAPBASE_mul(*pt1, *pt1, w);
    }
    for ( pt1 = vec_tmp_z, pt2 = vec_z, ui = 0; ui < lenZ; ui++, pt1++, pt2++ ) {
      SNAPBASE_mul(*pt1, mone, *pt2);
    }
    // STEP3-a: min|| .... ||_2
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    for ( jpvt_pt = jpvt, ui = 0; ui < lenX + lenZ; jpvt_pt++, ui++ ) {
      *jpvt_pt = 0;
    }
    // solve min|| .... ||_2 and vec_rz := [del_z, del_x]!!!
    if ( mode_transform == SNAP_STLNGCDEX_GIVENS ) {
      if ( SNAP_gelsy(SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
        printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
        goto post;
      }
    } else {
#ifdef SNAP_SUPD_T
      if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
        printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
        goto post;
      }
#elif defined(SNAP_SUPDC_T)
      if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(YPk), SNAPMTX_cols(YPk), 1, SNAPMTX_elems(YPk), SNAPMTX_csize(YPk), vec_rz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
        printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd_ex), ": LAPACKE_zgelsy could not compute the solution.\n");
        goto post;
      }
#endif // SNAP_SUPD_T
    }
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3-b: x = x + del_x, z = z + del_z
    for ( pt1 = vec_zx, pt2 = vec_rz, ui = 0; ui < lenX + lenZ; pt1++, pt2++, ui++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }
    // STEP3-c: E_k, h_k from z, Y_k from x, A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    // actual: loop condition check and A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    SNAPMTX_BLAS_nrm2(norm_g, lenZ, vec_rz, 1);
    SNAPMTX_BLAS_nrm2(norm_f, lenX, vec_rz + lenZ, 1);
    if ( SNAPBASEREAL_cmp(norm_f, threshold) <= 0 && SNAPBASEREAL_cmp(norm_g, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // A_k = A_k + E_k
    for ( uj = 0; uj < SNAP_degree(poly_ing)-d; uj++ ) {
      for ( pt2 = Ak + (uj + 1)*ldAk + uj, pt3 = vec_rz, uk = 0; uk <= SNAP_degree(poly_inf); uk++, pt2 += ldAk, pt3++ ) {
        SNAPBASE_add(*pt2, *pt2, *pt3);
      }
    }
    for ( ; uj < lenX; uj++ ) {
      for ( pt2 = Ak + (uj - SNAP_degree(poly_ing) + d)*ldAk + uj, pt3 = vec_rz + SNAP_degree(poly_inf) + 1, uk = 0; uk <= SNAP_degree(poly_ing); uk++, pt2 += ldAk, pt3++ ) {
        SNAPBASE_add(*pt2, *pt2, *pt3);
      }
    }
    // b_k = b_k + h_k
    for ( pt1 = bk, pt2 = vec_rz, ui = 0; ui <= SNAP_degree(poly_inf); pt1 += ldbk, pt2++, ui++ ) {
      SNAPBASE_add(*pt1, *pt1, *pt2);
    }
    //
    // r := bk - Ak x = bk - Ak.vec_x
    for ( pt1 = vec_r, pt2 = bk, ui = 0; ui < lenR; pt1++, pt2+=ldbk, ui++ ) {
      SNAPBASE_set(*pt1, *pt2);
    }
#ifdef SNAP_SUPD_T
    cblas_dgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, mone, Ak, ldAk, vec_x, 1, one, vec_r, 1);
#elif defined(SNAP_SUPDC_T)
    cblas_zgemv(CblasRowMajor, CblasNoTrans, rAk, cAk, &mone, Ak, ldAk, vec_x, 1, &one, vec_r, 1);
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3: until ||del_x||_2 <= tol and ||del_z||_2 <= tol
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in agcd_stlngcd_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in agcd_stlngcd_ex: converged within %d iterations.\n", loop_count);
  }
  // STEP4: output polys formed by b_k and A_k
  for ( pt1 = SNAP_coefp(poly_f, SNAP_degree(poly_inf)), pt2 = bk, ui = 0; ui <= SNAP_degree(poly_inf); pt1--, pt2 += ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
  for ( pt1 = SNAP_coefp(poly_g, SNAP_degree(poly_ing)), pt2 = Ak + SNAP_degree(poly_ing) - d, ui = 0; ui <= SNAP_degree(poly_ing); pt1--, pt2 += ldbk, ui++ ) {
    SNAPBASE_set(*pt1, *pt2);
  }
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(Sk));
  for ( pt1 = tau, pt2 = zvect, pt3 = yvect, ui=0; ui<SNAPMTX_csize(Sk); pt1++, pt2++, pt3++, ui++ ) {
    SNAPBASE_init(*pt1);
    SNAPBASE_init(*pt2);
    SNAPBASE_init(*pt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: zgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: mlapack_frgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(Sk), SNAPMTX_cols(Sk), SNAPMTX_elems(Sk), SNAPMTX_csize(Sk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#endif // defined
  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(Sk, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(Sk, zvect, yvect, sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {

#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
      if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
      ) {
        result_code = -1;
      }
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( pt1 = tau, pt2 = zvect, pt3 = yvect, ui=0; ui<SNAPMTX_csize(Sn); pt1++, pt2++, pt3++, ui++ ) {
    SNAPBASE_clear(*pt1);
    SNAPBASE_clear(*pt2);
    SNAPBASE_clear(*pt3);
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(YPk);
  free(jpvt);
  SNAPMTX_clear(zx);
  SNAPMTX_clear(rz);
  SNAPMTX_clear(Sk);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(mw);
  SNAPBASEREAL_clear(w);
  SNAPBASE_clear(mone);
  SNAPBASE_clear(one);
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef(poly_d, 0, one);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

/* ====================== STLN-GCD in ISSAC 2006 ======================== */

/*!
 *  @brief Constructs the matrix H(s,t) s.t. H(s,t)(f,g)^t=Syl_r(f,g)(s,t)^t
 *  @param[out] matH The matrix H which should have the correct size. This does not fill zeros.
 *  @param[in] vecS The conjugate vector of s.
 *  @param[in] vecT The conjugate vector of t.
 *  @param[in] degF The degree of f.
 *  @param[in] degG The degree of g.
 *  @param[in] r The order of subresultant matrix (r=0 means Sylvester).
 */
int SNAP_agcd_stlngcd2_convolution_like_matrix(SNAPMTX_t matH, SNAPBASE_t *vecS, SNAPBASE_t *vecT, unsigned int degF, unsigned int degG, unsigned int r) {
  SNAPBASE_t *ptS, *ptT, *ptH;
  unsigned int i, j;
  for ( i = 0; i <= degF; i++ ) {
    for ( j = 0, ptH = SNAPMTX_elemp(matH, i, i), ptS = vecS; j < degG - r; j++, ptH += SNAPMTX_csize(matH), ptS++ ) {
#if defined(SNAP_SUPD_T)
      SNAPBASE_set(*ptH, *ptS);
#elif defined(SNAP_SUPDC_T)
      SNAPBASE_conj(*ptH, *ptS);
#else
      IMPLEMENT HERE!!
#endif
    }
  }
  for ( i = 0; i <= degG; i++ ) {
    for ( j = 0, ptH = SNAPMTX_elemp(matH, i, i + degF + 1), ptT = vecT; j < degF - r; j++, ptH += SNAPMTX_csize(matH), ptT++ ) {
#if defined(SNAP_SUPD_T)
      SNAPBASE_set(*ptH, *ptT);
#elif defined(SNAP_SUPDC_T)
      SNAPBASE_conj(*ptH, *ptT);
#else
      IMPLEMENT HERE!!
#endif
    }
  }
  return 0;
}

/*!
 *  @brief Build the matrix Ak that is the Subresultant matrix of order k but it lacks the specified column.
 *  @param[out] matAk The matrix Ak which should have the correct size. This does not fill zeros.
 *  @param[out] vec_b The vector b which should have the correct size. This does not fill zeros.
 *  @param[in] poly_f This function returns Syl_r(poly_f,poly_g).
 *  @param[in] poly_g This function returns Syl_r(poly_f,poly_g).
 *  @param[in] r The order of Subresultant mapping.
 *  @param[in] posB The column index s.t. matAk does not have this column.
 */
int SNAP_agcd_stlngcd2_init_Ak(SNAPMTXREAL_t matAk, SNAPBASEREAL_t *vec_b, SNAP_t poly_f, SNAP_t poly_g, unsigned int r, unsigned int posB) {
  SNAPBASEREAL_t *pt;
  SNAPBASE_t *bpt;
  unsigned int i, j, rcpos;
#if defined(SNAP_SUPDC_T)
  unsigned int rstep_b = SNAPMTXREAL_rows(matAk)/2;
  unsigned int rstep = SNAPMTXREAL_csize(matAk)*rstep_b;
  unsigned int cstep = SNAPMTXREAL_cols(matAk)/2;
#endif // defined
  // poly_f
  for ( rcpos=0, i=0; i<SNAP_degree(poly_g)-r; i++ ) {
    if ( i == posB ) {
      for ( j=0, pt=vec_b+i, bpt=SNAP_coefp(poly_f,SNAP_degree(poly_f)); j<=SNAP_degree(poly_f); j++, pt++,bpt-- ) {
#if defined(SNAP_SUPD_T)
        SNAPBASEREAL_set(*pt, *bpt);
#elif defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*pt, SNAPBASE_real(*bpt));
        SNAPBASEREAL_set(*(pt+rstep_b), SNAPBASE_imag(*bpt));
#else
        IMPLEMENT HERE!!
#endif
      }
    } else {
      for ( j=0, pt=SNAPMTXREAL_elemp(matAk, i, rcpos), bpt=SNAP_coefp(poly_f,SNAP_degree(poly_f)); j<=SNAP_degree(poly_f); j++, pt+=SNAPMTXREAL_csize(matAk),bpt-- ) {
#if defined(SNAP_SUPD_T)
        SNAPBASEREAL_set(*pt, *bpt);
#elif defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*pt, SNAPBASE_real(*bpt));
        SNAPBASEREAL_set(*(pt+rstep), SNAPBASE_imag(*bpt));
        SNAPBASEREAL_set(*(pt+rstep+cstep), SNAPBASE_real(*bpt));
        SNAPBASEREAL_neg(*(pt+cstep), SNAPBASE_imag(*bpt));
#else
        IMPLEMENT HERE!!
#endif
      }
      rcpos++;
    }
  }
  // poly_g
  for ( i=0; i<SNAP_degree(poly_f)-r; i++ ) {
    if ( SNAP_degree(poly_g)-r+i == posB ) {
      for ( j=0, pt=vec_b+i, bpt=SNAP_coefp(poly_g,SNAP_degree(poly_g)); j<=SNAP_degree(poly_g); j++, pt++,bpt-- ) {
#if defined(SNAP_SUPD_T)
        SNAPBASEREAL_set(*pt, *bpt);
#elif defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*pt, SNAPBASE_real(*bpt));
        SNAPBASEREAL_set(*(pt+rstep_b), SNAPBASE_imag(*bpt));
#else
        IMPLEMENT HERE!!
#endif
      }
    } else {
      for ( j=0, pt=SNAPMTXREAL_elemp(matAk, i, rcpos), bpt=SNAP_coefp(poly_g,SNAP_degree(poly_g)); j<=SNAP_degree(poly_g); j++, pt+=SNAPMTXREAL_csize(matAk),bpt-- ) {
#if defined(SNAP_SUPD_T)
        SNAPBASEREAL_set(*pt, *bpt);
#elif defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*pt, SNAPBASE_real(*bpt));
        SNAPBASEREAL_set(*(pt+rstep), SNAPBASE_imag(*bpt));
        SNAPBASEREAL_set(*(pt+rstep+cstep), SNAPBASE_real(*bpt));
        SNAPBASEREAL_neg(*(pt+cstep), SNAPBASE_imag(*bpt));
#else
        IMPLEMENT HERE!!
#endif
      }
      rcpos++;
    }
  }
  return 0;
}

/*!
 *  @brief Update the matrix Ak that is the Subresultant matrix of order k but it lacks the specified column.
 *  @param[out] matAk The matrix Ak which should have the correct size. This does not fill zeros.
 *  @param[out] vec_b The vector b which should have the correct size. This does not fill zeros.
 *  @param[in] vecZ The coefficient update vector of (f_R g_R f_I g_I).
 *  @param[in] degF The degree of f.
 *  @param[in] degG The degree of g.
 *  @param[in] r The order of subresultant matrix (r=0 means Sylvester).
 *  @param[in] posB The column index s.t. matAk does not have this column.
 */
int SNAP_agcd_stlngcd2_update_Ak(SNAPMTXREAL_t matAk, SNAPBASEREAL_t *vec_b, SNAPBASEREAL_t *vecZ, unsigned int degF, unsigned int degG, unsigned int r, unsigned int posB) {
  SNAPBASEREAL_t *pt1, *ptF, *ptG;
  unsigned int i, j, rcpos;
#if defined(SNAP_SUPDC_T)
  unsigned int rstep_b = SNAPMTXREAL_rows(matAk)/2;
  unsigned int rstep = SNAPMTXREAL_csize(matAk)*rstep_b;
  unsigned int cstep = SNAPMTXREAL_cols(matAk)/2;
#endif // defined
  // reduce a compile warning
  ptF=vecZ;
  // poly_f
  for ( rcpos=0, i=0; i<degG-r; i++ ) {
    if ( i == posB ) {
      for ( j=0, pt1=vec_b+i, ptF=vecZ; j<=degF; j++, pt1++,ptF++ ) {
        SNAPBASEREAL_add(*pt1, *pt1, *ptF);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_add(*(pt1+rstep_b), *(pt1+rstep_b), *(ptF+degF+degG+2));
#endif
      }
    } else {
      for ( j=0, pt1=SNAPMTXREAL_elemp(matAk, i, rcpos), ptF=vecZ; j<=degF; j++, pt1+=SNAPMTXREAL_csize(matAk),ptF++ ) {
        SNAPBASEREAL_add(*pt1, *pt1, *ptF);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_add(*(pt1+rstep), *(pt1+rstep), *(ptF+degF+degG+2));
        SNAPBASEREAL_add(*(pt1+rstep+cstep), *(pt1+rstep+cstep), *ptF);
        SNAPBASEREAL_sub(*(pt1+cstep), *(pt1+cstep), *(ptF+degF+degG+2));
#endif
      }
      rcpos++;
    }
  }
  // poly_g
  for ( i=0; i<degF-r; i++ ) {
    if ( degG-r+i == posB ) {
      for ( j=0, pt1=vec_b+i, ptG=ptF; j<=degG; j++, pt1++,ptG++ ) {
        SNAPBASEREAL_add(*pt1, *pt1, *ptG);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_add(*(pt1+rstep_b), *(pt1+rstep_b), *(ptG+degF+degG+2));
#endif
      }
    } else {
      for ( j=0, pt1=SNAPMTXREAL_elemp(matAk, i, rcpos), ptG=ptF; j<=degG; j++, pt1+=SNAPMTXREAL_csize(matAk),ptG++ ) {
        SNAPBASEREAL_add(*pt1, *pt1, *ptG);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_add(*(pt1+rstep), *(pt1+rstep), *(ptG+degF+degG+2));
        SNAPBASEREAL_add(*(pt1+rstep+cstep), *(pt1+rstep+cstep), *ptG);
        SNAPBASEREAL_sub(*(pt1+cstep), *(pt1+cstep), *(ptG+degF+degG+2));
#endif
      }
      rcpos++;
    }
  }
  return 0;
}

/*!
 *  @brief Constructs the matrix YPk(s,t).
 *  @param[out] matYPk The matrix YPk which should have the correct size. This does not fill zeros.
 *  @param[in] vecX The vector of (vecS_R vecT_R vecS_I vecT_I)
 *  @param[in] degF The degree of f.
 *  @param[in] degG The degree of g.
 *  @param[in] r The order of subresultant matrix (r=0 means Sylvester).
 *  @param[in] posB The column index s.t. matAk does not have this column. This means -1 at this position of (s t) in YPk.
 */
int SNAP_agcd_stlngcd2_convolution_like_matrix_YPk(SNAPMTXREAL_t matYPk, SNAPBASEREAL_t *vecX, unsigned int degF, unsigned int degG, unsigned int r, unsigned int posB) {
  SNAPBASEREAL_t *ptS, *ptT, *ptYPk;
#if defined(SNAP_SUPDC_T)
  unsigned int rstep = SNAPMTXREAL_csize(matYPk)*(degF+degG-r);
  unsigned int cstep = degF+degG+2;
#endif // defined
  unsigned int i, j;
  for ( i = 0; i <= degF; i++ ) {
    for ( j = 0, ptYPk = SNAPMTXREAL_elemp(matYPk, i, i), ptS = vecX; j < degG - r; j++, ptYPk += SNAPMTX_csize(matYPk), ptS++ ) {
      if ( posB == j ) {
        SNAPBASEREAL_set_si(*ptYPk, -1);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set_si(*(ptYPk+rstep+cstep), -1);
#endif // defined
        ptS--;
      } else {
        SNAPBASEREAL_set(*ptYPk, *ptS);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*(ptYPk+rstep+cstep), *ptS);
#endif // defined
      }
    }
  }
  for ( i = 0; i <= degG; i++ ) {
    for ( j = 0, ptYPk = SNAPMTX_elemp(matYPk, i, i + degF + 1), ptT = ptS; j < degF - r; j++, ptYPk += SNAPMTX_csize(matYPk), ptT++ ) {
      if ( posB + r == degG + j ) {
        SNAPBASEREAL_set_si(*ptYPk, -1);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set_si(*(ptYPk+rstep+cstep), -1);
#endif // defined
        ptT--;
      } else {
        SNAPBASEREAL_set(*ptYPk, *ptT);
#if defined(SNAP_SUPDC_T)
        SNAPBASEREAL_set(*(ptYPk+rstep+cstep), *ptT);
#endif // defined
      }
    }
  }
#if defined(SNAP_SUPDC_T)
  for ( i = 0; i <= degF; i++ ) {
    for ( j = 0, ptYPk = SNAPMTXREAL_elemp(matYPk, i, i), ptS = ptT; j < degG - r; j++, ptYPk += SNAPMTX_csize(matYPk), ptS++ ) {
      if ( posB == j ) {
        SNAPBASEREAL_set_si(*(ptYPk+rstep), 0);
        SNAPBASEREAL_set_si(*(ptYPk+cstep), 0);
        ptS--;
      } else {
        SNAPBASEREAL_set(*(ptYPk+rstep), *ptS);
        SNAPBASEREAL_neg(*(ptYPk+cstep), *ptS);
      }
    }
  }
  for ( i = 0; i <= degG; i++ ) {
    for ( j = 0, ptYPk = SNAPMTX_elemp(matYPk, i, i + degF + 1), ptT = ptS; j < degF - r; j++, ptYPk += SNAPMTX_csize(matYPk), ptT++ ) {
      if ( posB + r == degG + j ) {
        SNAPBASEREAL_set_si(*(ptYPk+rstep), 0);
        SNAPBASEREAL_set_si(*(ptYPk+cstep), 0);
        ptT--;
      } else {
        SNAPBASEREAL_set(*(ptYPk+rstep), *ptT);
        SNAPBASEREAL_neg(*(ptYPk+cstep), *ptT);
      }
    }
  }
#endif // defined
  return 0;
}

/*============ IMPLEMENTATION NOTES ============*/
/*
  v := smallest singular vector of Syl_r(f,g).
  H(h) s.t. H(h).(f g)^T = Syl_r(f,g).(h)
  initial z = -H(v)^T.(H(v).H(v)^T)^(-1).Syl_r(f,g).v
  t := the position of the maximum of absolute values in v.
  initial x = v/(-v[t]).
  Syl_r(f,g) = (A_1 b A_2) where b is the t-th column.
  A = (A_1 A_2).

  how to compute z, x, b and A.
  z := v; H := H(v); HH := H.H^T;
  z := Syl_r(f,g).z;
  solve HH.x = z and z := x;
  z := -H^T.z;
  x := v/(-v[t]);
  b := Syl_r(f,g)[t];
  A := (A_1 A_2);

  complex case note:
  - use complex as is, for the initial vector
  - use real expression, for the loop

*/
/*!
 *  @brief STLNGCD algorithm for two univariate polynomials.
 *  @details This is an implementation of the STLNGCD algorithm proposed by Kaltofen et.al.
 *   in 2006. (see ISSAC 2006, pp. 169-176).
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning In the original paper, the modified GKO method is used for LS.
 *   However, we use LAPACK routine for LS since our fastgcd routine is not better than
 *   uvgcd in computational time and also there is no significant difference for not
 *   large matrices in the original paper on solving LS by the modified GKO method.
 *  @warning In the original paper, the term "the first singular vector" is used but
 *   its meaning is the singular vector corresponding to the smallest singular value
 *   according to the preliminary implementation by Lihong. Hence in our implementation,
 *   we use the singular vector corresponding to the smallest singular value for that.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] tol The tolerance.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_stlngcd2(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t tol) {
  // local variables
  SNAPMTX_t matSk, matVT, matH, matHH;
  SNAPMTXREAL_t vecS, vecZX, vecRDZ, matAk, vecB, matYPk;
  SNAPBASE_t *bpt1, *bpt2, *bpt3;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance, w, sigma_m1;
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_t rho;
#endif // SNAPLIB_STLN_USE_REFINEMENT;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g;
#if defined(SNAP_SUPDC_T)
  SNAPBASE_t one, mone, zero, tempComplex;
#endif // defined
  SNAPBASEREAL_t *vec_zx, *vec_z, *vec_x, *vec_rdz, *vec_r, *vec_dz, *vec_b;
  SNAPBASEREAL_t *pt1, *pt2;
  int result_code = 0;
  unsigned int i, imax, j;
  unsigned int lenZinC, lenZinR, lenXinC, lenXinR, lenRinC, lenRinR;
  unsigned int posB;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // defined
  unsigned int loop_condition, loop_count;
#if defined(SNAP_SUPDC_T)
  unsigned int realonly_case;
#endif // defined
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // initializations
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  // initialization
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
#if defined(SNAP_SUPD_T)
  lenZinR = lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenXinR = lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenRinR = lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
#elif defined(SNAP_SUPDC_T)
  lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenZinR = lenZinC*2;
  lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenXinR = lenXinC*2;
  lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
  lenRinR = lenRinC*2;
  SNAPBASE_init_set_si(one, 1);
  SNAPBASE_init_set_si(mone, -1);
  SNAPBASE_init_set_si(zero, 0);
  SNAPBASE_init(tempComplex);
#else
  IMPLEMENT HERE!!
#endif
  SNAPMTX_init2(matSk, lenRinC, lenXinC+1);
  SNAPMTXREAL_init2(vecS, 1, lenXinC+1);
  SNAPMTX_init2(matVT, lenXinC+1, lenXinC+1);
  SNAPMTX_init2(matH, lenRinC, lenZinC);
  SNAPMTX_init2(matHH, lenRinC, lenRinC);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#if defined(SNAP_SUPDC_T)
  realonly_case = SNAPLIB_TRUE;
  for ( bpt1 = SNAP_coefs(poly_f), i = 0; realonly_case && i <= SNAP_degree(poly_f); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  for ( bpt1 = SNAP_coefs(poly_g), i = 0; realonly_case && i <= SNAP_degree(poly_g); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  if ( realonly_case ) {
    printf("realonly_case!!!\n");
    supdc_randpoly(poly_f, SNAP_degree(poly_f), tol, 1.0);
    for ( bpt1 = SNAP_coefs(poly_f), bpt2 = SNAP_coefs(poly_inf), i = 0; i <= SNAP_degree(poly_f); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
    supdc_randpoly(poly_g, SNAP_degree(poly_g), tol, 1.0);
    for ( bpt1 = SNAP_coefs(poly_g), bpt2 = SNAP_coefs(poly_ing), i = 0; i <= SNAP_degree(poly_g); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
  }
#endif // defined
  SNAPMTXREAL_init2(vecZX, 1, lenZinR+lenXinR);
  vec_zx = vec_z = SNAPMTXREAL_elems(vecZX);
  vec_x = vec_zx + lenZinR;
  SNAPMTXREAL_init2(vecRDZ, 1, lenRinR+lenZinR);
  vec_rdz = vec_r = SNAPMTXREAL_elems(vecRDZ);
  vec_dz = vec_rdz + lenRinR;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*(lenXinR + lenZinR));
#endif // defined
  SNAPMTXREAL_init2(matAk, lenRinR, lenXinR);
  SNAPMTXREAL_init2(vecB, 1, lenRinR);
  vec_b = SNAPMTXREAL_elems(vecB);
  SNAPBASEREAL_init_set_d(w, SNAPLIB_STLN_PENALTY_VALUE);
  SNAPMTXREAL_init2(matYPk, lenRinR+lenZinR, lenZinR+lenXinR);
  /* ====================== initial vectos by Lagrangian multipliers ====================== */
  // Sk(f,g)
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);

#ifdef SNAPLIB_DEBUG
  printf("matSk(f,g):\n");
#ifdef SNAP_SUPD_T
  show_sdmd(matSk);
#elif defined(SNAP_SUPDC_T)
  show_sdmdc(matSk);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG

  // Construct Sk = U.S.VT as U in Sk, S in vecS and VT in matVT.
  // matVT[-1]^h means the singular vector "v" will be used.
  // Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#else
  IMPLEMENT HERE!!
#endif
    printf("ERROR in agcd_stlngcd2: Xgesdd could not compute the SVD.\n");
    result_code = -1;
    goto post;
  }
  // Construct H(v)
  SNAP_agcd_stlngcd2_convolution_like_matrix(matH, SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1);

#ifdef SNAPLIB_DEBUG
  printf("matH(v):\n");
#ifdef SNAP_SUPD_T
  show_sdmd(matH);
#elif defined(SNAP_SUPDC_T)
  show_sdmdc(matH);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG

  // Compute HH as H(v).H(v)^h
#if defined(SNAP_SUPD_T)
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), 1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), 0.0, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#elif defined(SNAP_SUPDC_T)
  cblas_zgemm(CblasRowMajor, CblasNoTrans, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), &one, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), &zero, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#else
  IMPLEMENT HERE!!
#endif
  // Construct the initial vector X
  posB = 0;
  for ( i=1, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), bpt2=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 1); i<=lenXinC; i++, bpt2++ ) {
    if ( SNAPBASE_cmpabs(*bpt1, *bpt2) < 0 ) {
      posB = i;
      bpt1 = bpt2;
    }
  }
  bpt2 = SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, posB);
  for ( i=0, pt1=vec_x, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0); i<=lenXinC; i++, pt1++, bpt1++ ) {
    if ( i == posB ) {
      pt1--;
      continue;
    }
#if defined(SNAP_SUPD_T)
    SNAPBASEREAL_neg(*pt1, *bpt1);
    SNAPBASEREAL_div(*pt1, *pt1, *bpt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_neg(tempComplex, *bpt1);
    SNAPBASE_div(tempComplex, tempComplex, *bpt2);
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(tempComplex));
    SNAPBASEREAL_set(*(pt1+lenXinC), SNAPBASE_imag(tempComplex));
#else
    IMPLEMENT HERE!!
#endif
  }
  // Compute Sk[;-1] := Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
  pt1 = SNAPMTXREAL_elemp(vecS, 0, SNAPMTX_cols(matSk) -1);
  for ( i=0, bpt1=SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1); i<lenRinC; i++, bpt1+=SNAPMTX_csize(matSk)) {
    SNAPBASE_mul_basereal(*bpt1, *bpt1, *pt1);
  }
  // Solve solve HH.x = z and z := x; stored in U[;-1]!!
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_rows(matSk); i++, jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2), ": LAPACKE_zgelsy could not compute the solution.\n");
    goto post;
  }
#endif // SNAP_SUPD_T
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  // Compute vec_z := -H^T.z; as -matH^t * U[;-1]!!
#if defined(SNAP_SUPD_T)
  cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), -1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), 0.0, vec_z, 1);
#elif defined(SNAP_SUPDC_T)
  cblas_zgemv(CblasRowMajor, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), &mone, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), &zero, SNAPMTX_elems(matVT), 1);
  for ( i=0, bpt1=SNAPMTX_elems(matVT), pt1=vec_z; i<lenZinC; i++, bpt1++, pt1++ ) {
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(*bpt1));
    SNAPBASEREAL_set(*(pt1+lenZinC), SNAPBASE_imag(*bpt1));
  }
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  /* ====================== preparation for the main loop ====================== */
  // NEED (done): matAk, vec_z, vec_x, vec_b, vec_r
  SNAP_agcd_stlngcd2_init_Ak(matAk, vec_b, poly_inf, poly_ing, d-1, posB);
  SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_z, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
  // vec_r := bk - Ak x = vec_b - matAk.vec_x
  for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt1, *pt2);
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T

#ifdef SNAPLIB_DEBUG
  printf("vec_z:\n");
  show_dlist(vec_z, lenZinR);
#endif // SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
  printf("vec_x(without posB=%d):\n", posB);
  show_dlist(vec_x, lenXinR);
#endif // SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
  printf("vec_r:\n");
  show_dlist(vec_r, lenRinR);
  printf("norm(vec_r)=%e\n", cblas_dnrm2(lenRinR, vec_r, 1));
#endif // SNAPLIB_DEBUG

  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < SNAPLIB_STLN_MAXIMUM_LOOP ) {
    // NEED (outside): matAk, vec_z, vec_x, vec_b, vec_r
    // NEED (inside): matYPk, vec_rdz
    // reset YPk
    imax = SNAPMTXREAL_csize(matYPk)*SNAPMTXREAL_rsize(matYPk);
    for ( i=0, pt1=SNAPMTXREAL_elems(matYPk); i < imax; i++,pt1++ ) {
      SNAPBASEREAL_set_d(*pt1, 0.0);
    }
    // the left-bottom block is I_{m+n+2}
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, lenRinR, 0); i < lenZinR; i++, pt1+=SNAPMTXREAL_csize(matYPk)+1 ) {
      SNAPBASEREAL_set_d(*pt1, 1.0);
    }
    // the upper-right block is w(Ak+Ek)=w*matAk
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, 0, lenZinR), pt2=SNAPMTXREAL_elems(matAk); i < lenRinR; i++, pt1+=lenZinR ) {
      for ( j=0; j < lenXinR; j++, pt1++, pt2++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt2, w);
      }
    }
    // the upper-left block is w(Yk-Pk)
    SNAP_agcd_stlngcd2_convolution_like_matrix_YPk(matYPk, vec_x, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    for ( i=0; i < lenRinR; i++ ) {
      for ( j=0, pt1=SNAPMTXREAL_elemp(matYPk, i, 0); j < lenZinR; j++, pt1++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
      }
    }
    // form "-" [-wr, z]
    for ( i=0, pt1=vec_r; i < lenRinR; i++,pt1++ ) {
      SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
    }
    for ( i=0, pt1=vec_dz, pt2=vec_z; i < lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_mul_si(*pt1, *pt2, -1);
    }

#ifdef SNAPLIB_DEBUG
    printf("YPk:\n");
    show_sdmd(matYPk);
#endif // SNAPLIB_DEBUG

#ifdef SNAPLIB_DEBUG
    printf(" - [-wr, z]:\n");
    show_dlist(vec_rdz, lenRinR + lenZinR);
#endif // SNAPLIB_DEBUG

    // min|| YPk.(dz dx) + (-w*r z) ||_2 by QR as LS
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTXREAL_cols(matYPk); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    // solve min|| .... ||_2 and vec_rdz := [del_z, del_x]!!!
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTXREAL_rows(matYPk), SNAPMTXREAL_cols(matYPk), 1, SNAPMTXREAL_elems(matYPk), SNAPMTXREAL_csize(matYPk), vec_rdz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3-b: x = x + del_x, z = z + del_z
    for ( i=0, pt1=vec_zx, pt2=vec_rdz; i < lenXinR + lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_add(*pt1, *pt1, *pt2);
    }

#ifdef SNAPLIB_DEBUG
    printf("vec_z vec_x:\n");
    show_dlist(vec_zx, lenXinR + lenZinR);
    printf("vec_delz vec_delx:\n");
    show_dlist(vec_rdz, lenXinR + lenZinR);
#endif // SNAPLIB_DEBUG

    // STEP3-c: E_k, h_k from z, Y_k from x, A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    // actual: loop condition check and A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    SNAPMTXREAL_BLAS_nrm2(norm_g, lenZinR, vec_rdz, 1);
    SNAPMTXREAL_BLAS_nrm2(norm_f, lenXinR, vec_rdz + lenZinR, 1);

#ifdef SNAPLIB_DEBUG
    printf("norm_delz: %e, norm_delx: %e, tol: %e\n", norm_g, norm_f, tol);
#endif // SNAPLIB_DEBUG

    if ( SNAPBASEREAL_cmp(norm_f, tol) <= 0 && SNAPBASEREAL_cmp(norm_g, tol) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // A_k = A_k + E_k
    // b_k = b_k + h_k
    SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_rdz, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    // vec_r := bk - Ak x = vec_b - matAk.vec_x
    for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T

#ifdef SNAPLIB_DEBUG
    printf("vec_r:\n");
    show_dlist(vec_r, lenRinR);
    printf("norm(vec_r)=%e\n", cblas_dnrm2(lenRinR, vec_r, 1));
#endif // SNAPLIB_DEBUG

    // STEP3: until ||del_x||_2 <= tol and ||del_z||_2 <= tol
  }
  printf("STLN: #iterations: %u\n", loop_count);
#ifdef SNAPLIB_DEBUG
  if ( loop_count >= SNAPLIB_STLN_MAXIMUM_LOOP ) {
    printf("maximum loop reached!\n");
  }
#endif // SNAPLIB_DEBUG
  // STEP4: output polys formed by b_k and A_k
  for ( i=0, bpt1=SNAP_coefp(poly_f, SNAP_degree(poly_inf)), bpt2=SNAP_coefp(poly_inf, SNAP_degree(poly_inf)), pt2=vec_z; i <= SNAP_degree(poly_inf); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }
  for ( i=0, bpt1=SNAP_coefp(poly_g, SNAP_degree(poly_ing)), bpt2=SNAP_coefp(poly_ing, SNAP_degree(poly_ing)); i <= SNAP_degree(poly_ing); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }

#ifdef SNAPLIB_DEBUG
#ifdef SNAP_SUPD_T
  printf("F[x]: ");
  show_dlist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dlist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#elif defined(SNAP_SUPDC_T)
  printf("F[x]: ");
  show_dclist(SNAP_coefs(poly_f), SNAP_degree(poly_f)+1);
  printf("G[x]: ");
  show_dclist(SNAP_coefs(poly_g), SNAP_degree(poly_g)+1);
#endif // SNAP_SUPD_T
#endif // SNAPLIB_DEBUG

  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, bpt1=SNAPMTX_elems(matSk); i < SNAPMTX_csize(matSk)*SNAPMTX_rsize(matSk); i++,bpt1++ ) {
    SNAPBASE_set_d(*bpt1, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_init(rho);
#endif // SNAPLIB_STLN_USE_REFINEMENT
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_init(*bpt1);
    SNAPBASE_init(*bpt2);
    SNAPBASE_init(*bpt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: zgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: mlapack_frgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#endif // defined
  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {

    //#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
    //#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
#ifdef SNAPLIB_STLN_USE_REFINEMENT
    if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
      SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
    ) {
      result_code = -1;
    }
#endif // SNAPLIB_STLN_USE_REFINEMENT
  } else {
    result_code = -1;
  }
post_uvgcd:
#ifdef SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_clear(rho);
#endif // SNAPLIB_STLN_USE_REFINEMENT
  SNAPBASEREAL_clear(sigma_m1);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_clear(*bpt1);
    SNAPBASE_clear(*bpt2);
    SNAPBASE_clear(*bpt3);
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSk);
  SNAPMTX_clear(matVT);
  SNAPMTX_clear(matH);
  SNAPMTX_clear(matHH);
  SNAPMTXREAL_clear(vecS);
  SNAPMTXREAL_clear(vecZX);
  SNAPMTXREAL_clear(vecRDZ);
  SNAPMTXREAL_clear(matAk);
  SNAPMTXREAL_clear(vecB);
  SNAPMTXREAL_clear(matYPk);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(w);
#if defined(SNAP_SUPDC_T)
  SNAPBASE_clear(one);
  SNAPBASE_clear(mone);
  SNAPBASE_clear(zero);
  SNAPBASE_clear(tempComplex);
#endif // defined
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  free(jpvt);
#endif // defined
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

/*!
 *  @brief STLNGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the STLNGCD algorithm proposed by Kaltofen et.al.
 *   in 2006. (see ISSAC 2006, pp. 169-176).
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning In the original paper, the modified GKO method is used for LS.
 *   However, we use LAPACK routine for LS since our fastgcd routine is not better than
 *   uvgcd in computational time and also there is no significant difference for not
 *   large matrices in the original paper on solving LS by the modified GKO method.
 *  @warning In the original paper, the term "the first singular vector" is used but
 *   its meaning is the singular vector corresponding to the smallest singular value
 *   according to the preliminary implementation by Lihong. Hence in our implementation,
 *   we use the singular vector corresponding to the smallest singular value for that.
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_stlngcd2_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  // local variables
  SNAPMTX_t matSk, matVT, matH, matHH;
  SNAPMTXREAL_t vecS, vecZX, vecRDZ, matAk, vecB, matYPk;
  SNAPBASE_t *bpt1, *bpt2, *bpt3;
  SNAPBASEREAL_t norm_f, norm_g, relative_tolerance, w, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g;
#if defined(SNAP_SUPDC_T)
  SNAPBASE_t one, mone, zero, tempComplex;
#endif // defined
  SNAPBASEREAL_t *vec_zx, *vec_z, *vec_x, *vec_rdz, *vec_r, *vec_dz, *vec_b;
  SNAPBASEREAL_t *pt1, *pt2;
  int result_code = 0;
  unsigned int i, imax, j;
  unsigned int lenZinC, lenZinR, lenXinC, lenXinR, lenRinC, lenRinR;
  unsigned int posB;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  lapack_int *jpvt, *jpvt_pt;
  lapack_int rank;
#endif // defined
  unsigned int loop_condition, loop_count;
#if defined(SNAP_SUPDC_T)
  unsigned int realonly_case;
#endif // defined
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // initializations
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  mode_type mode_refinement;
  // initialization
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_UVGCD;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
#if defined(SNAP_SUPD_T)
  lenZinR = lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenXinR = lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenRinR = lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
#elif defined(SNAP_SUPDC_T)
  lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenZinR = lenZinC*2;
  lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenXinR = lenXinC*2;
  lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
  lenRinR = lenRinC*2;
  SNAPBASE_init_set_si(one, 1);
  SNAPBASE_init_set_si(mone, -1);
  SNAPBASE_init_set_si(zero, 0);
  SNAPBASE_init(tempComplex);
#else
  IMPLEMENT HERE!!
#endif
  SNAPMTX_init2(matSk, lenRinC, lenXinC+1);
  SNAPMTXREAL_init2(vecS, 1, lenXinC+1);
  SNAPMTX_init2(matVT, lenXinC+1, lenXinC+1);
  SNAPMTX_init2(matH, lenRinC, lenZinC);
  SNAPMTX_init2(matHH, lenRinC, lenRinC);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAPBASEREAL_init_set_d(relative_tolerance, 1.0e-15);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
  if ( SNAPBASEREAL_cmp(norm_f, norm_g) < 0 ) {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_f);
  } else {
    SNAPBASEREAL_div(relative_tolerance, relative_tolerance, norm_g);
  }
#if defined(SNAP_SUPDC_T)
  realonly_case = SNAPLIB_TRUE;
  for ( bpt1 = SNAP_coefs(poly_f), i = 0; realonly_case && i <= SNAP_degree(poly_f); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  for ( bpt1 = SNAP_coefs(poly_g), i = 0; realonly_case && i <= SNAP_degree(poly_g); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  if ( realonly_case ) {
    printf("realonly_case!!!\n");
    supdc_randpoly(poly_f, SNAP_degree(poly_f), threshold, 1.0);
    for ( bpt1 = SNAP_coefs(poly_f), bpt2 = SNAP_coefs(poly_inf), i = 0; i <= SNAP_degree(poly_f); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
    supdc_randpoly(poly_g, SNAP_degree(poly_g), threshold, 1.0);
    for ( bpt1 = SNAP_coefs(poly_g), bpt2 = SNAP_coefs(poly_ing), i = 0; i <= SNAP_degree(poly_g); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
  }
#endif // defined
  SNAPMTXREAL_init2(vecZX, 1, lenZinR+lenXinR);
  vec_zx = vec_z = SNAPMTXREAL_elems(vecZX);
  vec_x = vec_zx + lenZinR;
  SNAPMTXREAL_init2(vecRDZ, 1, lenRinR+lenZinR);
  vec_rdz = vec_r = SNAPMTXREAL_elems(vecRDZ);
  vec_dz = vec_rdz + lenRinR;
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  jpvt = (lapack_int *) malloc(sizeof(lapack_int)*(lenXinR + lenZinR));
#endif // defined
  SNAPMTXREAL_init2(matAk, lenRinR, lenXinR);
  SNAPMTXREAL_init2(vecB, 1, lenRinR);
  vec_b = SNAPMTXREAL_elems(vecB);
  SNAPBASEREAL_init_set_d(w, SNAPLIB_STLN_PENALTY_VALUE);
  SNAPMTXREAL_init2(matYPk, lenRinR+lenZinR, lenZinR+lenXinR);
  /* ====================== initial vectos by Lagrangian multipliers ====================== */
  // Sk(f,g)
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);
  // Construct Sk = U.S.VT as U in Sk, S in vecS and VT in matVT.
  // matVT[-1]^h means the singular vector "v" will be used.
  // Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#else
  IMPLEMENT HERE!!
#endif
    printf("ERROR in agcd_stlngcd2_ex: Xgesdd could not compute the SVD.\n");
    result_code = -1;
    goto post;
  }
  // Construct H(v)
  SNAP_agcd_stlngcd2_convolution_like_matrix(matH, SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1);
  // Compute HH as H(v).H(v)^h
#if defined(SNAP_SUPD_T)
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), 1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), 0.0, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#elif defined(SNAP_SUPDC_T)
  cblas_zgemm(CblasRowMajor, CblasNoTrans, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), &one, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), &zero, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#else
  IMPLEMENT HERE!!
#endif
  // Construct the initial vector X
  posB = 0;
  for ( i=1, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), bpt2=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 1); i<=lenXinC; i++, bpt2++ ) {
    if ( SNAPBASE_cmpabs(*bpt1, *bpt2) < 0 ) {
      posB = i;
      bpt1 = bpt2;
    }
  }
  bpt2 = SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, posB);
  for ( i=0, pt1=vec_x, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0); i<=lenXinC; i++, pt1++, bpt1++ ) {
    if ( i == posB ) {
      pt1--;
      continue;
    }
#if defined(SNAP_SUPD_T)
    SNAPBASEREAL_neg(*pt1, *bpt1);
    SNAPBASEREAL_div(*pt1, *pt1, *bpt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_neg(tempComplex, *bpt1);
    SNAPBASE_div(tempComplex, tempComplex, *bpt2);
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(tempComplex));
    SNAPBASEREAL_set(*(pt1+lenXinC), SNAPBASE_imag(tempComplex));
#else
    IMPLEMENT HERE!!
#endif
  }
  // Compute Sk[;-1] := Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
  pt1 = SNAPMTXREAL_elemp(vecS, 0, SNAPMTX_cols(matSk) -1);
  for ( i=0, bpt1=SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1); i<lenRinC; i++, bpt1+=SNAPMTX_csize(matSk)) {
    SNAPBASE_mul_basereal(*bpt1, *bpt1, *pt1);
  }
  // Solve solve HH.x = z and z := x; stored in U[;-1]!!
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  for ( i=0, jpvt_pt=jpvt; i < SNAPMTX_rows(matSk); i++, jpvt_pt++ ) {
    *jpvt_pt = 0;
  }
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
    goto post;
  }
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgelsy(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), jpvt, relative_tolerance, &rank) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2_ex), ": LAPACKE_zgelsy could not compute the solution.\n");
    goto post;
  }
#endif // SNAP_SUPD_T
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  // Compute vec_z := -H^T.z; as -matH^t * U[;-1]!!
#if defined(SNAP_SUPD_T)
  cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), -1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), 0.0, vec_z, 1);
#elif defined(SNAP_SUPDC_T)
  cblas_zgemv(CblasRowMajor, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), &mone, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), &zero, SNAPMTX_elems(matVT), 1);
  for ( i=0, bpt1=SNAPMTX_elems(matVT), pt1=vec_z; i<lenZinC; i++, bpt1++, pt1++ ) {
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(*bpt1));
    SNAPBASEREAL_set(*(pt1+lenZinC), SNAPBASE_imag(*bpt1));
  }
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  /* ====================== preparation for the main loop ====================== */
  // NEED (done): matAk, vec_z, vec_x, vec_b, vec_r
  SNAP_agcd_stlngcd2_init_Ak(matAk, vec_b, poly_inf, poly_ing, d-1, posB);
  SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_z, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
  // vec_r := bk - Ak x = vec_b - matAk.vec_x
  for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt1, *pt2);
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < SNAPLIB_STLN_MAXIMUM_LOOP ) {
    // NEED (outside): matAk, vec_z, vec_x, vec_b, vec_r
    // NEED (inside): matYPk, vec_rdz
    // reset YPk
    imax = SNAPMTXREAL_csize(matYPk)*SNAPMTXREAL_rsize(matYPk);
    for ( i=0, pt1=SNAPMTXREAL_elems(matYPk); i < imax; i++,pt1++ ) {
      SNAPBASEREAL_set_d(*pt1, 0.0);
    }
    // the left-bottom block is I_{m+n+2}
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, lenRinR, 0); i < lenZinR; i++, pt1+=SNAPMTXREAL_csize(matYPk)+1 ) {
      SNAPBASEREAL_set_d(*pt1, 1.0);
    }
    // the upper-right block is w(Ak+Ek)=w*matAk
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, 0, lenZinR), pt2=SNAPMTXREAL_elems(matAk); i < lenRinR; i++, pt1+=lenZinR ) {
      for ( j=0; j < lenXinR; j++, pt1++, pt2++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt2, w);
      }
    }
    // the upper-left block is w(Yk-Pk)
    SNAP_agcd_stlngcd2_convolution_like_matrix_YPk(matYPk, vec_x, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    for ( i=0; i < lenRinR; i++ ) {
      for ( j=0, pt1=SNAPMTXREAL_elemp(matYPk, i, 0); j < lenZinR; j++, pt1++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
      }
    }
    // form "-" [-wr, z]
    for ( i=0, pt1=vec_r; i < lenRinR; i++,pt1++ ) {
      SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
    }
    for ( i=0, pt1=vec_dz, pt2=vec_z; i < lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_mul_si(*pt1, *pt2, -1);
    }
    // min|| YPk.(dz dx) + (-w*r z) ||_2 by QR as LS
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    for ( i=0, jpvt_pt=jpvt; i < SNAPMTXREAL_cols(matYPk); i++,jpvt_pt++ ) {
      *jpvt_pt = 0;
    }
    // solve min|| .... ||_2 and vec_rdz := [del_z, del_x]!!!
    if ( LAPACKE_dgelsy(LAPACK_ROW_MAJOR, SNAPMTXREAL_rows(matYPk), SNAPMTXREAL_cols(matYPk), 1, SNAPMTXREAL_elems(matYPk), SNAPMTXREAL_csize(matYPk), vec_rdz, 1, jpvt, relative_tolerance, &rank) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2_ex), ": LAPACKE_dgelsy could not compute the solution.\n");
      goto post;
    }
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3-b: x = x + del_x, z = z + del_z
    for ( i=0, pt1=vec_zx, pt2=vec_rdz; i < lenXinR + lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_add(*pt1, *pt1, *pt2);
    }
    // STEP3-c: E_k, h_k from z, Y_k from x, A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    // actual: loop condition check and A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    SNAPMTXREAL_BLAS_nrm2(norm_g, lenZinR, vec_rdz, 1);
    SNAPMTXREAL_BLAS_nrm2(norm_f, lenXinR, vec_rdz + lenZinR, 1);
    if ( SNAPBASEREAL_cmp(norm_f, threshold) <= 0 && SNAPBASEREAL_cmp(norm_g, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // A_k = A_k + E_k
    // b_k = b_k + h_k
    SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_rdz, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    // vec_r := bk - Ak x = vec_b - matAk.vec_x
    for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3: until ||del_x||_2 <= tol and ||del_z||_2 <= tol
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in agcd_stlngcd2_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in agcd_stlngcd2_ex: converged within %d iterations.\n", loop_count);
  }
  // STEP4: output polys formed by b_k and A_k
  for ( i=0, bpt1=SNAP_coefp(poly_f, SNAP_degree(poly_inf)), bpt2=SNAP_coefp(poly_inf, SNAP_degree(poly_inf)), pt2=vec_z; i <= SNAP_degree(poly_inf); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }
  for ( i=0, bpt1=SNAP_coefp(poly_g, SNAP_degree(poly_ing)), bpt2=SNAP_coefp(poly_ing, SNAP_degree(poly_ing)); i <= SNAP_degree(poly_ing); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, bpt1=SNAPMTX_elems(matSk); i < SNAPMTX_csize(matSk)*SNAPMTX_rsize(matSk); i++,bpt1++ ) {
    SNAPBASE_set_d(*bpt1, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_init(*bpt1);
    SNAPBASE_init(*bpt2);
    SNAPBASE_init(*bpt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: zgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: mlapack_frgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#endif // defined
  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
      if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
      ) {
        result_code = -1;
      }
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_clear(*bpt1);
    SNAPBASE_clear(*bpt2);
    SNAPBASE_clear(*bpt3);
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSk);
  SNAPMTX_clear(matVT);
  SNAPMTX_clear(matH);
  SNAPMTX_clear(matHH);
  SNAPMTXREAL_clear(vecS);
  SNAPMTXREAL_clear(vecZX);
  SNAPMTXREAL_clear(vecRDZ);
  SNAPMTXREAL_clear(matAk);
  SNAPMTXREAL_clear(vecB);
  SNAPMTXREAL_clear(matYPk);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(relative_tolerance);
  SNAPBASEREAL_clear(w);
#if defined(SNAP_SUPDC_T)
  SNAPBASE_clear(one);
  SNAPBASE_clear(mone);
  SNAPBASE_clear(zero);
  SNAPBASE_clear(tempComplex);
#endif // defined
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  free(jpvt);
#endif // defined
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

/*!
 *  @brief STLNGCD algorithm for two univariate polynomials with several options.
 *  @details This is an implementation of the STLNGCD algorithm proposed by Kaltofen et.al.
 *   in 2006. (see ISSAC 2006, pp. 169-176).
 *   In the final stage, we use UVGCD's scheme to find the approximate GCD.
 *  @warning In the original paper, the modified GKO method is used for LS.
 *   However, we use LAPACK routine for LS since our fastgcd routine is not better than
 *   uvgcd in computational time and also there is no significant difference for not
 *   large matrices in the original paper on solving LS by the modified GKO method.
 *  @warning In the original paper, the term "the first singular vector" is used but
 *   its meaning is the singular vector corresponding to the smallest singular value
 *   according to the preliminary implementation by Lihong. Hence in our implementation,
 *   we use the singular vector corresponding to the smallest singular value for that.
 *  @warning This uses gels instead of gelsy!!
 *  @param[in] poly_inf The input polynomial f[x].
 *  @param[in] poly_ing The input polynomial g[x].
 *  @param[out] poly_d The approximate GCD found.
 *  @param[out] poly_f1 The approximate cofactor found for poly_inf.
 *  @param[out] poly_g1 The approximate cofactor found for poly_ing.
 *  @param[in] d The degree of approximate GCD to be found.
 *  @param[in] threshold The threshold for iterations.
 *  @param[in] max_iterations The maximum number of iterations.
 *  @param[in] mode_opt UVGCD's refinement if (mode_opt%10) is not 0.
 *  @return 0 if success or -1 if failure.
 */
int SNAP_agcd_stlngcd2x_ex(SNAP_t poly_inf, SNAP_t poly_ing, SNAP_t poly_d, SNAP_t poly_f1, SNAP_t poly_g1, const unsigned int d, const SNAPBASEREAL_t threshold, const unsigned int max_iterations, const int mode_opt) {
  // local variables
  SNAPMTX_t matSk, matVT, matH, matHH;
  SNAPMTXREAL_t vecS, vecZX, vecRDZ, matAk, vecB, matYPk;
  SNAPBASE_t *bpt1, *bpt2, *bpt3;
  SNAPBASEREAL_t norm_f, norm_g, w, sigma_m1;
  SNAPBASEREAL_t rho;
  SNAPBASE_t *tau, *yvect, *zvect;
  SNAP_t poly_f, poly_g;
#if defined(SNAP_SUPDC_T)
  SNAPBASE_t one, mone, zero, tempComplex;
#endif // defined
  SNAPBASEREAL_t *vec_zx, *vec_z, *vec_x, *vec_rdz, *vec_r, *vec_dz, *vec_b;
  SNAPBASEREAL_t *pt1, *pt2;
  int result_code = 0;
  unsigned int i, imax, j;
  unsigned int lenZinC, lenZinR, lenXinC, lenXinR, lenRinC, lenRinR;
  unsigned int posB;
  unsigned int loop_condition, loop_count;
#if defined(SNAP_SUPDC_T)
  unsigned int realonly_case;
#endif // defined
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randstate_t rstate;
#endif
  // initializations
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  gmp_randinit_mt(rstate);
  gmp_randseed_ui(rstate, (unsigned int)time(NULL));
#else
  srand((unsigned int)time(NULL));
#endif
  mode_type mode_refinement;
  // initialization
  if ( mode_opt % 10 == 0 ) {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_NONE;
  } else {
    mode_refinement = SNAP_STLNGCDEX_REFINEMENT_UVGCD;
  }
  SNAP_init_set(poly_f, poly_inf);
  SNAP_init_set(poly_g, poly_ing);
#if defined(SNAP_SUPD_T)
  lenZinR = lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenXinR = lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenRinR = lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
#elif defined(SNAP_SUPDC_T)
  lenZinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)+2;
  lenZinR = lenZinC*2;
  lenXinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-2*d+1;
  lenXinR = lenXinC*2;
  lenRinC = SNAP_degree(poly_inf)+SNAP_degree(poly_ing)-d+1;
  lenRinR = lenRinC*2;
  SNAPBASE_init_set_si(one, 1);
  SNAPBASE_init_set_si(mone, -1);
  SNAPBASE_init_set_si(zero, 0);
  SNAPBASE_init(tempComplex);
#else
  IMPLEMENT HERE!!
#endif
  SNAPMTX_init2(matSk, lenRinC, lenXinC+1);
  SNAPMTXREAL_init2(vecS, 1, lenXinC+1);
  SNAPMTX_init2(matVT, lenXinC+1, lenXinC+1);
  SNAPMTX_init2(matH, lenRinC, lenZinC);
  SNAPMTX_init2(matHH, lenRinC, lenRinC);
  SNAPBASEREAL_init(norm_f);
  SNAPBASEREAL_init(norm_g);
  SNAP_2norm(norm_f, poly_inf);
  SNAP_2norm(norm_g, poly_ing);
#if defined(SNAP_SUPDC_T)
  realonly_case = SNAPLIB_TRUE;
  for ( bpt1 = SNAP_coefs(poly_f), i = 0; realonly_case && i <= SNAP_degree(poly_f); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  for ( bpt1 = SNAP_coefs(poly_g), i = 0; realonly_case && i <= SNAP_degree(poly_g); bpt1++, i++ ) {
    if ( !SNAPBASEREAL_zeroq(SNAPBASE_imag(*bpt1)) ) {
      realonly_case = SNAPLIB_FALSE;
    }
  }
  if ( realonly_case ) {
    printf("realonly_case!!!\n");
    supdc_randpoly(poly_f, SNAP_degree(poly_f), threshold, 1.0);
    for ( bpt1 = SNAP_coefs(poly_f), bpt2 = SNAP_coefs(poly_inf), i = 0; i <= SNAP_degree(poly_f); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
    supdc_randpoly(poly_g, SNAP_degree(poly_g), threshold, 1.0);
    for ( bpt1 = SNAP_coefs(poly_g), bpt2 = SNAP_coefs(poly_ing), i = 0; i <= SNAP_degree(poly_g); bpt1++, bpt2++, i++ ) {
      *bpt1 = cimag(*bpt1)*_Complex_I + *bpt2;
    }
  }
#endif // defined
  SNAPMTXREAL_init2(vecZX, 1, lenZinR+lenXinR);
  vec_zx = vec_z = SNAPMTXREAL_elems(vecZX);
  vec_x = vec_zx + lenZinR;
  SNAPMTXREAL_init2(vecRDZ, 1, lenRinR+lenZinR);
  vec_rdz = vec_r = SNAPMTXREAL_elems(vecRDZ);
  vec_dz = vec_rdz + lenRinR;
  SNAPMTXREAL_init2(matAk, lenRinR, lenXinR);
  SNAPMTXREAL_init2(vecB, 1, lenRinR);
  vec_b = SNAPMTXREAL_elems(vecB);
  SNAPBASEREAL_init_set_d(w, SNAPLIB_STLN_PENALTY_VALUE);
  SNAPMTXREAL_init2(matYPk, lenRinR+lenZinR, lenZinR+lenXinR);
  /* ====================== initial vectos by Lagrangian multipliers ====================== */
  // Sk(f,g)
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);
  // Construct Sk = U.S.VT as U in Sk, S in vecS and VT in matVT.
  // matVT[-1]^h means the singular vector "v" will be used.
  // Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
#if defined(SNAP_SUPD_T)
  if ( LAPACKE_dgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#elif defined(SNAP_SUPDC_T)
  if ( LAPACKE_zgesdd( LAPACK_ROW_MAJOR, 'O', SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk),
                       SNAPMTXREAL_elems(vecS), NULL, 1, SNAPMTX_elems(matVT), SNAPMTX_csize(matVT) ) != 0 ) {
#else
  IMPLEMENT HERE!!
#endif
    printf("ERROR in agcd_stlngcd2_ex: Xgesdd could not compute the SVD.\n");
    result_code = -1;
    goto post;
  }
  // Construct H(v)
  SNAP_agcd_stlngcd2_convolution_like_matrix(matH, SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, SNAP_degree(poly_ing)-d+1), SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1);
  // Compute HH as H(v).H(v)^h
#if defined(SNAP_SUPD_T)
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), 1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), 0.0, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#elif defined(SNAP_SUPDC_T)
  cblas_zgemm(CblasRowMajor, CblasNoTrans, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_rows(matH), SNAPMTX_cols(matH), &one, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elems(matH), SNAPMTX_csize(matH), &zero, SNAPMTX_elems(matHH), SNAPMTX_csize(matHH));
#else
  IMPLEMENT HERE!!
#endif
  // Construct the initial vector X
  posB = 0;
  for ( i=1, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0), bpt2=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 1); i<=lenXinC; i++, bpt2++ ) {
    if ( SNAPBASE_cmpabs(*bpt1, *bpt2) < 0 ) {
      posB = i;
      bpt1 = bpt2;
    }
  }
  bpt2 = SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, posB);
  for ( i=0, pt1=vec_x, bpt1=SNAPMTX_elemp(matVT, SNAPMTX_rows(matVT)-1, 0); i<=lenXinC; i++, pt1++, bpt1++ ) {
    if ( i == posB ) {
      pt1--;
      continue;
    }
#if defined(SNAP_SUPD_T)
    SNAPBASEREAL_neg(*pt1, *bpt1);
    SNAPBASEREAL_div(*pt1, *pt1, *bpt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_neg(tempComplex, *bpt1);
    SNAPBASE_div(tempComplex, tempComplex, *bpt2);
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(tempComplex));
    SNAPBASEREAL_set(*(pt1+lenXinC), SNAPBASE_imag(tempComplex));
#else
    IMPLEMENT HERE!!
#endif
  }
  // Compute Sk[;-1] := Sk.v is stored as U[;-1]*S[-1] = Sk[;-1]*vecS[-1].
  pt1 = SNAPMTXREAL_elemp(vecS, 0, SNAPMTX_cols(matSk) -1);
  for ( i=0, bpt1=SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1); i<lenRinC; i++, bpt1+=SNAPMTX_csize(matSk)) {
    SNAPBASE_mul_basereal(*bpt1, *bpt1, *pt1);
  }
  // Solve solve HH.x = z and z := x; stored in U[;-1]!!
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#if defined(SNAP_SUPD_T)
  if ( clapack_dgels(CblasColMajor, CblasTrans, SNAPMTX_cols(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_rows(matSk) ) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2x_ex), ": clapack_dgels could not compute the solution.\n");
    goto post;
  }
#elif defined(SNAP_SUPDC_T)
  for ( i=0, bpt1=SNAPMTX_elems(matSk); i < SNAPMTX_csize(matSk)*SNAPMTX_rows(matSk); i++, bpt1++ ) {
    SNAPBASE_conj(*bpt1, *bpt1);
  }
  if ( clapack_zgels(CblasColMajor, CblasConjTrans, SNAPMTX_cols(matSk), SNAPMTX_rows(matSk), 1, SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_rows(matSk) ) != 0 ) {
    printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2x_ex), ": clapack_zgels could not compute the solution.\n");
    goto post;
  }
#endif // SNAP_SUPD_T
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  // Compute vec_z := -H^T.z; as -matH^t * U[;-1]!!
#if defined(SNAP_SUPD_T)
  cblas_dgemv(CblasRowMajor, CblasTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), -1.0, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), 0.0, vec_z, 1);
#elif defined(SNAP_SUPDC_T)
  cblas_zgemv(CblasRowMajor, CblasConjTrans, SNAPMTX_rows(matH), SNAPMTX_cols(matH), &mone, SNAPMTX_elems(matH), SNAPMTX_csize(matH), SNAPMTX_elemp(matSk, 0, SNAPMTX_cols(matSk)-1), SNAPMTX_csize(matSk), &zero, SNAPMTX_elems(matVT), 1);
  for ( i=0, bpt1=SNAPMTX_elems(matVT), pt1=vec_z; i<lenZinC; i++, bpt1++, pt1++ ) {
    SNAPBASEREAL_set(*pt1, SNAPBASE_real(*bpt1));
    SNAPBASEREAL_set(*(pt1+lenZinC), SNAPBASE_imag(*bpt1));
  }
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  /* ====================== preparation for the main loop ====================== */
  // NEED (done): matAk, vec_z, vec_x, vec_b, vec_r
  SNAP_agcd_stlngcd2_init_Ak(matAk, vec_b, poly_inf, poly_ing, d-1, posB);
  SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_z, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
  // vec_r := bk - Ak x = vec_b - matAk.vec_x
  for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
    SNAPBASEREAL_set(*pt1, *pt2);
  }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
  IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
  /* ====================== main loop!! ====================== */
  loop_count = 0;
  loop_condition = SNAPLIB_TRUE;
  while ( loop_condition && loop_count++ < SNAPLIB_STLN_MAXIMUM_LOOP ) {
    // NEED (outside): matAk, vec_z, vec_x, vec_b, vec_r
    // NEED (inside): matYPk, vec_rdz
    // reset YPk
    imax = SNAPMTXREAL_csize(matYPk)*SNAPMTXREAL_rsize(matYPk);
    for ( i=0, pt1=SNAPMTXREAL_elems(matYPk); i < imax; i++,pt1++ ) {
      SNAPBASEREAL_set_d(*pt1, 0.0);
    }
    // the left-bottom block is I_{m+n+2}
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, lenRinR, 0); i < lenZinR; i++, pt1+=SNAPMTXREAL_csize(matYPk)+1 ) {
      SNAPBASEREAL_set_d(*pt1, 1.0);
    }
    // the upper-right block is w(Ak+Ek)=w*matAk
    for ( i=0, pt1=SNAPMTXREAL_elemp(matYPk, 0, lenZinR), pt2=SNAPMTXREAL_elems(matAk); i < lenRinR; i++, pt1+=lenZinR ) {
      for ( j=0; j < lenXinR; j++, pt1++, pt2++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt2, w);
      }
    }
    // the upper-left block is w(Yk-Pk)
    SNAP_agcd_stlngcd2_convolution_like_matrix_YPk(matYPk, vec_x, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    for ( i=0; i < lenRinR; i++ ) {
      for ( j=0, pt1=SNAPMTXREAL_elemp(matYPk, i, 0); j < lenZinR; j++, pt1++ ) {
        SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
      }
    }
    // form "-" [-wr, z]
    for ( i=0, pt1=vec_r; i < lenRinR; i++,pt1++ ) {
      SNAPBASEREAL_mul_basereal(*pt1, *pt1, w);
    }
    for ( i=0, pt1=vec_dz, pt2=vec_z; i < lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_mul_si(*pt1, *pt2, -1);
    }
    // min|| YPk.(dz dx) + (-w*r z) ||_2 by QR as LS
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    // solve min|| .... ||_2 and vec_rdz := [del_z, del_x]!!!
    if ( clapack_dgels(CblasColMajor, CblasTrans, SNAPMTX_cols(matYPk), SNAPMTX_rows(matYPk), 1, SNAPMTX_elems(matYPk), SNAPMTX_csize(matYPk), vec_rdz, SNAPMTX_rows(matYPk) ) != 0 ) {
      printf("%s%s%s", "ERROR in ", SNAP2str(SNAP_agcd_stlngcd2_ex), ": clapack_dgels could not compute the solution.\n");
      goto post;
    }
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3-b: x = x + del_x, z = z + del_z
    for ( i=0, pt1=vec_zx, pt2=vec_rdz; i < lenXinR + lenZinR; i++,pt1++,pt2++ ) {
      SNAPBASEREAL_add(*pt1, *pt1, *pt2);
    }
    // STEP3-c: E_k, h_k from z, Y_k from x, A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    // actual: loop condition check and A_k = A_k + E_k, b_k = b_k + h_k, r = b_k - A_k x
    SNAPMTXREAL_BLAS_nrm2(norm_g, lenZinR, vec_rdz, 1);
    SNAPMTXREAL_BLAS_nrm2(norm_f, lenXinR, vec_rdz + lenZinR, 1);
    if ( SNAPBASEREAL_cmp(norm_f, threshold) <= 0 && SNAPBASEREAL_cmp(norm_g, threshold) <= 0 ) {
      loop_condition = SNAPLIB_FALSE;
    }
    // A_k = A_k + E_k
    // b_k = b_k + h_k
    SNAP_agcd_stlngcd2_update_Ak(matAk, vec_b, vec_rdz, SNAP_degree(poly_inf), SNAP_degree(poly_ing), d-1, posB);
    // vec_r := bk - Ak x = vec_b - matAk.vec_x
    for ( i=0, pt1=vec_r, pt2=vec_b; i < lenRinR; i++, pt1++,pt2++ ) {
      SNAPBASEREAL_set(*pt1, *pt2);
    }
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
    cblas_dgemv(CblasRowMajor, CblasNoTrans, SNAPMTXREAL_rows(matAk), SNAPMTXREAL_cols(matAk), -1.0, SNAPMTXREAL_elems(matAk), SNAPMTXREAL_csize(matAk), vec_x, 1, 1.0, vec_r, 1);
#else
    IMPLEMENT HERE!!
#endif // SNAP_SUPDC_T
    // STEP3: until ||del_x||_2 <= tol and ||del_z||_2 <= tol
  }
  if ( loop_count > max_iterations ) {
    printf("INFO in agcd_stlngcd2_ex: didn't converge within %d iterations.\n", max_iterations);
  } else {
    printf("INFO in agcd_stlngcd2_ex: converged within %d iterations.\n", loop_count);
  }
  // STEP4: output polys formed by b_k and A_k
  for ( i=0, bpt1=SNAP_coefp(poly_f, SNAP_degree(poly_inf)), bpt2=SNAP_coefp(poly_inf, SNAP_degree(poly_inf)), pt2=vec_z; i <= SNAP_degree(poly_inf); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }
  for ( i=0, bpt1=SNAP_coefp(poly_g, SNAP_degree(poly_ing)), bpt2=SNAP_coefp(poly_ing, SNAP_degree(poly_ing)); i <= SNAP_degree(poly_ing); i++,bpt1--,bpt2--,pt2++ ) {
#if defined(SNAP_SUPD_T)
    SNAPBASE_add(*bpt1, *bpt2, *pt2);
#elif defined(SNAP_SUPDC_T)
    SNAPBASE_add_basereal_basereal(*bpt1, *bpt2, *pt2, *(pt2+lenZinC));
#else
    IMPLEMENT HERE!!
#endif // defined
  }
  // WE COMPUTE THE APPROXIMATE GCD by UVGCD SCHEME.
  for ( i=0, bpt1=SNAPMTX_elems(matSk); i < SNAPMTX_csize(matSk)*SNAPMTX_rsize(matSk); i++,bpt1++ ) {
    SNAPBASE_set_d(*bpt1, 0.0);
  }
  SNAPMTX_subresultant_mapping_matrix_dsc(matSk, poly_f, poly_g, d-1);
  SNAPBASEREAL_init(sigma_m1);
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_init(rho);
  }
  tau = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  zvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  yvect = (SNAPBASE_t*) malloc(sizeof(SNAPBASE_t)*SNAPMTX_csize(matSk));
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_init(*bpt1);
    SNAPBASE_init(*bpt2);
    SNAPBASE_init(*bpt3);
  }
  // QR decomposition
#if defined(SNAP_SUPD_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_dgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_dgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: dgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPDC_T)
  if (
#ifdef SNAPLIB_USE_PATCHED_CLAPACK
    clapack_zgeqrf(CblasRowMajor, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#else
    LAPACKE_zgeqrf(LAPACK_ROW_MAJOR, SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0
#endif
  ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: zgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T)
  if ( snap_mlapack_frgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: mlapack_frgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFC_T)
  if ( snap_mlapack_fcgeqrf(SNAPMTX_rows(matSk), SNAPMTX_cols(matSk), SNAPMTX_elems(matSk), SNAPMTX_csize(matSk), tau) != 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: mlapack_fcgeqrf could not compute the QR decomposition.\n");
    goto post_uvgcd;
  }
#endif // defined
  // After this, zvect is the singular vector
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, &sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  if ( SNAP_agcd_uvgcd_iteration_7_3(matSk, zvect, yvect, sigma_m1, rstate) < 0 ) {
    result_code = -1;
    printf("ERROR in agcd_stlngcd2_ex: agcd_uvgcd_iteration_7_3 could not work correctly.\n");
    goto post_uvgcd;
  }
#endif
  // from 0 to deg(g)-init_r-1, deg(f)-init_r-1
  SNAP_set_rbase(poly_g1, zvect, SNAP_degree(poly_ing)-d+1 );
  SNAP_set_rbase(poly_f1, zvect+SNAP_degree(poly_ing)-d+1, SNAP_degree(poly_inf)-d+1 );
  SNAP_self_neg(poly_f1);
  if ( SNAP_agcd_uvgcd_solving_10_2(poly_f, poly_g, poly_d, poly_f1, poly_g1, d-1) == 0 ) {
#ifdef SNAPLIB_DEBUG
    SNAP_t dum;
    double a;
    SNAP_init(dum);
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_f, dum);
    printf("w/o newton, Residual of F[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_f1, poly_d);
    SNAP_sub(dum, poly_inf, dum);
    printf("w/o newton, Residual of inF[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_g, dum);
    printf("w/o newton, Residual of G[X]: %le\n", SNAP_2norm(a,dum));
    SNAP_mul(dum, poly_g1, poly_d);
    SNAP_sub(dum, poly_ing, dum);
    printf("w/o newton, Residual of inG[X]: %le\n", SNAP_2norm(a,dum));

    SNAP_clear(dum);
#endif // SNAPLIB_DEBUG
    // getting a better approximate GCD.
    if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
      if (
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, &rho) < 0
#elif defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
        SNAP_agcd_uvgcd_iteration_9_2(poly_inf, poly_ing, poly_d, poly_f1, poly_g1, rho) < 0
#endif
      ) {
        result_code = -1;
      }
    }
  } else {
    result_code = -1;
  }
post_uvgcd:
  if ( mode_refinement == SNAP_STLNGCDEX_REFINEMENT_UVGCD ) {
    SNAPBASEREAL_clear(rho);
  }
  SNAPBASEREAL_clear(sigma_m1);
#if defined(SNAP_SUPFR_T) || defined(SNAP_SUPFC_T)
  for ( i=0, bpt1=tau, bpt2=zvect, bpt3=yvect; i < SNAPMTX_csize(matSk); i++,bpt1++,bpt2++,bpt3++ ) {
    SNAPBASE_clear(*bpt1);
    SNAPBASE_clear(*bpt2);
    SNAPBASE_clear(*bpt3);
  }
  gmp_randclear(rstate);
#endif
  free(zvect);
  free(yvect);
  free(tau);
post:
  SNAPMTX_clear(matSk);
  SNAPMTX_clear(matVT);
  SNAPMTX_clear(matH);
  SNAPMTX_clear(matHH);
  SNAPMTXREAL_clear(vecS);
  SNAPMTXREAL_clear(vecZX);
  SNAPMTXREAL_clear(vecRDZ);
  SNAPMTXREAL_clear(matAk);
  SNAPMTXREAL_clear(vecB);
  SNAPMTXREAL_clear(matYPk);
  SNAPBASEREAL_clear(norm_f);
  SNAPBASEREAL_clear(norm_g);
  SNAPBASEREAL_clear(w);
#if defined(SNAP_SUPDC_T)
  SNAPBASE_clear(one);
  SNAPBASE_clear(mone);
  SNAPBASE_clear(zero);
  SNAPBASE_clear(tempComplex);
#endif // defined
  SNAP_clear(poly_g);
  SNAP_clear(poly_f);
  if ( result_code < 0 ) {
    SNAP_degree(poly_d) = 0;
    SNAP_set_coef_d(poly_d, 0, 1.0);
    SNAP_set(poly_f1, poly_inf);
    SNAP_set(poly_g1, poly_ing);
  }
  return result_code;
}

















