/*
 * Copyright (c) 2015, Kosaku Nagasaka <nagasaka@main.h.kobe-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <snap.h>
#include <float.h>
#if defined(SNAP_SUPD_T) || defined(SNAP_SUPDC_T)
#include <fftw3.h>
#include <cblas.h>
#include <clapack.h>
#include <lapacke.h>
#endif

#define SNAPLIB_MIN(a,b) (((a)<(b)) ? (a):(b))
#define SNAPLIB_MAX(a,b) (((a)>(b)) ? (a):(b))
#define SNAP_QR_THRESHOLD 1.0e-14

/*!
 *  @brief This macro applies the conjugate transpose of Givens rotation with (c,s) to the vector (a,b)^t (i.e. G^h v), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_H(c,s,a,b,t1,t2) {SNAPBASECOMPLEX_conj((t1),(c));SNAPBASECOMPLEX_mul((t1),(t1),(a));SNAPBASECOMPLEX_mul((t2),(s),(b));SNAPBASECOMPLEX_sub((t1),(t1),(t2));SNAPBASECOMPLEX_mul((b),(b),(c));SNAPBASECOMPLEX_conj((t2),(s));SNAPBASECOMPLEX_mul((a),(a),(t2));SNAPBASECOMPLEX_add((b),(b),(a));SNAPBASECOMPLEX_set((a),(t1));}
/*!
 *  @brief This macro applies the Givens rotation with (c,s) to the vector (a,b)^t (i.e. G v), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS(c,s,a,b,t1,t2) {SNAPBASECOMPLEX_mul((t1),(c),(a));SNAPBASECOMPLEX_mul((t2),(s),(b));SNAPBASECOMPLEX_add((t1),(t1),(t2));SNAPBASECOMPLEX_conj((t2),(c));SNAPBASECOMPLEX_mul((b),(b),(t2));SNAPBASECOMPLEX_conj((t2),(s));SNAPBASECOMPLEX_mul((a),(a),(t2));SNAPBASECOMPLEX_sub((b),(b),(a));SNAPBASECOMPLEX_set((a),(t1));}
/*!
 *  @brief This macro applies the conjugate transpose of Givens rotation with (c,s) to the vector (a,b) (i.e. v G^h), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_HR(c,s,a,b,t1,t2) {SNAPBASECOMPLEX_conj((t1),(c));SNAPBASECOMPLEX_mul((t1),(t1),(a));SNAPBASECOMPLEX_conj((t2),(s));SNAPBASECOMPLEX_mul((t2),(t2),(b));SNAPBASECOMPLEX_add((t1),(t1),(t2));SNAPBASECOMPLEX_mul((b),(b),(c));SNAPBASECOMPLEX_mul((a),(a),(s));SNAPBASECOMPLEX_sub((b),(b),(a));SNAPBASECOMPLEX_set((a),(t1));}
/*!
 *  @brief This macro applies the Givens rotation with (c,s) to the vector (a,b) (i.e. v G), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_R(c,s,a,b,t1,t2) {SNAPBASECOMPLEX_mul((t1),(c),(a));SNAPBASECOMPLEX_conj((t2),(s));SNAPBASECOMPLEX_mul((t2),(t2),(b));SNAPBASECOMPLEX_sub((t1),(t1),(t2));SNAPBASECOMPLEX_conj((t2),(c));SNAPBASECOMPLEX_mul((b),(b),(t2));SNAPBASECOMPLEX_mul((a),(a),(s));SNAPBASECOMPLEX_add((b),(b),(a));SNAPBASECOMPLEX_set((a),(t1));}
/*!
 *  @brief This macro applies the conjugate transpose of Givens rotation with (c2,s2) to the Givens rotation with (c1,s1) (i.e. G1G2^h), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_GIVENS_HR(c1,s1,c2,s2,t1,t2) {SNAPBASECOMPLEX_conj((t1),(c2));SNAPBASECOMPLEX_mul((t1),(t1),(c1));SNAPBASECOMPLEX_conj((t2),(s2));SNAPBASECOMPLEX_mul((t2),(t2),(s1));SNAPBASECOMPLEX_add((t1),(t1),(t2));SNAPBASECOMPLEX_mul((c1),(c1),(s2));SNAPBASECOMPLEX_mul((s1),(s1),(c2));SNAPBASECOMPLEX_sub((s1),(s1),(c1));SNAPBASECOMPLEX_set((c1),(t1));}
/*!
 *  @brief This macro applies the Givens rotation with (c2,s2) to the Givens rotation with (c1,s1) (i.e. G2G1), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_GIVENS(c1,s1,c2,s2,t1,t2) {SNAPBASECOMPLEX_mul((t1),(c1),(c2));SNAPBASECOMPLEX_conj((t2),(s1));SNAPBASECOMPLEX_mul((t2),(t2),(s2));SNAPBASECOMPLEX_sub((t1),(t1),(t2));SNAPBASECOMPLEX_conj((t2),(c1));SNAPBASECOMPLEX_mul((c1),(t2),(s2));SNAPBASECOMPLEX_mul((s1),(s1),(c2));SNAPBASECOMPLEX_add((s1),(s1),(c1));SNAPBASECOMPLEX_set((c1),(t1));}
/*!
 *  @brief This macro applies the Givens rotation with (c2,s2) to the Givens rotation with (c1,s1) (i.e. G1G2), with the temporal variables t1 and t2.
 */
#define APPLY_GIVENS_GIVENS_R(c1,s1,c2,s2,t1,t2) {SNAPBASECOMPLEX_mul((t1),(c1),(c2));SNAPBASECOMPLEX_conj((t2),(s2));SNAPBASECOMPLEX_mul((t2),(t2),(s1));SNAPBASECOMPLEX_sub((t1),(t1),(t2));SNAPBASECOMPLEX_conj((t2),(c2));SNAPBASECOMPLEX_mul((s1),(t2),(s1));SNAPBASECOMPLEX_mul((c1),(s2),(c1));SNAPBASECOMPLEX_add((s1),(s1),(c1));SNAPBASECOMPLEX_set((c1),(t1));}
/*!
 *  @brief This computes ij-th element of W=VH.
 */
#define MAT_W_ELEM_SET(w, i, j, genV, genH, r, pt1, pt2) {pt1=SNAPMTXCOMPLEX_elemp(genV,i,0);pt2=SNAPMTXCOMPLEX_elemp(genH,0,j);SNAPBASECOMPLEX_mul(w,*pt1,*pt2);for(spi4w=1,pt1++,pt2+=SNAPMTXCOMPLEX_csize(genH);spi4w<r;spi4w++,pt1++,pt2+=SNAPMTXCOMPLEX_csize(genH)){SNAPBASECOMPLEX_addmul(w,*pt1,*pt2);}}

// The following one definition is just for Code::Blocks. Erase this before release.
#ifndef SNAP_SUPDC_T
#define SNAP_SUPDC_T
#endif // SNAP_SUPDC_T

#define SNAP_QR_DEBUG
#define SNAP_QR_GENERATE_SYL
// The following macro enables the Givens rotation with consideration of magnitudes of ca and cb.
#define SNAP_QR_USE_RELATIVE_GIVENS
// The following macro forces the use of the Given Rotations which makes the resulting value in reals in pull-through related operations.
#define SNAP_QR_USE_GIVENS_R_ONLY
// The following macro enables the use of the first column as the second Givens rotation in pull-through related operations.
#define SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
// The following macro enables the computation of the last Givens rotation in pull-through related operations, by the Given rotation.
// #define SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
// The following macro enables the use of the last row as the last Givens rotation in pull-through related operations.
#define SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH

#if defined(SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH) && defined(SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH)
#undef SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
#endif // defined

#ifdef SNAP_QR_DEBUG
#define PRINT_MAT_SYL {printf("matSYL:\n");for(q1=0;q1<sdmdc_rows(matSYL);q1++){for(q2=0;q2<sdmdc_cols(matSYL);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(matSYL,q1,q2)),cimag(sdmdc_elem(matSYL,q1,q2)));}printf("\n");}}
#define PRINT_MAT_SYL_trinorm(a,k) {printf("tri(matSYL)_F: ");(a)=0.0;for(q1=sdmdc_rows(matSYL)-k-1;q1<sdmdc_rows(matSYL);q1++){for(q2=0;q2+sdmdc_rows(matSYL)-k<q1&&q2<q1&&q2<sdmdc_cols(matSYL);q2++){(a)+=cabs(sdmdc_elem(matSYL,q1,q2))*cabs(sdmdc_elem(matSYL,q1,q2));}}printf("%e\n",sqrt(creal((a))));}
#define PRINT_MAT_SYL_REAL {printf("matSYL:\n");for(q1=0;q1<sdmdc_rows(matSYL);q1++){for(q2=0;q2<sdmdc_cols(matSYL);q2++){printf("%+0.1e,",creal(sdmdc_elem(matSYL,q1,q2)));}printf("\n");}}
#define PRINT_MAT_R {printf("matR:\n");for(q1=0;q1<sdmdc_rows(matR);q1++){for(q2=0;q2<sdmdc_cols(matR);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(matR,q1,q2)),cimag(sdmdc_elem(matR,q1,q2)));}printf("\n");}}
#define PRINT_MAT_R_trinorm(a) {printf("tri(matR)_F: ");(a)=0.0;for(q1=0;q1<sdmdc_rows(matR);q1++){for(q2=0;q2<q1&&q2<sdmdc_cols(matR);q2++){(a)+=cabs(sdmdc_elem(matR,q1,q2))*cabs(sdmdc_elem(matR,q1,q2));}}printf("%e\n",sqrt(creal((a))));}
#define PRINT_MAT_R_row(sr,er,c) {printf("matR:\n");for(q1=(sr);q1<=(er);q1++){printf("(%+e%+eI),",creal(sdmdc_elem(matR,q1,(c))),cimag(sdmdc_elem(matR,q1,(c))));}printf("\n");}
#define PRINT_MAT_Q {printf("matQ:\n");for(q1=0;q1<sdmdc_rows(matQ);q1++){for(q2=0;q2<sdmdc_cols(matQ);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(matQ,q1,q2)),cimag(sdmdc_elem(matQ,q1,q2)));}printf("\n");}}
// #define PRINT_MAT_W {printf("matW:\n");for(q1=0;q1<sdmdc_rows(matW);q1++){for(q2=0;q2<sdmdc_cols(matW);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(matW,q1,q2)),cimag(sdmdc_elem(matW,q1,q2)));}printf("\n");}}
#define PRINT_MAT_W(w,pt1,pt2) {printf("matW:\n");for(q1=0;q1<sdmdc_rows(genV);q1++){for(q2=0;q2<sdmdc_cols(genH);q2++){MAT_W_ELEM_SET(w,q1,q2,genV,genH,r,pt1,pt2)printf("(%+e%+eI),",creal(w),cimag(w));}printf("\n");}}
#define PRINT_LEFT_S {printf("leftS:\n");for(q1=0;q1<sdmdc_rows(leftS);q1++){for(q2=0;q2<sdmdc_cols(leftS);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(leftS,q1,q2)),cimag(sdmdc_elem(leftS,q1,q2)));}printf("\n");}}
#define PRINT_RIGHT_S {printf("rightS:\n");for(q1=0;q1<sdmdc_rows(rightS);q1++){for(q2=0;q2<sdmdc_cols(rightS);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(rightS,q1,q2)),cimag(sdmdc_elem(rightS,q1,q2)));}printf("\n");}}
#define PRINT_SCALAR_S {printf("scalarS:\n");for(q1=0;q1<sdmdc_rows(scalarS);q1++){for(q2=0;q2<sdmdc_cols(scalarS);q2++){printf("(%+e%+eI),",creal(sdmdc_elem(scalarS,q1,q2)),cimag(sdmdc_elem(scalarS,q1,q2)));}printf("\n");}}
#define PRINT_LEFT_S18 {printf("leftS:\n");for(q1=0;q1<sdmdc_rows(leftS);q1++){for(q2=0;q2<sdmdc_cols(leftS);q2++){printf("(%+0.18e%+0.18eI),",creal(sdmdc_elem(leftS,q1,q2)),cimag(sdmdc_elem(leftS,q1,q2)));}printf("\n");}}
#define PRINT_RIGHT_S18 {printf("rightS:\n");for(q1=0;q1<sdmdc_rows(rightS);q1++){for(q2=0;q2<sdmdc_cols(rightS);q2++){printf("(%+0.18e%+0.18eI),",creal(sdmdc_elem(rightS,q1,q2)),cimag(sdmdc_elem(rightS,q1,q2)));}printf("\n");}}
#define PRINT_SCALAR_S18 {printf("scalarS:\n");for(q1=0;q1<sdmdc_rows(scalarS);q1++){for(q2=0;q2<sdmdc_cols(scalarS);q2++){printf("(%+0.18e%+0.18eI),",creal(sdmdc_elem(scalarS,q1,q2)),cimag(sdmdc_elem(scalarS,q1,q2)));}printf("\n");}}
void print_rebuildS(unsigned int n, unsigned int r, sdmdc_t matQ, sdmdc_t lisD) {
  // Q^hDQ
  unsigned int i,j,k,q1,q2;
  sdmdc_t matQh, matY;
  sdmdc_init2(matQh, n, n);
  sdmdc_init2(matY, n, n);
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matQh, j, i) = conj(sdmdc_elem(matQ, i, j));
    }
  }
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matQh, i, j ) *= sdmdc_elem(lisD,0,j);
    }
  }
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matY, i, j) = 0;
      for ( k = 0; k < n; k++ ) {
        sdmdc_elem(matY, i, j) +=sdmdc_elem(matQh, i, k )*sdmdc_elem(matQ, k, j );
      }
    }
  }
  printf("Q^hDQ from Q and D:\n");
  for(q1=0; q1<sdmdc_rows(matY); q1++) {
    for(q2=0; q2<sdmdc_cols(matY); q2++) {
      printf("(%+e%+eI),",creal(sdmdc_elem(matY,q1,q2)),cimag(sdmdc_elem(matY,q1,q2)));
    }
    printf("\n");
  }
  sdmdc_clear(matY);
  sdmdc_clear(matQh);
}
void SNAP_alg_qr_dgv2010refine_apply_S(const unsigned int s, SNAPMTXCOMPLEX_t leftS, const unsigned int rL, SNAPMTXCOMPLEX_t scalarS, SNAPMTXCOMPLEX_t rightS, const unsigned int rR, SNAPMTXCOMPLEX_t matR, const unsigned int k);
void print_computedS(unsigned int n, unsigned int r, sdmdc_t scalarS, sdmdc_t leftS, sdmdc_t rightS) {
  unsigned int i,j,q1,q2;
  sdmdc_t matX, matY;
  sdmdc_init2(matX, n, n);
  sdmdc_init2(matY, n, n);
  for ( i = 0; i < n; i++ ) {
    sdmdc_elem(matX, i, i) = 1;
  }
  for ( i = 0; i < n; i++ ) {
    SNAP_alg_qr_dgv2010refine_apply_S(0, leftS, r, scalarS, rightS, r-1, matX, i);
  }
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matY, i, j) = conj(sdmdc_elem(matX, j, i));
    }
  }
  printf("Q^hDQ from scalarS, leftS and rightS:\n");
  for(q1=0; q1<sdmdc_rows(matY); q1++) {
    for(q2=0; q2<sdmdc_cols(matY); q2++) {
      printf("(%+e%+eI),",creal(sdmdc_elem(matY,q1,q2)),cimag(sdmdc_elem(matY,q1,q2)));
    }
    printf("\n");
  }
  sdmdc_clear(matY);
  sdmdc_clear(matX);
}
void print_computedS_embed(unsigned int n, unsigned int r, sdmdc_t scalarS, sdmdc_t leftS, sdmdc_t rightS) {
  unsigned int i,j,q1,q2;
  sdmdc_t matX, matY;
  sdmdc_init2(matX, n, n);
  sdmdc_init2(matY, n, n);
  for ( i = 0; i < n; i++ ) {
    sdmdc_elem(matX, i, i) = 1;
  }
  for ( i = 0; i < n; i++ ) {
    SNAP_alg_qr_dgv2010refine_apply_S(0, leftS, r, scalarS, rightS, r, matX, i);
  }
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matY, i, j) = conj(sdmdc_elem(matX, j, i));
    }
  }
  printf("Q^hDQ from scalarS, leftS and rightS:\n");
  for(q1=0; q1<sdmdc_rows(matY); q1++) {
    for(q2=0; q2<sdmdc_cols(matY); q2++) {
      printf("(%+e%+eI),",creal(sdmdc_elem(matY,q1,q2)),cimag(sdmdc_elem(matY,q1,q2)));
    }
    printf("\n");
  }
  sdmdc_clear(matY);
  sdmdc_clear(matX);
}
void print_computedS_wG(unsigned int n, unsigned int r, sdmdc_t scalarS, sdmdc_t leftS, sdmdc_t rightS, double complex *g, unsigned int s) {
  unsigned int i,j,q1,q2;
  double complex t1,t2;
  sdmdc_t matX, matY;
  sdmdc_init2(matX, n, n);
  sdmdc_init2(matY, n, n);
  for ( i = 0; i < n; i++ ) {
    sdmdc_elem(matX, i, i) = 1;
  }
  for ( i = 0; i < n; i++ ) {
    APPLY_GIVENS_H(*g, *(g+1), sdmdc_elem(matX,s,i), sdmdc_elem(matX,s+1,i),t1,t2)
  }
  for ( i = 0; i < n; i++ ) {
    SNAP_alg_qr_dgv2010refine_apply_S(0, leftS, r, scalarS, rightS, r-1, matX, i);
  }
  for ( i = 0; i < n; i++ ) {
    for ( j = 0; j < n; j++ ) {
      sdmdc_elem(matY, i, j) = conj(sdmdc_elem(matX, j, i));
    }
  }
  printf("Q^hDQ from scalarS, leftS and rightS:\n");
  for(q1=0; q1<sdmdc_rows(matY); q1++) {
    for(q2=0; q2<sdmdc_cols(matY); q2++) {
      printf("(%+e%+eI),",creal(sdmdc_elem(matY,q1,q2)),cimag(sdmdc_elem(matY,q1,q2)));
    }
    printf("\n");
  }
  sdmdc_clear(matY);
  sdmdc_clear(matX);
}
#endif // SNAP_QR_DEBUG

/*!
 *  @brief This computes the generators of sylvester matrix of f and g with the displacement operator Z_1 S - S Z_0 where Z is an upper-right shift operator.
 *  @param[in] poly_inf The input polynomial f[x] of degree n.
 *  @param[in] poly_ing The input polynomial g[x] of degree m.
 *  @param[in,out] genV V must have dim(n+m,2) and zero entries. The generator V of Sylvester matrix S s.t. Z_1 S - S Z_0 = VH on exit.
 *  @param[in,out] genH H must have dim(2,n+m) and zero entries. The generator H of Sylvester matrix S s.t. Z_1 S - S Z_0 = VH on exit.
 */
void SNAP_alg_qr_dgv2010refine_make_sylvester_generators(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXCOMPLEX_t genV, SNAPMTXCOMPLEX_t genH) {
  unsigned int i, n, m;
  SNAPBASECOMPLEX_t *pt1, *pt2;

  n = SNAP_degree(poly_inf);
  m = SNAP_degree(poly_ing);
  /* |--- for genV, we just assign two 1. */
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genV,m-1,0), 1);
  SNAPBASECOMPLEX_set_ui(SNAPMTXCOMPLEX_elem(genV,n+m-1,1), 1);
  /* |--- for genH[0,0..n+m-1] */
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(genH), pt2 = SNAP_coefp(poly_ing, m); i <= m; i++, pt1++, pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  for ( i = m, pt1--, pt2 = SNAP_coefp(poly_inf, n); i < n+m; i++, pt1++, pt2-- ) {
    SNAPBASECOMPLEX_sub_base(*pt1, *pt1, *pt2);
  }
  /* |--- for genH[1,0..n+m-1] */
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(genH, 1, 0), pt2 = SNAP_coefp(poly_inf, n); i <= n; i++, pt1++, pt2-- ) {
    SNAPBASECOMPLEX_set_base(*pt1, *pt2);
  }
  for ( i = n, pt1--, pt2 = SNAP_coefp(poly_ing, m); i < n+m; i++, pt1++, pt2-- ) {
    SNAPBASECOMPLEX_sub_base(*pt1, *pt1, *pt2);
  }
}

/*!
 *  @brief This applies (scalarS leftS rightS)^h to the given vector whose target indices from sr to er.
 *  @param[in] s s is the first row index to be applied.
 *  @param[in] leftS leftS must be constructed as S_left in SNAP_alg_qr_dgv2010refine.
 *  @param[in] rL rL is the length of each unitary representation of leftS.
 *  @param[in] scalarS scalarS must be constructed as S_left in SNAP_alg_qr_dgv2010refine.
 *  @param[in] rightS rightS must be constructed as S_left in SNAP_alg_qr_dgv2010refine.
 *  @param[in] rR rR is the length of each unitary representation of rightS.
 *  @param[in,out] matR matR is the matrix to which this applies (scalarS leftS rightS)^h.
 *  @param[in] k k is the column index of matR to be applied.
 */
void SNAP_alg_qr_dgv2010refine_apply_S(const unsigned int s, SNAPMTXCOMPLEX_t leftS, const unsigned int rL, SNAPMTXCOMPLEX_t scalarS, SNAPMTXCOMPLEX_t rightS, const unsigned int rR, SNAPMTXCOMPLEX_t matR, const unsigned int k) {
  unsigned int i, j, n, er, en;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3, ctmp1, ctmp2;
  SNAPBASECOMPLEX_init(ctmp1);
  SNAPBASECOMPLEX_init(ctmp2);

  n = SNAPMTXCOMPLEX_rows(matR);
#ifdef SNAP_QR_DEBUG
//  unsigned int q1;
//  printf("=========== apply_S =========\n");
//  PRINT_MAT_R_row(s,n-1,k)
#endif // SNAP_QR_DEBUG
  // apply scalarS
  for ( i = s, pt1 = SNAPMTXCOMPLEX_elemp(scalarS, 0, s), pt2 = SNAPMTXCOMPLEX_elemp(matR, s, k); i < n; i++, pt1++, pt2+=SNAPMTXCOMPLEX_csize(matR) ) {
    SNAPBASECOMPLEX_conj(ctmp1, *pt1);
    SNAPBASECOMPLEX_mul(*pt2, *pt2, ctmp1);
  }
#ifdef SNAP_QR_DEBUG
//  PRINT_MAT_R_row(s,n-1,k)
#endif // SNAP_QR_DEBUG
  // apply leftS in (n-1)x(2*rL) // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
  // do (rL-1) times at first, then do the rest things.
  er = SNAPLIB_MIN(n-s, rL);
  en = n - s - 1;
  for ( i = 0; i < er; i++ ) {
    pt1 = SNAPMTXCOMPLEX_elemp(leftS, n-i-2, 0);
    pt2 = SNAPMTXCOMPLEX_elemp(matR, n-i-2,k);
    pt3 = SNAPMTXCOMPLEX_elemp(matR, n-i-1,k);
    for ( j = 0; j <= i; j++, pt1+=2, pt2+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
      APPLY_GIVENS_H(*pt1,*(pt1+1),*pt2,*pt3,ctmp1,ctmp2)
    }
  }
  for ( ; i < en; i++ ) {
    pt1 = SNAPMTXCOMPLEX_elemp(leftS, n-i-2, 0);
    pt2 = SNAPMTXCOMPLEX_elemp(matR, n-i-2,k);
    pt3 = SNAPMTXCOMPLEX_elemp(matR, n-i-1,k);
    for ( j = 0; j < rL; j++, pt1+=2, pt2+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
      APPLY_GIVENS_H(*pt1,*(pt1+1),*pt2,*pt3,ctmp1,ctmp2)
    }
  }
#ifdef SNAP_QR_DEBUG
//  PRINT_MAT_R_row(s,n-1,k)
#endif // SNAP_QR_DEBUG
  // apply rightS in (n-rL-1)x(2*rR) // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
  en = SNAPLIB_MIN(en, n-rL-1);
  er = SNAPLIB_MIN(n-s, rR);
  i = en - 1;
  if ( i >= er ) {
    do {
      pt1 = SNAPMTXCOMPLEX_elemp(rightS, n-rL-2-i, 2*(rR-1));
      pt2 = SNAPMTXCOMPLEX_elemp(matR, n-2-i+rR-1,k); // n-i-2+rL-1
      pt3 = SNAPMTXCOMPLEX_elemp(matR, n-1-i+rR-1,k); // n-i-1+rL-1
      for ( j = 0; j < rR; j++, pt1-=2, pt2-=SNAPMTXCOMPLEX_csize(matR), pt3-=SNAPMTXCOMPLEX_csize(matR) ) {
        APPLY_GIVENS_H(*pt1,*(pt1+1),*pt2,*pt3,ctmp1,ctmp2)
      }
    } while ( i-- > er );
  }
  if ( er != 0 ) {
    do {
      pt1 = SNAPMTXCOMPLEX_elemp(rightS, n-rL-2-i, 2*i);
      pt2 = SNAPMTXCOMPLEX_elemp(matR, n-2,k); // n-i-2+i
      pt3 = SNAPMTXCOMPLEX_elemp(matR, n-1,k); // n-i-1+i
      for ( j = 0; j <= i; j++, pt1-=2, pt2-=SNAPMTXCOMPLEX_csize(matR), pt3-=SNAPMTXCOMPLEX_csize(matR) ) {
        APPLY_GIVENS_H(*pt1,*(pt1+1),*pt2,*pt3,ctmp1,ctmp2)
      }
    } while ( i-- > 0 );
  }
#ifdef SNAP_QR_DEBUG
//  PRINT_MAT_R_row(s,n-1,k)
#endif // SNAP_QR_DEBUG
  SNAPBASECOMPLEX_clear(ctmp2);
  SNAPBASECOMPLEX_clear(ctmp1);
}

/*!
 *  @brief This computes a Givens rotation ((c,s),(-sbar,cbar)).((ca),(cb))=((r),(0)) for complexes, where r is a real number.
 *  @warning CLARTG or CROTG may be faster than this. But we need to make r a real number.
 *  @param[in,out] ca ca will be overwritten by the resulting r on exit.
 *  @param[in,out] cb cb will be overwritten by 0.
 *  @param[out] c c is the cosine term which should not share the pointer with other given variables.
 *  @param[out] s s is the sine term which should not share the pointer with other given variables.
 *  @param[in] ctmp ctmp is a work area which should not share the pointer with other given variables.
 */
void SNAP_alg_qr_dgv2010refine_givens_R(SNAPBASECOMPLEX_t *ca, SNAPBASECOMPLEX_t *cb, SNAPBASECOMPLEX_t *c, SNAPBASECOMPLEX_t *s, SNAPBASECOMPLEX_t *ctmp) {
#ifdef SNAP_QR_USE_RELATIVE_GIVENS
  double complex some;
  double complex *ctmp2;
  ctmp2 = &some;
  if ( SNAPBASECOMPLEX_cmpabs_basereal(*cb, SNAP_QR_THRESHOLD) < 0 ) {
    SNAPBASECOMPLEX_conj(*c, *ca);      // (c,s)=(bar{ca}/abs{ca},0) ==> g(ca,0)= (abs(ca), 0)
    SNAPBASECOMPLEX_abs(*ca, *ca);
    SNAPBASECOMPLEX_div(*c, *c, *ca);
    SNAPBASECOMPLEX_set_ui(*s, 0);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
    return;
  }
  if ( SNAPBASECOMPLEX_cmpabs(*ca, *cb) < 0 ) {
    SNAPBASECOMPLEX_div(*ctmp2, *ca, *cb);
    SNAPBASECOMPLEX_conj(*c, *ctmp2);
    SNAPBASECOMPLEX_set_ui(*s, 1);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp2, *c);
    SNAPBASECOMPLEX_add_d(*ctmp, *ctmp, 1.0);
    SNAPBASECOMPLEX_sqrt(*ctmp, *ctmp);

    SNAPBASECOMPLEX_conj(*ctmp2, *ctmp2);
    SNAPBASECOMPLEX_mul(*ctmp2, *ctmp2, *ca);
    SNAPBASECOMPLEX_add(*ca, *ctmp2, *cb);

    SNAPBASECOMPLEX_conj(*ctmp2, *cb);
    SNAPBASECOMPLEX_div(*ctmp, *ctmp, *ctmp2);
    SNAPBASECOMPLEX_abs(*ctmp2, *cb);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp, *ctmp2);

    SNAPBASECOMPLEX_div(*c, *c, *ctmp);
    SNAPBASECOMPLEX_div(*s, *s, *ctmp);
    SNAPBASECOMPLEX_div(*ca, *ca, *ctmp);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
  } else {
    SNAPBASECOMPLEX_div(*ctmp2, *cb, *ca);
    SNAPBASECOMPLEX_conj(*s, *ctmp2);
    SNAPBASECOMPLEX_set_ui(*c, 1);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp2, *s);
    SNAPBASECOMPLEX_add_d(*ctmp, *ctmp, 1.0);
    SNAPBASECOMPLEX_sqrt(*ctmp, *ctmp);

    SNAPBASECOMPLEX_conj(*ctmp2, *ctmp2);
    SNAPBASECOMPLEX_mul(*ctmp2, *ctmp2, *cb);
    SNAPBASECOMPLEX_add(*ca, *ctmp2, *ca);

    SNAPBASECOMPLEX_conj(*ctmp2, *ca);
    SNAPBASECOMPLEX_div(*ctmp, *ctmp, *ctmp2);
    SNAPBASECOMPLEX_abs(*ctmp2, *ca);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp, *ctmp2);

    SNAPBASECOMPLEX_div(*c, *c, *ctmp);
    SNAPBASECOMPLEX_div(*s, *s, *ctmp);
    SNAPBASECOMPLEX_div(*ca, *ca, *ctmp);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
  }
#else
  if ( SNAPBASECOMPLEX_cmpabs_basereal(*cb, SNAP_QR_THRESHOLD) < 0 ) {
    if ( SNAPBASECOMPLEX_cmpabs_basereal(*ca, SNAP_QR_THRESHOLD) < 0 ) {
      SNAPBASECOMPLEX_set_ui(*c, 1);      // (c,s)=(1,0) = id.
      SNAPBASECOMPLEX_set_ui(*s, 0);
      SNAPBASECOMPLEX_set_ui(*ca, 0);
      SNAPBASECOMPLEX_set_ui(*cb, 0);
    } else {
      SNAPBASECOMPLEX_conj(*c, *ca);      // (c,s)=(bar{ca}/abs{ca},0) ==> g(ca,0)= (abs(ca), 0)
      SNAPBASECOMPLEX_abs(*ca, *ca);
      SNAPBASECOMPLEX_div(*c, *c, *ca);
      SNAPBASECOMPLEX_set_ui(*s, 0);
    }
    return;
  } else if ( SNAPBASECOMPLEX_cmpabs_basereal(*ca, SNAP_QR_THRESHOLD) < 0 ) {
    SNAPBASECOMPLEX_conj(*s, *cb);        // (c,s)=(0,bar(cb)/abs(cb)) ==> g(0,cb)=(abs(cb),0)
    SNAPBASECOMPLEX_abs(*cb, *cb);
    SNAPBASECOMPLEX_div(*s, *s, *cb);
    SNAPBASECOMPLEX_set_ui(*c, 0);
    return;
  }
  SNAPBASECOMPLEX_abs(*c, *ca);
  SNAPBASECOMPLEX_abs(*s, *cb);
  SNAPBASECOMPLEX_add(*ctmp, *c, *s); // |ca|+|cb|
  SNAPBASECOMPLEX_div(*c, *ca, *ctmp);
  SNAPBASECOMPLEX_abs(*c, *c);
  SNAPBASECOMPLEX_mul(*c, *c, *c);
  SNAPBASECOMPLEX_div(*s, *cb, *ctmp);
  SNAPBASECOMPLEX_abs(*s, *s);
  SNAPBASECOMPLEX_mul(*s, *s, *s);
  SNAPBASECOMPLEX_add(*c, *c, *s);
  SNAPBASECOMPLEX_sqrt(*c, *c);
  SNAPBASECOMPLEX_mul(*ctmp, *ctmp, *c); // (|ca|+|cb|)sqrt(|ca/(|ca|+|cb|)|^2+|cb/(|ca|+|cb|)|^2)
  SNAPBASECOMPLEX_conj(*c, *ca);
  SNAPBASECOMPLEX_div(*c, *c, *ctmp);
  SNAPBASECOMPLEX_conj(*s, *cb);
  SNAPBASECOMPLEX_div(*s, *s, *ctmp);
  SNAPBASECOMPLEX_set(*ca, *ctmp);
  SNAPBASECOMPLEX_set_ui(*cb, 0);
#endif // SNAP_QR_USE_RELATIVE_GIVENS
}

/*!
 *  @brief This computes a Givens rotation ((c,s),(-sbar,cbar)).((ca),(cb))=((r),(0)) for complexes, where r is a complex number.
 *  @warning CLARTG or CROTG may be faster than this. But we would like to make c a complex number.
 *  @param[in,out] ca ca will be overwritten by the resulting r on exit.
 *  @param[in,out] cb cb will be overwritten by 0.
 *  @param[out] c c is the cosine term which should not share the pointer with other given variables.
 *  @param[out] s s is the sine term which should not share the pointer with other given variables.
 *  @param[in] ctmp ctmp is a work area which should not share the pointer with other given variables.
 */
void SNAP_alg_qr_dgv2010refine_givens_C(SNAPBASECOMPLEX_t *ca, SNAPBASECOMPLEX_t *cb, SNAPBASECOMPLEX_t *c, SNAPBASECOMPLEX_t *s, SNAPBASECOMPLEX_t *ctmp) {
#ifdef SNAP_QR_USE_RELATIVE_GIVENS
  double complex some;
  double complex *ctmp2;
  ctmp2 = &some;
  if ( SNAPBASECOMPLEX_cmpabs_basereal(*cb, SNAP_QR_THRESHOLD) < 0 ) {
    SNAPBASECOMPLEX_set_ui(*c, 1);  // (c,s)=(1,0) ==> id.
    SNAPBASECOMPLEX_set_ui(*s, 0);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
    return;
  }
  if ( SNAPBASECOMPLEX_cmpabs(*ca, *cb) < 0 ) {
    SNAPBASECOMPLEX_div(*ctmp2, *ca, *cb);
    SNAPBASECOMPLEX_conj(*c, *ctmp2);
    SNAPBASECOMPLEX_set_ui(*s, 1);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp2, *c);
    SNAPBASECOMPLEX_add_d(*ctmp, *ctmp, 1.0);
    SNAPBASECOMPLEX_sqrt(*ctmp, *ctmp);
    SNAPBASECOMPLEX_div(*c, *c, *ctmp);
    SNAPBASECOMPLEX_div(*s, *s, *ctmp);
    SNAPBASECOMPLEX_conj(*ctmp2, *ctmp2);
    SNAPBASECOMPLEX_mul(*ctmp2, *ctmp2, *ca);
    SNAPBASECOMPLEX_add(*ca, *ctmp2, *cb);
    SNAPBASECOMPLEX_div(*ca, *ca, *ctmp);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
  } else {
    SNAPBASECOMPLEX_div(*ctmp2, *cb, *ca);
    SNAPBASECOMPLEX_conj(*s, *ctmp2);
    SNAPBASECOMPLEX_set_ui(*c, 1);
    SNAPBASECOMPLEX_mul(*ctmp, *ctmp2, *s);
    SNAPBASECOMPLEX_add_d(*ctmp, *ctmp, 1.0);
    SNAPBASECOMPLEX_sqrt(*ctmp, *ctmp);
    SNAPBASECOMPLEX_div(*c, *c, *ctmp);
    SNAPBASECOMPLEX_div(*s, *s, *ctmp);
    SNAPBASECOMPLEX_conj(*ctmp2, *ctmp2);
    SNAPBASECOMPLEX_mul(*ctmp2, *ctmp2, *cb);
    SNAPBASECOMPLEX_add(*ca, *ctmp2, *ca);
    SNAPBASECOMPLEX_div(*ca, *ca, *ctmp);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
  }
#else
  if ( SNAPBASECOMPLEX_cmpabs_basereal(*cb, SNAP_QR_THRESHOLD) < 0 ) {
    SNAPBASECOMPLEX_set_ui(*c, 1);  // (c,s)=(1,0) ==> id.
    SNAPBASECOMPLEX_set_ui(*s, 0);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
    return;
  } else if ( SNAPBASECOMPLEX_cmpabs_basereal(*ca, SNAP_QR_THRESHOLD) < 0 ) {
    SNAPBASECOMPLEX_set_ui(*c, 0);  // (c,s)=(0,1) ==> g(0,cb)=(cb,0)
    SNAPBASECOMPLEX_set_ui(*s, 1);
    SNAPBASECOMPLEX_set(*ca, *cb);
    SNAPBASECOMPLEX_set_ui(*cb, 0);
    return;
  }
  SNAPBASECOMPLEX_abs(*c, *ca);
  SNAPBASECOMPLEX_abs(*s, *cb);
  SNAPBASECOMPLEX_add(*ctmp, *c, *s); // |ca|+|cb|
  SNAPBASECOMPLEX_div(*c, *ca, *ctmp);
  SNAPBASECOMPLEX_abs(*c, *c);
  SNAPBASECOMPLEX_mul(*c, *c, *c);
  SNAPBASECOMPLEX_div(*s, *cb, *ctmp);
  SNAPBASECOMPLEX_abs(*s, *s);
  SNAPBASECOMPLEX_mul(*s, *s, *s);
  SNAPBASECOMPLEX_add(*c, *c, *s);
  SNAPBASECOMPLEX_sqrt(*c, *c);
  SNAPBASECOMPLEX_mul(*ctmp, *ctmp, *c); // (|ca|+|cb|)sqrt(|ca/(|ca|+|cb|)|^2+|cb/(|ca|+|cb|)|^2)
  SNAPBASECOMPLEX_div(*c, *ca, *ctmp);
  SNAPBASECOMPLEX_conj(*s, *ca);
  SNAPBASECOMPLEX_div(*ca, *ca, *s);
  SNAPBASECOMPLEX_conj(*s, *cb);
  SNAPBASECOMPLEX_mul(*s, *s, *ca);
  SNAPBASECOMPLEX_div(*s, *s, *ctmp);
  SNAPBASECOMPLEX_mul(*ca, *ca, *ctmp);
  SNAPBASECOMPLEX_set_ui(*cb, 0);
#endif // SNAP_QR_USE_RELATIVE_GIVENS
}

/*!
 *  @brief This computes x s.t. Ax=0 by the least square.
 *  @warning This does not preserve a11,a12,a21 and a22.
 *  @param[in,out] a11 a11 is the element of A, and a temporal value on exit.
 *  @param[in,out] a12 a12 is the element of A, and a temporal value on exit.
 *  @param[in,out] a21 a21 is the element of A, and a temporal value on exit.
 *  @param[in,out] a22 a22 is the element of A, and a temporal value on exit.
 *  @param[out] x1 x1 is the element of x.
 *  @param[out] x2 x2 is the element of x.
 *  @param[in] w w is the temporal work area which must be an array of length 3 at least.
 */
void SNAP_alg_qr_dgv2010refine_givens_LS(SNAPBASECOMPLEX_t *a11, SNAPBASECOMPLEX_t *a12, SNAPBASECOMPLEX_t *a21, SNAPBASECOMPLEX_t *a22, SNAPBASECOMPLEX_t *x1, SNAPBASECOMPLEX_t *x2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptmp1, *ptmp2, *ptmp3;
  ptmp1 = w++;
  ptmp2 = w++;
  ptmp3 = w;

  SNAPBASECOMPLEX_abs(*ptmp1, *a11);
  SNAPBASECOMPLEX_abs(*ptmp2, *a21);
  SNAPBASECOMPLEX_add(*ptmp1, *ptmp1, *ptmp2);
  SNAPBASECOMPLEX_abs(*ptmp2, *a12);
  SNAPBASECOMPLEX_abs(*ptmp3, *a22);
  SNAPBASECOMPLEX_add(*ptmp2, *ptmp2, *ptmp3);
  if ( SNAPBASECOMPLEX_cmp_realpart(*ptmp1, *ptmp2) > 0 ) {
    SNAP_alg_qr_dgv2010refine_givens_C(a11, a21, x1, x2, ptmp3);
    APPLY_GIVENS(*x1,*x2,*a12,*a22,*ptmp1,*ptmp2);
    SNAP_alg_qr_dgv2010refine_givens_C(a11, a12, ptmp1, ptmp2, ptmp3);
    SNAPBASECOMPLEX_conj(*x2, *ptmp1);
    SNAPBASECOMPLEX_conj(*x1, *ptmp2);
    SNAPBASECOMPLEX_neg(*x1, *x1);
  } else {
    SNAP_alg_qr_dgv2010refine_givens_C(a12, a22, x1, x2, ptmp3);
    APPLY_GIVENS(*x1,*x2,*a11,*a21,*ptmp1,*ptmp2);
    SNAP_alg_qr_dgv2010refine_givens_C(a11, a12, ptmp1, ptmp2, ptmp3);
    SNAPBASECOMPLEX_conj(*x2, *ptmp1);
    SNAPBASECOMPLEX_conj(*x1, *ptmp2);
    SNAPBASECOMPLEX_neg(*x1, *x1);
  }
}

/*!
 *  @brief This computes x s.t. Givens(c,s)^h A=((x,0),(0,x)) by the least square and reflect it as in DGV2010.
 *  @warning This does not preserve a11,a12,a21 and a22.
 *  @param[in,out] a11 a11 is the element of A, and (1,1) element of Givens(c,s)^h A on exit.
 *  @param[in,out] a12 a12 is the element of A, and a temporal value on exit.
 *  @param[in,out] a21 a21 is the element of A, and a temporal value on exit.
 *  @param[in,out] a22 a22 is the element of A, and (2,2) element of Givens(c,s)^h A on exit.
 *  @param[out] c c is the cosine term of Givens(c,s)^h.
 *  @param[out] s s is the sine term of Givens(c,s)^h.
 *  @param[in] w w is the temporal work area which must be an array of length 7=4+3 at least.
 */
void SNAP_alg_qr_dgv2010refine_givens_LS4RF(SNAPBASECOMPLEX_t *a11, SNAPBASECOMPLEX_t *a12, SNAPBASECOMPLEX_t *a21, SNAPBASECOMPLEX_t *a22, SNAPBASECOMPLEX_t *c, SNAPBASECOMPLEX_t *s, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptmp1, *ptmp2, *ptmp3, *ptmp4;
  ptmp1 = w++;
  ptmp2 = w++;
  ptmp3 = w++;
  ptmp4 = w++;
  SNAPBASECOMPLEX_conj(*ptmp1, *a21);
  SNAPBASECOMPLEX_conj(*ptmp2, *a11);
  SNAPBASECOMPLEX_set(*ptmp3, *a12);
  SNAPBASECOMPLEX_neg(*ptmp4, *a22);
  SNAP_alg_qr_dgv2010refine_givens_LS(ptmp1, ptmp2, ptmp3, ptmp4, c, s, w);
  SNAPBASECOMPLEX_conj(*c, *c);             // c = the actual solution of Givens(c,s)^h A=((x,0),(0,x)).

  SNAPBASECOMPLEX_conj(*ptmp1, *c);         // t1 = conj(c)
  SNAPBASECOMPLEX_mul(*a11, *a11, *ptmp1);  // a11 = a11 * t1 = conj(c)*a11
  SNAPBASECOMPLEX_mul(*ptmp1, *a21, *s);    // t1 = a21 * s
  SNAPBASECOMPLEX_sub(*a11, *a11, *ptmp1);  // a11 = a11 - t1 = conj(c)*a11 - a21*s

  SNAPBASECOMPLEX_mul(*a22, *a22, *c);      // a22 = a22 * c
  SNAPBASECOMPLEX_conj(*ptmp1, *s);         // t1 = conj(s)
  SNAPBASECOMPLEX_mul(*ptmp1, *ptmp1, *a12);// t1 = t1 * a12 = conj(s)*a12
  SNAPBASECOMPLEX_add(*a22, *a22, *ptmp1);  // a22 = a22 + t1 = a22*c + conj(s)*a12
}

/*!
 *  @brief This computes the refactorization of unitary transformation by the pull-through lemma (i.e. g0g1g2 = g1'g2'g0').
 *  @param[in,out] g0 g0 will be overwritten by the resulting g0'.
 *  @param[in,out] g1 g1 will be overwritten by the resulting g1'.
 *  @param[in,out] g2 g2 will be overwritten by the resulting g2'.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pulldownthrough(SNAPBASECOMPLEX_t *g0, SNAPBASECOMPLEX_t *g1, SNAPBASECOMPLEX_t *g2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptw11, *ptw12, *ptw13, *ptw21, *ptw22, *ptw23, *ptw31, *ptw32, *ptw33, *pg0c, *pg0s, *pg1c, *pg1s, *pg2c, *pg2s, *ptmp1, *ptmp2;
  ptw11 = w++;
  ptw12 = w++;
  ptw13 = w++;
  ptw21 = w++;
  ptw22 = w++;
  ptw23 = w++;
  ptw31 = w++;
  ptw32 = w++;
  ptw33 = w++;
  ptmp1 = w++;
  ptmp2 = w;
  pg0c = g0++;
  pg0s = g0;
  pg1c = g1++;
  pg1s = g1;
  pg2c = g2++;
  pg2s = g2;
  // construct g0.g1.g2
  SNAPBASECOMPLEX_set(*ptw11, *pg2c);
  SNAPBASECOMPLEX_conj(*ptw21, *pg2s);
  SNAPBASECOMPLEX_neg(*ptw21, *ptw21);
  SNAPBASECOMPLEX_conj(*ptw32, *pg1s);
  SNAPBASECOMPLEX_neg(*ptw32, *ptw32);
  SNAPBASECOMPLEX_mul(*ptw31, *ptw32, *ptw21);
  SNAPBASECOMPLEX_mul(*ptw21, *ptw21, *pg1c);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw11, *ptw21, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw12, *pg2s);
  SNAPBASECOMPLEX_conj(*ptw22, *pg2c);
  SNAPBASECOMPLEX_mul(*ptw32, *ptw32, *ptw22);
  SNAPBASECOMPLEX_mul(*ptw22, *ptw22, *pg1c);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw23, *pg1s);
  SNAPBASECOMPLEX_conj(*ptw33, *pg1c);
  SNAPBASECOMPLEX_mul(*ptw13, *ptw23, *pg0s);
  SNAPBASECOMPLEX_conj(*ptmp1, *pg0c);
  SNAPBASECOMPLEX_mul(*ptw23, *ptw23, *ptmp1);
#ifdef SNAP_QR_DEBUG
  printf("========== pulldownthrough whole matrix =========\n");
  printf("(%+.18e%+.18eI),",creal(*(ptw11)),cimag(*(ptw11)));
  printf("(%+.18e%+.18eI),",creal(*(ptw12)),cimag(*(ptw12)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw13)),cimag(*(ptw13)));
  printf("(%+.18e%+.18eI),",creal(*(ptw21)),cimag(*(ptw21)));
  printf("(%+.18e%+.18eI),",creal(*(ptw22)),cimag(*(ptw22)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw23)),cimag(*(ptw23)));
  printf("(%+.18e%+.18eI),",creal(*(ptw31)),cimag(*(ptw31)));
  printf("(%+.18e%+.18eI),",creal(*(ptw32)),cimag(*(ptw32)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw33)),cimag(*(ptw33)));
  printf("\n");
#endif // SNAP_QR_DEBUG
  // refactorize g1'.g2'.g0'
#ifdef SNAP_QR_USE_GIVENS_R_ONLY
  SNAP_alg_qr_dgv2010refine_givens_R(ptw21, ptw31, pg1c, pg1s, ptmp1);
#else
  SNAP_alg_qr_dgv2010refine_givens_C(ptw21, ptw31, pg1c, pg1s, ptmp1);
#endif // SNAP_QR_USE_GIVENS_R_ONLY
  APPLY_GIVENS(*pg1c, *pg1s, *ptw22, *ptw32, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg1c, *pg1s, *ptw23, *ptw33, *ptmp1, *ptmp2)
#ifdef SNAP_QR_DEBUG
  printf("========== givens =========\n");
  printf("(%+.18e%+.18eI),",creal(*(pg1c)),cimag(*(pg1c)));
  printf("(%+.18e%+.18eI)\n",creal(*(pg1s)),cimag(*(pg1s)));
  printf("========== pulldownthrough whole matrix =========\n");
  printf("(%+.18e%+.18eI),",creal(*(ptw11)),cimag(*(ptw11)));
  printf("(%+.18e%+.18eI),",creal(*(ptw12)),cimag(*(ptw12)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw13)),cimag(*(ptw13)));
  printf("(%+.18e%+.18eI),",creal(*(ptw21)),cimag(*(ptw21)));
  printf("(%+.18e%+.18eI),",creal(*(ptw22)),cimag(*(ptw22)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw23)),cimag(*(ptw23)));
  printf("(%+.18e%+.18eI),",creal(*(ptw31)),cimag(*(ptw31)));
  printf("(%+.18e%+.18eI),",creal(*(ptw32)),cimag(*(ptw32)));
  printf("(%+.18e%+.18eI)\n",creal(*(ptw33)),cimag(*(ptw33)));
  printf("\n");
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg2c, *ptw11);
  SNAPBASECOMPLEX_conj(*pg2s, *ptw21);
#else
  SNAP_alg_qr_dgv2010refine_givens_R(ptw11, ptw21, pg2c, pg2s, ptmp1);
#endif // SNAP_QR_USE_FIRST_ROW_FOR_SECONDPULLTHROUGH
  APPLY_GIVENS(*pg2c, *pg2s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg2c, *pg2s, *ptw13, *ptw23, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*ptw33, *ptw33);
  SNAPBASECOMPLEX_add(*ptw22, *ptw22, *ptw33);
  SNAPBASECOMPLEX_div_d(*ptw22, *ptw22, 2);
  SNAPBASECOMPLEX_conj(*ptw23, *ptw23);
  SNAPBASECOMPLEX_sub(*ptw32, *ptw32, *ptw23);
  SNAPBASECOMPLEX_div_d(*ptw32, *ptw32, 2);
  SNAP_alg_qr_dgv2010refine_givens_R(ptw22, ptw32, pg0c, pg0s, ptmp1);
  SNAPBASECOMPLEX_conj(*pg0c, *pg0c);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#elif defined(SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH)
  SNAPBASECOMPLEX_conj(*pg0c, *ptw33);
  SNAPBASECOMPLEX_conj(*pg0s, *ptw32);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#else
  SNAPBASECOMPLEX_set(*pg0c, *ptw22);
  SNAPBASECOMPLEX_set(*pg0s, *ptw23);
#endif // SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg1c, *pg1c);
  SNAPBASECOMPLEX_neg(*pg1s, *pg1s);
  SNAPBASECOMPLEX_conj(*pg2c, *pg2c);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
}

/*!
 *  @brief This computes the refactorization of unitary transformation by the pull-through lemma (i.e. g0s1g1s2g2 = s1g1's2g2'g0').
 *  @param[in,out] g0 g0 will be overwritten by the resulting g0'.
 *  @param[in] s1 s1 is the lower scaling factor whose absolute value is 1.
 *  @param[in,out] g1 g1 will be overwritten by the resulting g1'.
 *  @param[in] s2 s2 is the upper scaling factor whose absolute value is 1.
 *  @param[in,out] g2 g2 will be overwritten by the resulting g2'.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pulldownthrough_ws(SNAPBASECOMPLEX_t *g0, SNAPBASECOMPLEX_t *s1, SNAPBASECOMPLEX_t *g1, SNAPBASECOMPLEX_t *s2, SNAPBASECOMPLEX_t *g2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptw11, *ptw12, *ptw13, *ptw21, *ptw22, *ptw23, *ptw31, *ptw32, *ptw33, *pg0c, *pg0s, *pg1c, *pg1s, *pg2c, *pg2s, *ptmp1, *ptmp2;
  ptw11 = w++;
  ptw12 = w++;
  ptw13 = w++;
  ptw21 = w++;
  ptw22 = w++;
  ptw23 = w++;
  ptw31 = w++;
  ptw32 = w++;
  ptw33 = w++;
  ptmp1 = w++;
  ptmp2 = w;
  pg0c = g0++;
  pg0s = g0;
  pg1c = g1++;
  pg1s = g1;
  pg2c = g2++;
  pg2s = g2;
  // we must change g0s before computing.
  SNAPBASECOMPLEX_conj(*ptmp1, *s2);
  SNAPBASECOMPLEX_mul(*pg0s, *pg0s, *ptmp1);
  SNAPBASECOMPLEX_mul(*pg0s, *pg0s, *s1);
  // construct g0.g1.g2
  SNAPBASECOMPLEX_set(*ptw11, *pg2c);
  SNAPBASECOMPLEX_conj(*ptw21, *pg2s);
  SNAPBASECOMPLEX_neg(*ptw21, *ptw21);
  SNAPBASECOMPLEX_conj(*ptw32, *pg1s);
  SNAPBASECOMPLEX_neg(*ptw32, *ptw32);
  SNAPBASECOMPLEX_mul(*ptw31, *ptw32, *ptw21);
  SNAPBASECOMPLEX_mul(*ptw21, *ptw21, *pg1c);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw11, *ptw21, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw12, *pg2s);
  SNAPBASECOMPLEX_conj(*ptw22, *pg2c);
  SNAPBASECOMPLEX_mul(*ptw32, *ptw32, *ptw22);
  SNAPBASECOMPLEX_mul(*ptw22, *ptw22, *pg1c);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw23, *pg1s);
  SNAPBASECOMPLEX_conj(*ptw33, *pg1c);
  SNAPBASECOMPLEX_mul(*ptw13, *ptw23, *pg0s);
  SNAPBASECOMPLEX_conj(*ptmp1, *pg0c);
  SNAPBASECOMPLEX_mul(*ptw23, *ptw23, *ptmp1);
  // refactorize g1'.g2'.g0'
#ifdef SNAP_QR_USE_GIVENS_R_ONLY
  SNAP_alg_qr_dgv2010refine_givens_R(ptw21, ptw31, pg1c, pg1s, ptmp1);
#else
  SNAP_alg_qr_dgv2010refine_givens_C(ptw21, ptw31, pg1c, pg1s, ptmp1);
#endif // SNAP_QR_USE_GIVENS_R_ONLY
  APPLY_GIVENS(*pg1c, *pg1s, *ptw22, *ptw32, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg1c, *pg1s, *ptw23, *ptw33, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg2c, *ptw11);
  SNAPBASECOMPLEX_conj(*pg2s, *ptw21);
#else
  SNAP_alg_qr_dgv2010refine_givens_R(ptw11, ptw21, pg2c, pg2s, ptmp1);
#endif // SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  APPLY_GIVENS(*pg2c, *pg2s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg2c, *pg2s, *ptw13, *ptw23, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*ptw33, *ptw33);
  SNAPBASECOMPLEX_add(*ptw22, *ptw22, *ptw33);
  SNAPBASECOMPLEX_div_d(*ptw22, *ptw22, 2);
  SNAPBASECOMPLEX_conj(*ptw23, *ptw23);
  SNAPBASECOMPLEX_sub(*ptw32, *ptw32, *ptw23);
  SNAPBASECOMPLEX_div_d(*ptw32, *ptw32, 2);
  SNAP_alg_qr_dgv2010refine_givens_R(ptw22, ptw32, pg0c, pg0s, ptmp1);
  SNAPBASECOMPLEX_conj(*pg0c, *pg0c);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#elif defined(SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH)
  SNAPBASECOMPLEX_conj(*pg0c, *ptw33);
  SNAPBASECOMPLEX_conj(*pg0s, *ptw32);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#else
  SNAPBASECOMPLEX_set(*pg0c, *ptw22);
  SNAPBASECOMPLEX_set(*pg0s, *ptw23);
#endif // SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg1c, *pg1c);
  SNAPBASECOMPLEX_neg(*pg1s, *pg1s);
  SNAPBASECOMPLEX_conj(*pg2c, *pg2c);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
}

/*!
 *  @brief This apply the scalars to the Givens rotation on the left.
 *  @param[in,out] g0 g0 will be overwritten by the resulting g0'.
 *  @param[in] s1 s1 is the lower scaling factor whose absolute value is 1.
 *  @param[in] s2 s2 is the upper scaling factor whose absolute value is 1.
 *  @param[in] w w is the temporal work area which must be an array of length 1 at least.
 */
void SNAP_alg_qr_dgv2010refine_apply_scalar(SNAPBASECOMPLEX_t *g0, SNAPBASECOMPLEX_t *s1, SNAPBASECOMPLEX_t *s2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *pg0s, *ptmp1;
  ptmp1 = w;
  pg0s = ++g0;
  // we must change g0s before computing.
  SNAPBASECOMPLEX_conj(*ptmp1, *s2);
  SNAPBASECOMPLEX_mul(*pg0s, *pg0s, *ptmp1);
  SNAPBASECOMPLEX_mul(*pg0s, *pg0s, *s1);
}

/*!
 *  @brief This pulls the given Givens rotations through the given unitary structure to the right.
 *  @param[in] r r is the displacement rank.
 *  @param[in,out] g g is the Givens rotation will be pulled through, and the resulting rotation on exit.
 *  @param[in,out] scalarS1 scalarS1 is the lower scaling factor of leftS1.
 *  @param[in,out] leftS1 leftS1 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] scalarS2 scalarS2 is the upper scaling factor of leftS2.
 *  @param[in,out] leftS2 leftS2 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] num_op num_op is the number of required single pull-through operations.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pulldownthroughright(const unsigned int r, SNAPBASECOMPLEX_t *g, SNAPBASECOMPLEX_t *scalarS1, SNAPBASECOMPLEX_t *leftS1, SNAPBASECOMPLEX_t *scalarS2, SNAPBASECOMPLEX_t *leftS2, const unsigned int num_op, SNAPBASECOMPLEX_t *w) {
  unsigned int i;
  if ( num_op == 0 ) {
    SNAP_alg_qr_dgv2010refine_apply_scalar(g, scalarS1, scalarS2, w);
    return;
  }
  SNAP_alg_qr_dgv2010refine_pulldownthrough_ws(g, scalarS1, leftS1, scalarS2, leftS2, w);
#ifdef SNAP_QR_DEBUG
  printf("========== leftS1, leftS2 =========\n");
  printf("(%+.18e%+.18eI),",creal(*(leftS1)),cimag(*(leftS1)));
  printf("(%+.18e%+.18eI),",creal(*(leftS1+1)),cimag(*(leftS1+1)));
  printf("(%+.18e%+.18eI),",creal(*(leftS2)),cimag(*(leftS2)));
  printf("(%+.18e%+.18eI),",creal(*(leftS2+1)),cimag(*(leftS2+1)));
  printf("\n");
#endif // SNAP_QR_DEBUG
  for ( i = 1, leftS1+=2, leftS2+=2; i < num_op; i++, leftS1+=2, leftS2+=2 ) {
    SNAP_alg_qr_dgv2010refine_pulldownthrough(g, leftS1, leftS2, w);
#ifdef SNAP_QR_DEBUG
    printf("========== leftS1, leftS2 =========\n");
    printf("(%+.18e%+.18eI),",creal(*(leftS1)),cimag(*(leftS1)));
    printf("(%+.18e%+.18eI),",creal(*(leftS1+1)),cimag(*(leftS1+1)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2)),cimag(*(leftS2)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2+1)),cimag(*(leftS2+1)));
    printf("\n");
#endif // SNAP_QR_DEBUG
  }
}

/*!
 *  @brief This computes the refactorization of unitary transformation by the pull-through lemma (i.e. g0g1g2 = g1'g2'g0').
 *  @param[in,out] g0 g0 will be overwritten by the resulting g0'.
 *  @param[in,out] g1 g1 will be overwritten by the resulting g1'.
 *  @param[in,out] g2 g2 will be overwritten by the resulting g2'.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pullupthrough(SNAPBASECOMPLEX_t *g0, SNAPBASECOMPLEX_t *g1, SNAPBASECOMPLEX_t *g2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptw11, *ptw12, *ptw13, *ptw21, *ptw22, *ptw23, *ptw31, *ptw32, *ptw33, *pg0c, *pg0s, *pg1c, *pg1s, *pg2c, *pg2s, *ptmp1, *ptmp2;
  ptw11 = w++;
  ptw12 = w++;
  ptw13 = w++;
  ptw21 = w++;
  ptw22 = w++;
  ptw23 = w++;
  ptw31 = w++;
  ptw32 = w++;
  ptw33 = w++;
  ptmp1 = w++;
  ptmp2 = w;
  pg0c = g0++;
  pg0s = g0;
  pg1c = g1++;
  pg1s = g1;
  pg2c = g2++;
  pg2s = g2;
  // construct g0.g1.g2
  SNAPBASECOMPLEX_set(*ptw11, *pg1c);
  SNAPBASECOMPLEX_conj(*ptw21, *pg1s);
  SNAPBASECOMPLEX_neg(*ptw21, *ptw21);
  SNAPBASECOMPLEX_conj(*ptw31, *pg0s);
  SNAPBASECOMPLEX_neg(*ptw31, *ptw31);
  SNAPBASECOMPLEX_mul(*ptw31, *ptw31, *ptw21);
  SNAPBASECOMPLEX_mul(*ptw21, *ptw21, *pg0c);
  SNAPBASECOMPLEX_set(*ptw22, *pg2c);
  SNAPBASECOMPLEX_conj(*ptw32, *pg2s);
  SNAPBASECOMPLEX_neg(*ptw32, *ptw32);
  SNAPBASECOMPLEX_mul(*ptw12, *ptw22, *pg1s);
  SNAPBASECOMPLEX_conj(*ptmp1, *pg1c);
  SNAPBASECOMPLEX_mul(*ptw22, *ptw22, *ptmp1);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw22, *ptw32, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw23, *pg2s);
  SNAPBASECOMPLEX_conj(*ptw33, *pg2c);
  SNAPBASECOMPLEX_mul(*ptw13, *ptw23, *pg1s);
  SNAPBASECOMPLEX_conj(*ptmp1, *pg1c);
  SNAPBASECOMPLEX_mul(*ptw23, *ptw23, *ptmp1);
  APPLY_GIVENS(*pg0c, *pg0s, *ptw23, *ptw33, *ptmp1, *ptmp2)
  // refactorize g1'.g2'.g0'
#ifdef SNAP_QR_USE_GIVENS_R_ONLY
  SNAP_alg_qr_dgv2010refine_givens_R(ptw23, ptw13, pg1c, pg1s, ptmp1);
#else
  SNAP_alg_qr_dgv2010refine_givens_C(ptw23, ptw13, pg1c, pg1s, ptmp1);
#endif // SNAP_QR_USE_GIVENS_R_ONLY
  SNAPBASECOMPLEX_conj(*pg1c, *pg1c);
  SNAPBASECOMPLEX_conj(*pg1s, *pg1s);
  SNAPBASECOMPLEX_neg(*pg1s, *pg1s);
  APPLY_GIVENS(*pg1c, *pg1s, *ptw11, *ptw21, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg1c, *pg1s, *ptw12, *ptw22, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  SNAPBASECOMPLEX_set(*pg2c, *ptw33);
  SNAPBASECOMPLEX_neg(*pg2s, *ptw23);
#else
  SNAP_alg_qr_dgv2010refine_givens_R(ptw33, ptw23, pg2c, pg2s, ptmp1);
  SNAPBASECOMPLEX_conj(*pg2c, *pg2c);
  SNAPBASECOMPLEX_conj(*pg2s, *pg2s);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
#endif // SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  APPLY_GIVENS(*pg2c, *pg2s, *ptw21, *ptw31, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg2c, *pg2s, *ptw22, *ptw32, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*ptw22, *ptw22);
  SNAPBASECOMPLEX_add(*ptw11, *ptw11, *ptw22);
  SNAPBASECOMPLEX_div_d(*ptw11, *ptw11, 2);
  SNAPBASECOMPLEX_conj(*ptw12, *ptw12);
  SNAPBASECOMPLEX_sub(*ptw21, *ptw21, *ptw12);
  SNAPBASECOMPLEX_div_d(*ptw21, *ptw21, 2);
  SNAP_alg_qr_dgv2010refine_givens_R(ptw11, ptw21, pg0c, pg0s, ptmp1);
  SNAPBASECOMPLEX_conj(*pg0c, *pg0c);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#elif defined(SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH)
  SNAPBASECOMPLEX_conj(*pg0c, *ptw22);
  SNAPBASECOMPLEX_conj(*pg0s, *ptw21);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
#else
  SNAPBASECOMPLEX_set(*pg0c, *ptw11);
  SNAPBASECOMPLEX_set(*pg0s, *ptw12);
#endif // SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg1c, *pg1c);
  SNAPBASECOMPLEX_neg(*pg1s, *pg1s);
  SNAPBASECOMPLEX_conj(*pg2c, *pg2c);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
}

/*!
 *  @brief This pulls the given Givens rotations through the given unitary structure to the right.
 *  @param[in] r r is the displacement rank.
 *  @param[in,out] g g is the Givens rotation will be pulled through, and the resulting rotation on exit.
 *  @param[in,out] rightS1 rightS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] rightS2 rightS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] num_op num_op is the number of required single pull-through operations.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pullupthroughright(const unsigned int r, SNAPBASECOMPLEX_t *g, SNAPBASECOMPLEX_t *rightS1, SNAPBASECOMPLEX_t *rightS2, const unsigned int num_op, SNAPBASECOMPLEX_t *w) {
  unsigned int i;
  if ( num_op == 0 ) return;
  for ( i = 0, rightS1+=2*(r-2), rightS2+=2*(r-2); i < num_op; i++, rightS1-=2, rightS2-=2 ) {
    SNAP_alg_qr_dgv2010refine_pullupthrough(g, rightS1, rightS2, w);
  }
}

/*!
 *  @brief This computes the refactorization of unitary transformation by the pull-through lemma (i.e. g1g2g0 = g0'g1'g2').
 *  @param[in,out] g0 g0 will be overwritten by the resulting g0'.
 *  @param[in,out] g1 g1 will be overwritten by the resulting g1'.
 *  @param[in,out] g2 g2 will be overwritten by the resulting g2'.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pulldownthrough_left(SNAPBASECOMPLEX_t *g0, SNAPBASECOMPLEX_t *g1, SNAPBASECOMPLEX_t *g2, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *ptw11, *ptw12, *ptw13, *ptw21, *ptw22, *ptw23, *ptw31, *ptw32, *ptw33, *pg0c, *pg0s, *pg1c, *pg1s, *pg2c, *pg2s, *ptmp1, *ptmp2;
  ptw11 = w++;
  ptw12 = w++;
  ptw13 = w++;
  ptw21 = w++;
  ptw22 = w++;
  ptw23 = w++;
  ptw31 = w++;
  ptw32 = w++;
  ptw33 = w++;
  ptmp1 = w++;
  ptmp2 = w;
  pg0c = g0++;
  pg0s = g0;
  pg1c = g1++;
  pg1s = g1;
  pg2c = g2++;
  pg2s = g2;
  // construct g1.g2.g0
  SNAPBASECOMPLEX_set(*ptw11, *pg0c);
  SNAPBASECOMPLEX_conj(*ptw21, *pg0s);
  SNAPBASECOMPLEX_neg(*ptw21, *ptw21);
  SNAPBASECOMPLEX_conj(*ptw32, *pg2s);
  SNAPBASECOMPLEX_neg(*ptw32, *ptw32);
  SNAPBASECOMPLEX_mul(*ptw31, *ptw32, *ptw21);
  SNAPBASECOMPLEX_mul(*ptw21, *ptw21, *pg2c);
  APPLY_GIVENS(*pg1c, *pg1s, *ptw11, *ptw21, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw12, *pg0s);
  SNAPBASECOMPLEX_conj(*ptw22, *pg0c);
  SNAPBASECOMPLEX_mul(*ptw32, *ptw32, *ptw22);
  SNAPBASECOMPLEX_mul(*ptw22, *ptw22, *pg2c);
  APPLY_GIVENS(*pg1c, *pg1s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  SNAPBASECOMPLEX_set(*ptw23, *pg2s);
  SNAPBASECOMPLEX_conj(*ptw33, *pg2c);
  SNAPBASECOMPLEX_mul(*ptw13, *ptw23, *pg1s);
  SNAPBASECOMPLEX_conj(*ptmp1, *pg1c);
  SNAPBASECOMPLEX_mul(*ptw23, *ptw23, *ptmp1);
  // refactorize g0'.g1'.g2'
#ifdef SNAP_QR_USE_GIVENS_R_ONLY
  SNAP_alg_qr_dgv2010refine_givens_R(ptw21, ptw31, pg0c, pg0s, ptmp1);
#else
  SNAP_alg_qr_dgv2010refine_givens_C(ptw21, ptw31, pg0c, pg0s, ptmp1);
#endif // SNAP_QR_USE_GIVENS_R_ONLY
  APPLY_GIVENS(*pg0c, *pg0s, *ptw22, *ptw32, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg0c, *pg0s, *ptw23, *ptw33, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg1c, *ptw11);
  SNAPBASECOMPLEX_conj(*pg1s, *ptw21);
#else
  SNAP_alg_qr_dgv2010refine_givens_R(ptw11, ptw21, pg1c, pg1s, ptmp1);
#endif // SNAP_QR_USE_FIRST_COLUMN_FOR_SECONDPULLTHROUGH
  APPLY_GIVENS(*pg1c, *pg1s, *ptw12, *ptw22, *ptmp1, *ptmp2)
  APPLY_GIVENS(*pg1c, *pg1s, *ptw13, *ptw23, *ptmp1, *ptmp2)
#ifdef SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*ptw33, *ptw33);
  SNAPBASECOMPLEX_add(*ptw22, *ptw22, *ptw33);
  SNAPBASECOMPLEX_div_d(*ptw22, *ptw22, 2);
  SNAPBASECOMPLEX_conj(*ptw23, *ptw23);
  SNAPBASECOMPLEX_sub(*ptw32, *ptw32, *ptw23);
  SNAPBASECOMPLEX_div_d(*ptw32, *ptw32, 2);
  SNAP_alg_qr_dgv2010refine_givens_R(ptw22, ptw32, pg2c, pg2s, ptmp1);
  SNAPBASECOMPLEX_conj(*pg2c, *pg2c);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
#elif defined(SNAP_QR_USE_LAST_ROW_FOR_LASTPULLTHROUGH)
  SNAPBASECOMPLEX_conj(*pg2c, *ptw33);
  SNAPBASECOMPLEX_conj(*pg2s, *ptw32);
  SNAPBASECOMPLEX_neg(*pg2s, *pg2s);
#else
  SNAPBASECOMPLEX_set(*pg2c, *ptw22);
  SNAPBASECOMPLEX_set(*pg2s, *ptw23);
#endif // SNAP_QR_USE_GIVENS_FOR_LASTPULLTHROUGH
  SNAPBASECOMPLEX_conj(*pg0c, *pg0c);
  SNAPBASECOMPLEX_neg(*pg0s, *pg0s);
  SNAPBASECOMPLEX_conj(*pg1c, *pg1c);
  SNAPBASECOMPLEX_neg(*pg1s, *pg1s);
}

/*!
 *  @brief This pulls the given Givens rotations through the given unitary structure to the left.
 *  @param[in] r r is the displacement rank.
 *  @param[in,out] g g is the Givens rotation will be pulled through, and the resulting rotation on exit.
 *  @param[in,out] rightS1 rightS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] rightS2 rightS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] num_op num_op is the number of required single pull-through operations.
 *  @param[in] w w is the temporal work area which must be an array of length 11=3x3+2 at least.
 */
void SNAP_alg_qr_dgv2010refine_pulldownthroughleft(const unsigned int r, SNAPBASECOMPLEX_t *g, SNAPBASECOMPLEX_t *rightS1, SNAPBASECOMPLEX_t *rightS2, const unsigned int num_op, SNAPBASECOMPLEX_t *w) {
  unsigned int i;
  if ( num_op == 0 ) return;
  for ( i = 0; i < num_op; i++, rightS1+=2, rightS2+=2 ) {
    SNAP_alg_qr_dgv2010refine_pulldownthrough_left(g, rightS1, rightS2, w);
  }
}

/*!
 *  @brief This chases the given Givens rotations through the given unitary structure to the right (embeded version).
 *  @param[in] r r is the displacement rank.
 *  @param[in,out] g g is the Givens rotation will be pulled through, and the resulting rotation on exit.
 *  @param[in,out] leftS1 leftS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] leftS2 leftS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] scalarS1 scalarS1 is the pointer to the upper scalar of scalarS.
 *  @param[in] scalarS2 scalarS2 is the pointer to the lower scalar of scalarS.
 *  @param[in,out] rightS1 rightS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] rightS2 rightS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] matX matX is the temporal work area which must be an array of length (r+2)x2 at least.
 *  @param[in] matY matY is the temporal work area which must be an array of length (r+2)x2 at least.
 *  @param[in] w w is the temporal work area which must be an array of length 21=4+7+2x5 at least.
 */
void SNAP_alg_qr_dgv2010refine_chasing_embed(const unsigned int r, SNAPBASECOMPLEX_t *g, SNAPBASECOMPLEX_t *leftS1, SNAPBASECOMPLEX_t *leftS2, SNAPBASECOMPLEX_t *scalarS1, SNAPBASECOMPLEX_t *scalarS2, SNAPBASECOMPLEX_t *rightS1, SNAPBASECOMPLEX_t *rightS2, SNAPBASECOMPLEX_t *matX, SNAPBASECOMPLEX_t *matY, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *vecX1, *vecX2, *vecY1, *vecY2, *pt1, *pt2, *pt3, *pt4, *pt5, *pt6, *pt7, *pt8, *ptmp1, *ptmp2, *ptmp3, *ptmp4, *aaG, *newH, *qtmp1, *qtmp2, *qtmp3;
  unsigned int i;
  if ( r < 2 ) {
    printf("Error: r should be larger than 1.\n");
    exit(EXIT_FAILURE);
  }
  ptmp1 = w;
  ptmp2 = w+1;
  ptmp3 = w+2;
  ptmp4 = w+3;
  aaG = w + 11;
  newH = aaG + 2;
  qtmp1 = newH + 2;
  qtmp2 = qtmp1 + 2;
  qtmp3 = qtmp2 + 2;
#ifdef SNAP_QR_DEBUG
  printf("========== scalarS1, scalarS2 =========\n");
  printf("(%+.18e%+.18eI),",creal(*(scalarS1)),cimag(*(scalarS1)));
  printf("(%+.18e%+.18eI),",creal(*(scalarS2)),cimag(*(scalarS2)));
  printf("\n========== leftS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*2)),cimag(*(leftS1+i*2)));
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*2+1)),cimag(*(leftS1+i*2+1)));
  }
  printf("\n========== leftS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*2)),cimag(*(leftS2+i*2)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*2+1)),cimag(*(leftS2+i*2+1)));
  }
  printf("\n========== rightS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*2)),cimag(*(rightS1+i*2)));
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*2+1)),cimag(*(rightS1+i*2+1)));
  }
  printf("\n========== rightS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*2)),cimag(*(rightS2+i*2)));
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*2+1)),cimag(*(rightS2+i*2+1)));
  }
  printf("\n");
#endif // SNAP_QR_DEBUG
  // initializations for vecX1, vecX2, vecY1, vecY2
  vecX1 = matX;
  vecX2 = matX+r+2;
  vecY1 = matY;
  vecY2 = matY+r+2;
  for ( i = 0, pt1=vecX1, pt2=vecX2, pt3=vecY1, pt4=vecY2; i < r+2; i++, pt1++, pt2++, pt3++, pt4++ ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
    SNAPBASECOMPLEX_set_ui(*pt3, 0);
    SNAPBASECOMPLEX_set_ui(*pt4, 0);
  }
  SNAPBASECOMPLEX_set_ui(*(vecX1+r), 1);
  SNAPBASECOMPLEX_set_ui(*(vecX2+r+1), 1);
  SNAPBASECOMPLEX_set_ui(*(vecY1), 1);
  SNAPBASECOMPLEX_set_ui(*(vecY2+1), 1);
  // construction of vecX1 and vecX2
  for ( i = 0, pt1=vecX1+r-1, pt2=vecX1+r, pt3=vecX1+r+1, pt5=vecX2+r, pt6=vecX2+r+1, pt7=leftS1+2*(r-1), pt8=leftS2+2*(r-1); i < r; i++, pt1--, pt2--, pt3--, pt5--, pt6--, pt7-=2, pt8-=2 ) {
    APPLY_GIVENS(*pt7,*(pt7+1),*pt1,*pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt8,*(pt8+1),*pt2,*pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt8,*(pt8+1),*pt5,*pt6,*ptmp1,*ptmp2)
  }
  // construction of vecY1 and vecY2
  for ( i = 0, pt1=vecY1, pt2=vecY1+1, pt4=vecY2, pt5=vecY2+1, pt6=vecY2+2, pt7=rightS1, pt8=rightS2; i < r; i++, pt1++, pt2++, pt4++, pt5++, pt6++, pt7+=2, pt8+=2 ) {
    APPLY_GIVENS(*pt8,*(pt8+1),*pt5,*pt6,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt7,*(pt7+1),*pt4,*pt5,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt7,*(pt7+1),*pt1,*pt2,*ptmp1,*ptmp2)
  }
#ifdef SNAP_QR_DEBUG
  unsigned int j;
  printf("========== vecX1 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecX1+j)),cimag(*(vecX1+j)));
  printf("\n========== vecX2 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecX2+j)),cimag(*(vecX2+j)));
  printf("\n========== vecY1 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecY1+j)),cimag(*(vecY1+j)));
  printf("\n========== vecY2 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecY2+j)),cimag(*(vecY2+j)));
  printf("\n");
#endif // SNAP_QR_DEBUG
  // compute G_i
  SNAPBASECOMPLEX_conj(*ptmp3, *(vecX2+1));             // p3 = bar{x2}_2
  SNAPBASECOMPLEX_mul(*ptmp1, *ptmp3, *(vecY2+r));      // p1 = p3*y2_{r+1} = y2_{r+1}*bar{x2}_2
  SNAPBASECOMPLEX_conj(*ptmp2, *(vecX1+1));             // p2 = bar{x1}_2
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *(vecY2+r+1));    // p2 = p2*y2_{r+2} = y2_{r+2}*bar{x1}_2
  SNAPBASECOMPLEX_sub(*ptmp1, *ptmp1, *ptmp2);          // p1 = p1 - p2 = y2_{r+1}*bar{x2}_2 - y2{r+2}*bar{x1}_2
  SNAPBASECOMPLEX_conj(*ptmp4, *scalarS1);              // p4 = bar{alpha1}
  SNAPBASECOMPLEX_conj(*ptmp2, *(vecX1));               // p2 = bar{x1}_1
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *ptmp4);          // p2 = p2*p4 = bar{alpha1}*bar{x1}_1
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *scalarS2);       // p2 = p2*alpha2 = bar{alpha1}*alpha2*bar{x1}_1
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *(vecY2+r+1));    // p2 = p2*y2_{r+2} = bar{alpha1}*alpha2*bar{x1}_1*y2_{r+2}
  SNAPBASECOMPLEX_mul(*ptmp3, *ptmp3, *(vecY1+r));      // p3 = p3*y1_{r+1} = bar{x2}_2*y1_{r+1}
  SNAPBASECOMPLEX_sub(*ptmp2, *ptmp2, *ptmp3);          // p2 = p2 - p3 = bar{alpha1}*alpha2*bar{x1}_1*y2_{r+2} - bar{x2}_2*y1_{r+1}
#ifdef SNAP_QR_DEBUG
  printf("========== ca,cb for newG =========\n");
  printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(ptmp1)),cimag(*(ptmp1)),creal(*(ptmp2)),cimag(*(ptmp2)));
#endif // SNAP_QR_DEBUG
  if ( SNAPBASECOMPLEX_zeroq(*ptmp1) && SNAPBASECOMPLEX_zeroq(*ptmp2) ) {
    printf("Error: some implicit condition is not satisfied.\n");
    exit(EXIT_FAILURE);
  }
  SNAP_alg_qr_dgv2010refine_givens_C(ptmp1, ptmp2, g, g+1, ptmp3);
#ifdef SNAP_QR_DEBUG
  printf("========== newG =========\n");
  printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(g)),cimag(*(g)),creal(*(g+1)),cimag(*(g+1)));
#endif // SNAP_QR_DEBUG
  // copy the cosine term to aaG
  SNAPBASECOMPLEX_set(*aaG, *g);
  // compute H
  // SNAPBASECOMPLEX_conj(*ptmp4, *scalarS1);           // p4 = bar{alpha1}
  SNAPBASECOMPLEX_mul(*ptmp4, *ptmp4, *scalarS2);       // p4 = p4*alpha2 = bar{alpha1}*alpha2
  SNAPBASECOMPLEX_mul(*(aaG+1), *ptmp4, *(g+1));        // aaGs = p4*sg = bar{alpha1}*alpha2*sg
  SNAPBASECOMPLEX_mul(*ptmp1, *(aaG+1), *(vecX1+1));    // p1 = aaGs*x1_2 = bar{alpha1}*alpha2*sg*x1_2
  SNAPBASECOMPLEX_mul(*ptmp2, *g, *vecX1);              // p2 = cg*x1_1
  SNAPBASECOMPLEX_add(*ptmp1, *ptmp1, *ptmp2);          // p1 = p1 + p2 = cg*x1_1 + bar{alpha1}*alpha2*sg*x1_2
  SNAPBASECOMPLEX_conj(*ptmp2, *(vecY1+r));             // p2 = bar{y1}_{r+1}
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *g);              // p2 = p2*cg = cg*bar{y1}_{r+1}
  SNAPBASECOMPLEX_conj(*ptmp3, *(vecY2+r));             // p3 = bar{y2}_{r+1}
  SNAPBASECOMPLEX_mul(*ptmp3, *ptmp3, *(g+1));          // p3 = p3*sg = sg*bar{y2}_{r+1}
  SNAPBASECOMPLEX_add(*ptmp3, *ptmp2, *ptmp3);          // p3 = p2 + p3 = cg*bar{y1}_{r+1} + sg*bar{y2}_{r+1}
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp4, *(g+1));          // p2 = p4*sg = bar{alpha1}*alpha2*sg
  SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *(vecX2+1));      // p2 = p2*x2_2 = bar{alpha1}*alpha2*sg*x_{2,r+2}
  SNAPBASECOMPLEX_conj(*ptmp4, *(vecY2+r+1));           // p4 = bar{y2}_{r+2}
  SNAPBASECOMPLEX_mul(*ptmp4, *ptmp4, *(g+1));          // p4 = p4*sg = sg*bar{y2}_{r+2}
  SNAP_alg_qr_dgv2010refine_givens_LS( ptmp1, ptmp2, ptmp3, ptmp4, newH+1, newH, w+4);
  SNAPBASECOMPLEX_conj(*newH, *newH);
#ifdef SNAP_QR_DEBUG
  printf("========== newH =========\n");
  printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(newH)),cimag(*(newH)),creal(*(newH+1)),cimag(*(newH+1)));
#endif // SNAP_QR_DEBUG
  // refactorize X
  // ----> apply tilde{G}_i to X from the left
  APPLY_GIVENS(*aaG,*(aaG+1),*vecX1, *(vecX1+1),*ptmp1,*ptmp2)
  APPLY_GIVENS(*aaG,*(aaG+1),*vecX2, *(vecX2+1),*ptmp1,*ptmp2)
  // ----> apply H to X from the right and PRESERVE the last column of X in vecX2.
  SNAPBASECOMPLEX_conj(*ptmp4, *newH); // p4 = bar{ch}
  for ( i = 0, pt1=vecX1, pt2=vecX2; i < r+2; i++,pt1++,pt2++ ) {
    SNAPBASECOMPLEX_mul(*ptmp1, *pt1, *(newH+1)); // p1 = pt1*sh = x21*sh
    SNAPBASECOMPLEX_mul(*ptmp2, *pt2, *ptmp4);    // p2 = pt2*p4 = x22*bar{ch}
    SNAPBASECOMPLEX_add(*pt2, *ptmp1, *ptmp2);    // x22 = p1 - p2 = x21*sh-x22*bar{ch}
  }
  pt1 = vecX1;    // p1 = x1i
  pt2 = vecX1+1;  // p2 = x2i
  pt3 = vecX1+2;  // p3 = x3i
  // ----> generate the first column of X and store it into vecX1.
  SNAPBASECOMPLEX_set_ui(*pt1, 1);                            // p1 = 1 ==> x1i = 1
  SNAPBASECOMPLEX_set_ui(*pt2, 0);                            // p2 = 0 ==> x2i = 0
  SNAPBASECOMPLEX_set_ui(*pt3, 0);                            // p3 = 0 ==> x3i = 0
  APPLY_GIVENS(*leftS1,*(leftS1+1),*pt1, *pt2,*ptmp1,*ptmp2)  // (p1,p2)^t = LS11.(p1,p2)^t
  APPLY_GIVENS(*leftS2,*(leftS2+1),*pt2, *pt3,*ptmp1,*ptmp2)  // (p2,p3)^t = LS21.(p2,p3)^t
  APPLY_GIVENS(*aaG,*(aaG+1),*pt1, *pt2,*ptmp1,*ptmp2)            // (p1,p2)^t = tilde{G}_i.(p1,p2)^t
  // ----> generate lowerS and upperS, and update leftS1 and leftS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, vecX2+1, pt3, vecX2+2, qtmp3, qtmp3+1, w);  // Givens(q31,q32)^h((p2,x22),(p3,x32))=Givens(c,s)^h((x2i,x22),(p3i,x32)) --> ((x,0),(0,x))
  SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);                                                   // q21 = conj(q31)
  SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));                                            // q22 = -q32       ==> Givens(q2) = Givens(q3)^h
  SNAP_alg_qr_dgv2010refine_pullupthrough(qtmp2, aaG, leftS2, w);                         // q2.tilde{G}_i.LS21 ==> tilde{G}_i'.LS21'.q2'
  APPLY_GIVENS_GIVENS(*leftS1,*(leftS1+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
  SNAPBASECOMPLEX_mul(*(leftS2+1), *(leftS2+1), *leftS1);
  SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);                                                    // q11 = q31
  SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));                                            // q12 = q32        ==> Givens(q1) = Givens(q3)
  SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);                    // Givens(q2).(p1,p2)^t = Givens(q2).(x1i,x2i)^t
  SNAPBASECOMPLEX_conj(*leftS1, *qtmp2);                                                  // LS11 = conj(q2)
  SNAPBASECOMPLEX_neg(*(leftS1+1), *(qtmp2+1));                                           // LS12 = -q2       ==> LS1 = Givens(q2)^h
  // ----> repeat (r-2) times
  for ( i = 2, pt4=leftS1+2, pt5=leftS2+2, pt6=leftS2, pt7=vecX2+2, pt8=vecX2+3; i < r; i++, pt4+=2, pt5+=2, pt6+=2, pt7++, pt8++ ) {
    // ----> generate the first column of X and store it into vecX1.
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
    SNAPBASECOMPLEX_set_ui(*pt3, 0);
    APPLY_GIVENS(*pt4,*(pt4+1),*pt1, *pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt5,*(pt5+1),*pt2, *pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt6,*(pt6+1),*pt1, *pt2,*ptmp1,*ptmp2)
    // ----> generate lowerS and upperS, and update leftS1 and leftS2
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, pt7, pt3, pt8, qtmp3, qtmp3+1, w);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_pullupthrough(qtmp2, pt6, pt5, w);
    APPLY_GIVENS_GIVENS(*pt4,*(pt4+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
    SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt4);
    SNAPBASECOMPLEX_set(*pt6, *qtmp1);
    SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
    SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
    SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_conj(*pt4, *qtmp2);
    SNAPBASECOMPLEX_neg(*(pt4+1), *(qtmp2+1));
  }
  // ----> the last
  // ----> generate the first column of X and store it into vecX1.
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  SNAPBASECOMPLEX_set_ui(*pt3, 0);
  APPLY_GIVENS(*pt4,*(pt4+1),*pt1, *pt2,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt5,*(pt5+1),*pt2, *pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt6,*(pt6+1),*pt1, *pt2,*ptmp1,*ptmp2)
  // ----> generate lowerS and upperS, and update leftS1 and leftS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, pt7, pt3, pt8, pt5, pt5+1, w);
  SNAPBASECOMPLEX_mul(*pt2, *pt2, *pt8);          // to make the resulting pt8 being 1.
  SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt8);
  SNAPBASECOMPLEX_conj(*pt8, *pt8);
  SNAPBASECOMPLEX_mul(*pt5, *pt5, *pt8);
  SNAPBASECOMPLEX_set(*pt6, *qtmp1);
  SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
  SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);
  SNAPBASECOMPLEX_conj(*pt4, *qtmp2);
  SNAPBASECOMPLEX_neg(*(pt4+1), *(qtmp2+1));
  // refactorize Y
  // ----> G_{i} ==> G_{i}^h
  SNAPBASECOMPLEX_conj(*aaG, *g);
  SNAPBASECOMPLEX_neg(*(aaG+1), *(g+1));
  // ----> apply H^h to Y from the left and G^h to Y from the right and PRESERVE the first column of Y in vecY1.
  SNAPBASECOMPLEX_conj(*ptmp3, *(aaG+1));
  for ( i = 0, pt1=vecY1, pt2=vecY2, pt3=aaG; i < r+2; i++,pt1++,pt2++ ) {
    SNAPBASECOMPLEX_mul(*ptmp1, *pt1, *pt3);
    SNAPBASECOMPLEX_mul(*ptmp2, *pt2, *ptmp3);
    SNAPBASECOMPLEX_sub(*pt1, *ptmp1, *ptmp2);
  }
  // now newH denotes H^h
  SNAPBASECOMPLEX_conj(*newH, *newH);
  SNAPBASECOMPLEX_neg(*(newH+1), *(newH+1));
  APPLY_GIVENS(*newH, *(newH+1), *(vecY1+r), *(vecY1+r+1), *ptmp1, *ptmp2)
  pt1 = vecY2;
  pt2 = vecY2+1;
  pt3 = vecY2+2;
// ----> generate the last column of Y and store it into vecY2.
  SNAPBASECOMPLEX_set_ui(*pt1, 0);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  SNAPBASECOMPLEX_set_ui(*pt3, 1);
  APPLY_GIVENS(*(rightS2+2*(r-1)),*(rightS2+2*(r-1)+1),*pt2, *pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*(rightS1+2*(r-1)),*(rightS1+2*(r-1)+1),*pt1, *pt2,*ptmp1,*ptmp2)
  APPLY_GIVENS(*newH,*(newH+1),*pt2, *pt3,*ptmp1,*ptmp2)
  // ----> generate lowerS and upperS, and update rightS1 and rightS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(vecY1+r-1, pt1, vecY1+r, pt2, qtmp3, qtmp3+1, w);
  SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
  SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
  SNAP_alg_qr_dgv2010refine_pulldownthrough(qtmp2, newH, rightS1+2*(r-1), w);
  APPLY_GIVENS_GIVENS(*(rightS2+2*(r-1)),*(rightS2+2*(r-1)+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
  SNAPBASECOMPLEX_mul(*(rightS1+2*(r-1)+1), *(rightS1+2*(r-1)+1), *(rightS2+2*(r-1)));
  SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
  SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
  SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
  SNAPBASECOMPLEX_set(*(rightS2+2*(r-1)), *qtmp2);
  SNAPBASECOMPLEX_conj(*qtmp2, *qtmp2);
  SNAPBASECOMPLEX_conj(*(qtmp2+1), *(qtmp2+1));
  SNAPBASECOMPLEX_set(*(rightS2+2*(r-1)+1), *(qtmp2+1));
  SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp2+1));
  // ----> repeat (r-2) times
  for ( i = 2, pt4=rightS2+2*(r-2), pt5=rightS1+2*(r-2), pt6=rightS1+2*(r-1), pt7=vecY1+r-2, pt8=vecY1+r-1; i < r; i++, pt4-=2, pt5-=2, pt6-=2, pt7--, pt8-- ) {
    // ----> generate the last column of Y and store it into vecY2.
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
    SNAPBASECOMPLEX_set_ui(*pt3, 1);
    APPLY_GIVENS(*pt4,*(pt4+1),*pt2, *pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt5,*(pt5+1),*pt1, *pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt6,*(pt6+1),*pt2, *pt3,*ptmp1,*ptmp2)
    // ----> generate lowerS and upperS, and update rightS1 and rightS2
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt7, pt1, pt8, pt2, qtmp3, qtmp3+1, w);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_pulldownthrough(qtmp2, pt6, pt5, w);
    APPLY_GIVENS_GIVENS(*pt4,*(pt4+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
    SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt4);
    SNAPBASECOMPLEX_set(*pt6, *qtmp1);
    SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
    SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
    SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_set(*pt4, *qtmp2);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp2);
    SNAPBASECOMPLEX_conj(*(qtmp2+1), *(qtmp2+1));
    SNAPBASECOMPLEX_set(*(pt4+1), *(qtmp2+1));
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp2+1));
  }
  // ----> the last
  // ----> generate the last column of Y and store it into vecY2.
  SNAPBASECOMPLEX_set_ui(*pt1, 0);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  SNAPBASECOMPLEX_set_ui(*pt3, 1);
  APPLY_GIVENS(*pt4,*(pt4+1),*pt2, *pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt5,*(pt5+1),*pt1, *pt2,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt6,*(pt6+1),*pt2, *pt3,*ptmp1,*ptmp2)
  // ----> generate lowerS and upperS, and update rightS1 and rightS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt7, pt1, pt8, pt2, pt5, pt5+1, w);
  SNAPBASECOMPLEX_mul(*pt5, *pt5, *pt7);          // to make the resulting vecY1+r-2 being 1.
  SNAPBASECOMPLEX_mul(*pt2, *pt2, *pt7);
  SNAPBASECOMPLEX_conj(*pt7, *pt7);
  SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt7);
  SNAPBASECOMPLEX_set(*pt6, *qtmp1);
  SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
  SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
  SNAPBASECOMPLEX_set(*pt4, *qtmp2);
  SNAPBASECOMPLEX_conj(*(pt4+1), *(qtmp2+1));
#ifdef SNAP_QR_DEBUG
  printf("========== leftS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r)),cimag(*(leftS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r+1)),cimag(*(leftS1+i*r+1)));
  }
  printf("\n========== leftS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r)),cimag(*(leftS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r+1)),cimag(*(leftS2+i*r+1)));
  }
  printf("\n========== rightS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r)),cimag(*(rightS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r+1)),cimag(*(rightS1+i*r+1)));
  }
  printf("\n========== rightS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r)),cimag(*(rightS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r+1)),cimag(*(rightS2+i*r+1)));
  }
  printf("\n");
#endif // SNAP_QR_DEBUG
}

/*!
 *  @brief This chases the given Givens rotations through the given unitary structure to the right.
 *  @param[in] r r is the displacement rank.
 *  @param[in,out] g g is the Givens rotation will be pulled through, and the resulting rotation on exit.
 *  @param[in,out] leftS1 leftS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] leftS2 leftS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] rightS1 rightS1 is the upper list of Givens rotations, and modified by pull-through on exit.
 *  @param[in,out] rightS2 rightS2 is the lower list of Givens rotations, and modified by pull-through on exit.
 *  @param[in] matX matX is the temporal work area which must be an array of length (r+2)x2 at least.
 *  @param[in] matY matY is the temporal work area which must be an array of length (r+1)x2 at least.
 *  @param[in] w w is the temporal work area which must be an array of length 21=3x3+2+2x5 at least.
 */
void SNAP_alg_qr_dgv2010refine_chasing(const unsigned int r, SNAPBASECOMPLEX_t *g, SNAPBASECOMPLEX_t *leftS1, SNAPBASECOMPLEX_t *leftS2, SNAPBASECOMPLEX_t *rightS1, SNAPBASECOMPLEX_t *rightS2, SNAPBASECOMPLEX_t *matX, SNAPBASECOMPLEX_t *matY, SNAPBASECOMPLEX_t *w) {
  SNAPBASECOMPLEX_t *vecX1, *vecX2, *vecY1, *vecY2, *pt1, *pt2, *pt3, *pt4, *pt5, *pt6, *pt7, *pt8, *ptmp1, *ptmp2, *ptmp3, *ptmp4, *newg, *newH, *qtmp1, *qtmp2, *qtmp3;
  unsigned int i;
  ptmp1 = w;
  ptmp2 = w+1;
  ptmp3 = w+2;
  ptmp4 = w+3;
  newg = w + 11;
  newH = newg + 2;
  qtmp1 = newH + 2;
  qtmp2 = qtmp1 + 2;
  qtmp3 = qtmp2 + 2;
#ifdef SNAP_QR_DEBUG
  printf("========== leftS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r)),cimag(*(leftS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r+1)),cimag(*(leftS1+i*r+1)));
  }
  printf("\n========== leftS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r)),cimag(*(leftS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r+1)),cimag(*(leftS2+i*r+1)));
  }
  printf("\n========== rightS1 =========\n");
  for ( i = 0; i < r-1; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r)),cimag(*(rightS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r+1)),cimag(*(rightS1+i*r+1)));
  }
  printf("\n========== rightS2 =========\n");
  for ( i = 0; i < r-1; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r)),cimag(*(rightS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r+1)),cimag(*(rightS2+i*r+1)));
  }
  printf("\n");
#endif // SNAP_QR_DEBUG
// initializations for vecX1, vecX2, vecY1, vecY2
  vecX1 = matX;
  vecX2 = matX+r+2;
  vecY1 = matY;
  vecY2 = matY+r+1;
  for ( i = 0, pt1=vecX1, pt2=vecX2; i < r+2; i++, pt1++, pt2++ ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
  }
  for ( i = 0, pt1=vecY1, pt2=vecY2; i < r+1; i++, pt1++, pt2++ ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
  }
  SNAPBASECOMPLEX_set_ui(*(vecX1+r), 1);
  SNAPBASECOMPLEX_set_ui(*(vecX2+r+1), 1);
  SNAPBASECOMPLEX_set_ui(*(vecY1), 1);
  SNAPBASECOMPLEX_set_ui(*(vecY2+1), 1);
// construction of vecX1 and vecX2
  for ( i = 0, pt1=vecX1+r-1, pt2=vecX1+r, pt3=vecX1+r+1, pt5=vecX2+r, pt6=vecX2+r+1, pt7=leftS1+2*(r-1), pt8=leftS2+2*(r-1); i < r; i++, pt1--, pt2--, pt3--, pt5--, pt6--, pt7-=2, pt8-=2 ) {
    APPLY_GIVENS(*pt7,*(pt7+1),*pt1,*pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt8,*(pt8+1),*pt2,*pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt8,*(pt8+1),*pt5,*pt6,*ptmp1,*ptmp2)
  }
  APPLY_GIVENS(*g,*(g+1),*pt2,*pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*g,*(g+1),*pt5,*pt6,*ptmp1,*ptmp2)
// construction of vecY1 and vecY2
  for ( i = 1, pt1=vecY1, pt2=vecY1+1, pt4=vecY2, pt5=vecY2+1, pt6=vecY2+2, pt7=rightS1, pt8=rightS2; i < r; i++, pt1++, pt2++, pt4++, pt5++, pt6++, pt7+=2, pt8+=2 ) {
    APPLY_GIVENS(*pt8,*(pt8+1),*pt5,*pt6,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt7,*(pt7+1),*pt4,*pt5,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt7,*(pt7+1),*pt1,*pt2,*ptmp1,*ptmp2)
  }
#ifdef SNAP_QR_DEBUG
  unsigned int j;
  printf("========== vecX1 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecX1+j)),cimag(*(vecX1+j)));
  printf("\n========== vecX2 =========\n");
  for ( j=0; j<r+2; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecX2+j)),cimag(*(vecX2+j)));
  printf("\n========== vecY1 =========\n");
  for ( j=0; j<r+1; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecY1+j)),cimag(*(vecY1+j)));
  printf("\n========== vecY2 =========\n");
  for ( j=0; j<r+1; j++ )
    printf("(%+.18e%+.18eI),",creal(*(vecY2+j)),cimag(*(vecY2+j)));
  printf("\n");
#endif // SNAP_QR_DEBUG
  if ( SNAPBASECOMPLEX_zeroq(*(vecY2+r)) ) {
    printf("Error: some implicit condition is not satisfied.\n");
    exit(EXIT_FAILURE);
  }
// compute G_{i+1} and H (NOTE: override g is postponed)
  if ( SNAPBASECOMPLEX_zeroq(*vecX2) ) {
    SNAPBASECOMPLEX_set_ui(*newg, 1);
    SNAPBASECOMPLEX_set_ui(*(newg+1), 0);
    SNAPBASECOMPLEX_set_ui(*newH, 1);
    SNAPBASECOMPLEX_set_ui(*(newH+1), 0);
  } else if ( SNAPBASECOMPLEX_zeroq(*vecX1) ) {
    SNAPBASECOMPLEX_set(*newg, *(vecY1+r-1));           // g1 = y11
    SNAPBASECOMPLEX_set(*(newg+1), *(vecY2+r-1));       // g2 = y12
    SNAP_alg_qr_dgv2010refine_givens_C(newg, newg+1, ptmp1, ptmp2, ptmp3);
    SNAPBASECOMPLEX_neg(*newg, *ptmp2);                 // x1 = -s
    SNAPBASECOMPLEX_set(*(newg+1), *ptmp1);             // x2 = c
    SNAPBASECOMPLEX_set_ui(*newH, 0);
    SNAPBASECOMPLEX_set_ui(*(newH+1), 1);
  } else {
    SNAPBASECOMPLEX_conj(*ptmp3, *vecX2);               // p3 = bar{x12}
    SNAPBASECOMPLEX_mul(*ptmp1, *ptmp3, *(vecY2+r-1));  // p1 = p3*y12 = y12*bar{x12}
    SNAPBASECOMPLEX_conj(*ptmp2, *vecX1);               // p2 = bar{x11}
    SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *(vecY2+r));    // p2 = p2*y22 = y22*bar{x11}
    SNAPBASECOMPLEX_sub(*ptmp1, *ptmp2, *ptmp1);        // p1 = p2 - p1 = y22*bar{x11} - y12*bar{x12}
    SNAPBASECOMPLEX_neg(*ptmp3, *ptmp3);                // p3 = -p3 = -bar{x12}
    SNAPBASECOMPLEX_mul(*ptmp2, *ptmp3, *(vecY1+r-1));  // p2 = p3*y11 = -y11*bar{x12}
#ifdef SNAP_QR_DEBUG
    printf("========== GCS =========\n");
    printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(ptmp2)),cimag(*(ptmp2)),creal(*(ptmp1)),cimag(*(ptmp1)));
#endif // SNAP_QR_DEBUG
    SNAP_alg_qr_dgv2010refine_givens_C(ptmp2, ptmp1, newg, ptmp4, ptmp3);
    SNAPBASECOMPLEX_set(*(newg+1), *newg);              // x2 = c
    SNAPBASECOMPLEX_neg(*newg, *ptmp4);                 // x1 = -s

    SNAPBASECOMPLEX_conj(*ptmp1, *(vecY1+r-1));         // p1 = bar{y11}
    SNAPBASECOMPLEX_mul(*ptmp1, *ptmp1, *newg);         // p1 = p1*cg = cg*bar{y11}
    SNAPBASECOMPLEX_conj(*ptmp2, *(vecY2+r-1));         // p2 = bar{y12}
    SNAPBASECOMPLEX_mul(*ptmp2, *ptmp2, *(newg+1));     // p2 = p2*sg = sg*bar{y12}
    SNAPBASECOMPLEX_add(*ptmp4, *ptmp1, *ptmp2);        // p4 = p1 + p2 = cg*bar{y11} + sg*bar{y12}
    SNAPBASECOMPLEX_neg(*ptmp4, *ptmp4);                // p4 = -p4 = - cg*bar{y11} - sg*bar{y12}
    SNAPBASECOMPLEX_conj(*ptmp3, *(vecY2+r));           // p3 = bar{y22}
    SNAPBASECOMPLEX_mul(*ptmp3, *ptmp3, *(newg+1));     // p3 = p3*sg = sg*bar{y22}
    SNAPBASECOMPLEX_neg(*ptmp2, *vecX1);                // p2 = -x11
    SNAPBASECOMPLEX_set(*ptmp1, *vecX2);                // p1 = x12
#ifdef SNAP_QR_DEBUG
    printf("========== LS =========\n");
    printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(ptmp1)),cimag(*(ptmp1)),creal(*(ptmp2)),cimag(*(ptmp2)));
    printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(ptmp3)),cimag(*(ptmp3)),creal(*(ptmp4)),cimag(*(ptmp4)));
#endif // SNAP_QR_DEBUG
    SNAP_alg_qr_dgv2010refine_givens_LS( ptmp1, ptmp2, ptmp3, ptmp4, newH, newH+1, w+4);
  }
#ifdef SNAP_QR_DEBUG
  printf("========== newG =========\n");
  printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(newg)),cimag(*(newg)),creal(*(newg+1)),cimag(*(newg+1)));
  printf("========== newH =========\n");
  printf("(%+.18e%+.18eI), (%+.18e%+.18eI)\n",creal(*(newH)),cimag(*(newH)),creal(*(newH+1)),cimag(*(newH+1)));
#endif // SNAP_QR_DEBUG
// refactorize X
// ----> apply H^h to X from the right and PRESERVE the last column of X in vecX2.
  for ( i = 1, pt1=vecX1+1, pt2=vecX2+1, pt3=newH+1, pt4=newH; i < r+2; i++,pt1++,pt2++ ) {
    SNAPBASECOMPLEX_mul(*ptmp1, *pt1, *pt3);    // t1 = p1*p3 = x21*hs
    SNAPBASECOMPLEX_mul(*ptmp2, *pt2, *pt4);    // t2 = p2*p4 = x22*hc
    SNAPBASECOMPLEX_sub(*pt2, *ptmp2, *ptmp1);  // x22 = t2-t1 = x22*hc-x21*hs
  }
  pt1 = vecX1;    // p1 = x1i
  pt2 = vecX1+1;  // p2 = x2i
  pt3 = vecX1+2;  // p3 = x3i
// ----> generate the first column of X and store it into vecX1.
  SNAPBASECOMPLEX_set_ui(*pt1, 1);                            // p1 = 1 ==> x1i = 1
  SNAPBASECOMPLEX_set_ui(*pt2, 0);                            // p2 = 0 ==> x2i = 0
  SNAPBASECOMPLEX_set_ui(*pt3, 0);                            // p3 = 0 ==> x3i = 0
  APPLY_GIVENS(*leftS1,*(leftS1+1),*pt1, *pt2,*ptmp1,*ptmp2)  // (p1,p2)^t = LS11.(p1,p2)^t
  APPLY_GIVENS(*leftS2,*(leftS2+1),*pt2, *pt3,*ptmp1,*ptmp2)  // (p2,p3)^t = LS21.(p2,p3)^t
  APPLY_GIVENS(*g,*(g+1),*pt1, *pt2,*ptmp1,*ptmp2)            // (p1,p2)^t = Gi.(p1,p2)^t
// ----> generate lowerS and upperS, and update leftS1 and leftS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, vecX2+1, pt3, vecX2+2, qtmp3, qtmp3+1, w);  // Givens(q31,q32)^h((p2,x22),(p3,x32))=Givens(c,s)^h((x2i,x22),(p3i,x32)) --> ((x,0),(0,x))
  SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);                                                   // q21 = conj(q31)
  SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));                                            // q22 = -q32       ==> Givens(q2) = Givens(q3)^h
  SNAP_alg_qr_dgv2010refine_pullupthrough(qtmp2, g, leftS2, w);                           // q2.Gi.LS21 ==> Gi'.LS21'.q2'
  APPLY_GIVENS_GIVENS(*leftS1,*(leftS1+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
  SNAPBASECOMPLEX_mul(*(leftS2+1), *(leftS2+1), *leftS1);
  SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);                                                    // q11 = q31
  SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));                                            // q12 = q32        ==> Givens(q1) = Givens(q3)
  SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);                    // Givens(q2).(p1,p2)^t = Givens(q2).(x1i,x2i)^t
  SNAPBASECOMPLEX_conj(*leftS1, *qtmp2);                                                  // LS11 = conj(q2)
  SNAPBASECOMPLEX_neg(*(leftS1+1), *(qtmp2+1));                                           // LS12 = -q2       ==> LS1 = Givens(q2)^h
// ----> repeat (r-2) times
  for ( i = 2, pt4=leftS1+2, pt5=leftS2+2, pt6=leftS2, pt7=vecX2+2, pt8=vecX2+3; i < r; i++, pt4+=2, pt5+=2, pt6+=2, pt7++, pt8++ ) {
    // ----> generate the first column of X and store it into vecX1.
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
    SNAPBASECOMPLEX_set_ui(*pt3, 0);
    APPLY_GIVENS(*pt4,*(pt4+1),*pt1, *pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt5,*(pt5+1),*pt2, *pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt6,*(pt6+1),*pt1, *pt2,*ptmp1,*ptmp2)
    // ----> generate lowerS and upperS, and update leftS1 and leftS2
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, pt7, pt3, pt8, qtmp3, qtmp3+1, w);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_pullupthrough(qtmp2, pt6, pt5, w);
    APPLY_GIVENS_GIVENS(*pt4,*(pt4+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
    SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt4);
    SNAPBASECOMPLEX_set(*pt6, *qtmp1);
    SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
    SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
    SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_conj(*pt4, *qtmp2);
    SNAPBASECOMPLEX_neg(*(pt4+1), *(qtmp2+1));
  }
// ----> the last
// ----> generate the first column of X and store it into vecX1.
  SNAPBASECOMPLEX_set_ui(*pt1, 1);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  SNAPBASECOMPLEX_set_ui(*pt3, 0);
  APPLY_GIVENS(*pt4,*(pt4+1),*pt1, *pt2,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt5,*(pt5+1),*pt2, *pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*pt6,*(pt6+1),*pt1, *pt2,*ptmp1,*ptmp2)
// ----> generate lowerS and upperS, and update leftS1 and leftS2
  SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt2, pt7, pt3, pt8, pt5, pt5+1, w);
  SNAPBASECOMPLEX_mul(*pt2, *pt2, *pt8);          // to make the resulting pt8 being 1.
  SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt8);
  SNAPBASECOMPLEX_conj(*pt8, *pt8);
  SNAPBASECOMPLEX_mul(*pt5, *pt5, *pt8);
  SNAPBASECOMPLEX_set(*pt6, *qtmp1);
  SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
  SNAP_alg_qr_dgv2010refine_givens_R(pt1, pt2, qtmp2, qtmp2+1, ptmp3);
  SNAPBASECOMPLEX_conj(*pt4, *qtmp2);
  SNAPBASECOMPLEX_neg(*(pt4+1), *(qtmp2+1));
// override g by G_{i+1}
  SNAPBASECOMPLEX_set(*g, *newg);
  SNAPBASECOMPLEX_set(*(g+1), *(newg+1));
// refactorize Y
// ----> G_{i+1} ==> G_{i+1}^h
  SNAPBASECOMPLEX_conj(*newg, *newg);
  SNAPBASECOMPLEX_neg(*(newg+1), *(newg+1));
// ----> apply H to Y from the left and G^h to Y from the right and PRESERVE the first column of Y in vecY1.
  SNAPBASECOMPLEX_conj(*ptmp3, *(newg+1));
  for ( i = 0, pt1=vecY1, pt2=vecY2, pt3=newg; i < r+1; i++,pt1++,pt2++ ) {
    SNAPBASECOMPLEX_mul(*ptmp1, *pt1, *pt3);
    SNAPBASECOMPLEX_mul(*ptmp2, *pt2, *ptmp3);
    SNAPBASECOMPLEX_sub(*pt1, *ptmp1, *ptmp2);
  }
  APPLY_GIVENS(*newH, *(newH+1), *(vecY1+r-1), *(vecY1+r), *ptmp1, *ptmp2)
  pt1 = vecY2;
  pt2 = vecY2+1;
  pt3 = vecY2+2;
// ----> generate the last column of Y and store it into vecY2.
  SNAPBASECOMPLEX_set_ui(*pt1, 0);
  SNAPBASECOMPLEX_set_ui(*pt2, 0);
  SNAPBASECOMPLEX_set_ui(*pt3, 1);
  APPLY_GIVENS(*(rightS2+2*(r-2)),*(rightS2+2*(r-2)+1),*pt2, *pt3,*ptmp1,*ptmp2)
  APPLY_GIVENS(*(rightS1+2*(r-2)),*(rightS1+2*(r-2)+1),*pt1, *pt2,*ptmp1,*ptmp2)
  APPLY_GIVENS(*newH,*(newH+1),*pt2, *pt3,*ptmp1,*ptmp2)
  if ( r == 2 ) {
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(vecY1+r-2, pt1, vecY1+r-1, pt2, rightS1, rightS1+1, w);
    SNAPBASECOMPLEX_mul(*rightS1, *rightS1, *(vecY1+r-2));          // to make the resulting vecY1+r-2 being 1.
    SNAPBASECOMPLEX_mul(*pt2, *pt2, *(vecY1+r-2));
    SNAPBASECOMPLEX_conj(*(vecY1+r-2), *(vecY1+r-2));
    SNAPBASECOMPLEX_mul(*(rightS1+1), *(rightS1+1), *(vecY1+r-2));
    SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_set(*rightS2, *qtmp2);
    SNAPBASECOMPLEX_conj(*(rightS2+1), *(qtmp2+1));
  } else {
    // ----> generate lowerS and upperS, and update rightS1 and rightS2
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(vecY1+r-2, pt1, vecY1+r-1, pt2, qtmp3, qtmp3+1, w);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_pulldownthrough(qtmp2, newH, rightS1+2*(r-2), w);
    APPLY_GIVENS_GIVENS(*(rightS2+2*(r-2)),*(rightS2+2*(r-2)+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
    SNAPBASECOMPLEX_mul(*(rightS1+2*(r-2)+1), *(rightS1+2*(r-2)+1), *(rightS2+2*(r-2)));
    SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
    SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
    SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_set(*(rightS2+2*(r-2)), *qtmp2);
    SNAPBASECOMPLEX_conj(*qtmp2, *qtmp2);
    SNAPBASECOMPLEX_conj(*(qtmp2+1), *(qtmp2+1));
    SNAPBASECOMPLEX_set(*(rightS2+2*(r-2)+1), *(qtmp2+1));
    SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp2+1));
    // ----> repeat (r-3) times
    for ( i = 3, pt4=rightS2+2*(r-3), pt5=rightS1+2*(r-3), pt6=rightS1+2*(r-2), pt7=vecY1+r-3, pt8=vecY1+r-2; i < r; i++, pt4-=2, pt5-=2, pt6-=2, pt7--, pt8-- ) {
      // ----> generate the last column of Y and store it into vecY2.
      SNAPBASECOMPLEX_set_ui(*pt1, 0);
      SNAPBASECOMPLEX_set_ui(*pt2, 0);
      SNAPBASECOMPLEX_set_ui(*pt3, 1);
      APPLY_GIVENS(*pt4,*(pt4+1),*pt2, *pt3,*ptmp1,*ptmp2)
      APPLY_GIVENS(*pt5,*(pt5+1),*pt1, *pt2,*ptmp1,*ptmp2)
      APPLY_GIVENS(*pt6,*(pt6+1),*pt2, *pt3,*ptmp1,*ptmp2)
      // ----> generate lowerS and upperS, and update rightS1 and rightS2
      SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt7, pt1, pt8, pt2, qtmp3, qtmp3+1, w);
      SNAPBASECOMPLEX_conj(*qtmp2, *qtmp3);
      SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp3+1));
      SNAP_alg_qr_dgv2010refine_pulldownthrough(qtmp2, pt6, pt5, w);
      APPLY_GIVENS_GIVENS(*pt4,*(pt4+1),*qtmp2,*(qtmp2+1),*ptmp1,*ptmp2)
      SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt4);
      SNAPBASECOMPLEX_set(*pt6, *qtmp1);
      SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
      SNAPBASECOMPLEX_set(*qtmp1, *qtmp3);
      SNAPBASECOMPLEX_set(*(qtmp1+1), *(qtmp3+1));
      SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
      SNAPBASECOMPLEX_set(*pt4, *qtmp2);
      SNAPBASECOMPLEX_conj(*qtmp2, *qtmp2);
      SNAPBASECOMPLEX_conj(*(qtmp2+1), *(qtmp2+1));
      SNAPBASECOMPLEX_set(*(pt4+1), *(qtmp2+1));
      SNAPBASECOMPLEX_neg(*(qtmp2+1), *(qtmp2+1));
    }
    // ----> the last
    // ----> generate the last column of Y and store it into vecY2.
    SNAPBASECOMPLEX_set_ui(*pt1, 0);
    SNAPBASECOMPLEX_set_ui(*pt2, 0);
    SNAPBASECOMPLEX_set_ui(*pt3, 1);
    APPLY_GIVENS(*pt4,*(pt4+1),*pt2, *pt3,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt5,*(pt5+1),*pt1, *pt2,*ptmp1,*ptmp2)
    APPLY_GIVENS(*pt6,*(pt6+1),*pt2, *pt3,*ptmp1,*ptmp2)
    // ----> generate lowerS and upperS, and update rightS1 and rightS2
    SNAP_alg_qr_dgv2010refine_givens_LS4RF(pt7, pt1, pt8, pt2, pt5, pt5+1, w);
    SNAPBASECOMPLEX_mul(*pt5, *pt5, *pt7);          // to make the resulting vecY1+r-2 being 1.
    SNAPBASECOMPLEX_mul(*pt2, *pt2, *pt7);
    SNAPBASECOMPLEX_conj(*pt7, *pt7);
    SNAPBASECOMPLEX_mul(*(pt5+1), *(pt5+1), *pt7);
    SNAPBASECOMPLEX_set(*pt6, *qtmp1);
    SNAPBASECOMPLEX_set(*(pt6+1), *(qtmp1+1));
    SNAP_alg_qr_dgv2010refine_givens_R(pt3, pt2, qtmp2, qtmp2+1, ptmp3);
    SNAPBASECOMPLEX_set(*pt4, *qtmp2);
    SNAPBASECOMPLEX_conj(*(pt4+1), *(qtmp2+1));
  }
#ifdef SNAP_QR_DEBUG
  printf("========== leftS1 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r)),cimag(*(leftS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS1+i*r+1)),cimag(*(leftS1+i*r+1)));
  }
  printf("\n========== leftS2 =========\n");
  for ( i = 0; i < r; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r)),cimag(*(leftS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(leftS2+i*r+1)),cimag(*(leftS2+i*r+1)));
  }
  printf("\n========== rightS1 =========\n");
  for ( i = 0; i < r-1; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r)),cimag(*(rightS1+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS1+i*r+1)),cimag(*(rightS1+i*r+1)));
  }
  printf("\n========== rightS2 =========\n");
  for ( i = 0; i < r-1; i++ ) {
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r)),cimag(*(rightS2+i*r)));
    printf("(%+.18e%+.18eI),",creal(*(rightS2+i*r+1)),cimag(*(rightS2+i*r+1)));
  }
  printf("\n");
#endif // SNAP_QR_DEBUG
}

/*!
 *  @brief The function computes a QR factorization of a matrix, given by its generators, by the DGV2010 refine algorithm (embeded version).
 *  @warning All the matrices must be of the row major and this routine can accept n=m at least now.
 *  @param[in] n n is the row length of A.
 *  @param[in] m m is the column length of A.
 *  @param[in] r r is the displacement order of A w.r.t. Z_1 and Z_0.
 *  @param[in] lisD D must be 1d array of length n and the last two elements are complex conjugates each other.
 *  @param[in,out] genV V must be dim(n,r) array satisfying D A - A Z_0 = VH. Overwritten by temporal elements on exit.
 *  @param[in] genH H must be dim(r,m) array satisfying D A - A Z_0 = VH.
 *  @param[in,out] matQ Q must have dim(n,n) and zero entries. The unitary matrix of A=QR on exit.
 *  @param[in,out] matR R must have dim(n,m) and zero entries. The upper triangular matrix of A=QR on exit.
 */
#ifdef SNAP_QR_GENERATE_SYL
void SNAP_alg_qr_dgv2010refine_embed(const unsigned int n, const unsigned int m, const unsigned int r, SNAPMTXCOMPLEX_t lisD, SNAPMTXCOMPLEX_t genV, SNAPMTXCOMPLEX_t genH, SNAPMTXCOMPLEX_t matQ, SNAPMTXCOMPLEX_t matR, SNAPMTXCOMPLEX_t matSYL) {
#else
void SNAP_alg_qr_dgv2010refine_embed(const unsigned int n, const unsigned int m, const unsigned int r, SNAPMTXCOMPLEX_t lisD, SNAPMTXCOMPLEX_t genV, SNAPMTXCOMPLEX_t genH, SNAPMTXCOMPLEX_t matQ, SNAPMTXCOMPLEX_t matR) {
#endif // SNAP_QR_GENERATE_SYL
//  SNAPMTXCOMPLEX_t leftS, rightS, scalarS, givensG, matW, matTMP;
  SNAPMTXCOMPLEX_t leftS, rightS, scalarS, givensG, matX, matY, matTMP;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3, *pt4, *ctmp1, *ctmp2, *ctmp3, *ctmp4, *ptmp;
  unsigned int i, j, k, pcv1, pcv2, pcv3, pcv4, spi4w;
#ifdef SNAP_QR_DEBUG
  unsigned int q1,q2;
  SNAPMTXCOMPLEX_t matI;
#endif // SNAP_QR_DEBUG
  if ( n != m ) {
    printf("Error: n and m shoule be the same integer.\n");
    exit(EXIT_FAILURE);
  }
  // allocate memories. NOTE: all the elements are now 0.
  SNAPMTXCOMPLEX_init2(leftS, n-1, 2*r);                      // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
  SNAPMTXCOMPLEX_init2(rightS, n-r-1, 2*r);                   // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
  SNAPMTXCOMPLEX_init2(scalarS, 1, n);                        // accesor: i
  SNAPMTXCOMPLEX_init2(givensG, r, 2);                        // accesor: i,j   G(i) => (c,s)
  SNAPMTXCOMPLEX_init2(matX, r+2, 2);
  SNAPMTXCOMPLEX_init2(matY, r+2, 2);
  ptmp = (SNAPBASECOMPLEX_t *) malloc(sizeof(SNAPBASECOMPLEX_t)*4);
  ctmp1 = ptmp;
  ctmp2 = ptmp+1;
  ctmp3 = ptmp+2;
  ctmp4 = ptmp+3;
  SNAPBASECOMPLEX_init(*ctmp1);
  SNAPBASECOMPLEX_init(*ctmp2);
  SNAPBASECOMPLEX_init(*ctmp3);
  SNAPBASECOMPLEX_init(*ctmp4);
  SNAPMTXCOMPLEX_init2(matTMP, 1, 21); // for pulldownthrough and misc.
  // (2-a) initialization at the allocation site.
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(leftS); i < r*(n-1); i++, pt1+=2 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(rightS); i < r*(n-r-1); i++, pt1+=2 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(matQ); i < n; i++, pt1+=SNAPMTXCOMPLEX_csize(matQ)+1 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  // (2-b) initialization after the allocation.
  k = 0;
  MAT_W_ELEM_SET(*ctmp1, n-1, 0, genV, genH, r, pt1, pt2)
  SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-1, 0), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-1));
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(scalarS), pt2 = SNAPMTXCOMPLEX_elems(lisD); i < n; i++, pt1++, pt2++ ) {
    SNAPBASECOMPLEX_set(*pt1, *pt2);
  }
#ifdef SNAP_QR_DEBUG
  printf("========== after initialization =========\n");
//  PRINT_MAT_W PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S PRINT_SCALAR_S
  PRINT_MAT_W(*ctmp1,pt1,pt2) PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S PRINT_SCALAR_S18
#endif // SNAP_QR_DEBUG
  while (1) {
    // (3-a) Increments.
    k++;
    // (3-b) Embed the previous result into the larger one.
    // (3-b-i) Embed the previous result into the larger one.
    MAT_W_ELEM_SET(*ctmp1, n-k-1, 0, genV, genH, r, pt1, pt2)
    SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-k-1, 0), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-k-1));
    for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, 0); i < k-1; i++, pt1++ ) {
      MAT_W_ELEM_SET(*ctmp1, n-k-1, i+1, genV, genH, r, pt2, pt3)
      SNAPBASECOMPLEX_add(*ctmp1, *ctmp1, *pt1);
      SNAPBASECOMPLEX_div(*(pt1+1), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-k-1));
    }
    // (3-b-ii) Embed the previous result into the larger one.
    if ( k > 1) {
      for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, k), pt3 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, k-1); i < k+1; i++, pt1+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
        MAT_W_ELEM_SET(*ctmp1, n+i-k-1, k, genV, genH, r, pt2, pt4)
        SNAPBASECOMPLEX_add(*pt1, *ctmp1, *pt3);
      }
#ifdef SNAP_QR_DEBUG
      printf("========== before applyS =========\n");
      PRINT_MAT_R
#endif // SNAP_QR_DEBUG
      SNAP_alg_qr_dgv2010refine_apply_S(n-k-1, leftS, r, scalarS, rightS, r, matR, k);
    } else {
      MAT_W_ELEM_SET(*ctmp1, n-2, 1, genV, genH, r, pt1, pt2)
      SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-2, 1), *ctmp1, SNAPMTXCOMPLEX_elem(matR, n-2, 0));
      SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-2, 1), SNAPMTXCOMPLEX_elem(matR, n-2, 1), SNAPMTXCOMPLEX_elem(lisD, 0, n-2));

      MAT_W_ELEM_SET(*ctmp1, n-1, 1, genV, genH, r, pt1, pt2)
      SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-1, 1), *ctmp1, SNAPMTXCOMPLEX_elem(matR, n-1, 0));
      SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-1, 1), SNAPMTXCOMPLEX_elem(matR, n-1, 1), SNAPMTXCOMPLEX_elem(lisD, 0, n-1));
    }
#ifdef SNAP_QR_DEBUG
    printf("========== after embedding (k)=(%u) =========\n", k);
    PRINT_MAT_R
#endif // SNAP_QR_DEBUG
    // (4) first r-th Givens rotations.
    pcv1 = SNAPLIB_MIN(r, k);
    for ( i = 0; i < pcv1; i++ ) {
      // (4-a)
      // (givensG, r, 2) // accesor: i,j   G(i) => (c,s)
      SNAP_alg_qr_dgv2010refine_givens_C(SNAPMTXCOMPLEX_elemp(matR, n-k+i-1, i), SNAPMTXCOMPLEX_elemp(matR, n-k+i, i), SNAPMTXCOMPLEX_elemp(givensG, i, 0), SNAPMTXCOMPLEX_elemp(givensG, i, 1), ctmp1);
#ifdef SNAP_QR_DEBUG
      printf("========== Givens (i)=(%u) =========\n", i);
      printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, i, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, i, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, i, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, i, 1)));
#endif // SNAP_QR_DEBUG
      // (4-b)
      pcv2 = SNAPLIB_MIN(r, k+1);
      for ( j = i+1; j < pcv2; j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matR, n-k+i, j),*ctmp1,*ctmp2)
      }
      // (4-c)
      pcv2 = k+1;
      pcv3 = SNAPLIB_MAX(r+1,pcv2-r);
      for ( j = pcv3 - 1; j < pcv2; j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matR, n-k+i, j),*ctmp1,*ctmp2)
      }
#ifdef SNAP_QR_GENERATE_SYL
      for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matSYL, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matSYL, n-k+i, j),*ctmp1,*ctmp2)
      }
#endif // SNAP_QR_GENERATE_SYL
      // (4-d) Q:=QG^h and V:=GV
      for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i-1), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
        APPLY_GIVENS_HR(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *pt1, *pt2, *ctmp1, *ctmp2)
      }
      for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-k+i-1, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-k+i, 0); j < r; j++, pt1++, pt2++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *pt1, *pt2, *ctmp1, *ctmp2)
      }
      // (4-e) leftS(n-k,i):=leftS(n-k,i)G^h, (leftS, n-1, 2*r), accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
      APPLY_GIVENS_GIVENS_HR( SNAPMTXCOMPLEX_elem(leftS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(leftS, n-k-1, 2*i+1), SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *ctmp1, *ctmp2)
    }
#ifdef SNAP_QR_DEBUG
    printf("========== after first Givens (k)=(%u) =========\n", k);
    PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S
#ifdef SNAP_QR_GENERATE_SYL
    PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
    // (5) Pull through the first r Givens rotations into rightS.
    for ( i = 0; i < pcv1; i++ ) {
      SNAPBASECOMPLEX_set(*ctmp1, SNAPMTXCOMPLEX_elem(givensG, i, 0));
      SNAPBASECOMPLEX_set(*ctmp2, SNAPMTXCOMPLEX_elem(givensG, i, 1));
#ifdef SNAP_QR_DEBUG
      printf("========== INTO RIGHT-Givens (i)=(%u) =========\n", i);
      printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
      if ( k <= r ) {
        SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), k-i-1, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== (a)INTO RIGHT-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)), SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
      } else if ( k < 2*r ) {
        if ( i < k-r ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), r, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
          printf("========== (b)INTO RIGHT-Givens (i)=(%u) =========\n", i);
          printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), k-i-1, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
          printf("========== (c)INTO RIGHT-Givens (i)=(%u) =========\n", i);
          printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)), SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
      } else {
        SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), r, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== (d)INTO RIGHT-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
      }
    }
#ifdef SNAP_QR_DEBUG
    printf("========== after pull through into rightS (k)=(%u) =========\n", k);
    PRINT_LEFT_S PRINT_RIGHT_S
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_DEBUG
    print_rebuildS(n, r, matQ, lisD);
    print_computedS_embed(n, r, scalarS, leftS, rightS);
#endif // SNAP_QR_DEBUG
    // (6) Chasing
    if ( k > 2*r ) {
      for ( i = r; i < k - r; i++ ) {
#ifdef SNAP_QR_DEBUG
        printf("========== chasing (k)=(%u) and (leftS1,S2)=(%u,%u), (rightS1,S2)=(%u,%u) =========\n", k, n-k+i-1, n-k+i,n-r-k+i,n-r-k+i+1 );
        PRINT_SCALAR_S18 PRINT_LEFT_S18 PRINT_RIGHT_S18
#endif // SNAP_QR_DEBUG
        // (6-a and 6-b)
        SNAP_alg_qr_dgv2010refine_chasing_embed(r, SNAPMTXCOMPLEX_elemp(givensG, 0, 0), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i-1, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i, 0), SNAPMTXCOMPLEX_elems(matX), SNAPMTXCOMPLEX_elems(matY), SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== Chasing-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, 0, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, 0, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, 0, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, 0, 1)));
#endif // SNAP_QR_DEBUG
        // (6-c)
        pcv2 = k+1;
        pcv1 = SNAPLIB_MAX(r+1, pcv2-r);
        for ( j = pcv1-1; j < pcv2; j++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, 0, 0), SNAPMTXCOMPLEX_elem(givensG, 0, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matR, n-k+i, j),*ctmp1,*ctmp2)
        }
#ifdef SNAP_QR_GENERATE_SYL
        for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, 0, 0), SNAPMTXCOMPLEX_elem(givensG, 0, 1), SNAPMTXCOMPLEX_elem(matSYL, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matSYL, n-k+i, j),*ctmp1,*ctmp2)
        }
#endif // SNAP_QR_GENERATE_SYL
        // (6-d) Q:=QG^h and V:=GV
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i-1), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
          APPLY_GIVENS_HR(SNAPMTXCOMPLEX_elem(givensG, 0, 0), SNAPMTXCOMPLEX_elem(givensG, 0, 1), *pt1, *pt2, *ctmp1, *ctmp2)
        }
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-k+i-1, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-k+i, 0); j < r; j++, pt1++, pt2++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, 0, 0), SNAPMTXCOMPLEX_elem(givensG, 0, 1), *pt1, *pt2, *ctmp1, *ctmp2)
        }
#ifdef SNAP_QR_DEBUG
        printf("========== DONE chasing (k)=(%u) and (leftS1,S2)=(%u,%u), (rightS1,S2)=(%u,%u) =========\n", k, n-k+i-1, n-k+i,n-r-k+i,n-r-k+i+1 );
        PRINT_SCALAR_S18 PRINT_LEFT_S18 PRINT_RIGHT_S18
#endif // SNAP_QR_DEBUG
      }
#ifdef SNAP_QR_DEBUG
      printf("========== after chasing (k)=(%u) =========\n", k);
      PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S
#ifdef SNAP_QR_GENERATE_SYL
      PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#ifdef SNAP_QR_DEBUG
      print_rebuildS(n, r, matQ, lisD);
      print_computedS_embed(n, r, scalarS, leftS, rightS);
#endif // SNAP_QR_DEBUG
#endif // SNAP_QR_DEBUG
    }
    // (7) last r Givens rotations.
    if ( k > r ) { // we use unsigned integers hence this if statement is required.
      pcv1 = SNAPLIB_MIN(r, k-r);
      pcv2 = 1;
      pcv4 = k+1;
      for ( i = pcv1; i >= pcv2; i-- ) {
#ifdef SNAP_QR_DEBUG
        printf("========== LAST GIVENS (i)=(%u) =========\n", i);
#ifdef SNAP_QR_GENERATE_SYL
        PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
        // G:=(ctmp1, ctmp2)!!
        // (7-a)
        SNAP_alg_qr_dgv2010refine_givens_C(SNAPMTXCOMPLEX_elemp(matR, n-i-1, k-i), SNAPMTXCOMPLEX_elemp(matR, n-i, k-i), ctmp1, ctmp2, ctmp3);
#ifdef SNAP_QR_DEBUG
        printf("========== Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        // (7-b)
        pcv3 = SNAPLIB_MAX(2*r+1-i, pcv4-i+1);
        for ( j = pcv3-1; j < pcv4; j++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, SNAPMTXCOMPLEX_elem(matR, n-i-1, j), SNAPMTXCOMPLEX_elem(matR, n-i, j),*ctmp3,*ctmp4)
        }
#ifdef SNAP_QR_GENERATE_SYL
        for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, SNAPMTXCOMPLEX_elem(matSYL, n-i-1, j), SNAPMTXCOMPLEX_elem(matSYL, n-i, j),*ctmp3,*ctmp4)
        }
#endif // SNAP_QR_GENERATE_SYL
        // (7-c) Q:=QG^h and V:=GV
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-i-1), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-i); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
          APPLY_GIVENS_HR(*ctmp1, *ctmp2, *pt1, *pt2, *ctmp3, *ctmp4)
        }
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-i-1, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-i, 0); j < r; j++, pt1++, pt2++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, *pt1, *pt2, *ctmp3, *ctmp4)
        }
        // (7-d)
        // backup the Givens rotation
        SNAPBASECOMPLEX_set(SNAPMTXCOMPLEX_elem(givensG, 0, 0), *ctmp1);
        SNAPBASECOMPLEX_set(SNAPMTXCOMPLEX_elem(givensG, 0, 1), *ctmp2);
        if ( i > 1 ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i), SNAPMTXCOMPLEX_elemp(leftS, n-i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i-1), SNAPMTXCOMPLEX_elemp(leftS, n-i-1, 0), i-1, SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          SNAP_alg_qr_dgv2010refine_apply_scalar(ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i-1), SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
        // (7-e)
        // restore the Givens rotation
        SNAPBASECOMPLEX_conj(*ctmp1, SNAPMTXCOMPLEX_elem(givensG, 0, 0));
        SNAPBASECOMPLEX_neg(*ctmp2, SNAPMTXCOMPLEX_elem(givensG, 0, 1));
        if ( i > 1 ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughleft(r, ptmp, SNAPMTXCOMPLEX_elemp(rightS, n-r-i-1, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-i, 0), i-1, SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS_R(SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          APPLY_GIVENS_GIVENS_R(SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
      }
#ifdef SNAP_QR_DEBUG
      printf("========== after last r Givens rotations (k)=(%u) =========\n", k);
      PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S
#ifdef SNAP_QR_GENERATE_SYL
      PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
    }
#ifdef SNAP_QR_DEBUG
    print_rebuildS(n, r, matQ, lisD);
    print_computedS_embed(n, r, scalarS, leftS, rightS);
#ifdef SNAP_QR_GENERATE_SYL
    PRINT_MAT_SYL_trinorm(*ctmp1,k)
#endif // SNAP_QR_GENERATE_SYL
    SNAPMTXCOMPLEX_init2(matI, n, n);
    printf("norm(I-matI=I-Q^hQ)_F: ");
    *ctmp1 = 0.0;
    for(q1=0; q1<sdmdc_rows(matI); q1++) {
      for(q2=0; q2<sdmdc_cols(matI); q2++) {
        SNAPBASECOMPLEX_set_d(SNAPMTXCOMPLEX_elem(matI,q1,q1),0);
        for ( i = 0; i<sdmdc_rows(matQ); i++ ) {
          sdmdc_elem(matI,q1,q2)+=conj(sdmdc_elem(matQ,i,q1))*sdmdc_elem(matQ,i,q2);
        }
        if ( q1 == q2 ) {
          sdmdc_elem(matI,q1,q2)-=1.0;
        }
        *ctmp1 += cabs(sdmdc_elem(matI,q1,q2))*cabs(sdmdc_elem(matI,q1,q2));
      }
    }
    printf("%e\n", sqrt(*ctmp1));
    SNAPMTXCOMPLEX_clear(matI);
#endif // SNAP_QR_DEBUG
    // (8) loop or not
    if ( k+1 >= n ) {
      break;
    }
  }
  // (9) reconstruct matR
  pcv1 = n-r;
  for ( j = r; j < pcv1; j++ ) {
    for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, 0, j), pt3 = SNAPMTXCOMPLEX_elemp(matR, 0, j-1); i < n; i++, pt1+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
      MAT_W_ELEM_SET(*ctmp1, i, j, genV, genH, r, pt2, pt4)
      SNAPBASECOMPLEX_add(*pt1, *ctmp1, *pt3);
    }
    SNAP_alg_qr_dgv2010refine_apply_S(0, leftS, r, scalarS, rightS, r, matR, j);
  }
#ifdef SNAP_QR_DEBUG
  printf("========== after reconstruction of matR =========\n");
  PRINT_MAT_R
  PRINT_MAT_R_trinorm(*ctmp1)
#ifdef SNAP_QR_GENERATE_SYL
  PRINT_MAT_SYL
  for ( i=0; i<n; i++ ) {
    for ( j=0; j<m; j++ ) {
      SNAPBASECOMPLEX_sub(SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matR,i,j));
    }
  }
  PRINT_MAT_SYL
  for ( i=0; i<n; i++ ) {
    for ( j=0; j<m; j++ ) {
      SNAPBASECOMPLEX_abs(SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matR,i,j));
    }
  }
  PRINT_MAT_SYL_REAL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
  // free the memories allocated.
  SNAPMTXCOMPLEX_clear(matTMP);
  //  SNAPMTXCOMPLEX_clear(matW);
  SNAPBASECOMPLEX_clear(*ctmp4);
  SNAPBASECOMPLEX_clear(*ctmp3);
  SNAPBASECOMPLEX_clear(*ctmp2);
  SNAPBASECOMPLEX_clear(*ctmp1);
  free(ptmp);
  SNAPMTXCOMPLEX_clear(matY);
  SNAPMTXCOMPLEX_clear(matX);
  SNAPMTXCOMPLEX_clear(givensG);
  SNAPMTXCOMPLEX_clear(scalarS);
  SNAPMTXCOMPLEX_clear(rightS);
  SNAPMTXCOMPLEX_clear(leftS);
}

/*!
 *  @brief The function computes a QR factorization of a matrix, given by its generators, by the DGV2010 refine algorithm.
 *  @warning All the matrices must be of the row major.
 *  @param[in] n n is the row length of A.
 *  @param[in] m m is the column length of A.
 *  @param[in] r r is the displacement order of A w.r.t. Z_1 and Z_0.
 *  @param[in] lisD D must be 1d array of length n and the last two elements are complex conjugates each other.
 *  @param[in,out] genV V must be dim(n,r) array satisfying D A - A Z_0 = VH. Overwritten by temporal elements on exit.
 *  @param[in] genH H must be dim(r,m) array satisfying D A - A Z_0 = VH.
 *  @param[in,out] matQ Q must have dim(n,n) and zero entries. The unitary matrix of A=QR on exit.
 *  @param[in,out] matR R must have dim(n,m) and zero entries. The upper triangular matrix of A=QR on exit.
 */
#ifdef SNAP_QR_GENERATE_SYL
void SNAP_alg_qr_dgv2010refine(const unsigned int n, const unsigned int m, const unsigned int r, SNAPMTXCOMPLEX_t lisD, SNAPMTXCOMPLEX_t genV, SNAPMTXCOMPLEX_t genH, SNAPMTXCOMPLEX_t matQ, SNAPMTXCOMPLEX_t matR, SNAPMTXCOMPLEX_t matSYL) {
#else
void SNAP_alg_qr_dgv2010refine(const unsigned int n, const unsigned int m, const unsigned int r, SNAPMTXCOMPLEX_t lisD, SNAPMTXCOMPLEX_t genV, SNAPMTXCOMPLEX_t genH, SNAPMTXCOMPLEX_t matQ, SNAPMTXCOMPLEX_t matR) {
#endif // SNAP_QR_GENERATE_SYL
//  SNAPMTXCOMPLEX_t leftS, rightS, scalarS, givensG, matW, matTMP;
  SNAPMTXCOMPLEX_t leftS, rightS, scalarS, givensG, matX, matY, matTMP;
  SNAPBASECOMPLEX_t *pt1, *pt2, *pt3, *pt4, *ctmp1, *ctmp2, *ctmp3, *ctmp4, *ptmp;
  unsigned int i, j, k, kbar, betak, betakbar, pcv1, pcv2, pcv3, pcv4, spi4w;
#ifdef SNAP_QR_DEBUG
  unsigned int q1,q2;
  SNAPMTXCOMPLEX_t matI;
#endif // SNAP_QR_DEBUG
  // allocate memories. NOTE: all the elements are now 0.
  SNAPMTXCOMPLEX_init2(leftS, n-1, 2*r);                      // accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
  SNAPMTXCOMPLEX_init2(rightS, n-r-1, 2*(r-1));               // accesor: i,2*j S(i,2*j) => (c,s)
  SNAPMTXCOMPLEX_init2(scalarS, 1, n);                        // accesor: i
  SNAPMTXCOMPLEX_init2(givensG, r, 2);                        // accesor: i,j   G(i) => (c,s)
  SNAPMTXCOMPLEX_init2(matX, r+2, 2);
  SNAPMTXCOMPLEX_init2(matY, r+1, 2);
  ptmp = (SNAPBASECOMPLEX_t *) malloc(sizeof(SNAPBASECOMPLEX_t)*4);
  ctmp1 = ptmp;
  ctmp2 = ptmp+1;
  ctmp3 = ptmp+2;
  ctmp4 = ptmp+3;
  SNAPBASECOMPLEX_init(*ctmp1);
  SNAPBASECOMPLEX_init(*ctmp2);
  SNAPBASECOMPLEX_init(*ctmp3);
  SNAPBASECOMPLEX_init(*ctmp4);
//  SNAPMTXCOMPLEX_init2(matW, n, m);
  SNAPMTXCOMPLEX_init2(matTMP, 1, 21); // for pulldownthrough
  // pre-computed W=VH
#ifdef SNAP_SUPDC_T
//  *ctmp1 = 1.0;
//  *ctmp2 = 0.0;
//  cblas_zgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, n, m, r, ctmp1, SNAPMTXCOMPLEX_elems(genV), SNAPMTXCOMPLEX_csize(genV), SNAPMTXCOMPLEX_elems(genH), SNAPMTXCOMPLEX_csize(genH), ctmp2, SNAPMTXCOMPLEX_elems(matW), SNAPMTXCOMPLEX_csize(matW));
#else
  // ERROR: The called fuction has not been implemented.
  return false;
#endif // SNAP_SUPDC_T
  // (2-a) initialization at the allocation site.
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(leftS); i < r*(n-1); i++, pt1+=2 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(rightS); i < (r-1)*(n-r-1); i++, pt1+=2 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  // THIS IS NOT REQUIRED since OVERWRITTEN
  // for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(scalarS); i < n; i++, pt1++ ) {
  //   SNAPBASECOMPLEX_set_ui(*pt1, 1);
  // }
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(matQ); i < n; i++, pt1+=SNAPMTXCOMPLEX_csize(matQ)+1 ) {
    SNAPBASECOMPLEX_set_ui(*pt1, 1);
  }
  // (2-b) initialization after the allocation.
  k = kbar = 0;
//  SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-1, 0), SNAPMTXCOMPLEX_elem(matW, n-1, 0), SNAPMTXCOMPLEX_elem(lisD, 0, n-1));
  MAT_W_ELEM_SET(*ctmp1, n-1, 0, genV, genH, r, pt1, pt2)
  SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-1, 0), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-1));
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(scalarS), pt2 = SNAPMTXCOMPLEX_elems(lisD); i < n; i++, pt1++, pt2++ ) {
    SNAPBASECOMPLEX_set(*pt1, *pt2);
  }

// TEST
//  *ctmp1 = cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*3)/(11));
//  SNAPMTXCOMPLEX_elem(leftS, 0, 0) = *ctmp1;
//  *ctmp2 = conj(*ctmp1);
//  SNAPMTXCOMPLEX_elem(scalarS, 0, 0) *= *ctmp2;
//  SNAPMTXCOMPLEX_elem(scalarS, 0, 1) *= *ctmp1;

  // THIS IS NOT REQUIRED
  // SNAPBASECOMPLEX_set(SNAPMTXCOMPLEX_elem(leftS, n-2, 0), SNAPMTXCOMPLEX_elem(lisD, 0, n-2));
#ifdef SNAP_QR_DEBUG
  printf("========== after initialization =========\n");
//  PRINT_MAT_W PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S PRINT_SCALAR_S
  PRINT_MAT_W(*ctmp1,pt1,pt2) PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S PRINT_SCALAR_S18
#endif // SNAP_QR_DEBUG
  while (1) {
    // (3-a) Increments. NOTE: in this implementation, we assume square case at first.
    betak = betakbar = 1;
    k += betak;
    kbar += betakbar;
    // (3-b) Embed the previous result into the larger one.
    if ( betak > 0 ) {
//      SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-k-1, 0), SNAPMTXCOMPLEX_elem(matW, n-k-1, 0), SNAPMTXCOMPLEX_elem(lisD, 0, n-k-1));
      MAT_W_ELEM_SET(*ctmp1, n-k-1, 0, genV, genH, r, pt1, pt2)
      SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-k-1, 0), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-k-1));
      for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, 0); i < kbar-1; i++, pt1++ ) {
//        SNAPBASECOMPLEX_add(*ctmp1, SNAPMTXCOMPLEX_elem(matW, n-k-1, i+1), *pt1);
        MAT_W_ELEM_SET(*ctmp1, n-k-1, i+1, genV, genH, r, pt2, pt3)
        SNAPBASECOMPLEX_add(*ctmp1, *ctmp1, *pt1);
        SNAPBASECOMPLEX_div(*(pt1+1), *ctmp1, SNAPMTXCOMPLEX_elem(lisD, 0, n-k-1));
      }
    }
    if ( betakbar > 0 ) {
      if ( kbar > 1) {
//        for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, kbar), pt2 = SNAPMTXCOMPLEX_elemp(matW, n-k-1, kbar), pt3 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, kbar-1); i < k+1; i++, pt1+=SNAPMTXCOMPLEX_csize(matR), pt2+=SNAPMTXCOMPLEX_csize(matW), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
        for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, kbar), pt3 = SNAPMTXCOMPLEX_elemp(matR, n-k-1, kbar-1); i < k+1; i++, pt1+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
          MAT_W_ELEM_SET(*ctmp1, n+i-k-1, kbar, genV, genH, r, pt2, pt4)
          SNAPBASECOMPLEX_add(*pt1, *ctmp1, *pt3);
        }
#ifdef SNAP_QR_DEBUG
        printf("========== before applyS =========\n");
        PRINT_MAT_R
#endif // SNAP_QR_DEBUG
        SNAP_alg_qr_dgv2010refine_apply_S(n-k-1, leftS, r, scalarS, rightS, r-1, matR, kbar);
      } else {
//        SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-2, 1), SNAPMTXCOMPLEX_elem(matW, n-2, 1), SNAPMTXCOMPLEX_elem(matR, n-2, 0));
        MAT_W_ELEM_SET(*ctmp1, n-2, 1, genV, genH, r, pt1, pt2)
        SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-2, 1), *ctmp1, SNAPMTXCOMPLEX_elem(matR, n-2, 0));
        SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-2, 1), SNAPMTXCOMPLEX_elem(matR, n-2, 1), SNAPMTXCOMPLEX_elem(lisD, 0, n-2));
//        SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-1, 1), SNAPMTXCOMPLEX_elem(matW, n-1, 1), SNAPMTXCOMPLEX_elem(matR, n-1, 0));
        MAT_W_ELEM_SET(*ctmp1, n-1, 1, genV, genH, r, pt1, pt2)
        SNAPBASECOMPLEX_add(SNAPMTXCOMPLEX_elem(matR, n-1, 1), *ctmp1, SNAPMTXCOMPLEX_elem(matR, n-1, 0));
        SNAPBASECOMPLEX_div(SNAPMTXCOMPLEX_elem(matR, n-1, 1), SNAPMTXCOMPLEX_elem(matR, n-1, 1), SNAPMTXCOMPLEX_elem(lisD, 0, n-1));
      }
    }
#ifdef SNAP_QR_DEBUG
    printf("========== after embedding (k,kbar)=(%u,%u) =========\n", k, kbar);
    PRINT_MAT_R
#endif // SNAP_QR_DEBUG
    // (4) and (5) Restore step. NOTE: not implemented since we assume square case at first.
    // (6) first r-th Givens rotations.
    pcv1 = SNAPLIB_MIN(r, k);
    pcv1 = SNAPLIB_MIN(pcv1, kbar);
    for ( i = 0; i < pcv1; i++ ) {
      // (6-a)S
      // (givensG, r, 2) // accesor: i,j   G(i) => (c,s)
      SNAP_alg_qr_dgv2010refine_givens_C(SNAPMTXCOMPLEX_elemp(matR, n-k+i-1, i), SNAPMTXCOMPLEX_elemp(matR, n-k+i, i), SNAPMTXCOMPLEX_elemp(givensG, i, 0), SNAPMTXCOMPLEX_elemp(givensG, i, 1), ctmp1);
#ifdef SNAP_QR_DEBUG
      printf("========== Givens (i)=(%u) =========\n", i);
      printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, i, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, i, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, i, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, i, 1)));
#endif // SNAP_QR_DEBUG
      // (6-b)
      pcv2 = SNAPLIB_MIN(r, kbar+1);
      for ( j = i+1; j < pcv2; j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matR, n-k+i, j),*ctmp1,*ctmp2)
      }
      // (6-c)
      pcv2 = SNAPLIB_MIN(k+1,kbar+1);
      pcv3 = SNAPLIB_MAX(r+1,pcv2-r+1);
      for ( j = pcv3 - 1; j < pcv2; j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matR, n-k+i, j),*ctmp1,*ctmp2)
      }
#ifdef SNAP_QR_GENERATE_SYL
      for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), SNAPMTXCOMPLEX_elem(matSYL, n-k+i-1, j), SNAPMTXCOMPLEX_elem(matSYL, n-k+i, j),*ctmp1,*ctmp2)
      }
#endif // SNAP_QR_GENERATE_SYL
      // (6-d) Q:=QG^h and V:=GV
      for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i-1), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
        APPLY_GIVENS_HR(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *pt1, *pt2, *ctmp1, *ctmp2)
      }
      for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-k+i-1, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-k+i, 0); j < r; j++, pt1++, pt2++ ) {
        APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *pt1, *pt2, *ctmp1, *ctmp2)
      }
      // (6-e) leftS(n-k,i):=leftS(n-k,i)G^h, (leftS, n-1, 2*r), accesor: i,2*j S(i,2*j) => (c,s) => ((c,s),(-sbar,cbar))
      APPLY_GIVENS_GIVENS_HR( SNAPMTXCOMPLEX_elem(leftS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(leftS, n-k-1, 2*i+1), SNAPMTXCOMPLEX_elem(givensG, i, 0), SNAPMTXCOMPLEX_elem(givensG, i, 1), *ctmp1, *ctmp2)
    }
#ifdef SNAP_QR_DEBUG
//    SNAPBASECOMPLEX_set(*ctmp3, SNAPMTXCOMPLEX_elem(matR, n-k+i-1, i));
//    SNAPBASECOMPLEX_set(*ctmp4, SNAPMTXCOMPLEX_elem(matR, n-k+i, i));
//    dcomplex hoge, hage;
//    SNAP_alg_qr_dgv2010refine_givens_C(ctmp3, ctmp4, &hoge, &hage, ctmp1);
//    printf("========== Givens (i)=(%u) =========\n", i);
//    printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(hoge),cimag(hoge),creal(hage),cimag(hage));
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_DEBUG
    printf("========== after first Givens (k,kbar)=(%u,%u) =========\n", k, kbar);
    PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S
#ifdef SNAP_QR_GENERATE_SYL
    PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
    // (7) Pull through the first (r-1) Givens rotations into rightS.
    for ( i = 0; i < pcv1; i++ ) {
      SNAPBASECOMPLEX_set(*ctmp1, SNAPMTXCOMPLEX_elem(givensG, i, 0));
      SNAPBASECOMPLEX_set(*ctmp2, SNAPMTXCOMPLEX_elem(givensG, i, 1));
#ifdef SNAP_QR_DEBUG
      printf("========== INTO RIGHT-Givens (i)=(%u) =========\n", i);
      printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
      if ( k <= r ) {
        SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), k-i-1, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== (a)INTO RIGHT-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)), SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
      } else if ( k < 2*r ) {
        if ( i < k-r ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), r, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
          printf("========== (b)INTO RIGHT-Givens (i)=(%u) =========\n", i);
          printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), k-i-1, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
          printf("========== (c)INTO RIGHT-Givens (i)=(%u) =========\n", i);
          printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)), SNAPMTXCOMPLEX_elem(leftS, n-k+i-1, 2*(k-i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
      } else if ( i+1 < r ) {
        SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), r, SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== (d)INTO RIGHT-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i), SNAPMTXCOMPLEX_elem(rightS, n-k-1, 2*i+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
      }
    }
#ifdef SNAP_QR_DEBUG
    printf("========== after pull through into rightS (k,kbar)=(%u,%u) =========\n", k, kbar);
    PRINT_LEFT_S PRINT_RIGHT_S
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_DEBUG
    print_rebuildS(n, r, matQ, lisD);
    print_computedS(n, r, scalarS, leftS, rightS);
    if ( k >= 2*r && r <= kbar+1 ) {
      print_computedS_wG(n, r, scalarS, leftS, rightS, SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), n-k+r-1-1);
    }
#endif // SNAP_QR_DEBUG
    // (8) Chasing
    if ( k >= 2*r && r <= kbar+1 ) {
      // G = SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1)
      // (8-b)
      for ( i = r - 1; i < k - r; i++ ) {
#ifdef SNAP_QR_DEBUG
        printf("========== chasing (k,kbar)=(%u,%u) and (leftS1,S2)=(%u,%u), (rightS1,S2)=(%u,%u) =========\n", k, kbar, n-k+i-1, n-k+i,n-r-k+i,n-r-k+i+1 );
        PRINT_SCALAR_S18 PRINT_LEFT_S18 PRINT_RIGHT_S18
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_DEBUG
        printf("========== Givens to be chased =========\n");
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)));
#endif // SNAP_QR_DEBUG
        // (8-b-i) and (8-b-ii) apply the scalars
        SNAP_alg_qr_dgv2010refine_apply_scalar(SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elems(matTMP));
#ifdef SNAP_QR_DEBUG
        printf("========== Givens modifed by scalar  =========\n");
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)));
#endif // SNAP_QR_DEBUG
        SNAP_alg_qr_dgv2010refine_chasing(r, SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i+1, 0), SNAPMTXCOMPLEX_elems(matX), SNAPMTXCOMPLEX_elems(matY), SNAPMTXCOMPLEX_elems(matTMP));
        /*
                        // (8-b-i)
                        SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i), SNAPMTXCOMPLEX_elemp(leftS, n-k+i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-k+i-1), SNAPMTXCOMPLEX_elemp(leftS, n-k+i-1, 0), r, SNAPMTXCOMPLEX_elems(matTMP));
                        // (8-b-ii)
                        SNAP_alg_qr_dgv2010refine_pullupthroughright(r, SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-k+i+1, 0), r-1, SNAPMTXCOMPLEX_elems(matTMP));
        */
#ifdef SNAP_QR_DEBUG
        printf("========== Chasing-Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 0)),creal(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)),cimag(SNAPMTXCOMPLEX_elem(givensG, r-1, 1)));
#endif // SNAP_QR_DEBUG
        // (8-b-iii)
        pcv2 = SNAPLIB_MIN(k+1, kbar+1);
        pcv1 = SNAPLIB_MAX(r+1, pcv2-r+1);
        for ( j = pcv1-1; j < pcv2; j++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1), SNAPMTXCOMPLEX_elem(matR, n-k+i, j), SNAPMTXCOMPLEX_elem(matR, n-k+i+1, j),*ctmp1,*ctmp2)
        }
#ifdef SNAP_QR_GENERATE_SYL
        for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1), SNAPMTXCOMPLEX_elem(matSYL, n-k+i, j), SNAPMTXCOMPLEX_elem(matSYL, n-k+i+1, j),*ctmp1,*ctmp2)
        }
#endif // SNAP_QR_GENERATE_SYL
        // (8-b-iv) Q:=QG^h and V:=GV
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-k+i+1); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
          APPLY_GIVENS_HR(SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1), *pt1, *pt2, *ctmp1, *ctmp2)
        }
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-k+i, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-k+i+1, 0); j < r; j++, pt1++, pt2++ ) {
          APPLY_GIVENS(SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1), *pt1, *pt2, *ctmp1, *ctmp2)
        }
#ifdef SNAP_QR_DEBUG
        printf("========== DONE chasing (k,kbar)=(%u,%u) and (leftS1,S2)=(%u,%u), (rightS1,S2)=(%u,%u) =========\n", k, kbar, n-k+i-1, n-k+i,n-r-k+i,n-r-k+i+1 );
        PRINT_SCALAR_S18 PRINT_LEFT_S18 PRINT_RIGHT_S18
#endif // SNAP_QR_DEBUG
      }
      // (8-c)
      SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, SNAPMTXCOMPLEX_elemp(givensG, r-1, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-r), SNAPMTXCOMPLEX_elemp(leftS, n-r, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-r-1), SNAPMTXCOMPLEX_elemp(leftS, n-r-1, 0), r-1, SNAPMTXCOMPLEX_elems(matTMP));
      APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-r-1, 2*(r-1)), SNAPMTXCOMPLEX_elem(leftS, n-r-1, 2*(r-1)+1), SNAPMTXCOMPLEX_elem(givensG, r-1, 0), SNAPMTXCOMPLEX_elem(givensG, r-1, 1), *ctmp3, *ctmp4)
#ifdef SNAP_QR_DEBUG
      printf("========== after chasing (k,kbar)=(%u,%u) =========\n", k, kbar);
      PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S
#ifdef SNAP_QR_GENERATE_SYL
      PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#ifdef SNAP_QR_DEBUG
      print_rebuildS(n, r, matQ, lisD);
      print_computedS(n, r, scalarS, leftS, rightS);
#endif // SNAP_QR_DEBUG
#endif // SNAP_QR_DEBUG
    }
    // (9) last (r-1) Givens rotations.
    if ( k > r && kbar > r ) { // we use unsigned integers hence this if statement is required.
      pcv1 = SNAPLIB_MIN(r-1, k-r);
      pcv1 = SNAPLIB_MIN(pcv1, kbar-r);
      if ( k > kbar ) {
        pcv2 = SNAPLIB_MAX(k - kbar, 1);
      } else {
        pcv2 = 1;
      }
      pcv4 = SNAPLIB_MIN(k+1,kbar+1);
      for ( i = pcv1; i >= pcv2; i-- ) {
        // G:=(ctmp1, ctmp2)!!
        // (9-a)
        SNAP_alg_qr_dgv2010refine_givens_C(SNAPMTXCOMPLEX_elemp(matR, n-i-1, k-i), SNAPMTXCOMPLEX_elemp(matR, n-i, k-i), ctmp1, ctmp2, ctmp3);
#ifdef SNAP_QR_DEBUG
        printf("========== Givens (i)=(%u) =========\n", i);
        printf("(%+.18e%+.18eI),(%+.18e%+.18eI)\n",creal(*ctmp1),cimag(*ctmp1),creal(*ctmp2),cimag(*ctmp2));
#endif // SNAP_QR_DEBUG
        // (9-b)
        pcv3 = SNAPLIB_MAX(2*r+1-i, pcv4-i+1);
        for ( j = pcv3-1; j < pcv4; j++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, SNAPMTXCOMPLEX_elem(matR, n-i-1, j), SNAPMTXCOMPLEX_elem(matR, n-i, j),*ctmp3,*ctmp4)
        }
#ifdef SNAP_QR_GENERATE_SYL
        for ( j = 0; j < SNAPMTXCOMPLEX_cols(matSYL); j++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, SNAPMTXCOMPLEX_elem(matSYL, n-i-1, j), SNAPMTXCOMPLEX_elem(matSYL, n-i, j),*ctmp3,*ctmp4)
        }
#endif // SNAP_QR_GENERATE_SYL
        // (9-c) Q:=QG^h and V:=GV
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-i-1), pt2 = SNAPMTXCOMPLEX_elemp(matQ, 0, n-i); j < n; j++, pt1+=SNAPMTXCOMPLEX_csize(matQ), pt2+=SNAPMTXCOMPLEX_csize(matQ) ) {
          APPLY_GIVENS_HR(*ctmp1, *ctmp2, *pt1, *pt2, *ctmp3, *ctmp4)
        }
        for ( j = 0, pt1 = SNAPMTXCOMPLEX_elemp(genV, n-i-1, 0), pt2 = SNAPMTXCOMPLEX_elemp(genV, n-i, 0); j < r; j++, pt1++, pt2++ ) {
          APPLY_GIVENS(*ctmp1, *ctmp2, *pt1, *pt2, *ctmp3, *ctmp4)
        }
        // (9-d)
        // backup the Givens rotation
        SNAPBASECOMPLEX_set(SNAPMTXCOMPLEX_elem(givensG, 0, 0), *ctmp1);
        SNAPBASECOMPLEX_set(SNAPMTXCOMPLEX_elem(givensG, 0, 1), *ctmp2);
        if ( i > 1 ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughright(r, ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i), SNAPMTXCOMPLEX_elemp(leftS, n-i, 0), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i-1), SNAPMTXCOMPLEX_elemp(leftS, n-i-1, 0), i-1, SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          SNAP_alg_qr_dgv2010refine_apply_scalar(ptmp, SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i), SNAPMTXCOMPLEX_elemp(scalarS, 0, n-i-1), SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS(SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(leftS, n-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
        // (9-e)
        // restore the Givens rotation
        SNAPBASECOMPLEX_conj(*ctmp1, SNAPMTXCOMPLEX_elem(givensG, 0, 0));
        SNAPBASECOMPLEX_neg(*ctmp2, SNAPMTXCOMPLEX_elem(givensG, 0, 1));
        if ( i > 1 ) {
          SNAP_alg_qr_dgv2010refine_pulldownthroughleft(r, ptmp, SNAPMTXCOMPLEX_elemp(rightS, n-r-i-1, 0), SNAPMTXCOMPLEX_elemp(rightS, n-r-i, 0), i-1, SNAPMTXCOMPLEX_elems(matTMP));
          APPLY_GIVENS_GIVENS_R(SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        } else {
          APPLY_GIVENS_GIVENS_R(SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)), SNAPMTXCOMPLEX_elem(rightS, n-r-i-1, 2*(i-1)+1), *ctmp1, *ctmp2, *ctmp3, *ctmp4)
        }
      }
#ifdef SNAP_QR_DEBUG
      printf("========== after last (r-1) Givens rotations (k,kbar)=(%u,%u) =========\n", k, kbar);
      PRINT_MAT_Q PRINT_MAT_R PRINT_LEFT_S PRINT_RIGHT_S
#ifdef SNAP_QR_GENERATE_SYL
      PRINT_MAT_SYL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
    }
#ifdef SNAP_QR_DEBUG
    print_rebuildS(n, r, matQ, lisD);
    print_computedS(n, r, scalarS, leftS, rightS);
#ifdef SNAP_QR_GENERATE_SYL
    PRINT_MAT_SYL_trinorm(*ctmp1,k)
#endif // SNAP_QR_GENERATE_SYL
    SNAPMTXCOMPLEX_init2(matI, n, n);
    printf("norm(I-matI=I-Q^hQ)_F: ");
    *ctmp1 = 0.0;
    for(q1=0; q1<sdmdc_rows(matI); q1++) {
      for(q2=0; q2<sdmdc_cols(matI); q2++) {
        SNAPBASECOMPLEX_set_d(SNAPMTXCOMPLEX_elem(matI,q1,q1),0);
        for ( i = 0; i<sdmdc_rows(matQ); i++ ) {
          sdmdc_elem(matI,q1,q2)+=conj(sdmdc_elem(matQ,i,q1))*sdmdc_elem(matQ,i,q2);
        }
        if ( q1 == q2 ) {
          sdmdc_elem(matI,q1,q2)-=1.0;
        }
        *ctmp1 += cabs(sdmdc_elem(matI,q1,q2))*cabs(sdmdc_elem(matI,q1,q2));
      }
    }
    printf("%e\n", sqrt(*ctmp1));
    SNAPMTXCOMPLEX_clear(matI);
#endif // SNAP_QR_DEBUG
    // (10) loop or not
    if ( k+1 >= n && kbar+1 >= SNAPLIB_MIN(m, k+1) ) {
      break;
    }
  }
  // (11) reconstruct matR
  pcv1 = SNAPLIB_MIN(n-r, m);
  for ( j = r; j < pcv1; j++ ) {
    for ( i = 0, pt1 = SNAPMTXCOMPLEX_elemp(matR, 0, j), pt3 = SNAPMTXCOMPLEX_elemp(matR, 0, j-1); i < n; i++, pt1+=SNAPMTXCOMPLEX_csize(matR), pt3+=SNAPMTXCOMPLEX_csize(matR) ) {
      MAT_W_ELEM_SET(*ctmp1, i, j, genV, genH, r, pt2, pt4)
      SNAPBASECOMPLEX_add(*pt1, *ctmp1, *pt3);
    }
    SNAP_alg_qr_dgv2010refine_apply_S(0, leftS, r, scalarS, rightS, r-1, matR, j);
  }
#ifdef SNAP_QR_DEBUG
  printf("========== after reconstruction of matR =========\n");
  PRINT_MAT_R
  PRINT_MAT_R_trinorm(*ctmp1)
#ifdef SNAP_QR_GENERATE_SYL
  PRINT_MAT_SYL
  for ( i=0; i<n; i++ ) {
    for ( j=0; j<m; j++ ) {
      SNAPBASECOMPLEX_sub(SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matR,i,j));
    }
  }
  PRINT_MAT_SYL
  for ( i=0; i<n; i++ ) {
    for ( j=0; j<m; j++ ) {
      SNAPBASECOMPLEX_abs(SNAPMTXCOMPLEX_elem(matSYL,i,j),SNAPMTXCOMPLEX_elem(matR,i,j));
    }
  }
  PRINT_MAT_SYL_REAL
#endif // SNAP_QR_GENERATE_SYL
#endif // SNAP_QR_DEBUG
  // free the memories allocated.
  SNAPMTXCOMPLEX_clear(matTMP);
//  SNAPMTXCOMPLEX_clear(matW);
  SNAPBASECOMPLEX_clear(*ctmp4);
  SNAPBASECOMPLEX_clear(*ctmp3);
  SNAPBASECOMPLEX_clear(*ctmp2);
  SNAPBASECOMPLEX_clear(*ctmp1);
  free(ptmp);
  SNAPMTXCOMPLEX_clear(matY);
  SNAPMTXCOMPLEX_clear(matX);
  SNAPMTXCOMPLEX_clear(givensG);
  SNAPMTXCOMPLEX_clear(scalarS);
  SNAPMTXCOMPLEX_clear(rightS);
  SNAPMTXCOMPLEX_clear(leftS);
}

/*!
 *  @brief The function computes a QR factorization of the Sylvester matrix of f and g, by the DGV2010 refine algorithm.
 *  @warning All the matrices must be of the row major.
 *  @param[in] poly_inf The input polynomial f[x] of degree n.
 *  @param[in] poly_ing The input polynomial g[x] of degree m.
 *  @param[in,out] matQ Q must have dim(n+m,n+m) and zero entries. The unitary matrix of Sylvester matrix on exit.
 *  @param[in,out] matR R must have dim(n+m,n+m) and zero entries. The upper triangular matrix of Sylvester matrix on exit.
 */
void SNAP_alg_qr_dgv2010refine_sylvester(SNAP_t poly_inf, SNAP_t poly_ing, SNAPMTXCOMPLEX_t matQ, SNAPMTXCOMPLEX_t matR) {
  unsigned int i, n, m;
  SNAPMTXCOMPLEX_t genV, genH, lisD;
  SNAPBASECOMPLEX_t *pt1, sqrtn;
#ifdef SNAP_QR_GENERATE_SYL
  fftw_plan fftp;
  int fftdim[1];
  double sqrtofnm;
  unsigned int s1,s2;
  SNAPMTXCOMPLEX_t matSYL;
#endif // SNAP_QR_GENERATE_SYL

  n = SNAP_degree(poly_inf);
  m = SNAP_degree(poly_ing);
  /* The generatos of Sylvester w.r.t. Z_1 and Z_0, upper-right shift operators. */
  SNAPMTXCOMPLEX_init2(genV, n+m, 2);
  SNAPMTXCOMPLEX_init2(genH, 2, n+m);
  SNAP_alg_qr_dgv2010refine_make_sylvester_generators(poly_inf, poly_ing, genV, genH);

/* // if test with randomly generated generators (time.h is required).
srand((unsigned int)time(NULL));
for (s1=0; s1<sdmdc_rows(genV); s1++ ) {
for (s2=0; s2<sdmdc_cols(genV); s2++ ) {
sdmdc_elem(genV,s1,s2)=((double) rand()/((double)RAND_MAX+1))+((double) rand()/((double)RAND_MAX+1))*_Complex_I;
}
}
for (s1=0; s1<sdmdc_rows(genH); s1++ ) {
for (s2=0; s2<sdmdc_cols(genH); s2++ ) {
sdmdc_elem(genH,s1,s2)=((double) rand()/((double)RAND_MAX+1))+((double) rand()/((double)RAND_MAX+1))*_Complex_I;
}
}
*/

#ifdef SNAP_QR_GENERATE_SYL
  SNAPMTXCOMPLEX_init(matSYL);
  SNAPMTX_sylvester_matrix_transpose(matSYL, poly_inf, poly_ing);
#ifdef SNAP_SUPDC_T
  printf("Sylvester Matrix:\n");
  for (s1=0; s1<sdmdc_rows(matSYL); s1++ ) {
    for (s2=0; s2<sdmdc_cols(matSYL); s2++ ) {
      printf("(%+e%+eI),", creal(sdmdc_elem(matSYL,s1,s2)), cimag(sdmdc_elem(matSYL,s1,s2)));
    }
    printf("\n");
  }
#endif // SNAP_SUPDC_T
#endif // SNAP_QR_GENERATE_SYL

  /* Multiply F^h on the left where F is Fourier matrix (0,0 to n-1,n-1) but swapped 1st and (n-2)th columns. */
  SNAPBASECOMPLEX_sqrt_ui(sqrtn, n+m);
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(genV); i < n+m; i++, pt1+=SNAPMTXCOMPLEX_csize(genV) ) {
#ifdef SNAP_SUPDC_T
    SNAPBASECOMPLEX_set(*pt1, cexp((dcomplex) (-2*3.1415926535897932384626433832795029L*_Complex_I*i*(m-1)/(n+m)))/sqrtn);
    SNAPBASECOMPLEX_set(*(pt1+1), cexp((dcomplex) (-2*3.1415926535897932384626433832795029L*_Complex_I*i*(n+m-1)/(n+m)))/sqrtn);
#else
    // ERROR: The called fuction has not been implemented.
    return false;
#endif // SNAP_SUPDC_T
  }
  // THIS IS NOT REQUIRED.
  // SNAPBASECOMPLEX_swap( SNAPMTXCOMPLEX_elem(genV, 1, 0), SNAPMTXCOMPLEX_elem(genV, n+m-2, 0));
  // SNAPBASECOMPLEX_swap( SNAPMTXCOMPLEX_elem(genV, 1, 1), SNAPMTXCOMPLEX_elem(genV, n+m-2, 1));

#ifdef SNAP_QR_GENERATE_SYL
  // FFTW
  fftdim[0] = SNAPMTXCOMPLEX_rows(matSYL);
  fftp = fftw_plan_many_dft(1, fftdim, SNAPMTXCOMPLEX_cols(matSYL), SNAPMTXCOMPLEX_elems(matSYL), fftdim, SNAPMTXCOMPLEX_csize(matSYL), 1, SNAPMTXCOMPLEX_elems(matSYL), fftdim, SNAPMTXCOMPLEX_csize(matSYL), 1, FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_execute(fftp);
  fftw_destroy_plan(fftp);
  sqrtofnm = sqrt(n+m);
  for (s1=0; s1<sdmdc_rows(matSYL); s1++ ) {
    for (s2=0; s2<sdmdc_cols(matSYL); s2++ ) {
      SNAPBASECOMPLEX_div_basereal(sdmdc_elem(matSYL,s1,s2),sdmdc_elem(matSYL,s1,s2), sqrtofnm);
    }
  }
  // THIS IS NOT REQUIRED.
  // for (s2=0; s2<sdmdc_cols(matSYL); s2++ ) {
  //   SNAPBASECOMPLEX_swap( SNAPMTXCOMPLEX_elem(matSYL, 1, s2), SNAPMTXCOMPLEX_elem(matSYL, n+m-2, s2));
  // }
#ifdef SNAP_SUPDC_T
  printf("Modified Sylvester Matrix:\n");
  for (s1=0; s1<sdmdc_rows(matSYL); s1++ ) {
    for (s2=0; s2<sdmdc_cols(matSYL); s2++ ) {
      printf("(%+e%+eI),", creal(sdmdc_elem(matSYL,s1,s2)), cimag(sdmdc_elem(matSYL,s1,s2)));
    }
    printf("\n");
  }
#endif // SNAP_SUPDC_T
#endif // SNAP_QR_GENERATE_SYL

  SNAPMTXCOMPLEX_init2(lisD, 1, n+m);
  for ( i = 0, pt1 = SNAPMTXCOMPLEX_elems(lisD); i < n+m; i++, pt1++ ) {
#ifdef SNAP_SUPDC_T
    SNAPBASECOMPLEX_set(*pt1, cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i)/(n+m)));
    // THIS IS NOT REQUIRED.
    // if ( i != 1 && i != n+m-2 ) {
    //   SNAPBASECOMPLEX_set(*pt1, cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*i)/(n+m)));
    // } else {
    //   if ( i == 1 ) {
    //     SNAPBASECOMPLEX_set(*pt1, cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I*(n+m-2))/(n+m)));
    //   } else {
    //     SNAPBASECOMPLEX_set(*pt1, cexp((dcomplex) (2*3.1415926535897932384626433832795029L*_Complex_I)/(n+m)));
    //   }
    // }
#else
    // ERROR: The called fuction has not been implemented.
    return false;
#endif // SNAP_SUPDC_T
  }
#ifdef SNAP_QR_DEBUG
  unsigned int q1,q2;
  printf("Modified Generators of Sylvester Matrix: V:\n");
  for (q1=0; q1<sdmdc_rows(genV); q1++ ) {
    for (q2=0; q2<sdmdc_cols(genV); q2++ ) {
      printf("(%+.18e%+.18eI),", creal(sdmdc_elem(genV,q1,q2)), cimag(sdmdc_elem(genV,q1,q2)));
    }
    printf("\n");
  }
  printf("Modified Generators of Sylvester Matrix: H:\n");
  for (q1=0; q1<sdmdc_rows(genH); q1++ ) {
    for (q2=0; q2<sdmdc_cols(genH); q2++ ) {
      printf("(%+.18e%+.18eI),", creal(sdmdc_elem(genH,q1,q2)), cimag(sdmdc_elem(genH,q1,q2)));
    }
    printf("\n");
  }
  printf("D of Sylvester Matrix:\n");
  for (q1=0; q1<sdmdc_rows(lisD); q1++ ) {
    for (q2=0; q2<sdmdc_cols(lisD); q2++ ) {
      printf("(%+f%+fI),", creal(sdmdc_elem(lisD,q1,q2)), cimag(sdmdc_elem(lisD,q1,q2)));
    }
    printf("\n");
  }
#endif // SNAP_QR_DEBUG
#ifdef SNAP_QR_GENERATE_SYL
//  SNAP_alg_qr_dgv2010refine(n+m, n+m, 2, lisD, genV, genH, matQ, matR, matSYL);
  SNAP_alg_qr_dgv2010refine_embed(n+m, n+m, 2, lisD, genV, genH, matQ, matR, matSYL);
#else
//  SNAP_alg_qr_dgv2010refine(n+m, n+m, 2, lisD, genV, genH, matQ, matR);
  SNAP_alg_qr_dgv2010refine_embed(n+m, n+m, 2, lisD, genV, genH, matQ, matR);
#endif // SNAP_QR_GENERATE_SYL

#ifdef SNAP_QR_GENERATE_SYL
  SNAPMTXCOMPLEX_clear(matSYL);
#endif // SNAP_QR_GENERATE_SYL
  SNAPMTXCOMPLEX_clear(genV);
  SNAPMTXCOMPLEX_clear(genH);
  SNAPMTXCOMPLEX_clear(lisD);
}

